Package: CausalMixGPD
Type: Package
Title: Bayesian Nonparametric Conditional Density Modeling in Causal
        Inference and Clustering with a Heavy-Tail Extension
Version: 0.8.0
Maintainer: Arnab Aich <aaich@fsu.edu>
Authors@R: c(
    person(given = "Arnab", family = "Aich", role = c("aut", "cre"),
           email = "aaich@fsu.edu",
           comment = c(ORCID = "0009-0005-7801-6701"))
    )
Description: The presence of a heavy tail is a feature of many scenarios when risk management
            involves extremely rare events. While parametric distributions may give adequate
            representation of the mode of data, they are likely to misrepresent heavy tails,
            and completely nonparametric approaches lack a rigorous mechanism for
            tail extrapolation; see Pickands (1975) <doi:10.1214/aos/1176343003>. The
            package 'CausalMixGPD' implements tools for Bayesian analysis of heavy-tailed
            outcomes by combining Dirichlet process mixture models for the body of the
            distribution with optional generalized Pareto tails. The method allows
            for unconditional and covariate-modulated mixtures, implements MCMC estimation
            using 'nimble', and extends to mixtures of different arms' outcomes with
            application to causal inference in the Rubin (1974)
            <doi:10.1037/h0037350> framework. Posterior summaries include density
            functions, quantiles, expected values, survival functions, and causal effects,
            with an emphasis on tail quantiles and functional measures sensitive to the
            tail.
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.0.0)
Imports: nimble, stats, utils, ggplot2
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, here, cli, coda,
        ggmcmc, future, future.apply, crayon, DT, kableExtra, plotly,
        MatchIt, codetools
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Config/testthat/edition: 3
URL: https://arnabaich96.github.io/CausalMixGPD/pkgdown/,
        https://doi.org/10.5281/zenodo.19671881
BugReports: https://arnabaich96.github.io/CausalMixGPD/
NeedsCompilation: no
Packaged: 2026-05-05 07:45:28 UTC; arnab
Author: Arnab Aich [aut, cre] (ORCID: <https://orcid.org/0009-0005-7801-6701>)
Repository: CRAN
Date/Publication: 2026-05-05 10:30:02 UTC
Built: R 4.6.0; ; 2026-05-05 11:22:16 UTC; unix

Package: CausalMixGPD
Type: Package
Title: Bayesian Nonparametric Conditional Density Modeling in Causal
        Inference and Clustering with a Heavy-Tail Extension
Version: 0.8.0
Maintainer: Arnab Aich <aaich@fsu.edu>
Authors@R: c(
    person(given = "Arnab", family = "Aich", role = c("aut", "cre"),
           email = "aaich@fsu.edu",
           comment = c(ORCID = "0009-0005-7801-6701"))
    )
Description: The presence of a heavy tail is a feature of many scenarios when risk management
            involves extremely rare events. While parametric distributions may give adequate
            representation of the mode of data, they are likely to misrepresent heavy tails,
            and completely nonparametric approaches lack a rigorous mechanism for
            tail extrapolation; see Pickands (1975) <doi:10.1214/aos/1176343003>. The
            package 'CausalMixGPD' implements tools for Bayesian analysis of heavy-tailed
            outcomes by combining Dirichlet process mixture models for the body of the
            distribution with optional generalized Pareto tails. The method allows
            for unconditional and covariate-modulated mixtures, implements MCMC estimation
            using 'nimble', and extends to mixtures of different arms' outcomes with
            application to causal inference in the Rubin (1974)
            <doi:10.1037/h0037350> framework. Posterior summaries include density
            functions, quantiles, expected values, survival functions, and causal effects,
            with an emphasis on tail quantiles and functional measures sensitive to the
            tail.
License: GPL-3
Encoding: UTF-8
LazyData: true
Depends: R (>= 4.0.0)
Imports: nimble, stats, utils, ggplot2
Suggests: testthat (>= 3.0.0), knitr, rmarkdown, here, cli, coda,
        ggmcmc, future, future.apply, crayon, DT, kableExtra, plotly,
        MatchIt, codetools
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Config/testthat/edition: 3
URL: https://arnabaich96.github.io/CausalMixGPD/pkgdown/,
        https://doi.org/10.5281/zenodo.19671881
BugReports: https://arnabaich96.github.io/CausalMixGPD/
NeedsCompilation: no
Packaged: 2026-05-05 07:45:28 UTC; arnab
Author: Arnab Aich [aut, cre] (ORCID: <https://orcid.org/0009-0005-7801-6701>)
Repository: CRAN
Date/Publication: 2026-05-05 10:30:02 UTC
Built: R 4.5.2; ; 2026-05-05 11:30:50 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(ate,causalmixgpd_causal_fit)
S3method(ate,default)
S3method(cate,causalmixgpd_causal_fit)
S3method(cate,default)
S3method(cluster_profiles,dpmixgpd_cluster_fit)
S3method(cluster_profiles,dpmixgpd_cluster_labels)
S3method(cluster_profiles,summary.dpmixgpd_cluster_fit)
S3method(cluster_profiles,summary.dpmixgpd_cluster_labels)
S3method(cqte,causalmixgpd_causal_fit)
S3method(cqte,default)
S3method(fitted,mixgpd_fit)
S3method(params,causalmixgpd_causal_fit)
S3method(params,causalmixgpd_ps_fit)
S3method(params,mixgpd_fit)
S3method(plot,causalmixgpd_ate)
S3method(plot,causalmixgpd_causal_fit)
S3method(plot,causalmixgpd_causal_predict)
S3method(plot,causalmixgpd_qte)
S3method(plot,dpmixgpd_cluster_bundle)
S3method(plot,dpmixgpd_cluster_fit)
S3method(plot,dpmixgpd_cluster_labels)
S3method(plot,dpmixgpd_cluster_psm)
S3method(plot,mixgpd_fit)
S3method(plot,mixgpd_fitted)
S3method(plot,mixgpd_predict)
S3method(predict,causalmixgpd_causal_fit)
S3method(predict,dpmixgpd_cluster_fit)
S3method(predict,mixgpd_fit)
S3method(print,causalmixgpd_ate)
S3method(print,causalmixgpd_bundle)
S3method(print,causalmixgpd_causal_bundle)
S3method(print,causalmixgpd_causal_fit)
S3method(print,causalmixgpd_causal_fit_plots)
S3method(print,causalmixgpd_causal_predict_plots)
S3method(print,causalmixgpd_ps_bundle)
S3method(print,causalmixgpd_ps_fit)
S3method(print,causalmixgpd_qte)
S3method(print,dpmixgpd_cluster_bundle)
S3method(print,dpmixgpd_cluster_fit)
S3method(print,dpmixgpd_cluster_labels)
S3method(print,dpmixgpd_cluster_psm)
S3method(print,mixgpd_ess_summary)
S3method(print,mixgpd_fit)
S3method(print,mixgpd_fit_plots)
S3method(print,mixgpd_fitted_plots)
S3method(print,mixgpd_params)
S3method(print,mixgpd_params_pair)
S3method(print,mixgpd_predict)
S3method(print,mixgpd_predict_plots)
S3method(print,mixgpd_summary)
S3method(print,summary.causalmixgpd_ate)
S3method(print,summary.causalmixgpd_causal_fit)
S3method(print,summary.causalmixgpd_ps_fit)
S3method(print,summary.causalmixgpd_qte)
S3method(qte,causalmixgpd_causal_fit)
S3method(qte,default)
S3method(residuals,mixgpd_fit)
S3method(summary,causalmixgpd_ate)
S3method(summary,causalmixgpd_bundle)
S3method(summary,causalmixgpd_causal_bundle)
S3method(summary,causalmixgpd_causal_fit)
S3method(summary,causalmixgpd_ps_bundle)
S3method(summary,causalmixgpd_ps_fit)
S3method(summary,causalmixgpd_qte)
S3method(summary,dpmixgpd_cluster_bundle)
S3method(summary,dpmixgpd_cluster_fit)
S3method(summary,dpmixgpd_cluster_labels)
S3method(summary,dpmixgpd_cluster_psm)
S3method(summary,mixgpd_ess_summary)
S3method(summary,mixgpd_fit)
export(ate)
export(ate_rmean)
export(att)
export(build_causal_bundle)
export(build_nimble_bundle)
export(bundle)
export(cate)
export(check_glue_validity)
export(cluster_profiles)
export(cqte)
export(dAmoroso)
export(dAmorosoGpd)
export(dAmorosoMix)
export(dAmorosoMixGpd)
export(dCauchy)
export(dCauchyMix)
export(dGammaGpd)
export(dGammaMix)
export(dGammaMixGpd)
export(dGpd)
export(dInvGauss)
export(dInvGaussGpd)
export(dInvGaussMix)
export(dInvGaussMixGpd)
export(dLaplaceGpd)
export(dLaplaceMix)
export(dLaplaceMixGpd)
export(dLognormalGpd)
export(dLognormalMix)
export(dLognormalMixGpd)
export(dNormGpd)
export(dNormMix)
export(dNormMixGpd)
export(damoroso)
export(damorosogpd)
export(damorosomix)
export(damorosomixgpd)
export(dcauchy_vec)
export(dcauchymix)
export(dgammagpd)
export(dgammamix)
export(dgammamixgpd)
export(dgpd)
export(dinvgauss)
export(dinvgaussgpd)
export(dinvgaussmix)
export(dinvgaussmixgpd)
export(dlaplacegpd)
export(dlaplacemix)
export(dlaplacemixgpd)
export(dlognormalgpd)
export(dlognormalmix)
export(dlognormalmixgpd)
export(dnormgpd)
export(dnormmix)
export(dnormmixgpd)
export(dpmgpd)
export(dpmgpd.causal)
export(dpmgpd.cluster)
export(dpmix)
export(dpmix.causal)
export(dpmix.cluster)
export(ess_summary)
export(get_kernel_registry)
export(get_tail_registry)
export(init_kernel_registry)
export(kernel_support_table)
export(mcmc)
export(pAmoroso)
export(pAmorosoGpd)
export(pAmorosoMix)
export(pAmorosoMixGpd)
export(pCauchy)
export(pCauchyMix)
export(pGammaGpd)
export(pGammaMix)
export(pGammaMixGpd)
export(pGpd)
export(pInvGauss)
export(pInvGaussGpd)
export(pInvGaussMix)
export(pInvGaussMixGpd)
export(pLaplaceGpd)
export(pLaplaceMix)
export(pLaplaceMixGpd)
export(pLognormalGpd)
export(pLognormalMix)
export(pLognormalMixGpd)
export(pNormGpd)
export(pNormMix)
export(pNormMixGpd)
export(pamoroso)
export(pamorosogpd)
export(pamorosomix)
export(pamorosomixgpd)
export(params)
export(pcauchy_vec)
export(pcauchymix)
export(pgammagpd)
export(pgammamix)
export(pgammamixgpd)
export(pgpd)
export(pinvgauss)
export(pinvgaussgpd)
export(pinvgaussmix)
export(pinvgaussmixgpd)
export(plaplacegpd)
export(plaplacemix)
export(plaplacemixgpd)
export(plognormalgpd)
export(plognormalmix)
export(plognormalmixgpd)
export(pnormgpd)
export(pnormmix)
export(pnormmixgpd)
export(qAmoroso)
export(qAmorosoGpd)
export(qAmorosoMix)
export(qAmorosoMixGpd)
export(qCauchy)
export(qCauchyMix)
export(qGammaGpd)
export(qGammaMix)
export(qGammaMixGpd)
export(qGpd)
export(qInvGauss)
export(qInvGaussGpd)
export(qInvGaussMix)
export(qInvGaussMixGpd)
export(qLaplaceGpd)
export(qLaplaceMix)
export(qLaplaceMixGpd)
export(qLognormalGpd)
export(qLognormalMix)
export(qLognormalMixGpd)
export(qNormGpd)
export(qNormMix)
export(qNormMixGpd)
export(qamoroso)
export(qamorosogpd)
export(qamorosomix)
export(qamorosomixgpd)
export(qcauchy_vec)
export(qcauchymix)
export(qgammagpd)
export(qgammamix)
export(qgammamixgpd)
export(qgpd)
export(qinvgauss)
export(qinvgaussgpd)
export(qinvgaussmix)
export(qinvgaussmixgpd)
export(qlaplacegpd)
export(qlaplacemix)
export(qlaplacemixgpd)
export(qlognormalgpd)
export(qlognormalmix)
export(qlognormalmixgpd)
export(qnormgpd)
export(qnormmix)
export(qnormmixgpd)
export(qte)
export(qtt)
export(rAmoroso)
export(rAmorosoGpd)
export(rAmorosoMix)
export(rAmorosoMixGpd)
export(rCauchy)
export(rCauchyMix)
export(rGammaGpd)
export(rGammaMix)
export(rGammaMixGpd)
export(rGpd)
export(rInvGauss)
export(rInvGaussGpd)
export(rInvGaussMix)
export(rInvGaussMixGpd)
export(rLaplaceGpd)
export(rLaplaceMix)
export(rLaplaceMixGpd)
export(rLognormalGpd)
export(rLognormalMix)
export(rLognormalMixGpd)
export(rNormGpd)
export(rNormMix)
export(rNormMixGpd)
export(ramoroso)
export(ramorosogpd)
export(ramorosomix)
export(ramorosomixgpd)
export(rcauchy_vec)
export(rcauchymix)
export(rgammagpd)
export(rgammamix)
export(rgammamixgpd)
export(rgpd)
export(rinvgauss)
export(rinvgaussgpd)
export(rinvgaussmix)
export(rinvgaussmixgpd)
export(rlaplacegpd)
export(rlaplacemix)
export(rlaplacemixgpd)
export(rlognormalgpd)
export(rlognormalmix)
export(rlognormalmixgpd)
export(rnormgpd)
export(rnormmix)
export(rnormmixgpd)
export(run_mcmc_bundle_manual)
export(run_mcmc_causal)
export(sim_bulk_tail)
export(sim_causal_qte)
export(sim_survival_tail)
import(ggplot2)
import(nimble)
importFrom(nimble,ddexp)
importFrom(nimble,nimNumeric)
importFrom(nimble,pdexp)
importFrom(nimble,qdexp)
importFrom(nimble,rdexp)
importFrom(stats,coef)
importFrom(stats,dcauchy)
importFrom(stats,density)
importFrom(stats,dgamma)
importFrom(stats,dlnorm)
importFrom(stats,dnorm)
importFrom(stats,fitted)
importFrom(stats,pcauchy)
importFrom(stats,pgamma)
importFrom(stats,plnorm)
importFrom(stats,plogis)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qcauchy)
importFrom(stats,qgamma)
importFrom(stats,qlnorm)
importFrom(stats,qlogis)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,rcauchy)
importFrom(stats,residuals)
importFrom(stats,rexp)
importFrom(stats,rgamma)
importFrom(stats,rlnorm)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,uniroot)
importFrom(utils,capture.output)
importFrom(utils,head)

This folder contains precomputed artifacts used when building package vignettes.

The vignettes avoid storing full fitted MCMC objects because those can be very
large and unstable for package distribution and checks. Instead, we ship only
lightweight outputs (prediction summaries, effect objects, small tables, and
static diagnostic figures).

1) Lightweight vignette output bundles
   - one_arm_outputs.rds      — prediction/summary outputs for vignettes/cmgpd_one_arm.Rmd
   - causal_outputs.rds       — effect objects + prediction outputs for vignettes/cmgpd_causal.Rmd
   - clustering_outputs.rds   — PSM/label objects + profile tables for vignettes/cmgpd_clustering.Rmd

   Supporting static diagnostics (examples):
   - one_arm_fit_summary.txt
   - one_arm_alpha_trace.png
   - one_arm_alpha_density.png
   - causal_fit_summary.txt
   - clustering_fit_summary.txt

   Regenerate artifacts with:
     Rscript data-raw/build_vignette_fits.R
   from the package root (requires pkgload, nimble, and full MCMC dependencies).

2) Legacy CSV / figure exports (older static tables)
   - unconditional_quantiles.csv, conditional_quantiles.csv
   - causal_ate.csv, causal_qte.csv
   - causal_qte.png (if present)

Prebuilt vignette HTML may also be shipped under inst/doc/ where applicable.

End users do not need to regenerate these files to read the vignettes.

# CausalMixGPD replication entrypoint
#
# This script is the single package-level replication driver. It runs compact
# versions of the package workflows used in the manuscript and writes the
# resulting tables/figures to an output directory. Longer manuscript builds may
# source this file or reuse the same public API calls with larger MCMC settings.

args <- commandArgs(trailingOnly = TRUE)
out_dir <- if (length(args)) args[[1]] else file.path(getwd(), "CausalMixGPD-replication")
dir.create(out_dir, recursive = TRUE, showWarnings = FALSE)

required <- c("ggplot2")
missing <- required[!vapply(required, requireNamespace, logical(1), quietly = TRUE)]
if (length(missing)) {
  stop("Install required package(s): ", paste(missing, collapse = ", "), call. = FALSE)
}

if (file.exists("DESCRIPTION") &&
    any(grepl("^Package:\\s*CausalMixGPD\\s*$", readLines("DESCRIPTION", warn = FALSE)))) {
  if (!requireNamespace("devtools", quietly = TRUE)) {
    stop("Install 'devtools' to run replication from a source checkout.", call. = FALSE)
  }
  devtools::load_all(quiet = TRUE)
} else {
  if (!requireNamespace("CausalMixGPD", quietly = TRUE)) {
    stop("Install 'CausalMixGPD' or run this script from the package source root.", call. = FALSE)
  }
  library(CausalMixGPD)
}

set.seed(2026)
replication_seed <- 2026
mcmc_rep <- list(
  niter = 80,
  nburnin = 40,
  thin = 1,
  nchains = 1,
  seed = 2026,
  show_progress = FALSE,
  quiet = TRUE
)

write_table <- function(x, name) {
  utils::write.csv(x, file.path(out_dir, name), row.names = FALSE)
}

write_text <- function(x, name) {
  writeLines(x, file.path(out_dir, name), useBytes = TRUE)
}

save_plot <- function(p, name, width = 7, height = 5) {
  if (requireNamespace("ggplot2", quietly = TRUE)) {
    ggplot2::ggsave(file.path(out_dir, name), plot = p, width = width, height = height)
  }
}

pkg_version <- tryCatch(as.character(utils::packageVersion("CausalMixGPD")), error = function(e) NA_character_)
output_manifest <- data.frame(
  file = character(),
  workflow = character(),
  manuscript_use = character(),
  stringsAsFactors = FALSE
)
record_output <- function(file, workflow, manuscript_use) {
  output_manifest <<- rbind(
    output_manifest,
    data.frame(
      file = file,
      workflow = workflow,
      manuscript_use = manuscript_use,
      stringsAsFactors = FALSE
    )
  )
}

# One-arm workflow -----------------------------------------------------------
data("nc_pos200_k3", package = "CausalMixGPD")
idx_one <- seq_len(30)
fit_one <- dpmix(
  y = nc_pos200_k3$y[idx_one],
  kernel = "gamma",
  backend = "sb",
  components = 3,
  mcmc = mcmc_rep
)
write_table(summary(fit_one)$table, "one_arm_summary.csv")
record_output("one_arm_summary.csv", "one-arm", "Package overview one-arm posterior summary")
pred_one <- predict(fit_one, type = "quantile", index = c(0.25, 0.5, 0.75), show_progress = FALSE)
write_table(pred_one$fit, "one_arm_quantiles.csv")
record_output("one_arm_quantiles.csv", "one-arm", "Package overview quantile prediction table")
save_plot(plot(pred_one), "one_arm_quantiles.png")
record_output("one_arm_quantiles.png", "one-arm", "Package overview quantile prediction figure")

# Clustering workflow --------------------------------------------------------
data("nc_realX100_p3_k2", package = "CausalMixGPD")
dat_cl <- data.frame(
  y = nc_realX100_p3_k2$y[1:30],
  nc_realX100_p3_k2$X[1:30, , drop = FALSE]
)
fit_cluster <- dpmix.cluster(
  y ~ x1 + x2 + x3,
  data = dat_cl,
  kernel = "normal",
  type = "param",
  components = 3,
  mcmc = mcmc_rep
)
labels_cluster <- predict(fit_cluster, type = "label", return_scores = TRUE)
write_table(cluster_profiles(labels_cluster), "cluster_profiles.csv")
record_output("cluster_profiles.csv", "clustering", "Cluster profile table via public accessor")
save_plot(plot(labels_cluster, type = "sizes"), "cluster_sizes.png")
record_output("cluster_sizes.png", "clustering", "Cluster assignment size figure")

# Causal workflow ------------------------------------------------------------
data("causal_pos500_p3_k2", package = "CausalMixGPD")
idx_causal <- seq_len(40)
fit_causal <- dpmix.causal(
  y = causal_pos500_p3_k2$y[idx_causal],
  X = causal_pos500_p3_k2$X[idx_causal, , drop = FALSE],
  treat = causal_pos500_p3_k2$A[idx_causal],
  kernel = "gamma",
  backend = "sb",
  components = 3,
  mcmc = mcmc_rep
)
ate_rep <- ate(fit_causal, interval = "credible", show_progress = FALSE)
write_table(summary(ate_rep)$effect_table, "causal_ate.csv")
record_output("causal_ate.csv", "causal", "Causal ATE summary table")
save_plot(plot(ate_rep, type = "effect"), "causal_ate.png")
record_output("causal_ate.png", "causal", "Causal ATE effect figure")

write_table(output_manifest, "manifest.csv")
write_text(
  c(
    "CausalMixGPD replication run",
    paste("Package version:", pkg_version),
    paste("Seed:", replication_seed),
    paste("Output directory:", normalizePath(out_dir, winslash = "/")),
    "",
    "Session info:",
    utils::capture.output(utils::sessionInfo())
  ),
  "session-info.txt"
)
record_output("manifest.csv", "metadata", "Output-to-manuscript mapping")
record_output("session-info.txt", "metadata", "R session, package version, seed, and platform")
write_table(output_manifest, "manifest.csv")

cat("Replication outputs written to:", normalizePath(out_dir, winslash = "/"), "\n")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.8,
  out.width = "85%",
  fig.align = "center",
  message = FALSE,
  warning = FALSE
)
.pkg_extdata <- function(name) {
  f <- system.file("extdata", name, package = "CausalMixGPD")
  if (nzchar(f)) {
    return(f)
  }
  vdir <- knitr::current_input(dir = TRUE)
  pkg_root <- normalizePath(file.path(vdir, ".."), winslash = "/", mustWork = TRUE)
  alt <- file.path(pkg_root, "inst", "extdata", name)
  if (file.exists(alt)) {
    return(normalizePath(alt, winslash = "/", mustWork = TRUE))
  }
  stop("Missing extdata: ", name, ".", call. = FALSE)
}

## ----libraries----------------------------------------------------------------
library(CausalMixGPD)
library(ggplot2)

## ----mcmc-settings------------------------------------------------------------
mcmc_vig <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2026
)
# Settings used when regenerating `inst/extdata/causal_*.csv` (see tools/.Rscripts/build_vignette_fits.R).
mcmc_out <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2027
)
mcmc_ps <- list(
  niter = 1000,
  nburnin = 250,
  thin = 2,
  nchains = 2,
  seed = 2028
)

## ----load-data----------------------------------------------------------------
data("causal_pos500_p3_k2", package = "CausalMixGPD")
causal_dat <- data.frame(
  y = causal_pos500_p3_k2$y,
  A = causal_pos500_p3_k2$A,
  causal_pos500_p3_k2$X
)

## ----exploratory-plots, fig.height=4.8, fig.width=7---------------------------
ggplot(causal_dat, aes(x = y, colour = factor(A), fill = factor(A))) +
  geom_histogram(aes(y = after_stat(density)), bins = 25,
                 alpha = 0.25, position = "identity") +
  labs(
    x = "y",
    y = "Density",
    colour = "A",
    fill = "A",
    title = "Observed outcome distributions by treatment arm"
  ) +
  theme_minimal()

## ----fit-causal-display, eval=FALSE-------------------------------------------
# fit_causal <- dpmgpd.causal(
#   formula = y ~ x1 + x2 + x3,
#   data = causal_dat,
#   treat = "A",
#   backend = "crp",
#   kernel = "gamma",
#   components = 6,
#   PS = "logit",
#   ps_scale = "logit",
#   ps_summary = "mean",
#   mcmc_outcome = mcmc_out,
#   mcmc_ps = mcmc_ps,
#   parallel_arms = FALSE
# )

## ----ate-summary-display, eval=FALSE------------------------------------------
# ate_fit <- ate(fit_causal, interval = "hpd", level = 0.95, show_progress = FALSE)
# knitr::kable(summary(ate_fit)$effect_table, format = "html", digits = 4)

## ----ate-summary, echo=FALSE--------------------------------------------------
ate_tab <- read.csv(.pkg_extdata("causal_ate.csv"))
knitr::kable(ate_tab, format = "html", digits = 4)

## ----qte-summary-display, eval=FALSE------------------------------------------
# qte_fit <- qte(
#   fit_causal,
#   probs = c(0.25, 0.50, 0.75, 0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   show_progress = FALSE
# )
# knitr::kable(summary(qte_fit)$effect_table, format = "html", digits = 4)

## ----qte-summary, echo=FALSE--------------------------------------------------
qte_tab <- read.csv(.pkg_extdata("causal_qte.csv"))
knitr::kable(qte_tab, format = "html", digits = 4)

## ----qte-plot-display, eval=FALSE---------------------------------------------
# qte_fit <- qte(
#   fit_causal,
#   probs = c(0.25, 0.50, 0.75, 0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   show_progress = FALSE
# )
# plot(qte_fit, type = "effect")

## ----qte-plot, echo=FALSE, fig.height=5, fig.width=7--------------------------
plot(
  qte_tab$prob,
  qte_tab$estimate,
  type = "b",
  pch = 19,
  xlab = "Quantile level",
  ylab = "Estimated QTE",
  ylim = range(c(qte_tab$lower, qte_tab$upper)),
  main = "Quantile treatment effect curve"
)
segments(qte_tab$prob, qte_tab$lower, qte_tab$prob, qte_tab$upper, lwd = 1.2)
abline(h = 0, lty = 2, col = "grey50")

## ----newdata-grid-------------------------------------------------------------
qs <- c(0.25, 0.50, 0.75)
Xgrid <- expand.grid(lapply(causal_dat[, c("x1", "x2", "x3")], quantile, probs = qs))
head(Xgrid)

## ----conditional-qte-table-display, eval=FALSE--------------------------------
# Xgrid <- as.data.frame(lapply(
#   causal_dat[, c("x1", "x2", "x3")],
#   stats::quantile,
#   probs = c(0.25, 0.50, 0.75),
#   na.rm = TRUE
# ))
# rownames(Xgrid) <- c("low", "mid", "high")
# cqte_fit <- cqte(
#   fit_causal,
#   newdata = Xgrid,
#   probs = c(0.25, 0.50, 0.75, 0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   show_progress = FALSE
# )
# knitr::kable(summary(cqte_fit)$effect_table, format = "html", digits = 4)

## ----conditional-qte-table, echo=FALSE----------------------------------------
cond_q <- read.csv(.pkg_extdata("conditional_quantiles.csv"))
knitr::kable(cond_q, format = "html", digits = 4)

## ----session-info-------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.8,
  out.width = "85%",
  fig.align = "center",
  message = FALSE,
  warning = FALSE
)
.pkg_extdata <- function(name) {
  f <- system.file("extdata", name, package = "CausalMixGPD")
  if (nzchar(f)) {
    return(f)
  }
  vdir <- knitr::current_input(dir = TRUE)
  pkg_root <- normalizePath(file.path(vdir, ".."), winslash = "/", mustWork = TRUE)
  alt <- file.path(pkg_root, "inst", "extdata", name)
  if (file.exists(alt)) {
    return(normalizePath(alt, winslash = "/", mustWork = TRUE))
  }
  stop("Missing extdata: ", name, ". Run data-raw/build_vignette_fits.R", call. = FALSE)
}

## ----libraries----------------------------------------------------------------
library(CausalMixGPD)
library(ggplot2)

## ----mcmc-settings------------------------------------------------------------
mcmc_vig <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2026
)

## ----load-data----------------------------------------------------------------
data("nc_posX100_p3_k2", package = "CausalMixGPD")
onearm_dat <- data.frame(
  y = nc_posX100_p3_k2$y,
  nc_posX100_p3_k2$X
)


## ----exploratory-plots, fig.height=4.5, fig.width=7---------------------------
p1 <- ggplot(onearm_dat, aes(x = y)) +
  geom_histogram(aes(y = after_stat(density)), bins = 25,
                 fill = "grey85", colour = "grey35") +
  geom_density(linewidth = 0.9) +
  labs(x = "y", y = "Density",
       title = "Observed response distribution") +
  theme_minimal()

p2 <- ggplot(onearm_dat, aes(sample = y)) +
  stat_qq() +
  stat_qq_line(colour = 2) +
  labs(title = "Normal Q-Q plot for the response") +
  theme_minimal()

p1
p2

## ----fit-spliced-display, eval=FALSE------------------------------------------
# fit_spliced <- dpmgpd(
#   formula = y ~ x1 + x2 + x3,
#   data = onearm_dat,
#   backend = "sb",
#   kernel = "lognormal",
#   components = 5,
#   mcmc = mcmc_vig
# )

## ----fit-spliced, echo=FALSE--------------------------------------------------
one_arm_out <- readRDS(.pkg_extdata("one_arm_outputs.rds"))

## ----summary-spliced-display, eval=FALSE--------------------------------------
# summary(fit_spliced)

## ----summary-spliced, echo=FALSE----------------------------------------------
fsum <- system.file("extdata", "one_arm_fit_summary.txt", package = "CausalMixGPD")
if (!nzchar(fsum)) {
  alt <- normalizePath(
    file.path(knitr::current_input(dir = TRUE), "..", "inst", "extdata", "one_arm_fit_summary.txt"),
    winslash = "/",
    mustWork = FALSE
  )
  if (file.exists(alt)) fsum <- alt
}
if (nzchar(fsum) && file.exists(fsum)) {
  cat(paste(readLines(fsum, warn = FALSE, encoding = "UTF-8"), collapse = "\n"))
} else {
  cat("Precomputed fit summary not found; run tools/.Rscripts/build_vignette_fits.R.\n")
}

## ----plot-spliced-trace-display, eval=FALSE-----------------------------------
# plot(fit_spliced, family = c("traceplot", "density"), params = "alpha")

## ----plot-spliced-trace, echo=FALSE, fig.height=5.2, fig.width=7--------------
knitr::include_graphics("assets/one_arm_alpha_trace.png")

## ----plot-spliced-density-display, eval=FALSE---------------------------------
# # Posterior density for monitored parameters (e.g. alpha); see plot.mixgpd_fit().
# plot(fit_spliced, family = "density", params = "alpha")

## ----plot-spliced-density, echo=FALSE, fig.height=5, fig.width=7--------------
knitr::include_graphics("assets/one_arm_alpha_density.png")

## ----newdata-grid-display, eval=FALSE-----------------------------------------
# x_new <- as.data.frame(lapply(
#   onearm_dat[, c("x1", "x2", "x3")],
#   stats::quantile,
#   probs = c(0.25, 0.50, 0.75),
#   na.rm = TRUE
# ))
# rownames(x_new) <- c("q25", "q50", "q75")
# x_new

## ----newdata-grid, echo=FALSE-------------------------------------------------
x_new <- one_arm_out$x_new
x_new

## ----density-prediction-display, eval=FALSE-----------------------------------
# y_grid <- seq(min(onearm_dat$y), quantile(onearm_dat$y, 0.99), length.out = 120)
# pred_dens <- predict(
#   fit_spliced,
#   newdata = x_new,
#   y = y_grid,
#   type = "density",
#   interval = "credible",
#   level = 0.95,
#   store_draws = FALSE
# )
# plot(pred_dens, type = "density", facet = "covariate")

## ----density-prediction, echo=FALSE-------------------------------------------
pred_dens <- one_arm_out$pred_dens
plot(pred_dens, type = "density", facet = "covariate")

## ----survival-prediction-display, eval=FALSE----------------------------------
# y_grid <- seq(min(onearm_dat$y), quantile(onearm_dat$y, 0.99), length.out = 120)
# pred_surv <- predict(
#   fit_spliced,
#   newdata = x_new,
#   y = y_grid,
#   type = "survival",
#   interval = "credible",
#   level = 0.95,
#   store_draws = FALSE
# )
# plot(pred_surv)

## ----survival-prediction, echo=FALSE------------------------------------------
pred_surv <- one_arm_out$pred_surv
plot(pred_surv)

## ----quantile-prediction-display, eval=FALSE----------------------------------
# pred_quant <- predict(
#   fit_spliced,
#   newdata = x_new,
#   type = "quantile",
#   index = c(0.25, 0.50, 0.75, 0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   store_draws = FALSE
# )

## ----quantile-prediction, echo=FALSE------------------------------------------
pred_quant <- one_arm_out$pred_quant

## ----quantile-prediction-plot-display, eval=FALSE-----------------------------
# plot(pred_quant)

## ----quantile-prediction-plot, echo=FALSE, fig.height=5.5, fig.width=7--------
plot(pred_quant)

## ----mean-prediction-display, eval=FALSE--------------------------------------
# pred_mean <- predict(
#   fit_spliced,
#   newdata = x_new,
#   type = "mean",
#   interval = "hpd",
#   level = 0.90,
#   store_draws = FALSE
# )

## ----mean-prediction, echo=FALSE----------------------------------------------
pred_mean <- one_arm_out$pred_mean

## ----mean-prediction-plot-display, eval=FALSE---------------------------------
# plot(pred_mean)

## ----mean-prediction-plot, echo=FALSE, fig.height=5.2, fig.width=7------------
plot(pred_mean)

## ----rmean-prediction-display, eval=FALSE-------------------------------------
# cutoff_val <- as.numeric(stats::quantile(onearm_dat$y, 0.95))
# pred_rmean <- predict(
#   fit_spliced,
#   newdata = x_new,
#   type = "rmean",
#   cutoff = cutoff_val,
#   interval = "hpd",
#   level = 0.90,
#   store_draws = FALSE
# )

## ----rmean-prediction, echo=FALSE---------------------------------------------
cutoff_val <- one_arm_out$cutoff_val

## ----rmean-table-display, eval=FALSE------------------------------------------
# knitr::kable(pred_rmean$fit_df, format = "html", row.names = FALSE, digits = 4)

## ----rmean-table, echo=FALSE--------------------------------------------------
knitr::kable(
  one_arm_out$pred_rmean_fit_df,
  format = "html",
  row.names = FALSE,
  digits = 4
)

## ----fit-bulk-only-display, eval=FALSE----------------------------------------
# fit_bulk <- dpmix(
#   formula = y ~ x1 + x2 + x3,
#   data = onearm_dat,
#   backend = "sb",
#   kernel = "lognormal",
#   components = 5,
#   mcmc = mcmc_vig
# )

## ----fit-bulk-only, echo=FALSE------------------------------------------------
cat("Bulk-only fit computed offline; see quantile comparison below.\n")

## ----compare-quantiles-display, eval=FALSE------------------------------------
# quant_bulk_fit_df <- predict(
#   fit_bulk,
#   newdata = x_new,
#   type = "quantile",
#   index = c(0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   store_draws = FALSE
# )$fit_df
# quant_spliced_fit_df <- predict(
#   fit_spliced,
#   newdata = x_new,
#   type = "quantile",
#   index = c(0.90, 0.95),
#   interval = "credible",
#   level = 0.95,
#   store_draws = FALSE
# )$fit_df

## ----compare-quantiles, echo=FALSE--------------------------------------------
quant_bulk_fit_df <- one_arm_out$quant_bulk_fit_df
quant_spliced_fit_df <- one_arm_out$quant_spliced_fit_df

## ----compare-quantiles-plot-display, eval=FALSE-------------------------------
# qb <- quant_bulk_fit_df
# qb$model <- "dpmix (bulk only)"
# qs <- quant_spliced_fit_df
# qs$model <- "dpmgpd (spliced)"
# qc <- rbind(qb, qs)
# qc$profile <- rownames(x_new)[as.integer(qc$id)]
# ggplot(qc, aes(x = profile, y = estimate, colour = model, group = model)) +
#   geom_line(linewidth = 0.8) +
#   geom_point(size = 2) +
#   facet_wrap(~index, scales = "free_y", ncol = 2) +
#   labs(
#     x = "Covariate profile",
#     y = "Posterior mean quantile",
#     title = "Upper quantiles: bulk-only vs spliced tail"
#   ) +
#   theme_minimal() +
#   theme(axis.text.x = element_text(angle = 20, hjust = 1))

## ----compare-quantiles-plot, echo=FALSE, fig.height=5, fig.width=7------------
qb <- quant_bulk_fit_df
qb$model <- "dpmix (bulk only)"
qs <- quant_spliced_fit_df
qs$model <- "dpmgpd (spliced)"
qc <- rbind(qb, qs)
qc$profile <- rownames(x_new)[as.integer(qc$id)]
ggplot(qc, aes(x = profile, y = estimate, colour = model, group = model)) +
  geom_line(linewidth = 0.8) +
  geom_point(size = 2) +
  facet_wrap(~index, scales = "free_y", ncol = 2) +
  labs(
    x = "Covariate profile",
    y = "Posterior mean quantile",
    title = "Upper quantiles: bulk-only vs spliced tail"
  ) +
  theme_minimal() +
  theme(axis.text.x = element_text(angle = 20, hjust = 1))

## ----session-info-------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.8,
  out.width = "85%",
  fig.align = "center",
  message = FALSE,
  warning = FALSE
)

.pkg_extdata <- function(name) {
  f <- system.file("extdata", name, package = "CausalMixGPD")
  if (nzchar(f)) {
    return(f)
  }
  vdir <- knitr::current_input(dir = TRUE)
  pkg_root <- normalizePath(file.path(vdir, ".."), winslash = "/", mustWork = TRUE)
  alt <- file.path(pkg_root, "inst", "extdata", name)
  if (file.exists(alt)) {
    return(normalizePath(alt, winslash = "/", mustWork = TRUE))
  }
  stop("Missing extdata: ", name, ". Run tools/.Rscripts/build_vignette_fits.R", call. = FALSE)
}

## ----libraries----------------------------------------------------------------
library(CausalMixGPD)
library(ggplot2)

## ----mcmc-settings------------------------------------------------------------
mcmc_vig <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2026
)
# Used when regenerating clustering vignette artifacts (tools/.Rscripts/build_vignette_fits.R).
mcmc_clust <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2029
)

## ----load-data----------------------------------------------------------------
data("nc_realX100_p3_k2", package = "CausalMixGPD")
dat_cl <- data.frame(
  y = nc_realX100_p3_k2$y,
  nc_realX100_p3_k2$X
)

## ----split-data-display, eval=FALSE-------------------------------------------
# set.seed(123)
# idx_train <- sample(seq_len(nrow(dat_cl)), size = 80)
# train_dat <- dat_cl[idx_train, ]
# test_dat <- dat_cl[-idx_train, ]
# nrow(train_dat)
# nrow(test_dat)

## ----split-data, echo=FALSE---------------------------------------------------
clust_out <- readRDS(.pkg_extdata("clustering_outputs.rds"))
train_dat <- clust_out$train_dat
test_dat <- clust_out$test_dat

nrow(train_dat)
nrow(test_dat)

## ----exploratory-plots, fig.height=4.8, fig.width=7---------------------------
ggplot(train_dat, aes(x = y)) +
  geom_histogram(aes(y = after_stat(density)), bins = 25,
                 fill = "grey85", colour = "grey35") +
  geom_density(linewidth = 0.9) +
  labs(x = "y", y = "Density", title = "Training-sample response distribution") +
  theme_minimal()

## ----fit-cluster-display, eval=FALSE------------------------------------------
# fit_cluster <- dpmix.cluster(
#   formula = y ~ x1 + x2 + x3,
#   data = train_dat,
#   kernel = "laplace",
#   type = "both",
#   components = 8,
#   mcmc = mcmc_clust
# )

## ----fit-cluster, echo=FALSE--------------------------------------------------
fit_cluster <- clust_out$fit_cluster

## ----psm-prediction-display, eval=FALSE---------------------------------------
# z_train_psm <- predict(fit_cluster, type = "psm")
# z_train_psm

## ----psm-prediction, echo=FALSE-----------------------------------------------
z_train_psm <- clust_out$psm_obj
z_train_psm

## ----psm-plot-display, eval=FALSE---------------------------------------------
# plot(z_train_psm, type = "summary")

## ----psm-plot, echo=FALSE, fig.height=5.2, fig.width=7------------------------
plot(z_train_psm, type = "summary")

## ----label-prediction-train-display, eval=FALSE-------------------------------
# z_train_lab <- predict(fit_cluster, type = "label", return_scores = TRUE)
# z_train_lab

## ----label-prediction-train, echo=FALSE---------------------------------------
z_train_lab <- clust_out$train_lab
z_train_lab

## ----label-plot-train-display, eval=FALSE-------------------------------------
# plot(z_train_lab, type = "summary")

## ----label-plot-train, echo=FALSE, fig.height=5, fig.width=7------------------
plot(z_train_lab, type = "summary")

## ----label-prediction-test-display, eval=FALSE--------------------------------
# z_test <- predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
# z_test

## ----label-prediction-test, echo=FALSE----------------------------------------
z_test <- clust_out$test_lab
z_test

## ----test-cluster-profiles-display, eval=FALSE--------------------------------
# summary(
#   predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
# )
# cluster_profiles(
#   predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
# )

## ----test-cluster-profiles, echo=FALSE----------------------------------------
cluster_profiles(z_test)

## ----sizes-plot-test-display, eval=FALSE--------------------------------------
# plot(z_test, type = "sizes")

## ----sizes-plot-test, echo=FALSE, fig.height=4.8, fig.width=7-----------------
plot(z_test, type = "sizes")

## ----session-info-------------------------------------------------------------
sessionInfo()

# helper-test-levels.R
# Tiered testing infrastructure for CausalMixGPD
#
# Test Tiers:
#   - "cran": Fast unit/contract tests only (~seconds to minutes). Default for R CMD check.
#   - "ci":   Integration tests (Tier A + B). For CI pipelines.
#   - "full": Exhaustive combo grids (Tier A + B + C). Nightly/local only.
#
# Environment variable: DPMIXGPD_TEST_LEVEL (default: "cran")
#
# Usage in test files:
#   skip_if_not_test_level("ci")   # Skip unless level is "ci" or "full"
#   skip_if_not_test_level("full") # Skip unless level is "full"

#' Get current test level
#' @return Character: "cran", "ci", or "full"
#' @keywords internal
get_test_level <- function() {

  level <- tolower(Sys.getenv("DPMIXGPD_TEST_LEVEL", "cran"))
  if (!level %in% c("cran", "ci", "full")) {
    warning("Unknown DPMIXGPD_TEST_LEVEL '", level, "'; defaulting to 'cran'")
    level <- "cran"
  }
  level
}

#' Check if current test level meets minimum requirement
#' @param min_level Minimum required level: "cran", "ci", or "full"
#' @return Logical
#' @keywords internal
test_level_at_least <- function(min_level) {
  levels <- c("cran" = 1L, "ci" = 2L, "full" = 3L)
  current <- get_test_level()
  current_rank <- levels[[current]]
  min_rank <- levels[[min_level]]
  if (is.na(min_rank)) stop("Unknown test level: ", min_level)
  current_rank >= min_rank
}

#' Skip test if current level is below minimum
#' @param min_level Minimum required level: "ci" or "full"
#' @keywords internal
skip_if_not_test_level <- function(min_level) {
  if (!test_level_at_least(min_level)) {
    testthat::skip(paste0(
      "Test level '", min_level, "' required (current: '", get_test_level(), "'). ",
      "Set DPMIXGPD_TEST_LEVEL='", min_level, "' to run."
    ))
  }
}

#' Skip slow/integration tests (convenience wrapper)
#' @keywords internal
skip_if_cran <- function() {
  skip_if_not_test_level("ci")
}

#' Skip exhaustive tests (convenience wrapper)
#' @keywords internal
skip_if_not_full <- function() {
  skip_if_not_test_level("full")
}

#' Standard minimal MCMC settings for fast tests
#' @keywords internal
mcmc_fast <- function(seed = 1L) {
  list(niter = 20L, nburnin = 5L, thin = 1L, nchains = 1L, seed = seed)
}

#' Standard MCMC settings for integration tests (slightly longer)
#' @keywords internal
mcmc_integration <- function(seed = 1L) {
  list(niter = 50L, nburnin = 15L, thin = 1L, nchains = 1L, seed = seed)
}

#' Standard MCMC settings for exhaustive tests
#' @keywords internal
mcmc_full <- function(seed = 1L) {
  list(niter = 200L, nburnin = 50L, thin = 2L, nchains = 1L, seed = seed)
}

#' Representative kernel/backend/GPD combos for integration tests
#' Covers the main branches without exhaustive enumeration
#' @keywords internal
representative_combos <- function() {
  list(
    # Normal: most common, test both backends
    list(kernel = "normal", backend = "sb",  GPD = FALSE, label = "normal_sb_bulk"),
    list(kernel = "normal", backend = "crp", GPD = TRUE,  label = "normal_crp_gpd"),
    # Gamma: positive support, different code path

    list(kernel = "gamma",  backend = "sb",  GPD = TRUE,  label = "gamma_sb_gpd"),
    # Lognormal: another positive kernel
    list(kernel = "lognormal", backend = "crp", GPD = FALSE, label = "lognormal_crp_bulk"),
    # Spliced: component-level GPD flexibility
    list(kernel = "gamma", backend = "spliced", GPD = TRUE, label = "gamma_spliced_gpd")
  )
}

#' Representative causal combos for integration tests
#' @keywords internal
representative_causal_combos <- function() {
  list(
    # Standard observational with PS
    list(
      kernel = c("normal", "normal"),
      backend = c("sb", "sb"),
      GPD = c(FALSE, FALSE),
      PS = "logit",
      label = "normal_sb_logit"
    ),
    # RCT without PS
    list(
      kernel = c("gamma", "gamma"),
      backend = c("crp", "crp"),
      GPD = c(TRUE, TRUE),
      PS = FALSE,
      label = "gamma_crp_gpd_rct"
    ),
    # Mixed backends
    list(
      kernel = c("lognormal", "normal"),
      backend = c("sb", "crp"),
      GPD = c(FALSE, FALSE),
      PS = "probit",
      label = "mixed_probit"
    )
  )
}

# Message at load time if verbose
if (Sys.getenv("DPMIXGPD_TEST_VERBOSE", "0") %in% c("1", "true", "TRUE")) {
  message("[CausalMixGPD tests] Test level: ", get_test_level())
}

test_that("predict applies large-N defaults and supports cache hits", {
  skip_if_not_test_level("full")
  set.seed(531)
  n <- 120
  X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
  y <- abs(0.4 * X[, 1] + stats::rnorm(n)) + 0.1
  fit <- dpmgpd(
    x = y, X = X, backend = "sb", kernel = "normal", components = 6,
    mcmc = list(niter = 100, nburnin = 30, thin = 1, nchains = 1, seed = 17),
    monitor_latent = TRUE
  )
  X_new <- X[rep(seq_len(nrow(X)), length.out = 20050), , drop = FALSE]

  p1 <- suppressMessages(predict(fit, newdata =X_new, type = "quantile", index = c(0.5, 0.9), parallel = FALSE))
  p2 <- suppressMessages(predict(fit, newdata =X_new, type = "quantile", index = c(0.5, 0.9), parallel = FALSE))

  expect_s3_class(p1, "mixgpd_predict")
  expect_s3_class(p2, "mixgpd_predict")
  expect_equal(dim(p1$fit), dim(p2$fit))
  expect_true(isTRUE(p2$diagnostics$cache_hit))
})

test_that("cluster S3 methods run without error", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(789)
  dat <- data.frame(
    y = abs(stats::rnorm(18)) + 0.2,
    x1 = stats::rnorm(18),
    x2 = stats::runif(18)
  )

  fit <- dpmix.cluster(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    components = 4,
    type = "weights",
    mcmc = mcmc_fast(seed = 30L)
  )
  b <- fit$bundle
  lbl <- predict(fit, type = "label", return_scores = TRUE)
  lbl_new <- predict(fit, newdata = dat[1:6, , drop = FALSE], type = "label", return_scores = TRUE)
  psm <- predict(fit, type = "psm")

  expect_output(print(b), "Cluster bundle")
  expect_silent(summary(b))
  expect_s3_class(plot(b, plotly = FALSE), "ggplot")

  expect_output(print(fit), "Cluster fit")
  fit_sum <- summary(fit)
  expect_s3_class(fit_sum, "summary.dpmixgpd_cluster_fit")
  expect_true(is.data.frame(fit_sum$cluster_profiles))
  expect_true(all(c("cluster", "n", "y_mean", "y_sd", "x1_mean", "x1_sd", "x2_mean", "x2_sd", "certainty_mean", "certainty_sd") %in% names(fit_sum$cluster_profiles)))
  expect_true(all(diff(as.integer(fit_sum$cluster_sizes)) <= 0))
  expect_s3_class(plot(fit, which = "psm", plotly = FALSE), "ggplot")
  expect_s3_class(plot(fit, which = "k", plotly = FALSE), "ggplot")
  expect_s3_class(plot(fit, which = "sizes", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")
  expect_s3_class(plot(fit, which = "summary", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")

  expect_output(print(lbl), "Cluster labels")
  lbl_sum <- summary(lbl, top_n = 2L)
  expect_s3_class(lbl_sum, "summary.dpmixgpd_cluster_labels")
  expect_true(is.data.frame(lbl_sum$cluster_profiles))
  expect_lte(nrow(lbl_sum$cluster_profiles), 2L)
  expect_true(all(diff(as.integer(lbl_sum$cluster_sizes)) <= 0))
  expect_s3_class(plot(lbl, type = "sizes", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")
  expect_s3_class(plot(lbl, type = "certainty", plotly = FALSE), "ggplot")
  expect_s3_class(plot(lbl, type = "summary", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")

  expect_s3_class(lbl_new, "dpmixgpd_cluster_labels")
  expect_true(is.list(lbl_new$train_reference))
  expect_s3_class(plot(lbl_new, type = "summary", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")

  expect_output(print(psm), "Cluster PSM")
  expect_silent(summary(psm))
  expect_s3_class(plot(psm, psm_max_n = nrow(psm$psm), plotly = FALSE), "ggplot")

  if (requireNamespace("plotly", quietly = TRUE)) {
    expect_true(inherits(plot(lbl, type = "sizes", plotly = TRUE), "plotly"))
  }
})

test_that("cluster fit summary rounds cluster profiles to 3 decimals", {
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(321)
  dat <- data.frame(
    y = abs(stats::rnorm(18)) + 0.2,
    x1 = stats::rnorm(18),
    x2 = stats::runif(18)
  )

  fit <- dpmix.cluster(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    components = 4,
    type = "weights",
    mcmc = mcmc_fast(seed = 31L)
  )

  fit_sum <- summary(fit, top_n = 2L)
  prof <- fit_sum$cluster_profiles
  num_cols <- vapply(prof, is.numeric, logical(1))

  expect_true(all(prof[num_cols] == lapply(prof[num_cols], round, digits = 3L)))
})

test_that("monitor profile supports core/full policy", {
  skip_if_not_test_level("ci")
  set.seed(511)
  y <- abs(stats::rnorm(80)) + 0.1
  b_core <- build_nimble_bundle(
    y = y, backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    monitor = "core"
  )
  b_full <- build_nimble_bundle(
    y = y, backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    monitor = "full"
  )
  expect_false(any(grepl("^z\\[", b_core$monitors)))
  expect_true(any(grepl("^z\\[", b_full$monitors)))
  expect_true(any(grepl("^v\\[", b_full$monitors)))
})

test_that("mcmc overrides accept z_update_every", {
  skip_if_not_test_level("ci")
  set.seed(512)
  y <- abs(stats::rnorm(80)) + 0.1
  fit <- dpmgpd(
    x = y,
    backend = "sb",
    kernel = "normal",
    components = 6,
    mcmc = list(niter = 100, nburnin = 30, thin = 1, nchains = 1, seed = 14, z_update_every = 2)
  )
  expect_equal(fit$mcmc$z_update_every, 2L)
})

#' Compile model specification (pre-NIMBLE)
#'
#' Build a fully normalized model specification for later bundle construction.
#' This function is structural: it decides node types, dimensions, defaults, and
#' how parameters are represented (fixed / dist / link) but does not run NIMBLE.
#'
#' The model size is controlled by a single user-facing parameter:
#' \code{components}. For the stick-breaking (SB) backend it is the truncation
#' level of the finite mixture. For the CRP backend it is the maximum number of
#' represented clusters in the finite NIMBLE model.
#'
#' @param y Numeric outcome vector.
#' @param X Optional design matrix (N x P). Can be matrix or data.frame.
#' @param backend Dirichlet process representation:
#'   \itemize{
#'     \item \code{"sb"}: stick-breaking truncation
#'     \item \code{"crp"}: Chinese Restaurant Process
#'   }
#' @param kernel Kernel name; must exist in \code{get_kernel_registry()}.
#' @param GPD Logical; include GPD tail if TRUE.
#' @param components Integer >= 2; single truncation parameter used for both backends.
#' @param param_specs Optional list to override defaults. Expected structure:
#' \preformatted{
#' list(
#'   bulk = list(
#'     <param> = list(mode="fixed"/"dist"/"link", ...),
#'     ...
#'   ),
#'   gpd = list(
#'     threshold = list(...),
#'     tail_scale = list(...),
#'     tail_shape = list(...),
#'     ...
#'   ),
#'   concentration = list( ... ) # optional alpha override
#'   ps = list(
#'     prior = list(dist = "normal", args = list(mean = 0, sd = 2))
#'   )
#' )
#' }
#' @param alpha_random Logical; if TRUE, the DP concentration parameter (\eqn{\kappa}) is stochastic with default Gamma(1,1) prior.
#' @param ps Optional numeric vector of propensity scores (length N). When provided, the
#'   compiled spec will include \code{beta_ps_<param>} coefficients for link-mode parameters.
#' @param ... Unused; accepted for forward compatibility.
#'
#' @return A named list \code{spec} containing \code{meta}, \code{kernel_info},
#' \code{signatures}, and a canonical \code{plan}.
#' @keywords internal
#' @noRd
compile_model_spec <- function(
    y,
    X = NULL,
    backend = c("sb", "crp", "spliced"),
    kernel,
    GPD = FALSE,
    components,
    param_specs = NULL,
    alpha_random = TRUE,
    ps = NULL,
    ...
) {

  backend <- match.arg(backend, choices = allowed_backends)

  y <- as.numeric(y)
  if (!length(y)) stop("y must be a non-empty numeric vector.", call. = FALSE)
  N <- length(y)

  has_X <- !is.null(X)
  if (has_X) {
    if (!is.matrix(X)) X <- as.matrix(X)
    if (nrow(X) != N) stop("X must have the same number of rows as length(y).", call. = FALSE)
    P <- ncol(X)
    if (P < 1) stop("X must have at least one column.", call. = FALSE)
    if (!is.null(colnames(X))) {
      .validate_nimble_reserved_names(colnames(X), "X column names")
    }
  } else {
    P <- 0L
  }

  # ps is optional; validate if provided
  if (!is.null(ps)) {
    ps <- as.numeric(ps)
    if (length(ps) != N) stop("ps must have the same length as y.", call. = FALSE)
  }

  if (missing(components) || is.null(components)) {
    stop("components is required (single truncation parameter for both backends).", call. = FALSE)
  }
  components <- as.integer(components)
  if (!is.finite(components) || components < 2L) {
    stop("components must be an integer >= 2.", call. = FALSE)
  }

  # Kernel registry lookup + validation
  krn <- get_kernel_registry()
  kernel <- match.arg(kernel, choices = names(krn))
  kinfo <- krn[[kernel]]

  if (!is_allowed_kernel(kernel)) stop(sprintf("Kernel '%s' is not supported.", kernel), call. = FALSE)
  if (is.null(kinfo$bulk_params) || !length(kinfo$bulk_params)) {
    stop("Kernel registry entry is missing bulk_params.", call. = FALSE)
  }
  if (is.null(kinfo$param_types) || !length(kinfo$param_types)) {
    stop("Kernel registry entry is missing param_types.", call. = FALSE)
  }

  # signatures are required for likelihood emission
  check_gpd_contract(GPD, kernel)
  if (is.null(kinfo$signatures) || is.null(kinfo$signatures[[backend]])) {
    stop("Kernel registry entry is missing signatures for this backend.", call. = FALSE)
  }
  if (GPD) {
    if (is.null(kinfo$signatures[[backend]]$gpd)) {
      stop("Requested GPD=TRUE but kernel registry has no GPD signature for this backend.", call. = FALSE)
    }
  } else {
    if (is.null(kinfo$signatures[[backend]]$bulk)) {
      stop("Requested GPD=FALSE but kernel registry has no bulk signature for this backend.", call. = FALSE)
    }
  }

  # ---- helpers for default priors by parameter type ----
  default_prior_by_type <- function(type, support = NULL) {
    type <- as.character(type)
    support <- as.character(support %||% "")
    if (support %in% c("positive_location", "positive_scale", "positive_shape", "positive_sd")) {
      return(list(dist = "gamma", args = list(shape = 2, rate = 1)))
    }
    if (type %in% c("location")) {
      list(dist = "normal", args = list(mean = 0, sd = 5))
    } else if (type %in% c("scale")) {
      list(dist = "gamma", args = list(shape = 2, rate = 1))
    } else if (type %in% c("shape", "sd")) {
      list(dist = "invgamma", args = list(shape = 2, scale = 1))
    } else {
      # fallback: weak normal
      list(dist = "normal", args = list(mean = 0, sd = 5))
    }
  }

  default_beta_prior <- function(kind = c("bulk", "threshold", "tail_scale", "tail_shape")) {
    kind <- match.arg(kind)
    if (kind == "threshold") {
      list(dist = "normal", args = list(mean = 0, sd = 0.2))
    } else if (kind == "tail_scale") {
      list(dist = "normal", args = list(mean = 0, sd = 0.5))
    } else if (kind == "tail_shape") {
      list(dist = "normal", args = list(mean = 0, sd = 0.3))
    } else {
      list(dist = "normal", args = list(mean = 0, sd = 2))
    }
  }

  validate_link_spec <- function(param_name,
                                 link,
                                 link_power = NULL,
                                 support = NULL,
                                 beta_prior = NULL,
                                 link_dist = NULL) {
    link <- as.character(link %||% "identity")
    support <- as.character(support %||% "")

    if (!link %in% c("identity", "exp", "log", "softplus", "power")) {
      stop(sprintf("Unsupported link '%s' for '%s'.", link, param_name), call. = FALSE)
    }

    bp <- beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 1))
    bp_dist <- as.character(bp$dist %||% "normal")
    if (!identical(bp_dist, "normal")) {
      stop(
        sprintf(
          "link-mode coefficient prior for '%s' must use dist = 'normal'; got '%s'.",
          param_name,
          bp_dist
        ),
        call. = FALSE
      )
    }

    bp_mean <- bp$args$mean %||% 0
    bp_sd <- bp$args$sd %||% NA_real_
    if (!is.numeric(bp_mean) || length(bp_mean) != 1L || !is.finite(bp_mean)) {
      stop(sprintf("link-mode coefficient prior mean for '%s' must be a finite scalar.", param_name), call. = FALSE)
    }
    if (!is.numeric(bp_sd) || length(bp_sd) != 1L || !is.finite(bp_sd) || bp_sd <= 0) {
      stop(sprintf("link-mode coefficient prior sd for '%s' must be a positive finite scalar.", param_name), call. = FALSE)
    }

    if (identical(link, "log")) {
      stop(
        sprintf(
          "link '%s' is not supported for '%s' with unrestricted normal coefficient priors; use identity, exp, softplus, or a supported power link instead.",
          link,
          param_name
        ),
        call. = FALSE
      )
    }

    if (identical(link, "power")) {
      pw <- suppressWarnings(as.numeric(link_power))
      if (length(pw) != 1L || !is.finite(pw)) {
        stop(sprintf("power link for '%s' requires a finite numeric link_power.", param_name), call. = FALSE)
      }
      if (abs(pw - round(pw)) > sqrt(.Machine$double.eps) || pw <= 0) {
        stop(
          sprintf(
            "power link for '%s' requires a positive integer link_power when coefficient priors are normal.",
            param_name
          ),
          call. = FALSE
        )
      }
      if (support %in% c("positive_location", "positive_scale", "positive_shape", "positive_sd") &&
          as.integer(round(pw)) %% 2L != 0L) {
        stop(
          sprintf(
            "positive-support parameter '%s' requires an even integer power link, or use exp/softplus.",
            param_name
          ),
          call. = FALSE
        )
      }
    }

    if (support %in% c("positive_location", "positive_scale", "positive_shape", "positive_sd") &&
        !link %in% c("exp", "softplus", "power")) {
      stop(
        sprintf(
          "link '%s' is not appropriate for positive-support parameter '%s'; use exp, softplus, or an even-integer power link.",
          link,
          param_name
        ),
        call. = FALSE
      )
    }

    if (!is.null(link_dist)) {
      ld_dist <- as.character(link_dist$dist %||% "")
      if (!nzchar(ld_dist) || !identical(ld_dist, "lognormal")) {
        stop(sprintf("Only lognormal link_dist is supported for '%s'.", param_name), call. = FALSE)
      }
    }

    invisible(TRUE)
  }

  # ---- normalize/merge user param specs ----
  param_specs <- param_specs %||% list()
  user_bulk <- param_specs$bulk %||% list()
  user_gpd  <- param_specs$gpd  %||% list()
  user_conc <- param_specs$concentration %||% list()
  user_ps   <- param_specs$ps %||% list()

  # ---- concentration (alpha) plan ----
  conc_plan <- NULL
  if (length(user_conc)) {
    # user override must supply mode
    mode <- user_conc$mode %||% NA_character_
    if (!is.character(mode) || length(mode) != 1L) stop("concentration$mode must be a single character.", call. = FALSE)
    if (!mode %in% c("fixed", "dist")) stop("concentration$mode must be 'fixed' or 'dist'.", call. = FALSE)

    if (mode == "fixed") {
      if (is.null(user_conc$value) || !is.numeric(user_conc$value) || length(user_conc$value) != 1L) {
        stop("concentration fixed mode requires concentration$value (single numeric).", call. = FALSE)
      }
      conc_plan <- list(mode = "fixed", value = as.numeric(user_conc$value))
    } else {
      # dist
      dist <- user_conc$dist %||% "gamma"
      args <- user_conc$args %||% list(shape = 1, rate = 1)
      conc_plan <- list(mode = "dist", dist = dist, args = args)
    }
  } else {
    # default behavior
    if (isTRUE(alpha_random)) {
      conc_plan <- list(mode = "dist", dist = "gamma", args = list(shape = 1, rate = 1))
    } else {
      # fixed alpha default (can be overridden later by user_conc)
      conc_plan <- list(mode = "fixed", value = 1)
    }
  }

  # ---- bulk parameter plans ----
  bulk_params <- kinfo$bulk_params
  fixed_defaults <- kinfo$fixed_defaults %||% list()
  defaults_X <- kinfo$defaults_X %||% list()

  bulk_plan <- vector("list", length(bulk_params))
  names(bulk_plan) <- bulk_params

  for (nm in bulk_params) {
    # fixed-by-default (e.g. amoroso shape1=1)
    if (!is.null(fixed_defaults[[nm]])) {
      bulk_plan[[nm]] <- list(mode = "fixed", value = fixed_defaults[[nm]])
      next
    }

    # user override?
    u <- user_bulk[[nm]]
    if (!is.null(u)) {
      mode <- u$mode %||% NA_character_
      if (!is.character(mode) || length(mode) != 1L) stop(sprintf("bulk[%s] mode must be a single character.", nm), call. = FALSE)
      if (!mode %in% c("fixed", "dist", "link")) stop(sprintf("bulk[%s] mode must be fixed/dist/link.", nm), call. = FALSE)

      if (mode == "fixed") {
        if (is.null(u$value)) stop(sprintf("bulk[%s] fixed mode requires value.", nm), call. = FALSE)
        bulk_plan[[nm]] <- list(mode = "fixed", value = u$value)
      } else if (mode == "dist") {
        dist <- u$dist %||% default_prior_by_type(kinfo$param_types[[nm]], kinfo$bulk_support[[nm]])$dist
        args <- u$args %||% default_prior_by_type(kinfo$param_types[[nm]], kinfo$bulk_support[[nm]])$args
        bulk_plan[[nm]] <- list(mode = "dist", dist = dist, args = args)
      } else {
        # link
        if (!has_X) stop(sprintf("bulk[%s] requested link mode but X is NULL.", nm), call. = FALSE)
        link <- u$link %||% "identity"
        link_power <- u$link_power %||% NULL
        beta_prior <- u$beta_prior %||% default_beta_prior("bulk")
        validate_link_spec(
          param_name = nm,
          link = link,
          link_power = link_power,
          support = kinfo$bulk_support[[nm]],
          beta_prior = beta_prior
        )
        bulk_plan[[nm]] <- list(mode = "link", link = link, link_power = link_power, beta_prior = beta_prior)
      }
      next
    }

    # defaults
    if (has_X) {
      # if defaults_X provides link behavior for this param, use it; otherwise fall back to dist
      dx <- defaults_X[[nm]]
      if (!is.null(dx) && is.list(dx) && identical(dx$mode, "link")) {
        validate_link_spec(
          param_name = nm,
          link = dx$link %||% "identity",
          link_power = dx$link_power %||% NULL,
          support = kinfo$bulk_support[[nm]],
          beta_prior = dx$beta_prior %||% default_beta_prior("bulk")
        )
        bulk_plan[[nm]] <- list(
          mode = "link",
          link = dx$link %||% "identity",
          link_power = dx$link_power %||% NULL,
          beta_prior = dx$beta_prior %||% default_beta_prior("bulk")
        )
      } else {
        pr <- default_prior_by_type(kinfo$param_types[[nm]], kinfo$bulk_support[[nm]])
        bulk_plan[[nm]] <- list(mode = "dist", dist = pr$dist, args = pr$args)
      }
    } else {
      pr <- default_prior_by_type(kinfo$param_types[[nm]], kinfo$bulk_support[[nm]])
      bulk_plan[[nm]] <- list(mode = "dist", dist = pr$dist, args = pr$args)
    }
  }

  # ---- GPD plan ----
  gpd_plan <- list()
  if (isTRUE(GPD)) {
    # threshold
    thr_u <- user_gpd$threshold
    if (!is.null(thr_u)) {
      # allow fixed/dist/link; link may optionally include link_dist
      mode <- thr_u$mode %||% NA_character_
      if (!mode %in% c("fixed", "dist", "link")) stop("gpd$threshold mode must be fixed/dist/link.", call. = FALSE)

      if (mode == "fixed") {
        gpd_plan$threshold <- list(mode = "fixed", value = thr_u$value)
      } else if (mode == "dist") {
        gpd_plan$threshold <- list(mode = "dist", dist = thr_u$dist, args = thr_u$args)
      } else {
        if (!has_X) stop("gpd$threshold link mode requires X.", call. = FALSE)
        thr_link <- thr_u[["link", exact = TRUE]] %||% "identity"
        thr_link_power <- thr_u[["link_power", exact = TRUE]] %||% NULL
        thr_beta_prior <- thr_u[["beta_prior", exact = TRUE]] %||% default_beta_prior("threshold")
        thr_link_dist <- thr_u[["link_dist", exact = TRUE]] %||% NULL
        validate_link_spec(
          param_name = "threshold",
          link = thr_link,
          link_power = thr_link_power,
          support = "real",
          beta_prior = thr_beta_prior,
          link_dist = thr_link_dist
        )
        # if user supplies link_dist, keep it; otherwise allow default below
        gpd_plan$threshold <- list(
          mode = "link",
          link = thr_link,
          link_power = thr_link_power,
          beta_prior = thr_beta_prior,
          link_dist = thr_link_dist
        )
      }
    } else {
      # default threshold behavior:
      # * all backends with X: threshold follows a link model
      # * when link_dist is present, threshold is lognormal around the link mean
      # * no X: scalar dist-mode threshold
      if (has_X) {
        gpd_plan$threshold <- list(
          mode = "link",
          link = "identity",
          beta_prior = default_beta_prior("threshold")
        )
        gpd_plan$threshold$link_dist <- list(
          dist = "lognormal",
          mean_arg = "meanlog",
          sd_name = "sdlog_u"
        )
        gpd_plan$sdlog_u <- list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
      } else {
        gpd_plan$threshold <- list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
      }
    }
    if (!is.null(gpd_plan$threshold$link_dist) &&
        identical(gpd_plan$threshold$link_dist$dist, "lognormal")) {
      sdlog_u_u <- user_gpd[["sdlog_u", exact = TRUE]] %||% NULL
      if (is.null(sdlog_u_u)) {
        gpd_plan$sdlog_u <- gpd_plan$sdlog_u %||%
          list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
      } else {
        sdlog_mode <- sdlog_u_u$mode %||% "dist"
        if (!identical(sdlog_mode, "dist")) {
          stop("gpd$sdlog_u must be dist-mode when using threshold link_dist.", call. = FALSE)
        }
        gpd_plan$sdlog_u <- list(
          mode = "dist",
          dist = sdlog_u_u$dist %||% "invgamma",
          args = sdlog_u_u$args %||% list(shape = 2, scale = 1)
        )
      }
    }

    # tail_scale
    ts_u <- user_gpd$tail_scale
    if (!is.null(ts_u)) {
      mode <- ts_u$mode %||% NA_character_
      if (!mode %in% c("fixed", "dist", "link")) stop("gpd$tail_scale mode must be fixed/dist/link.", call. = FALSE)
      if (mode == "link" && !has_X) stop("gpd$tail_scale link mode requires X.", call. = FALSE)
      if (mode == "link") {
        validate_link_spec(
          param_name = "tail_scale",
          link = ts_u$link %||% "exp",
          link_power = ts_u$link_power %||% NULL,
          support = "positive_scale",
          beta_prior = ts_u$beta_prior %||% default_beta_prior("tail_scale")
        )
        gpd_plan$tail_scale <- list(
          mode = "link",
          link = ts_u$link %||% "exp",
          link_power = ts_u$link_power %||% NULL,
          beta_prior = ts_u$beta_prior %||% default_beta_prior("tail_scale")
        )
      } else if (mode == "dist") {
        gpd_plan$tail_scale <- list(mode = "dist", dist = ts_u$dist, args = ts_u$args)
      } else {
        gpd_plan$tail_scale <- list(mode = "fixed", value = ts_u$value)
      }
    } else {
      if (has_X) {
        gpd_plan$tail_scale <- list(mode = "link", link = "exp", beta_prior = default_beta_prior("tail_scale"))
      } else {
        gpd_plan$tail_scale <- list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
      }
    }

    # tail_shape
    tsh_u <- user_gpd$tail_shape
    if (!is.null(tsh_u)) {
      mode <- tsh_u$mode %||% NA_character_
      if (!mode %in% c("fixed", "dist", "link")) stop("gpd$tail_shape mode must be fixed/dist/link.", call. = FALSE)
      if (mode == "link" && !has_X) stop("gpd$tail_shape link mode requires X.", call. = FALSE)
      if (mode == "link") {
        validate_link_spec(
          param_name = "tail_shape",
          link = tsh_u$link %||% "identity",
          link_power = tsh_u$link_power %||% NULL,
          support = "real",
          beta_prior = tsh_u$beta_prior %||% default_beta_prior("tail_shape")
        )
        gpd_plan$tail_shape <- list(
          mode = "link",
          link = tsh_u$link %||% "identity",
          link_power = tsh_u$link_power %||% NULL,
          beta_prior = tsh_u$beta_prior %||% default_beta_prior("tail_shape")
        )
      } else if (mode == "fixed") {
        gpd_plan$tail_shape <- list(mode = "fixed", value = tsh_u$value)
      } else {
        gpd_plan$tail_shape <- list(mode = "dist", dist = tsh_u$dist, args = tsh_u$args)
      }
    } else {
      gpd_plan$tail_shape <- list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 0.2))
    }

    # ---- Spliced backend validation ----
    # Spliced backend enforces component-level GPD parameterization with flexible modes
    if (backend == "spliced") {
      # Ensure all GPD params support component-level specification (automatically satisfied by structure)
      # Link mode requires X - already validated above for each parameter
      # Store level metadata for consistency
      for (param_name in c("threshold", "tail_scale", "tail_shape")) {
        if (!is.null(gpd_plan[[param_name]])) {
          gpd_plan[[param_name]]$level <- "component"
        }
      }
    }
  }

  # ---- finalize plan ----
  ps_plan <- NULL
  if (!is.null(ps)) {
    ps_plan <- list(prior = user_ps$prior %||% list(dist = "normal", args = list(mean = 0, sd = 2)))
  }

  plan <- list(
    backend = backend,
    kernel = kernel,
    GPD = isTRUE(GPD),
    has_X = has_X,
    N = N,
    P = as.integer(P),
    components = components,
    concentration = conc_plan,
    bulk = bulk_plan,
    gpd = gpd_plan,
    ps = ps_plan
  )

  spec <- list(
    meta = list(
      backend = backend,
      kernel = kernel,
      GPD = isTRUE(GPD),
      has_X = has_X,
      has_ps = !is.null(ps),
      N = N,
      P = as.integer(P),
      components = components,
      custom_build = !is.null(param_specs) && length(param_specs) > 0
    ),
    kernel_info = kinfo,
    signatures = kinfo$signatures[[backend]],
    plan = plan
  )

  class(spec) <- c("causalmixgpd_spec", "list")
  spec
}

.cluster_parse_formula <- function(formula, data, na.action = stats::na.omit) {
  if (missing(formula) || is.null(formula)) {
    stop("'formula' is required.", call. = FALSE)
  }
  if (missing(data) || is.null(data)) {
    stop("'data' is required with 'formula'.", call. = FALSE)
  }

  mf <- stats::model.frame(formula, data = data, na.action = na.action)
  y <- stats::model.response(mf)
  if (is.null(y)) stop("Could not extract response from formula.", call. = FALSE)
  y <- as.numeric(y)

  trm <- stats::terms(mf)
  term_labels <- attr(trm, "term.labels") %||% character(0)
  X <- NULL
  ctr <- NULL
  X_cols <- character(0)

  if (length(term_labels) > 0L) {
    design <- .formula_design_matrix(
      trm = trm,
      mf = mf,
      drop_intercept = TRUE,
      drop_terms = "id",
      coerce_double = TRUE
    )
    X <- design$X

    if (is.null(X)) {
      X <- NULL
    } else {
      X_cols <- design$X_cols
      ctr <- design$contrasts
    }
  }

  resp_name <- tryCatch(as.character(formula[[2L]]), error = function(e) "y")

  list(
    y = y,
    X = X,
    terms = trm,
    xlevels = stats::.getXlevels(trm, mf),
    contrasts = ctr,
    X_cols = X_cols,
    response = resp_name,
    has_X = !is.null(X)
  )
}

.cluster_default_link <- function(kernel_info, param_name) {
  support <- as.character((kernel_info$bulk_support %||% list())[[param_name]] %||% "")
  if (support %in% c("positive_location", "positive_scale", "positive_shape", "positive_sd")) {
    return("exp")
  }
  "identity"
}

.cluster_link_bulk_specs <- function(kernel, beta_sd = 2) {
  kinfo <- get_kernel_registry()[[kernel]]
  if (is.null(kinfo)) stop(sprintf("Kernel '%s' not found.", kernel), call. = FALSE)

  defaults_X <- kinfo$defaults_X %||% list()
  out <- list()
  bulk_params <- kinfo$bulk_params %||% character(0)
  for (nm in bulk_params) {
    dx <- defaults_X[[nm]] %||% list()
    if (!identical(dx$mode %||% NA_character_, "link")) next
    out[[nm]] <- list(
      mode = "link",
      link = dx$link %||% .cluster_default_link(kinfo, nm),
      beta_prior = list(dist = "normal", args = list(mean = 0, sd = beta_sd))
    )
  }
  out
}

.cluster_validate_type <- function(type, default) {
  choices <- c("weights", "param", "both")
  default <- match.arg(default, choices = choices)
  if (missing(type) || is.null(type) || length(type) < 1L) type <- default
  match.arg(type, choices = choices)
}

.cluster_validate_type_requirements <- function(type, has_X, components_missing) {
  if (type %in% c("weights", "both") && !isTRUE(has_X)) {
    stop(sprintf("type='%s' requires covariates in the formula.", type), call. = FALSE)
  }
  if (type %in% c("weights", "both") && isTRUE(components_missing)) {
    stop(sprintf("type='%s' requires an explicit 'components' value.", type), call. = FALSE)
  }
}

.cluster_default_beta_prior <- function(param) {
  if (identical(param, "threshold")) return(list(dist = "normal", args = list(mean = 0, sd = 0.2)))
  if (identical(param, "tail_scale")) return(list(dist = "normal", args = list(mean = 0, sd = 0.5)))
  if (identical(param, "tail_shape")) return(list(dist = "normal", args = list(mean = 0, sd = 0.3)))
  list(dist = "normal", args = list(mean = 0, sd = 2))
}

.cluster_split_overrides <- function(x, bulk_names, gpd_names) {
  x <- x %||% list()
  if (!is.list(x)) stop("Overrides must be supplied as a list.", call. = FALSE)

  out <- list(bulk = list(), gpd = list(), concentration = NULL)

  nms <- names(x) %||% character(0)
  if (!length(nms) && length(x)) {
    stop("Unnamed override entries are not supported.", call. = FALSE)
  }
  if (!length(nms)) return(out)

  bad <- setdiff(nms, c("bulk", "gpd", "concentration", "alpha", bulk_names, gpd_names))
  if (length(bad)) {
    stop(sprintf("Unknown override names: %s", paste(bad, collapse = ", ")), call. = FALSE)
  }

  if ("bulk" %in% nms) {
    out$bulk <- x$bulk %||% list()
  }
  if ("gpd" %in% nms) {
    out$gpd <- x$gpd %||% list()
  }

  for (nm in intersect(nms, bulk_names)) {
    out$bulk[[nm]] <- x[[nm]]
  }
  for (nm in intersect(nms, gpd_names)) {
    out$gpd[[nm]] <- x[[nm]]
  }

  out$concentration <- x$concentration %||% x$alpha %||% NULL
  out
}

.cluster_normalize_link_entry <- function(entry) {
  if (is.character(entry) && length(entry) == 1L) {
    return(list(mode = "link", link = entry))
  }
  if (!is.list(entry)) {
    stop("Link overrides must be strings or lists.", call. = FALSE)
  }
  mode <- entry$mode %||% "link"
  if (!identical(mode, "link")) {
    stop("Link overrides must have mode='link'.", call. = FALSE)
  }
  out <- utils::modifyList(list(mode = "link"), entry)
  out
}

.cluster_apply_link_overrides <- function(spec, link, has_X) {
  if (is.null(link)) return(spec)
  if (!isTRUE(has_X)) {
    stop("`link` overrides require covariates in the formula.", call. = FALSE)
  }

  bulk_names <- names(spec$plan$bulk %||% list())
  gpd_names <- names(spec$plan$gpd %||% list())
  parsed <- .cluster_split_overrides(link, bulk_names = bulk_names, gpd_names = gpd_names)

  for (nm in names(parsed$bulk)) {
    ent <- .cluster_normalize_link_entry(parsed$bulk[[nm]])
    cur <- spec$plan$bulk[[nm]] %||% list()
    ent$beta_prior <- ent$beta_prior %||% cur$beta_prior %||% .cluster_default_beta_prior(nm)
    spec$plan$bulk[[nm]] <- utils::modifyList(cur, ent)
  }

  for (nm in names(parsed$gpd)) {
    ent <- .cluster_normalize_link_entry(parsed$gpd[[nm]])
    cur <- spec$plan$gpd[[nm]] %||% list()
    ent$beta_prior <- ent$beta_prior %||% cur$beta_prior %||% .cluster_default_beta_prior(nm)
    spec$plan$gpd[[nm]] <- utils::modifyList(cur, ent)
  }

  spec
}

.cluster_normalize_prior_entry <- function(entry, current_mode, param_name = NULL) {
  is_link_mode <- identical(current_mode, "link")

  if (is.numeric(entry) && length(entry) == 1L) {
    return(list(mode = "fixed", value = as.numeric(entry)))
  }
  if (is.character(entry) && length(entry) == 1L) {
    if (is_link_mode) {
      return(list(mode = "link", beta_prior = list(dist = entry, args = list())))
    }
    return(list(mode = "dist", dist = entry, args = list()))
  }
  if (!is.list(entry)) {
    stop("Prior overrides must be numeric, character, or list entries.", call. = FALSE)
  }

  if (!is.null(entry$mode)) {
    return(entry)
  }
  if (!is.null(entry$value)) {
    return(list(mode = "fixed", value = entry$value))
  }
  if (!is.null(entry$beta_prior)) {
    return(list(mode = "link", beta_prior = entry$beta_prior))
  }
  if (!is.null(entry$dist) || !is.null(entry$args)) {
    if (is_link_mode) {
      return(list(
        mode = "link",
        beta_prior = list(
          dist = entry$dist %||% "normal",
          args = entry$args %||% .cluster_default_beta_prior(param_name)$args
        )
      ))
    }
    return(list(
      mode = "dist",
      dist = entry$dist %||% "normal",
      args = entry$args %||% list(mean = 0, sd = 2)
    ))
  }

  stop("Could not interpret prior override entry.", call. = FALSE)
}

.cluster_apply_prior_overrides <- function(spec, priors) {
  if (is.null(priors)) return(spec)

  bulk_names <- names(spec$plan$bulk %||% list())
  gpd_names <- names(spec$plan$gpd %||% list())
  parsed <- .cluster_split_overrides(priors, bulk_names = bulk_names, gpd_names = gpd_names)

  for (nm in names(parsed$bulk)) {
    cur <- spec$plan$bulk[[nm]] %||% list()
    ent <- .cluster_normalize_prior_entry(parsed$bulk[[nm]], current_mode = cur$mode %||% NULL, param_name = nm)
    spec$plan$bulk[[nm]] <- utils::modifyList(cur, ent)
  }

  for (nm in names(parsed$gpd)) {
    cur <- spec$plan$gpd[[nm]] %||% list()
    ent <- .cluster_normalize_prior_entry(parsed$gpd[[nm]], current_mode = cur$mode %||% NULL, param_name = nm)
    spec$plan$gpd[[nm]] <- utils::modifyList(cur, ent)
  }

  if (!is.null(parsed$concentration)) {
    cc <- parsed$concentration
    if (is.numeric(cc) && length(cc) == 1L) {
      spec$plan$concentration <- list(mode = "fixed", value = as.numeric(cc))
    } else if (is.list(cc) && !is.null(cc$mode)) {
      spec$plan$concentration <- cc
    } else if (is.list(cc)) {
      if (!is.null(cc$value)) {
        spec$plan$concentration <- list(mode = "fixed", value = cc$value)
      } else {
        spec$plan$concentration <- list(
          mode = "dist",
          dist = cc$dist %||% "gamma",
          args = cc$args %||% list(shape = 1, rate = 1)
        )
      }
    } else if (is.character(cc) && length(cc) == 1L) {
      spec$plan$concentration <- list(mode = "dist", dist = cc, args = list())
    } else {
      stop("Invalid concentration prior override.", call. = FALSE)
    }
  }

  spec
}

codegen_cluster_model <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta))
  spec$cluster <- spec$cluster %||% list()
  type <- spec$cluster$type %||% "weights"
  spec$cluster$gating <- isTRUE(spec$cluster$gating %||% (type %in% c("weights", "both")))
  spec$cluster$param_link <- isTRUE(spec$cluster$param_link %||% (type %in% c("param", "both")))
  backend_target <- if (identical(type, "param")) "crp" else "sb"
  spec$meta$backend <- backend_target
  spec$plan$backend <- backend_target
  build_code_from_spec(spec)
}

build_cluster_bundle <- function(formula,
                                 data,
                                 kernel,
                                 GPD,
                                 type = c("weights", "param", "both"),
                                 default = "weights",
                                 link = NULL,
                                 priors = NULL,
                                 components = 10L,
                                 mcmc = list(niter = 2000, nburnin = 500, thin = 1, nchains = 1, seed = 1),
                                 param_specs = NULL,
                                 epsilon = 0.025,
                                 alpha_random = TRUE,
                                 monitor = c("core", "full"),
                                 monitor_v = FALSE,
                                 monitor_latent = TRUE,
                                 ...) {
  type <- .cluster_validate_type(type, default)
  parsed <- .cluster_parse_formula(formula = formula, data = data)
  components_missing <- missing(components) || is.null(components)
  .cluster_validate_type_requirements(
    type = type,
    has_X = parsed$has_X,
    components_missing = components_missing
  )

  if (components_missing) {
    components <- 10L
  }
  components <- as.integer(components)
  if (!is.finite(components) || components < 2L) {
    stop("'components' must be an integer >= 2.", call. = FALSE)
  }

  backend <- if (identical(type, "param")) "crp" else "sb"
  monitor <- match.arg(monitor)
  if (identical(monitor, "full")) {
    monitor_latent <- TRUE
    if (identical(backend, "sb")) monitor_v <- TRUE
  }
  monitor_latent <- TRUE

  user_param_specs <- param_specs %||% list()
  cluster_gating <- type %in% c("weights", "both")
  cluster_param_link <- (type %in% c("param", "both")) && isTRUE(parsed$has_X)
  if (cluster_param_link) {
    linked_bulk <- .cluster_link_bulk_specs(kernel = kernel)
    user_param_specs$bulk <- linked_bulk
  }

  spec <- compile_model_spec(
    y = parsed$y,
    X = parsed$X,
    backend = backend,
    kernel = kernel,
    GPD = isTRUE(GPD),
    components = components,
    param_specs = user_param_specs,
    alpha_random = alpha_random
  )
  spec <- .cluster_apply_link_overrides(spec, link = link, has_X = parsed$has_X)
  spec <- .cluster_apply_prior_overrides(spec, priors = priors)

  spec$cluster <- list(
    type = type,
    default = default,
    gating = cluster_gating,
    param_link = cluster_param_link,
    formula = formula,
    formula_meta = list(
      terms = parsed$terms,
      xlevels = parsed$xlevels,
      contrasts = parsed$contrasts,
      X_cols = parsed$X_cols,
      response = parsed$response
    )
  )

  code <- .wrap_nimble_code(codegen_cluster_model(spec))
  constants <- build_constants_from_spec(spec)
  dimensions <- build_dimensions_from_spec(spec)
  data_list <- build_data_from_inputs(y = parsed$y, X = parsed$X)
  inits <- build_inits_from_spec(spec, y = parsed$y, X = parsed$X)
  monitors <- build_monitors_from_spec(
    spec,
    monitor_v = isTRUE(monitor_v),
    monitor_latent = TRUE
  )

  out <- list(
    spec = spec,
    code = code,
    constants = constants,
    dimensions = dimensions,
    data = data_list,
    inits = inits,
    monitors = monitors,
    monitor_policy = list(
      monitor = monitor,
      monitor_latent = TRUE,
      monitor_v = isTRUE(monitor_v)
    ),
    mcmc = mcmc,
    epsilon = as.numeric(epsilon)[1],
    cluster = list(type = type),
    call = match.call()
  )
  class(out) <- c("dpmixgpd_cluster_bundle", "causalmixgpd_bundle", "list")
  out
}

#' Fit a clustering-only bulk model
#'
#' Build and fit a Dirichlet-process mixture for clustering without causal estimands or posterior
#' prediction for a response surface. This interface focuses on latent partition recovery from a
#' formula specification and returns a cluster-fit object that can be summarized, plotted, or
#' converted into labels and posterior similarity matrices with [predict.dpmixgpd_cluster_fit()].
#'
#' @param formula Model formula. The response must be present in `data`.
#' @param data Data frame containing the response and optional predictors.
#' @param type Clustering mode:
#'   \itemize{
#'     \item \code{"weights"}: links mixture weights to predictors
#'     \item \code{"param"}: links kernel parameters to predictors
#'     \item \code{"both"}: links both weights and kernel parameters to predictors
#'   }
#' @param default Default mode used when `type` is omitted.
#' @param mcmc MCMC control list passed into the cluster bundle.
#' @param ... Additional arguments passed to `build_cluster_bundle()`, including kernel settings,
#'   prior overrides, component counts, and monitoring controls.
#'
#' @return Object of class `dpmixgpd_cluster_fit`.
#'
#' @details
#' The fitted model targets a latent partition \eqn{z_1, \dots, z_n} with component-specific kernel
#' parameters. Depending on `type`, predictors can enter through the gating probabilities
#' \deqn{
#' \Pr(z_i = k \mid x_i) = \pi_k(x_i)
#' }
#' or through linked kernel parameters for each component. The returned fit stores posterior draws
#' of the latent cluster labels and associated parameters; the representative clustering is extracted
#' later by [predict.dpmixgpd_cluster_fit()] using Dahl's least-squares rule.
#'
#' Use `type = "weights"` or `type = "both"` only when the formula includes predictors and when an
#' explicit number of `components` is supplied. Otherwise the builder stops before fitting.
#'
#' @seealso [dpmgpd.cluster()], [predict.dpmixgpd_cluster_fit()],
#'   [summary.dpmixgpd_cluster_fit()], [plot.dpmixgpd_cluster_fit()],
#'   [build_nimble_bundle()], [dpmix()].
#' @family cluster workflow
#' @examples
#' \donttest{
#' data("nc_realX100_p3_k2", package = "CausalMixGPD")
#' dat <- data.frame(y = nc_realX100_p3_k2$y[1:20],
#'                   nc_realX100_p3_k2$X[1:20, , drop = FALSE])
#' fit <- dpmix.cluster(
#'   y ~ x1 + x2 + x3,
#'   data = dat,
#'   kernel = "normal",
#'   type = "param",
#'   components = 3,
#'   mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1)
#' )
#' summary(fit)
#' }
#' @export
dpmix.cluster <- function(formula,
                          data,
                          type = c("weights", "param", "both"),
                          default = "weights",
                          mcmc = list(),
                          ...) {
  type <- .cluster_validate_type(type, default)
  bundle <- do.call(
    build_cluster_bundle,
    c(
      list(
        formula = formula,
        data = data,
        GPD = FALSE,
        type = type,
        default = default,
        mcmc = mcmc
      ),
      list(...)
    )
  )
  fit <- run_cluster_mcmc(bundle)
  fit$call <- match.call()
  fit
}

#' Fit a clustering-only bulk-tail model
#'
#' Variant of [dpmix.cluster()] that augments the cluster kernel with a generalized Pareto tail.
#' This is the clustering analogue of the spliced bulk-tail workflow used by [dpmgpd()].
#'
#' @inheritParams dpmix.cluster
#'
#' @return Object of class `dpmixgpd_cluster_fit`.
#'
#' @details
#' For observations above a component-specific threshold, the component density is spliced as
#' \deqn{
#' f(y) = (1 - F_{bulk}(u)) g_{GPD}(y \mid u, \sigma_u, \xi_u), \qquad y \ge u,
#' }
#' so cluster assignment can be informed by both central behavior and tail behavior.
#'
#' This interface is preferable when cluster separation is driven by upper-tail differences rather
#' than bulk-only shape or location differences.
#'
#' @seealso [dpmix.cluster()], [predict.dpmixgpd_cluster_fit()],
#'   [dpmgpd()], [sim_bulk_tail()].
#' @family cluster workflow
#' @examples
#' \donttest{
#' data("nc_posX100_p3_k2", package = "CausalMixGPD")
#' dat <- data.frame(y = nc_posX100_p3_k2$y[1:20],
#'                   nc_posX100_p3_k2$X[1:20, , drop = FALSE])
#' fit <- dpmgpd.cluster(
#'   y ~ x1 + x2 + x3,
#'   data = dat,
#'   kernel = "gamma",
#'   type = "param",
#'   components = 3,
#'   mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1)
#' )
#' cluster_profiles(fit)
#' }
#' @export
dpmgpd.cluster <- function(formula,
                           data,
                           type = c("weights", "param", "both"),
                           default = "weights",
                           mcmc = list(),
                           ...) {
  type <- .cluster_validate_type(type, default)
  bundle <- do.call(
    build_cluster_bundle,
    c(
      list(
        formula = formula,
        data = data,
        GPD = TRUE,
        type = type,
        default = default,
        mcmc = mcmc
      ),
      list(...)
    )
  )
  fit <- run_cluster_mcmc(bundle)
  fit$call <- match.call()
  fit
}

#' Extract Cluster Profiles
#'
#' Access descriptive cluster profiles without reaching into summary internals.
#'
#' @param object A cluster fit, cluster-label object, or corresponding summary object.
#' @param ... Additional arguments passed to summary methods for fitted or label objects.
#' @return A data frame of cluster-level descriptive summaries, or \code{NULL}
#'   when profiles are unavailable.
#' @family cluster workflow
#' @export
cluster_profiles <- function(object, ...) {
  UseMethod("cluster_profiles")
}

#' @export
cluster_profiles.dpmixgpd_cluster_fit <- function(object, ...) {
  cluster_profiles(summary(object, ...))
}

#' @export
cluster_profiles.dpmixgpd_cluster_labels <- function(object, ...) {
  cluster_profiles(summary(object, ...))
}

#' @export
cluster_profiles.summary.dpmixgpd_cluster_fit <- function(object, ...) {
  object$cluster_profiles %||% NULL
}

#' @export
cluster_profiles.summary.dpmixgpd_cluster_labels <- function(object, ...) {
  object$cluster_profiles %||% NULL
}






#' Print a cluster bundle
#'
#' Compact display for a `dpmixgpd_cluster_bundle` before MCMC is run.
#'
#' @details
#' A cluster bundle is the pre-sampling representation of the latent partition
#' model. It stores the formula-derived design, kernel choice, truncation level,
#' and the rule by which predictors enter the clustering model, but it does not
#' yet contain posterior draws of the latent labels \eqn{z_1, \dots, z_n}.
#'
#' `print()` is intentionally brief. It is meant to confirm that the bundle
#' matches the requested clustering structure before you run MCMC with
#' `run_cluster_mcmc()` or the higher-level wrappers [dpmix.cluster()] and
#' [dpmgpd.cluster()].
#'
#' @param x A cluster bundle.
#' @param ... Unused.
#'
#' @return `x`, invisibly.
#'
#' @seealso [dpmix.cluster()], [dpmgpd.cluster()], [summary.dpmixgpd_cluster_bundle()].
#' @family cluster workflow
#' @export
print.dpmixgpd_cluster_bundle <- function(x, ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_bundle"))
  meta <- x$spec$meta %||% list()
  cl <- x$spec$cluster %||% list()
  cat("Cluster bundle\n")
  cat("type      :", cl$type %||% "weights", "\n")
  cat("link mode :", if (isTRUE(cl$param_link)) "param|x" else "none", "\n")
  cat("kernel    :", meta$kernel %||% "?", "\n")
  cat("GPD       :", isTRUE(meta$GPD), "\n")
  cat("n         :", as.integer(meta$N %||% NA_integer_), "\n")
  cat("components:", as.integer(meta$components %||% NA_integer_), "\n")
  invisible(x)
}

#' Summarize a cluster bundle
#'
#' Report the modeling choices encoded in a cluster bundle before fitting.
#'
#' @details
#' This summary is a pre-flight check for the clustering workflow. It reports
#' the latent-partition design, the chosen kernel family, whether a GPD tail
#' will be spliced above the threshold, the effective sample and predictor
#' dimensions, and the monitor set that will be carried into MCMC.
#'
#' Because no posterior simulation has occurred yet, the summary describes only
#' the assumed model structure. Quantities such as representative labels,
#' pairwise co-clustering probabilities, and cluster-specific summaries become
#' available only after the fitted object has been created and post-processed.
#'
#' @param object A cluster bundle.
#' @param ... Unused.
#'
#' @return Summary list containing kernel choice, GPD flag, dimensions, component count, and
#'   monitor configuration.
#'
#' @seealso [print.dpmixgpd_cluster_bundle()], [plot.dpmixgpd_cluster_bundle()],
#'   [predict.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @export
summary.dpmixgpd_cluster_bundle <- function(object, ...) {
  stopifnot(inherits(object, "dpmixgpd_cluster_bundle"))
  meta <- object$spec$meta %||% list()
  cl <- object$spec$cluster %||% list()
  out <- list(
    type = cl$type %||% "weights",
    link_mode = if (isTRUE(cl$param_link)) "param|x" else "none",
    kernel = meta$kernel %||% NA_character_,
    GPD = isTRUE(meta$GPD),
    N = as.integer(meta$N %||% NA_integer_),
    P = as.integer(meta$P %||% 0L),
    components = as.integer(meta$components %||% NA_integer_),
    monitors = object$monitors %||% character(0)
  )
  class(out) <- c("summary.dpmixgpd_cluster_bundle", "list")
  out
}

#' Plot a cluster bundle
#'
#' Produce a compact graphical summary of the cluster bundle metadata.
#'
#' @details
#' The bundle plot is a metadata display rather than an inferential graphic. It
#' mirrors the structural fields reported by `print()` and `summary()` in a
#' single panel so the pre-MCMC clustering specification can be reviewed in a
#' figure-oriented workflow or notebook.
#'
#' Because the object has not been sampled yet, no representative partition or
#' posterior uncertainty is shown here. Use [plot.dpmixgpd_cluster_fit()],
#' [plot.dpmixgpd_cluster_labels()], or [plot.dpmixgpd_cluster_psm()] after
#' fitting when you need substantive clustering output.
#'
#' @param x A cluster bundle.
#' @param plotly Logical; if `TRUE`, convert the `ggplot2` output to a `plotly` /
#'   `htmlwidget` representation via `.wrap_plotly()`. Defaults to
#'   `getOption("CausalMixGPD.plotly", FALSE)`.
#' @param ... Unused.
#'
#' @return A `ggplot2` object or a `plotly`/`htmlwidget` object when `plotly = TRUE`.
#'
#' @seealso [summary.dpmixgpd_cluster_bundle()], [dpmix.cluster()], [dpmgpd.cluster()].
#' @family cluster workflow
#' @rdname plot.dpmixgpd_cluster_fit
#' @export
plot.dpmixgpd_cluster_bundle <- function(x,
                                         plotly = getOption("CausalMixGPD.plotly", FALSE),
                                         ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_bundle"))
  sm <- summary(x)
  txt <- c(
    sprintf("Type: %s", sm$type),
    sprintf("Link mode: %s", sm$link_mode),
    sprintf("Kernel: %s", sm$kernel),
    sprintf("GPD: %s", sm$GPD),
    sprintf("N: %d  P: %d", sm$N, sm$P),
    sprintf("Components: %d", sm$components)
  )
  .cluster_require_ggplot()
  p <- ggplot2::ggplot(
    data.frame(x = 0, y = 1, label = paste(txt, collapse = "\n"), stringsAsFactors = FALSE),
    ggplot2::aes(x = x, y = y, label = label)
  ) +
    ggplot2::geom_text(hjust = 0, vjust = 1, family = "mono") +
    ggplot2::coord_cartesian(xlim = c(0, 1), ylim = c(0, 1), clip = "off") +
    ggplot2::labs(title = "Cluster Bundle") +
    ggplot2::theme_void() +
    ggplot2::theme(
      plot.title = ggplot2::element_text(face = "bold", hjust = 0),
      plot.margin = ggplot2::margin(15, 15, 15, 15)
    )
  .cluster_maybe_wrap_plotly(p, plotly = plotly)
}

#' Print a cluster fit
#'
#' Compact display for a fitted clustering object.
#'
#' @details
#' A fitted cluster object contains posterior draws for the latent labels and
#' associated component parameters. The printed header identifies the model
#' family that generated those draws, including whether the fit used a bulk-only
#' kernel or a spliced bulk-tail specification.
#'
#' The printed `components` value is the truncation size used by the sampler. It
#' is not the same thing as the number of occupied clusters in the Dahl
#' representative partition, which is computed later by
#' [predict.dpmixgpd_cluster_fit()] and summarized by
#' [summary.dpmixgpd_cluster_fit()].
#'
#' @param x A cluster fit.
#' @param ... Unused.
#'
#' @return `x`, invisibly.
#'
#' @seealso [summary.dpmixgpd_cluster_fit()], [predict.dpmixgpd_cluster_fit()],
#'   [plot.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @export
print.dpmixgpd_cluster_fit <- function(x, ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_fit"))
  meta <- x$spec$meta %||% list()
  cl <- x$spec$cluster %||% list()
  cat("Cluster fit\n")
  cat("type      :", cl$type %||% "weights", "\n")
  cat("link mode :", if (isTRUE(cl$param_link)) "param|x" else "none", "\n")
  cat("kernel    :", meta$kernel %||% "?", "\n")
  cat("GPD       :", isTRUE(meta$GPD), "\n")
  cat("n         :", as.integer(meta$N %||% NA_integer_), "\n")
  cat("components:", as.integer(meta$components %||% NA_integer_), "\n")
  invisible(x)
}

#' Summarize a cluster fit
#'
#' Summarize the posterior clustering induced by the Dahl representative partition.
#'
#' @param object A cluster fit.
#' @param burnin Number of initial posterior draws to discard.
#' @param thin Keep every `thin`-th posterior draw.
#' @param top_n Number of populated clusters to profile when descriptive summaries are available.
#' @param order_by Ordering rule for descriptive cluster profiles:
#'   \itemize{
#'     \item \code{"size"}: decreasing cluster size
#'     \item \code{"label"}: ascending cluster label
#'   }
#' @param vars Optional character vector of numeric columns to summarize within each cluster.
#' @param ... Unused.
#'
#' @return Summary list with the number of retained clusters, cluster sizes, optional
#'   cluster-level descriptive summaries, and the burn-in/thinning settings used to construct the
#'   summary.
#'
#' @details
#' This summary is based on [predict.dpmixgpd_cluster_fit()] with `type = "label"`. The reported
#' cluster count \eqn{K^*} is the number of unique labels in the representative partition rather
#' than the number of components available in the truncated sampler.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [plot.dpmixgpd_cluster_fit()],
#'   [summary.dpmixgpd_cluster_labels()].
#' @family cluster workflow
#' @export
summary.dpmixgpd_cluster_fit <- function(object,
                                        burnin = NULL,
                                        thin = NULL,
                                        top_n = 5L,
                                        order_by = c("size", "label"),
                                        vars = NULL,
                                        ...) {
  stopifnot(inherits(object, "dpmixgpd_cluster_fit"))
  lbl <- predict(object, type = "label", burnin = burnin, thin = thin, return_scores = TRUE)
  tab <- .cluster_size_table(lbl$labels, order_by = "size")
  lbl_sum <- summary(lbl, top_n = top_n, order_by = order_by, vars = vars)
  cluster_profiles <- lbl_sum$cluster_profiles
  if (is.data.frame(cluster_profiles) && nrow(cluster_profiles)) {
    num_cols <- vapply(cluster_profiles, is.numeric, logical(1))
    cluster_profiles[num_cols] <- lapply(cluster_profiles[num_cols], round, digits = 3L)
  }
  out <- list(
    K_star = length(tab),
    cluster_sizes = tab,
    cluster_profiles = cluster_profiles,
    certainty = lbl_sum$certainty,
    source = lbl$source,
    burnin = lbl$burnin,
    thin = lbl$thin
  )
  class(out) <- c("summary.dpmixgpd_cluster_fit", "list")
  out
}

#' Plot a cluster fit
#'
#' Visualize either the posterior similarity matrix, the posterior number of occupied clusters, the
#' size distribution of the representative clusters, or cluster-specific response summaries.
#'
#' @details
#' This plot method exposes the main posterior diagnostics for clustering. The
#' `which = "k"` view tracks the number of occupied clusters across retained
#' draws, `which = "psm"` visualizes pairwise co-clustering probabilities,
#' `which = "sizes"` displays the size profile of the representative partition,
#' and `which = "summary"` shows response summaries conditional on the selected
#' representative labels.
#'
#' The representative partition is obtained from
#' [predict.dpmixgpd_cluster_fit()] using Dahl's least-squares rule. As a
#' result, the `sizes` and `summary` views describe that representative
#' clustering rather than the full posterior distribution over partitions.
#'
#' @param x A cluster fit.
#' @param which Plot type:
#'   \itemize{
#'     \item \code{"psm"}: posterior similarity matrix heatmap
#'     \item \code{"k"}: posterior number of occupied clusters
#'     \item \code{"sizes"}: bar chart of representative cluster sizes
#'     \item \code{"summary"}: cluster-specific response summaries
#'   }
#' @param burnin Number of initial posterior draws to discard.
#' @param thin Keep every `thin`-th posterior draw.
#' @param psm_max_n Maximum training sample size allowed for PSM plotting.
#' @param top_n Number of populated representative clusters to display for
#'   \code{which = "sizes"} or \code{which = "summary"}. Use \code{NULL} to
#'   display all populated clusters.
#' @param order_by Ordering rule for cluster displays:
#'   \itemize{
#'     \item \code{"size"}: decreasing cluster size
#'     \item \code{"label"}: ascending cluster label
#'   }
#' @param plotly Logical; if `TRUE`, convert the `ggplot2` output to a `plotly` /
#'   `htmlwidget` representation via `.wrap_plotly()`. Defaults to
#'   `getOption("CausalMixGPD.plotly", FALSE)`.
#' @param ... Unused.
#'
#' @return A `ggplot2` object or a `plotly`/`htmlwidget` object when `plotly = TRUE`.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [summary.dpmixgpd_cluster_fit()],
#'   [plot.dpmixgpd_cluster_psm()], [plot.dpmixgpd_cluster_labels()].
#' @family cluster workflow
#' @export
plot.dpmixgpd_cluster_fit <- function(x,
                                      which = c("psm", "k", "sizes", "summary"),
                                      burnin = NULL,
                                      thin = NULL,
                                      psm_max_n = 2000L,
                                      top_n = 5L,
                                      order_by = c("size", "label"),
                                      plotly = getOption("CausalMixGPD.plotly", FALSE),
                                      ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_fit"))
  which <- match.arg(which)
  .cluster_require_ggplot()

  if (identical(which, "psm")) {
    psm <- predict(x, type = "psm", burnin = burnin, thin = thin, psm_max_n = psm_max_n)
    return(plot(psm, psm_max_n = psm_max_n, plotly = plotly, ...))
  }

  if (identical(which, "k")) {
    z <- extract_z_draws(x$samples, burnin = burnin, thin = thin)
    k_draw <- apply(z, 1, function(v) length(unique(as.integer(v))))
    df <- data.frame(draw = seq_along(k_draw), K = as.numeric(k_draw))
    pal <- .plot_palette(2L)
    p <- ggplot2::ggplot(df, ggplot2::aes(x = draw, y = K)) +
      ggplot2::geom_line(color = pal[1], linewidth = 0.8) +
      ggplot2::geom_point(color = pal[1], size = 1.8) +
      .plot_theme() +
      ggplot2::labs(x = "Draw", y = "K", title = "Clusters per draw")
    return(.cluster_maybe_wrap_plotly(p, plotly = plotly))
  }

  lbl <- predict(x, type = "label", burnin = burnin, thin = thin)
  if (identical(which, "summary")) {
    return(plot(
      lbl,
      type = "summary",
      top_n = top_n,
      order_by = order_by,
      plotly = plotly,
      ...
    ))
  }

  plot(
    lbl,
    type = "sizes",
    top_n = top_n,
    order_by = order_by,
    plotly = plotly,
    title = "Dahl cluster sizes",
    ...
  )
}

#' Print cluster labels
#'
#' Compact display for a representative clustering.
#'
#' @details
#' A `dpmixgpd_cluster_labels` object represents one partition, usually the Dahl
#' representative partition for the training data or the induced allocation of
#' `newdata` to those representative clusters. The printed output therefore
#' describes the selected labels and their sizes, not the full posterior
#' uncertainty over alternative partitions.
#'
#' When the object comes from `predict(..., return_scores = TRUE)`, richer
#' assignment information is carried alongside the labels and can be inspected
#' with [summary.dpmixgpd_cluster_labels()] or
#' [plot.dpmixgpd_cluster_labels()].
#'
#' @param x Cluster labels object.
#' @param ... Unused.
#'
#' @return `x`, invisibly.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [summary.dpmixgpd_cluster_labels()],
#'   [plot.dpmixgpd_cluster_labels()].
#' @family cluster workflow
#' @export
print.dpmixgpd_cluster_labels <- function(x, ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_labels"))
  tab <- .cluster_size_table(x$labels, order_by = "size")
  cat("Cluster labels (", x$source %||% "train", ")\n", sep = "")
  cat("n         :", length(x$labels), "\n")
  cat("components:", x$components %||% length(tab), "\n")
  cat("sizes     :", paste(sprintf("%s:%s", names(tab), as.integer(tab)), collapse = ", "), "\n")
  invisible(x)
}

#' Summarize cluster labels
#'
#' Summarize a representative clustering for training data or new observations.
#'
#' @param object Cluster labels object.
#' @param top_n Number of populated clusters to profile when attached data are available.
#' @param order_by Ordering rule for descriptive cluster profiles:
#'   \itemize{
#'     \item \code{"size"}: decreasing cluster size
#'     \item \code{"label"}: ascending cluster label
#'   }
#' @param vars Optional character vector of numeric columns to summarize within each cluster.
#' @param ... Unused.
#'
#' @return Summary list containing cluster sizes, optional cluster-level descriptive summaries, and,
#'   when available, assignment-certainty summaries.
#'
#' @details
#' If score or probability matrices are attached, certainty is summarized by the rowwise maxima
#' \eqn{\max_k p_{ik}}, which quantify how strongly each observation is assigned to its selected
#' cluster. When the labels object also carries attached training or prediction data, the summary
#' includes descriptive mean/sd profiles for the first populated clusters.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [plot.dpmixgpd_cluster_labels()],
#'   [summary.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @export
summary.dpmixgpd_cluster_labels <- function(object,
                                            top_n = 5L,
                                            order_by = c("size", "label"),
                                            vars = NULL,
                                            ...) {
  stopifnot(inherits(object, "dpmixgpd_cluster_labels"))
  tab <- .cluster_size_table(object$labels, order_by = "size")
  score_mat <- object$scores %||% object$probs
  max_prob <- if (is.matrix(score_mat)) apply(score_mat, 1, max) else NA_real_
  out <- list(
    source = object$source %||% "train",
    n = length(object$labels),
    components = object$components %||% length(tab),
    cluster_sizes = tab,
    cluster_profiles = .cluster_profile_table(
      data = object$data %||% NULL,
      labels = object$labels,
      score_mat = score_mat,
      top_n = top_n,
      order_by = order_by,
      vars = vars
    ),
    certainty = if (all(is.na(max_prob))) NULL else summary(max_prob)
  )
  class(out) <- c("summary.dpmixgpd_cluster_labels", "list")
  out
}

#' Plot cluster labels
#'
#' Visualize representative cluster sizes, assignment certainty, or cluster-specific response
#' summaries. For `type = "summary"`, the response view is shown as boxplots ordered by
#' cluster size or label. When `x` comes from `predict(..., newdata = ...)`, only clusters
#' represented in the new sample are displayed.
#'
#' @details
#' This method visualizes the representative partition stored in a
#' `dpmixgpd_cluster_labels` object. The `sizes` view emphasizes the empirical
#' distribution of the selected clusters, the `certainty` view summarizes the
#' assignment scores \eqn{\max_k p_{ik}}, and the `summary` view compares the
#' attached response data across representative clusters.
#'
#' For new-data prediction, the plots are always interpreted relative to the
#' representative training clusters. That is why only clusters observed in the
#' predicted sample are shown even though the training partition may contain
#' additional occupied groups.
#'
#' @param x Cluster labels object.
#' @param type Plot type:
#'   \itemize{
#'     \item \code{"sizes"}: bar chart of representative cluster sizes
#'     \item \code{"certainty"}: assignment certainty distribution
#'     \item \code{"summary"}: cluster-specific response boxplots
#'   }
#' @param top_n Number of populated representative clusters to display for
#'   \code{type = "sizes"} or \code{type = "summary"}. Use \code{NULL} to
#'   display all populated clusters.
#' @param order_by Ordering rule for cluster displays:
#'   \itemize{
#'     \item \code{"size"}: decreasing cluster size
#'     \item \code{"label"}: ascending cluster label
#'   }
#' @param plotly Logical; if `TRUE`, convert the `ggplot2` output to a `plotly` /
#'   `htmlwidget` representation via `.wrap_plotly()`. Defaults to
#'   `getOption("CausalMixGPD.plotly", FALSE)`.
#' @param ... Unused.
#'
#' @return A `ggplot2` object or a `plotly`/`htmlwidget` object when `plotly = TRUE`.
#'
#' @seealso [summary.dpmixgpd_cluster_labels()], [predict.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @rdname plot.dpmixgpd_cluster_fit
#' @export
plot.dpmixgpd_cluster_labels <- function(x,
                                         type = c("sizes", "certainty", "summary"),
                                         top_n = 5L,
                                         order_by = c("size", "label"),
                                         plotly = getOption("CausalMixGPD.plotly", FALSE),
                                         ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_labels"))
  type <- match.arg(type)
  .cluster_require_ggplot()
  if (identical(type, "sizes")) {
    title <- list(...)$title %||% "Cluster sizes"
    ref_labels <- x$train_reference$labels %||% x$labels
    return(.cluster_plot_sizes(
      x$labels,
      title = title,
      top_n = top_n,
      order_by = order_by,
      reference_levels = .cluster_order_levels(ref_labels, order_by = order_by),
      plotly = plotly
    ))
  }
  if (identical(type, "summary")) {
    return(.cluster_plot_summary_labels(
      x,
      top_n = top_n,
      order_by = order_by,
      plotly = plotly
    ))
  }
  score_mat <- x$scores %||% x$probs
  if (!is.matrix(score_mat)) {
    warning("No score matrix available for certainty plot.", call. = FALSE)
    return(invisible(x))
  }
  max_prob <- apply(score_mat, 1, max)
  .cluster_plot_certainty(max_prob, plotly = plotly)
}

#' Print a cluster posterior similarity matrix
#'
#' Compact display for a posterior similarity matrix.
#'
#' @details
#' A posterior similarity matrix records pairwise co-clustering probabilities on
#' the training sample. Its \eqn{(i,j)} entry is the posterior probability that
#' observations \eqn{i} and \eqn{j} share the same latent cluster across the
#' retained MCMC draws.
#'
#' The printed header reports only matrix size and bookkeeping information. Use
#' [summary.dpmixgpd_cluster_psm()] for numerical summaries and
#' [plot.dpmixgpd_cluster_psm()] for a visual heatmap.
#'
#' @param x Cluster PSM object.
#' @param ... Unused.
#'
#' @return `x`, invisibly.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [summary.dpmixgpd_cluster_psm()],
#'   [plot.dpmixgpd_cluster_psm()].
#' @family cluster workflow
#' @export
print.dpmixgpd_cluster_psm <- function(x, ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_psm"))
  n <- nrow(x$psm %||% matrix(0, 0, 0))
  cat("Cluster PSM\n")
  cat("n         :", n, "\n")
  cat("components:", x$components %||% NA_integer_, "\n")
  cat("draw_index:", x$draw_index %||% NA_integer_, "\n")
  invisible(x)
}

#' Summarize a cluster posterior similarity matrix
#'
#' Summarize pairwise posterior co-clustering probabilities.
#'
#' @param object Cluster PSM object.
#' @param ... Unused.
#'
#' @return Summary list with matrix size and basic summaries of the similarity entries.
#'
#' @details
#' The diagonal of a posterior similarity matrix is always close to one, while off-diagonal values
#' near one indicate highly stable co-clustering across retained posterior draws.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [plot.dpmixgpd_cluster_psm()],
#'   [summary.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @export
summary.dpmixgpd_cluster_psm <- function(object, ...) {
  stopifnot(inherits(object, "dpmixgpd_cluster_psm"))
  p <- object$psm
  out <- list(
    n = nrow(p),
    components = object$components %||% NA_integer_,
    min = min(p, na.rm = TRUE),
    mean = mean(p, na.rm = TRUE),
    max = max(p, na.rm = TRUE),
    diagonal_mean = mean(diag(p), na.rm = TRUE)
  )
  class(out) <- c("summary.dpmixgpd_cluster_psm", "list")
  out
}

#' Plot a cluster posterior similarity matrix
#'
#' Heatmap of pairwise posterior co-clustering probabilities.
#'
#' @details
#' The heatmap visualizes the matrix
#' \deqn{
#' \mathrm{PSM}_{ij} \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}),
#' }
#' so larger values indicate pairs of observations that are stably allocated to
#' the same cluster over the retained posterior draws.
#'
#' Because the PSM is an \eqn{n \times n} object, plotting and even storing it
#' becomes expensive for large `n`. The `psm_max_n` argument is therefore a
#' deliberate guard against accidental quadratic memory use.
#'
#' @param x Cluster PSM object.
#' @param psm_max_n Maximum allowed matrix size for plotting.
#' @param order_by Ordering rule for rows and columns:
#'   \itemize{
#'     \item \code{"label"}: order by representative cluster labels when available
#'     \item \code{"hclust"}: order by hierarchical clustering of \code{1 - PSM}
#'     \item \code{"input"}: preserve input order
#'   }
#' @param plotly Logical; if `TRUE`, convert the `ggplot2` output to a `plotly` /
#'   `htmlwidget` representation via `.wrap_plotly()`. Defaults to
#'   `getOption("CausalMixGPD.plotly", FALSE)`.
#' @param ... Unused.
#'
#' @return A `ggplot2` object or a `plotly`/`htmlwidget` object when `plotly = TRUE`.
#'
#' @seealso [predict.dpmixgpd_cluster_fit()], [summary.dpmixgpd_cluster_psm()],
#'   [plot.dpmixgpd_cluster_fit()].
#' @family cluster workflow
#' @rdname plot.dpmixgpd_cluster_fit
#' @export
plot.dpmixgpd_cluster_psm <- function(x,
                                      psm_max_n = x$psm_max_n %||% 2000L,
                                      order_by = c("label", "hclust", "input"),
                                      plotly = getOption("CausalMixGPD.plotly", FALSE),
                                      ...) {
  stopifnot(inherits(x, "dpmixgpd_cluster_psm"))
  .cluster_require_ggplot()
  order_by <- match.arg(order_by)
  psm_max_n <- as.integer(psm_max_n)[1]
  if (!is.finite(psm_max_n) || psm_max_n < 1L) {
    stop("'psm_max_n' must be an integer >= 1.", call. = FALSE)
  }
  n <- nrow(x$psm %||% matrix(0, 0, 0))
  if (n > psm_max_n) {
    stop(
      sprintf(
        "PSM plot blocked: n=%d exceeds psm_max_n=%d. Increase 'psm_max_n' to plot.",
        n,
        psm_max_n
      ),
      call. = FALSE
    )
  }
  .cluster_plot_psm(x$psm, labels = x$labels %||% NULL, order_by = order_by, plotly = plotly)
}




.cluster_draw_indices <- function(n_draws, burnin = NULL, thin = NULL) {
  burnin <- as.integer(burnin %||% 0L)
  thin <- as.integer(thin %||% 1L)
  if (!is.finite(burnin) || burnin < 0L) stop("'burnin' must be >= 0.", call. = FALSE)
  if (!is.finite(thin) || thin < 1L) stop("'thin' must be >= 1.", call. = FALSE)
  if (burnin >= n_draws) stop("'burnin' is too large for available draws.", call. = FALSE)
  seq.int(from = burnin + 1L, to = n_draws, by = thin)
}

.cluster_samples_to_matrix <- function(samples) {
  if (!requireNamespace("coda", quietly = TRUE)) {
    stop("Package 'coda' is required.", call. = FALSE)
  }
  smp <- samples
  if (inherits(smp, "mcmc")) smp <- coda::mcmc.list(smp)
  if (!inherits(smp, "mcmc.list")) stop("Expected 'mcmc' or 'mcmc.list' samples.", call. = FALSE)
  do.call(rbind, lapply(smp, as.matrix))
}

.cluster_extract_z_from_matrix <- function(draw_mat) {
  z_cols <- grep("^z\\[[0-9]+\\]$", colnames(draw_mat))
  if (!length(z_cols)) stop("No z[i] columns found in samples.", call. = FALSE)
  z_names <- colnames(draw_mat)[z_cols]
  z_idx <- as.integer(sub("^z\\[([0-9]+)\\]$", "\\1", z_names))
  z_cols <- z_cols[order(z_idx)]
  z <- draw_mat[, z_cols, drop = FALSE]
  mode(z) <- "integer"
  z
}

extract_z_draws <- function(samples, burnin = NULL, thin = NULL) {
  draw_mat <- .cluster_samples_to_matrix(samples)
  idx <- .cluster_draw_indices(nrow(draw_mat), burnin = burnin, thin = thin)
  z <- .cluster_extract_z_from_matrix(draw_mat[idx, , drop = FALSE])
  z
}

compute_psm <- function(z_draws) {
  if (exists(".compute_psm", mode = "function")) return(.compute_psm(z_draws))
  n_iter <- nrow(z_draws)
  n_obs <- ncol(z_draws)
  psm <- matrix(0, n_obs, n_obs)
  for (s in seq_len(n_iter)) {
    z <- z_draws[s, ]
    psm <- psm + outer(z, z, "==") * 1.0
  }
  psm / n_iter
}

dahl_labels <- function(z_draws, psm) {
  if (exists(".dahl_representative", mode = "function")) return(.dahl_representative(z_draws, psm))

  n_iter <- nrow(z_draws)
  ssq <- numeric(n_iter)
  for (s in seq_len(n_iter)) {
    z <- z_draws[s, ]
    A <- outer(z, z, "==") * 1.0
    ssq[s] <- sum((A - psm)^2)
  }
  s_star <- which.min(ssq)
  z_hat <- as.integer(z_draws[s_star, ])
  labels <- match(z_hat, unique(z_hat))
  list(draw_index = s_star, labels = labels, K = length(unique(labels)))
}

.cluster_compute_scores <- function(z_draws, labels, psm) {
  if (exists(".compute_cluster_probs", mode = "function")) {
    return(.compute_cluster_probs(z_draws, labels, psm))
  }
  n_obs <- length(labels)
  K <- length(unique(labels))
  probs <- matrix(0, nrow = n_obs, ncol = K)
  for (k in seq_len(K)) {
    idx <- which(labels == k)
    for (i in seq_len(n_obs)) probs[i, k] <- mean(psm[i, idx])
  }
  rs <- rowSums(probs)
  rs[rs <= 0] <- 1
  probs / rs
}

.cluster_compute_probs <- .cluster_compute_scores

.cluster_order_levels <- function(labels, order_by = c("size", "label")) {
  order_by <- match.arg(order_by)
  tab <- table(as.character(labels))
  lev <- names(tab)
  lev_num <- suppressWarnings(as.numeric(lev))
  if (identical(order_by, "size")) {
    ord <- order(-as.integer(tab), ifelse(is.na(lev_num), Inf, lev_num), lev)
  } else {
    ord <- order(ifelse(is.na(lev_num), Inf, lev_num), lev)
  }
  lev[ord]
}

.cluster_validate_top_n <- function(top_n) {
  if (is.null(top_n)) return(NULL)
  top_n <- as.integer(top_n)[1]
  if (!is.finite(top_n) || top_n < 1L) {
    stop("'top_n' must be an integer >= 1.", call. = FALSE)
  }
  top_n
}

.cluster_select_levels <- function(labels,
                                   top_n = 5L,
                                   order_by = c("size", "label")) {
  keep <- .cluster_order_levels(labels, order_by = order_by)
  top_n <- .cluster_validate_top_n(top_n)
  if (!is.null(top_n)) keep <- head(keep, top_n)
  keep
}

.cluster_size_table <- function(labels, order_by = c("size", "label"), levels = NULL) {
  labels_chr <- as.character(labels)
  if (is.null(levels)) {
    levels <- .cluster_order_levels(labels_chr, order_by = order_by)
  } else {
    levels <- as.character(levels)
  }
  table(factor(labels_chr, levels = levels))
}

.cluster_data_frame_from_design <- function(design, formula_meta = list()) {
  if (is.null(design)) return(NULL)

  response_name <- formula_meta$response %||% "y"
  out <- data.frame(
    stats::setNames(list(as.numeric(design$y %||% numeric(0))), response_name),
    check.names = FALSE
  )

  X <- design$X %||% NULL
  X_cols <- formula_meta$X_cols %||% character(0)
  if (!is.null(X)) {
    X_df <- as.data.frame(X, check.names = FALSE, stringsAsFactors = FALSE)
    if (length(X_cols) == ncol(X_df)) names(X_df) <- X_cols
    out <- data.frame(out, X_df, check.names = FALSE)
  }

  out
}

.cluster_training_data_frame <- function(fit) {
  stopifnot(inherits(fit, "dpmixgpd_cluster_fit"))
  bundle <- fit$bundle %||% list()
  design <- bundle$data %||% NULL
  formula_meta <- ((fit$spec %||% list())$cluster %||% list())$formula_meta %||% list()
  .cluster_data_frame_from_design(design = design, formula_meta = formula_meta)
}

.cluster_training_reference <- function(fit, labels) {
  data <- .cluster_training_data_frame(fit)
  if (is.null(data) || !length(labels)) return(NULL)
  list(
    labels = as.integer(labels),
    data = data
  )
}

.cluster_response_name <- function(data) {
  data <- as.data.frame(data, check.names = FALSE, stringsAsFactors = FALSE)
  nm <- names(data)[1]
  if (is.null(nm) || !nzchar(nm)) "y" else as.character(nm)
}

.cluster_response_values <- function(data, response_name = NULL) {
  if (is.null(data)) return(numeric(0))
  data <- as.data.frame(data, check.names = FALSE, stringsAsFactors = FALSE)
  response_name <- as.character(response_name %||% .cluster_response_name(data))[1]
  if (!(response_name %in% names(data))) return(numeric(0))
  as.numeric(data[[response_name]])
}

.cluster_plot_ylim <- function(values) {
  vals <- as.numeric(values)
  vals <- vals[is.finite(vals)]
  if (!length(vals)) return(c(0, 1))

  rng <- range(vals)
  pad <- if (rng[1] == rng[2]) {
    max(abs(rng[1]), 1) * 0.08
  } else {
    diff(rng) * 0.06
  }
  c(rng[1] - pad, rng[2] + pad)
}

.cluster_scale_values <- function(levels, reference_levels = NULL) {
  lev <- as.character(levels %||% character(0))
  if (!length(lev)) return(stats::setNames(character(0), character(0)))

  ref <- as.character(reference_levels %||% lev)
  ref <- unique(ref)
  if (!length(ref)) ref <- lev

  stats::setNames(.plot_palette(length(ref)), ref)[lev]
}

.cluster_hist_breaks <- function(values, bins = 18L) {
  vals <- as.numeric(values)
  vals <- vals[is.finite(vals)]
  if (!length(vals)) return(c(0, 1))

  bins <- as.integer(bins %||% 18L)
  if (!is.finite(bins) || bins < 4L) bins <- 18L

  rng <- range(vals)
  if (rng[1] == rng[2]) {
    pad <- max(abs(rng[1]), 1) * 0.25
    return(seq(rng[1] - pad, rng[2] + pad, length.out = bins + 1L))
  }

  br <- pretty(rng, n = bins)
  br <- sort(unique(as.numeric(br[is.finite(br)])))
  if (length(br) < 2L) {
    br <- seq(rng[1], rng[2], length.out = bins + 1L)
  }
  if (min(br) > rng[1]) br <- c(rng[1], br)
  if (max(br) < rng[2]) br <- c(br, rng[2])
  br <- sort(unique(br))

  if (length(br) < 2L) {
    return(seq(rng[1], rng[2], length.out = bins + 1L))
  }
  br
}

.cluster_hist_max_counts <- function(df, levels, breaks) {
  lev <- as.character(levels %||% character(0))
  out <- stats::setNames(rep(0, length(lev)), lev)
  if (is.null(df) || !nrow(df) || length(breaks) < 2L) return(out)

  for (cl in lev) {
    vals <- as.numeric(df$response[as.character(df$cluster) == cl])
    vals <- vals[is.finite(vals)]
    if (!length(vals)) next
    hist_obj <- suppressWarnings(graphics::hist(vals, breaks = breaks, plot = FALSE))
    out[[cl]] <- max(c(0, hist_obj$counts), na.rm = TRUE)
  }

  out
}

.cluster_require_ggplot <- function() {
  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package 'ggplot2' is required for plotting. Install it first.", call. = FALSE)
  }
}

.cluster_maybe_wrap_plotly <- function(p, plotly = getOption("CausalMixGPD.plotly", FALSE)) {
  if (!isTRUE(plotly)) return(p)
  old_options <- options(CausalMixGPD.plotly = TRUE)
  on.exit(options(old_options), add = TRUE)
  .wrap_plotly(p)
}

.cluster_plot_sizes <- function(labels,
                                title = "Cluster sizes",
                                top_n = 5L,
                                order_by = c("size", "label"),
                                reference_levels = NULL,
                                plotly = getOption("CausalMixGPD.plotly", FALSE)) {
  .cluster_require_ggplot()
  keep <- .cluster_select_levels(labels, top_n = top_n, order_by = order_by)
  tab <- .cluster_size_table(labels, levels = keep)
  df <- data.frame(
    cluster = factor(names(tab), levels = names(tab)),
    size = as.integer(tab),
    stringsAsFactors = FALSE
  )
  cluster_values <- .cluster_scale_values(
    names(tab),
    reference_levels = reference_levels %||% keep
  )
  p <- ggplot2::ggplot(
    df,
    ggplot2::aes(x = cluster, y = size, fill = cluster, color = cluster)
  ) +
    ggplot2::geom_col(linewidth = 0.4, show.legend = FALSE) +
    .plot_theme() +
    ggplot2::scale_fill_manual(values = cluster_values, drop = FALSE) +
    ggplot2::scale_color_manual(values = cluster_values, drop = FALSE) +
    ggplot2::scale_x_discrete(labels = function(v) paste0("C", v)) +
    ggplot2::labs(x = "Cluster", y = "Size", title = title) +
    ggplot2::theme(legend.position = "none")
  .cluster_maybe_wrap_plotly(p, plotly = plotly)
}

.cluster_plot_certainty <- function(max_prob,
                                    plotly = getOption("CausalMixGPD.plotly", FALSE)) {
  .cluster_require_ggplot()
  df <- data.frame(max_prob = as.numeric(max_prob))
  df <- df[is.finite(df$max_prob), , drop = FALSE]
  pal <- .plot_palette(2L)
  p <- ggplot2::ggplot(df, ggplot2::aes(x = max_prob)) +
    ggplot2::geom_histogram(bins = 20, fill = pal[1], color = pal[2], alpha = 0.75) +
    .plot_theme() +
    ggplot2::scale_x_continuous(limits = c(0, 1)) +
    ggplot2::labs(
      x = "Max assignment probability",
      y = "Count",
      title = "Label certainty"
    )
  .cluster_maybe_wrap_plotly(p, plotly = plotly)
}

.cluster_psm_order <- function(psm, labels = NULL, order_by = c("label", "hclust", "input")) {
  order_by <- match.arg(order_by)
  n <- nrow(psm)
  if (identical(order_by, "input") || n < 2L) return(seq_len(n))
  if (identical(order_by, "label") && !is.null(labels) && length(labels) == n) {
    labels_chr <- as.character(labels)
    labels_num <- suppressWarnings(as.numeric(labels_chr))
    row_score <- rowMeans(psm, na.rm = TRUE)
    return(order(ifelse(is.na(labels_num), Inf, labels_num), labels_chr, -row_score, seq_len(n)))
  }
  if (n > 2L) {
    return(tryCatch(
      stats::hclust(stats::as.dist(1 - psm), method = "average")$order,
      error = function(e) seq_len(n)
    ))
  }
  seq_len(n)
}

.cluster_plot_psm <- function(psm,
                              labels = NULL,
                              order_by = c("label", "hclust", "input"),
                              plotly = getOption("CausalMixGPD.plotly", FALSE)) {
  .cluster_require_ggplot()
  order_by <- match.arg(order_by)
  ord <- .cluster_psm_order(psm, labels = labels, order_by = order_by)
  psm <- psm[ord, ord, drop = FALSE]
  n <- nrow(psm)
  df <- expand.grid(
    row = seq_len(n),
    col = seq_len(n),
    KEEP.OUT.ATTRS = FALSE
  )
  df$prob <- as.numeric(psm)

  p <- ggplot2::ggplot(df, ggplot2::aes(x = col, y = row, fill = prob)) +
    ggplot2::geom_raster() +
    ggplot2::scale_fill_viridis_c(limits = c(0, 1), option = "C") +
    ggplot2::scale_y_reverse(expand = c(0, 0)) +
    ggplot2::scale_x_continuous(expand = c(0, 0)) +
    ggplot2::coord_fixed() +
    .plot_theme() +
    ggplot2::labs(
      x = "Observation index",
      y = "Observation index",
      fill = "PSM",
      title = if (identical(order_by, "input")) {
        "Posterior Similarity Matrix"
      } else {
        sprintf("Posterior Similarity Matrix (ordered by %s)", order_by)
      }
    ) +
    ggplot2::theme(
      axis.text = ggplot2::element_blank(),
      axis.ticks = ggplot2::element_blank()
    )
  .cluster_maybe_wrap_plotly(p, plotly = plotly)
}

.cluster_summary_plot_data <- function(object,
                                      top_n = 5L,
                                      order_by = c("size", "label")) {
  stopifnot(inherits(object, "dpmixgpd_cluster_labels"))
  data <- object$data %||% NULL
  if (is.null(data) || !nrow(as.data.frame(data))) {
    warning("No attached data available for summary plot.", call. = FALSE)
    return(NULL)
  }

  data <- as.data.frame(data, check.names = FALSE, stringsAsFactors = FALSE)
  if (nrow(data) != length(object$labels)) {
    warning("Attached data and labels must have the same number of rows for summary plot.", call. = FALSE)
    return(NULL)
  }

  response_name <- .cluster_response_name(data)
  levels <- .cluster_select_levels(object$labels, top_n = top_n, order_by = order_by)

  main_df <- data.frame(
    cluster = factor(as.character(object$labels), levels = levels),
    response = .cluster_response_values(data, response_name = response_name),
    sample = if (identical(object$source %||% "", "newdata")) "newdata" else "train",
    stringsAsFactors = FALSE
  )
  main_df <- main_df[is.finite(main_df$response) & !is.na(main_df$cluster), , drop = FALSE]

  if (!nrow(main_df)) {
    warning("No finite response values available for summary plot.", call. = FALSE)
    return(NULL)
  }

  list(
    response_name = response_name,
    has_reference = FALSE,
    levels = levels,
    main_df = main_df,
    ref_df = NULL,
    ylim = .cluster_plot_ylim(main_df$response),
    title = if (identical(object$source %||% "", "newdata")) {
      "Cluster response summary: newdata"
    } else {
      "Cluster response summary"
    }
  )
}

.cluster_plot_summary_labels <- function(object,
                                         top_n = 5L,
                                         order_by = c("size", "label"),
                                         plotly = getOption("CausalMixGPD.plotly", FALSE)) {
  .cluster_require_ggplot()
  dat <- .cluster_summary_plot_data(object, top_n = top_n, order_by = order_by)
  if (is.null(dat)) return(invisible(object))
  ref_labels <- if (identical(object$source %||% "", "newdata")) {
    (object$train_reference %||% list())$labels %||% object$labels
  } else {
    object$labels
  }
  cluster_values <- .cluster_scale_values(
    dat$levels,
    reference_levels = .cluster_order_levels(ref_labels, order_by = order_by)
  )
  p <- ggplot2::ggplot(
    dat$main_df,
    ggplot2::aes(x = cluster, y = response, fill = cluster, color = cluster)
  ) +
    ggplot2::geom_boxplot(
      width = 0.72,
      alpha = 0.82,
      outlier.alpha = 0.45,
      linewidth = 0.45,
      show.legend = FALSE
    ) +
    ggplot2::scale_fill_manual(values = cluster_values, drop = FALSE) +
    ggplot2::scale_color_manual(values = cluster_values, drop = FALSE) +
    ggplot2::scale_x_discrete(
      drop = FALSE,
      labels = function(v) paste0("C", v)
    ) +
    ggplot2::coord_cartesian(ylim = dat$ylim) +
    .plot_theme() +
    ggplot2::labs(
      x = "Cluster",
      y = dat$response_name,
      title = dat$title
    ) +
    ggplot2::theme(legend.position = "none")

  .cluster_maybe_wrap_plotly(p, plotly = plotly)
}

.cluster_profile_table <- function(data,
                                   labels,
                                   score_mat = NULL,
                                   top_n = 5L,
                                   order_by = c("size", "label"),
                                   vars = NULL) {
  if (is.null(data)) return(NULL)

  order_by <- match.arg(order_by)
  data <- as.data.frame(data, check.names = FALSE, stringsAsFactors = FALSE)
  if (!nrow(data) || !length(labels)) return(NULL)
  if (nrow(data) != length(labels)) {
    stop("Cluster labels and attached data must have the same number of rows.", call. = FALSE)
  }

  if (is.null(vars)) {
    vars <- names(data)[vapply(data, is.numeric, logical(1))]
  } else {
    vars <- as.character(vars)
    bad <- setdiff(vars, names(data))
    if (length(bad)) {
      stop(sprintf("Unknown profiling variables: %s", paste(bad, collapse = ", ")), call. = FALSE)
    }
    vars <- vars[vapply(data[vars], is.numeric, logical(1))]
  }

  keep <- .cluster_select_levels(labels, top_n = top_n, order_by = order_by)

  max_prob <- if (is.matrix(score_mat)) apply(score_mat, 1, max) else NULL
  rows <- lapply(keep, function(cl) {
    idx <- which(as.character(labels) == cl)
    row <- list(
      cluster = paste0("C", cl),
      n = length(idx)
    )

    if (length(vars)) {
      for (nm in vars) {
        x <- as.numeric(data[[nm]][idx])
        row[[paste0(nm, "_mean")]] <- mean(x, na.rm = TRUE)
        row[[paste0(nm, "_sd")]] <- if (sum(is.finite(x)) > 1L) stats::sd(x, na.rm = TRUE) else NA_real_
      }
    }

    if (!is.null(max_prob)) {
      p <- max_prob[idx]
      row$certainty_mean <- mean(p, na.rm = TRUE)
      row$certainty_sd <- if (sum(is.finite(p)) > 1L) stats::sd(p, na.rm = TRUE) else NA_real_
    }

    as.data.frame(row, check.names = FALSE)
  })

  out <- do.call(rbind, rows)
  row.names(out) <- NULL
  out
}

.cluster_check_new_factor_levels <- function(meta, newdata) {
  xl <- meta$xlevels %||% list()
  if (!length(xl)) return(invisible(NULL))
  for (nm in names(xl)) {
    if (!(nm %in% names(newdata))) next
    vals <- as.character(newdata[[nm]])
    vals <- unique(vals[!is.na(vals)])
    bad <- setdiff(vals, as.character(xl[[nm]]))
    if (length(bad)) {
      stop(
        sprintf(
          "newdata has unseen factor levels for '%s': %s",
          nm,
          paste(bad, collapse = ", ")
        ),
        call. = FALSE
      )
    }
  }
  invisible(NULL)
}

.cluster_build_design <- function(meta, newdata) {
  if (!is.data.frame(newdata)) newdata <- as.data.frame(newdata)
  response_name <- meta$response %||% "y"
  if (!(response_name %in% names(newdata))) {
    stop(sprintf("newdata must include response column '%s'.", response_name), call. = FALSE)
  }

  y_new <- as.numeric(newdata[[response_name]])
  if (anyNA(y_new)) stop("newdata response contains NA.", call. = FALSE)

  trm <- meta$terms
  has_X <- length(meta$X_cols %||% character(0)) > 0L
  if (!has_X) {
    return(list(y = y_new, X = NULL))
  }

  rhs <- stats::delete.response(trm)
  .cluster_check_new_factor_levels(meta = meta, newdata = newdata)
  mf_new <- tryCatch(
    stats::model.frame(
      rhs,
      data = newdata,
      xlev = meta$xlevels %||% NULL,
      na.action = stats::na.fail
    ),
    error = function(e) {
      msg <- conditionMessage(e)
      if (grepl("new level", msg, fixed = TRUE)) {
        stop(sprintf("newdata has unseen factor levels: %s", msg), call. = FALSE)
      }
      stop(sprintf("Failed to build model frame for newdata: %s", conditionMessage(e)), call. = FALSE)
    }
  )
  mm <- stats::model.matrix(rhs, data = mf_new, contrasts.arg = meta$contrasts %||% NULL)
  if ("(Intercept)" %in% colnames(mm)) {
    mm <- mm[, colnames(mm) != "(Intercept)", drop = FALSE]
  }
  X_cols <- meta$X_cols %||% character(0)
  miss <- setdiff(X_cols, colnames(mm))
  if (length(miss)) {
    stop(sprintf("newdata is missing required predictors: %s", paste(miss, collapse = ", ")), call. = FALSE)
  }
  mm <- mm[, X_cols, drop = FALSE]
  storage.mode(mm) <- "double"
  list(y = y_new, X = mm)
}

.cluster_link_apply <- function(eta, link = "identity", link_power = NULL) {
  link <- as.character(link %||% "identity")
  if (identical(link, "identity")) return(eta)
  if (identical(link, "exp")) return(exp(eta))
  if (identical(link, "log")) return(log(eta))
  if (identical(link, "softplus")) return(log1p(exp(eta)))
  if (identical(link, "power")) {
    pw <- as.numeric(link_power)
    if (!is.finite(pw) || length(pw) != 1L) stop("Invalid power link exponent.", call. = FALSE)
    return(eta ^ pw)
  }
  stop(sprintf("Unsupported link '%s'.", link), call. = FALSE)
}

.cluster_extract_beta_component <- function(draw_row, base, comp, P) {
  out <- rep(0, P)
  for (p in seq_len(P)) {
    nm <- sprintf("%s[%d,%d]", base, comp, p)
    if (!(nm %in% names(draw_row))) {
      nm <- sprintf("%s[%d, %d]", base, comp, p)
    }
    if (nm %in% names(draw_row)) out[p] <- as.numeric(draw_row[[nm]])
  }
  out
}

.cluster_extract_beta_global <- function(draw_row, base, P) {
  out <- rep(0, P)
  for (p in seq_len(P)) {
    nm <- sprintf("%s[%d]", base, p)
    if (nm %in% names(draw_row)) out[p] <- as.numeric(draw_row[[nm]])
  }
  out
}

.cluster_extract_beta_auto <- function(draw_row, base, comp, P) {
  if (P < 1L) return(numeric(0))
  nn <- names(draw_row)
  has_component <- any(grepl(sprintf("^%s\\[%d,\\s*[0-9]+\\]$", base, comp), nn))
  if (isTRUE(has_component)) {
    return(.cluster_extract_beta_component(draw_row, base = base, comp = comp, P = P))
  }
  .cluster_extract_beta_global(draw_row, base = base, P = P)
}

.cluster_softmax <- function(logits) {
  logits <- as.numeric(logits)
  if (!length(logits)) return(numeric(0))
  shift <- max(logits)
  ex <- exp(logits - shift)
  s <- sum(ex)
  if (!is.finite(s) || s <= 0) return(rep(1 / length(logits), length(logits)))
  ex / s
}

.cluster_extract_gating_draw <- function(draw_row, K, P) {
  K <- as.integer(K)
  P <- as.integer(P)
  if (K < 2L || P < 1L) return(NULL)
  eta <- rep(0, K - 1L)
  B <- matrix(0, nrow = K - 1L, ncol = P)
  nn <- names(draw_row)

  for (j in seq_len(K - 1L)) {
    nm_eta <- sprintf("eta[%d]", j)
    if (nm_eta %in% nn) eta[j] <- as.numeric(draw_row[[nm_eta]])
    for (p in seq_len(P)) {
      nm <- sprintf("B[%d,%d]", j, p)
      if (!(nm %in% nn)) nm <- sprintf("B[%d, %d]", j, p)
      if (nm %in% nn) B[j, p] <- as.numeric(draw_row[[nm]])
    }
  }

  if (!any(grepl("^eta\\[[0-9]+\\]$", nn)) || !any(grepl("^B\\[[0-9]+,\\s*[0-9]+\\]$", nn))) {
    return(NULL)
  }
  list(eta = eta, B = B)
}

.cluster_gating_weights <- function(gating_draw, x_row) {
  if (is.null(gating_draw)) return(NULL)
  eta <- gating_draw$eta
  B <- gating_draw$B
  K <- length(eta) + 1L
  x_row <- as.numeric(x_row)
  lin <- as.numeric(B %*% x_row)
  logits <- c(eta + lin, 0)
  out <- .cluster_softmax(logits)
  if (length(out) != K) return(NULL)
  out
}

.cluster_resolve_density_fun <- function(spec) {
  meta <- spec$meta %||% list()
  kernel <- meta$kernel
  GPD <- isTRUE(meta$GPD)
  kdef <- get_kernel_registry()[[kernel]]
  if (is.null(kdef)) stop(sprintf("Kernel '%s' not found.", kernel), call. = FALSE)
  d_name <- if (GPD) kdef$crp$d_gpd else kdef$crp$d_base
  if (is.null(d_name) || isTRUE(is.na(d_name))) stop("Could not resolve density function.", call. = FALSE)

  ns_pkg <- getNamespace("CausalMixGPD")
  ns_stats <- getNamespace("stats")
  ns_nimble <- getNamespace("nimble")

  if (exists(d_name, envir = ns_pkg, inherits = FALSE)) return(get(d_name, envir = ns_pkg))
  if (exists(d_name, envir = ns_stats, inherits = FALSE)) return(get(d_name, envir = ns_stats))
  if (exists(d_name, envir = ns_nimble, inherits = FALSE)) return(get(d_name, envir = ns_nimble))
  stop(sprintf("Density function '%s' is unavailable.", d_name), call. = FALSE)
}

.cluster_component_density <- function(spec, draw_row, k, x_row, y_val, density_fun) {
  meta <- spec$meta %||% list()
  plan <- spec$plan %||% list()
  bulk <- plan$bulk %||% list()
  arg_order <- if (isTRUE(meta$GPD)) spec$signatures$gpd$args else spec$signatures$bulk$args
  P <- as.integer(meta$P %||% 0L)

  args <- list(x = as.numeric(y_val))

  for (nm in arg_order) {
    if (nm %in% names(bulk)) {
      ent <- bulk[[nm]]
      mode <- ent$mode %||% "dist"
      if (identical(mode, "link")) {
        b <- .cluster_extract_beta_component(draw_row, paste0("beta_", nm), comp = k, P = P)
        eta <- if (P > 0L) sum(x_row * b) else 0
        args[[nm]] <- .cluster_link_apply(eta, ent$link, ent$link_power)
      } else {
        col_nm <- sprintf("%s[%d]", nm, k)
        if (!(col_nm %in% names(draw_row))) {
          args[[nm]] <- as.numeric(ent$value %||% NA_real_)
        } else {
          args[[nm]] <- as.numeric(draw_row[[col_nm]])
        }
      }
      next
    }

    if (nm %in% c("threshold", "tail_scale", "tail_shape")) {
      gpd <- plan$gpd %||% list()
      ent <- gpd[[nm]] %||% list(mode = "dist")
      if (identical(ent$mode, "link")) {
        b <- .cluster_extract_beta_auto(
          draw_row = draw_row,
          base = paste0("beta_", nm),
          comp = k,
          P = P
        )
        eta <- if (P > 0L) sum(x_row * b) else 0
        args[[nm]] <- .cluster_link_apply(eta, ent$link, ent$link_power)
      } else if (identical(ent$mode, "fixed")) {
        args[[nm]] <- as.numeric(ent$value)
      } else {
        # dist mode: prefer scalar draw, fallback to first indexed value
        if (nm %in% names(draw_row)) {
          args[[nm]] <- as.numeric(draw_row[[nm]])
        } else {
          col_nm <- sprintf("%s[%d]", nm, k)
          if (col_nm %in% names(draw_row)) {
            args[[nm]] <- as.numeric(draw_row[[col_nm]])
          } else {
            args[[nm]] <- NA_real_
          }
        }
      }
      next
    }
  }

  fm <- names(formals(density_fun))
  if ("log" %in% fm && is.null(args$log)) args$log <- FALSE

  val <- suppressWarnings(tryCatch(as.numeric(do.call(density_fun, args))[1], error = function(e) NA_real_))
  if (!is.finite(val) || val < 0) val <- 0
  val
}

predict_labels_newdata <- function(fit, newdata, burnin = NULL, thin = NULL) {
  stopifnot(inherits(fit, "dpmixgpd_cluster_fit"))
  spec <- fit$spec
  meta <- spec$meta %||% list()
  formula_meta <- (spec$cluster %||% list())$formula_meta %||% list()
  newdat <- .cluster_build_design(formula_meta, newdata = newdata)

  draw_mat <- .cluster_samples_to_matrix(fit$samples)
  idx <- .cluster_draw_indices(nrow(draw_mat), burnin = burnin, thin = thin)
  draw_sub <- draw_mat[idx, , drop = FALSE]
  z_draws <- .cluster_extract_z_from_matrix(draw_sub)

  cache_key <- paste("cache", length(idx), burnin %||% 0L, thin %||% 1L, sep = "_")
  cache_env <- fit$cache_env %||% new.env(parent = emptyenv())
  if (exists(cache_key, envir = cache_env, inherits = FALSE)) {
    tr <- get(cache_key, envir = cache_env, inherits = FALSE)
    if (is.null(tr$scores_train) && !is.null(tr$probs_train)) tr$scores_train <- tr$probs_train
  } else {
    psm <- compute_psm(z_draws)
    dahl <- dahl_labels(z_draws, psm)
    scores_train <- .cluster_compute_scores(z_draws, dahl$labels, psm)
    tr <- list(psm = psm, dahl = dahl, scores_train = scores_train)
    assign(cache_key, tr, envir = cache_env)
  }

  dahl_labels_train <- as.integer(tr$dahl$labels)
  Kd <- as.integer(tr$dahl$K)
  density_fun <- .cluster_resolve_density_fun(spec)
  n_new <- length(newdat$y)
  out_scores <- matrix(0, nrow = n_new, ncol = Kd)

  K <- as.integer(meta$components %||% max(z_draws, na.rm = TRUE))
  type <- (spec$cluster %||% list())$type %||% "weights"
  use_gating <- type %in% c("weights", "both")

  for (s in seq_len(nrow(draw_sub))) {
    draw_row <- draw_sub[s, ]
    z_s <- as.integer(z_draws[s, ])
    comp_sizes <- pmax(tabulate(z_s, nbins = K), 0)
    if (sum(comp_sizes) <= 0) comp_sizes <- rep(1, K)
    gating_draw <- if (isTRUE(use_gating) && !is.null(newdat$X)) {
      .cluster_extract_gating_draw(draw_row = draw_row, K = K, P = ncol(newdat$X))
    } else {
      NULL
    }
    w_nm <- paste0("w[", seq_len(K), "]")

    map_comp_to_dahl <- matrix(0, nrow = K, ncol = Kd)
    for (k in seq_len(K)) {
      idx_k <- which(z_s == k)
      if (!length(idx_k)) {
        map_comp_to_dahl[k, ] <- rep(1 / Kd, Kd)
      } else {
        map_comp_to_dahl[k, ] <- tabulate(dahl_labels_train[idx_k], nbins = Kd) / length(idx_k)
      }
    }

    for (i in seq_len(n_new)) {
      x_row <- if (is.null(newdat$X)) numeric(0) else as.numeric(newdat$X[i, ])
      weight_factor <- comp_sizes
      if (isTRUE(use_gating)) {
        w_new <- .cluster_gating_weights(gating_draw = gating_draw, x_row = x_row)
        if (!is.null(w_new)) {
          weight_factor <- pmax(w_new, 0)
        } else if (all(w_nm %in% names(draw_row))) {
          weight_factor <- pmax(as.numeric(draw_row[w_nm]), 0)
        }
      }
      if (sum(weight_factor) <= 0) weight_factor <- rep(1 / K, K)
      like_k <- numeric(K)
      for (k in seq_len(K)) {
        like_k[k] <- .cluster_component_density(
          spec = spec,
          draw_row = draw_row,
          k = k,
          x_row = x_row,
          y_val = newdat$y[i],
          density_fun = density_fun
        )
      }
      post_k <- weight_factor * pmax(like_k, 0)
      if (!any(is.finite(post_k)) || sum(post_k) <= 0) {
        post_k <- weight_factor
      } else {
        post_k <- post_k / sum(post_k)
      }
      out_scores[i, ] <- out_scores[i, ] + as.numeric(post_k %*% map_comp_to_dahl)
    }
  }

  out_scores <- out_scores / nrow(draw_sub)
  rs <- rowSums(out_scores)
  rs[rs <= 0] <- 1
  out_scores <- out_scores / rs
  labels <- max.col(out_scores, ties.method = "first")

  list(
    labels = as.integer(labels),
    scores = out_scores,
    data = .cluster_data_frame_from_design(design = newdat, formula_meta = formula_meta),
    K = Kd,
    cache = tr
  )
}

run_cluster_mcmc <- function(bundle, ...) {
  stopifnot(inherits(bundle, "dpmixgpd_cluster_bundle"))
  base_fit <- run_mcmc_bundle_manual(bundle, ...)

  out <- list(
    call = match.call(),
    spec = bundle$spec,
    bundle = bundle,
    base_fit = base_fit,
    samples = base_fit$samples %||% (base_fit$mcmc %||% list())$samples,
    mcmc = base_fit$mcmc %||% list(),
    timing = utils::modifyList(
      base_fit$timing %||% list(),
      list(total = (base_fit$timing %||% list())$total %||%
             sum(unlist((base_fit$timing %||% list())[c("build", "compile", "mcmc")]), na.rm = TRUE))
    ),
    cache_env = new.env(parent = emptyenv()),
    psm = NULL,
    dahl = NULL
  )
  class(out) <- c("dpmixgpd_cluster_fit", "causalmixgpd_fit", "list")
  out
}

## S3 -------------------------------------------------------------------------

#' Predict labels or similarity matrices from a cluster fit
#'
#' Convert posterior draws from a `dpmixgpd_cluster_fit` object into either a representative
#' clustering or a posterior similarity matrix (PSM). This is the main post-processing step for
#' the cluster workflow after [dpmix.cluster()] or [dpmgpd.cluster()].
#'
#' @param object A fitted cluster object.
#' @param newdata Optional new data containing the response and predictors required by the original
#'   formula. New-data prediction is available only for `type = "label"`.
#' @param type Prediction target:
#'   \itemize{
#'     \item \code{"label"}: representative partition via
#'       Dahl's least-squares rule
#'     \item \code{"psm"}: posterior similarity matrix on the
#'       training sample
#'   }
#' @param burnin Number of initial posterior draws to discard.
#' @param thin Keep every `thin`-th posterior draw.
#' @param return_scores Logical; if `TRUE` and `type = "label"`, include the matrix of Dahl-cluster
#'   assignment scores.
#' @param psm_max_n Maximum training sample size allowed for `type = "psm"`.
#' @param ... Unused.
#'
#' @return A `dpmixgpd_cluster_labels` object when `type = "label"` or a
#'   `dpmixgpd_cluster_psm` object when `type = "psm"`.
#'
#' @details
#' Let \eqn{z_i^{(s)}} denote the latent cluster label for observation \eqn{i} at posterior draw
#' \eqn{s}. The posterior similarity matrix is
#' \deqn{
#' \mathrm{PSM}_{ij} = \Pr(z_i = z_j \mid y) \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}).
#' }
#' The returned label solution is the Dahl representative partition, obtained by choosing the draw
#' whose adjacency matrix is closest to the PSM in squared error.
#'
#' For `newdata`, the function combines draw-specific component weights and component densities to
#' produce posterior assignment scores relative to the representative training clusters. Returned
#' `newdata` label objects also carry the training labels and response data needed for comparative
#' `plot(..., type = "summary")` displays. A PSM is not defined for `newdata`, so `type = "psm"`
#' is restricted to the training sample.
#'
#' Computing the PSM is \eqn{O(n^2)} in the training sample size, so `psm_max_n` guards against
#' accidental large matrix allocations.
#'
#' @seealso [dpmix.cluster()], [dpmgpd.cluster()], [summary.dpmixgpd_cluster_fit()],
#'   [plot.dpmixgpd_cluster_fit()], [summary.dpmixgpd_cluster_labels()],
#'   [summary.dpmixgpd_cluster_psm()].
#' @family cluster workflow
#' @export
predict.dpmixgpd_cluster_fit <- function(object,
                                         newdata = NULL,
                                         type = c("label", "psm"),
                                         burnin = NULL,
                                         thin = NULL,
                                         return_scores = FALSE,
                                         psm_max_n = 2000L,
                                         ...) {
  stopifnot(inherits(object, "dpmixgpd_cluster_fit"))
  type <- match.arg(type)
  psm_max_n <- as.integer(psm_max_n)[1]
  if (!is.finite(psm_max_n) || psm_max_n < 1L) {
    stop("'psm_max_n' must be an integer >= 1.", call. = FALSE)
  }

  if (!is.null(newdata) && identical(type, "psm")) {
    stop("type='psm' is only available for training data (newdata=NULL).", call. = FALSE)
  }

  draw_mat <- .cluster_samples_to_matrix(object$samples)
  idx <- .cluster_draw_indices(nrow(draw_mat), burnin = burnin, thin = thin)
  draw_sub <- draw_mat[idx, , drop = FALSE]
  z_draws <- .cluster_extract_z_from_matrix(draw_sub)

  cache_key <- paste("cache", length(idx), burnin %||% 0L, thin %||% 1L, sep = "_")
  cache_env <- object$cache_env %||% new.env(parent = emptyenv())
  if (exists(cache_key, envir = cache_env, inherits = FALSE)) {
    tr <- get(cache_key, envir = cache_env, inherits = FALSE)
    if (is.null(tr$scores_train) && !is.null(tr$probs_train)) tr$scores_train <- tr$probs_train
  } else {
    n_train <- ncol(z_draws)
    if (identical(type, "psm") && n_train > psm_max_n) {
      stop(
        sprintf(
          "PSM is O(n^2): n=%d exceeds psm_max_n=%d. Increase 'psm_max_n' or use type='label'.",
          n_train,
          psm_max_n
        ),
        call. = FALSE
      )
    }
    psm <- compute_psm(z_draws)
    dahl <- dahl_labels(z_draws, psm)
    scores_train <- .cluster_compute_scores(z_draws, dahl$labels, psm)
    tr <- list(psm = psm, dahl = dahl, scores_train = scores_train)
    assign(cache_key, tr, envir = cache_env)
  }

  if (identical(type, "psm")) {
    n_train <- ncol(z_draws)
    if (n_train > psm_max_n) {
      stop(
        sprintf(
          "PSM is O(n^2): n=%d exceeds psm_max_n=%d. Increase 'psm_max_n' or use type='label'.",
          n_train,
          psm_max_n
        ),
        call. = FALSE
      )
    }
    out <- list(
      psm = tr$psm,
      labels = as.integer(tr$dahl$labels),
      components = as.integer(tr$dahl$K),
      draw_index = as.integer(tr$dahl$draw_index),
      burnin = as.integer(burnin %||% 0L),
      thin = as.integer(thin %||% 1L),
      psm_max_n = psm_max_n
    )
    class(out) <- c("dpmixgpd_cluster_psm", "list")
    return(out)
  }

  if (is.null(newdata)) {
    out <- list(
      labels = as.integer(tr$dahl$labels),
      components = as.integer(tr$dahl$K),
      data = .cluster_training_data_frame(object),
      source = "train",
      burnin = as.integer(burnin %||% 0L),
      thin = as.integer(thin %||% 1L)
    )
    if (isTRUE(return_scores)) out$scores <- tr$scores_train
    class(out) <- c("dpmixgpd_cluster_labels", "list")
    return(out)
  }

  pred <- predict_labels_newdata(fit = object, newdata = newdata, burnin = burnin, thin = thin)
  out <- list(
    labels = as.integer(pred$labels),
    components = as.integer(pred$K),
    data = pred$data,
    train_reference = .cluster_training_reference(object, labels = tr$dahl$labels),
    source = "newdata",
    burnin = as.integer(burnin %||% 0L),
    thin = as.integer(thin %||% 1L)
  )
  if (isTRUE(return_scores)) out$scores <- pred$scores
  class(out) <- c("dpmixgpd_cluster_labels", "list")
  out
}

skip_if_not_test_level("ci")

test_that("cluster wrappers fit and predict across type modes", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(123)
  dat <- data.frame(
    y = abs(stats::rnorm(24)) + 0.25,
    x1 = stats::rnorm(24),
    x2 = stats::runif(24)
  )

  for (tp in c("weights", "param", "both")) {
    fit <- dpmix.cluster(
      y ~ x1 + x2,
      data = dat,
      kernel = "normal",
      components = 4,
      type = tp,
      mcmc = mcmc_fast(seed = 10L)
    )
    expect_s3_class(fit, "dpmixgpd_cluster_fit")
    expect_true(is.list(fit$timing))
    expect_true(all(c("build", "compile", "mcmc", "cache_hit", "total") %in% names(fit$timing)))

    psm <- predict(fit, type = "psm")
    expect_s3_class(psm, "dpmixgpd_cluster_psm")
    expect_true(is.matrix(psm$psm))
    expect_equal(nrow(psm$psm), ncol(psm$psm))
    expect_equal(unname(diag(psm$psm)), rep(1, nrow(psm$psm)), tolerance = 1e-8)
    expect_true(all(psm$psm >= -1e-10 & psm$psm <= 1 + 1e-10))

    lbl_train <- predict(fit, type = "label")
    expect_s3_class(lbl_train, "dpmixgpd_cluster_labels")
    expect_equal(length(lbl_train$labels), nrow(dat))
    expect_false("scores" %in% names(lbl_train))

    lbl_train_scores <- predict(fit, type = "label", return_scores = TRUE)
    expect_true(is.matrix(lbl_train_scores$scores))
    expect_equal(rowSums(lbl_train_scores$scores), rep(1, nrow(dat)), tolerance = 1e-8)

    nd <- dat[1:6, c("y", "x1", "x2")]
    lbl_new <- predict(fit, newdata = nd, type = "label")
    expect_s3_class(lbl_new, "dpmixgpd_cluster_labels")
    expect_equal(length(lbl_new$labels), nrow(nd))
    expect_false("scores" %in% names(lbl_new))

    lbl_new_scores <- predict(fit, newdata = nd, type = "label", return_scores = TRUE)
    expect_true(is.matrix(lbl_new_scores$scores))
    expect_equal(rowSums(lbl_new_scores$scores), rep(1, nrow(nd)), tolerance = 1e-8)
    expect_true(is.list(lbl_new_scores$train_reference))
    expect_equal(lbl_new_scores$train_reference$labels, predict(fit, type = "label")$labels)
  }
})

test_that("validation and guard behavior for clustering types", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(99)
  dat <- data.frame(
    y = abs(stats::rnorm(30)) + 0.1,
    x1 = stats::rnorm(30),
    x2 = stats::runif(30)
  )

  expect_error(
    dpmix.cluster(y ~ 1, data = dat, kernel = "normal", type = "weights", components = 4, mcmc = mcmc_fast(seed = 1L)),
    "requires covariates"
  )
  expect_error(
    dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", type = "weights", mcmc = mcmc_fast(seed = 1L)),
    "explicit 'components'"
  )
  expect_error(
    dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", type = "both", mcmc = mcmc_fast(seed = 1L)),
    "explicit 'components'"
  )

  expect_warning(
    fit_param <- dpmix.cluster(y ~ 1, data = dat, kernel = "normal", type = "param", mcmc = mcmc_fast(seed = 2L)),
    "using default components"
  )
  expect_s3_class(fit_param, "dpmixgpd_cluster_fit")

  expect_error(
    predict(fit_param, type = "psm", psm_max_n = 10L),
    "PSM is O\\(n\\^2\\)"
  )
})

test_that("cluster link and priors overrides are applied to spec", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("build_cluster_bundle", mode = "function"))

  set.seed(111)
  dat <- data.frame(
    y = abs(stats::rnorm(20)) + 0.2,
    x1 = stats::rnorm(20),
    x2 = stats::runif(20)
  )

  b <- build_cluster_bundle(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    GPD = TRUE,
    type = "both",
    components = 4,
    link = list(
      bulk = list(mean = "identity"),
      gpd = list(tail_scale = list(link = "exp"))
    ),
    priors = list(
      bulk = list(sd = list(dist = "gamma", args = list(shape = 2, rate = 1))),
      gpd = list(
        tail_shape = list(dist = "normal", args = list(mean = 0, sd = 0.25)),
        tail_scale = list(dist = "normal", args = list(mean = 0, sd = 0.7))
      ),
      concentration = list(dist = "gamma", args = list(shape = 3, rate = 2))
    ),
    mcmc = mcmc_fast(seed = 7L)
  )

  expect_equal(b$spec$plan$bulk$mean$mode, "link")
  expect_equal(b$spec$plan$bulk$mean$link, "identity")
  expect_equal(b$spec$plan$bulk$sd$mode, "dist")
  expect_equal(b$spec$plan$bulk$sd$dist, "gamma")
  expect_equal(b$spec$plan$bulk$sd$args$shape, 2)
  expect_equal(b$spec$plan$gpd$tail_scale$mode, "link")
  expect_equal(b$spec$plan$gpd$tail_scale$link, "exp")
  expect_equal(b$spec$plan$gpd$tail_scale$beta_prior$dist, "normal")
  expect_equal(b$spec$plan$gpd$tail_scale$beta_prior$args$sd, 0.7)
  expect_equal(b$spec$plan$gpd$tail_shape$mode, "dist")
  expect_equal(b$spec$plan$gpd$tail_shape$dist, "normal")
  expect_equal(b$spec$plan$concentration$mode, "dist")
  expect_equal(b$spec$plan$concentration$dist, "gamma")
  expect_equal(b$spec$plan$concentration$args$shape, 3)
})

test_that("weights and both monitor gating while param does not", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(202)
  dat <- data.frame(
    y = abs(stats::rnorm(26)) + 0.2,
    x1 = stats::rnorm(26),
    x2 = stats::runif(26)
  )

  .to_mat <- function(smp) {
    if (inherits(smp, "mcmc.list")) {
      return(do.call(rbind, lapply(smp, as.matrix)))
    }
    as.matrix(smp)
  }

  fit_w <- dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", components = 4, type = "weights", mcmc = mcmc_fast(seed = 3L))
  fit_b <- dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", components = 4, type = "both", mcmc = mcmc_fast(seed = 4L))
  fit_p <- dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", components = 4, type = "param", mcmc = mcmc_fast(seed = 5L))

  cn_w <- colnames(.to_mat(fit_w$samples))
  cn_b <- colnames(.to_mat(fit_b$samples))
  cn_p <- colnames(.to_mat(fit_p$samples))

  expect_true(any(grepl("^eta\\[[0-9]+\\]$", cn_w)))
  expect_true(any(grepl("^B\\[[0-9]+,\\s*[0-9]+\\]$", cn_w)))
  expect_true(any(grepl("^eta\\[[0-9]+\\]$", cn_b)))
  expect_true(any(grepl("^B\\[[0-9]+,\\s*[0-9]+\\]$", cn_b)))
  expect_false(any(grepl("^eta\\[[0-9]+\\]$", cn_p)))
  expect_false(any(grepl("^B\\[[0-9]+,\\s*[0-9]+\\]$", cn_p)))
})

test_that("weights labels depend on x through gating scores", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmix.cluster", mode = "function"))

  set.seed(321)
  n <- 40
  x1 <- stats::rnorm(n)
  x2 <- stats::rnorm(n)
  y <- abs(stats::rnorm(n, mean = ifelse(x1 > 0, 2, 0.5), sd = 0.4)) + 0.1
  dat <- data.frame(y = y, x1 = x1, x2 = x2)

  fit <- dpmix.cluster(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    components = 4,
    type = "weights",
    mcmc = mcmc_fast(seed = 6L)
  )

  nd <- data.frame(
    y = c(stats::median(y), stats::median(y)),
    x1 = c(-2, 2),
    x2 = c(-1, 1)
  )
  pred <- predict(fit, newdata = nd, type = "label", return_scores = TRUE)
  expect_true(is.matrix(pred$scores))
  expect_gt(sum(abs(pred$scores[1, ] - pred$scores[2, ])), 1e-6)
})

test_that("component-level beta extraction is preferred when present", {
  skip_if_not_test_level("ci")
  fn <- getFromNamespace(".cluster_extract_beta_auto", "CausalMixGPD")

  draw_row <- c(
    "beta_tail_scale[1,1]" = 2.0,
    "beta_tail_scale[1,2]" = -1.0,
    "beta_tail_scale[2,1]" = 0.5,
    "beta_tail_scale[2,2]" = 0.25,
    "beta_tail_scale[1]" = 99.0,
    "beta_tail_scale[2]" = 98.0
  )

  b1 <- fn(draw_row = draw_row, base = "beta_tail_scale", comp = 1L, P = 2L)
  b2 <- fn(draw_row = draw_row, base = "beta_tail_scale", comp = 2L, P = 2L)
  expect_equal(as.numeric(b1), c(2.0, -1.0), tolerance = 1e-12)
  expect_equal(as.numeric(b2), c(0.5, 0.25), tolerance = 1e-12)
})

test_that("dpmgpd.cluster supports label and psm prediction", {
  skip_if_not_test_level("ci")
  skip_if_not(exists("dpmgpd.cluster", mode = "function"))

  set.seed(404)
  dat <- data.frame(
    y = abs(stats::rnorm(20)) + 0.5,
    x1 = stats::rnorm(20),
    x2 = stats::runif(20)
  )

  fit <- dpmgpd.cluster(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    components = 4,
    type = "weights",
    mcmc = mcmc_fast(seed = 20L)
  )
  expect_s3_class(fit, "dpmixgpd_cluster_fit")
  expect_s3_class(predict(fit, type = "psm"), "dpmixgpd_cluster_psm")
  expect_s3_class(predict(fit, type = "label"), "dpmixgpd_cluster_labels")
})

test_that("newdata unseen factor levels get explicit error", {
  fn <- getFromNamespace(".cluster_build_design", "CausalMixGPD")

  train <- data.frame(
    y = c(1, 2, 3),
    g = factor(c("a", "b", "a"))
  )
  trm <- stats::terms(y ~ g, data = train)
  mf <- stats::model.frame(trm, data = train)
  mm <- stats::model.matrix(stats::delete.response(trm), data = mf)
  X_cols <- setdiff(colnames(mm), "(Intercept)")
  meta <- list(
    terms = trm,
    xlevels = stats::.getXlevels(trm, mf),
    contrasts = attr(mm, "contrasts"),
    X_cols = X_cols,
    response = "y"
  )

  nd <- data.frame(
    y = 1,
    g = factor("c", levels = c("a", "b", "c"))
  )
  expect_error(
    fn(meta = meta, newdata = nd),
    "unseen factor levels"
  )
})

cluster_parse_formula <- getFromNamespace(".cluster_parse_formula", "CausalMixGPD")
cluster_default_link <- getFromNamespace(".cluster_default_link", "CausalMixGPD")
cluster_link_bulk_specs <- getFromNamespace(".cluster_link_bulk_specs", "CausalMixGPD")
cluster_default_beta_prior <- getFromNamespace(".cluster_default_beta_prior", "CausalMixGPD")
cluster_split_overrides <- getFromNamespace(".cluster_split_overrides", "CausalMixGPD")
cluster_normalize_link_entry <- getFromNamespace(".cluster_normalize_link_entry", "CausalMixGPD")
cluster_normalize_prior_entry <- getFromNamespace(".cluster_normalize_prior_entry", "CausalMixGPD")
cluster_apply_prior_overrides <- getFromNamespace(".cluster_apply_prior_overrides", "CausalMixGPD")
cluster_order_levels <- getFromNamespace(".cluster_order_levels", "CausalMixGPD")
cluster_size_table <- getFromNamespace(".cluster_size_table", "CausalMixGPD")
cluster_data_frame_from_design <- getFromNamespace(".cluster_data_frame_from_design", "CausalMixGPD")
cluster_training_reference <- getFromNamespace(".cluster_training_reference", "CausalMixGPD")
cluster_response_name <- getFromNamespace(".cluster_response_name", "CausalMixGPD")
cluster_response_values <- getFromNamespace(".cluster_response_values", "CausalMixGPD")
cluster_plot_ylim <- getFromNamespace(".cluster_plot_ylim", "CausalMixGPD")
cluster_scale_values <- getFromNamespace(".cluster_scale_values", "CausalMixGPD")
cluster_hist_breaks <- getFromNamespace(".cluster_hist_breaks", "CausalMixGPD")
cluster_hist_max_counts <- getFromNamespace(".cluster_hist_max_counts", "CausalMixGPD")
cluster_summary_plot_data <- getFromNamespace(".cluster_summary_plot_data", "CausalMixGPD")
cluster_profile_table <- getFromNamespace(".cluster_profile_table", "CausalMixGPD")
cluster_build_design <- getFromNamespace(".cluster_build_design", "CausalMixGPD")
cluster_softmax <- getFromNamespace(".cluster_softmax", "CausalMixGPD")
cluster_gating_weights <- getFromNamespace(".cluster_gating_weights", "CausalMixGPD")
cluster_resolve_density_fun <- getFromNamespace(".cluster_resolve_density_fun", "CausalMixGPD")
cluster_component_density <- getFromNamespace(".cluster_component_density", "CausalMixGPD")

test_that("cluster helper edge cases cover additional validation branches", {
  dat <- data.frame(
    y = c(1, 2, 3),
    id = c("a", "b", "c")
  )

  expect_error(cluster_parse_formula(~ id, data = dat), "Could not extract response")
  parsed_id <- cluster_parse_formula(y ~ id, data = dat)
  expect_equal(parsed_id$response, "y")

  expect_equal(cluster_default_link(list(bulk_support = list(sd = "positive_sd")), "sd"), "exp")
  expect_equal(cluster_default_link(list(bulk_support = list(mean = "real")), "mean"), "identity")
  expect_error(cluster_link_bulk_specs("not-a-kernel"), "Kernel 'not-a-kernel' not found")
  expect_equal(cluster_default_beta_prior("tail_scale")$args$sd, 0.5)
  expect_equal(cluster_default_beta_prior("tail_shape")$args$sd, 0.3)

  expect_error(cluster_split_overrides("bad", bulk_names = "mean", gpd_names = "tail_scale"), "supplied as a list")
  expect_error(cluster_split_overrides(list(1), bulk_names = "mean", gpd_names = "tail_scale"), "Unnamed override entries")
  empty_overrides <- cluster_split_overrides(list(), bulk_names = "mean", gpd_names = "tail_scale")
  expect_equal(empty_overrides$bulk, list())
  expect_equal(empty_overrides$gpd, list())
  expect_null(empty_overrides$concentration)

  expect_error(cluster_normalize_link_entry(list(mode = "fixed")), "mode='link'")
  expect_equal(cluster_normalize_prior_entry("gamma", current_mode = "link")$beta_prior$dist, "gamma")
  expect_equal(cluster_normalize_prior_entry(list(mode = "dist", dist = "gamma"), current_mode = "dist")$dist, "gamma")
  expect_equal(
    cluster_normalize_prior_entry(
      list(beta_prior = list(dist = "normal", args = list(sd = 1))),
      current_mode = "dist"
    )$mode,
    "link"
  )
  expect_equal(
    cluster_normalize_prior_entry(
      list(args = list(mean = 1)),
      current_mode = "dist"
    )$dist,
    "normal"
  )
  expect_equal(
    cluster_normalize_prior_entry(
      list(args = list(mean = 0)),
      current_mode = "link",
      param_name = "tail_scale"
    )$beta_prior$dist,
    "normal"
  )

  spec <- list(
    plan = list(
      bulk = list(mean = list(mode = "dist")),
      gpd = list(),
      concentration = list(mode = "dist", dist = "gamma")
    )
  )

  out_char <- cluster_apply_prior_overrides(spec, priors = list(concentration = "gamma"))
  expect_equal(out_char$plan$concentration$dist, "gamma")

  out_mode <- cluster_apply_prior_overrides(
    spec,
    priors = list(concentration = list(mode = "fixed", value = 2))
  )
  expect_equal(out_mode$plan$concentration$value, 2)

  out_dist <- cluster_apply_prior_overrides(
    spec,
    priors = list(concentration = list(args = list(shape = 2, rate = 3)))
  )
  expect_equal(out_dist$plan$concentration$mode, "dist")
  expect_equal(out_dist$plan$concentration$args$shape, 2)
  expect_equal(out_dist$plan$concentration$args$rate, 3)

  expect_error(
    cluster_apply_prior_overrides(spec, priors = list(concentration = TRUE)),
    "Invalid concentration prior override"
  )
})

test_that("cluster fallback matrix helpers run when fast-path helpers are unavailable", {
  z_draws <- matrix(
    c(
      1, 1, 2,
      1, 2, 2
    ),
    nrow = 2,
    byrow = TRUE
  )

  testthat::local_mocked_bindings(
    .compute_psm = 0,
    .dahl_representative = 0,
    .compute_cluster_probs = 0,
    .package = "CausalMixGPD"
  )

  psm <- compute_psm(z_draws)
  expect_equal(
    psm,
    matrix(
      c(
        1, 0.5, 0,
        0.5, 1, 0.5,
        0, 0.5, 1
      ),
      nrow = 3,
      byrow = TRUE
    )
  )

  dahl <- dahl_labels(z_draws, psm)
  expect_equal(dahl$draw_index, 1L)
  expect_equal(dahl$labels, c(1L, 1L, 2L))

  scores <- getFromNamespace(".cluster_compute_scores", "CausalMixGPD")(z_draws, dahl$labels, psm)
  expect_equal(rowSums(scores), rep(1, 3), tolerance = 1e-8)
  expect_gt(scores[1, 1], scores[1, 2])
})

test_that("cluster summary and profile helpers cover non-plot edge branches", {
  expect_equal(cluster_order_levels(c("10", "2", "x"), order_by = "label"), c("2", "10", "x"))

  tab <- cluster_size_table(c("a", "b", "a"), levels = c("b", "a", "c"))
  expect_equal(as.integer(tab), c(1L, 2L, 0L))

  expect_null(cluster_data_frame_from_design(NULL))

  design_df <- cluster_data_frame_from_design(
    design = list(y = c(1, 2), X = matrix(c(10, 20, 30, 40), nrow = 2)),
    formula_meta = list(response = "resp", X_cols = c("x1", "x2"))
  )
  expect_equal(names(design_df), c("resp", "x1", "x2"))

  fit_stub <- structure(
    list(
      bundle = list(data = NULL),
      spec = list(cluster = list(formula_meta = list()))
    ),
    class = c("dpmixgpd_cluster_fit", "list")
  )
  expect_null(cluster_training_reference(fit_stub, integer(0)))

  blank_name <- data.frame(y = c(1, 2), check.names = FALSE)
  names(blank_name) <- ""
  expect_equal(cluster_response_name(blank_name), "y")
  expect_equal(cluster_response_values(data.frame(a = 1:2), response_name = "missing"), numeric(0))
  expect_equal(cluster_plot_ylim(c(NA_real_, NA_real_)), c(0, 1))
  expect_length(cluster_scale_values(character(0)), 0L)
  expect_equal(cluster_hist_breaks(c(NA_real_, NA_real_)), c(0, 1))

  hist_counts <- cluster_hist_max_counts(
    data.frame(cluster = c("1", "1"), response = c(1, 2)),
    levels = c("1", "2"),
    breaks = seq(0, 2, by = 1)
  )
  expect_equal(unname(hist_counts), c(1, 0))

  lbl_no_data <- structure(
    list(labels = c("1", "2"), components = 2L, source = "train"),
    class = c("dpmixgpd_cluster_labels", "list")
  )
  expect_warning(expect_null(cluster_summary_plot_data(lbl_no_data)), "No attached data available")

  lbl_mismatch <- structure(
    list(labels = c("1", "2"), data = data.frame(y = 1), source = "train"),
    class = c("dpmixgpd_cluster_labels", "list")
  )
  expect_warning(expect_null(cluster_summary_plot_data(lbl_mismatch)), "same number of rows")

  lbl_nonfinite <- structure(
    list(labels = c("1", "2"), data = data.frame(y = c(NA_real_, NaN)), source = "train"),
    class = c("dpmixgpd_cluster_labels", "list")
  )
  expect_warning(expect_null(cluster_summary_plot_data(lbl_nonfinite)), "No finite response values")

  lbl_new <- structure(
    list(
      labels = c("10", "10", "3"),
      data = data.frame(y = c(2, 3, 4)),
      source = "newdata",
      train_reference = list(
        labels = c("3", "10", "10"),
        data = data.frame(y = c(1, 2, 3))
      )
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )
  summary_dat <- cluster_summary_plot_data(lbl_new)
  expect_false(summary_dat$has_reference)
  expect_equal(summary_dat$levels, c("10", "3"))
  expect_equal(cluster_summary_plot_data(lbl_new, top_n = 1L, order_by = "size")$levels, "10")
  expect_equal(cluster_summary_plot_data(lbl_new, top_n = 1L, order_by = "label")$levels, "3")

  expect_null(cluster_profile_table(NULL, labels = c(1L, 2L)))
  expect_null(cluster_profile_table(data.frame(y = numeric(0)), labels = integer(0)))
  expect_error(cluster_profile_table(data.frame(y = 1:2), labels = 1:3), "same number of rows")
  expect_error(cluster_profile_table(data.frame(y = 1:2), labels = 1:2, vars = "missing"), "Unknown profiling variables")
  expect_error(cluster_profile_table(data.frame(y = 1:2), labels = 1:2, top_n = 0), "integer >= 1")

  prof <- cluster_profile_table(
    data = data.frame(y = c(1, 2), grp = c("a", "b")),
    labels = c("2", "10"),
    score_mat = matrix(c(0.9, 0.1, 0.2, 0.8), nrow = 2, byrow = TRUE),
    top_n = 1L,
    order_by = "label",
    vars = c("y", "grp")
  )
  expect_equal(nrow(prof), 1L)
  expect_true(all(c("certainty_mean", "certainty_sd") %in% names(prof)))
})

test_that("cluster newdata builders and density helpers cover edge branches", {
  train <- data.frame(
    y = c(1, 2, 3),
    x1 = c(0, 1, 0),
    g = factor(c("a", "b", "a"))
  )
  trm <- stats::terms(y ~ x1 + g, data = train)
  mf <- stats::model.frame(trm, data = train)
  mm <- stats::model.matrix(stats::delete.response(trm), data = mf)
  meta <- list(
    terms = trm,
    xlevels = stats::.getXlevels(trm, mf),
    contrasts = attr(mm, "contrasts"),
    X_cols = setdiff(colnames(mm), "(Intercept)"),
    response = "y"
  )

  meta_no_x <- list(
    terms = stats::terms(y ~ 1, data = train),
    xlevels = list(),
    contrasts = NULL,
    X_cols = character(0),
    response = "y"
  )

  out_no_x <- cluster_build_design(meta_no_x, data.frame(y = c(1, 2)))
  expect_null(out_no_x$X)
  expect_equal(out_no_x$y, c(1, 2))

  expect_error(cluster_build_design(meta, data.frame(x1 = 1, g = "a")), "response column")
  expect_error(cluster_build_design(meta_no_x, data.frame(y = c(1, NA_real_))), "response contains NA")
  expect_error(cluster_build_design(meta, data.frame(y = 1)), "Failed to build model frame")

  meta_missing <- utils::modifyList(meta, list(X_cols = c(meta$X_cols, "x_missing")))
  expect_error(
    cluster_build_design(
      meta_missing,
      data.frame(y = 1, x1 = 0, g = factor("a", levels = c("a", "b")))
    ),
    "missing required predictors"
  )

  expect_equal(cluster_softmax(c(Inf, Inf)), c(0.5, 0.5))

  testthat::local_mocked_bindings(
    .cluster_softmax = function(logits) 1,
    .package = "CausalMixGPD"
  )
  expect_null(cluster_gating_weights(list(eta = 0, B = matrix(1, nrow = 1)), x_row = 1))

  expect_error(cluster_resolve_density_fun(list(meta = list(kernel = "nope", GPD = FALSE))), "Kernel 'nope' not found")

  testthat::local_mocked_bindings(
    get_kernel_registry = function() {
      list(dummy = list(crp = list(d_base = NA_character_, d_gpd = NA_character_)))
    },
    .package = "CausalMixGPD"
  )
  expect_error(cluster_resolve_density_fun(list(meta = list(kernel = "dummy", GPD = FALSE))), "Could not resolve density function")

  testthat::local_mocked_bindings(
    get_kernel_registry = function() {
      list(dummy = list(crp = list(d_base = "definitely_not_a_density", d_gpd = "definitely_not_a_density")))
    },
    .package = "CausalMixGPD"
  )
  expect_error(cluster_resolve_density_fun(list(meta = list(kernel = "dummy", GPD = FALSE))), "is unavailable")

  spec_gpd <- list(
    meta = list(GPD = TRUE, P = 1L),
    plan = list(
      bulk = list(
        mean = list(mode = "link", link = "identity"),
        sd = list(mode = "fixed", value = 1)
      ),
      gpd = list(
        threshold = list(mode = "fixed", value = 0.5),
        tail_scale = list(mode = "link", link = "exp"),
        tail_shape = list(mode = "dist")
      )
    ),
    signatures = list(gpd = list(args = c("mean", "sd", "threshold", "tail_scale", "tail_shape")))
  )

  dens_fun <- function(x, mean, sd, threshold, tail_scale, tail_shape, log = FALSE) {
    out <- x + mean + sd + threshold + tail_scale + tail_shape
    if (isTRUE(log)) log(out) else out
  }

  dens_link <- cluster_component_density(
    spec = spec_gpd,
    draw_row = c(
      "beta_mean[1,1]" = 0.2,
      "sd[1]" = 1.1,
      "beta_tail_scale[1]" = 0,
      "tail_shape" = 0.3
    ),
    k = 1L,
    x_row = 2,
    y_val = 1,
    density_fun = dens_fun
  )
  expect_true(is.finite(dens_link))
  expect_gt(dens_link, 0)

  dens_indexed <- cluster_component_density(
    spec = spec_gpd,
    draw_row = c(
      "beta_mean[1,1]" = 0,
      "sd[1]" = 1,
      "beta_tail_scale[1]" = 0,
      "tail_shape[1]" = 0.25
    ),
    k = 1L,
    x_row = 1,
    y_val = 1,
    density_fun = function(...) -1
  )
  expect_equal(dens_indexed, 0)

  dens_missing <- cluster_component_density(
    spec = spec_gpd,
    draw_row = c(
      "beta_mean[1,1]" = 0,
      "sd[1]" = 1,
      "beta_tail_scale[1]" = 0
    ),
    k = 1L,
    x_row = 1,
    y_val = 1,
    density_fun = function(...) stop("boom")
  )
  expect_equal(dens_missing, 0)
})

test_that("cluster prediction helpers cover cache fallbacks and actual run assembly", {
  meta_no_x <- list(
    terms = stats::terms(y ~ 1, data = data.frame(y = 1)),
    xlevels = list(),
    contrasts = NULL,
    X_cols = character(0),
    response = "y"
  )

  cache_env <- new.env(parent = emptyenv())
  assign(
    "cache_1_0_1",
    list(
      dahl = list(labels = c(1L, 2L), K = 2L),
      probs_train = matrix(c(0.9, 0.1, 0.2, 0.8), nrow = 2, byrow = TRUE)
    ),
    envir = cache_env
  )

  fit_newdata <- structure(
    list(
      spec = list(
        meta = list(kernel = "normal", GPD = FALSE, components = 3L),
        cluster = list(type = "weights", formula_meta = meta_no_x)
      ),
      samples = "fake_samples",
      cache_env = cache_env
    ),
    class = c("dpmixgpd_cluster_fit", "list")
  )

  testthat::local_mocked_bindings(
    .cluster_samples_to_matrix = function(samples) {
      matrix(
        c(0, 0, 0, 0, 0),
        nrow = 1,
        dimnames = list(NULL, c("z[1]", "z[2]", "w[1]", "w[2]", "w[3]"))
      )
    },
    .cluster_draw_indices = function(n_draws, burnin = NULL, thin = NULL) 1L,
    .cluster_extract_z_from_matrix = function(draw_sub) matrix(c(0L, 0L), nrow = 1),
    .cluster_resolve_density_fun = function(spec) function(...) 1,
    .cluster_component_density = function(...) NA_real_,
    .package = "CausalMixGPD"
  )

  pred <- predict_labels_newdata(fit_newdata, newdata = data.frame(y = c(1, 2)))
  expect_equal(rowSums(pred$scores), rep(1, 2), tolerance = 1e-8)

  expect_equal(pred$cache$scores_train, pred$cache$probs_train)

  bundle <- structure(list(spec = list(meta = list(kernel = "normal"))), class = c("dpmixgpd_cluster_bundle", "list"))

  testthat::local_mocked_bindings(
    run_mcmc_bundle_manual = function(bundle, ...) {
      list(
        samples = NULL,
        mcmc = list(samples = "chain"),
        timing = list(build = 0.1, compile = 0.2, mcmc = 0.3)
      )
    },
    .package = "CausalMixGPD"
  )

  fit_run <- run_cluster_mcmc(bundle)
  expect_s3_class(fit_run, "dpmixgpd_cluster_fit")
  expect_equal(fit_run$samples, "chain")
  expect_equal(fit_run$timing$total, 0.6, tolerance = 1e-8)

  fit_cached <- structure(
    list(
      spec = list(meta = list()),
      samples = "fake_samples",
      cache_env = new.env(parent = emptyenv())
    ),
    class = c("dpmixgpd_cluster_fit", "list")
  )
  assign(
    "cache_1_0_1",
    list(
      psm = diag(2),
      dahl = list(labels = c(1L, 2L), K = 2L, draw_index = 1L),
      probs_train = matrix(c(0.9, 0.1, 0.2, 0.8), nrow = 2, byrow = TRUE)
    ),
    envir = fit_cached$cache_env
  )

  testthat::local_mocked_bindings(
    .cluster_samples_to_matrix = function(samples) {
      matrix(c(1, 2), nrow = 1, dimnames = list(NULL, c("z[1]", "z[2]")))
    },
    .cluster_draw_indices = function(n_draws, burnin = NULL, thin = NULL) 1L,
    .cluster_extract_z_from_matrix = function(draw_sub) matrix(c(1L, 2L), nrow = 1),
    .cluster_training_data_frame = function(object) data.frame(y = c(1, 2)),
    .package = "CausalMixGPD"
  )

  lbl <- predict(fit_cached, type = "label", return_scores = TRUE)
  expect_true(is.matrix(lbl$scores))
  expect_equal(lbl$scores, matrix(c(0.9, 0.1, 0.2, 0.8), nrow = 2, byrow = TRUE))

  fit_guard <- structure(
    list(
      spec = list(meta = list()),
      samples = "fake_samples",
      cache_env = new.env(parent = emptyenv())
    ),
    class = c("dpmixgpd_cluster_fit", "list")
  )

  testthat::local_mocked_bindings(
    .cluster_samples_to_matrix = function(samples) {
      matrix(c(1, 1, 2), nrow = 1, dimnames = list(NULL, c("z[1]", "z[2]", "z[3]")))
    },
    .cluster_draw_indices = function(n_draws, burnin = NULL, thin = NULL) 1L,
    .cluster_extract_z_from_matrix = function(draw_sub) matrix(c(1L, 1L, 2L), nrow = 1),
    .package = "CausalMixGPD"
  )

  expect_error(predict(fit_guard, type = "psm", psm_max_n = 2L), "PSM is O\\(n\\^2\\)")
})

#' Simulate positive bulk-tail data
#'
#' Generate synthetic outcomes with a light-to-moderate bulk and a heavier upper tail.
#' The sample is assembled from a lognormal-gamma bulk and a shifted tail sample, then sorted.
#' This generator is intended for examples, help pages, and workflow checks
#' rather than as a formal generative model matching the full package
#' hierarchy exactly.
#'
#' @param n Integer sample size.
#' @param tail_prob Approximate tail probability \eqn{\Pr(X > u)} used to split the sample into
#'   bulk and tail draws.
#' @param seed Optional random seed for reproducibility.
#'
#' @return Numeric vector of length `n` containing positive outcomes sorted in ascending order.
#'
#' @details
#' The generator approximates a spliced sample
#' \deqn{
#' X \sim (1 - \pi_u) F_{bulk} + \pi_u F_{tail},
#' }
#' where \eqn{\pi_u =} `tail_prob`. The bulk component is itself a simple two-component mixture,
#' while the tail component is a shifted positive distribution that produces larger values.
#'
#' Use this helper when you need a fast toy sample for [bundle()], [dpmix()], or [dpmgpd()].
#' It should not be interpreted as posterior predictive simulation from a fitted object.
#'
#' @seealso [sim_causal_qte()], [sim_survival_tail()], [bundle()], [dpmgpd()].
#' @family simulation helpers
#' @importFrom stats rbinom rexp
#' @export
sim_bulk_tail <- function(n = 200, tail_prob = 0.12, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  bulk_n <- ceiling(n * (1 - tail_prob))
  tail_n <- n - bulk_n
  bulk <- c(
    rlnorm(bulk_n %/% 2, meanlog = 0, sdlog = 0.5),
    rgamma(bulk_n - bulk_n %/% 2, shape = 2, scale = 1.5)
  )
  tail <- stats::rgamma(tail_n, shape = 0.9, scale = 5) + 5
  sort(c(bulk, tail))
}

#' Simulate causal quantile-treatment-effect data
#'
#' Generate a treatment indicator, covariates, and a continuous outcome with both location and
#' tail heterogeneity. The resulting structure is intended for examples involving
#' [dpmix.causal()], [dpmgpd.causal()], [qte()], and [cqte()].
#'
#' @param n Integer sample size.
#' @param seed Optional random seed.
#'
#' @return List with components `y`, `t`, and `X`; `A` is included as a backward-compatible alias
#'   for `t`.
#'
#' @details
#' Treatment assignment is generated from a logistic propensity score
#' \deqn{
#' \Pr(T = 1 \mid X) = \operatorname{logit}^{-1}(\eta(X)),
#' }
#' and the observed outcome combines baseline covariate effects, an average treatment shift, and
#' a covariate-dependent tail amplification for treated units. This produces data where marginal
#' and conditional quantile effects differ across the outcome distribution.
#'
#' The returned list can be converted directly into the arguments expected by the causal fitting
#' wrappers after minor formatting.
#'
#' @seealso [sim_bulk_tail()], [dpmgpd.causal()], [qte()], [cqte()].
#' @family simulation helpers
#' @export
sim_causal_qte <- function(n = 300, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  X <- data.frame(
    x1 = rnorm(n, 0, 1),
    x2 = runif(n, -1, 1),
    x3 = rnorm(n)
  )
  lin_ps <- 0.2 + 0.6 * X$x1 - 0.4 * X$x2
  A <- rbinom(n, 1, stats::plogis(lin_ps))
  y_base <- 2 + 0.5 * X$x1 + 0.8 * X$x3 + rnorm(n)
  tail_effect <- 1.5 * X$x2 * (A == 1)
  y <- y_base + tail_effect + 2 * A + rexp(n, rate = 0.5)
  list(y = y, t = A, X = X, A = A)
}

#' Simulate censored survival-style tail data
#'
#' Generate event times, censoring times, an event indicator, and covariates for examples where
#' right tail behavior and positive support matter.
#'
#' @param n Integer sample size.
#' @param seed Optional random seed.
#'
#' @return Data frame containing observed time `time`, event indicator `status`, and covariates.
#'
#' @details
#' Event times are sampled from an exponential model with covariate-dependent mean, then censored
#' by an independent uniform censoring time. The observed time is
#' \deqn{
#' \tilde{T} = \min(T, C), \qquad \Delta = I(T \le C).
#' }
#'
#' This helper is mainly for experimentation and stress-testing positive-support kernels; it does
#' not implement a dedicated survival model from the package API.
#'
#' @seealso [sim_bulk_tail()], [build_nimble_bundle()], [dpmgpd()].
#' @family simulation helpers
#' @export
sim_survival_tail <- function(n = 250, seed = NULL) {
  if (!is.null(seed)) set.seed(seed)
  X <- data.frame(
    x1 = rnorm(n),
    x2 = rbinom(n, 1, 0.4)
  )
  base <- exp(3 - 0.5 * X$x1 + 0.3 * X$x2)
  time <- stats::rexp(n, rate = 1 / base)
  censor <- stats::runif(n, 0, 10)
  status <- as.integer(time <= censor)
  data.frame(time = pmin(time, censor), status = status, X)
}









#' Real-Line Bulk-Only Data Set With Two Components
#'
#' Real-line, bulk-only mixture dataset with K=2 components and no covariates.
#' Intended for non-causal bulk-only vignettes (normal/laplace/cauchy, GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{NULL.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_real200_k2$y)
#' @keywords datasets
"nc_real200_k2"

#' Positive-Support Bulk-Only Data Set With Three Components
#'
#' Positive-support, bulk-only mixture dataset with K=3 components and no covariates.
#' Intended for non-causal bulk-only positive-kernel vignettes (gamma/lognormal/invgauss/amoroso).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{NULL.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_pos200_k3$y)
#' @keywords datasets
"nc_pos200_k3"

#' Positive-Support Tail Data Set With Four Components
#'
#' Positive-support, tail-designed mixture dataset with K=4 components and no covariates.
#' Intended for GPD vignettes (gamma/lognormal/invgauss/amoroso with GPD=TRUE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{NULL.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_pos_tail200_k4$y)
#' @keywords datasets
"nc_pos_tail200_k4"

#' Positive-Support Covariate Data Set With Three Predictors
#'
#' Positive-support dataset with covariates (p=3) and K=2 mixture components.
#' Intended for covariate and prediction vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{data.frame with x1-x3.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_posX100_p3_k2$X)
#' @keywords datasets
"nc_posX100_p3_k2"

#' Positive-Support Covariate Data Set With Four Predictors
#'
#' Positive-support dataset with covariates (p=4) and K=3 mixture components.
#' Intended for covariate and prediction vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{data.frame with x1-x4.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_posX100_p4_k3$X)
#' @keywords datasets
"nc_posX100_p4_k3"

#' Positive-Support Covariate Data Set With Five Predictors
#'
#' Positive-support dataset with covariates (p=5) and K=4 mixture components.
#' Intended for covariate and prediction vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{data.frame with x1-x5.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_posX100_p5_k4$X)
#' @keywords datasets
"nc_posX100_p5_k4"

#' Real-Line Covariate Data Set With Three Predictors
#'
#' Real-line dataset with covariates (p=3) and K=2 mixture components.
#' Intended for covariate and prediction vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{data.frame with x1-x3.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_realX100_p3_k2$X)
#' @keywords datasets
"nc_realX100_p3_k2"

#' Real-Line Covariate Data Set With Five Predictors
#'
#' Real-line dataset with covariates (p=5) and K=3 mixture components.
#' Intended for covariate and prediction vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{X}{data.frame with x1-x5.}
#'   \item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
#'   \item{truth}{List with kernel, weights, params, threshold, tail_params.}
#' }
#' @examples
#' head(nc_realX100_p5_k3$X)
#' @keywords datasets
"nc_realX100_p5_k3"

#' Positive-Support Causal Data Set With Three Predictors
#'
#' Causal dataset (N=500, p=3) with the same positive-support kernel for both arms.
#' Intended for same-kernel causal baselines (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{A}{Binary treatment indicator (0/1).}
#'   \item{X}{data.frame with x1-x3.}
#'   \item{meta}{List with N, support, p, K0, K1, tail, exceed_frac.}
#'   \item{truth}{List with kernel0, kernel1, params0, params1, tail_params.}
#' }
#' @examples
#' head(causal_pos500_p3_k2$X)
#' @keywords datasets
"causal_pos500_p3_k2"

#' Alternative Positive-Support Causal Data Set With Three Predictors
#'
#' Causal dataset (N=500, p=3) with different positive-support kernels by arm.
#' Intended for alternating-kernel causal vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{A}{Binary treatment indicator (0/1).}
#'   \item{X}{data.frame with x1-x3.}
#'   \item{meta}{List with N, support, p, K0, K1, tail, exceed_frac.}
#'   \item{truth}{List with kernel0, kernel1, params0, params1, tail_params.}
#' }
#' @examples
#' head(causal_alt_pos500_p3_k3$X)
#' @keywords datasets
"causal_alt_pos500_p3_k3"

#' Alternative Real-Line Causal Data Set With Four Predictors
#'
#' Causal dataset (N=500, p=4) with different real-line kernels by arm.
#' Intended for alternating-kernel causal vignettes (GPD=FALSE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{A}{Binary treatment indicator (0/1).}
#'   \item{X}{data.frame with x1-x4.}
#'   \item{meta}{List with N, support, p, K0, K1, tail, exceed_frac.}
#'   \item{truth}{List with kernel0, kernel1, params0, params1, tail_params.}
#' }
#' @examples
#' head(causal_alt_real500_p4_k2$X)
#' @keywords datasets
"causal_alt_real500_p4_k2"

#' Alternative Positive-Support Tail Causal Data Set With Five Predictors
#'
#' Causal dataset (N=500, p=5) with different positive-support kernels by arm
#' and tail-designed exceedances (GPD=TRUE).
#'
#' @format A list with:
#' \describe{
#'   \item{y}{Numeric outcome vector.}
#'   \item{A}{Binary treatment indicator (0/1).}
#'   \item{X}{data.frame with x1-x5.}
#'   \item{meta}{List with N, support, p, K0, K1, tail, exceed_frac.}
#'   \item{truth}{List with kernel0, kernel1, params0, params1, tail_params.}
#' }
#' @examples
#' head(causal_alt_pos500_p5_k4_tail$X)
#' @keywords datasets
"causal_alt_pos500_p5_k4_tail"

#' Posterior Summary Utilities (internal)
#'
#' Helper functions for summarizing posterior draws: compute credible/HPD intervals,
#' format fit headers, and create summary tables with mean/sd/quantiles.
#'
#' @name summary-utils
#' @keywords internal
#' @noRd
NULL

#' Compute credible or HPD interval for draws
#'
#' @param draws Numeric vector of posterior draws.
#' @param level Numeric in (0, 1); credible level (default 0.95).
#' @param type Character; interval type:
#'   \itemize{
#'     \item \code{"credible"}: equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @return Named numeric vector with \code{lower} and \code{upper}.
#' @keywords internal
#' @noRd
.compute_interval <- function(draws, level = 0.95, type = c("credible", "hpd")) {
  type <- match.arg(type)
  draws <- draws[is.finite(draws)]
  if (length(draws) < 2L) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  if (type == "credible") {
    probs <- c((1 - level) / 2, (1 + level) / 2)
    q <- stats::quantile(draws, probs = probs, na.rm = TRUE)
    c(lower = unname(q[1]), upper = unname(q[2]))
  } else {
    # HPD via coda
    if (!requireNamespace("coda", quietly = TRUE)) {
      stop("Package 'coda' is required for HPD intervals.", call. = FALSE)
    }
    hpd <- coda::HPDinterval(coda::as.mcmc(draws), prob = level)
    c(lower = hpd[1, "lower"], upper = hpd[1, "upper"])
  }
}

#' Summarize posterior draws (mean + quantiles)
#'
#' @param draws Numeric vector, matrix, or array with draws in last dimension.
#' @param probs Numeric quantile probs.
#' @param interval Character or NULL; interval type:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile
#'       intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @return List with estimate, lower, upper, and q.
#' @keywords internal
#' @noRd
.posterior_summarize <- function(draws, probs = c(0.025, 0.5, 0.975),
                                 interval = "credible") {
  # Handle NULL interval (no interval computation)
  if (is.null(interval)) {
    interval <- "none"
  } else {
    interval <- match.arg(interval, choices = c("credible", "hpd"))
  }
  probs <- as.numeric(probs)

  # Compute credible level from probs (for HPD)
  level <- probs[length(probs)] - probs[1]
  if (!is.finite(level) || level <= 0 || level >= 1) level <- 0.95

  # Helper to compute intervals for a row of draws
  .row_interval <- function(row) {
    if (interval == "none") {
      c(NA_real_, NA_real_)
    } else if (interval == "credible") {
      q <- stats::quantile(row, probs = probs, na.rm = TRUE, names = FALSE)
      c(q[1], q[length(probs)])
    } else {
      iv <- .compute_interval(row, level = level, type = "hpd")
      c(iv["lower"], iv["upper"])
    }
  }

  if (is.null(dim(draws))) {
    mat <- matrix(as.numeric(draws), nrow = 1)
    qmat <- t(apply(mat, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
    iv <- .row_interval(as.numeric(draws))
    return(list(
      estimate = rowMeans(mat, na.rm = TRUE),
      lower = iv[1],
      upper = iv[2],
      q = qmat
    ))
  }

  dims <- dim(draws)
  if (length(dims) == 2L) {
    qmat <- t(apply(draws, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
    ivmat <- t(apply(draws, 1, .row_interval))
    return(list(
      estimate = rowMeans(draws, na.rm = TRUE),
      lower = ivmat[, 1],
      upper = ivmat[, 2],
      q = qmat
    ))
  }

  Sdim <- dims[length(dims)]
  mat <- matrix(draws, nrow = prod(dims[-length(dims)]), ncol = Sdim)
  qmat <- t(apply(mat, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
  ivmat <- t(apply(mat, 1, .row_interval))
  estimate <- rowMeans(mat, na.rm = TRUE)
  lower <- ivmat[, 1]
  upper <- ivmat[, 2]
  dim(estimate) <- dims[-length(dims)]
  dim(lower) <- dims[-length(dims)]
  dim(upper) <- dims[-length(dims)]
  list(estimate = estimate, lower = lower, upper = upper, q = qmat)
}

#' Format a short header for printing
#'
#' @param x A mixgpd_fit.
#' @return Character vector lines.
#' @keywords internal
#' @noRd
.format_fit_header <- function(x) {
  spec <- x$spec %||% list()
  meta <- spec$meta %||% list()

  backend <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  kernel  <- meta$kernel  %||% spec$kernel$key %||% "<unknown>"

  gpd_val <- meta$GPD %||% meta$gpd %||% spec$dispatch$GPD
  gpd_txt <- if (isTRUE(gpd_val)) "TRUE" else if (identical(gpd_val, FALSE)) "FALSE" else "<unknown>"

  y <- x$data$y %||% x$y %||% NULL
  n <- if (!is.null(y)) length(y) else (meta$N %||% spec$N %||% NA_integer_)
  Kmax <- meta$Kmax %||% spec$Kmax %||% NA_integer_

  eps <- .get_epsilon(x, epsilon = NULL)
  lines <- c(
    sprintf("MixGPD fit | backend: %s | kernel: %s | GPD tail: %s",
            .backend_label(backend), .kernel_label(kernel), gpd_txt),
    sprintf("n = %s | components = %s | epsilon = %s",
            ifelse(is.na(n), "<unknown>", n),
            ifelse(is.na(meta$components %||% NA_integer_), "<unknown>", meta$components),
            ifelse(is.na(eps), "<unknown>", eps))
  )

  m <- x$mcmc %||% list()
  it <- m$niter %||% NA_integer_
  nb <- m$nburnin %||% NA_integer_
  th <- m$thin %||% NA_integer_
  ch <- m$nchains %||% NA_integer_
  if (!all(is.na(c(it, nb, th, ch)))) {
    lines <- c(lines, sprintf("MCMC: niter=%s, nburnin=%s, thin=%s, nchains=%s",
                              ifelse(is.na(it), "?", it),
                              ifelse(is.na(nb), "?", nb),
                              ifelse(is.na(th), "?", th),
                              ifelse(is.na(ch), "?", ch)))
  }

  lines
}

#' Summarize posterior draws for selected parameters
#'
#' @param object mixgpd_fit
#' @param pars character vector; if NULL uses all non-v parameters
#' @param probs quantiles to report
#' @return data.frame with mean/sd/quantiles + ess/rhat where available
#' @keywords internal
#' @noRd
.summarize_posterior <- function(object, pars = NULL, probs = c(0.025, 0.5, 0.975)) {
  stopifnot(inherits(object, "mixgpd_fit"))

  if (!requireNamespace("coda", quietly = TRUE)) stop("Need 'coda'.", call. = FALSE)

  mat <- .extract_draws(object, pars = NULL, chains = "stack", epsilon = NULL)
  eps <- .get_epsilon(object, epsilon = NULL)

  if (is.null(pars)) {
    pars <- colnames(mat)

    spec <- object$spec %||% list()
    plan <- spec$plan %||% list()
    bulk <- plan$bulk %||% list()
    gpd <- plan$gpd %||% list()

    cn <- pars
    keep <- cn %in% "alpha"
    keep <- keep | grepl("^w\\[[0-9]+\\]$", cn)

    for (nm in names(bulk)) {
      ent <- bulk[[nm]] %||% list()
      mode <- ent$mode %||% NA_character_
      if (identical(mode, "link")) {
        keep <- keep | grepl(paste0("^beta_", nm, "\\["), cn)
      } else {
        keep <- keep | grepl(paste0("^", nm, "\\[[0-9]+\\]$"), cn)
      }
    }

    if (!is.null(gpd$threshold)) {
      thr_mode <- gpd$threshold$mode %||% NA_character_
      if (identical(thr_mode, "link")) {
        keep <- keep | grepl("^beta_threshold\\[", cn)
        if (!is.null(gpd$threshold$link_dist) &&
            identical(gpd$threshold$link_dist$dist, "lognormal")) {
          keep <- keep | cn == "sdlog_u"
        }
      } else {
        keep <- keep | cn == "threshold" | grepl("^threshold\\[[0-9]+\\]$", cn)
      }
    }

    if (!is.null(gpd$tail_scale)) {
      ts_mode <- gpd$tail_scale$mode %||% NA_character_
      if (identical(ts_mode, "link")) {
        keep <- keep | grepl("^beta_tail_scale\\[", cn)
      } else if (ts_mode %in% c("dist", "fixed")) {
        keep <- keep | cn == "tail_scale"
      }
    }

    if (!is.null(gpd$tail_shape)) {
      keep <- keep | cn == "tail_shape"
    }

    pars <- cn[keep]
    mat <- mat[, pars, drop = FALSE]
  } else {
    pars <- gsub("^weight\\[", "w[", pars)
    .match_summary_pars <- function(tokens, all_params) {
      hits <- character(0)
      for (tok in tokens) {
        if (tok %in% all_params) {
          hits <- c(hits, tok)
          next
        }
        tok_esc <- gsub("([][{}()+*^$|\\\\?.])", "\\\\\\1", tok)
        pref <- grep(paste0("^", tok_esc, "(\\[|$)"), all_params, value = TRUE)
        if (!length(pref) && identical(tok, "weights")) {
          pref <- grep("^w\\[", all_params, value = TRUE)
        }
        if (!length(pref)) {
          stop("Unknown params: ", tok, call. = FALSE)
        }
        hits <- c(hits, pref)
      }
      unique(hits)
    }
    pars <- .match_summary_pars(pars, colnames(mat))
    mat <- mat[, pars, drop = FALSE]
  }
  wpars <- pars[grepl("^w\\[[0-9]+\\]$", pars)]
  if (length(wpars)) {
    pars <- c(wpars, setdiff(pars, wpars))
    mat <- mat[, pars, drop = FALSE]
  }
  if (length(wpars) && is.finite(eps) && eps > 0) {
    wmat <- mat[, wpars, drop = FALSE]
    wmat[wmat < eps] <- NA_real_
    keep_w <- apply(wmat, 2, function(v) any(is.finite(v)))
    w_keep <- wpars[keep_w]
    if (length(w_keep)) {
      mat[, w_keep] <- wmat[, w_keep, drop = FALSE]
      pars <- c(w_keep, setdiff(pars, wpars))
      mat <- mat[, pars, drop = FALSE]
    } else {
      pars <- setdiff(pars, wpars)
      mat <- mat[, pars, drop = FALSE]
    }
  }

  thr_cols <- grep("^threshold\\[[0-9]+\\]$", colnames(mat), value = TRUE)
  if (length(thr_cols) >= 1) {
    thr_vec <- if (length(thr_cols) == 1) {
      as.numeric(mat[, thr_cols[1]])
    } else {
      rowMeans(mat[, thr_cols, drop = FALSE], na.rm = TRUE)
    }
    mat <- mat[, setdiff(colnames(mat), thr_cols), drop = FALSE]
    mat <- cbind(mat, threshold = thr_vec)
    pars <- c(setdiff(pars, thr_cols), "threshold")
  }

  meanv <- colMeans(mat, na.rm = TRUE)
  sdv   <- apply(mat, 2, stats::sd, na.rm = TRUE)

  qmat <- t(apply(mat, 2, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
  colnames(qmat) <- paste0("q", formatC(probs, format = "f", digits = 3))

  out <- data.frame(
    parameter = pars,
    mean = as.numeric(meanv[pars]),
    sd   = as.numeric(sdv[pars]),
    qmat[pars, , drop = FALSE],
    stringsAsFactors = FALSE
  )

  ess_vec <- rep(NA_real_, ncol(mat))
  for (j in seq_len(ncol(mat))) {
    v <- mat[, j]
    v <- v[is.finite(v)]
    if (length(v) >= 3L) {
      ess_vec[j] <- as.numeric(coda::effectiveSize(coda::mcmc(v)))
    }
  }
  names(ess_vec) <- colnames(mat)
  out$ess <- as.numeric(ess_vec[out$parameter])
  out$parameter <- sub("^w\\[", "weights[", out$parameter)

  rownames(out) <- NULL
  out
}

# Formatting helpers for consistent numeric printing
# These are internal (not exported).

#' Format numbers to 3 decimals without trailing zeros
#' @param x numeric vector
#' @return character vector
#' @noRd
fmt3 <- function(x) {
  formatC(x, digits = 3, format = "f", drop0trailing = TRUE)
}

#' Format numbers to 3 decimals, switching to scientific for large values
#' @param x numeric vector
#' @param digits number of digits after decimal
#' @param big absolute value threshold to switch to scientific notation
#' @return character vector
#' @noRd
fmt3_sci <- function(x, digits = 3, big = 1e4) {
  out <- formatC(x, digits = digits, format = "f", drop0trailing = TRUE)
  if (!length(x)) return(out)
  big_idx <- is.finite(x) & abs(x) >= big
  if (any(big_idx)) {
    out[big_idx] <- formatC(x[big_idx], digits = digits, format = "e")
  }
  out
}

#' @noRd
fmt3_vec <- function(x) {
  if (!length(x)) return("")
  paste(fmt3(x), collapse = ", ")
}

#' @noRd
format_df3 <- function(df) {
  if (!is.data.frame(df)) return(df)
  num_cols <- vapply(df, is.numeric, logical(1))
  df[num_cols] <- lapply(df[num_cols], fmt3)
  df
}

#' @noRd
format_df3_sci <- function(df, digits = 3, big = 1e4) {
  if (!is.data.frame(df)) return(df)
  num_cols <- vapply(df, is.numeric, logical(1))
  df[num_cols] <- lapply(df[num_cols], fmt3_sci, digits = digits, big = big)
  df
}

#' @noRd
format_mat3 <- function(mat) {
  if (!is.matrix(mat)) return(mat)
  out <- apply(mat, 2, fmt3)
  dim(out) <- dim(mat)
  dimnames(out) <- dimnames(mat)
  out
}

#' @noRd
format_mat3_sci <- function(mat, digits = 3, big = 1e4) {
  if (!is.matrix(mat)) return(mat)
  out <- apply(mat, 2, fmt3_sci, digits = digits, big = big)
  dim(out) <- dim(mat)
  dimnames(out) <- dimnames(mat)
  out
}

#' @noRd
.is_knitr_output <- function() {
  isTRUE(getOption("knitr.in.progress")) &&
    requireNamespace("knitr", quietly = TRUE)
}

#' @noRd
.knitr_asis <- function(...) {
  if (!requireNamespace("knitr", quietly = TRUE)) return(NULL)
  pieces <- list(...)
  flat <- unlist(lapply(pieces, function(x) {
    if (is.null(x)) return(character(0))
    if (inherits(x, "knitr_kable")) return(as.character(x))
    if (is.character(x)) return(x)
    as.character(x)
  }), use.names = FALSE)
  knitr::asis_output(paste(flat, collapse = "\n"))
}

#' @noRd
.kable_fmt <- function() {
  if (!requireNamespace("knitr", quietly = TRUE)) return("markdown")
  if (knitr::is_latex_output()) return("latex")
  if (knitr::is_html_output()) return("html")
  "markdown"
}

#' @noRd
.kable_table <- function(df, row.names = TRUE) {
  if (!requireNamespace("knitr", quietly = TRUE)) return(NULL)
  kbl <- knitr::kable(df, align = "c", row.names = row.names, format = .kable_fmt())
  if (requireNamespace("kableExtra", quietly = TRUE)) {
    kbl <- kableExtra::kable_styling(kbl, full_width = FALSE, position = "center")
  }
  kbl
}

#' @noRd
.dt_view_table <- function(df, row.names = TRUE, digits = 3, min_rows = 10L, min_cols = 8L) {
  # Goal: still print a regular data.frame to the console, but if DT is installed
  # and we're in an interactive session (not knitr), also open an interactive view.
  if (!interactive()) return(invisible(NULL))
  if (isTRUE(getOption("knitr.in.progress"))) return(invisible(NULL))
  if (!requireNamespace("DT", quietly = TRUE)) return(invisible(NULL))

  df <- as.data.frame(df)
  if (NROW(df) < as.integer(min_rows) || NCOL(df) < as.integer(min_cols)) {
    return(invisible(NULL))
  }

  num_cols <- vapply(df, is.numeric, logical(1))
  dt <- DT::datatable(
    df,
    rownames = isTRUE(row.names),
    class = "compact stripe hover",
    options = list(
      pageLength = min(25L, NROW(df)),
      scrollX = TRUE,
      autoWidth = TRUE,
      columnDefs = list(list(className = "dt-center", targets = "_all"))
    )
  )
  if (any(num_cols)) {
    dt <- DT::formatRound(dt, columns = names(df)[num_cols], digits = digits)
  }

  # In RStudio this typically opens the Viewer; otherwise it may open a browser tab.
  withCallingHandlers(
    print(dt),
    warning = function(w) {
      # Log suppressed warnings for debugging without showing them as warnings
      .cmgpd_message("DT view warning (suppressed): ", conditionMessage(w))
      invokeRestart("muffleWarning")
    }
  )
  invisible(NULL)
}

#' @noRd
print_fmt3 <- function(x, ...) {
  args <- list(...)
  row_names <- if (!is.null(args$row.names)) args$row.names else TRUE
  if (is.data.frame(x)) {
    df_raw <- x
    df <- format_df3(x)
    if (.is_knitr_output() && isTRUE(getOption("causalmixgpd.knitr.kable", FALSE))) {
      kbl <- .kable_table(df, row.names = row_names)
      if (!is.null(kbl)) return(print(kbl))
    }
    .dt_view_table(df_raw, row.names = row_names, digits = 3)
    return(print(df, quote = FALSE, ...))
  }
  if (is.matrix(x)) {
    mat_raw <- x
    mat <- format_mat3(x)
    if (.is_knitr_output() && isTRUE(getOption("causalmixgpd.knitr.kable", FALSE))) {
      kbl <- .kable_table(as.data.frame(mat), row.names = row_names)
      if (!is.null(kbl)) return(print(kbl))
    }
    .dt_view_table(as.data.frame(mat_raw), row.names = row_names, digits = 3)
    return(print(noquote(mat), ...))
  }
  if (is.numeric(x)) {
    return(print(noquote(fmt3(x)), ...))
  }
  print(x, ...)
}

#' @noRd
print_fmt3_sci <- function(x, digits = 3, big = 1e4, ...) {
  args <- list(...)
  row_names <- if (!is.null(args$row.names)) args$row.names else TRUE
  if (is.data.frame(x)) {
    df_raw <- x
    df <- format_df3_sci(x, digits = digits, big = big)
    if (.is_knitr_output() && isTRUE(getOption("causalmixgpd.knitr.kable", FALSE))) {
      kbl <- .kable_table(df, row.names = row_names)
      if (!is.null(kbl)) return(print(kbl))
    }
    .dt_view_table(df_raw, row.names = row_names, digits = digits)
    return(print(df, quote = FALSE, ...))
  }
  if (is.matrix(x)) {
    mat_raw <- x
    mat <- format_mat3_sci(x, digits = digits, big = big)
    if (.is_knitr_output() && isTRUE(getOption("causalmixgpd.knitr.kable", FALSE))) {
      kbl <- .kable_table(as.data.frame(mat), row.names = row_names)
      if (!is.null(kbl)) return(print(kbl))
    }
    .dt_view_table(as.data.frame(mat_raw), row.names = row_names, digits = digits)
    return(print(noquote(mat), ...))
  }
  if (is.numeric(x)) {
    return(print(noquote(fmt3_sci(x, digits = digits, big = big)), ...))
  }
  print(x, ...)
}


#' Plot styling and theme helpers (internal)
#'
#' Internal helper functions for consistent plot styling across the package.
#' These functions provide default color palettes, themes, and plotly conversion
#' utilities.
#'
#' @name viz-theme
#' @keywords internal
#' @noRd
NULL

#' Get default color palette
#'
#' Returns a consistent color palette for plotting. Uses colorblind-friendly
#' Wong palette as base.
#'
#' @param n Integer, number of colors needed. If NULL or greater than 8,
#'   repeats the base palette.
#'
#' @return Character vector of hex colors.
#' @keywords internal
#' @noRd
.plot_palette <- function(n = 8L) {
  base <- c(
    "#0072B2", # blue
    "#D55E00", # vermillion
    "#009E73", # green
    "#CC79A7", # purple
    "#56B4E9", # sky blue
    "#E69F00", # orange
    "#000000", # black
    "#999999"  # gray
  )
  n <- as.integer(n %||% length(base))
  if (n <= length(base)) return(base[seq_len(n)])
  rep_len(base, n)
}

#' Get default ggplot2 theme
#'
#' Returns a minimal theme with package-specific styling.
#'
#' @return A ggplot2 theme object.
#' @keywords internal
#' @noRd
.plot_theme <- function() {
  ggplot2::theme_minimal(base_size = 16) +
    ggplot2::theme(
      panel.grid.minor = ggplot2::element_blank(),
      plot.title = ggplot2::element_text(face = "bold", size = 20),
      axis.title = ggplot2::element_text(size = 17),
      axis.text = ggplot2::element_text(size = 14),
      legend.title = ggplot2::element_text(size = 15),
      legend.text = ggplot2::element_text(size = 13),
      strip.text = ggplot2::element_text(size = 15),
      legend.position = "top"
    )
}

#' Strip fill scales from ggplot object
#'
#' Removes fill aesthetics scales from a ggplot object. Used to work around
#' plotly conversion issues with certain fill scales.
#'
#' @param p A ggplot object.
#' @return The modified ggplot object.
#' @keywords internal
#' @noRd
.strip_fill_scales <- function(p) {
  if (!inherits(p, "ggplot")) return(p)
  if (is.null(p$scales) || !length(p$scales$scales)) return(p)
  keep <- vapply(p$scales$scales, function(s) {
    !("fill" %in% (s$aesthetics %||% character()))
  }, logical(1))
  p$scales$scales <- p$scales$scales[keep]
  p
}

#' Safely convert ggplot to plotly
#'
#' Attempts to convert a ggplot to plotly, with fallback to stripping
#' fill scales if the initial conversion fails.
#'
#' @param p A ggplot object.
#' @return A plotly object or the original ggplot if conversion fails.
#' @keywords internal
#' @noRd
.safe_ggplotly <- function(p) {
  tryCatch(
    plotly::ggplotly(p),
    error = function(e) {
      p2 <- .strip_fill_scales(p)
      tryCatch(plotly::ggplotly(p2), error = function(e2) p2)
    }
  )
}

#' Wrap plots in plotly if requested
#'
#' Conditionally converts ggplot objects to plotly based on user option.
#' Respects `options(CausalMixGPD.plotly = TRUE)` setting. Handles both
#' single ggplot objects and lists of plots.
#'
#' @param p A ggplot object or list of ggplot objects.
#' @return Either plotly object(s) or the original plot(s).
#' @keywords internal
#' @noRd
.wrap_plotly <- function(p) {
  # Default to static output; opt in via options(CausalMixGPD.plotly = TRUE).
  if (isTRUE(getOption("CausalMixGPD.plotly", FALSE)) &&
      requireNamespace("plotly", quietly = TRUE)) {
    if (is.list(p) && !inherits(p, "ggplot")) {
      # List of plots - wrap each, preserve class
      result <- lapply(p, function(plt) {
        if (inherits(plt, "ggplot")) .safe_ggplotly(plt) else plt
      })
      # Preserve original class attributes
      class(result) <- class(p)
      result
    } else if (inherits(p, "ggplot")) {
      # Single ggplot - wrap it
      .safe_ggplotly(p)
    } else {
      # Not a ggplot, return as-is
      p
    }
  } else {
    # plotly not available - return original
    p
  }
}

#' Internal global declarations
#'
#' Declares global variables to satisfy R CMD check when using non-standard
#' evaluation or generated column names.
#'
#' @name globals
#' @keywords internal
#' @importFrom stats uniroot pgamma predict rgamma density quantile coef setNames fitted plogis qlogis residuals sd
#' @importFrom utils head
#' @import nimble ggplot2
NULL

utils::globalVariables(c(
  ".kernel_registry_env",
  ".kernel_registry_ready",
  "ddirch",
  "dcat",
  "dbeta",
  "inprod",
  "v",
  "pow",
  "ddexp",
  "pdexp",
  "rdexp",
  "components",
  "BULK_DECL_PLACEHOLDER",
  "BULK_BLOCK",
  "CONC_PLACEHOLDER",
  "GPD_BLOCK",
  "HASX_BETA_BLOCK",
  "HASX_DET_BLOCK",
  "LIKELIHOOD_BLOCK",
  "MIXING_BLOCK",
  "logit<-",
  "probit<-",
  "S",
  "S_lower",
  "S_upper",
  "arm",
  "box_y",
  "cluster",
  "count",
  "dataset",
  "max_prob",
  "stick_breaking",
  # ggplot2 NSE variables
  "Chain",
  "draw",
  "hover",
  "K",
  "label",
  "Parameter",
  "prob",
  "conc_plan",
  "estimate",
  "fitted",
  "group",
  "id",
  "index",
  "kernel",
  "lower",
  "observed",
  "plogis",
  "ps",
  "qlogis",
  "response",
  "residuals",
  "sd",
  "size",
  "survival",
  "type",
  "upper",
  "value",
  "x",
  "x_plot",
  "y",
  "z",
  "measure"
))
#' Package hooks
#'
#' Internal package initialization.
#'
#' @details
#' The load hook initializes package-wide defaults that should exist as soon as
#' the namespace is attached. In particular, it ensures that the kernel and tail
#' registries are ready for later model-building code and sets the package plotly
#' option if the user has not already chosen one.
#'
#' @keywords internal
.onLoad <- function(libname, pkgname) {
  # Initialize kernel registry if available
  # (may not be loaded during roxygen documentation generation)
  init_kernel_registry_fun <- get0(
    "init_kernel_registry",
    mode = "function",
    inherits = TRUE
  )
  if (!is.null(init_kernel_registry_fun)) {
    init_kernel_registry_fun()
  }
  if (is.null(getOption("CausalMixGPD.plotly"))) {
    options(CausalMixGPD.plotly = FALSE)
  }
  # Wrap exported functions if utility is available
  wrap_exported_silent_fun <- get0(
    ".wrap_exported_silent",
    mode = "function",
    inherits = TRUE
  )
  if (!is.null(wrap_exported_silent_fun)) {
    wrap_exported_silent_fun(pkgname, opt_name = "CausalMixGPD.silent")
  }
  invisible()
}

library(testthat)
testthat::set_max_fails(Inf)

# Keep package checks at CRAN-level tests, regardless of ambient CI vars.
is_pkg_check <- nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_")) ||
  nzchar(Sys.getenv("RCMDCHECK")) ||
  identical(tolower(Sys.getenv("NOT_CRAN", "")), "false")

if (is_pkg_check) {
  Sys.setenv(DPMIXGPD_TEST_LEVEL = "cran")
}

# During explicit coverage runs, default to CI-level tests unless overridden.
if (!is_pkg_check &&
    (nzchar(Sys.getenv("DPMIXGPD_COVERAGE")) ||
     nzchar(Sys.getenv("COVERAGE")) ||
     nzchar(Sys.getenv("R_COVR")) ||
     any(grepl("covr", loadedNamespaces(), ignore.case = TRUE))) &&
    !nzchar(Sys.getenv("DPMIXGPD_TEST_LEVEL"))) {
  Sys.setenv(DPMIXGPD_TEST_LEVEL = "ci")
}

test_check("CausalMixGPD")

#' Amoroso mixture distribution
#'
#' Finite mixture of Amoroso components for flexible positive-support bulk modeling.
#'
#' The mixture density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k f_{Amoroso}(x \mid a_k, \theta_k, \alpha_k, \beta_k),
#' }
#' with normalized weights \eqn{\tilde{w}_k}. These scalar functions are NIMBLE-compatible; for
#' vectorized R usage, use [amoroso_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions treat the weights
#'   as non-negative and normalize them internally when needed.
#' @param loc Numeric vector of length \eqn{K} giving component locations.
#' @param scale Numeric vector of length \eqn{K} giving component scales. Negative values flip support
#'   for the corresponding component.
#' @param shape1 Numeric vector of length \eqn{K} giving the first Amoroso shape parameter for each component.
#' @param shape2 Numeric vector of length \eqn{K} giving the second Amoroso shape parameter for each component.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Mixture density/CDF/RNG functions return numeric scalars. `qAmorosoMix()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' The Amoroso family is especially useful for positive-support data because it can reproduce a wide
#' range of skewed and heavy-right-tail shapes while remaining analytically tractable through its
#' gamma transformation. The mixture CDF is
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k F_{Amoroso}(x \mid a_k,\theta_k,\alpha_k,\beta_k),
#' }
#' and random generation proceeds by selecting a component and sampling from that component.
#'
#' Closed-form mixture quantiles are not available, so \code{qAmorosoMix()} inverts the mixture CDF
#' numerically. The analytical mixture mean is the weighted average of the component means,
#' \eqn{a_k + \theta_k \Gamma(\alpha_k + 1/\beta_k) / \Gamma(\alpha_k)}, whenever those component
#' moments exist.
#'
#' @seealso [amoroso_mixgpd()], [amoroso_gpd()], [amoroso_lowercase()],
#'   [build_nimble_bundle()], [kernel_support_table()].
#' @family amoroso kernel families
#'
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' loc <- c(0, 1, 2)
#' scale <- c(1.0, 1.2, 1.6)
#' shape1 <- c(2, 4, 6)
#' shape2 <- c(1.0, 1.2, 1.5)
#'
#' dAmorosoMix(2.0, w, loc, scale, shape1, shape2, log = 0)
#' pAmorosoMix(2.0, w, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)
#' qAmorosoMix(0.50, w, loc, scale, shape1, shape2)
#' qAmorosoMix(0.95, w, loc, scale, shape1, shape2)
#' replicate(10, rAmorosoMix(1, w, loc, scale, shape1, shape2))
#' @rdname amoroso_mix
#' @name amoroso_mix
#' @aliases dAmorosoMix pAmorosoMix rAmorosoMix qAmorosoMix
NULL


#' @describeIn amoroso_mix Density Function of Amoroso Mixture Distribution
#' @export
dAmorosoMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) {
      if (log == 1) return(-1.0e300) else return(0.0)
    }
    ww <- w / wsum

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + ww[j] * dAmoroso(x, loc[j], scale[j], shape1[j], shape2[j], 0)
    }
    if (s0 < eps) s0 <- eps
    if (log == 1) return(log(s0)) else return(s0)
  }
)

#' @describeIn amoroso_mix Cumulative Distribution Function of Amoroso Mixture Distribution
#' @export
pAmorosoMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) {
      if (log.p != 0) return(-1.0e300) else return(0.0)
    }
    ww <- w / wsum

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + ww[j] * pAmoroso(q, loc[j], scale[j], shape1[j], shape2[j], 1, 0)
    }

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn amoroso_mix Random Generation for Amoroso Mixture Distribution
#' @export
rAmorosoMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) return(0.0)

    u <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0

    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (u <= cw) {
          idx <- j
          found <- 1
        }
      }
    }

    return(rAmoroso(1, loc[idx], scale[idx], shape1[idx], shape2[idx]))
  }
)

#' @describeIn amoroso_mix Quantile Function of Amoroso Mixture Distribution
#' @export
qAmorosoMix <- function(p, w, loc, scale, shape1, shape2,
                        lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)
  w <- as.numeric(w); w <- w / sum(w)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- -Inf; next }
    if (pi >= 1) { out[i] <- Inf; next }

    lo <- min(qAmoroso(pi, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE), na.rm = TRUE)
    hi <- max(qAmoroso(pi, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE), na.rm = TRUE)
    if (!is.finite(lo)) lo <- -1e10
    if (!is.finite(hi)) hi <- 1e10
    if (lo > hi) {
      tmp <- lo
      lo <- hi
      hi <- tmp
    }
    if (lo == hi) {
      lo <- lo - 1
      hi <- hi + 1
    }
    f_lo <- as.numeric(pAmorosoMix(lo, w, loc, scale, shape1, shape2, 1, 0) - pi)
    f_hi <- as.numeric(pAmorosoMix(hi, w, loc, scale, shape1, shape2, 1, 0) - pi)
    iter <- 0L
    while (is.finite(f_lo) && f_lo > 0 && lo > -1e20 && iter < 60L) {
      step <- max(1, abs(lo))
      lo <- lo - step
      f_lo <- as.numeric(pAmorosoMix(lo, w, loc, scale, shape1, shape2, 1, 0) - pi)
      iter <- iter + 1L
    }
    iter <- 0L
    while (is.finite(f_hi) && f_hi < 0 && hi < 1e20 && iter < 60L) {
      step <- max(1, abs(hi))
      hi <- hi + step
      f_hi <- as.numeric(pAmorosoMix(hi, w, loc, scale, shape1, shape2, 1, 0) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(lo) || !is.finite(hi) || lo >= hi || !is.finite(f_lo) || !is.finite(f_hi) || f_lo * f_hi > 0) {
      out[i] <- NA_real_
    } else {
      out[i] <- stats::uniroot(
        function(z) pAmorosoMix(z, w, loc, scale, shape1, shape2, 1, 0) - pi,
        interval = c(lo, hi),
        tol = tol, maxiter = maxiter
      )$root
    }
  }
  out
}

meanAmorosoMix <- function(w, loc, scale, shape1, shape2) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  loc <- as.numeric(loc)
  scale <- as.numeric(scale)
  shape1 <- as.numeric(shape1)
  shape2 <- as.numeric(shape2)
  if (!all(lengths(list(loc, scale, shape1, shape2)) == length(ww))) return(NA_real_)
  if (any(!is.finite(loc)) || any(!is.finite(scale)) || any(!is.finite(shape1)) || any(!is.finite(shape2))) return(NA_real_)
  if (any(scale <= 0) || any(shape1 <= 0) || any(shape2 <= 0)) return(NA_real_)
  comp_mean <- loc + scale * exp(lgamma(shape1 + 1 / shape2) - lgamma(shape1))
  sum(ww * comp_mean)
}

meanAmorosoMixTrunc <- function(w, loc, scale, shape1, shape2, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  loc <- as.numeric(loc)
  scale <- as.numeric(scale)
  shape1 <- as.numeric(shape1)
  shape2 <- as.numeric(shape2)
  u <- as.numeric(threshold)[1]
  if (!all(lengths(list(loc, scale, shape1, shape2)) == length(ww))) return(NA_real_)
  if (any(!is.finite(loc)) || any(!is.finite(scale)) || any(!is.finite(shape1)) || any(!is.finite(shape2)) || is.na(u)) return(NA_real_)
  if (any(scale <= 0) || any(shape1 <= 0) || any(shape2 <= 0)) return(NA_real_)
  if (is.infinite(u) && u > 0) return(meanAmorosoMix(w = ww, loc = loc, scale = scale, shape1 = shape1, shape2 = shape2))
  z <- pmax((u - loc) / scale, 0)^shape2
  cdf_u <- stats::pgamma(z, shape = shape1, scale = 1)
  moment_u <- exp(lgamma(shape1 + 1 / shape2) - lgamma(shape1)) *
    stats::pgamma(z, shape = shape1 + 1 / shape2, scale = 1)
  sum(ww * (loc * cdf_u + scale * moment_u))
}


#' Amoroso mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to an Amoroso mixture
#' bulk. Let \eqn{F_{mix}} denote the Amoroso mixture CDF. The spliced CDF is
#' \eqn{F(x)=F_{mix}(x)} for \eqn{x<threshold} and
#' \eqn{F(x)=F_{mix}(threshold) + \left\{1-F_{mix}(threshold)\right\}G(x)} for \eqn{x\ge threshold}, where \eqn{G}
#' is the GPD CDF for exceedances above \code{threshold}.
#'
#' The density, CDF, and RNG are implemented as \code{nimbleFunction}s for use in NIMBLE models.
#' The quantile function is an R function that uses numerical inversion in the bulk region and
#' the closed-form GPD quantile in the tail region.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions treat the weights
#'   as non-negative and normalize them internally when needed.
#' @param loc Numeric vector of length \eqn{K} giving component locations.
#' @param scale Numeric vector of length \eqn{K} giving component scales.
#' @param shape1 Numeric vector of length \eqn{K} giving the first Amoroso shape parameter for each component.
#' @param shape2 Numeric vector of length \eqn{K} giving the second Amoroso shape parameter for each component.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qAmorosoMixGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' The Amoroso mixture describes the bulk up to the threshold and the generalized Pareto describes
#' exceedances above it. If \eqn{F_{mix}(u)=p_u}, then the splice uses
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' f_{mix}(x), & x < u, \\
#' (1-p_u) g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' Bulk quantiles are computed numerically from the Amoroso mixture CDF and tail quantiles are
#' obtained from the GPD inverse after rescaling the tail probability.
#'
#' @seealso [amoroso_mix()], [amoroso_gpd()], [gpd()], [amoroso_lowercase()], [dpmgpd()].
#' @family amoroso kernel families
#'
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' loc <- c(0, 1, 2)
#' scale <- c(1.0, 1.2, 1.6)
#' shape1 <- c(2, 4, 6)
#' shape2 <- c(1.0, 1.2, 1.5)
#' threshold <- 3
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dAmorosoMixGpd(4.0, w, loc, scale, shape1, shape2,
#'               threshold, tail_scale, tail_shape, log = 0)
#' pAmorosoMixGpd(4.0, w, loc, scale, shape1, shape2,
#'               threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
#' qAmorosoMixGpd(0.50, w, loc, scale, shape1, shape2,
#'               threshold, tail_scale, tail_shape)
#' qAmorosoMixGpd(0.95, w, loc, scale, shape1, shape2,
#'               threshold, tail_scale, tail_shape)
#' replicate(10, rAmorosoMixGpd(1, w, loc, scale, shape1, shape2,
#'                             threshold, tail_scale, tail_shape))
#' @rdname amoroso_mixgpd
#' @name amoroso_mixgpd
#' @aliases dAmorosoMixGpd pAmorosoMixGpd rAmorosoMixGpd qAmorosoMixGpd
NULL



#' @describeIn amoroso_mixgpd Density Function of Amoroso Mixture Distribution with GPD Tail
#' @export
dAmorosoMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    if (x < threshold) return(dAmorosoMix(x, w, loc, scale, shape1, shape2, log))

    Fu <- pAmorosoMix(threshold, w, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps

    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn amoroso_mixgpd Cumulative Distribution Function of Amoroso Mixture Distribution with GPD Tail
#' @export
pAmorosoMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    if (q < threshold) return(pAmorosoMix(q, w, loc, scale, shape1, shape2, lower.tail, log.p))

    Fu <- pAmorosoMix(threshold, w, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn amoroso_mixgpd Random Generation for Amoroso Mixture Distribution with GPD Tail
#' @export
rAmorosoMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 loc = double(1),
                 scale = double(1),
                 shape1 = double(1),
                 shape2 = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pAmorosoMix(threshold, w, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rAmorosoMix(1, w, loc, scale, shape1, shape2))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn amoroso_mixgpd Quantile Function of Amoroso Mixture Distribution with GPD Tail
#' @export
qAmorosoMixGpd <- function(p, w, loc, scale, shape1, shape2,
                           threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE,
                           tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)
  w <- as.numeric(w); w <- w / sum(w)

  # Fu computed using the compiled mixture CDF at threshold (consistent with PD/R definitions)
  Fu <- pAmorosoMix(threshold, w, loc, scale, shape1, shape2, 1, 0)
  Fu <- max(min(as.numeric(Fu), 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qAmorosoMix(pi, w, loc, scale, shape1, shape2,
                            lower.tail = TRUE, log.p = FALSE, tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


#' Amoroso with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' Amoroso bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param loc Numeric scalar location parameter of the Amoroso bulk.
#' @param scale Numeric scalar scale parameter of the Amoroso bulk.
#' @param shape1 Numeric scalar first Amoroso shape parameter.
#' @param shape2 Numeric scalar second Amoroso shape parameter.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qAmorosoGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' This is the single-component Amoroso splice. The Amoroso law controls the distribution below the
#' threshold and the GPD controls exceedances above it, scaled so that the resulting CDF is
#' continuous at the threshold. The ordinary mean of the spliced law exists only when the tail
#' satisfies \eqn{\xi < 1}; otherwise users should rely on restricted means or quantile summaries.
#'
#' @seealso [amoroso_mix()], [amoroso_mixgpd()], [gpd()], [amoroso_lowercase()].
#' @family amoroso kernel families
#'
#' @examples
#' loc <- 0
#' scale <- 1.5
#' shape1 <- 2
#' shape2 <- 1.2
#' threshold <- 3
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dAmorosoGpd(4.0, loc, scale, shape1, shape2,
#'            threshold, tail_scale, tail_shape, log = 0)
#' pAmorosoGpd(4.0, loc, scale, shape1, shape2,
#'            threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
#' qAmorosoGpd(0.50, loc, scale, shape1, shape2,
#'            threshold, tail_scale, tail_shape)
#' qAmorosoGpd(0.95, loc, scale, shape1, shape2,
#'            threshold, tail_scale, tail_shape)
#' replicate(10, rAmorosoGpd(1, loc, scale, shape1, shape2,
#'                          threshold, tail_scale, tail_shape))
#' @rdname amoroso_gpd
#' @name amoroso_gpd
#' @aliases dAmorosoGpd pAmorosoGpd rAmorosoGpd qAmorosoGpd
NULL

#' @describeIn amoroso_gpd Density Function of Amoroso Distribution with GPD Tail
#' @export
dAmorosoGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    if (x < threshold) return(dAmoroso(x, loc, scale, shape1, shape2, log))

    Fu <- pAmoroso(threshold, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps

    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn amoroso_gpd Cumulative Distribution Function of Amoroso Distribution with GPD Tail
#' @export
pAmorosoGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    if (q < threshold) return(pAmoroso(q, loc, scale, shape1, shape2, lower.tail, log.p))

    Fu <- pAmoroso(threshold, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn amoroso_gpd Random Generation for Amoroso Distribution with GPD Tail
#' @export
rAmorosoGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pAmoroso(threshold, loc, scale, shape1, shape2, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rAmoroso(1, loc, scale, shape1, shape2))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn amoroso_gpd Quantile Function of Amoroso Distribution with GPD Tail
#' @export
#' @param tol Numeric tolerance for numerical inversion in \code{qAmorosoGpd}.
#' @param maxiter Maximum iterations for numerical inversion in \code{qAmorosoGpd}.
qAmorosoGpd <- function(p, loc, scale, shape1, shape2,
                        threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- pAmoroso(threshold, loc, scale, shape1, shape2, 1, 0)
  Fu <- max(min(as.numeric(Fu), 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      # bulk quantile via numerical inversion of base CDF
      lo <- qAmoroso(pi, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE)
      hi <- lo
      if (!is.finite(lo)) {
        lo <- -1e10
        hi <- 1e10
      }
      if (lo > hi) {
        tmp <- lo
        lo <- hi
        hi <- tmp
      }
      if (lo == hi) {
        lo <- lo - 1
        hi <- hi + 1
      }
      f_lo <- as.numeric(pAmoroso(lo, loc, scale, shape1, shape2, 1, 0) - pi)
      f_hi <- as.numeric(pAmoroso(hi, loc, scale, shape1, shape2, 1, 0) - pi)
      iter <- 0L
      while (is.finite(f_lo) && f_lo > 0 && lo > -1e20 && iter < 60L) {
        step <- max(1, abs(lo))
        lo <- lo - step
        f_lo <- as.numeric(pAmoroso(lo, loc, scale, shape1, shape2, 1, 0) - pi)
        iter <- iter + 1L
      }
      iter <- 0L
      while (is.finite(f_hi) && f_hi < 0 && hi < 1e20 && iter < 60L) {
        step <- max(1, abs(hi))
        hi <- hi + step
        f_hi <- as.numeric(pAmoroso(hi, loc, scale, shape1, shape2, 1, 0) - pi)
        iter <- iter + 1L
      }
      if (!is.finite(lo) || !is.finite(hi) || lo >= hi || !is.finite(f_lo) || !is.finite(f_hi) || f_lo * f_hi > 0) {
        out[i] <- NA_real_
      } else {
        out[i] <- stats::uniroot(
          function(z) pAmoroso(z, loc, scale, shape1, shape2, 1, 0) - pi,
          interval = c(lo, hi),
          tol = tol, maxiter = maxiter
        )$root
      }
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


# ==========================================================
# Lowercase vectorized R wrappers for Amoroso kernels
# ==========================================================

#' Lowercase vectorized Amoroso distribution functions
#'
#' Vectorized R wrappers for the scalar Amoroso-kernel topics in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param loc,scale,shape1,shape2 Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These are vectorized wrappers around the scalar Amoroso routines used internally by the package.
#' They preserve the same location-scale-shape parameterization and the same piecewise splice logic
#' for GPD tails. Quantile wrappers therefore continue to rely on the scalar numerical inversion or
#' scalar GPD inverse exactly as documented for the uppercase functions.
#'
#' @seealso [amoroso_mix()], [amoroso_mixgpd()], [amoroso_gpd()], [bundle()],
#'   [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.3, 0.1)
#' locs <- c(0.5, 0.5, 0.5)
#' scls <- c(1, 1.3, 1.6)
#' s1 <- c(2.5, 3, 4)
#' s2 <- c(1.2, 1.2, 1.2)
#'
#' # Amoroso mixture
#' damorosomix(c(1, 2, 3), w = w, loc = locs, scale = scls, shape1 = s1, shape2 = s2)
#' ramorosomix(5, w = w, loc = locs, scale = scls, shape1 = s1, shape2 = s2)
#'
#' @name amoroso_lowercase
#' @rdname amoroso_lowercase
NULL

# ---- Amoroso Mix lowercase wrappers ----

#' @describeIn amoroso_lowercase Amoroso mixture density (vectorized)
#' @export
damorosomix <- function(x, w, loc, scale, shape1, shape2, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dAmorosoMix(xi, w = w, loc = loc, scale = scale,
                                                 shape1 = shape1, shape2 = shape2, log = log_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso mixture distribution function (vectorized)
#' @export
pamorosomix <- function(q, w, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pAmorosoMix(qi, w = w, loc = loc, scale = scale,
                                                 shape1 = shape1, shape2 = shape2,
                                                 lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso mixture quantile function (vectorized)
#' @export
qamorosomix <- function(p, w, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  qAmorosoMix(p, w = w, loc = loc, scale = scale, shape1 = shape1, shape2 = shape2,
              lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn amoroso_lowercase Amoroso mixture random generation (vectorized)
#' @export
ramorosomix <- function(n, w, loc, scale, shape1, shape2) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rAmorosoMix(1L, w = w, loc = loc, scale = scale,
                                                         shape1 = shape1, shape2 = shape2)),
         numeric(1L))
}

# ---- Amoroso Mix + GPD lowercase wrappers ----

#' @describeIn amoroso_lowercase Amoroso mixture + GPD density (vectorized)
#' @export
damorosomixgpd <- function(x, w, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dAmorosoMixGpd(xi, w = w, loc = loc, scale = scale,
                                                    shape1 = shape1, shape2 = shape2,
                                                    threshold = threshold, tail_scale = tail_scale,
                                                    tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso mixture + GPD distribution function (vectorized)
#' @export
pamorosomixgpd <- function(q, w, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pAmorosoMixGpd(qi, w = w, loc = loc, scale = scale,
                                                    shape1 = shape1, shape2 = shape2,
                                                    threshold = threshold, tail_scale = tail_scale,
                                                    tail_shape = tail_shape,
                                                    lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso mixture + GPD quantile function (vectorized)
#' @export
qamorosomixgpd <- function(p, w, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qAmorosoMixGpd(p, w = w, loc = loc, scale = scale, shape1 = shape1, shape2 = shape2,
                 threshold = threshold, tail_scale = tail_scale, tail_shape = tail_shape,
                 lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn amoroso_lowercase Amoroso mixture + GPD random generation (vectorized)
#' @export
ramorosomixgpd <- function(n, w, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rAmorosoMixGpd(1L, w = w, loc = loc, scale = scale,
                                                           shape1 = shape1, shape2 = shape2,
                                                           threshold = threshold, tail_scale = tail_scale,
                                                           tail_shape = tail_shape)),
         numeric(1L))
}

# ---- Amoroso + GPD lowercase wrappers ----

#' @describeIn amoroso_lowercase Amoroso + GPD density (vectorized)
#' @export
damorosogpd <- function(x, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dAmorosoGpd(xi, loc = loc, scale = scale,
                                                 shape1 = shape1, shape2 = shape2,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso + GPD distribution function (vectorized)
#' @export
pamorosogpd <- function(q, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pAmorosoGpd(qi, loc = loc, scale = scale,
                                                 shape1 = shape1, shape2 = shape2,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape,
                                                 lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn amoroso_lowercase Amoroso + GPD quantile function (vectorized)
#' @export
qamorosogpd <- function(p, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qAmorosoGpd(p, loc = loc, scale = scale, shape1 = shape1, shape2 = shape2,
              threshold = threshold, tail_scale = tail_scale, tail_shape = tail_shape,
              lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn amoroso_lowercase Amoroso + GPD random generation (vectorized)
#' @export
ramorosogpd <- function(n, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rAmorosoGpd(1L, loc = loc, scale = scale,
                                                         shape1 = shape1, shape2 = shape2,
                                                         threshold = threshold, tail_scale = tail_scale,
                                                         tail_shape = tail_shape)),
         numeric(1L))
}

#' Cauchy mixture distribution
#'
#' Finite mixture of Cauchy components for symmetric heavy-tailed bulk modeling on the real line.
#'
#' The mixture density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k f_C(x \mid \ell_k, s_k),
#' }
#' with normalized weights \eqn{\tilde{w}_k}. These scalar functions are NIMBLE-compatible; for
#' vectorized R usage, use [cauchy_mix_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param location,scale Numeric vectors of length \eqn{K} giving component locations and scales.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qCauchyMix()` returns a numeric
#'   vector with the same length as `p`.
#'
#' @details
#' The mixture CDF is the weighted average of component CDFs,
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k
#' \left\{\frac{1}{2} + \frac{1}{\pi}\arctan\left(\frac{x-\ell_k}{s_k}\right)\right\}.
#' }
#' Random generation first selects a component according to the normalized weights and then draws
#' from the chosen Cauchy law by inverse-CDF sampling.
#'
#' Because each Cauchy component has undefined mean and variance, the mixture also lacks an
#' ordinary mean in general. That is why the package exposes Cauchy kernels for densities, CDFs,
#' quantiles, medians, survival functions, and restricted means, but not for ordinary predictive
#' means.
#'
#' @seealso [cauchy()], [cauchy_mix_lowercase()], [build_nimble_bundle()], [kernel_support_table()].
#' @family cauchy kernel families
#'
#' @examples
#' w <- c(0.50, 0.30, 0.20)
#' location <- c(-2, 0, 3)
#' scale <- c(1.0, 0.7, 1.5)
#'
#' dCauchyMix(0.5, w = w, location = location, scale = scale, log = FALSE)
#' pCauchyMix(0.5, w = w, location = location, scale = scale,
#'            lower.tail = TRUE, log.p = FALSE)
#' qCauchyMix(0.50, w = w, location = location, scale = scale)
#' qCauchyMix(0.95, w = w, location = location, scale = scale)
#' replicate(10, rCauchyMix(1, w = w, location = location, scale = scale))

#'
#' @rdname cauchy_mix
#' @name cauchy_mix
#' @aliases dCauchyMix pCauchyMix rCauchyMix qCauchyMix
#' @importFrom stats dcauchy pcauchy rcauchy qcauchy runif uniroot
NULL

#' @describeIn cauchy_mix Cauchy mixture density
#' @export
dCauchyMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    K <- length(w)

    # sum weights (allow unnormalized weights)
    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) {
      if (log == 1) return(-1.0e300) else return(0.0)
    }

    s0 <- 0.0
    for (j in 1:K) {
      if (scale[j] > 0.0) {
        z <- (x - location[j]) / scale[j]
        dj <- 1.0 / (pi * scale[j] * (1.0 + z * z))
        s0 <- s0 + (w[j] / wsum) * dj
      }
    }

    if (s0 < eps) s0 <- eps
    if (log == 1) return(log(s0))
    return(s0)
  }
)

#' @describeIn cauchy_mix Cauchy mixture distribution function
#' @export
pCauchyMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    K <- length(w)

    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) {
      if (log.p != 0) return(log(eps)) else return(eps)
    }

    cdf <- 0.0
    for (j in 1:K) {
      if (scale[j] > 0.0) {
        z <- (q - location[j]) / scale[j]
        pj <- 0.5 + atan(z) / pi
        cdf <- cdf + (w[j] / wsum) * pj
      }
    }

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn cauchy_mix Cauchy mixture random generation
#' @export
rCauchyMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)
    K <- length(w)

    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) return(0.0)

    # component draw using unnormalized weights
    u <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0
    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (u <= cw) {
          idx <- j
          found <- 1
        }
      }
    }

    if (scale[idx] <= 0.0) return(0.0)

    # inverse-CDF sampling
    uu <- runif(1, 0.0, 1.0)
    return(location[idx] + scale[idx] * tan(pi * (uu - 0.5)))
  }
)


#' @describeIn cauchy_mix Cauchy mixture quantile function
#' @export
qCauchyMix <- function(p, w, location, scale,
                       lower.tail = TRUE, log.p = FALSE,
                       tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- -Inf; next }
    if (pi >= 1) { out[i] <- Inf; next }

    lo <- min(stats::qcauchy(pi, location = location, scale = scale), na.rm = TRUE)
    hi <- max(stats::qcauchy(pi, location = location, scale = scale), na.rm = TRUE)
    if (!is.finite(lo)) lo <- -1e20
    if (!is.finite(hi)) hi <- 1e20
    f_lo <- as.numeric(pCauchyMix(lo, w = w, location = location, scale = scale,
                                 lower.tail = TRUE, log.p = FALSE) - pi)
    f_hi <- as.numeric(pCauchyMix(hi, w = w, location = location, scale = scale,
                                 lower.tail = TRUE, log.p = FALSE) - pi)
    iter <- 0L
    while (is.finite(f_lo) && f_lo > 0 && lo > -1e20 && iter < 60L) {
      step <- max(1, abs(lo))
      lo <- lo - step
      f_lo <- as.numeric(pCauchyMix(lo, w = w, location = location, scale = scale,
                                   lower.tail = TRUE, log.p = FALSE) - pi)
      iter <- iter + 1L
    }
    iter <- 0L
    while (is.finite(f_hi) && f_hi < 0 && hi < 1e20 && iter < 60L) {
      step <- max(1, abs(hi))
      hi <- hi + step
      f_hi <- as.numeric(pCauchyMix(hi, w = w, location = location, scale = scale,
                                   lower.tail = TRUE, log.p = FALSE) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(lo) || !is.finite(hi) || lo >= hi || !is.finite(f_lo) || !is.finite(f_hi) || f_lo * f_hi > 0) {
      out[i] <- NA_real_
    } else {
      out[i] <- stats::uniroot(
        function(z) pCauchyMix(z, w = w, location = location, scale = scale,
                               lower.tail = TRUE, log.p = FALSE) - pi,
        interval = c(lo, hi),
        tol = tol, maxiter = maxiter
      )$root
    }
  }
  out
}

meanCauchyMix <- function(w, location, scale) {
  stop("Mean is not supported for the Cauchy kernel; use type='rmean'.", call. = FALSE)
}


# ==========================================================
# Lowercase vectorized R wrappers for Cauchy mixture
# ==========================================================

#' Lowercase vectorized Cauchy mixture distribution functions
#'
#' Vectorized R wrappers for the scalar Cauchy mixture functions in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param location,scale Numeric vectors of component parameters.
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These are vectorized R wrappers around the scalar Cauchy-mixture routines.
#' They retain the same location-scale parameterization and the same inverse-CDF
#' logic for simulation and quantiles. The lowercase functions do not alter the
#' heavy-tail theory of the underlying Cauchy components; they apply the scalar
#' routines elementwise to vector inputs in R.
#'
#' @seealso [cauchy_mix()], [cauchy()], [bundle()], [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.3, 0.1)
#' loc <- c(-1, 0, 1)
#' scl <- c(1, 1.2, 2)
#'
#' dcauchymix(c(-2, 0, 2), w = w, location = loc, scale = scl)
#' rcauchymix(5, w = w, location = loc, scale = scl)
#'
#' @name cauchy_mix_lowercase
#' @rdname cauchy_mix_lowercase
NULL

#' @describeIn cauchy_mix_lowercase Cauchy mixture density (vectorized)
#' @export
dcauchymix <- function(x, w, location, scale, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dCauchyMix(xi, w = w, location = location, scale = scale, log = log_int)),
         numeric(1L))
}

#' @describeIn cauchy_mix_lowercase Cauchy mixture distribution function (vectorized)
#' @export
pcauchymix <- function(q, w, location, scale, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pCauchyMix(qi, w = w, location = location, scale = scale,
                                                lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn cauchy_mix_lowercase Cauchy mixture quantile function (vectorized)
#' @export
qcauchymix <- function(p, w, location, scale, lower.tail = TRUE, log.p = FALSE,
                       tol = 1e-10, maxiter = 200) {
  qCauchyMix(p, w = w, location = location, scale = scale, lower.tail = lower.tail,
             log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn cauchy_mix_lowercase Cauchy mixture random generation (vectorized)
#' @export
rcauchymix <- function(n, w, location, scale) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rCauchyMix(1L, w = w, location = location, scale = scale)),
         numeric(1L))
}

#' Build a causal bundle (design + two outcome arms)
#'
#' \code{build_causal_bundle()} is the detailed constructor behind
#' \code{\link{bundle}} for causal analyses. It prepares:
#' \itemize{
#'   \item a propensity score (PS) design block for \eqn{A \mid X},
#'   \item a control-arm outcome bundle for \eqn{Y(0)},
#'   \item a treated-arm outcome bundle for \eqn{Y(1)}.
#' }
#'
#' The outcome bundles reuse the one-arm DPM plus optional GPD machinery. The
#' PS block provides a shared adjustment object used by
#' \code{\link{run_mcmc_causal}} and
#' \code{\link{predict.causalmixgpd_causal_fit}}.
#'
#' @details
#' The causal bundle encodes the two arm-specific predictive laws
#' \eqn{F_0(y \mid x)} and \eqn{F_1(y \mid x)}. Downstream causal estimands are
#' functionals of these two distributions:
#' \deqn{\mathrm{ATE} = E\{Y(1)\} - E\{Y(0)\}, \qquad
#' \mathrm{QTE}(\tau) = Q_1(\tau) - Q_0(\tau).}
#'
#' When \code{PS} is enabled, the package estimates a propensity score model
#' \eqn{e(x) = \Pr(A = 1 \mid X = x)} and uses a posterior summary of that score
#' as an augmented covariate in the arm-specific outcome models. This mirrors
#' the workflow described in the manuscript vignette.
#'
#' @param y Numeric outcome vector.
#' @param X Design matrix or data.frame of covariates (N x P).
#' @param A Binary treatment indicator (length N, values 0/1).
#' @param backend Character; the Dirichlet process representation for outcome models:
#'   \itemize{
#'     \item \code{"sb"}: stick-breaking truncation
#'     \item \code{"crp"}: Chinese Restaurant Process
#'     \item \code{"spliced"}: CRP with GPD tail splicing
#'   }
#'   If length 2, the first entry is used for treated (\code{A=1}) and the
#'   second for control (\code{A=0}).
#' @param kernel Character kernel name for outcome models (must exist in
#'   \code{get_kernel_registry()}). If length 2:
#'   \itemize{
#'     \item first entry: used for treated (\code{A=1})
#'     \item second entry: used for control (\code{A=0})
#'   }
#' @param GPD Logical; include GPD tail for outcomes if TRUE. If length 2:
#'   \itemize{
#'     \item first entry: used for treated (\code{A=1})
#'     \item second entry: used for control (\code{A=0})
#'   }
#' @param components Integer >= 2; truncation parameter for outcome mixtures. If length 2:
#'   \itemize{
#'     \item first entry: used for treated (\code{A=1})
#'     \item second entry: used for control (\code{A=0})
#'   }
#' @param param_specs Outcome parameter overrides (same structure as
#'   \code{build_nimble_bundle()}):
#'   \itemize{
#'     \item a single list: used for both arms
#'     \item a list with \code{con} and \code{trt} entries: arm-specific overrides
#'   }
#' @param mcmc_outcome MCMC settings list for the outcome bundles.
#' @param mcmc_ps MCMC settings list for the PS model.
#' @param epsilon Numeric in [0,1) used by outcome bundles for posterior truncation
#'   summaries. If length 2:
#'   \itemize{
#'     \item first entry: used for treated (\code{A=1})
#'     \item second entry: used for control (\code{A=0})
#'   }
#' @param alpha_random Logical; whether the outcome-model DP concentration parameter \eqn{\kappa} is stochastic.
#' @param ps_prior Normal prior for PS coefficients. List with \code{mean} and \code{sd}.
#' @param include_intercept Logical; if TRUE, an intercept column is prepended to \code{X}
#'   in the PS model.
#' @param PS Character or logical; controls propensity score estimation:
#'   \itemize{
#'     \item \code{"logit"} (default): Logistic regression PS model
#'     \item \code{"probit"}: Probit regression PS model
#'     \item \code{"naive"}: Gaussian naive Bayes PS model
#'     \item \code{FALSE}: No PS estimation; outcome models use only \code{X}
#'   }
#'   The PS model choice is stored in bundle metadata for downstream use in
#'   prediction and summaries.
#' @param ps_scale Scale used when augmenting outcomes with PS:
#'   \itemize{
#'     \item \code{"logit"}: augment on the logit (log-odds) scale
#'     \item \code{"prob"}: augment on the probability scale
#'   }
#' @param ps_summary Posterior summary for PS:
#'   \itemize{
#'     \item \code{"mean"}: posterior mean of propensity scores
#'     \item \code{"median"}: posterior median of propensity scores
#'   }
#' @param ps_clamp Numeric epsilon for clamping PS values to \eqn{(\epsilon, 1-\epsilon)}.
#' @param monitor Character monitor profile:
#'   \itemize{
#'     \item \code{"core"} (default): monitors only the essential model parameters
#'     \item \code{"full"}: monitors all model nodes
#'   }
#' @param monitor_latent Logical; whether to monitor latent cluster labels (\code{z}) in outcome arms.
#' @param monitor_v Logical; whether to monitor stick-breaking \code{v} terms for SB outcomes.
#' @return A list of class \code{"causalmixgpd_causal_bundle"} containing the
#'   design bundle, two outcome bundles, training data, arm indices, and
#'   metadata required for posterior prediction and causal effect summaries.
#' @seealso \code{\link{bundle}}, \code{\link{run_mcmc_causal}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
#'   \code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
#' @examples
#' \donttest{
#' set.seed(1)
#' N <- 25
#' X <- cbind(x1 = rnorm(N), x2 = runif(N))
#' A <- rbinom(N, 1, plogis(0.3 + 0.5 * X[, 1]))
#' y <- rexp(N) + 0.1
#'
#' cb <- build_causal_bundle(
#'   y = y,
#'   X = X,
#'   A = A,
#'   backend = "sb",
#'   kernel = "gamma",
#'   GPD = TRUE,
#'   components = 3,
#'   PS = "probit"
#' )
#' }
#' @export
build_causal_bundle <- function(
    y,
    X,
    A,
    backend = c("sb", "crp", "spliced"),
    kernel,
    GPD = FALSE,
    components = 10L,
    param_specs = NULL,
    mcmc_outcome = list(niter = 2000, nburnin = 500, thin = 1, nchains = 1, seed = 1),
    mcmc_ps = list(niter = 1000, nburnin = 250, thin = 1, nchains = 1, seed = 1),
    epsilon = 0.025,
    alpha_random = TRUE,
    ps_prior = list(mean = 0, sd = 2),
    include_intercept = TRUE,
    PS = "logit",
    ps_scale = c("logit", "prob"),
    ps_summary = c("mean", "median"),
    ps_clamp = 1e-6,
    monitor = c("core", "full"),
    monitor_latent = FALSE,
    monitor_v = FALSE
) {

  .arm_value <- function(val, name) {
    if (length(val) == 1L) return(list(trt = val, con = val))
    if (length(val) == 2L) return(list(trt = val[[1]], con = val[[2]]))
    stop(sprintf("%s must be length 1 or length 2.", name), call. = FALSE)
  }

  backend <- .arm_value(backend, "backend")
  backend$trt <- match.arg(backend$trt, choices = c("sb", "crp", "spliced"))
  backend$con <- match.arg(backend$con, choices = c("sb", "crp", "spliced"))
  monitor <- match.arg(monitor)
  ps_scale <- match.arg(ps_scale)
  ps_summary <- match.arg(ps_summary)

  if (identical(monitor, "full")) {
    monitor_latent <- TRUE
    monitor_v <- TRUE
  }

  y <- as.numeric(y)
  if (!length(y)) stop("y must be a non-empty numeric vector.", call. = FALSE)

  has_x <- !is.null(X)
  if (has_x) {
    if (!is.matrix(X)) X <- as.matrix(X)
    if (ncol(X) == 0L) {
      X <- NULL
      has_x <- FALSE
    }
  }
  if (has_x && nrow(X) != length(y)) {
    stop("X must have the same number of rows as length(y).", call. = FALSE)
  }

  A <- as.integer(A)
  if (length(A) != length(y)) stop("A must have the same length as y.", call. = FALSE)
  if (anyNA(A) || !all(A %in% c(0L, 1L))) stop("A must be binary (0/1) with no NA.", call. = FALSE)

  # Validate and normalize PS parameter (disable PS when X is missing/empty)
  ps_model_type <- FALSE
  ps_choices <- c("logit", "probit", "naive")
  if (!has_x) {
    ps_model_type <- FALSE
  } else if (isFALSE(PS) || is.null(PS)) {
    ps_model_type <- FALSE
  } else if (isTRUE(PS)) {
    ps_model_type <- "logit"
  } else {
    ps_model_type <- match.arg(as.character(PS), choices = ps_choices)
  }

  idx_con <- which(A == 0L)
  idx_trt <- which(A == 1L)
  if (!length(idx_con) || !length(idx_trt)) {
    stop("Both treatment arms must have at least one observation.", call. = FALSE)
  }

  if (is.null(components)) {
    components <- 10L
  }
  components <- .arm_value(components, "components")
  components$trt <- as.integer(components$trt)
  components$con <- as.integer(components$con)
  if (!is.finite(components$trt) || components$trt < 2L) {
    stop("components (treated) must be an integer >= 2.", call. = FALSE)
  }
  if (!is.finite(components$con) || components$con < 2L) {
    stop("components (control) must be an integer >= 2.", call. = FALSE)
  }

  # Only build PS bundle if PS model is specified (not FALSE)
  ps_bundle <- NULL
  ps_placeholder <- NULL
  if (!isFALSE(ps_model_type)) {
    ps_spec <- list(
      model = ps_model_type,
      prior = list(
        mean = ps_prior$mean %||% 0,
        sd = ps_prior$sd %||% 2
      ),
      include_intercept = isTRUE(include_intercept)
    )
    ps_bundle <- .build_ps_bundle(A = A, X = X, spec = ps_spec, mcmc = mcmc_ps)
    ps_placeholder <- rep(0, length(y))
  }

  ps_con <- param_specs$con %||% param_specs
  ps_trt <- param_specs$trt %||% param_specs

  kernel <- .arm_value(kernel, "kernel")
  kchoices <- names(get_kernel_registry())
  kernel$trt <- match.arg(kernel$trt, choices = kchoices)
  kernel$con <- match.arg(kernel$con, choices = kchoices)

  GPD <- .arm_value(GPD, "GPD")
  GPD$trt <- isTRUE(GPD$trt)
  GPD$con <- isTRUE(GPD$con)

  epsilon <- .arm_value(epsilon, "epsilon")
  epsilon$trt <- as.numeric(epsilon$trt)
  epsilon$con <- as.numeric(epsilon$con)
  if (!is.finite(epsilon$trt) || epsilon$trt < 0 || epsilon$trt >= 1) {
    stop("epsilon (treated) must be in [0,1).", call. = FALSE)
  }
  if (!is.finite(epsilon$con) || epsilon$con < 0 || epsilon$con >= 1) {
    stop("epsilon (control) must be in [0,1).", call. = FALSE)
  }

  bundle_con <- build_nimble_bundle(
    y = y[idx_con],
    X = if (has_x) X[idx_con, , drop = FALSE] else NULL,
    ps = if (!is.null(ps_placeholder)) ps_placeholder[idx_con] else NULL,
    backend = backend$con,
    kernel = kernel$con,
    GPD = GPD$con,
    components = components$con,
    param_specs = ps_con,
    mcmc = mcmc_outcome,
    epsilon = epsilon$con,
    alpha_random = alpha_random,
    monitor = monitor,
    monitor_latent = monitor_latent,
    monitor_v = monitor_v
  )

  bundle_trt <- build_nimble_bundle(
    y = y[idx_trt],
    X = if (has_x) X[idx_trt, , drop = FALSE] else NULL,
    ps = if (!is.null(ps_placeholder)) ps_placeholder[idx_trt] else NULL,
    backend = backend$trt,
    kernel = kernel$trt,
    GPD = GPD$trt,
    components = components$trt,
    param_specs = ps_trt,
    mcmc = mcmc_outcome,
    epsilon = epsilon$trt,
    alpha_random = alpha_random,
    monitor = monitor,
    monitor_latent = monitor_latent,
    monitor_v = monitor_v
  )

  out <- list(
    design = ps_bundle,
    outcome = list(con = bundle_con, trt = bundle_trt),
    data = list(y = y, X = X, A = A),
    index = list(con = idx_con, trt = idx_trt),
    meta = list(
      backend = backend,
      kernel = kernel,
      GPD = GPD,
      components = components,
      epsilon = epsilon,
      has_x = has_x,
      needs_ps = !isFALSE(ps_model_type),
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      ps_clamp = ps_clamp,
      ps = list(
        enabled = !isFALSE(ps_model_type),
        include_in_outcome = !isFALSE(ps_model_type),
        model_type = ps_model_type
      ),
      monitor_policy = list(
        monitor = monitor,
        monitor_latent = isTRUE(monitor_latent),
        monitor_v = isTRUE(monitor_v)
      )
    ),
    call = match.call()
  )
  class(out) <- "causalmixgpd_causal_bundle"
  out
}

.run_ps_mcmc_bundle <- function(bundle, show_progress = TRUE, quiet = FALSE, timing = FALSE) {
  nimble_quiet <- isTRUE(show_progress) || isTRUE(quiet)

  stopifnot(inherits(bundle, "causalmixgpd_ps_bundle"))
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = isTRUE(quiet),
    label = "ps"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Validating PS MCMC inputs")

  if (!"package:nimble" %in% search()) {
    suppressPackageStartupMessages(base::require("nimble", quietly = TRUE, warn.conflicts = FALSE))
  }

  constants <- bundle$constants %||% list()
  data <- bundle$data %||% list()
  inits <- bundle$inits %||% list()
  m <- bundle$mcmc %||% list()
  model_type <- bundle$spec$model %||% "logit"
  code <- tryCatch(
    .ps_model_code(model_type),
    error = function(e) {
      code0 <- .extract_nimble_code(bundle$code)
      tryCatch({
        cleaned <- .deparse_without_covr(code0)
        expr <- parse(text = cleaned)[[1]]
        do.call(nimble::nimbleCode, list(expr))
      }, error = function(e2) {
        code0
      })
    }
  )
  monitors <- bundle$monitors %||% "beta"
  timing <- isTRUE(timing %||% m$timing %||% FALSE)
  tic <- function() proc.time()[["elapsed"]]
  timing_info <- list(build = 0, compile = 0, mcmc = 0, total = 0)
  t0_total <- tic()

  .ps_glm_fallback_fit <- function() {
    X <- as.matrix(data$X %||% matrix(0, nrow = length(data$A %||% integer(0)), ncol = 0L))
    A_obs <- as.integer(data$A %||% integer(0))
    if (!length(A_obs) || nrow(X) != length(A_obs)) {
      stop("PS fallback requires aligned A and X data.", call. = FALSE)
    }

    link <- if (identical(model_type, "probit")) "probit" else "logit"
    fit_glm <- stats::glm.fit(
      x = X,
      y = A_obs,
      family = stats::binomial(link = link)
    )

    beta_hat <- as.numeric(fit_glm$coefficients)
    beta_hat[!is.finite(beta_hat)] <- 0
    if (!length(beta_hat)) beta_hat <- 0

    niter <- as.integer(m$niter %||% 2000)
    nburnin <- as.integer(m$nburnin %||% 500)
    thin <- as.integer(m$thin %||% 1)
    nchains <- as.integer(m$nchains %||% 1)
    n_draws <- max(1L, as.integer(((niter - nburnin) / max(thin, 1L)) * max(nchains, 1L)))
    draws <- matrix(
      rep(beta_hat, each = n_draws),
      nrow = n_draws,
      ncol = length(beta_hat),
      byrow = FALSE
    )
    colnames(draws) <- sprintf("beta[%d]", seq_len(ncol(draws)))

    list(
      mcmc = list(samples = draws),
      bundle = bundle,
      timing = timing_info,
      call = match.call()
    )
  }

  .cmgpd_progress_step(progress_ctx, "Building PS NIMBLE model")
  t0_build <- tic()
  Rmodel <- .with_nimble_exports({
    .register_nimble_exports()
    tryCatch(
      {
        # Generated PS models are validated before this stage; disabling NIMBLE's
        # full check avoids a large startup penalty in manuscript workflows.
        .cmgpd_capture_nimble(
          nimble::nimbleModel(
            code = code,
            data = data,
            constants = constants,
            inits = inits,
            check = FALSE,
            calculate = FALSE
          ),
          suppress = nimble_quiet
        )
      },
      error = function(e) {
        msg <- conditionMessage(e)
        covr_reserved <- grepl("checkReservedVarNames", msg, fixed = TRUE) && grepl("if, if", msg, fixed = TRUE)
        if (!covr_reserved) stop(e)
        timing_info$build <- tic() - t0_build
        timing_info$total <- tic() - t0_total
        fit <- .ps_glm_fallback_fit()
        class(fit) <- "causalmixgpd_ps_fit"
        return(fit)
      }
    )
  })
  if (inherits(Rmodel, "causalmixgpd_ps_fit")) return(Rmodel)

  conf <- .cmgpd_capture_nimble(
    nimble::configureMCMC(
      Rmodel,
      monitors = monitors,
      enableWAIC = FALSE
    ),
    suppress = nimble_quiet
  )

  Rmcmc <- .cmgpd_capture_nimble(nimble::buildMCMC(conf), suppress = nimble_quiet)
  timing_info$build <- tic() - t0_build

  .cmgpd_progress_step(progress_ctx, "Compiling PS model")
  t0_compile <- tic()
  invisible(.cmgpd_capture_nimble(
    nimble::compileNimble(Rmodel, showCompilerOutput = FALSE),
    suppress = nimble_quiet
  ))
  Cmcmc  <- .cmgpd_capture_nimble(
    nimble::compileNimble(Rmcmc, project = Rmodel, showCompilerOutput = FALSE),
    suppress = nimble_quiet
  )
  timing_info$compile <- tic() - t0_compile

  niter   <- as.integer(m$niter   %||% 2000)
  nburnin <- as.integer(m$nburnin %||% 500)
  thin    <- as.integer(m$thin    %||% 1)
  nchains <- as.integer(m$nchains %||% 1)

  seed <- m$seed %||% NULL
  if (!is.null(seed)) {
    seed <- as.integer(seed)
    if (length(seed) == 1L && nchains > 1L) seed <- seed + seq_len(nchains) - 1L
    if (length(seed) != nchains) stop("mcmc$seed must be length 1 or length nchains.", call. = FALSE)
  }

  if (nchains > 1L) {
    inits_list <- vector("list", nchains)
    for (ch in seq_len(nchains)) {
      if (!is.null(seed)) set.seed(seed[ch])
      inits_list[[ch]] <- inits
    }
  } else {
    if (!is.null(seed)) set.seed(seed[1])
    inits_list <- inits
  }

  .cmgpd_progress_step(progress_ctx, "Running PS MCMC")
  t0_mcmc <- tic()
  res <- .with_nimble_exports(
    .cmgpd_capture_nimble(
      nimble::runMCMC(
        Cmcmc,
        niter = niter,
        nburnin = nburnin,
        thin = thin,
        nchains = nchains,
        inits = inits_list,
        setSeed = seed,
        progressBar = FALSE,
        samplesAsCodaMCMC = TRUE
      ),
      suppress = nimble_quiet
    )
  )
  timing_info$mcmc <- tic() - t0_mcmc
  timing_info$total <- tic() - t0_total

  .cmgpd_progress_step(progress_ctx, "Assembling PS fit")
  fit <- list(
    mcmc = list(samples = res),
    bundle = bundle,
    timing = timing_info,
    call = match.call()
  )
  class(fit) <- "causalmixgpd_ps_fit"
  fit
}

.ensure_causal_outcome_bundle_runtime_fields <- function(bundle, monitor_policy = NULL) {
  stopifnot(inherits(bundle, "causalmixgpd_bundle"))

  if (is.null(bundle$code)) {
    bundle$code <- .wrap_nimble_code(build_code_from_spec(bundle$spec))
  }
  if (is.null(bundle$constants)) {
    bundle$constants <- build_constants_from_spec(bundle$spec)
  }
  if (is.null(bundle$dimensions)) {
    bundle$dimensions <- build_dimensions_from_spec(bundle$spec)
  }
  if (is.null(bundle$monitors) || !length(bundle$monitors)) {
    pol <- bundle$monitor_policy %||% monitor_policy %||% list()
    bundle$monitors <- build_monitors_from_spec(
      bundle$spec,
      monitor_v = isTRUE(pol$monitor_v),
      monitor_latent = isTRUE(pol$monitor_latent)
    )
  }
  if (is.null(bundle$inits)) {
    bundle$inits <- build_inits_from_spec(
      bundle$spec,
      y = bundle$data$y,
      X = bundle$data$X %||% NULL
    )
  }

  bundle
}

#' Run posterior sampling for a causal bundle
#'
#' \code{run_mcmc_causal()} executes the PS block (when enabled) and the two
#' arm-specific outcome models prepared by \code{\link{build_causal_bundle}},
#' then returns a single \code{"causalmixgpd_causal_fit"} object.
#'
#' @details
#' The fitted object contains the posterior draws needed to evaluate arm-level
#' predictive distributions \eqn{F_1(y \mid x)} and
#' \eqn{F_0(y \mid x)}, followed by marginal or conditional causal
#' contrasts. When \code{PS = FALSE} in the bundle, the PS block is skipped and
#' outcome prediction uses only the original covariates.
#'
#' @param bundle A \code{"causalmixgpd_causal_bundle"} from \code{build_causal_bundle()}.
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @param quiet Logical; if TRUE, suppress step messages and progress display.
#' @param parallel_arms Logical; if TRUE, run control and treated outcome arms in parallel.
#' @param workers Optional integer workers for parallel arm execution.
#' @param timing Logical; if TRUE, return arm and total timings in \code{$timing}.
#' @param z_update_every Integer >= 1 passed to arm-level outcome MCMC.
#' @return A list of class \code{"causalmixgpd_causal_fit"} containing the
#'   fitted treated/control outcome models, optional PS fit, the original
#'   bundle, and timing metadata when requested.
#' @seealso \code{\link{build_causal_bundle}}, \code{\link{mcmc}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
#'   \code{\link{qte}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb)
#' }
#' @export
run_mcmc_causal <- function(bundle, show_progress = TRUE, quiet = FALSE,
                            parallel_arms = FALSE, workers = NULL, timing = FALSE,
                            z_update_every = NULL) {
  stopifnot(inherits(bundle, "causalmixgpd_causal_bundle"))
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 6L,
    enabled = isTRUE(show_progress),
    quiet = isTRUE(quiet),
    label = "causal"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Validating causal MCMC configuration")

  z_update_every <- as.integer(
    z_update_every %||%
      bundle$outcome$con$mcmc$z_update_every %||%
      bundle$outcome$trt$mcmc$z_update_every %||%
      1L
  )
  if (!is.finite(z_update_every) || z_update_every < 1L) {
    stop("'z_update_every' must be an integer >= 1.", call. = FALSE)
  }

  # Check if propensity scores are enabled
  ps_meta <- bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && isTRUE(bundle$meta$has_x)
  ps_summary <- bundle$meta$ps_summary %||% "mean"
  ps_scale <- bundle$meta$ps_scale %||% "logit"
  ps_clamp <- bundle$meta$ps_clamp %||% 1e-6

  ps_fit <- NULL
  ps_hat <- NULL
  ps_cov <- NULL
  ps_model <- NULL
  timing_info <- list(total = NA_real_, ps = NA_real_, con = NA_real_, trt = NA_real_, parallel_arms = FALSE)
  t0_total <- proc.time()[["elapsed"]]

  if (ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Running propensity score block")
    ps_fit <- .run_ps_mcmc_bundle(bundle$design, show_progress = show_progress, quiet = quiet, timing = timing)
    timing_info$ps <- ps_fit$timing$total %||% ps_fit$timing$mcmc %||% NA_real_
    ps_training_X <- bundle$data$X
    if (is.null(ps_training_X)) {
      stop("Training design matrix 'X' is missing from causal bundle.", call. = FALSE)
    }
    ps_training_X <- if (is.matrix(ps_training_X)) ps_training_X else as.matrix(ps_training_X)
    storage.mode(ps_training_X) <- "double"

    # Compute propensity scores and assign to outcome data
    ps_hat <- .compute_ps_from_fit(
      ps_fit = ps_fit,
      ps_bundle = bundle$design,
      X_new = ps_training_X,
      summary = ps_summary,
      clamp = ps_clamp
    )
    ps_cov <- .apply_ps_scale(ps_hat, scale = ps_scale, clamp = ps_clamp)
    idx_con <- bundle$index$con
    idx_trt <- bundle$index$trt
    bundle$outcome$con$data$ps <- ps_cov[idx_con]
    bundle$outcome$trt$data$ps <- ps_cov[idx_trt]

    # Prepare PS model for downstream prediction
    ps_model <- list(
      fit = ps_fit,
      bundle = bundle$design,
      scale = ps_scale,
      summary = ps_summary,
      clamp = ps_clamp
    )
  }
  if (!ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity score block")
  }

  .cmgpd_progress_step(progress_ctx, "Validating outcome-arm bundles")
  for (arm in c("con", "trt")) {
    bundle$outcome[[arm]] <- .ensure_causal_outcome_bundle_runtime_fields(
      bundle$outcome[[arm]],
      monitor_policy = bundle$meta$monitor_policy %||% list()
    )
  }
  parallel_arms <- isTRUE(parallel_arms)
  workers <- as.integer(workers %||% 2L)
  if (!is.finite(workers) || workers < 1L) workers <- 1L
  if (parallel_arms &&
      requireNamespace("future", quietly = TRUE) &&
      requireNamespace("future.apply", quietly = TRUE)) {
    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = min(workers, 2L))
    fit_list <- future.apply::future_lapply(c("con", "trt"), function(arm) {
      run_mcmc_bundle_manual(
        bundle$outcome[[arm]],
        show_progress = FALSE,
        quiet = TRUE,
        timing = timing,
        z_update_every = z_update_every
      )
    }, future.seed = TRUE)
    timing_info$parallel_arms <- TRUE
    timing_info$con <- fit_list[[1]]$timing$mcmc %||% NA_real_
    timing_info$trt <- fit_list[[2]]$timing$mcmc %||% NA_real_
    con_fit <- fit_list[[1]]
    trt_fit <- fit_list[[2]]
  } else {
    if (parallel_arms && !isTRUE(quiet) && !isTRUE(show_progress)) {
      warning("parallel_arms=TRUE requested but 'future'/'future.apply' are unavailable; running sequentially.",
              call. = FALSE)
    }
    .cmgpd_progress_step(progress_ctx, "Running control-arm outcome MCMC")
    t_con <- proc.time()[["elapsed"]]
    con_fit <- run_mcmc_bundle_manual(
      bundle$outcome$con,
      show_progress = show_progress,
      quiet = quiet,
      timing = timing,
      z_update_every = z_update_every
    )
    timing_info$con <- proc.time()[["elapsed"]] - t_con
    .cmgpd_progress_step(progress_ctx, "Running treated-arm outcome MCMC")
    t_trt <- proc.time()[["elapsed"]]
    trt_fit <- run_mcmc_bundle_manual(
      bundle$outcome$trt,
      show_progress = show_progress,
      quiet = quiet,
      timing = timing,
      z_update_every = z_update_every
    )
    timing_info$trt <- proc.time()[["elapsed"]] - t_trt
  }
  timing_info$total <- proc.time()[["elapsed"]] - t0_total

  # Attach PS model if available
  if (!is.null(ps_model)) {
    con_fit$ps_model <- ps_model
    trt_fit$ps_model <- ps_model
  }

  .cmgpd_progress_step(progress_ctx, "Assembling causal fit object")
  out <- list(
    ps_fit = ps_fit,
    outcome_fit = list(con = con_fit, trt = trt_fit),
    bundle = bundle,
    ps_hat = ps_hat,
    ps_cov = ps_cov,
    timing = timing_info,
    call = match.call()
  )
  class(out) <- c("causalmixgpd_causal_fit", "causalmixgpd_fit", "list")
  out
}

.causal_is_conditional_model <- function(fit) {
  meta <- fit$bundle$meta %||% list()
  has_x <- meta$has_x %||% NULL
  if (is.logical(has_x) && length(has_x) == 1L && !is.na(has_x)) {
    return(isTRUE(has_x))
  }
  X <- fit$bundle$data$X %||% NULL
  !is.null(X)
}

.causal_require_conditional <- function(fit, fn = c("cate", "cqte")) {
  fn <- match.arg(fn)
  if (.causal_is_conditional_model(fit)) return(invisible(TRUE))
  if (identical(fn, "cate")) {
    stop("cate() is available only for conditional causal models with covariates (X). For unconditional models, use ate()/att().", call. = FALSE)
  }
  stop("cqte() is available only for conditional causal models with covariates (X). For unconditional models, use qte()/qtt().", call. = FALSE)
}

.causal_validate_interval <- function(interval = "credible", level = 0.95) {
  compute_interval <- TRUE
  interval_use <- interval
  if (is.character(interval_use) && length(interval_use) == 1L && identical(tolower(interval_use), "none")) {
    compute_interval <- FALSE
    interval_use <- "credible"
  } else if (is.null(interval_use)) {
    compute_interval <- FALSE
    interval_use <- "credible"
  } else {
    interval_use <- match.arg(interval_use, choices = c("credible", "hpd"))
  }
  if (!is.numeric(level) || length(level) != 1L || !is.finite(level) || level <= 0 || level >= 1) {
    stop("'level' must be a numeric value between 0 and 1.", call. = FALSE)
  }
  list(compute_interval = compute_interval, interval = interval_use, level = as.numeric(level))
}

.causal_validate_fit <- function(fit) {
  if (!inherits(fit, "causalmixgpd_causal_fit")) {
    stop("'fit' must be a 'causalmixgpd_causal_fit' object.", call. = FALSE)
  }
  invisible(fit)
}

.causal_validate_probs <- function(probs) {
  probs <- as.numeric(probs)
  if (!length(probs) || anyNA(probs) || any(!is.finite(probs)) || any(probs <= 0 | probs >= 1)) {
    stop("'probs' must be a numeric vector with values strictly between 0 and 1.", call. = FALSE)
  }
  probs
}

.causal_validate_nsim_mean <- function(nsim_mean) {
  nsim_mean <- as.integer(nsim_mean)[1]
  if (!is.finite(nsim_mean) || nsim_mean < 1L) {
    stop("'nsim_mean' must be an integer >= 1.", call. = FALSE)
  }
  nsim_mean
}

.causal_warn_ignored_marginal_inputs <- function(fn, newdata = NULL, y = NULL, conditional_fn) {
  ignored <- character(0)
  if (!is.null(newdata)) ignored <- c(ignored, "'newdata'")
  if (!is.null(y)) ignored <- c(ignored, "'y'")
  if (!length(ignored)) return(invisible(FALSE))
  verb <- if (length(ignored) == 1L) "is" else "are"
  warning(
    sprintf(
      "%s() computes a marginal estimand over the training covariate distribution, so %s %s ignored. Use %s() for conditional effects at specified covariate values.",
      fn,
      paste(ignored, collapse = " and "),
      verb,
      conditional_fn
    ),
    call. = FALSE
  )
  invisible(TRUE)
}

.causal_resolve_conditional_x <- function(fit, newdata = NULL, fn = c("cate", "cqte")) {
  fn <- match.arg(fn)
  x_train <- fit$bundle$data$X %||% NULL
  x_pred <- newdata %||% x_train
  if (is.null(x_pred)) {
    stop(sprintf("%s() requires covariates in 'newdata' or training X in the fitted model.", fn), call. = FALSE)
  }
  x_pred
}

#' @export
cqte <- function(fit,
                 probs = c(0.1, 0.5, 0.9),
                 newdata = NULL,
                 interval = "credible",
                 level = 0.95,
                 show_progress = TRUE) {
  UseMethod("cqte")
}

#' @export
cqte.default <- function(fit,
                         probs = c(0.1, 0.5, 0.9),
                         newdata = NULL,
                         interval = "credible",
                         level = 0.95,
                         show_progress = TRUE) {
  .causal_validate_fit(fit)
}

#' Conditional quantile treatment effects
#'
#' \code{cqte()} evaluates treated-minus-control predictive quantiles at
#' user-supplied covariate rows.
#'
#' @details
#' For each prediction row \eqn{x}, the conditional quantile treatment effect is
#' \deqn{\mathrm{CQTE}(\tau, x) = Q_1(\tau \mid x) -
#' Q_0(\tau \mid x).}
#'
#' This estimand is available only for conditional causal models with
#' covariates. For marginal quantile contrasts over the empirical covariate
#' distribution, use \code{\link{qte}} or \code{\link{qtt}}.
#'
#' If the fit includes a PS block, the same PS adjustment is applied to both arm
#' predictions before differencing.
#'
#' @param fit A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.
#' @param probs Numeric vector of probabilities in (0, 1) specifying the quantile levels
#'   of the outcome distribution to estimate treatment effects at.
#' @param newdata Optional data.frame or matrix of covariates for prediction.
#'   If \code{NULL}, uses the training covariates stored in \code{fit}.
#' @param interval Character or NULL; type of credible interval:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @param level Numeric credible level for intervals (default 0.95 for 95 percent CI).
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @return An object of class \code{"causalmixgpd_qte"} containing the CQTE
#'   summary, the probability grid, and the treated/control prediction objects
#'   used to construct the effect. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{qte}}, \code{\link{qtt}}, \code{\link{cate}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' cqte(fit, probs = c(0.5, 0.9), newdata = X[1:5, , drop = FALSE])
#' cqte(fit, probs = c(0.5, 0.9), interval = "credible", level = 0.90)  # 90% CI
#' cqte(fit, probs = c(0.5, 0.9), interval = "hpd")  # HPD intervals
#' cqte(fit, probs = c(0.5, 0.9), interval = NULL)   # No intervals
#' }
#' @method cqte causalmixgpd_causal_fit
#' @aliases cqte
#' @export
cqte.causalmixgpd_causal_fit <- function(fit,
                probs = c(0.1, 0.5, 0.9),
                newdata = NULL,
                interval = "credible",
                level = 0.95,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  probs <- .causal_validate_probs(probs)
  iv <- .causal_validate_interval(interval = interval, level = level)
  .causal_require_conditional(fit, fn = "cqte")
  x_pred <- .causal_resolve_conditional_x(fit = fit, newdata = newdata, fn = "cqte")
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "cqte"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing CQTE inputs")

  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && isTRUE(fit$bundle$meta$has_x)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (!is.null(x_pred) && ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design

    # Fallback: try to retrieve PS model from outcome fits if causal fit missing it
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }

    # If PS model is still unavailable, warn and proceed without PS
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm quantiles")
  pr_trt <- predict(fit$outcome_fit$trt, newdata = x_pred, type = "quantile", index = probs,
                    ps = ps_cov,
                    interval = if (compute_interval) interval else NULL,
                    store_draws = TRUE,
                    show_progress = FALSE)
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm quantiles")
  pr_con <- predict(fit$outcome_fit$con, newdata = x_pred, type = "quantile", index = probs,
                    ps = ps_cov,
                    interval = if (compute_interval) interval else NULL,
                    store_draws = TRUE,
                    show_progress = FALSE)

  if (is.null(pr_trt$draws) || is.null(pr_con$draws)) {
    stop("CQTE requires stored posterior quantile draws; set store_draws=TRUE in predict().", call. = FALSE)
  }
  # Coerce unconditional quantile draws (M x S) into array (S x 1 x M)
  if (is.matrix(pr_trt$draws) && length(dim(pr_trt$draws)) == 2L) {
    M <- nrow(pr_trt$draws)
    S <- ncol(pr_trt$draws)
    arr <- array(NA_real_, dim = c(S, 1L, M))
    for (s in seq_len(S)) arr[s, 1L, ] <- pr_trt$draws[, s]
    pr_trt$draws <- arr
  }
  if (is.matrix(pr_con$draws) && length(dim(pr_con$draws)) == 2L) {
    M <- nrow(pr_con$draws)
    S <- ncol(pr_con$draws)
    arr <- array(NA_real_, dim = c(S, 1L, M))
    for (s in seq_len(S)) arr[s, 1L, ] <- pr_con$draws[, s]
    pr_con$draws <- arr
  }
  if (is.null(dim(pr_trt$draws))) pr_trt$draws <- array(pr_trt$draws, dim = c(length(pr_trt$draws), 1L, 1L))
  if (is.null(dim(pr_con$draws))) pr_con$draws <- array(pr_con$draws, dim = c(length(pr_con$draws), 1L, 1L))
  d_tr <- dim(pr_trt$draws)
  d_co <- dim(pr_con$draws)
  if (!identical(d_tr, d_co)) {
    common <- pmin(d_tr, d_co)
    if (any(common < 1L)) {
      stop("Treated and control posterior draws must have matching dimensions for CQTE.", call. = FALSE)
    }
    pr_trt$draws <- pr_trt$draws[
      seq_len(common[1]),
      seq_len(common[2]),
      seq_len(common[3]),
      drop = FALSE
    ]
    pr_con$draws <- pr_con$draws[
      seq_len(common[1]),
      seq_len(common[2]),
      seq_len(common[3]),
      drop = FALSE
    ]
  }

  diff_draws <- pr_trt$draws - pr_con$draws  # dims: S x n_pred x length(probs)
  .cmgpd_progress_step(progress_ctx, "Aggregating CQTE estimates")
  fit_mat <- apply(diff_draws, c(2, 3), mean, na.rm = TRUE)
  n_pred <- if (!is.null(dim(pr_trt$draws))) dim(pr_trt$draws)[2] else length(pr_trt$fit)
  n_prob <- length(probs)
  fit_mat <- matrix(fit_mat, nrow = n_pred, ncol = n_prob)
  profile_labels <- .causal_profile_labels(x_pred = x_pred, newdata = newdata)
  lower <- upper <- NULL
  if (compute_interval) {
    # Compute intervals for each (prediction, quantile) combination
    lower <- matrix(NA_real_, nrow = n_pred, ncol = n_prob)
    upper <- matrix(NA_real_, nrow = n_pred, ncol = n_prob)
    for (i in seq_len(n_pred)) {
      for (j in seq_len(n_prob)) {
        iv <- .compute_interval(diff_draws[, i, j], level = level, type = interval)
        lower[i, j] <- iv["lower"]
        upper[i, j] <- iv["upper"]
      }
    }
  }

  # Create CQTE fit data frame for convenience
  qte_fit <- data.frame(
    id = if (n_pred > 1L) rep(seq_len(n_pred), each = n_prob) else rep(1L, n_prob),
    index = rep(probs, times = n_pred),
    estimate = as.vector(t(fit_mat)),
    lower = if (!is.null(lower)) as.vector(t(lower)) else NA_real_,
    upper = if (!is.null(upper)) as.vector(t(upper)) else NA_real_
  )
  if (!is.null(profile_labels)) {
    qte_fit$profile <- rep(profile_labels, each = n_prob)
  }
  qte_fit <- .reorder_predict_cols(qte_fit)

  # Extract meta from causal fit
  meta <- fit$bundle$meta %||% list()

  out <- list(
    fit = fit_mat,
    fit_df = qte_fit,
    lower = lower,
    upper = upper,
    qte = list(fit = qte_fit, draws = diff_draws),
    probs = probs,
    grid = probs,
    trt = pr_trt,
    con = pr_con,
    trt_fit_df = pr_trt$fit_df %||% pr_trt$fit %||% NULL,
    con_fit_df = pr_con$fit_df %||% pr_con$fit %||% NULL,
    x = x_pred,
    profile = profile_labels,
    ps = ps_prob,
    n_pred = n_pred,
    level = level,
    interval = if (compute_interval) interval else "none",
    type = "cqte",
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  class(out) <- c("causalmixgpd_qte", "causalmixgpd_effect", "list")
  out
}


#' @export
cate <- function(fit,
                 newdata = NULL,
                 type = c("mean", "rmean"),
                 cutoff = NULL,
                 interval = "credible",
                 level = 0.95,
                 nsim_mean = 200L,
                 show_progress = TRUE) {
  UseMethod("cate")
}

#' @export
cate.default <- function(fit,
                         newdata = NULL,
                         type = c("mean", "rmean"),
                         cutoff = NULL,
                         interval = "credible",
                         level = 0.95,
                         nsim_mean = 200L,
                         show_progress = TRUE) {
  .causal_validate_fit(fit)
}

#' Conditional average treatment effects
#'
#' \code{cate()} evaluates treated-minus-control predictive means, or restricted
#' means, at user-supplied covariate rows.
#'
#' @details
#' For each prediction row \eqn{x}, the conditional average treatment effect is
#' \deqn{\mathrm{CATE}(x) = E\{Y(1) \mid x\} -
#' E\{Y(0) \mid x\}.}
#'
#' With \code{type = "rmean"}, the estimand becomes the conditional restricted
#' mean contrast
#' \deqn{E\{\min(Y(1), c) \mid x\} -
#' E\{\min(Y(0), c) \mid x\},}
#' which remains finite even when the ordinary mean is unstable under a heavy
#' GPD tail.
#' For outcome kernels with a finite analytical mean, the ordinary mean path is
#' analytical within each posterior draw; \code{rmean} remains a separate
#' simulation-based estimand.
#'
#' This estimand is available only for conditional causal models with
#' covariates. For marginal mean contrasts, use \code{\link{ate}} or
#' \code{\link{att}}.
#'
#' @param fit A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.
#' @param newdata Optional data.frame or matrix of covariates for prediction.
#'   If \code{NULL}, uses the training covariates stored in \code{fit}.
#' @param type Character; type of mean treatment effect:
#'   \itemize{
#'     \item \code{"mean"} (default): ordinary mean CATE
#'     \item \code{"rmean"}: restricted-mean CATE (requires \code{cutoff})
#'   }
#' @param cutoff Finite numeric cutoff for restricted mean; required for
#'   \code{type = "rmean"}, ignored otherwise.
#' @param interval Character or NULL; type of credible interval:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @param level Numeric credible level for intervals (default 0.95 for 95 percent CI).
#' @param nsim_mean Number of posterior predictive draws used by simulation-based
#'   mean targets. Ignored for analytical ordinary means.
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @return An object of class \code{"causalmixgpd_ate"} containing the CATE
#'   summary, optional intervals, and the treated/control prediction objects used
#'   to construct the effect. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{ate}}, \code{\link{att}}, \code{\link{cqte}},
#'   \code{\link{ate_rmean}}, \code{\link{predict.causalmixgpd_causal_fit}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' cate(fit, newdata = X[1:5, , drop = FALSE])
#' cate(fit, interval = "credible", level = 0.90)  # 90% CI
#' cate(fit, interval = "hpd")  # HPD intervals
#' cate(fit, interval = NULL)   # No intervals
#' }
#' @method cate causalmixgpd_causal_fit
#' @aliases cate
#' @export
cate.causalmixgpd_causal_fit <- function(fit,
                newdata = NULL,
                type = c("mean", "rmean"),
                cutoff = NULL,
                interval = "credible",
                level = 0.95,
                nsim_mean = 200L,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  type <- match.arg(type)
  nsim_mean <- .causal_validate_nsim_mean(nsim_mean)
  iv <- .causal_validate_interval(interval = interval, level = level)
  .causal_require_conditional(fit, fn = "cate")
  x_pred <- .causal_resolve_conditional_x(fit = fit, newdata = newdata, fn = "cate")
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "cate"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing CATE inputs")

  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && isTRUE(fit$bundle$meta$has_x)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (!is.null(x_pred) && ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design

    # Fallback: try to retrieve PS model from outcome fits if causal fit missing it
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }

    # If PS model is still unavailable, warn and proceed without PS
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm effects")
  pr_trt <- predict(fit$outcome_fit$trt, newdata = x_pred, type = type,
                    cutoff = cutoff,
                    ps = ps_cov, interval = if (compute_interval) interval else NULL,
                    level = level, nsim_mean = nsim_mean, store_draws = TRUE,
                    show_progress = FALSE)
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm effects")
  pr_con <- predict(fit$outcome_fit$con, newdata = x_pred, type = type,
                    cutoff = cutoff,
                    ps = ps_cov, interval = if (compute_interval) interval else NULL,
                    level = level, nsim_mean = nsim_mean, store_draws = TRUE,
                    show_progress = FALSE)

  if (is.null(pr_trt$draws) || is.null(pr_con$draws)) {
    stop("CATE requires stored posterior mean draws; set store_draws=TRUE in predict().", call. = FALSE)
  }
  if (is.null(dim(pr_trt$draws))) {
    pr_trt$draws <- matrix(pr_trt$draws, ncol = 1L)
  }
  if (is.null(dim(pr_con$draws))) {
    pr_con$draws <- matrix(pr_con$draws, ncol = 1L)
  }
  d_tr <- dim(pr_trt$draws)
  d_co <- dim(pr_con$draws)
  if (!identical(d_tr, d_co)) {
    common <- pmin(d_tr, d_co)
    if (any(common < 1L)) {
      stop("Treated and control posterior draws must have matching dimensions for CATE.", call. = FALSE)
    }
    pr_trt$draws <- pr_trt$draws[seq_len(common[1]), seq_len(common[2]), drop = FALSE]
    pr_con$draws <- pr_con$draws[seq_len(common[1]), seq_len(common[2]), drop = FALSE]
  }

  diff_draws <- pr_trt$draws - pr_con$draws  # dims: S x n_pred
  .cmgpd_progress_step(progress_ctx, "Aggregating CATE estimates")
  fit_vec <- colMeans(diff_draws, na.rm = TRUE)
  n_pred <- length(fit_vec)
  profile_labels <- .causal_profile_labels(x_pred = x_pred, newdata = newdata)
  lower <- upper <- NULL
  if (compute_interval) {
    # Compute intervals for each prediction point
    lower <- numeric(n_pred)
    upper <- numeric(n_pred)
    for (i in seq_len(n_pred)) {
      iv <- .compute_interval(diff_draws[, i], level = level, type = interval)
      lower[i] <- iv["lower"]
      upper[i] <- iv["upper"]
    }
  }

  # Create CATE fit data frame for convenience
  ate_fit <- data.frame(
    id = seq_len(n_pred),
    estimate = fit_vec,
    lower = if (!is.null(lower)) lower else NA_real_,
    upper = if (!is.null(upper)) upper else NA_real_
  )
  if (!is.null(profile_labels)) {
    ate_fit$profile <- profile_labels
  }
  ate_fit <- .reorder_predict_cols(ate_fit)

  # Extract meta from causal fit
  meta <- fit$bundle$meta %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && isTRUE(meta$has_x)

  out <- list(
    fit = fit_vec,
    fit_df = ate_fit,
    lower = lower,
    upper = upper,
    ate = list(fit = ate_fit, draws = diff_draws),
    grid = NULL,
    trt = pr_trt,
    con = pr_con,
    trt_fit_df = pr_trt$fit_df %||% pr_trt$fit %||% NULL,
    con_fit_df = pr_con$fit_df %||% pr_con$fit %||% NULL,
    x = x_pred,
    profile = profile_labels,
    ps = ps_prob,
    n_pred = n_pred,
    nsim_mean = .causal_mean_nsim(nsim_mean, pr_trt, pr_con),
    level = level,
    interval = if (compute_interval) interval else "none",
    type = "cate",
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  class(out) <- c("causalmixgpd_ate", "causalmixgpd_effect", "list")
  out
}

.causal_ate_draws_matrix <- function(draws) {
  if (is.null(draws)) return(NULL)
  if (is.null(dim(draws))) return(matrix(as.numeric(draws), ncol = 1L))
  if (length(dim(draws)) == 2L) return(draws)
  stop("Unexpected mean draw dimensions in causal treatment effects.", call. = FALSE)
}

.causal_mean_nsim <- function(default = NA_integer_, ...) {
  preds <- list(...)
  if (!length(preds)) return(default)
  methods <- vapply(preds, function(pr) {
    as.character((pr$diagnostics %||% list())$mean_method %||% NA_character_)
  }, character(1))
  if (all(!is.na(methods)) && all(methods == "analytic")) {
    return(NA_integer_)
  }
  as.integer(default)[1L]
}

.causal_effect_subset_index <- function(fit, n_pred, subset = c("all", "treated")) {
  subset <- match.arg(subset)
  if (n_pred <= 1L) return(1L)
  if (subset == "all") return(seq_len(n_pred))
  idx <- as.integer(fit$bundle$index$trt %||% integer(0))
  idx <- idx[is.finite(idx) & idx >= 1L & idx <= n_pred]
  idx <- unique(idx)
  if (!length(idx)) {
    stop("No treated rows available for treated-only treatment effect aggregation.", call. = FALSE)
  }
  idx
}

.causal_training_x_subset <- function(fit, subset = c("all", "treated")) {
  subset <- match.arg(subset)
  x_pred <- fit$bundle$data$X %||% NULL
  if (is.null(x_pred)) return(NULL)
  x_pred <- as.matrix(x_pred)
  if (subset == "all") return(x_pred)

  idx <- as.integer(fit$bundle$index$trt %||% integer(0))
  idx <- idx[is.finite(idx) & idx >= 1L & idx <= nrow(x_pred)]
  idx <- unique(idx)
  if (!length(idx)) {
    stop("No treated rows available for treated-only causal summaries.", call. = FALSE)
  }
  x_pred[idx, , drop = FALSE]
}

.causal_profile_labels <- function(x_pred, newdata = NULL) {
  if (is.null(newdata) || is.null(x_pred)) return(NULL)
  n_pred <- nrow(as.matrix(x_pred))
  rn <- rownames(x_pred)
  has_custom_names <- !is.null(rn) &&
    length(rn) == n_pred &&
    any(nzchar(rn)) &&
    !identical(rn, as.character(seq_len(n_pred)))
  if (has_custom_names) {
    return(as.character(rn))
  }
  paste("Profile", seq_len(n_pred))
}

.causal_summarize_draw_cols <- function(draw_mat, compute_interval, interval, level) {
  draw_mat <- as.matrix(draw_mat)
  estimate <- colMeans(draw_mat, na.rm = TRUE)
  lower <- upper <- rep(NA_real_, ncol(draw_mat))
  if (compute_interval) {
    for (j in seq_len(ncol(draw_mat))) {
      iv <- .compute_interval(draw_mat[, j], level = level, type = interval)
      lower[j] <- iv["lower"]
      upper[j] <- iv["upper"]
    }
  }
  list(estimate = estimate, lower = lower, upper = upper)
}

.causal_aggregate_qte <- function(obj, idx, effect_type = c("qte", "qtt")) {
  effect_type <- match.arg(effect_type)
  probs <- obj$probs %||% obj$grid %||% numeric(0)
  level <- obj$level %||% 0.95
  interval <- obj$interval %||% "none"
  compute_interval <- interval %in% c("credible", "hpd")

  trt_fit_draws <- .causal_ate_draws_matrix(obj$trt$draws %||% NULL)
  con_fit_draws <- .causal_ate_draws_matrix(obj$con$draws %||% NULL)
  if (is.null(trt_fit_draws) || is.null(con_fit_draws)) {
    stop("QTE aggregation requires treated/control posterior predictive draws.", call. = FALSE)
  }
  if (nrow(trt_fit_draws) != nrow(con_fit_draws)) {
    stop("Treated and control posterior draws must have the same number of rows.", call. = FALSE)
  }

  n_pred <- as.integer(obj$n_pred %||% 1L)
  if (n_pred > 1L) {
    if (ncol(trt_fit_draws) != ncol(con_fit_draws)) {
      stop("Treated and control posterior predictive draws must align by covariate row.", call. = FALSE)
    }
    idx <- unique(as.integer(idx))
    idx <- idx[idx >= 1L & idx <= ncol(trt_fit_draws)]
    if (!length(idx)) stop("No valid rows selected for QTE aggregation.", call. = FALSE)
    trt_pool <- trt_fit_draws[, idx, drop = FALSE]
    con_pool <- con_fit_draws[, idx, drop = FALSE]
  } else {
    # Unconditional case: each row already samples from the arm-level marginal.
    trt_pool <- trt_fit_draws
    con_pool <- con_fit_draws
  }

  .row_quantiles <- function(draw_mat, probs) {
    S <- nrow(draw_mat)
    M <- length(probs)
    out <- matrix(NA_real_, nrow = S, ncol = M)
    for (s in seq_len(S)) {
      xs <- as.numeric(draw_mat[s, ])
      xs <- xs[is.finite(xs)]
      if (!length(xs)) next
      out[s, ] <- as.numeric(stats::quantile(xs, probs = probs, names = FALSE, type = 7))
    }
    out
  }

  trt_q <- .row_quantiles(trt_pool, probs = probs)
  con_q <- .row_quantiles(con_pool, probs = probs)
  diff_q <- trt_q - con_q

  diff_summ <- .causal_summarize_draw_cols(diff_q, compute_interval = compute_interval, interval = interval, level = level)
  trt_summ <- .causal_summarize_draw_cols(trt_q, compute_interval = compute_interval, interval = interval, level = level)
  con_summ <- .causal_summarize_draw_cols(con_q, compute_interval = compute_interval, interval = interval, level = level)

  lower <- upper <- NULL
  if (compute_interval) {
    lower <- as.numeric(diff_summ$lower)
    upper <- as.numeric(diff_summ$upper)
  }

  qte_fit <- data.frame(
    index = probs,
    estimate = as.numeric(diff_summ$estimate),
    lower = if (!is.null(lower)) as.numeric(lower) else NA_real_,
    upper = if (!is.null(upper)) as.numeric(upper) else NA_real_
  )
  fit_tbl <- qte_fit[, c("index", "estimate", "lower", "upper"), drop = FALSE]

  trt_fit <- data.frame(
    index = probs,
    estimate = as.numeric(trt_summ$estimate),
    lower = as.numeric(trt_summ$lower),
    upper = as.numeric(trt_summ$upper)
  )
  con_fit <- data.frame(
    index = probs,
    estimate = as.numeric(con_summ$estimate),
    lower = as.numeric(con_summ$lower),
    upper = as.numeric(con_summ$upper)
  )

  S <- nrow(diff_q)
  M <- ncol(diff_q)
  trt_draws_out <- array(NA_real_, dim = c(S, 1L, M))
  con_draws_out <- array(NA_real_, dim = c(S, 1L, M))
  diff_draws_out <- array(NA_real_, dim = c(S, 1L, M))
  for (j in seq_len(M)) {
    trt_draws_out[, 1L, j] <- trt_q[, j]
    con_draws_out[, 1L, j] <- con_q[, j]
    diff_draws_out[, 1L, j] <- diff_q[, j]
  }

  pr_trt <- obj$trt
  pr_con <- obj$con
  pr_trt$fit <- trt_fit
  pr_con$fit <- con_fit
  pr_trt$draws <- trt_draws_out
  pr_con$draws <- con_draws_out

  out <- list(
    fit = fit_tbl,
    fit_df = fit_tbl,
    lower = lower,
    upper = upper,
    qte = list(fit = qte_fit, draws = diff_draws_out),
    probs = probs,
    grid = probs,
    trt = pr_trt,
    con = pr_con,
    trt_fit_df = pr_trt$fit_df %||% pr_trt$fit %||% NULL,
    con_fit_df = pr_con$fit_df %||% pr_con$fit %||% NULL,
    x = NULL,
    ps = NULL,
    n_pred = 1L,
    level = level,
    interval = interval,
    type = effect_type,
    meta = obj$meta %||% list()
  )
  class(out) <- c("causalmixgpd_qte", "causalmixgpd_effect", "list")
  out
}

.causal_aggregate_ate <- function(obj, idx, effect_type = c("ate", "att")) {
  effect_type <- match.arg(effect_type)
  level <- obj$level %||% 0.95
  interval <- obj$interval %||% "none"
  compute_interval <- interval %in% c("credible", "hpd")

  trt_draws <- .causal_ate_draws_matrix(obj$trt$draws %||% NULL)
  con_draws <- .causal_ate_draws_matrix(obj$con$draws %||% NULL)
  if (is.null(trt_draws) || is.null(con_draws)) {
    stop("ATE aggregation requires treated/control mean draws.", call. = FALSE)
  }
  if (!identical(dim(trt_draws), dim(con_draws))) {
    stop("Treated and control mean draws must match for aggregation.", call. = FALSE)
  }

  idx <- unique(as.integer(idx))
  idx <- idx[idx >= 1L & idx <= ncol(trt_draws)]
  if (!length(idx)) stop("No valid rows selected for ATE aggregation.", call. = FALSE)

  trt_avg <- rowMeans(trt_draws[, idx, drop = FALSE], na.rm = TRUE)
  con_avg <- rowMeans(con_draws[, idx, drop = FALSE], na.rm = TRUE)
  diff_avg <- trt_avg - con_avg

  diff_summ <- .causal_summarize_draw_cols(matrix(diff_avg, ncol = 1L), compute_interval = compute_interval, interval = interval, level = level)
  trt_summ <- .causal_summarize_draw_cols(matrix(trt_avg, ncol = 1L), compute_interval = compute_interval, interval = interval, level = level)
  con_summ <- .causal_summarize_draw_cols(matrix(con_avg, ncol = 1L), compute_interval = compute_interval, interval = interval, level = level)

  fit_vec <- as.numeric(diff_summ$estimate)
  lower <- upper <- NULL
  if (compute_interval) {
    lower <- as.numeric(diff_summ$lower)
    upper <- as.numeric(diff_summ$upper)
  }

  ate_fit <- data.frame(
    estimate = fit_vec,
    lower = if (!is.null(lower)) lower else NA_real_,
    upper = if (!is.null(upper)) upper else NA_real_
  )
  trt_fit <- data.frame(
    estimate = as.numeric(trt_summ$estimate),
    lower = as.numeric(trt_summ$lower),
    upper = as.numeric(trt_summ$upper)
  )
  con_fit <- data.frame(
    estimate = as.numeric(con_summ$estimate),
    lower = as.numeric(con_summ$lower),
    upper = as.numeric(con_summ$upper)
  )

  pr_trt <- obj$trt
  pr_con <- obj$con
  pr_trt$fit <- trt_fit
  pr_con$fit <- con_fit
  pr_trt$draws <- matrix(trt_avg, ncol = 1L)
  pr_con$draws <- matrix(con_avg, ncol = 1L)

  out <- list(
    fit = fit_vec,
    fit_df = ate_fit,
    lower = lower,
    upper = upper,
    ate = list(fit = ate_fit, draws = matrix(diff_avg, ncol = 1L)),
    grid = NULL,
    trt = pr_trt,
    con = pr_con,
    trt_fit_df = pr_trt$fit_df %||% pr_trt$fit %||% NULL,
    con_fit_df = pr_con$fit_df %||% pr_con$fit %||% NULL,
    x = NULL,
    ps = NULL,
    n_pred = 1L,
    nsim_mean = obj$nsim_mean %||% NA_integer_,
    level = level,
    interval = interval,
    type = effect_type,
    meta = obj$meta %||% list()
  )
  class(out) <- c("causalmixgpd_ate", "causalmixgpd_effect", "list")
  out
}

#' @export
qte <- function(fit,
                probs = c(0.1, 0.5, 0.9),
                newdata = NULL,
                y = NULL,
                interval = "credible",
                level = 0.95,
                show_progress = TRUE) {
  UseMethod("qte")
}

#' @export
qte.default <- function(fit,
                        probs = c(0.1, 0.5, 0.9),
                        newdata = NULL,
                        y = NULL,
                        interval = "credible",
                        level = 0.95,
                        show_progress = TRUE) {
  .causal_validate_fit(fit)
}

#' Quantile treatment effects, marginal over the empirical covariate distribution
#'
#' \code{qte()} returns the marginal quantile treatment effect implied by the
#' causal fit.
#'
#' @details
#' The package computes
#' \deqn{\mathrm{QTE}(\tau) = Q_1^{m}(\tau) - Q_0^{m}(\tau),}
#' where \eqn{Q_a^{m}(\tau)} is the arm-\eqn{a} posterior predictive marginal
#' quantile obtained by averaging over the empirical training covariate
#' distribution.
#'
#' For unconditional causal models (\code{X = NULL}), this reduces to a direct
#' contrast of the arm-level unconditional predictive distributions.
#'
#' @param fit A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.
#' @param probs Numeric vector of probabilities in (0, 1) specifying the quantile levels
#'   of the outcome distribution to estimate treatment effects at.
#' @param newdata Ignored for marginal estimands. If supplied, a warning is issued
#'   and training data are used.
#' @param y Ignored for marginal estimands. If supplied, a warning is issued
#'   and training data are used.
#' @param interval Character or NULL; type of credible interval:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @param level Numeric credible level for intervals (default 0.95 for 95 percent CI).
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @return An object of class \code{"causalmixgpd_qte"} containing the
#'   marginal QTE summary, the probability grid, and the arm-specific predictive
#'   objects used in the aggregation. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{qtt}}, \code{\link{cqte}}, \code{\link{ate}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' qte(fit, probs = c(0.5, 0.9))
#' }
#' @method qte causalmixgpd_causal_fit
#' @aliases qte
#' @export
qte.causalmixgpd_causal_fit <- function(fit,
                probs = c(0.1, 0.5, 0.9),
                newdata = NULL,
                y = NULL,
                interval = "credible",
                level = 0.95,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  probs <- .causal_validate_probs(probs)
  iv <- .causal_validate_interval(interval = interval, level = level)
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "qte"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing QTE inputs")

  .causal_warn_ignored_marginal_inputs(fn = "qte", newdata = newdata, y = y, conditional_fn = "cqte")
  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  x_pred <- .causal_training_x_subset(fit, subset = "all")
  n_pred <- if (!is.null(x_pred)) nrow(as.matrix(x_pred)) else 1L

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && .causal_is_conditional_model(fit) && !is.null(x_pred)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm draws")
  pr_trt <- predict(
    fit$outcome_fit$trt,
    newdata = x_pred,
    type = "fit",
    ps = ps_cov,
    interval = NULL,
    store_draws = TRUE,
    show_progress = FALSE
  )
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm draws")
  pr_con <- predict(
    fit$outcome_fit$con,
    newdata = x_pred,
    type = "fit",
    ps = ps_cov,
    interval = NULL,
    store_draws = TRUE,
    show_progress = FALSE
  )

  meta <- fit$bundle$meta %||% list()
  cq <- list(
    trt = pr_trt,
    con = pr_con,
    probs = probs,
    grid = probs,
    level = level,
    interval = if (compute_interval) interval else "none",
    n_pred = n_pred,
    x = NULL,
    ps = NULL,
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  idx <- seq_len(cq$n_pred %||% 1L)
  .cmgpd_progress_step(progress_ctx, "Aggregating QTE estimates")
  .causal_aggregate_qte(cq, idx = idx, effect_type = "qte")
}

#' Quantile treatment effects standardized to treated covariates
#'
#' \code{qtt()} computes the quantile treatment effect on the treated.
#'
#' @details
#' The estimand is
#' \deqn{\mathrm{QTT}(\tau) = Q_1^{t}(\tau) - Q_0^{t}(\tau),}
#' where marginalization is over the empirical covariate distribution of the
#' treated units only.
#'
#' @inheritParams qte.causalmixgpd_causal_fit
#' @return An object of class \code{"causalmixgpd_qte"} containing the QTT
#'   summary, the probability grid, and the arm-specific predictive objects used
#'   in the aggregation. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{qte}}, \code{\link{cqte}}, \code{\link{att}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' qtt(fit, probs = c(0.5, 0.9))
#' }
#' @export
qtt <- function(fit,
                probs = c(0.1, 0.5, 0.9),
                newdata = NULL,
                y = NULL,
                interval = "credible",
                level = 0.95,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  probs <- .causal_validate_probs(probs)
  iv <- .causal_validate_interval(interval = interval, level = level)
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "qtt"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing QTT inputs")

  .causal_warn_ignored_marginal_inputs(fn = "qtt", newdata = newdata, y = y, conditional_fn = "cqte")
  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  x_pred <- .causal_training_x_subset(fit, subset = "treated")
  n_pred <- if (!is.null(x_pred)) nrow(as.matrix(x_pred)) else 1L

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && .causal_is_conditional_model(fit) && !is.null(x_pred)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm draws")
  pr_trt <- predict(
    fit$outcome_fit$trt,
    newdata = x_pred,
    type = "fit",
    ps = ps_cov,
    interval = NULL,
    store_draws = TRUE,
    show_progress = FALSE
  )
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm draws")
  pr_con <- predict(
    fit$outcome_fit$con,
    newdata = x_pred,
    type = "fit",
    ps = ps_cov,
    interval = NULL,
    store_draws = TRUE,
    show_progress = FALSE
  )

  meta <- fit$bundle$meta %||% list()
  cq <- list(
    trt = pr_trt,
    con = pr_con,
    probs = probs,
    grid = probs,
    level = level,
    interval = if (compute_interval) interval else "none",
    n_pred = n_pred,
    x = NULL,
    ps = NULL,
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  idx <- seq_len(cq$n_pred %||% 1L)
  .cmgpd_progress_step(progress_ctx, "Aggregating QTT estimates")
  .causal_aggregate_qte(cq, idx = idx, effect_type = "qtt")
}

#' @export
ate <- function(fit,
                newdata = NULL,
                y = NULL,
                type = c("mean", "rmean"),
                cutoff = NULL,
                interval = "credible",
                level = 0.95,
                nsim_mean = 200L,
                show_progress = TRUE) {
  UseMethod("ate")
}

#' @export
ate.default <- function(fit,
                        newdata = NULL,
                        y = NULL,
                        type = c("mean", "rmean"),
                        cutoff = NULL,
                        interval = "credible",
                        level = 0.95,
                        nsim_mean = 200L,
                        show_progress = TRUE) {
  .causal_validate_fit(fit)
}

#' Average treatment effects, marginal over the empirical covariate distribution
#'
#' \code{ate()} computes the posterior predictive average treatment effect.
#'
#' @details
#' The default mean-scale estimand is
#' \deqn{\mathrm{ATE} = E\{Y(1)\} - E\{Y(0)\},}
#' where the expectation is taken with respect to the empirical training
#' covariate distribution for conditional models.
#'
#' When \code{type = "rmean"}, the function instead computes a restricted-mean
#' ATE using \eqn{E\{\min(Y(a), c)\}} for each arm.
#' For outcome kernels with a finite analytical mean, the ordinary mean path is
#' analytical within each posterior draw; \code{rmean} remains simulation-based.
#'
#' For unconditional causal models (\code{X = NULL}), the computation reduces to
#' a direct contrast of the unconditional treated and control predictive laws.
#'
#' @param fit A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.
#' @param newdata Ignored for marginal estimands. If supplied, a warning is issued
#'   and training data are used.
#' @param y Ignored for marginal estimands. If supplied, a warning is issued
#'   and training data are used.
#' @param type Character; type of mean treatment effect:
#'   \itemize{
#'     \item \code{"mean"} (default): ordinary mean ATE
#'     \item \code{"rmean"}: restricted-mean ATE (requires \code{cutoff})
#'   }
#' @param cutoff Finite numeric cutoff for restricted mean; required for
#'   \code{type = "rmean"}, ignored otherwise.
#' @param interval Character or NULL; type of credible interval:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @param level Numeric credible level for intervals (default 0.95 for 95 percent CI).
#' @param nsim_mean Number of posterior predictive draws used by simulation-based
#'   mean targets. Ignored for analytical ordinary means.
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @return An object of class \code{"causalmixgpd_ate"} containing the
#'   marginal ATE summary, optional intervals, and the arm-specific predictive
#'   objects used in the aggregation. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{att}}, \code{\link{cate}}, \code{\link{qte}},
#'   \code{\link{ate_rmean}}, \code{\link{predict.causalmixgpd_causal_fit}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' ate(fit, interval = "credible", level = 0.90, nsim_mean = 100)
#' }
#' @method ate causalmixgpd_causal_fit
#' @aliases ate
#' @export
ate.causalmixgpd_causal_fit <- function(fit,
                newdata = NULL,
                y = NULL,
                type = c("mean", "rmean"),
                cutoff = NULL,
                interval = "credible",
                level = 0.95,
                nsim_mean = 200L,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  type <- match.arg(type)
  nsim_mean <- .causal_validate_nsim_mean(nsim_mean)
  iv <- .causal_validate_interval(interval = interval, level = level)
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "ate"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing ATE inputs")

  .causal_warn_ignored_marginal_inputs(fn = "ate", newdata = newdata, y = y, conditional_fn = "cate")
  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  x_pred <- .causal_training_x_subset(fit, subset = "all")
  n_pred <- if (!is.null(x_pred)) nrow(as.matrix(x_pred)) else 1L

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && .causal_is_conditional_model(fit) && !is.null(x_pred)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm effects")
  pr_trt <- predict(
    fit$outcome_fit$trt,
    newdata = x_pred,
    type = type,
    cutoff = cutoff,
    ps = ps_cov,
    interval = if (compute_interval) interval else NULL,
    level = level,
    nsim_mean = nsim_mean,
    store_draws = TRUE,
    show_progress = FALSE
  )
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm effects")
  pr_con <- predict(
    fit$outcome_fit$con,
    newdata = x_pred,
    type = type,
    cutoff = cutoff,
    ps = ps_cov,
    interval = if (compute_interval) interval else NULL,
    level = level,
    nsim_mean = nsim_mean,
    store_draws = TRUE,
    show_progress = FALSE
  )

  meta <- fit$bundle$meta %||% list()
  ca <- list(
    trt = pr_trt,
    con = pr_con,
    level = level,
    interval = if (compute_interval) interval else "none",
    n_pred = n_pred,
    nsim_mean = .causal_mean_nsim(nsim_mean, pr_trt, pr_con),
    x = NULL,
    ps = NULL,
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  idx <- seq_len(ca$n_pred %||% 1L)
  .cmgpd_progress_step(progress_ctx, "Aggregating ATE estimates")
  .causal_aggregate_ate(ca, idx = idx, effect_type = "ate")
}

#' Average treatment effects standardized to treated covariates
#'
#' \code{att()} computes the average treatment effect on the treated.
#'
#' @details
#' The estimand is
#' \deqn{\mathrm{ATT} = E\{Y(1) - Y(0) \mid A = 1\},}
#' approximated by marginalizing over the empirical covariate distribution of
#' treated units.
#'
#' @inheritParams ate.causalmixgpd_causal_fit
#' @return An object of class \code{"causalmixgpd_ate"} containing the ATT
#'   summary, optional intervals, and the arm-specific predictive objects used
#'   in the aggregation. The returned object includes a top-level
#'   \code{$fit_df} data frame for direct extraction.
#' @seealso \code{\link{ate}}, \code{\link{qtt}}, \code{\link{cate}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb, show_progress = FALSE)
#' att(fit, interval = "credible", nsim_mean = 100)
#' }
#' @export
att <- function(fit,
                newdata = NULL,
                y = NULL,
                type = c("mean", "rmean"),
                cutoff = NULL,
                interval = "credible",
                level = 0.95,
                nsim_mean = 200L,
                show_progress = TRUE) {
  .causal_validate_fit(fit)
  type <- match.arg(type)
  nsim_mean <- .causal_validate_nsim_mean(nsim_mean)
  iv <- .causal_validate_interval(interval = interval, level = level)
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "att"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Preparing ATT inputs")

  .causal_warn_ignored_marginal_inputs(fn = "att", newdata = newdata, y = y, conditional_fn = "cate")
  compute_interval <- iv$compute_interval
  interval <- iv$interval
  level <- iv$level

  x_pred <- .causal_training_x_subset(fit, subset = "treated")
  n_pred <- if (!is.null(x_pred)) nrow(as.matrix(x_pred)) else 1L

  ps_meta <- fit$bundle$meta$ps %||% list()
  ps_enabled <- isTRUE(ps_meta$enabled) && .causal_is_conditional_model(fit) && !is.null(x_pred)
  ps_scale <- fit$bundle$meta$ps_scale %||% "logit"
  ps_summary <- fit$bundle$meta$ps_summary %||% "mean"
  ps_clamp <- fit$bundle$meta$ps_clamp %||% 1e-6

  ps_prob <- NULL
  ps_cov <- NULL
  if (ps_enabled) {
    .cmgpd_progress_step(progress_ctx, "Preparing propensity-score adjustment")
    ps_fit_use <- fit$ps_fit
    ps_bundle_use <- fit$bundle$design
    if (is.null(ps_fit_use)) {
      ps_model_try <- (fit$outcome_fit$trt$ps_model %||% fit$outcome_fit$con$ps_model %||% NULL)
      if (!is.null(ps_model_try)) {
        ps_fit_use <- ps_model_try$fit
        ps_bundle_use <- ps_model_try$bundle
      }
    }
    if (is.null(ps_fit_use) || is.null(ps_bundle_use)) {
      warning("Causal fit missing PS model; proceeding without PS adjustment.", call. = FALSE)
      ps_prob <- NULL
    } else {
      ps_prob <- .compute_ps_from_fit(
        ps_fit = ps_fit_use,
        ps_bundle = ps_bundle_use,
        X_new = x_pred,
        summary = ps_summary,
        clamp = ps_clamp
      )
      ps_cov <- .apply_ps_scale(ps_prob, scale = ps_scale, clamp = ps_clamp)
    }
  } else {
    .cmgpd_progress_step(progress_ctx, "Skipping propensity-score adjustment")
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated-arm effects")
  pr_trt <- predict(
    fit$outcome_fit$trt,
    newdata = x_pred,
    type = type,
    cutoff = cutoff,
    ps = ps_cov,
    interval = if (compute_interval) interval else NULL,
    level = level,
    nsim_mean = nsim_mean,
    store_draws = TRUE,
    show_progress = FALSE
  )
  .cmgpd_progress_step(progress_ctx, "Predicting control-arm effects")
  pr_con <- predict(
    fit$outcome_fit$con,
    newdata = x_pred,
    type = type,
    cutoff = cutoff,
    ps = ps_cov,
    interval = if (compute_interval) interval else NULL,
    level = level,
    nsim_mean = nsim_mean,
    store_draws = TRUE,
    show_progress = FALSE
  )

  meta <- fit$bundle$meta %||% list()
  ca <- list(
    trt = pr_trt,
    con = pr_con,
    level = level,
    interval = if (compute_interval) interval else "none",
    n_pred = n_pred,
    nsim_mean = .causal_mean_nsim(nsim_mean, pr_trt, pr_con),
    x = NULL,
    ps = NULL,
    meta = list(
      ps_enabled = ps_enabled,
      ps_scale = ps_scale,
      ps_summary = ps_summary,
      backend = meta$backend,
      kernel = meta$kernel,
      GPD = meta$GPD
    )
  )
  idx <- seq_len(ca$n_pred %||% 1L)
  .cmgpd_progress_step(progress_ctx, "Aggregating ATT estimates")
  .causal_aggregate_ate(ca, idx = idx, effect_type = "att")
}


#' Restricted-mean ATE helper
#'
#' \code{ate_rmean()} is a convenience wrapper for restricted-mean treatment
#' effects when the ordinary mean is unstable or undefined.
#'
#' @details
#' The restricted-mean estimand replaces \eqn{Y(a)} by
#' \eqn{\min\{Y(a), c\}}, so the contrast remains finite even when the fitted
#' GPD tail implies \eqn{\xi \ge 1}.
#'
#' @inheritParams cate.causalmixgpd_causal_fit
#' @param cutoff Finite numeric cutoff for the restricted mean.
#' @return A \code{"causalmixgpd_ate"} object computed via \code{\link{ate}}
#'   for unconditional fits or \code{\link{cate}} for conditional fits. The
#'   returned object includes a top-level \code{$fit_df} data frame for direct
#'   extraction.
#' @seealso \code{\link{ate}}, \code{\link{cate}}, \code{\link{predict.mixgpd_fit}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          GPD = TRUE, components = 3,
#'                          mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb)
#' ate_rm <- ate_rmean(fit, cutoff = 10, interval = "credible")
#' }
#' @export
ate_rmean <- function(fit,
                      newdata = NULL,
                      cutoff,
                      interval = "credible",
                      level = 0.95,
                      nsim_mean = 200L,
                      show_progress = TRUE) {
  .causal_validate_fit(fit)
  .causal_validate_interval(interval = interval, level = level)
  nsim_mean <- .causal_validate_nsim_mean(nsim_mean)
  if (.causal_is_conditional_model(fit)) {
    cate(fit = fit,
         newdata = newdata,
         type = "rmean",
         cutoff = cutoff,
         interval = interval,
         level = level,
         nsim_mean = nsim_mean,
         show_progress = show_progress)
  } else {
    ate(fit = fit,
        newdata = newdata,
        y = NULL,
        type = "rmean",
        cutoff = cutoff,
        interval = interval,
        level = level,
        nsim_mean = nsim_mean,
        show_progress = show_progress)
  }
}


#' Predict arm-specific and contrast-scale quantities from a causal fit
#'
#' \code{predict.causalmixgpd_causal_fit()} is the causal counterpart to
#' \code{\link{predict.mixgpd_fit}}. It coordinates the treated and control arm
#' predictions so that both sides use the same covariate rows and the same PS
#' adjustment.
#'
#' @details
#' For each prediction row \eqn{x}, the function evaluates arm-specific
#' posterior predictive quantities based on
#' \eqn{F_1(y \mid x)} and \eqn{F_0(y \mid x)}. Mean
#' and quantile outputs are returned on the treatment-effect scale, while
#' density, survival, and probability outputs retain both arm-specific curves.
#' For outcome kernels with a finite analytical mean, the mean path uses
#' analytical per-draw means; restricted means remain simulation-based.
#'
#' If a PS model is stored in the fit, the same estimated score is supplied to
#' both arms unless the user overrides it with \code{ps}. This is the main
#' prediction entry point used internally by \code{\link{ate}}, \code{\link{qte}},
#' \code{\link{cate}}, and \code{\link{cqte}}.
#'
#' @inheritParams predict.mixgpd_fit
#' @param object A \code{"causalmixgpd_causal_fit"} object returned by
#'   \code{run_mcmc_causal()}.
#' @param ps Optional numeric vector of propensity scores aligned with
#'   \code{newdata}. When provided, the supplied scores are used instead of
#'   recomputing them from the stored PS model (needed only for custom inputs).
#' @param id Optional identifier for prediction rows. Provide either a column name
#'   in \code{newdata} or a vector of length \code{nrow(newdata)}. The id column
#'   is excluded from analysis.
#' @param type Prediction type:
#'   \itemize{
#'     \item \code{"mean"}: posterior predictive mean treatment effect
#'     \item \code{"quantile"}: posterior predictive quantile treatment effect
#'     \item \code{"density"}: arm-specific posterior predictive densities
#'     \item \code{"survival"}: arm-specific posterior predictive survival functions
#'     \item \code{"prob"}: arm-specific posterior predictive probabilities
#'     \item \code{"sample"}: paired posterior predictive samples
#'   }
#' @param nsim Number of posterior predictive samples when \code{type = "sample"}.
#' @param interval Character or NULL; type of credible interval:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @param store_draws Logical; whether to store treatment-effect sample draws in
#'   the returned object when \code{type = "sample"}.
#' @param ... Additional arguments forwarded to per-arm
#'   \code{\link{predict.mixgpd_fit}} calls.
#' @return For \code{"mean"} and \code{"quantile"}, a causal prediction object
#'   whose \code{$fit} component reports treated-minus-control posterior
#'   summaries. For \code{"density"}, \code{"survival"}, and \code{"prob"},
#'   the \code{$fit} component contains side-by-side treated and control
#'   summaries evaluated on the supplied \code{y} grid. For \code{"sample"},
#'   the returned object contains paired treated, control, and treatment-effect
#'   posterior predictive samples. Sample outputs also include long-form data
#'   frames \code{$fit_df}, \code{$trt_fit_df}, and \code{$con_fit_df} for
#'   direct extraction.
#' @seealso \code{\link{predict.mixgpd_fit}}, \code{\link{ate}},
#'   \code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
#' @examples
#' \donttest{
#' N <- 25
#' X <- data.frame(x1 = stats::rnorm(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N)) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
#'                          mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
#' fit <- run_mcmc_causal(cb)
#' predict(fit, newdata = X[1:10, , drop = FALSE], type = "quantile", index = c(0.25, 0.5, 0.75))
#' predict(fit, newdata = X[1:10, ], type = "mean", interval = "hpd")  # HPD intervals
#' predict(fit, newdata = X[1:10, ], type = "mean", interval = NULL)   # No intervals
#' }
#' @export
predict.causalmixgpd_causal_fit <- function(object,
                                        newdata = NULL,
                                        y = NULL,
                                        ps = NULL,
                                        id = NULL,
                                        type = c("mean", "quantile", "density", "survival", "prob", "sample"),
                                        p = NULL,
                                        index = NULL,
                                        nsim = NULL,
                                        interval = "credible",
                                        probs = c(0.025, 0.5, 0.975),
                                        store_draws = TRUE,
                                        nsim_mean = 200L,
                                        ncores = 1L,
                                        show_progress = TRUE,
                                        ...) {
  stopifnot(inherits(object, "causalmixgpd_causal_fit"))
  dots <- list(...)
  progress_ctx <- .cmgpd_progress_start(
    total_steps = 5L,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "predict.causalmixgpd_causal_fit"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Resolving causal prediction inputs")

  x <- newdata
  id_info <- .resolve_predict_id(x, id = id)
  x <- id_info$x
  id_vec <- id_info$id

  type <- match.arg(type)

  # Handle interval: NULL/"none" means no interval, otherwise match to credible/hpd
  compute_interval <- TRUE
  if (is.character(interval) && length(interval) == 1L && identical(tolower(interval), "none")) {
    interval <- NULL
  }
  if (is.null(interval)) {
    compute_interval <- FALSE
    interval <- "credible"  # placeholder for downstream calls
  } else {
    interval <- match.arg(interval, choices = c("credible", "hpd"))
  }

  ncores_pred <- as.integer(ncores)
  if (!is.null(dots$workers)) {
    ncores_pred <- as.integer(dots$workers)
  } else if (isTRUE(dots$parallel)) {
    ncores_pred <- max(2L, ncores_pred)
  }
  if (is.na(ncores_pred) || ncores_pred < 1L) {
    stop("'ncores' must be an integer >= 1.", call. = FALSE)
  }
  ndraws_pred <- dots$ndraws_pred %||% NULL
  chunk_size <- dots$chunk_size %||% NULL

  bundle <- object$bundle %||% list()

  X_train <- bundle$data$X %||% NULL
  has_X <- !is.null(X_train)
  x_mat <- if (!is.null(x)) as.matrix(x) else NULL
  n_pred_default <- if (has_X) {
    nrow(X_train)
  } else if (type %in% c("density", "survival", "prob") && !is.null(y)) {
    length(as.numeric(y))
  } else {
    1L
  }
  n_pred <- if (!is.null(x_mat)) nrow(x_mat) else n_pred_default
  if (!is.null(id_vec) && length(id_vec) != n_pred) {
    stop("Length of 'id' must match the number of prediction rows.", call. = FALSE)
  }
  id_vals <- if (!is.null(id_vec)) id_vec else seq_len(n_pred)

  ps_enabled <- isTRUE(bundle$meta$ps$enabled) && has_X
  ps_include_in_outcome <- isTRUE(bundle$meta$ps$include_in_outcome)
  if (is.null(ps_include_in_outcome)) ps_include_in_outcome <- ps_enabled
  ps_model_type <- bundle$meta$ps$model_type %||% FALSE
  ps_scale <- bundle$meta$ps_scale %||% "logit"
  ps_summary <- bundle$meta$ps_summary %||% "mean"
  ps_clamp <- bundle$meta$ps_clamp %||% 1e-6

  if (type == "quantile") {
    # Alias index <-> p for consistency with predict.mixgpd_fit()
    if (!is.null(index) && is.null(p)) {
      p <- index
    } else if (!is.null(index) && !is.null(p)) {
      if (!isTRUE(all.equal(as.numeric(index), as.numeric(p)))) {
        stop("Provide only one of 'p' or 'index' for causal quantile prediction.", call. = FALSE)
      }
    }
    if (is.null(p) || length(p) == 0 || any(!is.finite(as.numeric(p)))) {
      stop("Causal predict for quantile requires one or more finite probabilities in 'p' or 'index'.", call. = FALSE)
    }
    p <- as.numeric(p)
    if (any(p <= 0 | p >= 1)) {
      stop("Probabilities in 'p'/'index' must be in (0, 1).", call. = FALSE)
    }
  }

  if (type %in% c("density", "survival", "prob")) {
    if (is.null(y)) stop("Causal predict for density/survival/prob requires 'y'.", call. = FALSE)
    y <- as.numeric(y)
    if (!length(y) || any(!is.finite(y))) stop("'y' must be a finite numeric vector.", call. = FALSE)
    if (length(y) != n_pred) {
      stop("Length of 'y' must match the number of prediction rows (nrow(x) or training X).", call. = FALSE)
    }
  }

  ps_full <- NULL
  ps_cov <- NULL
  .cmgpd_progress_step(progress_ctx, "Preparing propensity-score inputs")
  if (ps_enabled) {
    if (!is.null(ps)) {
      ps_full <- as.numeric(ps)
      eps <- as.numeric(ps_clamp)[1]
      if (is.finite(eps) && eps > 0) {
        ps_full <- pmin(pmax(ps_full, eps), 1 - eps)
      }
    } else if (!is.null(x_mat)) {
      if (is.null(object$ps_fit)) {
        stop(sprintf("Causal fit missing PS model (%s); cannot compute propensity scores for newdata.", ps_model_type), call. = FALSE)
      }
      ps_full <- .compute_ps_from_fit(
        ps_fit = object$ps_fit,
        ps_bundle = bundle$design,
        X_new = x_mat,
        summary = ps_summary,
        clamp = ps_clamp
      )
    } else {
      ps_full <- object$ps_hat %||% NULL
    }

    if (!is.null(ps_full) && length(ps_full) != n_pred) {
      stop("Length of 'ps' must equal the number of prediction rows (nrow(x)).", call. = FALSE)
    }
    if (!is.null(ps_full)) {
      ps_cov <- .apply_ps_scale(ps_full, scale = ps_scale, clamp = ps_clamp)
    }
  }

  ps_trt <- if (ps_enabled && ps_include_in_outcome) ps_cov else NULL
  ps_con <- if (ps_enabled && ps_include_in_outcome) ps_cov else NULL

  .return_out <- function(obj) {
    .cmgpd_progress_step(progress_ctx, "Assembling causal prediction output")
    obj
  }

  .extract_stats <- function(pr, n_pred) {
    fit <- pr$fit
    if (is.data.frame(fit)) {
      if ("id" %in% names(fit)) {
        fit <- fit[order(fit$id), , drop = FALSE]
      }
      est <- if ("estimate" %in% names(fit)) fit$estimate else as.numeric(fit[[1]])
      lower <- if ("lower" %in% names(fit)) fit$lower else rep(NA_real_, length(est))
      upper <- if ("upper" %in% names(fit)) fit$upper else rep(NA_real_, length(est))
    } else if (is.matrix(fit)) {
      est <- as.numeric(fit[, 1])
      lower <- rep(NA_real_, length(est))
      upper <- rep(NA_real_, length(est))
    } else {
      est <- as.numeric(fit)
      lower <- rep(NA_real_, length(est))
      upper <- rep(NA_real_, length(est))
    }
    if (length(est) == 1L && n_pred > 1L) {
      est <- rep(est, n_pred)
      lower <- rep(lower, n_pred)
      upper <- rep(upper, n_pred)
    }
    if (length(est) != n_pred) {
      stop("Unexpected prediction length in causal predict.", call. = FALSE)
    }
    list(estimate = est, lower = lower, upper = upper)
  }

  .extract_curve_stats <- function(pr, n_pred, id_vals, y_vec, value_name) {
    fit <- pr$fit
    if (!is.data.frame(fit)) {
      return(.extract_stats(pr, n_pred))
    }

    fit_df <- fit
    val_col <- if (value_name %in% names(fit_df)) {
      value_name
    } else if ("estimate" %in% names(fit_df)) {
      "estimate"
    } else {
      NULL
    }
    if (is.null(val_col)) {
      stop("Unexpected format from underlying predict() for causal curve prediction.", call. = FALSE)
    }

    if (!("id" %in% names(fit_df))) {
      if (nrow(fit_df) != n_pred) {
        stop("Unexpected prediction length in causal predict.", call. = FALSE)
      }
      fit_df$id <- id_vals
    }

    if ("y" %in% names(fit_df)) {
      pick_idx <- vapply(seq_along(id_vals), function(i) {
        rows_i <- which(fit_df$id == id_vals[i])
        if (!length(rows_i)) return(NA_integer_)

        y_i <- suppressWarnings(as.numeric(fit_df$y[rows_i]))
        if (!is.finite(y_vec[i])) return(rows_i[1])

        dist_i <- abs(y_i - y_vec[i])
        dist_i[!is.finite(dist_i)] <- Inf
        if (all(!is.finite(dist_i))) return(NA_integer_)

        rows_i[which.min(dist_i)]
      }, integer(1))
      fit_df <- fit_df[pick_idx[is.finite(pick_idx)], , drop = FALSE]
    } else {
      if (nrow(fit_df) != n_pred) {
        stop("Unexpected prediction length in causal predict.", call. = FALSE)
      }
      fit_df$y <- y_vec[match(fit_df$id, id_vals)]
    }

    fit_df <- unique(fit_df[, unique(c("id", "y", val_col, "lower", "upper")), drop = FALSE])
    fit_df <- fit_df[order(match(fit_df$id, id_vals)), , drop = FALSE]
    fit_df <- fit_df[!duplicated(fit_df$id), , drop = FALSE]

    if (nrow(fit_df) != n_pred) {
      stop("Unexpected prediction length in causal predict.", call. = FALSE)
    }

    est <- as.numeric(fit_df[[val_col]])
    lower <- if ("lower" %in% names(fit_df)) as.numeric(fit_df$lower) else rep(NA_real_, n_pred)
    upper <- if ("upper" %in% names(fit_df)) as.numeric(fit_df$upper) else rep(NA_real_, n_pred)
    list(estimate = est, lower = lower, upper = upper)
  }

  if (type %in% c("density", "survival", "prob")) {
    .cmgpd_progress_step(progress_ctx, "Predicting treated and control arms")
    pred_type <- if (type == "density") "density" else "survival"
    x_pred <- if (!is.null(x_mat)) x_mat else if (has_X) X_train else NULL
    y_vec <- y
    id_arg <- if (!is.null(x_pred)) id_vals else NULL

    pr_trt <- predict(
      object$outcome_fit$trt,
      newdata = x_pred,
      y = y_vec,
      ps = ps_trt,
      id = id_arg,
      type = pred_type,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = FALSE,
      show_progress = FALSE,
      ncores = ncores_pred,
      ...
    )
    pr_con <- predict(
      object$outcome_fit$con,
      newdata = x_pred,
      y = y_vec,
      ps = ps_con,
      id = id_arg,
      type = pred_type,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = FALSE,
      show_progress = FALSE,
      ncores = ncores_pred,
      ...
    )

    curve_value_name <- if (pred_type == "density") "density" else "survival"
    trt_stats <- .extract_curve_stats(pr_trt, n_pred, id_vals = id_vals, y_vec = y_vec, value_name = curve_value_name)
    con_stats <- .extract_curve_stats(pr_con, n_pred, id_vals = id_vals, y_vec = y_vec, value_name = curve_value_name)

    if (type == "prob") {
      trt_stats <- list(
        estimate = 1 - trt_stats$estimate,
        lower = 1 - trt_stats$upper,
        upper = 1 - trt_stats$lower
      )
      con_stats <- list(
        estimate = 1 - con_stats$estimate,
        lower = 1 - con_stats$upper,
        upper = 1 - con_stats$lower
      )
    }

    ps_col <- if (!is.null(ps_full)) ps_full else rep(NA_real_, n_pred)
    out <- data.frame(
      id = id_vals,
      y = y_vec,
      ps = ps_col,
      trt_estimate = trt_stats$estimate,
      trt_lower = trt_stats$lower,
      trt_upper = trt_stats$upper,
      con_estimate = con_stats$estimate,
      con_lower = con_stats$lower,
      con_upper = con_stats$upper,
      row.names = NULL
    )
    attr(out, "type") <- type
    class(out) <- c("causalmixgpd_causal_predict", class(out))
    return(.return_out(out))
  }

  id_arg <- if (!is.null(x)) id_vals else NULL
  sample_draw_idx <- NULL
  if (type == "sample") {
    nsim_use <- as.integer(nsim)
    if (is.na(nsim_use) || nsim_use < 1L) {
      nsim_use <- if (has_X) n_pred else length(bundle$data$y %||% object$outcome_fit$trt$data$y %||% 1L)
    }
    draw_trt <- .extract_draws_matrix(object$outcome_fit$trt)
    draw_con <- .extract_draws_matrix(object$outcome_fit$con)
    n_draws_common <- min(nrow(draw_trt), nrow(draw_con))
    if (!is.finite(n_draws_common) || n_draws_common < 1L) {
      stop("Posterior draws are unavailable for causal sample prediction.", call. = FALSE)
    }
    sample_draw_idx <- sample.int(n_draws_common, size = nsim_use, replace = TRUE)
    nsim <- nsim_use
  }

  .cmgpd_progress_step(progress_ctx, "Predicting treated arm")
  if (type == "sample") {
    pr_trt <- .predict_mixgpd(
      object$outcome_fit$trt,
      x = x,
      y = y,
      ps = ps_trt,
      id = id_arg,
      type = type,
      index = NULL,
      nsim = nsim,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = store_draws,
      nsim_mean = nsim_mean,
      show_progress = FALSE,
      ncores = ncores_pred,
      ndraws_pred = ndraws_pred,
      chunk_size = chunk_size,
      sample_draw_idx = sample_draw_idx
    )
  } else {
    pr_trt <- predict(
      object$outcome_fit$trt,
      newdata = x,
      y = y,
      ps = ps_trt,
      id = id_arg,
      type = type,
      index = if (type == "quantile") p else NULL,
      nsim = nsim,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = store_draws,
      nsim_mean = nsim_mean,
      show_progress = FALSE,
      ncores = ncores_pred,
      ...
    )
  }

  .cmgpd_progress_step(progress_ctx, "Predicting control arm")
  if (type == "sample") {
    pr_con <- .predict_mixgpd(
      object$outcome_fit$con,
      x = x,
      y = y,
      ps = ps_con,
      id = id_arg,
      type = type,
      index = NULL,
      nsim = nsim,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = store_draws,
      nsim_mean = nsim_mean,
      show_progress = FALSE,
      ncores = ncores_pred,
      ndraws_pred = ndraws_pred,
      chunk_size = chunk_size,
      sample_draw_idx = sample_draw_idx
    )
  } else {
    pr_con <- predict(
      object$outcome_fit$con,
      newdata = x,
      y = y,
      ps = ps_con,
      id = id_arg,
      type = type,
      index = if (type == "quantile") p else NULL,
      nsim = nsim,
      interval = if (compute_interval) interval else NULL,
      probs = probs,
      store_draws = store_draws,
      nsim_mean = nsim_mean,
      show_progress = FALSE,
      ncores = ncores_pred,
      ...
    )
  }

  if (type == "sample") {
    .extract_sample_fit <- function(pr, n_pred) {
      fit <- pr$fit
      mat <- if (is.null(dim(fit))) matrix(as.numeric(fit), nrow = 1L) else as.matrix(fit)
      if (nrow(mat) != n_pred) {
        if (n_pred == 1L && ncol(mat) == 1L) {
          mat <- matrix(as.numeric(mat), nrow = 1L)
        } else {
          stop("Unexpected sample dimensions in causal predict.", call. = FALSE)
        }
      }
      mat
    }

    trt_fit <- .extract_sample_fit(pr_trt, n_pred)
    con_fit <- .extract_sample_fit(pr_con, n_pred)
    if (!identical(dim(trt_fit), dim(con_fit))) {
      stop("Treated and control posterior predictive samples must have matching dimensions.", call. = FALSE)
    }

    eff_fit <- trt_fit - con_fit
    if (n_pred == 1L) {
      eff_fit_out <- as.numeric(eff_fit[1, ])
    } else {
      eff_fit_out <- eff_fit
    }

    ps_col <- if (!is.null(ps_full)) ps_full else rep(NA_real_, n_pred)
    out <- list(
      fit = eff_fit_out,
      fit_df = .values_to_long_df(eff_fit_out, id = id_vals, value_name = "sample"),
      draws = if (isTRUE(store_draws)) eff_fit_out else NULL,
      id = id_vals,
      ps = ps_col,
      nsim = ncol(trt_fit),
      trt = pr_trt,
      con = pr_con,
      trt_fit_df = pr_trt$fit_df %||% .values_to_long_df(trt_fit, id = id_vals, value_name = "sample"),
      con_fit_df = pr_con$fit_df %||% .values_to_long_df(con_fit, id = id_vals, value_name = "sample"),
      grid = NULL
    )
    attr(out, "type") <- type
    attr(out, "trt") <- pr_trt
    attr(out, "con") <- pr_con
    attr(out, "id") <- id_vals
    attr(out, "ps") <- ps_col
    class(out) <- "causalmixgpd_causal_predict"
    return(.return_out(out))
  }
  # If quantile draws are available, compute intervals on the treatment effect directly.
  if (type == "quantile" && !is.null(pr_trt$draws) && !is.null(pr_con$draws)) {
    normalize_draws <- function(draws) {
      if (is.null(dim(draws))) {
        return(array(as.numeric(draws), dim = c(length(draws), 1L, 1L)))
      }
      d <- dim(draws)
      if (length(d) == 2L) {
        # Unconditional case: draws is M x S -> convert to S x 1 x M
        M <- d[1]
        S <- d[2]
        arr <- array(NA_real_, dim = c(S, 1L, M))
        for (s in seq_len(S)) arr[s, 1L, ] <- draws[, s]
        return(arr)
      }
      if (length(d) == 3L) return(draws) # S x n_pred x M
      stop("Unexpected draw dimensions in causal quantile prediction.", call. = FALSE)
    }

    trt_draws <- normalize_draws(pr_trt$draws)
    con_draws <- normalize_draws(pr_con$draws)
    d_tr <- dim(trt_draws)
    d_co <- dim(con_draws)
    if (!identical(d_tr, d_co)) {
      common <- pmin(d_tr, d_co)
      if (any(common < 1L)) {
        stop("Treated and control posterior draws must have matching dimensions.", call. = FALSE)
      }
      trt_draws <- trt_draws[
        seq_len(common[1]),
        seq_len(common[2]),
        seq_len(common[3]),
        drop = FALSE
      ]
      con_draws <- con_draws[
        seq_len(common[1]),
        seq_len(common[2]),
        seq_len(common[3]),
        drop = FALSE
      ]
    }

    diff_draws <- trt_draws - con_draws  # S x n_pred x M
    # .posterior_summarize expects draws in last dimension
    diff_for_sum <- aperm(diff_draws, c(2, 3, 1))  # n_pred x M x S
    summ <- .posterior_summarize(
      diff_for_sum,
      probs = probs,
      interval = if (compute_interval) interval else NULL
    )

    est <- summ$estimate
    lower <- summ$lower
    upper <- summ$upper
    n_pred_eff <- if (is.matrix(est)) nrow(est) else length(est)
    id_eff <- id_vals[seq_len(min(length(id_vals), n_pred_eff))]
    if (length(id_eff) < n_pred_eff) {
      id_eff <- seq_len(n_pred_eff)
    }
    ps_col <- if (!is.null(ps_full)) {
      ps_full[seq_len(min(length(ps_full), n_pred_eff))]
    } else {
      rep(NA_real_, n_pred_eff)
    }

    if (length(p) > 1L) {
      out_df <- data.frame(
        id = rep(id_eff, each = length(p)),
        index = rep(p, times = n_pred_eff),
        ps = rep(ps_col, each = length(p)),
        estimate = as.vector(t(est)),
        lower = if (compute_interval) as.vector(t(lower)) else NA_real_,
        upper = if (compute_interval) as.vector(t(upper)) else NA_real_,
        row.names = NULL
      )
      out_df <- .reorder_predict_cols(out_df)
      attr(out_df, "type") <- type
      attr(out_df, "index") <- p
      attr(out_df, "trt") <- pr_trt
      attr(out_df, "con") <- pr_con
      class(out_df) <- c("causalmixgpd_causal_predict", class(out_df))
      return(.return_out(out_df))
    }

    est_vec <- as.numeric(est)
    lower_vec <- if (compute_interval) as.numeric(lower) else rep(NA_real_, n_pred_eff)
    upper_vec <- if (compute_interval) as.numeric(upper) else rep(NA_real_, n_pred_eff)
    out <- data.frame(
      id = id_eff,
      index = rep(p, length.out = n_pred_eff),
      ps = ps_col,
      estimate = est_vec,
      lower = lower_vec,
      upper = upper_vec,
      row.names = NULL
    )
    out <- .reorder_predict_cols(out)
    attr(out, "type") <- type
    attr(out, "index") <- p
    attr(out, "trt") <- pr_trt
    attr(out, "con") <- pr_con
    class(out) <- c("causalmixgpd_causal_predict", class(out))
    return(.return_out(out))
  }
  # Special handling for quantile with possibly multiple probabilities
  if (type == "quantile" && length(p) > 1L) {
    # Expect pr_trt$fit and pr_con$fit to be data.frames with rows ordered per index block:
    # index varies slowest in returned fit: index repeated each n_pred, with ids within block.
    fit_tr <- pr_trt$fit
    fit_co <- pr_con$fit
    if (!is.data.frame(fit_tr) || !is.data.frame(fit_co) ||
        !("index" %in% names(fit_tr)) || !("id" %in% names(fit_tr))) {
      stop("Unexpected format from underlying predict() for multiple quantiles.", call. = FALSE)
    }
    M <- length(p)
    # Ensure expected row count
    if (nrow(fit_tr) != n_pred * M || nrow(fit_co) != n_pred * M) {
      stop("Unexpected prediction length in causal predict.", call. = FALSE)
    }
    # Build matrices: columns = probs, rows = prediction rows
    trt_mat <- matrix(NA_real_, nrow = n_pred, ncol = M)
    trt_lower <- matrix(NA_real_, nrow = n_pred, ncol = M)
    trt_upper <- matrix(NA_real_, nrow = n_pred, ncol = M)
    con_mat <- matrix(NA_real_, nrow = n_pred, ncol = M)
    con_lower <- matrix(NA_real_, nrow = n_pred, ncol = M)
    con_upper <- matrix(NA_real_, nrow = n_pred, ncol = M)
    key_tr <- paste(fit_tr$id, fit_tr$index, sep = "|")
    key_co <- paste(fit_co$id, fit_co$index, sep = "|")
    key_ref <- paste(rep(id_vals, each = M), rep(p, times = n_pred), sep = "|")
    idx_tr <- match(key_ref, key_tr)
    idx_co <- match(key_ref, key_co)
    if (anyNA(idx_tr) || anyNA(idx_co)) {
      stop("Unexpected prediction ordering in causal predict.", call. = FALSE)
    }
    fit_tr <- fit_tr[idx_tr, , drop = FALSE]
    fit_co <- fit_co[idx_co, , drop = FALSE]
    trt_mat <- matrix(as.numeric(fit_tr$estimate), nrow = n_pred, ncol = M, byrow = TRUE)
    con_mat <- matrix(as.numeric(fit_co$estimate), nrow = n_pred, ncol = M, byrow = TRUE)
    trt_lower <- if ("lower" %in% names(fit_tr)) matrix(as.numeric(fit_tr$lower), nrow = n_pred, ncol = M, byrow = TRUE) else matrix(NA_real_, nrow = n_pred, ncol = M)
    trt_upper <- if ("upper" %in% names(fit_tr)) matrix(as.numeric(fit_tr$upper), nrow = n_pred, ncol = M, byrow = TRUE) else matrix(NA_real_, nrow = n_pred, ncol = M)
    con_lower <- if ("lower" %in% names(fit_co)) matrix(as.numeric(fit_co$lower), nrow = n_pred, ncol = M, byrow = TRUE) else matrix(NA_real_, nrow = n_pred, ncol = M)
    con_upper <- if ("upper" %in% names(fit_co)) matrix(as.numeric(fit_co$upper), nrow = n_pred, ncol = M, byrow = TRUE) else matrix(NA_real_, nrow = n_pred, ncol = M)
    diff_mat <- trt_mat - con_mat
    diff_lower <- trt_lower - con_upper
    diff_upper <- trt_upper - con_lower

    ps_col <- if (!is.null(ps_full)) ps_full else rep(NA_real_, n_pred)
    out_df <- data.frame(
      id = rep(id_vals, each = M),
      index = rep(p, times = n_pred),
      ps = rep(ps_col, each = M),
      estimate = as.vector(t(diff_mat)),
      lower = as.vector(t(diff_lower)),
      upper = as.vector(t(diff_upper)),
      row.names = NULL
    )
    out_df <- .reorder_predict_cols(out_df)
    attr(out_df, "type") <- type
    attr(out_df, "index") <- p
    attr(out_df, "trt") <- pr_trt
    attr(out_df, "con") <- pr_con
    class(out_df) <- c("causalmixgpd_causal_predict", class(out_df))
    return(.return_out(out_df))
  }

  trt_stats <- .extract_stats(pr_trt, n_pred)
  con_stats <- .extract_stats(pr_con, n_pred)

  diff_est <- trt_stats$estimate - con_stats$estimate
  diff_lower <- trt_stats$lower - con_stats$lower
  diff_upper <- trt_stats$upper - con_stats$upper

  ps_col <- if (!is.null(ps_full)) ps_full else rep(NA_real_, n_pred)
  out <- data.frame(
    id = id_vals,
    ps = ps_col,
    estimate = diff_est,
    lower = diff_lower,
    upper = diff_upper,
    row.names = NULL
  )
  out <- .reorder_predict_cols(out)
  attr(out, "type") <- type
  if (type == "quantile") attr(out, "index") <- p
  attr(out, "trt") <- pr_trt
  attr(out, "con") <- pr_con
  class(out) <- c("causalmixgpd_causal_predict", class(out))
  .return_out(out)
}


.ps_nimble_code <- function(lines) {
  txt <- paste(c("{", lines, "}"), collapse = "\n")
  expr <- parse(text = txt)[[1]]
  do.call(nimble::nimbleCode, list(expr))
}

.ps_model_code <- function(model = c("logit", "probit", "naive")) {
  model <- match.arg(model)

  if (identical(model, "logit")) {
    return(.ps_nimble_code(c(
      "for (i in 1:N) {",
      "  A[i] ~ dbern(pi[i])",
      "  logit(pi[i]) <- inprod(X[i, 1:P], beta[1:P])",
      "}",
      "for (j in 1:P) {",
      "  beta[j] ~ dnorm(beta_mean, sd = beta_sd)",
      "}"
    )))
  }

  if (identical(model, "probit")) {
    return(.ps_nimble_code(c(
      "for (i in 1:N) {",
      "  A[i] ~ dbern(pi[i])",
      "  probit(pi[i]) <- inprod(X[i, 1:P], beta[1:P])",
      "}",
      "for (j in 1:P) {",
      "  beta[j] ~ dnorm(beta_mean, sd = beta_sd)",
      "}"
    )))
  }

  .ps_nimble_code(c(
    "for (i in 1:N) {",
    "  A[i] ~ dbern(pi_prior)",
    "  for (j in 1:P) {",
    "    X[i, j] ~ dnorm(mu[A[i] + 1, j], sd = sigma[A[i] + 1, j])",
    "  }",
    "}",
    "pi_prior ~ dbeta(1, 1)",
    "for (k in 1:2) {",
    "  for (j in 1:P) {",
    "    mu[k, j] ~ dnorm(mu_mean, sd = mu_sd)",
    "    sigma[k, j] ~ dunif(sigma_min, sigma_max)",
    "  }",
    "}"
  ))
}

.build_ps_bundle <- function(A, X, spec, mcmc) {
  if (!is.matrix(X)) X <- as.matrix(X)
  N <- nrow(X)
  P <- ncol(X)

  if (isTRUE(spec$include_intercept)) {
    X <- cbind(`(Intercept)` = 1, X)
    P <- ncol(X)
  }

  model <- spec$model %||% "logit"
  if (!model %in% c("logit", "probit", "naive")) {
    stop("Unsupported PS model. Supported: logit, probit, naive.", call. = FALSE)
  }

  code <- .ps_model_code(model)

  constants <- list(N = N, P = P)

  # Set model-specific constants, data, inits, and monitors
  if (model %in% c("logit", "probit")) {
    constants$beta_mean <- as.numeric(spec$prior$mean %||% 0)
    constants$beta_sd <- as.numeric(spec$prior$sd %||% 2)
    data <- list(A = as.integer(A), X = X)
    inits <- list(beta = rep(0, P))
    monitors <- c("beta")
  } else if (model == "naive") {
    # Naive Bayes constants
    X_mean <- colMeans(X, na.rm = TRUE)
    X_sd <- apply(X, 2, sd, na.rm = TRUE)
    X_sd <- pmax(X_sd, 0.1)  # Prevent zero SD
    constants$mu_mean <- mean(X_mean, na.rm = TRUE)
    constants$mu_sd <- mean(X_sd, na.rm = TRUE)
    constants$sigma_min <- 0.05
    constants$sigma_max <- 5 * mean(X_sd, na.rm = TRUE)
    data <- list(A = as.integer(A), X = X)
    # Initialize with sample means/sds by treatment group
    init_mu <- matrix(0, nrow = 2, ncol = P)
    init_sigma <- matrix(1, nrow = 2, ncol = P)
    for (k in 0:1) {
      X_k <- X[A == k, , drop = FALSE]
      if (nrow(X_k) > 0) {
        init_mu[k + 1, ] <- colMeans(X_k, na.rm = TRUE)
        init_sigma[k + 1, ] <- pmax(apply(X_k, 2, sd, na.rm = TRUE), 0.1)
      }
    }
    inits <- list(pi_prior = 0.5, mu = init_mu, sigma = init_sigma)
    monitors <- c("pi_prior", "mu", "sigma")
  }

  # Store model type for downstream PS computation
  model_type_name <- if (model == "logit") "ps_logit"
                     else if (model == "probit") "ps_probit"
                     else if (model == "naive") "ps_naive"
                     else "ps_unknown"

  bundle <- list(
    spec = list(
      meta = list(type = model_type_name, include_intercept = isTRUE(spec$include_intercept)),
      model = model,
      prior = spec$prior
    ),
    code = code,
    constants = constants,
    dimensions = list(),
    data = data,
    inits = inits,
    monitors = monitors,
    mcmc = mcmc
  )
  class(bundle) <- "causalmixgpd_ps_bundle"
  bundle
}

.ps_design_matrix <- function(ps_bundle, X_new) {
  stopifnot(inherits(ps_bundle, "causalmixgpd_ps_bundle"))
  X_train <- ps_bundle$data$X
  if (is.null(X_train)) stop("PS bundle missing design matrix.", call. = FALSE)

  include_intercept <- isTRUE(ps_bundle$spec$meta$include_intercept)
  train_cols <- if (!is.null(colnames(X_train))) colnames(X_train) else character(ncol(X_train))
  base_cols <- if (include_intercept) setdiff(train_cols, "(Intercept)") else train_cols
  base_cols <- base_cols[!is.na(base_cols)]

  X_new <- as.matrix(X_new)
  storage.mode(X_new) <- "double"
  if (is.null(nrow(X_new)) || nrow(X_new) < 1) {
    stop("New data must have one or more rows.", call. = FALSE)
  }

  if (length(base_cols) > 0) {
    if (!is.null(colnames(X_new))) {
      if (!setequal(colnames(X_new), base_cols)) {
        stop("Column names of newdata do not match PS design.", call. = FALSE)
      }
      X_new <- X_new[, base_cols, drop = FALSE]
    } else {
      if (ncol(X_new) != length(base_cols)) {
        stop("Newdata must have the same number of columns as the original design.", call. = FALSE)
      }
    }
  } else {
    if (ncol(X_new) != 0L) {
      stop("No covariates expected in newdata for PS-only intercept models.", call. = FALSE)
    }
  }

  if (include_intercept) {
    out <- cbind(`(Intercept)` = 1, X_new)
    colnames(out) <- train_cols
  } else {
    out <- X_new
  }
  storage.mode(out) <- "double"
  out
}

.apply_ps_scale <- function(ps_prob, scale = c("prob", "logit"), clamp = 1e-6) {
  scale <- match.arg(scale)
  ps_prob <- as.numeric(ps_prob)
  if (scale == "prob") return(ps_prob)

  eps <- as.numeric(clamp)[1]
  if (is.finite(eps) && eps > 0) {
    ps_prob <- pmin(pmax(ps_prob, eps), 1 - eps)
  }
  qlogis(ps_prob)
}

.compute_ps_from_fit <- function(ps_fit, ps_bundle, X_new,
                                 summary = c("mean", "median"),
                                 clamp = 1e-6) {
  summary <- match.arg(summary)
  model_type <- ps_bundle$spec$model %||% "logit"
  samples <- as.matrix(ps_fit$mcmc$samples)

  if (model_type %in% c("logit", "probit")) {
    # Linear model PS: logit or probit
    design <- .ps_design_matrix(ps_bundle, X_new)
    beta_cols <- grep("^beta\\[[0-9]+\\]$", colnames(samples), value = TRUE)
    if (!length(beta_cols)) stop("PS beta draws not found.", call. = FALSE)

    beta_inds <- as.integer(sub("^beta\\[([0-9]+)\\]$", "\\1", beta_cols))
    order_idx <- order(beta_inds, na.last = NA)
    beta_mat <- samples[, beta_cols, drop = FALSE]
    beta_mat <- beta_mat[, order_idx, drop = FALSE]  # S x P

    if (ncol(design) != ncol(beta_mat)) {
      stop("Mismatch between PS design columns and beta coefficients.", call. = FALSE)
    }

    # Determine inverse link function
    inv_link <- if (model_type == "logit") plogis else if (model_type == "probit") pnorm else plogis

    # Posterior predictive PS: average inverse link over beta draws
    eta <- tcrossprod(beta_mat, design)  # S x n_pred
    ps_draws <- inv_link(eta)
    ps_vec <- if (summary == "mean") {
      colMeans(ps_draws)
    } else {
      apply(ps_draws, 2, stats::median)
    }
    eps <- as.numeric(clamp)[1]
    if (is.finite(eps) && eps > 0) {
      ps_vec <- pmin(pmax(ps_vec, eps), 1 - eps)
    }
    ps_vec

  } else if (model_type == "naive") {
    # Naive Bayes: use Bayes rule with learned feature distributions
    X_new <- as.matrix(X_new)
    storage.mode(X_new) <- "double"
    X_train <- ps_bundle$data$X %||% NULL
    if (is.null(X_train)) stop("PS bundle missing training design matrix.", call. = FALSE)

    include_intercept <- isTRUE(ps_bundle$spec$meta$include_intercept)
    if (include_intercept && ncol(X_new) == (ncol(X_train) - 1L)) {
      X_new <- cbind(`(Intercept)` = 1, X_new)
    }

    train_cols <- colnames(X_train)
    if (!is.null(train_cols) && length(train_cols) == ncol(X_new)) {
      new_cols <- colnames(X_new)
      if (!is.null(new_cols) && all(train_cols %in% new_cols)) {
        X_new <- X_new[, train_cols, drop = FALSE]
      }
    }

    if (ncol(X_new) != ncol(X_train)) {
      stop("Mismatch between PS training features and newdata columns.", call. = FALSE)
    }

    n_pred <- nrow(X_new)
    S <- nrow(samples)  # Number of MCMC iterations

    # Extract posterior samples
    pi_cols <- grep("^pi_prior$", colnames(samples), value = TRUE)
    mu_cols <- grep("^mu\\[", colnames(samples), value = TRUE)
    sigma_cols <- grep("^sigma\\[", colnames(samples), value = TRUE)

    if (!length(pi_cols) || !length(mu_cols) || !length(sigma_cols)) {
      stop("Naive Bayes PS samples missing (pi_prior, mu, or sigma).", call. = FALSE)
    }

    # Extract samples as matrices
    pi_samples <- samples[, pi_cols, drop = FALSE]
    mu_samples <- samples[, mu_cols, drop = FALSE]
    sigma_samples <- samples[, sigma_cols, drop = FALSE]

    # Initialize PS matrix
    ps_draws <- matrix(0, nrow = S, ncol = n_pred)

    # For each posterior sample, compute P(A=1|X) using Bayes rule
    for (s in 1:S) {
      pi_prior <- pi_samples[s, 1]  # P(A=1)

      # Reshape mu and sigma from samples to (K=2, P) matrices
      # mu and sigma are stored as mu[1,1], mu[1,2], ..., mu[2,1], mu[2,2], ...
      P <- ncol(X_new)
      mu_mat <- matrix(mu_samples[s, ], nrow = 2, ncol = P, byrow = FALSE)
      sigma_mat <- matrix(sigma_samples[s, ], nrow = 2, ncol = P, byrow = FALSE)

      # Compute likelihood P(X|A=k) for each treatment group
      # Using independence: P(X|A) = prod_j P(X_j|A)
      log_lik_t0 <- matrix(0, nrow = n_pred, ncol = P)
      log_lik_t1 <- matrix(0, nrow = n_pred, ncol = P)

      for (j in 1:P) {
        log_lik_t0[, j] <- dnorm(X_new[, j], mean = mu_mat[1, j], sd = sigma_mat[1, j], log = TRUE)
        log_lik_t1[, j] <- dnorm(X_new[, j], mean = mu_mat[2, j], sd = sigma_mat[2, j], log = TRUE)
      }

      # Sum across features (log scale)
      log_lik_t0_sum <- rowSums(log_lik_t0)
      log_lik_t1_sum <- rowSums(log_lik_t1)

      # Apply Bayes rule: P(A=1|X) = P(X|A=1)P(A=1) / [P(X|A=1)P(A=1) + P(X|A=0)P(A=0)]
      # Using log-sum-exp trick for numerical stability
      log_post_t0 <- log(1 - pi_prior) + log_lik_t0_sum
      log_post_t1 <- log(pi_prior) + log_lik_t1_sum

      # Convert back from log scale
      max_log <- pmax(log_post_t0, log_post_t1)
      ps_draws[s, ] <- exp(log_post_t1 - max_log) / (exp(log_post_t0 - max_log) + exp(log_post_t1 - max_log))
    }

    ps_vec <- if (summary == "mean") {
      colMeans(ps_draws)
    } else {
      apply(ps_draws, 2, stats::median)
    }
    eps <- as.numeric(clamp)[1]
    if (is.finite(eps) && eps > 0) {
      ps_vec <- pmin(pmax(ps_vec, eps), 1 - eps)
    }
    ps_vec
  }
}

# Combined tests from integration/
# Auto-generated by workspace refactor script

skip_if_not_test_level("ci")

# ===== BEGIN integration/test-bundle.R =====
test_that("Bundle generation: minimal y only", {

  set.seed(1)
  y <- abs(stats::rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6
  )

  expect_true(inherits(bundle, "causalmixgpd_bundle"), info = "bundle class (y only)")
  expect_true(is.list(bundle$spec$meta), info = "bundle spec meta exists (y only)")
  expect_false(isTRUE(bundle$spec$meta$has_X), info = "has_X is FALSE (y only)")
})

test_that("Bundle generation: with X", {

  set.seed(1)
  y <- abs(stats::rnorm(20)) + 0.1
  X <- data.frame(x1 = stats::rnorm(20), x2 = stats::runif(20))
  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6
  )

  expect_true(inherits(bundle, "causalmixgpd_bundle"), info = "bundle class (with X)")
  expect_true(isTRUE(bundle$spec$meta$has_X), info = "has_X is TRUE (with X)")
  expect_equal(bundle$spec$meta$P, ncol(X), info = "P matches ncol(X)")
})

test_that("Bundle generation: custom param_specs", {

  set.seed(1)
  y <- abs(stats::rnorm(20)) + 0.1
  X <- data.frame(x1 = stats::rnorm(20), x2 = stats::runif(20))

  param_specs <- list(
    bulk = list(
      mean = list(mode = "link", link = "identity")
    ),
    gpd = list(
      threshold = list(mode = "link", link = "identity"),
      tail_scale = list(mode = "link", link = "exp")
    )
  )

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "normal",
    GPD = TRUE,
    components = 6,
    param_specs = param_specs
  )

  plan <- bundle$spec$plan
  expect_equal(plan$bulk$mean$mode, "link", info = "bulk mean link mode")
  expect_equal(plan$gpd$threshold$mode, "link", info = "threshold link mode")
  expect_equal(plan$gpd$tail_scale$mode, "link", info = "tail_scale link mode")
})

test_that("Bundle generation: custom mcmc args", {

  set.seed(1)
  y <- abs(stats::rnorm(20)) + 0.1
  mcmc_args <- list(niter = 40, nburnin = 10, thin = 1, nchains = 1, seed = 1)

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6,
    mcmc = mcmc_args
  )

  expect_equal(bundle$mcmc$niter, 40, info = "mcmc niter stored")
  expect_equal(bundle$mcmc$nburnin, 10, info = "mcmc nburnin stored")
})

# ===== END integration/test-bundle.R =====

# ===== BEGIN integration/test-bundle-validation.R =====
# tests/testthat/test-bundle-validation.R
# Unit tests for build_nimble_bundle input validation and edge cases

# ======================================================================
# Input validation tests for build_nimble_bundle
# ======================================================================

test_that("build_nimble_bundle errors on empty y", {
  expect_error(
    build_nimble_bundle(
      y = numeric(0),
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4
    ),
    regexp = "non-empty"
  )
})

test_that("build_nimble_bundle errors on NULL y", {
  expect_error(
    build_nimble_bundle(
      y = NULL,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4
    ),
    regexp = "non-empty|length"
  )
})

test_that("build_nimble_bundle errors on components < 2", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  expect_error(
    build_nimble_bundle(
      y = y,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 1
    ),
    regexp = ">= 2"
  )
})

test_that("build_nimble_bundle errors on invalid epsilon", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1

  expect_error(
    build_nimble_bundle(
      y = y,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4,
      epsilon = -0.1
    ),
    regexp = "epsilon|\\[0.*1\\)"
  )

  expect_error(
    build_nimble_bundle(
      y = y,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4,
      epsilon = 1.5
    ),
    regexp = "epsilon|\\[0.*1\\)"
  )

  expect_error(
    build_nimble_bundle(
      y = y,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4,
      epsilon = NA
    ),
    regexp = "epsilon|numeric"
  )
})

test_that("build_nimble_bundle errors on ps length mismatch", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  ps <- runif(10)  # wrong length

  expect_error(
    build_nimble_bundle(
      y = y,
      ps = ps,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 4
    ),
    regexp = "same length"
  )
})

# ======================================================================
# Bundle structure tests
# ======================================================================

test_that("build_nimble_bundle returns correct class", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )
  expect_s3_class(bundle, "causalmixgpd_bundle")
})

test_that("build_nimble_bundle has required components", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_true("spec" %in% names(bundle))
  expect_true("code" %in% names(bundle))
  expect_true("constants" %in% names(bundle))
  expect_true("data" %in% names(bundle))
  expect_true("inits" %in% names(bundle))
  expect_true("monitors" %in% names(bundle))
  expect_true("mcmc" %in% names(bundle))
  expect_true("epsilon" %in% names(bundle))
})

test_that("build_nimble_bundle stores epsilon value", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    epsilon = 0.05
  )
  expect_equal(bundle$epsilon, 0.05)
})

test_that("build_nimble_bundle stores custom MCMC settings", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    mcmc = list(niter = 100, nburnin = 20, thin = 2, nchains = 2, seed = 42)
  )

  expect_equal(bundle$mcmc$niter, 100)
  expect_equal(bundle$mcmc$nburnin, 20)
  expect_equal(bundle$mcmc$thin, 2)
  expect_equal(bundle$mcmc$nchains, 2)
  expect_equal(bundle$mcmc$seed, 42)
})

# ======================================================================
# Backend-specific tests
# ======================================================================

test_that("build_nimble_bundle works with CRP backend", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "crp",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_equal(bundle$spec$meta$backend, "crp")
})

test_that("build_nimble_bundle works with GPD tail", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = TRUE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_true(bundle$spec$meta$GPD)
})

# ======================================================================
# Kernel-specific tests
# ======================================================================

test_that("build_nimble_bundle works with all kernels (sb, no GPD)", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  kernels <- c("normal", "lognormal", "gamma", "invgauss", "laplace", "amoroso", "cauchy")

  for (k in kernels) {
    bundle <- build_nimble_bundle(
      y = y,
      backend = "sb",
      kernel = k,
      GPD = FALSE,
      components = 4
    )
    expect_s3_class(bundle, "causalmixgpd_bundle")
    expect_equal(bundle$spec$meta$kernel, k)
  }
})

test_that("build_nimble_bundle works with X matrix", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  X <- matrix(rnorm(40), ncol = 2)
  colnames(X) <- c("x1", "x2")

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_true(bundle$spec$meta$has_X)
  expect_equal(bundle$spec$meta$P, 2)
})

test_that("build_nimble_bundle converts data.frame X to matrix", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  X <- data.frame(x1 = rnorm(20), x2 = runif(20))

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_true(bundle$spec$meta$has_X)
})

# ======================================================================
# alpha_random parameter tests
# ======================================================================

test_that("build_nimble_bundle respects alpha_random = FALSE", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    alpha_random = FALSE
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
})

test_that("build_nimble_bundle respects alpha_random = TRUE", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    alpha_random = TRUE
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
})

# ======================================================================
# param_specs override tests
# ======================================================================

test_that("build_nimble_bundle applies param_specs overrides", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  X <- matrix(rnorm(40), ncol = 2)
  colnames(X) <- c("x1", "x2")

  param_specs <- list(
    bulk = list(
      mean = list(mode = "link", link = "identity"),
      sd = list(mode = "dist")
    )
  )

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    param_specs = param_specs
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_equal(bundle$spec$plan$bulk$mean$mode, "link")
})

# ======================================================================
# J parameter alias tests
# ======================================================================

test_that("build_nimble_bundle uses components parameter correctly", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_equal(bundle$spec$meta$components, 6)
})

# ======================================================================
# PS augmentation tests
# ======================================================================

test_that("build_nimble_bundle works with propensity score vector", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  ps <- runif(20)

  bundle <- build_nimble_bundle(
    y = y,
    ps = ps,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  # PS should be included in data
  expect_true("ps" %in% names(bundle$data))
})

test_that("build_nimble_bundle works with X and ps combined", {
  set.seed(1)
  y <- abs(rnorm(20)) + 0.1
  X <- matrix(rnorm(40), ncol = 2)
  ps <- runif(20)

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    ps = ps,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")
  expect_true(bundle$spec$meta$has_X)
})

# ===== END integration/test-bundle-validation.R =====

# ===== BEGIN integration/test-fitted.R =====
# test-fitted.R
# Tests for fitted.mixgpd_fit, params(), and plot.mixgpd_fitted
#
# Tier B (ci): Requires MCMC fit

if (!exists(".cache_enabled")) {
  helper_path <- file.path("tests", "testthat", "helper-02-cache.R")
  if (file.exists(helper_path)) source(helper_path)
}

set.seed(123)
N <- 25
y <- abs(rnorm(N)) + 0.1
X <- data.frame(x1 = rnorm(N), x2 = runif(N))

mcmc_cfg <- list(niter = 60, nburnin = 20, thin = 1, nchains = 1, seed = 1)
cache_key <- NULL
if (exists(".cache_enabled") && isTRUE(.cache_enabled())) {
  key_str <- paste("fitted", "conditional", N, mcmc_cfg$niter, mcmc_cfg$nburnin,
                   mcmc_cfg$thin, mcmc_cfg$nchains, mcmc_cfg$seed, sep = "|")
  cache_key <- .cache_hash(key_str)
}
cached <- if (!is.null(cache_key)) .cache_get(cache_key) else NULL

# Only build the fit if we're at ci level or higher
# This prevents expensive MCMC compilation at cran level
fit <- NULL
if (test_level_at_least("ci")) {
  if (!is.null(cached) && inherits(cached$fit, "mixgpd_fit")) {
    fit <- cached$fit
  } else {
    bundle <- build_nimble_bundle(
      y = y,
      X = X,
      backend = "sb",
      kernel = "normal",
      GPD = FALSE,
      components = 5,
      mcmc = mcmc_cfg
    )
    fit <- run_mcmc_bundle_manual(bundle, show_progress = FALSE)
    if (!is.null(cache_key)) .cache_set(cache_key, list(fit = fit))
  }
}

test_that("Conditional fitted returns one value per observation", {
  skip_if_not_test_level("ci")

  ftd <- fitted(fit, type = "median", level = 0.9, seed = 1)
  expect_s3_class(ftd, "mixgpd_fitted")
  expect_equal(nrow(ftd), N)
  expect_true(all(is.finite(ftd$fit)))
})

test_that("Conditional fitted supports quantile type", {
  skip_if_not_test_level("ci")

  ftd <- fitted(fit, type = "quantile", p = 0.8, level = 0.9, seed = 1)
  expect_equal(nrow(ftd), N)
  expect_true(all(is.finite(ftd$fit)))
})

# =============================================================================
# Tests for params() extractor
# =============================================================================
test_that("params() returns mixgpd_params object with expected structure", {
  skip_if_not_test_level("ci")

  p <- params(fit)

  expect_s3_class(p, "mixgpd_params")
  expect_true(is.list(p))

  # Should have alpha (concentration parameter)
  expect_true("alpha" %in% names(p))
  expect_true(is.numeric(p$alpha))
  expect_true(is.finite(p$alpha))

  # Should have weights
  expect_true("w" %in% names(p))
  expect_true(is.numeric(p$w))
  expect_true(all(p$w >= 0))
  expect_true(abs(sum(p$w) - 1) < 0.1)  # Weights should sum to ~1
})

test_that("params() print method works", {
  skip_if_not_test_level("ci")

  p <- params(fit)

  expect_output(print(p), "Posterior mean parameters")
  expect_output(print(p), "alpha")
})

# =============================================================================
# Tests for plot.mixgpd_fitted
# =============================================================================
test_that("plot.mixgpd_fitted returns diagnostic plots", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("ggplot2")

  ftd <- fitted(fit, type = "mean", level = 0.9, seed = 1)

  expect_no_error({
    plots <- plot(ftd)
  })

  expect_s3_class(plots, "mixgpd_fitted_plots")
  expect_true(is.list(plots))
  expect_true("observed_fitted_plot" %in% names(plots))
  expect_true("residual_plot" %in% names(plots))

  # Both should be ggplot objects
  expect_true(inherits(plots$observed_fitted_plot, "ggplot"))
  expect_true(inherits(plots$residual_plot, "ggplot"))
})

# ===== END integration/test-fitted.R =====

# ===== BEGIN integration/test-pit-residuals.R =====
# test-pit-residuals.R
# Focused regression checks for PIT diagnostics and draw filtering.

test_that("PIT residuals include diagnostics and drop invalid draws", {
  skip_if_not_test_level("ci")

  fit <- .load_fixture("fit_cond_small.rds")

  smp <- fit$mcmc$samples %||% fit$samples
  if (is.null(smp) || !is.list(smp)) {
    skip("Fixture has no posterior samples to patch.")
  }

  kernel <- fit$spec$meta$kernel %||% fit$spec$kernel$key
  kdef <- CausalMixGPD::get_kernel_registry()[[kernel]] %||% list()
  bulk_support <- kdef$bulk_support %||% list()
  positive_params <- names(bulk_support)[bulk_support %in%
    c("positive_sd", "positive_scale", "positive_shape", "positive_location")]
  if (!length(positive_params)) {
    skip("No positive-support parameters found to invalidate.")
  }

  ch <- as.matrix(smp[[1]])
  target <- NULL
  for (nm in positive_params) {
    cand <- grep(paste0("^", nm, "\\["), colnames(ch), value = TRUE)
    if (length(cand)) {
      target <- cand[1]
      break
    }
  }
  if (is.null(target)) {
    skip("Could not find a bulk parameter column to invalidate.")
  }

  ch[1L, target] <- -abs(ch[1L, target]) - 1
  smp[[1]] <- coda::as.mcmc(ch)

  fit_patched <- fit
  if (!is.null(fit_patched$mcmc$samples)) fit_patched$mcmc$samples <- smp
  if (!is.null(fit_patched$samples)) fit_patched$samples <- smp

  res_mean <- residuals(fit_patched, type = "pit", pit = "bayes_mean", pit_seed = 1L)
  diag_mean <- attr(res_mean, "pit_diagnostics")
  expect_true(is.list(diag_mean))
  expect_true(diag_mean$n_draws_total >= diag_mean$n_draws_valid)
  expect_true(diag_mean$n_draws_dropped >= 1)
  expect_true(is.numeric(diag_mean$n_draws_used))
  expect_equal(length(diag_mean$n_draws_used), length(res_mean))
  expect_true(all(diag_mean$n_draws_used <= diag_mean$n_draws_total))
  expect_true(all(res_mean >= 0 & res_mean <= 1, na.rm = TRUE))

  res_draw <- residuals(fit_patched, type = "pit", pit = "bayes_draw", pit_seed = 1L)
  diag_draw <- attr(res_draw, "pit_diagnostics")
  expect_true(is.list(diag_draw))
  expect_true(diag_draw$n_draws_total >= diag_draw$n_draws_valid)
  expect_true(diag_draw$n_draws_dropped >= 1)
  expect_true(is.numeric(diag_draw$n_draws_used))
  expect_equal(length(diag_draw$n_draws_used), length(res_draw))
  expect_true(all(diag_draw$n_draws_used <= diag_draw$n_draws_total))
  expect_true(all(res_draw >= 0 & res_draw <= 1, na.rm = TRUE))
})

# ===== END integration/test-pit-residuals.R =====

# ===== BEGIN integration/test-predict-rmean.R =====
# test-predict-rmean.R
# Focused regression checks for mean vs restricted mean under heavy tails.

test_that("predict(type='mean') returns Inf when xi >= 1 and rmean stays finite", {
  skip_if_not_test_level("ci")

  fit <- .load_fixture("fit_gpd_small.rds")

  smp <- fit$mcmc$samples %||% fit$samples
  if (is.null(smp) || !is.list(smp)) {
    skip("Fixture has no posterior samples to patch.")
  }

  ch <- as.matrix(smp[[1]])
  if (!("tail_shape" %in% colnames(ch))) {
    skip("Fixture has no tail_shape draws (non-GPD or unexpected structure).")
  }

  ch[1L, "tail_shape"] <- 1.5
  smp[[1]] <- coda::as.mcmc(ch)

  fit_patched <- fit
  if (!is.null(fit_patched$mcmc$samples)) fit_patched$mcmc$samples <- smp
  if (!is.null(fit_patched$samples)) fit_patched$samples <- smp

  expect_warning(
    pred_mean <- predict(fit_patched, type = "mean", interval = "none", store_draws = FALSE),
    "infinite"
  )
  expect_equal(pred_mean$fit$estimate, Inf)

  pred_rmean <- predict(fit_patched, type = "rmean", cutoff = 50, interval = "none", store_draws = FALSE)
  expect_true(all(is.finite(pred_rmean$fit$estimate)))
})

# ===== END integration/test-predict-rmean.R =====

# ===== BEGIN integration/test-simulated-data.R =====
# Tests for simulated data functions (99-simulated-data.R)

# ============================================================================
# sim_bulk_tail() tests
# ============================================================================

test_that("sim_bulk_tail() returns numeric vector of correct length", {
  result <- sim_bulk_tail(n = 100)

  expect_true(is.numeric(result))
  expect_equal(length(result), 100L)
})
test_that("sim_bulk_tail() returns positive values", {
  result <- sim_bulk_tail(n = 200)

  expect_true(all(result > 0))
})

test_that("sim_bulk_tail() seed reproducibility works", {
  result1 <- sim_bulk_tail(n = 50, seed = 123)
  result2 <- sim_bulk_tail(n = 50, seed = 123)

  expect_equal(result1, result2)
})

test_that("sim_bulk_tail() different seeds produce different results", {
  result1 <- sim_bulk_tail(n = 50, seed = 123)
  result2 <- sim_bulk_tail(n = 50, seed = 456)

  expect_false(identical(result1, result2))
})

test_that("sim_bulk_tail() returns sorted values", {
  result <- sim_bulk_tail(n = 100, seed = 42)

  expect_equal(result, sort(result))
})

test_that("sim_bulk_tail() respects tail_prob parameter", {
  # With higher tail_prob, we expect more extreme values
  result_low <- sim_bulk_tail(n = 100, tail_prob = 0.05, seed = 1)
  result_high <- sim_bulk_tail(n = 100, tail_prob = 0.30, seed = 1)

  # Higher tail_prob should have larger maximum
  # (not always deterministic, but generally true)
  expect_true(is.numeric(result_low))
  expect_true(is.numeric(result_high))
})

# ============================================================================
# sim_causal_qte() tests
# ============================================================================

test_that("sim_causal_qte() returns list with required components", {
  result <- sim_causal_qte(n = 100)

  expect_true(is.list(result))
  expect_true("y" %in% names(result))
  expect_true("t" %in% names(result))
  expect_true("X" %in% names(result))
})

test_that("sim_causal_qte() y is numeric vector of correct length", {
  result <- sim_causal_qte(n = 150)

  expect_true(is.numeric(result$y))
  expect_equal(length(result$y), 150L)
})

test_that("sim_causal_qte() t is binary (0/1)", {
  result <- sim_causal_qte(n = 200, seed = 42)

  expect_true(is.numeric(result$t) || is.integer(result$t))
  expect_true(all(result$t %in% c(0L, 1L)))
  expect_equal(length(result$t), 200L)
})

test_that("sim_causal_qte() X is data frame with 3 columns", {
  result <- sim_causal_qte(n = 100)

  expect_true(is.data.frame(result$X))
  expect_equal(ncol(result$X), 3L)
  expect_equal(nrow(result$X), 100L)
  expect_equal(names(result$X), c("x1", "x2", "x3"))
})

test_that("sim_causal_qte() seed reproducibility works", {
  result1 <- sim_causal_qte(n = 50, seed = 789)
  result2 <- sim_causal_qte(n = 50, seed = 789)

  expect_equal(result1$y, result2$y)
  expect_equal(result1$t, result2$t)
  expect_equal(result1$X, result2$X)
})

test_that("sim_causal_qte() without seed produces different results", {
  set.seed(111)
  result1 <- sim_causal_qte(n = 50)
  set.seed(222)
  result2 <- sim_causal_qte(n = 50)

  expect_false(identical(result1$y, result2$y))
})

# ============================================================================
# sim_survival_tail() tests
# ============================================================================

test_that("sim_survival_tail() returns data frame", {
  result <- sim_survival_tail(n = 100)

  expect_true(is.data.frame(result))
})

test_that("sim_survival_tail() has correct columns", {
  result <- sim_survival_tail(n = 100)

  expect_true("time" %in% names(result))
  expect_true("status" %in% names(result))
  expect_true("x1" %in% names(result))
  expect_true("x2" %in% names(result))
})

test_that("sim_survival_tail() has correct number of rows", {
  result <- sim_survival_tail(n = 150)

  expect_equal(nrow(result), 150L)
})

test_that("sim_survival_tail() time is positive", {
  result <- sim_survival_tail(n = 200, seed = 42)

  expect_true(all(result$time > 0))
})

test_that("sim_survival_tail() status is binary", {
  result <- sim_survival_tail(n = 200, seed = 42)

  expect_true(all(result$status %in% c(0L, 1L)))
})

test_that("sim_survival_tail() x2 is binary", {
  result <- sim_survival_tail(n = 200, seed = 42)

  expect_true(all(result$x2 %in% c(0L, 1L)))
})

test_that("sim_survival_tail() seed reproducibility works", {
  result1 <- sim_survival_tail(n = 50, seed = 321)
  result2 <- sim_survival_tail(n = 50, seed = 321)

  expect_equal(result1, result2)
})

test_that("sim_survival_tail() different seeds produce different results", {
  result1 <- sim_survival_tail(n = 50, seed = 321)
  result2 <- sim_survival_tail(n = 50, seed = 654)

  expect_false(identical(result1$time, result2$time))
})

# ===== END integration/test-simulated-data.R =====

# ===== BEGIN integration/test-cluster-and-causal-coverage.R =====
test_that("cluster workflows cover fit, predict, and S3 methods from cluster.R", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if(
    identical(Sys.getenv("COVERAGE"), "1") &&
      identical(Sys.getenv("DPMIXGPD_SKIP_COVR_CLUSTER_WORKFLOWS"), "1"),
    "Skipping cluster workflow coverage block under covr"
  )

  set.seed(123)
  dat <- data.frame(
    y = abs(stats::rnorm(24)) + 0.25,
    x1 = stats::rnorm(24),
    x2 = stats::runif(24)
  )

  for (tp in c("weights", "param", "both")) {
    fit <- dpmix.cluster(
      y ~ x1 + x2,
      data = dat,
      kernel = "normal",
      components = 4,
      type = tp,
      mcmc = mcmc_fast(seed = 10L)
    )
    expect_s3_class(fit, "dpmixgpd_cluster_fit")

    psm <- predict(fit, type = "psm")
    expect_s3_class(psm, "dpmixgpd_cluster_psm")
    expect_true(is.matrix(psm$psm))
    expect_equal(nrow(psm$psm), ncol(psm$psm))
    expect_equal(unname(diag(psm$psm)), rep(1, nrow(psm$psm)), tolerance = 1e-8)

    lbl_train <- predict(fit, type = "label")
    expect_s3_class(lbl_train, "dpmixgpd_cluster_labels")
    expect_equal(length(lbl_train$labels), nrow(dat))

    lbl_scores <- predict(fit, type = "label", return_scores = TRUE)
    expect_true(is.matrix(lbl_scores$scores))
    expect_equal(rowSums(lbl_scores$scores), rep(1, nrow(dat)), tolerance = 1e-8)

    nd <- dat[1:6, c("y", "x1", "x2")]
    lbl_new <- predict(fit, newdata = nd, type = "label", return_scores = TRUE)
    expect_equal(length(lbl_new$labels), nrow(nd))
    expect_equal(rowSums(lbl_new$scores), rep(1, nrow(nd)), tolerance = 1e-8)

    expect_output(print(fit), "Cluster fit")
    expect_silent(summary(fit))
    expect_silent(plot(fit, which = "psm"))
    expect_silent(plot(fit, which = "k"))
    expect_silent(plot(fit, which = "sizes"))
    expect_output(print(lbl_scores), "Cluster labels")
    expect_silent(summary(lbl_scores))
    expect_silent(plot(lbl_scores, type = "sizes"))
    expect_silent(plot(lbl_scores, type = "certainty"))
    expect_output(print(psm), "Cluster PSM")
    expect_silent(summary(psm))
    expect_silent(plot(psm, psm_max_n = nrow(psm$psm)))
  }
})

test_that("cluster helpers cover design parsing and override branches", {
  skip_if_not_test_level("ci")
  skip_if(
    identical(Sys.getenv("COVERAGE"), "1") ||
      identical(Sys.getenv("DPMIXGPD_SKIP_COVR_CLUSTER_HELPERS"), "1"),
    "Skipping cluster helper coverage block under covr"
  )

  set.seed(111)
  dat <- data.frame(
    y = abs(stats::rnorm(20)) + 0.2,
    x1 = stats::rnorm(20),
    x2 = stats::runif(20)
  )

  expect_error(
    dpmix.cluster(y ~ 1, data = dat, kernel = "normal", type = "weights", components = 4, mcmc = mcmc_fast(seed = 1L)),
    "requires covariates"
  )
  expect_error(
    dpmix.cluster(y ~ x1 + x2, data = dat, kernel = "normal", type = "weights", mcmc = mcmc_fast(seed = 1L)),
    "explicit 'components'"
  )

  expect_warning(
    fit_param <- dpmix.cluster(y ~ 1, data = dat, kernel = "normal", type = "param", mcmc = mcmc_fast(seed = 2L)),
    "using default components"
  )
  expect_error(
    predict(fit_param, type = "psm", psm_max_n = 10L),
    "PSM is O\\(n\\^2\\)"
  )

  b <- build_cluster_bundle(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    GPD = TRUE,
    type = "both",
    components = 4,
    link = list(
      bulk = list(mean = "identity"),
      gpd = list(tail_scale = list(link = "exp"))
    ),
    priors = list(
      bulk = list(sd = list(dist = "gamma", args = list(shape = 2, rate = 1))),
      gpd = list(
        tail_shape = list(dist = "normal", args = list(mean = 0, sd = 0.25)),
        tail_scale = list(dist = "normal", args = list(mean = 0, sd = 0.7))
      ),
      concentration = list(dist = "gamma", args = list(shape = 3, rate = 2))
    ),
    mcmc = mcmc_fast(seed = 7L)
  )
  expect_equal(b$spec$plan$bulk$mean$link, "identity")
  expect_equal(b$spec$plan$gpd$tail_scale$link, "exp")
  expect_equal(b$spec$plan$concentration$dist, "gamma")

  fn_beta <- getFromNamespace(".cluster_extract_beta_auto", "CausalMixGPD")
  draw_row <- c(
    "beta_tail_scale[1,1]" = 2.0,
    "beta_tail_scale[1,2]" = -1.0,
    "beta_tail_scale[2,1]" = 0.5,
    "beta_tail_scale[2,2]" = 0.25,
    "beta_tail_scale[1]" = 99.0,
    "beta_tail_scale[2]" = 98.0
  )
  expect_equal(as.numeric(fn_beta(draw_row = draw_row, base = "beta_tail_scale", comp = 1L, P = 2L)), c(2.0, -1.0))

  fn_design <- getFromNamespace(".cluster_build_design", "CausalMixGPD")
  train <- data.frame(y = c(1, 2, 3), g = factor(c("a", "b", "a")))
  trm <- stats::terms(y ~ g, data = train)
  mf <- stats::model.frame(trm, data = train)
  mm <- stats::model.matrix(stats::delete.response(trm), data = mf)
  meta <- list(
    terms = trm,
    xlevels = stats::.getXlevels(trm, mf),
    contrasts = attr(mm, "contrasts"),
    X_cols = setdiff(colnames(mm), "(Intercept)"),
    response = "y"
  )
  nd <- data.frame(y = 1, g = factor("c", levels = c("a", "b", "c")))
  expect_error(fn_design(meta = meta, newdata = nd), "unseen factor levels")
})

test_that("causal bundle and fit workflows cover PS and arm-specific branches", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if(
    identical(Sys.getenv("COVERAGE"), "1") &&
      identical(Sys.getenv("DPMIXGPD_SKIP_COVR_CAUSAL_BRANCHES"), "1"),
    "Skipping causal coverage block under covr"
  )

  sim <- sim_causal_qte(n = 30, seed = 29)
  sim$y <- abs(sim$y) + 0.2

  bundle_ps <- build_causal_bundle(
    y = sim$y,
    X = as.matrix(sim$X),
    A = sim$t,
    backend = c("sb", "crp"),
    kernel = c("normal", "gamma"),
    GPD = c(FALSE, TRUE),
    components = c(3, 4),
    PS = "logit",
    mcmc_outcome = mcmc_fast(seed = 29L),
    mcmc_ps = mcmc_fast(seed = 31L)
  )
  expect_s3_class(bundle_ps, "causalmixgpd_causal_bundle")
  expect_equal(bundle_ps$meta$backend$trt, "sb")
  expect_equal(bundle_ps$meta$backend$con, "crp")

  bundle_no_ps <- build_causal_bundle(
    y = sim$y,
    X = as.matrix(sim$X),
    A = sim$t,
    backend = c("sb", "sb"),
    kernel = c("normal", "normal"),
    GPD = FALSE,
    components = c(3, 3),
    PS = FALSE,
    mcmc_outcome = mcmc_fast(seed = 33L)
  )
  expect_s3_class(bundle_no_ps, "causalmixgpd_causal_bundle")

  fit <- dpmix.causal(
    x = sim$y,
    X = as.matrix(sim$X),
    treat = sim$t,
    backend = c("sb", "crp"),
    kernel = c("normal", "gamma"),
    components = c(3, 3),
    PS = "logit",
    mcmc = c(mcmc_fast(seed = 35L), list(show_progress = FALSE))
  )
  expect_s3_class(fit, "causalmixgpd_causal_fit")
  expect_output(print(fit), "CausalMixGPD causal fit")
  expect_output(print(summary(fit)), "Outcome fits")
  expect_s3_class(params(fit), "mixgpd_params_pair")

  Xnew <- as.matrix(sim$X[1:4, , drop = FALSE])
  pred_mean <- predict(fit, newdata =Xnew, type = "mean", nsim_mean = 20L)
  pred_quant <- predict(fit, newdata =Xnew, type = "quantile", p = c(0.25, 0.75))
  expect_s3_class(pred_mean, "causalmixgpd_causal_predict")
  expect_s3_class(pred_quant, "causalmixgpd_causal_predict")

  expect_s3_class(cate(fit, newdata = Xnew, nsim_mean = 20L, interval = "credible"), "causalmixgpd_ate")
  expect_s3_class(cqte(fit, probs = c(0.25, 0.75), newdata = Xnew, interval = "credible"), "causalmixgpd_qte")
  expect_s3_class(ate(fit, nsim_mean = 20L, interval = "credible"), "causalmixgpd_ate")
  expect_s3_class(qte(fit, probs = c(0.25, 0.75), interval = "credible"), "causalmixgpd_qte")
})

# ===== END integration/test-cluster-and-causal-coverage.R =====

# Combined tests from ci/
# Auto-generated by workspace refactor script

# ===== BEGIN ci/test-site-map-tools.R =====
skip_if_not_test_level("ci")

find_package_root <- function() {
  source_root <- Sys.getenv("DPMIXGPD_SOURCE_ROOT", unset = "")
  if (nzchar(source_root)) {
    source_root <- normalizePath(source_root, winslash = "/", mustWork = FALSE)
    if (file.exists(file.path(source_root, "DESCRIPTION")) &&
        file.exists(file.path(source_root, "tools", "site-map", "extract_site_map.R"))) {
      return(source_root)
    }
  }

  current <- normalizePath(getwd(), winslash = "/", mustWork = TRUE)
  for (i in 0:8) {
    candidate <- current
    if (file.exists(file.path(candidate, "DESCRIPTION"))) {
      return(candidate)
    }
    current <- dirname(current)
  }
  stop("Could not locate package root for site-map tests.")
}

pkg_root <- find_package_root()
source(file.path(pkg_root, "tools", "site-map", "extract_site_map.R"), local = TRUE)
source(file.path(pkg_root, "tools", "site-map", "analyze_site_map.R"), local = TRUE)

write_fixture_file <- function(path, lines) {
  dir.create(dirname(path), recursive = TRUE, showWarnings = FALSE)
  writeLines(lines, con = path, useBytes = TRUE)
}

make_workflow_pages <- function(docs_root, break_chain = FALSE) {
  for (i in 1:20) {
    file_path <- file.path(docs_root, "workflows", sprintf("v%02d-topic.html", i))
    link_html <- ""
    if (i < 20) {
      if (!(break_chain && i == 10)) {
        link_html <- sprintf("<a href=\"v%02d-topic.html\">next</a>", i + 1L)
      }
    }
    write_fixture_file(
      file_path,
      c(
        "<html><body>",
        sprintf("<h1>v%02d</h1>", i),
        link_html,
        "</body></html>"
      )
    )
  }
}

test_that("extract_site_map handles strict internal links and normalization", {
  tmp_root <- tempfile("site_map_extract_")
  docs_root <- file.path(tmp_root, "docs")
  out_root <- file.path(tmp_root, "out")
  dir.create(docs_root, recursive = TRUE, showWarnings = FALSE)

  write_fixture_file(
    file.path(docs_root, "index.html"),
    c(
      "<html><body>",
      "<a href=\"workflows/v01-start-here.html\">workflow</a>",
      "<a href=\"missing-page.html\">broken</a>",
      "<a href=\"workflows/\">workflows-dir</a>",
      "<a href=\"https://example.com\">external</a>",
      "<a href=\"mailto:test@example.com\">mail</a>",
      "<a href=\"javascript:void(0)\">js</a>",
      "<a href=\"#top\">anchor</a>",
      "</body></html>"
    )
  )
  write_fixture_file(
    file.path(docs_root, "workflows", "index.html"),
    c("<html><body>", "<a href=\"../index.html\">home</a>", "</body></html>")
  )
  write_fixture_file(
    file.path(docs_root, "workflows", "v01-start-here.html"),
    c(
      "<html><body>",
      "<a href=\"../index.html#intro\">back-home</a>",
      "<a href=\"v02-introduction\">next-no-ext</a>",
      "</body></html>"
    )
  )
  write_fixture_file(
    file.path(docs_root, "workflows", "v02-introduction.html"),
    c("<html><body>", "ok", "</body></html>")
  )

  extract_site_map(site_root = docs_root, output_dir = out_root, verbose = FALSE)
  edges <- read.csv(file.path(out_root, "website_link_edges.csv"), stringsAsFactors = FALSE)

  expect_true(any(edges$href_raw == "missing-page.html" & !as.logical(edges$target_exists)))
  expect_true(any(edges$href_raw == "../index.html#intro" & edges$target == "index.html"))
  expect_true(any(edges$href_raw == "v02-introduction" & edges$target == "workflows/v02-introduction.html"))
  expect_true(any(edges$href_raw == "workflows/" & edges$target == "workflows/index.html"))

  # External/anchor hrefs are intentionally ignored by extractor.
  expect_false(any(grepl("^https?://", edges$href_raw)))
  expect_false(any(grepl("^mailto:", edges$href_raw)))
  expect_false(any(grepl("^javascript:", edges$href_raw)))
  expect_false(any(grepl("^#", edges$href_raw)))
})

test_that("analyze_site_map gate follows broken-link and workflow-chain policy", {
  tmp_root <- tempfile("site_map_analyze_")
  docs_root <- file.path(tmp_root, "docs")
  out_root <- file.path(tmp_root, "out")
  dir.create(docs_root, recursive = TRUE, showWarnings = FALSE)

  write_fixture_file(
    file.path(docs_root, "index.html"),
    c(
      "<html><body>",
      "<a href=\"workflows/v01-topic.html\">start</a>",
      "</body></html>"
    )
  )
  make_workflow_pages(docs_root = docs_root, break_chain = FALSE)

  extract_site_map(site_root = docs_root, output_dir = out_root, verbose = FALSE)
  ok_result <- analyze_site_map(site_root = docs_root, output_dir = out_root, verbose = FALSE)

  expect_true(isTRUE(ok_result$summary$gate_pass))
  expect_equal(ok_result$summary$broken_link_count, 0)
  expect_equal(ok_result$summary$workflow_chain_break_count, 0)

  # Rebuild with a broken chain and a broken link.
  unlink(tmp_root, recursive = TRUE, force = TRUE)
  dir.create(docs_root, recursive = TRUE, showWarnings = FALSE)
  write_fixture_file(
    file.path(docs_root, "index.html"),
    c(
      "<html><body>",
      "<a href=\"workflows/v01-topic.html\">start</a>",
      "<a href=\"does-not-exist.html\">broken</a>",
      "</body></html>"
    )
  )
  make_workflow_pages(docs_root = docs_root, break_chain = TRUE)

  extract_site_map(site_root = docs_root, output_dir = out_root, verbose = FALSE)
  bad_result <- analyze_site_map(site_root = docs_root, output_dir = out_root, verbose = FALSE)

  expect_false(isTRUE(bad_result$summary$gate_pass))
  expect_gt(bad_result$summary$broken_link_count, 0)
  expect_gt(bad_result$summary$workflow_chain_break_count, 0)
})

# ===== END ci/test-site-map-tools.R =====

# ===== BEGIN ci/test-vignette-coverage.R =====
# test-vignette-coverage.R
# =============================================================================
# Tests that exercise code paths from vignettes for coverage
# These tests use short MCMC runs to be fast while still exercising the code
# =============================================================================

# Use package's test level system - these run at "ci" level and above
# They will run during coverage calculation (DPMIXGPD_TEST_LEVEL="ci")
# but skip during R CMD check (DPMIXGPD_TEST_LEVEL="cran")

# Short MCMC settings for fast execution
mcmc_short <- list(niter = 50, nburnin = 10, thin = 1, nchains = 1, seed = 1)

# Helper to suppress MCMC output
quiet_run <- function(expr) {

  nullfile <- if (.Platform$OS.type == "windows") "NUL" else "/dev/null"
  utils::capture.output(result <- force(expr), file = nullfile)
  result
}

# =============================================================================
# Unconditional Models (from v05-v09 vignettes)
# =============================================================================

test_that("unconditional CRP bulk model works (v05 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_pos200_k3", package = "CausalMixGPD")
  y <- nc_pos200_k3$y[1:50]  # Use subset for speed

  bundle <- build_nimble_bundle(
    y = y,
    backend = "crp",
    kernel = "gamma",
    GPD = FALSE,
    components = 3,
    monitor_latent = TRUE,
    mcmc = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")

  # Test print and summary
  expect_output(print(bundle))
  expect_output(print(summary(bundle)))

  # Run MCMC
  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")

  # Test S3 methods
  expect_output(print(fit))
  summ <- summary(fit)
  expect_s3_class(summ, "mixgpd_summary")
  expect_output(print(summ))

  # Test params
  p <- params(fit)
  expect_s3_class(p, "mixgpd_params")
  expect_output(print(p))

  # Test predict methods
  pred_q <- predict(fit, type = "quantile", index = c(0.5, 0.9))
  expect_s3_class(pred_q, "mixgpd_predict")

  pred_d <- predict(fit, y = y[1:5], type = "density")
  expect_s3_class(pred_d, "mixgpd_predict")

  pred_s <- predict(fit, y = y[1:5], type = "survival")
  expect_s3_class(pred_s, "mixgpd_predict")
})

test_that("unconditional SB bulk model works (v07 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_pos200_k3", package = "CausalMixGPD")
  y <- nc_pos200_k3$y[1:50]

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "lognormal",
    GPD = FALSE,
    components = 3,
    mcmc = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")

  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")
  # fitted() not supported for unconditional; use predict() only
})

test_that("unconditional SB GPD model works (v09 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_pos200_k3", package = "CausalMixGPD")
  y <- nc_pos200_k3$y[1:50]

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "gamma",
    GPD = TRUE,
    components = 3,
    mcmc = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")

  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")
})

# =============================================================================
# Conditional Models (from v10-v13 vignettes)
# =============================================================================

test_that("conditional SB bulk model works (v11 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_posX100_p3_k2", package = "CausalMixGPD")
  y <- nc_posX100_p3_k2$y[1:40]
  X <- as.matrix(nc_posX100_p3_k2$X[1:40, ])

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "lognormal",
    GPD = FALSE,
    components = 3,
    mcmc = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")

  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")

  # Test conditional predict
  x_new <- X[1:5, , drop = FALSE]
  pred_mean <- predict(fit, newdata =x_new, type = "mean", nsim_mean = 20)
  expect_s3_class(pred_mean, "mixgpd_predict")
})

test_that("conditional SB GPD model works (v13 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_posX100_p3_k2", package = "CausalMixGPD")
  y <- nc_posX100_p3_k2$y[1:40]
  X <- as.matrix(nc_posX100_p3_k2$X[1:40, ])

  bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "sb",
    kernel = "lognormal",
    GPD = TRUE,
    components = 3,
    mcmc = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_bundle")

  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")
})

# =============================================================================
# Causal Models (from v14-v19 vignettes)
# =============================================================================

test_that("causal no-X CRP model works (v14 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:80]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:80]

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = NULL,
    kernel = "gamma",
    backend = "crp",
    PS = FALSE,
    GPD = FALSE,
    components = 3,
    monitor_latent = TRUE,
    mcmc_outcome = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_causal_bundle")
  expect_output(print(bundle))
  expect_output(print(summary(bundle)))

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))
  expect_s3_class(fit, "causalmixgpd_causal_fit")

  # Test causal S3 methods
  expect_output(print(fit))
  expect_output(print(summary(fit)))

  # Test causal params
  p <- params(fit)
  expect_s3_class(p, "mixgpd_params_pair")
  expect_output(print(p))

  # Test causal predict
  pred <- predict(fit, type = "mean", nsim_mean = 20)
  expect_s3_class(pred, "causalmixgpd_causal_predict")

  # Test QTE
  qte_result <- qte(fit, probs = c(0.5), interval = "credible")
  expect_s3_class(qte_result, "causalmixgpd_qte")
  expect_output(print(qte_result))
  summ_qte <- summary(qte_result)
  expect_output(print(summ_qte))

  # Test ATE
  ate_result <- ate(fit, interval = "credible", nsim_mean = 20)
  expect_s3_class(ate_result, "causalmixgpd_ate")
  expect_output(print(ate_result))
  summ_ate <- summary(ate_result)
  expect_output(print(summ_ate))
})

test_that("causal X no-PS SB model works (v15 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:80]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:80]
  X <- as.matrix(causal_alt_real500_p4_k2$X[1:80, 1:2])

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = X,
    kernel = "lognormal",
    backend = "sb",
    PS = FALSE,
    GPD = FALSE,
    components = 3,
    mcmc_outcome = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_causal_bundle")

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))
  expect_s3_class(fit, "causalmixgpd_causal_fit")

  # Test causal predict with multiple types
  pred_q <- predict(fit, type = "quantile", p = c(0.25, 0.75))
  expect_s3_class(pred_q, "causalmixgpd_causal_predict")

  pred_d <- predict(fit, newdata =X[1:5, , drop = FALSE], y = y[1:5], type = "density")
  expect_s3_class(pred_d, "causalmixgpd_causal_predict")

  pred_s <- predict(fit, newdata =X[1:5, , drop = FALSE], y = y[1:5], type = "survival")
  expect_s3_class(pred_s, "causalmixgpd_causal_predict")
})

test_that("causal no-X SB model works (v16 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:80]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:80]

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = NULL,
    kernel = "lognormal",
    backend = "sb",
    PS = FALSE,
    GPD = FALSE,
    components = 3,
    mcmc_outcome = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_causal_bundle")

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))
  expect_s3_class(fit, "causalmixgpd_causal_fit")

  # Test QTE with multiple quantiles
  qte_result <- qte(fit, probs = c(0.25, 0.5, 0.75), interval = "credible")
  expect_s3_class(qte_result, "causalmixgpd_qte")

  # Test plot methods for QTE (returns plot objects)
  qte_plots <- plot(qte_result)
  expect_type(qte_plots, "list")

  # Test ATE with HPD interval
  ate_result <- ate(fit, interval = "hpd", nsim_mean = 20)
  expect_s3_class(ate_result, "causalmixgpd_ate")

  # Test plot methods for ATE
  ate_plots <- plot(ate_result)
  expect_type(ate_plots, "list")
})

test_that("causal with GPD tails works (v17 coverage)", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:60]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:60]

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = NULL,
    kernel = "gamma",
    backend = "sb",
    PS = FALSE,
    GPD = TRUE,
    components = 3,
    mcmc_outcome = mcmc_short
  )

  expect_s3_class(bundle, "causalmixgpd_causal_bundle")

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))
  expect_s3_class(fit, "causalmixgpd_causal_fit")

  # Test predict with mean type (uses simulation fallback for gamma+GPD)
  pred_loc <- predict(fit, type = "mean", nsim_mean = 20L)
  expect_s3_class(pred_loc, "causalmixgpd_causal_predict")
})

test_that("causal fit plot method works", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:60]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:60]

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = NULL,
    kernel = "gamma",
    backend = "crp",
    PS = FALSE,
    GPD = FALSE,
    components = 3,
    monitor_latent = TRUE,
    mcmc_outcome = mcmc_short
  )

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))

  # Test plot method for causal fit
  fit_plots <- plot(fit, arm = "both")
  expect_type(fit_plots, "list")

  # Test plot for treatment arm only
  fit_plots_trt <- plot(fit, arm = "trt")
  expect_type(fit_plots_trt, "list")

  # Test plot for control arm only
  fit_plots_con <- plot(fit, arm = "con")
  expect_type(fit_plots_con, "list")
})

test_that("causal predict plot method works", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("causal_alt_real500_p4_k2", package = "CausalMixGPD")
  y <- abs(causal_alt_real500_p4_k2$y[1:60]) + 0.01
  T_vec <- causal_alt_real500_p4_k2$A[1:60]

  bundle <- build_causal_bundle(
    y = y,
    A = T_vec,
    X = NULL,
    kernel = "gamma",
    backend = "crp",
    PS = FALSE,
    GPD = FALSE,
    components = 3,
    monitor_latent = TRUE,
    mcmc_outcome = mcmc_short
  )

  fit <- quiet_run(run_mcmc_causal(bundle, show_progress = FALSE))

  # Test predict plot
  pred <- predict(fit, type = "quantile", p = 0.5)
  pred_plots <- plot(pred)
  expect_type(pred_plots, "list")
})

# =============================================================================
# Kernel Coverage (all 7 kernels)
# =============================================================================

test_that("all kernel types work with SB backend", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_pos200_k3", package = "CausalMixGPD")
  y_pos <- nc_pos200_k3$y[1:30]
  y_real <- y_pos - mean(y_pos)  # Center for real-support kernels

  # Positive support kernels
  for (kernel in c("gamma", "lognormal", "invgauss")) {
    bundle <- build_nimble_bundle(
      y = y_pos,
      backend = "sb",
      kernel = kernel,
      GPD = FALSE,
      components = 2,
      mcmc = list(niter = 30, nburnin = 5, thin = 1, nchains = 1, seed = 1)
    )
    expect_s3_class(bundle, "causalmixgpd_bundle")

    fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
    expect_s3_class(fit, "mixgpd_fit")
  }

  # Real support kernels
  for (kernel in c("normal", "laplace", "cauchy")) {
    bundle <- build_nimble_bundle(
      y = y_real,
      backend = "sb",
      kernel = kernel,
      GPD = FALSE,
      components = 2,
      mcmc = list(niter = 30, nburnin = 5, thin = 1, nchains = 1, seed = 1)
    )
    expect_s3_class(bundle, "causalmixgpd_bundle")

    fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
    expect_s3_class(fit, "mixgpd_fit")
  }
})

test_that("amoroso kernel works", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  data("nc_pos200_k3", package = "CausalMixGPD")
  y <- nc_pos200_k3$y[1:30]

  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "amoroso",
    GPD = FALSE,
    components = 2,
    mcmc = list(niter = 30, nburnin = 5, thin = 1, nchains = 1, seed = 1)
  )
  expect_s3_class(bundle, "causalmixgpd_bundle")

  fit <- quiet_run(run_mcmc_bundle_manual(bundle, show_progress = FALSE))
  expect_s3_class(fit, "mixgpd_fit")
})

# ===== END ci/test-vignette-coverage.R =====

test_that("reviewer-facing metadata and namespace are clean", {
  desc_path <- system.file("DESCRIPTION", package = "CausalMixGPD")
  if (!nzchar(desc_path)) {
    desc_path <- test_path("..", "..", "DESCRIPTION")
  }
  desc <- read.dcf(desc_path)[1, ]
  expect_false(grepl("zenodo.19620523", desc[["Description"]], fixed = TRUE))
  expect_false(grepl("nimble", desc[["Depends"]], fixed = TRUE))
  expect_true(grepl("nimble", desc[["Imports"]], fixed = TRUE))

  ns_path <- system.file("NAMESPACE", package = "CausalMixGPD")
  if (!nzchar(ns_path)) {
    ns_path <- test_path("..", "..", "NAMESPACE")
  }
  ns <- readLines(ns_path, warn = FALSE)
  expect_false(any(grepl("^export\\(build_(code|constants|dimensions|inits|monitors)_from_spec\\)", ns)))
  expect_true(any(ns == "export(cluster_profiles)"))
  expect_true(any(ns == "S3method(print,mixgpd_predict)"))
})

test_that("causal estimand validation happens before progress output", {
  bad_fit <- structure(list(), class = "not_a_causal_fit")
  expect_error(ate(bad_fit, show_progress = TRUE), "causalmixgpd_causal_fit")
  expect_error(qte(bad_fit, show_progress = TRUE), "causalmixgpd_causal_fit")
})

test_that("prediction print and rmean plotting methods are registered", {
  pr <- list(
    type = "rmean",
    fit = data.frame(id = 1, estimate = 1, lower = 0.5, upper = 1.5),
    draws = c(0.8, 1.0, 1.2)
  )
  class(pr) <- "mixgpd_predict"
  expect_output(print(pr), "MixGPD prediction")
  skip_if_not_installed("ggplot2")
  expect_s3_class(plot(pr), "mixgpd_predict_plots")
})

test_that("causal fit print omits unavailable timing instead of printing NA timings", {
  fit <- structure(
    list(
      bundle = list(
        meta = list(
          ps = list(enabled = FALSE),
          backend = list(trt = "sb", con = "sb"),
          kernel = list(trt = "normal", con = "normal"),
          GPD = list(trt = FALSE, con = FALSE)
        )
      ),
      timing = list(total = NA_real_, ps = NA_real_, con = NA_real_, trt = NA_real_)
    ),
    class = "causalmixgpd_causal_fit"
  )
  out <- capture.output(print(fit))
  expect_false(any(grepl("Timing \\(sec\\)", out)))
})

test_that("cluster profile accessor reads summary objects", {
  s <- structure(
    list(cluster_profiles = data.frame(cluster = 1L, y_mean = 2)),
    class = c("summary.dpmixgpd_cluster_labels", "list")
  )
  expect_equal(cluster_profiles(s)$y_mean, 2)
})

test_that("reviewer-facing classes include common S3 base classes", {
  one_arm <- structure(list(), class = c("mixgpd_fit", "causalmixgpd_fit", "list"))
  causal <- structure(list(), class = c("causalmixgpd_causal_fit", "causalmixgpd_fit", "list"))
  cluster <- structure(list(), class = c("dpmixgpd_cluster_fit", "causalmixgpd_fit", "list"))
  effect <- structure(list(), class = c("causalmixgpd_ate", "causalmixgpd_effect", "list"))

  expect_s3_class(one_arm, "causalmixgpd_fit")
  expect_s3_class(causal, "causalmixgpd_fit")
  expect_s3_class(cluster, "causalmixgpd_fit")
  expect_s3_class(effect, "causalmixgpd_effect")

  expect_identical(class(one_arm)[1], "mixgpd_fit")
  expect_identical(class(causal)[1], "causalmixgpd_causal_fit")
  expect_identical(class(cluster)[1], "dpmixgpd_cluster_fit")
  expect_identical(class(effect)[1], "causalmixgpd_ate")
})

test_that("reviewer-facing replication and examples use public entrypoints", {
  root <- normalizePath(test_path("..", ".."), winslash = "/", mustWork = TRUE)
  read_if_exists <- function(...) {
    path <- file.path(root, ...)
    if (!file.exists(path)) return(character())
    readLines(path, warn = FALSE)
  }

  article <- read_if_exists("manuscript", "CausalMixGPD_JSS_article.Rnw")
  if (length(article)) {
    expect_false(any(grepl("summary\\(z_test\\)\\$cluster_profiles", article)))
    expect_true(any(grepl("cluster_profiles\\(z_test\\)", article, fixed = FALSE)))
  }

  rd <- read_if_exists("man", "dpmix.causal.Rd")
  if (length(rd)) {
    expect_true(any(grepl("\\\\examples\\{", rd)))
  }

  man_dir <- file.path(root, "man")
  if (dir.exists(man_dir)) {
    titles <- unlist(lapply(list.files(man_dir, "\\.Rd$", full.names = TRUE), function(path) {
      raw <- paste(readLines(path, warn = FALSE), collapse = "\n")
      match <- regexec("\\\\title\\{([^}]*)\\}", raw, perl = TRUE)
      regmatches(raw, match)[[1]][2]
    }), use.names = FALSE)
    expect_false(any(grepl("^[a-z]", titles)))
  }
})

test_that("ess_summary returns stable structure for mixgpd_fit", {
  skip_if_not_test_level("ci")
  set.seed(501)
  y <- abs(stats::rnorm(90)) + 0.1
  b <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6,
    mcmc = list(niter = 120, nburnin = 30, thin = 1, nchains = 1, seed = 11, timing = TRUE)
  )
  fit <- run_mcmc_bundle_manual(b, show_progress = FALSE, timing = TRUE)
  es <- ess_summary(fit)

  expect_s3_class(es, "mixgpd_ess_summary")
  expect_true(all(c("table", "overall", "meta") %in% names(es)))
  expect_true(all(c("param", "chain", "ess", "seconds", "ess_per_sec") %in% names(es$table)))
  expect_true(nrow(es$table) >= 1L)
})

test_that("ess_summary supports causal fit arm-wise summaries", {
  skip_if_not_test_level("ci")
  set.seed(502)
  n <- 120
  X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
  A <- stats::rbinom(n, 1, 0.5)
  y <- abs(0.4 * X[, 1] + A + stats::rnorm(n)) + 0.1
  cb <- build_causal_bundle(
    y = y, X = X, A = A,
    backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    mcmc_outcome = list(niter = 100, nburnin = 30, thin = 1, nchains = 1, seed = 12),
    mcmc_ps = list(niter = 80, nburnin = 20, thin = 1, nchains = 1, seed = 13),
    PS = "logit"
  )
  fit <- run_mcmc_causal(cb, show_progress = FALSE, timing = TRUE)
  es <- ess_summary(fit)
  expect_true(all(c("control", "treated") %in% unique(es$table$arm)))
})

test_that("run_mcmc_causal forwards z_update_every to both arms", {
  skip_if_not_test_level("ci")
  set.seed(521)
  n <- 100
  X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
  A <- stats::rbinom(n, 1, 0.5)
  y <- abs(0.3 * X[, 1] + A + stats::rnorm(n)) + 0.1

  cb <- build_causal_bundle(
    y = y, X = X, A = A,
    backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    mcmc_outcome = list(niter = 90, nburnin = 20, thin = 1, nchains = 1, seed = 15),
    mcmc_ps = list(niter = 70, nburnin = 20, thin = 1, nchains = 1, seed = 16),
    PS = "logit"
  )
  fit <- run_mcmc_causal(cb, show_progress = FALSE, z_update_every = 3)
  expect_equal(fit$outcome_fit$con$mcmc$z_update_every, 3L)
  expect_equal(fit$outcome_fit$trt$mcmc$z_update_every, 3L)
})

test_that("monitoring policy defaults exclude latent z and opt-in works", {
  skip_if_not_test_level("ci")
  set.seed(101)
  y <- abs(stats::rnorm(80)) + 0.1

  b_default <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6
  )
  expect_false(any(grepl("^z\\[", b_default$monitors)))

  b_latent <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6,
    monitor_latent = TRUE
  )
  expect_true(any(grepl("^z\\[", b_latent$monitors)))
})

test_that("compile cache exposes timing and reuses build/compile", {
  skip_if_not_test_level("ci")
  set.seed(102)
  y <- abs(stats::rnorm(80)) + 0.1
  m <- list(niter = 120, nburnin = 30, thin = 1, nchains = 1, seed = 42, timing = TRUE)

  b <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 6,
    mcmc = m
  )

  fit1 <- run_mcmc_bundle_manual(b, show_progress = FALSE, timing = TRUE)
  fit2 <- run_mcmc_bundle_manual(b, show_progress = FALSE, timing = TRUE)

  expect_true(is.list(fit1$timing))
  expect_true(is.list(fit2$timing))
  expect_true(all(c("build", "compile", "mcmc") %in% names(fit1$timing)))
  expect_true(all(c("build", "compile", "mcmc") %in% names(fit2$timing)))
  expect_true(isTRUE(fit2$timing$cache_hit))
  expect_lte(unname(fit2$timing$build), unname(fit1$timing$build) + 1e-8)
  expect_lte(unname(fit2$timing$compile), unname(fit1$timing$compile) + 1e-8)
})

test_that("parallel chains path preserves output contract", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("future")
  skip_if_not_installed("future.apply")

  set.seed(103)
  y <- abs(stats::rnorm(60)) + 0.1
  b <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 5,
    mcmc = list(niter = 120, nburnin = 30, thin = 1, nchains = 2, seed = 11)
  )

  fit_seq <- run_mcmc_bundle_manual(b, show_progress = FALSE, parallel_chains = FALSE, timing = TRUE)
  fit_par <- run_mcmc_bundle_manual(b, show_progress = FALSE, parallel_chains = TRUE, workers = 2, timing = TRUE)

  expect_s3_class(fit_par, "mixgpd_fit")
  expect_equal(names(fit_par), names(fit_seq))
  expect_equal(fit_par$mcmc$nchains, 2)
  expect_true(length(unique(as.integer(fit_par$mcmc$seed))) > 1)
})

test_that("causal arms can run in parallel and preserve contract", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("future")
  skip_if_not_installed("future.apply")

  set.seed(104)
  n <- 120
  x1 <- stats::rnorm(n)
  x2 <- stats::runif(n)
  A <- stats::rbinom(n, 1, stats::plogis(0.2 + 0.4 * x1))
  y <- abs(0.7 * x1 + 0.2 * x2 + A + stats::rnorm(n)) + 0.1
  X <- cbind(x1 = x1, x2 = x2)

  cb <- build_causal_bundle(
    y = y, X = X, A = A,
    backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    mcmc_outcome = list(niter = 100, nburnin = 30, thin = 1, nchains = 1, seed = 21),
    mcmc_ps = list(niter = 80, nburnin = 20, thin = 1, nchains = 1, seed = 22),
    PS = "logit"
  )

  fit_seq <- run_mcmc_causal(cb, show_progress = FALSE, parallel_arms = FALSE, timing = TRUE)
  fit_par <- run_mcmc_causal(cb, show_progress = FALSE, parallel_arms = TRUE, workers = 2, timing = TRUE)

  expect_s3_class(fit_par, "causalmixgpd_causal_fit")
  expect_equal(names(fit_par), names(fit_seq))
  expect_true(is.list(fit_par$timing))
  expect_true(all(c("total", "ps", "con", "trt", "parallel_arms") %in% names(fit_par$timing)))
  expect_true(is.list(fit_par$ps_fit$timing))
  expect_true(all(c("build", "compile", "mcmc", "total") %in% names(fit_par$ps_fit$timing)))
})

test_that("predict supports ndraws_pred/chunk_size/parallel aliases", {
  skip_if_not_test_level("ci")
  set.seed(105)
  n <- 120
  X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
  y <- abs(0.5 * X[, 1] + stats::rnorm(n)) + 0.1
  b <- build_nimble_bundle(
    y = y, X = X, backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    mcmc = list(niter = 120, nburnin = 30, thin = 1, nchains = 1, seed = 31),
    monitor_latent = TRUE
  )
  fit <- run_mcmc_bundle_manual(b, show_progress = FALSE)
  X_new <- X[1:40, , drop = FALSE]

  p_seq <- predict(
    fit,
    x = X_new,
    type = "quantile",
    index = c(0.5, 0.9),
    ndraws_pred = 50,
    chunk_size = 10,
    parallel = FALSE
  )
  p_par <- predict(
    fit,
    x = X_new,
    type = "quantile",
    index = c(0.5, 0.9),
    ndraws_pred = 50,
    chunk_size = 10,
    parallel = TRUE,
    workers = 2
  )

  expect_s3_class(p_seq, "mixgpd_predict")
  expect_s3_class(p_par, "mixgpd_predict")
  expect_equal(names(p_seq$fit), names(p_par$fit))
  expect_equal(dim(p_seq$fit), dim(p_par$fit))

  expect_error(
    predict(
      fit,
      x = X_new,
      type = "quantile",
      index = c(0.5, 0.9),
      cred.level = 0.9
    ),
    "'cred.level' is no longer supported; use 'level' instead."
  )
})

test_that("parallel runtime does not leave global future plan modified", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("future")
  skip_if_not_installed("future.apply")

  old_plan <- future::plan()
  on.exit(future::plan(old_plan), add = TRUE)

  set.seed(106)
  y <- abs(stats::rnorm(50)) + 0.1
  b <- build_nimble_bundle(
    y = y, backend = "sb", kernel = "normal", GPD = FALSE, components = 5,
    mcmc = list(niter = 80, nburnin = 20, thin = 1, nchains = 2, seed = 9)
  )
  invisible(run_mcmc_bundle_manual(b, show_progress = FALSE, parallel_chains = TRUE, workers = 2))
  after_plan <- future::plan()
  expect_identical(class(after_plan), class(old_plan))
})

test_that("functional equivalence and contracts on seeded runs", {
  skip_if_not_test_level("full")

  tol <- 0.2
  set.seed(107)
  y <- abs(stats::rnorm(140)) + 0.1
  X <- cbind(x1 = stats::rnorm(140), x2 = stats::runif(140))
  A <- stats::rbinom(140, 1, 0.5)

  m <- list(niter = 600, nburnin = 150, thin = 2, nchains = 1, seed = 77)
  fit_u_old <- dpmgpd(y = y, backend = "sb", kernel = "normal", components = 6, mcmc = m, monitor_latent = TRUE)
  fit_u_new <- dpmgpd(y = y, backend = "sb", kernel = "normal", components = 6,
                      monitor_latent = TRUE,
                      mcmc = utils::modifyList(m, list(parallel_chains = FALSE)))
  mu_old <- mean(as.numeric(.extract_draws_matrix(fit_u_old)[, "alpha"]), na.rm = TRUE)
  mu_new <- mean(as.numeric(.extract_draws_matrix(fit_u_new)[, "alpha"]), na.rm = TRUE)
  expect_lte(abs(mu_old - mu_new), tol)
  expect_equal(class(fit_u_old), class(fit_u_new))
  expect_equal(sort(names(fit_u_old)), sort(names(fit_u_new)))

  cb <- build_causal_bundle(
    y = y, X = X, A = A, backend = "sb", kernel = "normal", GPD = FALSE, components = 6,
    mcmc_outcome = m, mcmc_ps = m, PS = "logit"
  )
  cf_seq <- run_mcmc_causal(cb, show_progress = FALSE, parallel_arms = FALSE)
  cf_par <- run_mcmc_causal(cb, show_progress = FALSE, parallel_arms = TRUE, workers = 2)
  expect_equal(class(cf_seq), class(cf_par))

  p_seq <- predict(fit_u_new, type = "quantile", index = c(0.5, 0.9), parallel = FALSE)
  p_old <- predict(fit_u_old, type = "quantile", index = c(0.5, 0.9), parallel = FALSE)
  expect_equal(names(p_seq$fit), names(p_old$fit))
  expect_equal(dim(p_seq$fit), dim(p_old$fit))
})

# Lightweight user-facing wrappers


.is_bundle <- function(x) {
  inherits(x, "causalmixgpd_bundle") || inherits(x, "causalmixgpd_causal_bundle")
}

.is_causal_bundle <- function(x) {
  inherits(x, "causalmixgpd_causal_bundle")
}

.coerce_treat <- function(treat) {
  if (is.null(treat)) return(NULL)

  if (is.factor(treat)) {
    if (nlevels(treat) != 2L) {
      stop("'treat' factor must have exactly 2 levels.", call. = FALSE)
    }
    treat <- as.integer(treat) - 1L
  } else if (is.logical(treat)) {
    treat <- as.integer(treat)
  } else if (is.character(treat)) {
    vals <- unique(treat)
    vals <- vals[!is.na(vals)]
    if (length(vals) != 2L) {
      stop("'treat' character input must be binary.", call. = FALSE)
    }
    treat <- as.integer(treat == vals[2L])
  } else {
    treat <- as.numeric(treat)
  }

  if (anyNA(treat)) stop("'treat' cannot contain NA.", call. = FALSE)
  if (!all(treat %in% c(0, 1))) stop("'treat' must be binary (0/1).", call. = FALSE)
  as.integer(treat)
}

.treat_arg_supplied <- function(call_args, treat_expr) {
  ("treat" %in% names(call_args)) && !identical(treat_expr, quote(NULL))
}

.formula_design_matrix <- function(trm,
                                   mf,
                                   drop_intercept = TRUE,
                                   drop_terms = character(0),
                                   coerce_double = FALSE) {
  term_labels <- attr(trm, "term.labels") %||% character(0)
  if (!length(term_labels)) {
    return(list(X = NULL, contrasts = NULL, X_cols = character(0)))
  }

  mm <- stats::model.matrix(trm, data = mf)
  keep <- rep(TRUE, ncol(mm))

  if (isTRUE(drop_intercept) && ncol(mm)) {
    keep <- keep & ((colnames(mm) %||% rep("", ncol(mm))) != "(Intercept)")
  }

  drop_terms <- tolower(as.character(drop_terms %||% character(0)))
  if (length(drop_terms) && ncol(mm)) {
    assign_idx <- attr(mm, "assign") %||% integer(0)
    drop_idx <- which(tolower(term_labels) %in% drop_terms)
    if (length(drop_idx) && length(assign_idx) == ncol(mm)) {
      keep <- keep & !(assign_idx %in% drop_idx)
    }
  }

  if (!any(keep)) {
    return(list(
      X = NULL,
      contrasts = attr(mm, "contrasts"),
      X_cols = character(0)
    ))
  }

  X <- mm[, keep, drop = FALSE]
  if (isTRUE(coerce_double)) storage.mode(X) <- "double"

  list(
    X = X,
    contrasts = attr(mm, "contrasts"),
    X_cols = colnames(X) %||% character(0)
  )
}

.parse_formula_yX <- function(formula, data, na.action = stats::na.omit, drop_intercept = TRUE) {
  if (is.null(data)) stop("'data' is required when 'formula' is provided.", call. = FALSE)

  mf <- stats::model.frame(formula, data = data, na.action = na.action)
  y <- stats::model.response(mf)
  if (is.null(y)) stop("Could not extract response 'y' from formula.", call. = FALSE)
  y <- as.numeric(y)

  trm <- stats::terms(mf)
  term_labels <- attr(trm, "term.labels") %||% character(0)
  is_unconditional <- length(term_labels) == 0L

  X <- NULL
  X_cols <- character(0)
  ctr <- NULL
  if (!is_unconditional) {
    design <- .formula_design_matrix(
      trm = trm,
      mf = mf,
      drop_intercept = drop_intercept,
      drop_terms = "id"
    )
    X <- design$X
    if (is.null(X)) {
      X <- NULL
      is_unconditional <- TRUE
    } else {
      X_cols <- design$X_cols
      ctr <- design$contrasts
    }
  }

  list(
    y = y,
    X = X,
    mf = mf,
    terms = trm,
    xlevels = stats::.getXlevels(trm, mf),
    contrasts = ctr,
    X_cols = X_cols,
    is_unconditional = is_unconditional
  )
}

.extract_treat_from_data <- function(mf, data, treat, treat_expr = NULL) {
  n <- nrow(mf)
  vals <- NULL

  if (is.character(treat) && length(treat) == 1L) {
    if (!is.null(mf[[treat]])) {
      vals <- mf[[treat]]
    } else if (!is.null(data) && !is.null(data[[treat]])) {
      vals <- data[[treat]]
    } else {
      stop(sprintf("Could not find treatment column '%s'.", treat), call. = FALSE)
    }
  } else if (!is.null(treat_expr) &&
             is.symbol(treat_expr) &&
             as.character(treat_expr) != "NULL" &&
             as.character(treat_expr) %in% names(mf)) {
    vals <- mf[[as.character(treat_expr)]]
  } else {
    vals <- treat
  }

  if (length(vals) == n) return(.coerce_treat(vals))

  # Align vectors provided at full data length to model.frame rows after NA filtering.
  if (!is.null(data) && length(vals) == nrow(data)) {
    ridx <- suppressWarnings(as.integer(rownames(mf)))
    if (!anyNA(ridx) && length(ridx) == n) {
      return(.coerce_treat(vals[ridx]))
    }
  }

  stop("'treat' length does not match model rows after NA handling.", call. = FALSE)
}

.bundle_has_any_gpd <- function(b) {
  if (.is_causal_bundle(b)) {
    gpd <- b$meta$GPD %||% list()
    return(isTRUE(gpd$trt) || isTRUE(gpd$con))
  }
  isTRUE(b$spec$meta$GPD)
}

.bundle_all_gpd <- function(b) {
  if (.is_causal_bundle(b)) {
    gpd <- b$meta$GPD %||% list()
    return(isTRUE(gpd$trt) && isTRUE(gpd$con))
  }
  isTRUE(b$spec$meta$GPD)
}

.strip_gpd_single_bundle <- function(b) {
  stopifnot(inherits(b, "causalmixgpd_bundle"))

  b$spec$meta$GPD <- FALSE
  b$spec$plan$GPD <- FALSE
  b$spec$plan$gpd <- list()

  b$code <- .wrap_nimble_code(build_code_from_spec(b$spec))
  b$constants <- build_constants_from_spec(b$spec)
  b$dimensions <- build_dimensions_from_spec(b$spec)
  b$inits <- build_inits_from_spec(b$spec, y = b$data$y, X = b$data$X %||% NULL)
  pol <- b$monitor_policy %||% list()
  b$monitors <- build_monitors_from_spec(
    b$spec,
    monitor_v = isTRUE(pol$monitor_v),
    monitor_latent = isTRUE(pol$monitor_latent)
  )

  b
}

.bundle_strip_gpd <- function(b) {
  if (.is_causal_bundle(b)) {
    b$outcome$con <- .strip_gpd_single_bundle(b$outcome$con)
    b$outcome$trt <- .strip_gpd_single_bundle(b$outcome$trt)
    b$meta$GPD <- list(trt = FALSE, con = FALSE)
    return(b)
  }
  .strip_gpd_single_bundle(b)
}

.normalize_mcmc_inputs <- function(args) {
  if (!length(args)) {
    return(list(
      overrides = list(),
      runner = list(),
      outcome_overrides = list(),
      ps_overrides = list()
    ))
  }
  nm <- names(args)
  if (is.null(nm)) nm <- rep("", length(args))
  if (any(!nzchar(nm))) stop("All mcmc arguments must be named.", call. = FALSE)

  override_names <- c(
    "niter", "nburn", "nburnin", "thin", "nchains", "seed", "waic",
    "parallel_chains", "parallel_arms", "workers", "timing", "z_update_every"
  )
  runner_names <- c("show_progress", "quiet", "parallel_chains", "parallel_arms", "workers", "timing")

  normalize_override_list <- function(x, label) {
    if (is.null(x)) return(list())
    if (!is.list(x)) stop(sprintf("'%s' must be a named list.", label), call. = FALSE)
    x_names <- names(x)
    if (is.null(x_names)) x_names <- rep("", length(x))
    if (any(!nzchar(x_names))) stop(sprintf("All entries in '%s' must be named.", label), call. = FALSE)
    unknown_x <- x_names[!(x_names %in% override_names)]
    if (length(unknown_x)) {
      stop(sprintf("Unknown %s argument(s): %s", label, paste(unique(unknown_x), collapse = ", ")), call. = FALSE)
    }
    if (!is.null(x$nburn) && is.null(x$nburnin)) {
      x$nburnin <- x$nburn
    }
    x$nburn <- NULL
    x
  }

  outcome_overrides <- normalize_override_list(args$mcmc_outcome %||% NULL, "mcmc_outcome")
  ps_overrides <- normalize_override_list(args$mcmc_ps %||% NULL, "mcmc_ps")
  args$mcmc_outcome <- NULL
  args$mcmc_ps <- NULL
  nm <- names(args)

  override_idx <- nm %in% override_names
  runner_idx <- nm %in% runner_names
  unknown <- nm[!(override_idx | runner_idx)]
  unknown <- unknown[nzchar(unknown)]
  if (length(unknown)) {
    stop(sprintf("Unknown mcmc argument(s): %s", paste(unique(unknown), collapse = ", ")), call. = FALSE)
  }

  overrides <- args[override_idx]
  if (!is.null(overrides$nburn) && is.null(overrides$nburnin)) {
    overrides$nburnin <- overrides$nburn
  }
  overrides$nburn <- NULL

  list(
    overrides = overrides,
    runner = args[runner_idx],
    outcome_overrides = outcome_overrides,
    ps_overrides = ps_overrides
  )
}

.apply_mcmc_overrides <- function(b, overrides, outcome_overrides = list(), ps_overrides = list()) {
  if (!length(overrides) && !length(outcome_overrides) && !length(ps_overrides)) return(b)

  if (.is_causal_bundle(b)) {
    outcome_merged <- utils::modifyList(overrides %||% list(), outcome_overrides %||% list())
    if (length(overrides) || length(outcome_overrides)) {
      b$outcome$con$mcmc <- utils::modifyList(b$outcome$con$mcmc %||% list(), outcome_merged)
      b$outcome$trt$mcmc <- utils::modifyList(b$outcome$trt$mcmc %||% list(), outcome_merged)
    }
    if (length(ps_overrides) && inherits(b$design, "causalmixgpd_ps_bundle")) {
      b$design$mcmc <- utils::modifyList(b$design$mcmc %||% list(), ps_overrides)
    }
    return(b)
  }

  if (length(outcome_overrides) || length(ps_overrides)) {
    stop("mcmc_outcome/mcmc_ps overrides are available only for causal bundles.", call. = FALSE)
  }

  b$mcmc <- utils::modifyList(b$mcmc %||% list(), overrides %||% list())
  b
}

.run_bundle_mcmc <- function(b, mcmc_args = list()) {
  if (!is.list(mcmc_args)) stop("'mcmc' must be a named list.", call. = FALSE)
  do.call(mcmc, c(list(b = b), mcmc_args))
}

.wrapper_mcmc_arg_names <- function() {
  c(
    "niter", "nburn", "nburnin", "thin", "nchains", "seed", "waic",
    "parallel_chains", "parallel_arms", "workers", "timing", "z_update_every",
    "mcmc_outcome", "mcmc_ps",
    "show_progress", "quiet"
  )
}

.collect_inline_mcmc_from_dots <- function(dots_expr, mcmc, eval_env) {
  if (is.null(dots_expr) || !length(dots_expr)) {
    return(list(mcmc = mcmc, names = character(0)))
  }

  dots_list <- as.list(dots_expr)
  dot_names <- names(dots_list)
  if (is.null(dot_names)) dot_names <- rep("", length(dots_list))

  inline_idx <- nzchar(dot_names) & (dot_names %in% .wrapper_mcmc_arg_names())
  if (!any(inline_idx)) {
    return(list(mcmc = mcmc, names = character(0)))
  }

  inline_vals <- lapply(dots_list[inline_idx], eval, envir = eval_env)
  parsed <- .normalize_mcmc_inputs(inline_vals)
  merged <- utils::modifyList(mcmc %||% list(), c(parsed$overrides, parsed$runner))
  if (length(parsed$outcome_overrides)) {
    merged$mcmc_outcome <- utils::modifyList(merged$mcmc_outcome %||% list(), parsed$outcome_overrides)
  }
  if (length(parsed$ps_overrides)) {
    merged$mcmc_ps <- utils::modifyList(merged$mcmc_ps %||% list(), parsed$ps_overrides)
  }

  list(mcmc = merged, names = dot_names[inline_idx])
}

#' Build the workflow bundle used by the package fitters
#'
#' \code{bundle()} is the main workflow constructor. It converts raw inputs,
#' a formula/data pair, or an already prepared bundle into the canonical
#' object consumed by \code{\link{mcmc}}, \code{\link{dpmix}},
#' \code{\link{dpmgpd}}, \code{\link{dpmix.causal}}, and
#' \code{\link{dpmgpd.causal}}.
#'
#' For one-arm models the returned object represents a bulk Dirichlet process
#' mixture, optionally augmented with a spliced generalized Pareto tail. For
#' causal models the returned object contains two arm-specific outcome bundles
#' plus an optional propensity score block.
#'
#' @details
#' The workflow is:
#' \enumerate{
#'   \item prepare a bundle with \code{bundle()},
#'   \item run posterior sampling with \code{\link{mcmc}} or one of the
#'   \code{dpmix*}/\code{dpmgpd*} wrappers,
#'   \item inspect the fitted object with \code{\link{summary.mixgpd_fit}},
#'   \code{\link{params}}, \code{\link{predict.mixgpd_fit}}, or the causal
#'   estimand helpers.
#' }
#'
#' Setting \code{GPD = TRUE} requests the spliced bulk-tail model with
#' conditional distribution
#' \deqn{F(y \mid x) = F_{\mathrm{bulk}}(y \mid x)\mathbf{1}\{y \le u(x)\} +
#' \left[p_u(x) + \{1 - p_u(x)\}F_{\mathrm{GPD}}(y \mid x)\right]\mathbf{1}\{y > u(x)\},}
#' where \eqn{p_u(x)} is the bulk probability below the threshold \eqn{u(x)}.
#'
#' See the manuscript vignette for the DPM hierarchy, SB/CRP representations,
#' and the spliced bulk-tail construction used throughout the package.
#'
#' @param y Either a response vector or an existing bundle.
#' @param X Optional design matrix/data.frame.
#' @param treat Optional binary treatment indicator.
#' @param data Optional data.frame used with \code{formula}.
#' @param formula Optional formula.
#' @param GPD Logical; include GPD tail in build mode.
#' @param ... Additional arguments passed to \code{build_nimble_bundle()} or
#'   \code{build_causal_bundle()}.
#' @return A \code{"causalmixgpd_bundle"} for one-arm models or a
#'   \code{"causalmixgpd_causal_bundle"} for causal models. The bundle stores
#'   code-generation inputs, monitor policy, and default MCMC settings, but it
#'   does not run MCMC.
#' @seealso \code{\link{build_nimble_bundle}}, \code{\link{build_causal_bundle}},
#'   \code{\link{mcmc}}, \code{\link{dpmix}}, \code{\link{dpmgpd}}.
#' @export
bundle <- function(y = NULL, X = NULL, treat = NULL, data = NULL, formula = NULL, GPD = FALSE, ...) {
  if (.is_bundle(y)) return(y)

  treat_expr <- substitute(treat)
  call_args <- as.list(match.call(expand.dots = FALSE))
  treat_supplied <- .treat_arg_supplied(call_args, treat_expr)

  y_vec <- NULL
  x_mat <- X
  t_vec <- NULL
  formula_meta <- NULL

  if (!is.null(formula)) {
    parsed <- .parse_formula_yX(formula = formula, data = data)
    y_vec <- parsed$y
    x_mat <- parsed$X
    if (treat_supplied) {
      treat_name <- if (is.symbol(treat_expr)) as.character(treat_expr) else NULL
      if (!is.null(treat_name) &&
          treat_name != "NULL" &&
          (treat_name %in% names(parsed$mf) || (!is.null(data) && treat_name %in% names(data)))) {
        treat_in <- treat_name
      } else {
        treat_in <- treat
      }
      t_vec <- .extract_treat_from_data(parsed$mf, data = data, treat = treat_in, treat_expr = treat_expr)
    }

    formula_meta <- list(
      terms = parsed$terms,
      xlevels = parsed$xlevels,
      contrasts = parsed$contrasts,
      X_cols = parsed$X_cols,
      treat = if (is.symbol(treat_expr)) as.character(treat_expr) else NULL
    )
  } else {
    if (is.null(y)) stop("Provide either 'y' (response vector), 'formula', or a bundle.", call. = FALSE)
    y_vec <- as.numeric(y)
    if (treat_supplied) t_vec <- .coerce_treat(treat)
  }

  if (is.data.frame(x_mat) && any(tolower(names(x_mat)) == "id")) {
    keep <- tolower(names(x_mat)) != "id"
    x_mat <- x_mat[, keep, drop = FALSE]
  }

  if (!is.null(x_mat) && !is.matrix(x_mat)) x_mat <- as.matrix(x_mat)
  if (length(y_vec) < 1L) stop("'y' must be a non-empty numeric vector.", call. = FALSE)
  if (!is.null(x_mat) && nrow(x_mat) != length(y_vec)) stop("nrow(X) must match length(y).", call. = FALSE)
  if (!is.null(t_vec) && length(t_vec) != length(y_vec)) stop("length(treat) must match length(y).", call. = FALSE)

  if (is.null(t_vec)) {
    b <- build_nimble_bundle(y = y_vec, X = x_mat, GPD = GPD, ...)
  } else {
    b <- build_causal_bundle(y = y_vec, X = x_mat, A = t_vec, GPD = GPD, ...)
  }

  if (!is.null(formula_meta)) {
    attr(b, "causalmixgpd_formula_meta") <- formula_meta
  }

  b
}

#' Run posterior sampling from a prepared bundle
#'
#' \code{mcmc()} is the generic workflow runner. It dispatches to
#' \code{\link{run_mcmc_bundle_manual}} for one-arm bundles and to
#' \code{\link{run_mcmc_causal}} for causal bundles.
#'
#' @details
#' This wrapper is useful when you want a two-stage workflow:
#' build first, inspect or modify the bundle, then sample. Named MCMC arguments
#' supplied through \code{...} override the settings stored in the bundle before
#' execution.
#'
#' The returned fit represents posterior draws from the finite SB/CRP
#' approximation encoded in the bundle. Downstream summaries therefore target
#' posterior predictive quantities such as \eqn{f(y \mid x)},
#' \eqn{F(y \mid x)}, and derived treatment-effect functionals.
#'
#' @param b A non-causal or causal bundle.
#' @param ... Optional MCMC overrides (\code{niter}, \code{nburnin}, \code{thin},
#'   \code{nchains}, \code{seed}, \code{waic}) and runner controls
#'   (\code{show_progress}, \code{quiet}).
#' @return A fitted object of class \code{"mixgpd_fit"} or
#'   \code{"causalmixgpd_causal_fit"}.
#' @seealso \code{\link{bundle}}, \code{\link{run_mcmc_bundle_manual}},
#'   \code{\link{run_mcmc_causal}}, \code{\link{predict.mixgpd_fit}}.
#' @export
mcmc <- function(b, ...) {
  if (!.is_bundle(b)) stop("'b' must be a causalmixgpd bundle object.", call. = FALSE)

  parsed <- .normalize_mcmc_inputs(list(...))
  b <- .apply_mcmc_overrides(
    b,
    parsed$overrides,
    outcome_overrides = parsed$outcome_overrides,
    ps_overrides = parsed$ps_overrides
  )

  if (.is_causal_bundle(b)) {
    # For causal workflows, parallel_chains is an arm-level MCMC override
    # consumed by run_mcmc_bundle_manual() inside run_mcmc_causal().
    parsed$runner$parallel_chains <- NULL
    allowed <- c("show_progress", "quiet", "parallel_arms", "workers", "timing")
    bad <- setdiff(names(parsed$runner), allowed)
    if (length(bad)) {
      stop(sprintf("Unsupported runner argument for causal bundles: %s", paste(bad, collapse = ", ")),
           call. = FALSE)
    }
    return(do.call(run_mcmc_causal, c(list(bundle = b), parsed$runner)))
  }

  allowed <- c("show_progress", "quiet", "parallel_chains", "workers", "timing")
  bad <- setdiff(names(parsed$runner), allowed)
  if (length(bad)) {
    stop(sprintf("Unsupported runner argument for non-causal bundles: %s", paste(bad, collapse = ", ")),
         call. = FALSE)
  }
  do.call(run_mcmc_bundle_manual, c(list(bundle = b), parsed$runner))
}

#' Fit a one-arm Dirichlet process mixture without a GPD tail
#'
#' \code{dpmix()} is the one-step convenience wrapper for the bulk-only model.
#' It combines \code{\link{bundle}} and \code{\link{mcmc}} for one-arm data.
#'
#' @details
#' The fitted model targets the posterior predictive bulk distribution
#' \deqn{f(y \mid x) = \int f(y \mid x, \theta)\,d\Pi(\theta),}
#' without the spliced tail augmentation used by \code{\link{dpmgpd}}.
#'
#' Use this wrapper when the outcome support is adequately modeled by the bulk
#' kernel alone. If you need threshold exceedance modeling or extreme-quantile
#' extrapolation, use \code{\link{dpmgpd}} instead.
#'
#' @param y Either a response vector or a bundle object.
#' @param X Optional design matrix/data.frame.
#' @param treat Optional binary treatment indicator. If supplied, this wrapper
#'   errors; use \code{dpmix.causal()} for causal models.
#' @param data Optional data.frame used with \code{formula}.
#' @param mcmc Named list of run arguments passed to \code{mcmc()} (including
#'   optional performance controls such as \code{parallel_chains},
#'   \code{workers}, \code{timing}, and \code{z_update_every}).
#' @param formula Optional formula.
#' @param ... Additional build arguments passed to \code{\link{build_nimble_bundle}}.
#' @return A fitted object of class \code{"mixgpd_fit"}.
#' @seealso \code{\link{build_nimble_bundle}},
#'   \code{\link{bundle}}, \code{\link{dpmgpd}},
#'   \code{\link{predict.mixgpd_fit}}, \code{\link{summary.mixgpd_fit}}.
#' @export
dpmix <- function(y = NULL, X = NULL, treat = NULL, data = NULL, mcmc = list(), formula = NULL, ...) {
  treat_expr <- substitute(treat)
  call_match <- match.call(expand.dots = FALSE)
  call_args <- as.list(call_match)

  if (.is_causal_bundle(y) || .treat_arg_supplied(call_args, treat_expr)) {
    stop(
      "dpmix() is for one-arm models. Use dpmix.causal() for causal models.",
      call. = FALSE
    )
  }

  inline <- .collect_inline_mcmc_from_dots(call_args$..., mcmc = mcmc, eval_env = parent.frame())
  mcmc <- inline$mcmc

  b <- NULL

  if (.is_bundle(y)) {
    b <- if (.bundle_has_any_gpd(y)) .bundle_strip_gpd(y) else y
    return(.run_bundle_mcmc(b, mcmc_args = mcmc))
  }

  bundle_call <- match.call(expand.dots = TRUE)
  bundle_call[[1L]] <- quote(bundle)
  bundle_call$mcmc <- NULL
  bundle_call$GPD <- FALSE
  if (length(inline$names)) {
    for (nm in unique(inline$names)) bundle_call[[nm]] <- NULL
  }

  b <- eval.parent(bundle_call)
  .run_bundle_mcmc(b, mcmc_args = mcmc)
}

#' Fit a one-arm Dirichlet process mixture with a spliced GPD tail
#'
#' \code{dpmgpd()} is the one-step convenience wrapper for the spliced
#' bulk-tail model. It combines \code{\link{bundle}} and \code{\link{mcmc}} for
#' one-arm data.
#'
#' @details
#' This wrapper targets the posterior predictive distribution obtained by
#' combining a flexible bulk DPM with a generalized Pareto exceedance model
#' above the threshold \eqn{u(x)}. In the tail region the predictive density is
#' proportional to
#' \deqn{\{1 - p_u(x)\} f_{\mathrm{GPD}}(y \mid x), \qquad y > u(x),}
#' where \eqn{p_u(x)} is the posterior bulk mass below the threshold.
#'
#' Use this wrapper when upper-tail behavior matters for inference, prediction,
#' or extrapolation of extreme quantiles and survival probabilities.
#'
#' @param y Either a response vector or a bundle object.
#' @param X Optional design matrix/data.frame.
#' @param treat Optional binary treatment indicator. If supplied, this wrapper
#'   errors; use \code{dpmgpd.causal()} for causal models.
#' @param data Optional data.frame used with \code{formula}.
#' @param mcmc Named list of run arguments passed to \code{mcmc()} (including
#'   optional performance controls such as \code{parallel_chains},
#'   \code{workers}, \code{timing}, and \code{z_update_every}).
#' @param formula Optional formula.
#' @param ... Additional build arguments passed to \code{\link{build_nimble_bundle}}.
#' @return A fitted object of class \code{"mixgpd_fit"}.
#' @seealso \code{\link{build_nimble_bundle}},
#'   \code{\link{bundle}}, \code{\link{dpmix}},
#'   \code{\link{predict.mixgpd_fit}}, \code{\link{summary.mixgpd_fit}}.
#' @export
dpmgpd <- function(y = NULL, X = NULL, treat = NULL, data = NULL, mcmc = list(), formula = NULL, ...) {
  treat_expr <- substitute(treat)
  call_match <- match.call(expand.dots = FALSE)
  call_args <- as.list(call_match)

  if (.is_causal_bundle(y) || .treat_arg_supplied(call_args, treat_expr)) {
    stop(
      "dpmgpd() is for one-arm models. Use dpmgpd.causal() for causal models.",
      call. = FALSE
    )
  }

  inline <- .collect_inline_mcmc_from_dots(call_args$..., mcmc = mcmc, eval_env = parent.frame())
  mcmc <- inline$mcmc

  if (.is_bundle(y)) {
    if (!.bundle_all_gpd(y)) {
      stop(
        "dpmgpd() requires a bundle with GPD enabled for all modeled arms; use dpmix(bundle) for non-GPD runs.",
        call. = FALSE
      )
    }
    return(.run_bundle_mcmc(y, mcmc_args = mcmc))
  }

  bundle_call <- match.call(expand.dots = TRUE)
  bundle_call[[1L]] <- quote(bundle)
  bundle_call$mcmc <- NULL
  bundle_call$GPD <- TRUE
  if (length(inline$names)) {
    for (nm in unique(inline$names)) bundle_call[[nm]] <- NULL
  }

  b <- eval.parent(bundle_call)
  .run_bundle_mcmc(b, mcmc_args = mcmc)
}

#' Fit a causal two-arm Dirichlet process mixture without a GPD tail
#'
#' \code{dpmix.causal()} fits a causal model with separate treated and control
#' outcome mixtures and, when requested, a propensity score block. It is the
#' bulk-only companion to \code{\link{dpmgpd.causal}}.
#'
#' @details
#' The resulting fit supports conditional outcome prediction
#' \eqn{F_a(y \mid x)} for \eqn{a \in \{0,1\}}, followed by causal
#' functionals such as \code{\link{ate}}, \code{\link{qte}},
#' \code{\link{cate}}, and \code{\link{cqte}}.
#'
#' @param y Either a response vector or a causal bundle object.
#' @param X Optional design matrix/data.frame.
#' @param treat Binary treatment indicator.
#' @param data Optional data.frame used with \code{formula}.
#' @param mcmc Named list of run arguments passed to \code{mcmc()} (including
#'   optional performance controls such as \code{parallel_arms},
#'   \code{workers}, \code{timing}, and \code{z_update_every}).
#' @param formula Optional formula.
#' @param ... Additional build arguments passed to \code{\link{build_causal_bundle}}.
#' @return A fitted object of class \code{"causalmixgpd_causal_fit"}.
#' @seealso \code{\link{build_causal_bundle}},
#'   \code{\link{bundle}}, \code{\link{dpmgpd.causal}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
#'   \code{\link{qte}}.
#' @examples
#' \donttest{
#' N <- 30
#' X <- data.frame(x1 = stats::rnorm(N), x2 = stats::runif(N))
#' A <- stats::rbinom(N, 1, 0.5)
#' y <- abs(stats::rnorm(N) + A + 0.5 * X$x1) + 0.1
#' mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#'
#' fit <- dpmix.causal(
#'   y = y, X = X, treat = A,
#'   backend = "sb", kernel = "normal",
#'   components = 3, mcmc = mcmc_small
#' )
#' summary(fit)
#' ate(fit, show_progress = FALSE)
#' }
#' @export
dpmix.causal <- function(y = NULL, X = NULL, treat = NULL, data = NULL, mcmc = list(), formula = NULL, ...) {
  treat_expr <- substitute(treat)
  call_match <- match.call(expand.dots = FALSE)
  call_args <- as.list(call_match)

  inline <- .collect_inline_mcmc_from_dots(call_args$..., mcmc = mcmc, eval_env = parent.frame())
  mcmc <- inline$mcmc

  b <- NULL

  if (.is_bundle(y)) {
    if (!.is_causal_bundle(y)) {
      stop("dpmix.causal() requires a causal bundle; use dpmix() for one-arm models.", call. = FALSE)
    }
    b <- if (.bundle_has_any_gpd(y)) .bundle_strip_gpd(y) else y
    return(.run_bundle_mcmc(b, mcmc_args = mcmc))
  }

  if (!.treat_arg_supplied(call_args, treat_expr)) {
    stop("dpmix.causal() requires 'treat' when building from raw inputs.", call. = FALSE)
  }

  bundle_call <- match.call(expand.dots = TRUE)
  bundle_call[[1L]] <- quote(bundle)
  bundle_call$mcmc <- NULL
  bundle_call$GPD <- FALSE
  if (length(inline$names)) {
    for (nm in unique(inline$names)) bundle_call[[nm]] <- NULL
  }

  b <- eval.parent(bundle_call)
  .run_bundle_mcmc(b, mcmc_args = mcmc)
}

#' Fit a causal two-arm Dirichlet process mixture with a spliced GPD tail
#'
#' \code{dpmgpd.causal()} is the highest-level causal fitting wrapper. It builds
#' or accepts a causal bundle, runs posterior sampling for the treated and
#' control arms, and returns a single causal fit ready for prediction and effect
#' estimation.
#'
#' @details
#' The arm-specific predictive distributions
#' \eqn{F_1(y \mid x)} and \eqn{F_0(y \mid x)} inherit
#' the spliced bulk-tail structure. Downstream causal estimands are computed as
#' functionals of these two predictive laws, for example
#' \deqn{\mathrm{QTE}(\tau) = Q_1(\tau) - Q_0(\tau), \qquad
#' \mathrm{ATE} = E(Y_1) - E(Y_0).}
#'
#' @param y Either a response vector or a causal bundle object.
#' @param X Optional design matrix/data.frame.
#' @param treat Binary treatment indicator.
#' @param data Optional data.frame used with \code{formula}.
#' @param mcmc Named list of run arguments passed to \code{mcmc()} (including
#'   optional performance controls such as \code{parallel_arms},
#'   \code{workers}, \code{timing}, and \code{z_update_every}).
#' @param formula Optional formula.
#' @param ... Additional build arguments passed to \code{\link{build_causal_bundle}}.
#' @return A fitted object of class \code{"causalmixgpd_causal_fit"}.
#' @seealso \code{\link{build_causal_bundle}},
#'   \code{\link{bundle}}, \code{\link{dpmix.causal}},
#'   \code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
#'   \code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
#' @examples
#' \donttest{
#' data("causal_pos500_p3_k2", package = "CausalMixGPD")
#' idx <- seq_len(30)
#' fit <- dpmgpd.causal(
#'   y = causal_pos500_p3_k2$y[idx],
#'   X = causal_pos500_p3_k2$X[idx, , drop = FALSE],
#'   treat = causal_pos500_p3_k2$A[idx],
#'   kernel = "gamma",
#'   backend = "sb",
#'   components = 3,
#'   mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1,
#'               show_progress = FALSE, quiet = TRUE)
#' )
#' ate(fit, show_progress = FALSE)
#' }
#' @export
dpmgpd.causal <- function(y = NULL, X = NULL, treat = NULL, data = NULL, mcmc = list(), formula = NULL, ...) {
  treat_expr <- substitute(treat)
  call_match <- match.call(expand.dots = FALSE)
  call_args <- as.list(call_match)

  inline <- .collect_inline_mcmc_from_dots(call_args$..., mcmc = mcmc, eval_env = parent.frame())
  mcmc <- inline$mcmc

  if (.is_bundle(y)) {
    if (!.is_causal_bundle(y)) {
      stop("dpmgpd.causal() requires a causal bundle; use dpmgpd() for one-arm models.", call. = FALSE)
    }
    if (!.bundle_all_gpd(y)) {
      stop(
        "dpmgpd.causal() requires a causal bundle with GPD enabled for all modeled arms; use dpmix.causal(bundle) for non-GPD runs.",
        call. = FALSE
      )
    }
    return(.run_bundle_mcmc(y, mcmc_args = mcmc))
  }

  if (!.treat_arg_supplied(call_args, treat_expr)) {
    stop("dpmgpd.causal() requires 'treat' when building from raw inputs.", call. = FALSE)
  }

  bundle_call <- match.call(expand.dots = TRUE)
  bundle_call[[1L]] <- quote(bundle)
  bundle_call$mcmc <- NULL
  bundle_call$GPD <- TRUE
  if (length(inline$names)) {
    for (nm in unique(inline$names)) bundle_call[[nm]] <- NULL
  }

  b <- eval.parent(bundle_call)
  .run_bundle_mcmc(b, mcmc_args = mcmc)
}

`%||%` <- function(a, b) if (!is.null(a)) a else b

if (!exists(".cache_enabled")) {
  helper_path <- file.path("tests", "testthat", "helper-02-cache.R")
  if (file.exists(helper_path)) source(helper_path)
}

.supports_gpd <- function(kernel, backend) {
  kinfo <- get_kernel_registry()[[kernel]]
  sig <- kinfo$signatures[[backend]]$gpd %||% NULL
  !is.null(sig) && !isTRUE(is.na(sig$dist_name))
}

.run_predict_case <- function(label, kernel, backend, gpd, has_X) {
  ok <- TRUE
  msg <- label
  warn_msgs <- character()

  withCallingHandlers({
    set.seed(1)
    N <- 20
    y <- abs(stats::rnorm(N)) + 0.1
    X <- if (has_X) {
      data.frame(x1 = stats::rnorm(N), x2 = stats::runif(N))
    } else {
      NULL
    }

    if (isTRUE(gpd) && !.supports_gpd(kernel, backend)) {
      err <- tryCatch({
        build_nimble_bundle(
          y = y,
          X = X,
          backend = backend,
          kernel = kernel,
          GPD = TRUE,
          components = 6,
          mcmc = list(niter = 10, nburnin = 5, thin = 1, nchains = 1, seed = 1)
        )
        NULL
      }, error = function(e) e)
      if (is.null(err)) stop("Expected error for unsupported GPD backend.")
      msg <- paste0(msg, " | expected error for unsupported GPD backend")
      return(list(ok = TRUE, msg = msg))
    }

    mcmc_cfg <- list(niter = 30, nburnin = 10, thin = 1, nchains = 1, seed = 1)
    cache_key <- NULL
    if (exists(".cache_enabled") && isTRUE(.cache_enabled())) {
      key_str <- paste(
        "predict", kernel, backend, gpd, has_X, N,
        mcmc_cfg$niter, mcmc_cfg$nburnin, mcmc_cfg$thin, mcmc_cfg$nchains, mcmc_cfg$seed,
        sep = "|"
      )
      cache_key <- .cache_hash(key_str)
    }
    cached <- if (!is.null(cache_key)) .cache_get(cache_key) else NULL
    if (!is.null(cached) && inherits(cached$fit, "mixgpd_fit")) {
      fit <- cached$fit
    } else {
      bundle <- build_nimble_bundle(
        y = y,
        X = X,
        backend = backend,
        kernel = kernel,
        GPD = gpd,
        components = 6,
        mcmc = mcmc_cfg
      )
      fit <- run_mcmc_bundle_manual(bundle, show_progress = FALSE)
      if (!is.null(cache_key)) .cache_set(cache_key, list(fit = fit))
    }

    if (!inherits(fit, "mixgpd_fit")) stop("fit is not a mixgpd_fit.")
    print(fit)
    s <- summary(fit)
    print(s)

    y_grid <- sort(y)
    p_grid <- c(0.5, 0.9)
    nsim <- 5L

    if (has_X) {
      X_new <- X[1:3, , drop = FALSE]
      pr_den <- predict(fit, newdata = X_new, y = y_grid, type = "density", ncores = 1)
      pr_surv <- predict(fit, newdata = X_new, y = y_grid, type = "survival", ncores = 1)
      pr_q <- predict(fit, newdata = X_new, type = "quantile", index = p_grid, ncores = 1)
      pr_samp <- predict(fit, newdata = X_new, type = "sample", nsim = nsim)
      pr_mean <- predict(fit, newdata = X_new, type = "mean", nsim_mean = 50)
      n_pred <- nrow(X_new)
    } else {
      pr_den <- predict(fit, y = y_grid, type = "density", ncores = 1)
      pr_surv <- predict(fit, y = y_grid, type = "survival", ncores = 1)
      pr_q <- predict(fit, type = "quantile", index = p_grid, ncores = 1)
      pr_samp <- predict(fit, type = "sample", nsim = nsim)
      pr_mean <- predict(fit, type = "mean", nsim_mean = 50)
      n_pred <- 1L
    }

    if (!is.data.frame(pr_den$fit)) stop("density fit must be data.frame.")
    if (!is.data.frame(pr_surv$fit)) stop("survival fit must be data.frame.")
    if (!is.data.frame(pr_q$fit)) stop("quantile fit must be data.frame.")
    if (!is.data.frame(pr_mean$fit)) stop("mean fit must be data.frame.")

    den_rows <- if (has_X) n_pred * length(y_grid) else length(y_grid)
    if (nrow(pr_den$fit) != den_rows) stop("density rows mismatch.")
    if (nrow(pr_surv$fit) != den_rows) stop("survival rows mismatch.")

    q_rows <- if (has_X) n_pred * length(p_grid) else length(p_grid)
    if (nrow(pr_q$fit) != q_rows) stop("quantile rows mismatch.")

    mean_rows <- if (has_X) n_pred else 1L
    if (nrow(pr_mean$fit) != mean_rows) stop("mean rows mismatch.")

    if (has_X) {
      if (!all(c("id", "index", "estimate", "lower", "upper") %in% names(pr_q$fit))) {
        stop("quantile columns mismatch (conditional).")
      }
    } else {
      if (!all(c("index", "estimate", "lower", "upper") %in% names(pr_q$fit))) {
        stop("quantile columns mismatch (unconditional).")
      }
    }

    den_col <- if ("density" %in% names(pr_den$fit)) "density" else "estimate"
    surv_col <- if ("survival" %in% names(pr_surv$fit)) "survival" else "estimate"

    if (any(!is.finite(pr_den$fit[[den_col]]))) stop("density has non-finite values.")
    if (any(pr_den$fit[[den_col]] < 0)) stop("density has negative values.")
    if (any(!is.finite(pr_surv$fit[[surv_col]]))) stop("survival has non-finite values.")
    if (any(pr_surv$fit[[surv_col]] < 0 | pr_surv$fit[[surv_col]] > 1)) {
      stop("survival outside [0,1].")
    }

    if (has_X) {
      if (!identical(dim(pr_samp$fit), c(n_pred, nsim))) stop("sample dims mismatch.")
    } else {
      if (!is.numeric(pr_samp$fit) || length(pr_samp$fit) != nsim) {
        stop("sample length mismatch.")
      }
    }

    if (has_X) {
      res <- residuals(fit, type = "pit")
      if (!is.numeric(res) || length(res) != length(y)) stop("residuals length mismatch.")
      ftd <- fitted(fit)
      ftd_n <- if (is.data.frame(ftd)) nrow(ftd) else length(ftd)
      if (ftd_n != length(y)) stop("fitted length mismatch.")
    }

    if (requireNamespace("ggmcmc", quietly = TRUE) && requireNamespace("coda", quietly = TRUE)) {
      plots <- plot(fit, family = c("traceplot"), params = "alpha")
      if (!is.list(plots)) stop("plot() did not return a list.")
    }
  }, warning = function(w) {
    warn_msgs <<- c(warn_msgs, conditionMessage(w))
    invokeRestart("muffleWarning")
  }, error = function(e) {
    ok <<- FALSE
    msg <<- paste0(msg, " | error: ", conditionMessage(e))
  })

  if (length(warn_msgs)) {
    msg <- paste0(msg, " | warnings: ", length(warn_msgs))
  }

  list(ok = ok, msg = msg)
}

# ============================================================
# 05-visualization-helpers.R
# Helper functions for type-specific prediction visualizations
# ============================================================


#' Plot quantile predictions with CI
#' @keywords internal
#' @noRd
.plot_quantile_pred <- function(pred, ...) {
  fit_df <- pred$fit

  if (!is.data.frame(fit_df)) {
    stop("Quantile prediction must return a data frame in $fit.", call. = FALSE)
  }

  plot_data <- fit_df
  has_id <- "id" %in% names(plot_data)

  if (has_id) {
    n_id <- length(unique(plot_data$id))
    pal <- .plot_palette(max(2L, n_id))
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = factor(index), y = estimate)) +
      ggplot2::geom_line(ggplot2::aes(group = id, color = factor(id)), linewidth = 0.7) +
      ggplot2::geom_point(ggplot2::aes(color = factor(id)), size = 3) +
      ggplot2::geom_errorbar(ggplot2::aes(ymin = lower, ymax = upper,
                                          group = id, color = factor(id)),
                             width = 0.2, linewidth = 1) +
      ggplot2::scale_color_manual(values = pal) +
      .plot_theme() +
      ggplot2::labs(
        title = "Quantile Predictions with Pointwise Credible Intervals",
        x = "Quantile Index",
        y = "Estimate"
      )
  } else {
    pal <- .plot_palette(2L)
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = factor(index), y = estimate)) +
      ggplot2::geom_point(color = pal[1], size = 3) +
      ggplot2::geom_errorbar(ggplot2::aes(ymin = lower, ymax = upper),
                             width = 0.2, linewidth = 1, color = pal[1]) +
      .plot_theme() +
      ggplot2::labs(
        title = "Quantile Predictions with Pointwise Credible Intervals",
        x = "Quantile Index",
        y = "Estimate"
      )
  }

  .wrap_plotly(p)
}

#' Plot sample predictions: histogram with density overlay
#' @keywords internal
#' @noRd
.plot_sample_pred <- function(pred, ...) {
  samples <- pred$fit

  if (!is.numeric(samples)) {
    stop("Sample prediction must return a numeric vector in $fit.", call. = FALSE)
  }

  plot_data <- data.frame(value = samples)

  pal <- .plot_palette(8L)
  p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = value)) +
    ggplot2::geom_histogram(ggplot2::aes(y = ggplot2::after_stat(density)),
                           bins = 30, alpha = 0.7, fill = pal[5], color = pal[7]) +
    ggplot2::geom_density(color = pal[2], linewidth = 1) +
    .plot_theme() +
    ggplot2::labs(
      title = "Posterior Predictive Samples",
      x = "Value",
      y = "Density"
    )

  .wrap_plotly(p)
}

#' Plot mean predictions: histogram with mean line
#' @keywords internal
#' @noRd
.plot_mean_pred <- function(pred, ...) {

  fit_df <- pred$fit

  # Extract estimate, lower, upper from data frame
  if (is.data.frame(fit_df)) {
    mean_val <- mean(fit_df$estimate, na.rm = TRUE)
    lower_val <- if ("lower" %in% names(fit_df)) mean(fit_df$lower, na.rm = TRUE) else NULL
    upper_val <- if ("upper" %in% names(fit_df)) mean(fit_df$upper, na.rm = TRUE) else NULL
  } else {
    # Fallback for old format
    mean_val <- mean(fit_df, na.rm = TRUE)
    lower_val <- NULL
    upper_val <- NULL
  }

  # Use posterior samples for histogram
  if (!is.null(pred$draws) && is.numeric(pred$draws) && length(pred$draws) > 1) {
    samples <- as.numeric(pred$draws)
    plot_data <- data.frame(value = samples)

    pal <- .plot_palette(8L)
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = value)) +
      ggplot2::geom_histogram(ggplot2::aes(y = ggplot2::after_stat(density)),
                             bins = 30, alpha = 0.7, fill = pal[5], color = pal[7]) +
      ggplot2::geom_density(color = pal[1], linewidth = 1) +
      ggplot2::geom_vline(xintercept = mean_val, color = pal[2],
                         linewidth = 1.2, linetype = "solid") +
      {if (!is.null(lower_val) && !is.null(upper_val) && !is.na(lower_val) && !is.na(upper_val)) {
        list(
          ggplot2::geom_vline(xintercept = lower_val, color = pal[6],
                             linewidth = 0.8, linetype = "dashed"),
          ggplot2::geom_vline(xintercept = upper_val, color = pal[6],
                             linewidth = 0.8, linetype = "dashed")
        )
      }} +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Predictive Mean Distribution",
        subtitle = paste0("Mean = ", round(mean_val, 3)),
        x = "Value",
        y = "Density"
      )
  } else {
    # If no samples available, show vertical lines only
    pal <- .plot_palette(8L)
    p <- ggplot2::ggplot(data.frame(x = mean_val), ggplot2::aes(x = x)) +
      ggplot2::geom_vline(xintercept = mean_val, color = pal[2],
                         linewidth = 2, linetype = "solid") +
      {if (!is.null(lower_val) && !is.null(upper_val) && !is.na(lower_val) && !is.na(upper_val)) {
        list(
          ggplot2::geom_vline(xintercept = lower_val, color = pal[6],
                             linewidth = 1, linetype = "dashed"),
          ggplot2::geom_vline(xintercept = upper_val, color = pal[6],
                             linewidth = 1, linetype = "dashed")
        )
      }} +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Mean Estimate",
        subtitle = paste0("Mean = ", round(mean_val, 3)),
        x = "Value",
        y = ""
      ) +
      ggplot2::theme(axis.text.y = ggplot2::element_blank())
  }

  .wrap_plotly(p)
}

#' Plot fit predictions (per-observation dots with CI)
#' @keywords internal
#' @noRd
.plot_fit_pred <- function(pred, ...) {
  fit_df <- pred$fit
  if (!is.data.frame(fit_df) || !all(c("id", "estimate") %in% names(fit_df))) {
    stop("Fit prediction must return a data.frame with columns 'id' and 'estimate'.", call. = FALSE)
  }

  plot_data <- fit_df
  pal <- .plot_palette(max(2L, length(unique(plot_data$id))))
  p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = factor(id), y = estimate)) +
    ggplot2::geom_point(ggplot2::aes(color = factor(id)), size = 3) +
    {if ("lower" %in% names(plot_data) && "upper" %in% names(plot_data)) {
      ggplot2::geom_errorbar(ggplot2::aes(ymin = lower, ymax = upper, group = id, color = factor(id)),
                             width = 0.2, linewidth = 1)
    }} +
    ggplot2::scale_color_manual(values = pal) +
    .plot_theme() +
    ggplot2::labs(
      title = "Per-observation Fit Estimates with Pointwise Credible Intervals",
      x = "Observation (id)",
      y = "Estimate"
    )

  .wrap_plotly(p)
}

#' Plot density predictions
#' @keywords internal
#' @noRd
.plot_density_pred <- function(pred, ...) {
  fit_val <- pred$fit

  # Handle both data frame and vector cases
  if (is.data.frame(fit_val)) {
    plot_data <- fit_val
    x_col <- if ("y" %in% names(plot_data)) "y" else if ("grid" %in% names(plot_data)) "grid" else names(plot_data)[1]
    y_col <- if ("density" %in% names(plot_data)) "density" else names(plot_data)[2]
    has_id <- "id" %in% names(plot_data)
  } else {
    y_grid <- pred$grid %||% seq_along(fit_val)
    plot_data <- data.frame(
      y = y_grid,
      density = as.numeric(fit_val)
    )
    x_col <- "y"
    y_col <- "density"
    has_id <- FALSE
  }

  if (has_id) {
    n_id <- length(unique(plot_data$id))
    pal <- .plot_palette(max(2L, n_id))
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = .data[[x_col]], y = .data[[y_col]])) +
      ggplot2::geom_line(ggplot2::aes(group = id, color = factor(id)), linewidth = 1) +
      ggplot2::geom_point(ggplot2::aes(color = factor(id)), size = 1.5) +
      ggplot2::scale_color_manual(values = pal) +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Predictive Density",
        x = "Value",
        y = "Density"
      )
  } else {
    pal <- .plot_palette(2L)
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = .data[[x_col]], y = .data[[y_col]])) +
      ggplot2::geom_line(color = pal[1], linewidth = 1) +
      ggplot2::geom_point(color = pal[1], size = 2) +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Predictive Density",
        x = "Value",
        y = "Density"
      )
  }

  .wrap_plotly(p)
}

#' Plot survival function predictions
#' @keywords internal
#' @noRd
.plot_survival_pred <- function(pred, ...) {

  fit_val <- pred$fit

  # Handle data frame format from prediction
  if (is.data.frame(fit_val)) {
    plot_data <- fit_val
    if (!("y" %in% names(plot_data))) names(plot_data)[1] <- "y"
    if (!("survival" %in% names(plot_data))) names(plot_data)[2] <- "survival"
    has_id <- "id" %in% names(plot_data)
  } else {
    # Handle vector format (old style)
    y_vals <- pred$grid %||% seq_along(fit_val)
    plot_data <- data.frame(
      y = y_vals,
      survival = as.numeric(fit_val)
    )
    has_id <- FALSE
  }

  # Sort by y values for proper survival curve
  plot_data <- plot_data[order(plot_data$y), ]

  if (has_id) {
    n_id <- length(unique(plot_data$id))
    pal <- .plot_palette(max(2L, n_id))
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = y, y = survival)) +
      ggplot2::geom_step(ggplot2::aes(group = id, color = factor(id)), direction = "hv",
                         linewidth = 1) +
      ggplot2::geom_point(ggplot2::aes(color = factor(id)), size = 1.5) +
      ggplot2::scale_color_manual(values = pal) +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Predictive Survival Function",
        x = "Value",
        y = "Survival Probability"
      ) +
      ggplot2::ylim(0, 1)
  } else {
    pal <- .plot_palette(2L)
    p <- ggplot2::ggplot(plot_data, ggplot2::aes(x = y, y = survival)) +
      ggplot2::geom_step(direction = "hv", color = pal[1], linewidth = 1) +
      ggplot2::geom_point(color = pal[1], size = 2) +
      .plot_theme() +
      ggplot2::labs(
        title = "Posterior Predictive Survival Function",
        x = "Value",
        y = "Survival Probability"
      ) +
      ggplot2::ylim(0, 1)
  }

  .wrap_plotly(p)
}

#' Plot location predictions (mean + median with CIs)
#' @keywords internal
#' @noRd
.plot_location_pred <- function(pred, ...) {

  fit_df <- pred$fit

  if (!is.data.frame(fit_df)) {
    stop("Location prediction 'fit' must be a data frame.", call. = FALSE)
  }

  has_id <- "id" %in% names(fit_df)

  # Reshape to long format for plotting: one row per (id x measure)
  rows <- list()
  for (i in seq_len(nrow(fit_df))) {
    id_val <- if (has_id) fit_df$id[i] else i
    if ("mean" %in% names(fit_df)) {
      rows[[length(rows) + 1L]] <- data.frame(
        id = id_val,
        measure = "mean",
        estimate = fit_df$mean[i],
        lower = if ("mean_lower" %in% names(fit_df)) fit_df$mean_lower[i] else NA_real_,
        upper = if ("mean_upper" %in% names(fit_df)) fit_df$mean_upper[i] else NA_real_,
        stringsAsFactors = FALSE
      )
    }
    if ("median" %in% names(fit_df)) {
      rows[[length(rows) + 1L]] <- data.frame(
        id = id_val,
        measure = "median",
        estimate = fit_df$median[i],
        lower = if ("median_lower" %in% names(fit_df)) fit_df$median_lower[i] else NA_real_,
        upper = if ("median_upper" %in% names(fit_df)) fit_df$median_upper[i] else NA_real_,
        stringsAsFactors = FALSE
      )
    }
  }
  plot_data <- do.call(rbind, rows)

  pal <- .plot_palette(max(2L, length(unique(plot_data$measure))))

  p <- ggplot2::ggplot(plot_data, ggplot2::aes(
    x = factor(id), y = estimate, color = measure, shape = measure
  )) +
    ggplot2::geom_point(size = 3, position = ggplot2::position_dodge(width = 0.3)) +
    {
      if (any(!is.na(plot_data$lower)) && any(!is.na(plot_data$upper))) {
        ggplot2::geom_errorbar(
          ggplot2::aes(ymin = lower, ymax = upper),
          width = 0.2,
          position = ggplot2::position_dodge(width = 0.3)
        )
      }
    } +
    ggplot2::scale_color_manual(values = pal) +
    .plot_theme() +
    ggplot2::labs(
      title = "Location Estimates (Mean & Median)",
      x = if (has_id) "Observation" else "Index",
      y = "Value"
    )

  .wrap_plotly(p)
}

.fixture_path <- function(name) {
  file.path("tests", "testthat", "fixtures", name)
}

.load_fixture <- function(name) {
  path <- .fixture_path(name)
  if (!file.exists(path)) {
    skip(sprintf("Fixture not found: %s", path))
  }
  readRDS(path)
}

# ==========================================================
# 1) Generalized Pareto distribution (GPD)
# ==========================================================

.mean_norm_mix_weights <- function(w) {
  w <- as.numeric(w)
  if (!length(w) || any(!is.finite(w))) return(NULL)
  wsum <- sum(w)
  if (!is.finite(wsum) || wsum <= 0) return(NULL)
  w / wsum
}

#' Generalized Pareto distribution
#'
#' Scalar generalized Pareto distribution (GPD) utilities for threshold exceedances above
#' `threshold`. These NIMBLE-compatible functions provide the tail component used by the spliced
#' bulk-tail families elsewhere in the package.
#'
#' The parameterization is
#' \deqn{
#' G(x) = 1 - \left(1 + \xi \frac{x - u}{\sigma_u}\right)^{-1 / \xi}, \qquad x \ge u,
#' }
#' where `threshold = u`, `scale = sigma_u > 0`, and `shape = xi`. When `shape` approaches zero,
#' the distribution reduces to the exponential tail limit.
#'
#' These uppercase NIMBLE-compatible functions are scalar (`x`/`q` and `n = 1`).
#' For vectorized R usage, use [base_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param threshold Numeric scalar threshold at which the GPD is attached.
#' @param scale Numeric scalar GPD scale parameter; must be positive.
#' @param shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return `dGpd()` returns a numeric scalar density, `pGpd()` returns a numeric scalar CDF,
#'   `rGpd()` returns one random draw, and `qGpd()` returns a numeric quantile.
#'
#' @details
#' The associated density is
#' \deqn{
#' g(x) = \frac{1}{\sigma_u}
#' \left(1 + \xi \frac{x - u}{\sigma_u}\right)^{-1/\xi - 1},
#' }
#' on the support where \eqn{1 + \xi (x-u)/\sigma_u > 0}. When \eqn{\xi = 0}, this reduces to the
#' exponential density with mean excess \eqn{\sigma_u}. The GPD has finite mean only when
#' \eqn{\xi < 1}, and finite variance only when \eqn{\xi < 1/2}. Those existence conditions matter
#' for downstream predictive means in the package: spliced models with \eqn{\xi \ge 1} require
#' restricted means rather than ordinary means.
#'
#' If a bulk distribution has CDF \eqn{F_{bulk}}, the package's spliced families use the tail construction
#' \deqn{
#' F(x) =
#' \left\{
#' \begin{array}{ll}
#' F_{bulk}(x), & x < u, \\
#' F_{bulk}(u) + \{1 - F_{bulk}(u)\} G(x), & x \ge u.
#' \end{array}
#' \right.
#' }
#'
#' @seealso [base_lowercase()], [normal_gpd()], [lognormal_gpd()], [gamma_gpd()],
#'   [InvGauss_gpd()], [laplace_gpd()], [amoroso_gpd()].
#' @family base tail distributions
#'
#' @examples
#' threshold <- 1
#' tail_scale <- 0.8
#' tail_shape <- 0.2
#'
#' dGpd(1.5, threshold, tail_scale, tail_shape, log = 0)
#' pGpd(1.5, threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
#' qGpd(0.50, threshold, tail_scale, tail_shape)
#' qGpd(0.95, threshold, tail_scale, tail_shape)
#' replicate(10, rGpd(1, threshold, tail_scale, tail_shape))
#' @rdname gpd
#' @name gpd
#' @aliases dGpd pGpd rGpd qGpd
#' @importFrom stats runif
NULL

#' @describeIn gpd Generalized Pareto density function
#' @export
dGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 threshold = double(0),
                 scale = double(0),
                 shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (scale <= 0.0) {
      if (log == 1) return(log(eps)) else return(eps)
    }
    if (x < threshold) {
      if (log == 1) return(log(eps)) else return(eps)
    }

    z <- (x - threshold) / scale
    val <- 0.0

    if (abs(shape) < 1e-12) {
      # Exponential limit
      val <- (1.0 / scale) * exp(-z)
    } else {
      t <- 1.0 + shape * z
      if (t <= 0.0) {
        if (log == 1) return(log(eps)) else return(eps)
      }
      val <- (1.0 / scale) * (t ^ (-1.0 / shape - 1.0))
    }

    if (val < eps) val <- eps
    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn gpd Generalized Pareto distribution function
#' @export
pGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 threshold = double(0),
                 scale = double(0),
                 shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (scale <= 0.0) {
      if (log.p != 0) return(log(eps)) else return(eps)
    }

    cdf <- 0.0
    if (q < threshold) {
      cdf <- 0.0
    } else {
      z <- (q - threshold) / scale
      if (abs(shape) < 1e-12) {
        cdf <- 1.0 - exp(-z)
      } else {
        t <- 1.0 + shape * z
        if (t <= 0.0) {
          cdf <- 1.0
        } else {
          cdf <- 1.0 - (t^(-1.0 / shape))
        }
      }
      if (is.nan(cdf)) cdf <- 0.0
      if (cdf < 0.0) cdf <- 0.0
      if (cdf > 1.0) cdf <- 1.0
    }

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn gpd Generalized Pareto random generation
#' @export
rGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 threshold = double(0),
                 scale = double(0),
                 shape = double(0)) {
    returnType(double(0))
    if (n != 1) return(0.0)
    if (scale <= 0.0) return(0.0)

    u <- runif(1, 0.0, 1.0)

    if (abs(shape) < 1e-12) {
      return(threshold - scale * log(1.0 - u))
    }
    return(threshold + (scale / shape) * ((1.0 - u)^(-shape) - 1.0))
  }
)

#' @describeIn gpd Generalized Pareto quantile function
#' @export
qGpd <- function(p, threshold, scale, shape,
                 lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) {
      out[i] <- threshold
      next
    }
    if (pi >= 1) {
      if (shape < 0) {
        out[i] <- threshold - scale / shape
      } else {
        out[i] <- Inf
      }
      next
    }

    if (abs(shape) < 1e-12) {
      out[i] <- threshold - scale * log(1.0 - pi)
    } else {
      out[i] <- threshold + (scale / shape) * ((1.0 - pi)^(-shape) - 1.0)
    }
  }
  out
}


# ==========================================================
# 2) Inverse Gaussian (custom base)
# ==========================================================

#' Inverse Gaussian (Wald) distribution
#'
#' Scalar inverse Gaussian utilities under the \eqn{(\mu, \lambda)} parameterization, where
#' `mean = mu > 0` and `shape = lambda > 0`. These functions are used directly and as building
#' blocks for inverse-Gaussian mixtures and spliced inverse-Gaussian-plus-GPD families.
#'
#' The density is
#' \deqn{
#' f(x) = \left(\frac{\lambda}{2 \pi x^3}\right)^{1/2}
#' \exp\left\{- \frac{\lambda (x - \mu)^2}{2 \mu^2 x}\right\}, \qquad x > 0.
#' }
#'
#' These uppercase NIMBLE-compatible functions are scalar (`x`/`q` and `n = 1`).
#' For vectorized R usage, use [base_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar giving the probability for the quantile.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param mean Numeric scalar mean parameter \eqn{\mu>0}.
#' @param shape Numeric scalar shape parameter \eqn{\lambda>0}.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return `dInvGauss()` returns a numeric scalar density, `pInvGauss()` returns a numeric scalar
#'   CDF, `rInvGauss()` returns one random draw, and `qInvGauss()` returns a numeric quantile.
#'
#' @details
#' The inverse Gaussian is the first-passage-time distribution of a Brownian motion with positive
#' drift. Under the \eqn{(\mu, \lambda)} parameterization used here, the mean is \eqn{E(X)=\mu} and
#' the variance is \eqn{\mathrm{Var}(X)=\mu^3/\lambda}. The implementation follows that
#' parameterization throughout the package, so inverse-Gaussian mixture and splice families inherit
#' the same interpretation.
#'
#' The distribution function is evaluated through the standard normal representation
#' \deqn{
#' F(x) =
#' \Phi\left(\sqrt{\frac{\lambda}{x}}\left(\frac{x}{\mu}-1\right)\right) +
#' \exp\left(\frac{2\lambda}{\mu}\right)
#' \Phi\left(-\sqrt{\frac{\lambda}{x}}\left(\frac{x}{\mu}+1\right)\right),
#' }
#' and the quantile is obtained numerically because no simple closed form is available.
#'
#' @seealso [InvGauss_mix()], [InvGauss_gpd()], [base_lowercase()], [build_nimble_bundle()].
#' @family base bulk distributions
#'
#' @examples
#' mean <- 2
#' shape <- 5
#'
#' dInvGauss(2.0, mean, shape, log = 0)
#' pInvGauss(2.0, mean, shape, lower.tail = 1, log.p = 0)
#' qInvGauss(0.50, mean, shape)
#' qInvGauss(0.95, mean, shape)
#' replicate(10, rInvGauss(1, mean, shape))

#'
#' @rdname InvGauss
#' @name InvGauss
#' @aliases dInvGauss pInvGauss rInvGauss qInvGauss
#' @importFrom stats pnorm rnorm runif uniroot
NULL

#' @describeIn InvGauss Inverse Gaussian density function
#' @export
dInvGauss <- nimble::nimbleFunction(
  run = function(x = double(0),
                 mean = double(0),
                 shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    # standard IG density
    z <- shape * (x - mean) * (x - mean) / (2.0 * mean * mean * x)
    logdens <- 0.5 * log(shape) - 0.5 * log(2.0 * pi) -
      1.5 * log(x) - z

    if (log == 1L) return(logdens)
    return(exp(logdens))
  }
)

#' @describeIn InvGauss Inverse Gaussian distribution function
#' @export
pInvGauss <- nimble::nimbleFunction(
  run = function(q = double(0),
                 mean = double(0),
                 shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    bad <- 0
    if (q <= 0.0) bad <- 1
    if (mean <= 0.0) bad <- 1
    if (shape <= 0.0) bad <- 1
    if (bad == 1) {
      if (lower.tail == 0L) {
        if (log.p == 1L) return(0.0)
        return(1.0)
      }
      if (log.p == 1L) return(-Inf)
      return(0.0)
    }

    z1 <- sqrt(shape / q) * (q / mean - 1.0)
    z2 <- -sqrt(shape / q) * (q / mean + 1.0)

    cdf <- pnorm(z1, 0.0, 1.0, 1L, 0L) +
      exp(2.0 * shape / mean) * pnorm(z2, 0.0, 1.0, 1L, 0L)

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0L) cdf <- 1.0 - cdf
    if (log.p == 1L) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn InvGauss Inverse Gaussian random generation
#' @export
rInvGauss <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 mean = double(0),
                 shape = double(0)) {
    returnType(double(0))
    if (n != 1L) return(0.0)

    v <- rnorm(1, 0.0, 1.0)
    y <- v * v
    x <- mean + (mean * mean * y) / (2.0 * shape) -
      (mean / (2.0 * shape)) * sqrt(4.0 * mean * shape * y + mean * mean * y * y)

    u <- runif(1, 0.0, 1.0)
    if (u <= mean / (mean + x)) return(x)
    return(mean * mean / x)
  }
)

#' @describeIn InvGauss Inverse Gaussian quantile function
#' @export
qInvGauss <- function(p, mean, shape,
                      lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) {
      out[i] <- 0
      next
    }
    if (pi >= 1) {
      out[i] <- Inf
      next
    }
    hi <- max(1, mean * 10)
    f0 <- as.numeric(pInvGauss(0, mean, shape) - pi)
    fhi <- as.numeric(pInvGauss(hi, mean, shape) - pi)
    iter <- 0L
    while (is.finite(fhi) && f0 * fhi > 0 && hi < 1e20 && iter < 60L) {
      hi <- hi * 2
      fhi <- as.numeric(pInvGauss(hi, mean, shape) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(fhi) || f0 * fhi > 0) {
      out[i] <- Inf
    } else {
      out[i] <- stats::uniroot(function(q) pInvGauss(q, mean, shape) - pi,
                               interval = c(0, hi),
                               tol = tol, maxiter = maxiter)$root
    }
  }
  out
}


# ==========================================================
# 3) Amoroso base kernels
# ==========================================================
# NOTE: You said Amoroso base functions already exist in your Amoroso script.
# We do not rename or redefine them here. This block provides documentation linkage only.

#' Amoroso distribution
#'
#' Scalar Amoroso utilities used as flexible positive-support base kernels and mixture components.
#' The package parameterization uses `loc`, `scale`, `shape1`, and `shape2`, allowing the family
#' to represent a broad range of skewed bulk shapes.
#'
#' Writing `loc = a`, `scale = theta`, `shape1 = alpha`, and `shape2 = beta`, the transformed
#' quantity \eqn{((X - a) / \theta)^\beta} follows a gamma law with shape \eqn{\alpha}.
#'
#' These uppercase NIMBLE-compatible functions are scalar (`x`/`q` and `n = 1`).
#' For vectorized R usage, use [base_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param loc Numeric scalar location parameter.
#' @param scale Numeric scalar scale parameter.
#' @param shape1 Numeric scalar first shape parameter.
#' @param shape2 Numeric scalar second shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return Density/CDF/RNG functions return numeric scalars. The quantile function returns a numeric
#'   scalar or vector matching the length of `p`.
#'
#' @details
#' The Amoroso family used in the package is defined by the density
#' \deqn{
#' f(x) =
#' \left|\frac{\beta}{\theta}\right|
#' \frac{z^{\alpha \beta - 1}\exp(-z^\beta)}{\Gamma(\alpha)},
#' \qquad
#' z = \frac{x-a}{\theta},
#' }
#' on the side of the location parameter determined by the sign of \eqn{\theta}. Equivalently,
#' \eqn{Z = ((X-a)/\theta)^\beta} follows a Gamma distribution with shape \eqn{\alpha} and unit
#' scale. That representation explains why the quantile function is computed from a gamma quantile
#' and then mapped back through the inverse transformation.
#'
#' The mean exists whenever \eqn{\alpha + 1/\beta} lies in the domain of the gamma function used by
#' the moment formula. In the package this family serves as a flexible positive-support bulk kernel
#' capable of reproducing gamma-like, Weibull-like, and other skewed shapes with a single
#' parameterization.
#'
#' @seealso [amoroso_mix()], [amoroso_gpd()], [base_lowercase()], [kernel_support_table()].
#' @family base bulk distributions
#'
#' @examples
#' loc <- 0
#' scale <- 1.5
#' shape1 <- 2
#' shape2 <- 1.2
#'
#' dAmoroso(1.0, loc, scale, shape1, shape2, log = 0)
#' pAmoroso(1.0, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)
#' qAmoroso(0.50, loc, scale, shape1, shape2)
#' qAmoroso(0.95, loc, scale, shape1, shape2)
#' replicate(10, rAmoroso(1, loc, scale, shape1, shape2))

#' @rdname amoroso
#' @name amoroso
#' @aliases dAmoroso pAmoroso rAmoroso qAmoroso
NULL

#' @describeIn amoroso Density Function of Amoroso Distribution
#' @export
dAmoroso <- nimble::nimbleFunction(
  run = function(x = double(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    if (scale == 0.0) {
      if (log == 1) return(log(eps)) else return(eps)
    }
    if (scale > 0.0) {
      if (x < loc) {
        if (log == 1) return(log(eps)) else return(eps)
      }
    }
    if (scale < 0.0) {
      if (x > loc) {
        if (log == 1) return(log(eps)) else return(eps)
      }
    }

    z <- (x - loc) / scale
    lik <- abs(shape2 / scale) * (z^(shape1 * shape2 - 1.0)) * exp(-z^shape2) / gamma(shape1)
    if (lik < eps) lik <- eps
    if (log == 1) return(log(lik)) else return(lik)
  }
)

#' @describeIn amoroso Distribution Function of Amoroso Distribution
#' @export
pAmoroso <- nimble::nimbleFunction(
  run = function(q = double(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    if (scale == 0.0) {
      if (log.p != 0) return(log(eps)) else return(eps)
    }
    if (scale > 0.0) {
      if (q <= loc) {
        cdf <- 0.0
        if (lower.tail == 0) cdf <- 1.0 - cdf
        if (log.p != 0) return(log(max(cdf, eps)))
        return(cdf)
      }
    }
    if (scale < 0.0) {
      if (q >= loc) {
        cdf <- 1.0
        if (lower.tail == 0) cdf <- 1.0 - cdf
        if (log.p != 0) return(log(max(cdf, eps)))
        return(cdf)
      }
    }

    z <- ((q - loc) / scale)^shape2
    cdf <- pgamma(z, shape = shape1, scale = 1.0)
    if (shape2 < 0) cdf <- 1.0 - cdf
    cdf <- max(min(cdf, 1.0), 0.0)
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)

#' @describeIn amoroso Quantile Function of Amoroso Distribution
#' @export
qAmoroso <- function(p, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p

  len <- max(length(p), length(loc), length(scale), length(shape1), length(shape2))
  p <- rep_len(p, len)
  loc <- rep_len(loc, len)
  scale <- rep_len(scale, len)
  shape1 <- rep_len(shape1, len)
  shape2 <- rep_len(shape2, len)

  p <- ifelse(shape2 < 0, 1 - p, p)
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(len)
  at_zero <- p <= 0
  at_one <- p >= 1
  out[at_zero] <- loc[at_zero]
  out[at_one] <- ifelse(scale[at_one] > 0, Inf, -Inf)

  mid <- !(at_zero | at_one)
  if (any(mid)) {
    z <- stats::qgamma(p[mid], shape = shape1[mid], scale = 1.0)
    out[mid] <- loc[mid] + scale[mid] * (z^(1 / shape2[mid]))
  }

  out
}

#' @describeIn amoroso Sample generating Function of Amoroso Distribution
#' @export
rAmoroso <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 loc = double(0),
                 scale = double(0),
                 shape1 = double(0),
                 shape2 = double(0)) {
    returnType(double(0))
    if (n != 1) return(loc)
    p <- runif(1, 0.0, 1.0)
    # call R quantile via nimbleFunction's R-call boundary isn't allowed; use gamma inverse and algebra
    if (shape2 < 0) p <- 1.0 - p
    z <- qgamma(p, shape = shape1, scale = 1.0, lower.tail = 1, log.p = 0)
    return(loc + scale * (z)^(1.0 / shape2))
  }
)


# ==========================================================
# 4) Cauchy base kernels
# ==========================================================

#' Cauchy distribution
#'
#' Scalar Cauchy utilities implemented for NIMBLE compatibility. These functions support symmetric
#' heavy-tailed kernels on the real line and feed directly into the finite Cauchy mixture family.
#'
#' The density is
#' \deqn{
#' f(x) = \frac{1}{\pi s \{1 + ((x - \ell)/s)^2\}},
#' }
#' where `location = ell` and `scale = s > 0`.
#'
#' These uppercase NIMBLE-compatible functions are scalar (`x`/`q` and `n = 1`).
#' For vectorized R usage, use [base_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param location Numeric scalar location parameter.
#' @param scale Numeric scalar scale parameter; must be positive.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return `dCauchy()` returns a numeric scalar density, `pCauchy()` returns a numeric scalar CDF,
#'   `rCauchy()` returns one random draw, and `qCauchy()` returns a numeric quantile.
#'
#' @details
#' The Cauchy law is a stable heavy-tailed distribution with undefined mean and variance. That is
#' why the package allows the Cauchy kernel only as a bulk distribution and deliberately does not
#' pair it with GPD tails in the kernel registry. For predictive summaries,
#' ordinary means are not available under Cauchy kernels; medians, quantiles,
#' survival curves, and restricted means remain well defined.
#'
#' The distribution function is
#' \deqn{
#' F(x) = \frac{1}{2} + \frac{1}{\pi}\arctan\left(\frac{x-\ell}{s}\right),
#' }
#' and the quantile is the corresponding inverse
#' \deqn{
#' Q(p) = \ell + s \tan\{\pi(p-1/2)\}.
#' }
#'
#' @seealso [cauchy_mix()], [base_lowercase()], [kernel_support_table()].
#' @family base bulk distributions
#'
#' @examples
#' location <- 0
#' scale <- 1.5
#'
#' dCauchy(0.5, location, scale, log = 0)
#' pCauchy(0.5, location, scale, lower.tail = 1, log.p = 0)
#' qCauchy(0.50, location, scale)
#' qCauchy(0.95, location, scale)
#' replicate(10, rCauchy(1, location, scale))
#'
#' @rdname cauchy
#' @name cauchy
#' @aliases dCauchy pCauchy rCauchy qCauchy
#' @importFrom stats runif
NULL

#' @describeIn cauchy Cauchy density function
#' @export
dCauchy <- nimble::nimbleFunction(
  run = function(x = double(0),
                 location = double(0),
                 scale = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    if (scale <= 0.0) {
      if (log == 1) return(log(eps)) else return(eps)
    }
    z <- (x - location) / scale
    val <- 1.0 / (pi * scale * (1.0 + z * z))
    if (val < eps) val <- eps
    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn cauchy Cauchy distribution function
#' @export
pCauchy <- nimble::nimbleFunction(
  run = function(q = double(0),
                 location = double(0),
                 scale = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300
    if (scale <= 0.0) {
      if (log.p != 0) return(log(eps)) else return(eps)
    }
    z <- (q - location) / scale
    cdf <- 0.5 + atan(z) / pi
    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) return(log(max(cdf, eps)))
    return(cdf)
  }
)

#' @describeIn cauchy Cauchy random generation
#' @export
rCauchy <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 location = double(0),
                 scale = double(0)) {
    returnType(double(0))
    if (n != 1) return(location)
    if (scale <= 0.0) return(location)
    u <- runif(1, 0.0, 1.0)
    return(location + scale * tan(pi * (u - 0.5)))
  }
)

#' @describeIn cauchy Cauchy quantile function
#' @export
qCauchy <- function(p, location, scale,
                    lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  # Handle boundary cases for vectorized input
  out <- location + scale * tan(pi * (p - 0.5))
  out[p <= 0] <- -Inf
  out[p >= 1] <- Inf
  out
}


# ==========================================================
# 5) Lowercase vectorized R wrappers for base kernels
# ==========================================================

#' Lowercase vectorized distribution functions (base kernels)
#'
#' Vectorized R wrappers around the scalar base-kernel functions defined in this file. These
#' helpers are intended for interactive R use, examples, testing, and checking numerical behavior
#' outside compiled NIMBLE code.
#'
#' The wrappers preserve the same parameterizations as the uppercase scalar functions, but accept
#' vector inputs for `x`, `q`, or `p` and allow `n > 1` for random generation.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param threshold,scale,shape,mean,loc,shape1,shape2,location
#'   Distribution parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are
#'   \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' Each lowercase helper is a vectorized R wrapper around the corresponding
#' uppercase scalar routine documented in this file. The wrapper keeps the same
#' parameterization and applies the scalar kernel repeatedly over the supplied
#' evaluation points or simulation index. These helpers are therefore
#' appropriate for interactive analysis, testing, and examples, whereas the
#' uppercase functions are the building blocks used inside NIMBLE model code.
#'
#' The wrappers do not change the underlying theory. For example, \code{qgpd()} still uses the
#' closed-form GPD inverse, \code{qinvgauss()} still performs numerical inversion of the inverse
#' Gaussian CDF, and \code{qamoroso()} still maps a gamma quantile through the Amoroso
#' transformation. Random-generation wrappers call the corresponding scalar RNG repeatedly when
#' \code{n > 1}.
#'
#' @seealso [gpd()], [InvGauss()], [amoroso()], [cauchy()], [build_nimble_bundle()],
#'   [kernel_support_table()].
#' @family vectorized kernel helpers
#'
#' @examples
#' # GPD
#' dgpd(c(1.5, 2.0, 2.5), threshold = 1, scale = 0.8, shape = 0.2)
#' pgpd(c(1.5, 2.0), threshold = 1, scale = 0.8, shape = 0.2)
#' qgpd(c(0.5, 0.9), threshold = 1, scale = 0.8, shape = 0.2)
#' rgpd(5, threshold = 1, scale = 0.8, shape = 0.2)
#'
#' # Inverse Gaussian
#' dinvgauss(c(1, 2, 3), mean = 2, shape = 5)
#' rinvgauss(5, mean = 2, shape = 5)
#'
#' # Amoroso
#' damoroso(c(1, 2), loc = 0, scale = 1.5, shape1 = 2, shape2 = 1.2)
#' ramoroso(5, loc = 0, scale = 1.5, shape1 = 2, shape2 = 1.2)
#'
#' # Cauchy
#' dcauchy_vec(c(-1, 0, 1), location = 0, scale = 1)
#' rcauchy_vec(5, location = 0, scale = 1)
#'
#' @name base_lowercase
#' @rdname base_lowercase
NULL

# ---- GPD lowercase wrappers ----

#' @describeIn base_lowercase GPD density (vectorized)
#' @export
dgpd <- function(x, threshold, scale, shape, log = FALSE) {

  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dGpd(xi, threshold, scale, shape, log_int)),
         numeric(1L))
}

#' @describeIn base_lowercase GPD distribution function (vectorized)
#' @export
pgpd <- function(q, threshold, scale, shape, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)

  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pGpd(qi, threshold, scale, shape, lt_int, lp_int)),
         numeric(1L))
}

#' @describeIn base_lowercase GPD quantile function (vectorized)
#' @export
qgpd <- function(p, threshold, scale, shape, lower.tail = TRUE, log.p = FALSE) {
  qGpd(p, threshold, scale, shape, lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn base_lowercase GPD random generation (vectorized)
#' @export
rgpd <- function(n, threshold, scale, shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rGpd(1L, threshold, scale, shape)),
         numeric(1L))
}

# ---- Inverse Gaussian lowercase wrappers ----

#' @describeIn base_lowercase Inverse Gaussian density (vectorized)
#' @export
dinvgauss <- function(x, mean, shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dInvGauss(xi, mean, shape, log_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Inverse Gaussian distribution function (vectorized)
#' @export
pinvgauss <- function(q, mean, shape, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pInvGauss(qi, mean, shape, lt_int, lp_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Inverse Gaussian quantile function (vectorized)
#' @export
qinvgauss <- function(p, mean, shape, lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-10, maxiter = 200) {
  qInvGauss(p, mean, shape, lower.tail = lower.tail, log.p = log.p,
            tol = tol, maxiter = maxiter)
}

#' @describeIn base_lowercase Inverse Gaussian random generation (vectorized)
#' @export
rinvgauss <- function(n, mean, shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rInvGauss(1L, mean, shape)),
         numeric(1L))
}

# ---- Amoroso lowercase wrappers ----

#' @describeIn base_lowercase Amoroso density (vectorized)
#' @export
damoroso <- function(x, loc, scale, shape1, shape2, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dAmoroso(xi, loc, scale, shape1, shape2, log_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Amoroso distribution function (vectorized)
#' @export
pamoroso <- function(q, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pAmoroso(qi, loc, scale, shape1, shape2, lt_int, lp_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Amoroso quantile function (vectorized)
#' @export
qamoroso <- function(p, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE) {
  qAmoroso(p, loc, scale, shape1, shape2, lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn base_lowercase Amoroso random generation (vectorized)
#' @export
ramoroso <- function(n, loc, scale, shape1, shape2) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rAmoroso(1L, loc, scale, shape1, shape2)),
         numeric(1L))
}

# ---- Cauchy lowercase wrappers ----

#' @describeIn base_lowercase Cauchy density (vectorized)
#' @export
dcauchy_vec <- function(x, location, scale, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dCauchy(xi, location, scale, log_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Cauchy distribution function (vectorized)
#' @export
pcauchy_vec <- function(q, location, scale, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pCauchy(qi, location, scale, lt_int, lp_int)),
         numeric(1L))
}

#' @describeIn base_lowercase Cauchy quantile function (vectorized)
#' @export
qcauchy_vec <- function(p, location, scale, lower.tail = TRUE, log.p = FALSE) {
  qCauchy(p, location, scale, lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn base_lowercase Cauchy random generation (vectorized)
#' @export
rcauchy_vec <- function(n, location, scale) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rCauchy(1L, location, scale)),
         numeric(1L))
}

if (!identical(Sys.getenv("DPMIXGPD_CI_COVERAGE_ONLY"), "1")) {
  testthat::skip("Coverage-only suite disabled. Set DPMIXGPD_CI_COVERAGE_ONLY=1 to run.")
}

`%||%` <- function(a, b) if (!is.null(a)) a else b

.coverage_quiet <- function(expr) {
  nullfile <- if (.Platform$OS.type == "windows") "NUL" else "/dev/null"
  utils::capture.output(result <- force(expr), file = nullfile)
  result
}

.coverage_cache <- new.env(parent = emptyenv())

.coverage_cached <- function(key, expr) {
  if (exists(key, envir = .coverage_cache, inherits = FALSE)) {
    return(get(key, envir = .coverage_cache, inherits = FALSE))
  }
  value <- force(expr)
  assign(key, value, envir = .coverage_cache)
  value
}

.coverage_mcmc <- function(seed = 1L, niter = 20L, nburnin = 5L) {
  list(niter = niter, nburnin = nburnin, thin = 1L, nchains = 1L, seed = seed)
}

.coverage_kernel_case <- function(kernel) {
  switch(
    kernel,
    normal = list(
      prefix = "norm",
      values = list(mean = c(-0.75, 0.25, 1.5), sd = c(0.6, 1.0, 1.4)),
      scalar = list(mean = 0.3, sd = 1.1),
      x = c(-1.5, 0.0, 1.5)
    ),
    gamma = list(
      prefix = "gamma",
      values = list(shape = c(2, 4, 6), scale = c(0.8, 1.2, 1.8)),
      scalar = list(shape = 3, scale = 1.4),
      x = c(0.5, 1.5, 3.0)
    ),
    lognormal = list(
      prefix = "lognormal",
      values = list(meanlog = c(-0.2, 0.3, 0.9), sdlog = c(0.4, 0.7, 1.0)),
      scalar = list(meanlog = 0.4, sdlog = 0.7),
      x = c(0.4, 1.2, 3.5)
    ),
    invgauss = list(
      prefix = "invgauss",
      values = list(mean = c(0.8, 1.5, 2.5), shape = c(2, 4, 8)),
      scalar = list(mean = 1.2, shape = 5),
      x = c(0.5, 1.0, 2.5)
    ),
    laplace = list(
      prefix = "laplace",
      values = list(location = c(-1, 0.5, 2), scale = c(0.6, 1.0, 1.4)),
      scalar = list(location = 0.4, scale = 1.0),
      x = c(-1.5, 0.0, 1.8)
    ),
    cauchy = list(
      prefix = "cauchy",
      values = list(location = c(-1, 0.5, 1.5), scale = c(0.8, 1.2, 1.6)),
      scalar = list(location = 0.25, scale = 1.1),
      x = c(-2.0, 0.0, 2.0)
    ),
    amoroso = list(
      prefix = "amoroso",
      values = list(
        loc = c(0, 0.2, 0.5),
        scale = c(1.0, 1.5, 2.0),
        shape1 = c(1.5, 2.0, 3.0),
        shape2 = c(0.8, 1.2, 1.6)
      ),
      scalar = list(loc = 0.1, scale = 1.4, shape1 = 2.0, shape2 = 1.2),
      x = c(0.3, 1.0, 3.0)
    )
  )
}

.coverage_numeric <- function(x, n) {
  testthat::expect_type(x, "double")
  testthat::expect_length(x, n)
  testthat::expect_true(all(is.finite(x) | is.infinite(x)))
}

.coverage_run_kernel_wrappers <- function(kernel) {
  case <- .coverage_kernel_case(kernel)
  prefix <- case$prefix
  w <- c(0.5, 0.3, 0.2)
  probs <- c(0.25, 0.75)
  tail_args <- list(threshold = max(case$x) * 0.7, tail_scale = 1.1, tail_shape = 0.15)

  d_mix <- get(paste0("d", prefix, "mix"), mode = "function")
  p_mix <- get(paste0("p", prefix, "mix"), mode = "function")
  q_mix <- get(paste0("q", prefix, "mix"), mode = "function")
  r_mix <- get(paste0("r", prefix, "mix"), mode = "function")

  .coverage_numeric(do.call(d_mix, c(list(x = case$x, w = w), case$values, list(log = FALSE))), length(case$x))
  .coverage_numeric(do.call(p_mix, c(list(q = case$x, w = w), case$values, list(lower.tail = TRUE, log.p = FALSE))), length(case$x))
  .coverage_numeric(do.call(q_mix, c(list(p = probs, w = w), case$values, list(lower.tail = TRUE, log.p = FALSE))), length(probs))
  .coverage_numeric(do.call(r_mix, c(list(n = 4L, w = w), case$values)), 4L)

  if (!identical(kernel, "cauchy")) {
    d_mix_gpd <- get(paste0("d", prefix, "mixgpd"), mode = "function")
    p_mix_gpd <- get(paste0("p", prefix, "mixgpd"), mode = "function")
    q_mix_gpd <- get(paste0("q", prefix, "mixgpd"), mode = "function")
    r_mix_gpd <- get(paste0("r", prefix, "mixgpd"), mode = "function")

    d_gpd <- get(paste0("d", prefix, "gpd"), mode = "function")
    p_gpd <- get(paste0("p", prefix, "gpd"), mode = "function")
    q_gpd <- get(paste0("q", prefix, "gpd"), mode = "function")
    r_gpd <- get(paste0("r", prefix, "gpd"), mode = "function")

    .coverage_numeric(do.call(d_mix_gpd, c(list(x = case$x, w = w), case$values, tail_args, list(log = FALSE))), length(case$x))
    .coverage_numeric(do.call(p_mix_gpd, c(list(q = case$x, w = w), case$values, tail_args, list(lower.tail = TRUE, log.p = FALSE))), length(case$x))
    .coverage_numeric(do.call(q_mix_gpd, c(list(p = probs, w = w), case$values, tail_args, list(lower.tail = TRUE, log.p = FALSE))), length(probs))
    .coverage_numeric(do.call(r_mix_gpd, c(list(n = 4L, w = w), case$values, tail_args)), 4L)

    .coverage_numeric(do.call(d_gpd, c(list(x = case$x), case$scalar, tail_args, list(log = FALSE))), length(case$x))
    .coverage_numeric(do.call(p_gpd, c(list(q = case$x), case$scalar, tail_args, list(lower.tail = TRUE, log.p = FALSE))), length(case$x))
    .coverage_numeric(do.call(q_gpd, c(list(p = probs), case$scalar, tail_args, list(lower.tail = TRUE, log.p = FALSE))), length(probs))
    .coverage_numeric(do.call(r_gpd, c(list(n = 4L), case$scalar, tail_args)), 4L)
  }
}

.coverage_conditional_fit <- function() {
  .coverage_cached("conditional_fit", {
    set.seed(41)
    n <- 24L
    X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n, -1, 1))
    y <- abs(1 + X[, 1] + 0.4 * X[, 2] + stats::rnorm(n, sd = 0.25)) + 0.2

    fit <- dpmgpd(
      y = y,
      X = X,
      backend = "crp",
      kernel = "gamma",
      components = 3,
      mcmc = c(.coverage_mcmc(seed = 41L), list(show_progress = FALSE, quiet = TRUE))
    )

    list(fit = fit, y = y, X = X)
  })
}

.coverage_unconditional_fit <- function() {
  .coverage_cached("unconditional_fit", {
    y <- sim_bulk_tail(n = 24, seed = 7)
    fit <- dpmix(
      y = y,
      backend = "sb",
      kernel = "normal",
      components = 3,
      mcmc = c(.coverage_mcmc(seed = 7L), list(show_progress = FALSE, quiet = TRUE))
    )

    list(fit = fit, y = y)
  })
}

.coverage_spliced_bundle <- function() {
  .coverage_cached("spliced_bundle", {
    set.seed(13)
    y <- sim_bulk_tail(n = 28, seed = 13)
    X <- cbind(x1 = stats::rnorm(28), x2 = stats::runif(28))

    build_nimble_bundle(
      y = y,
      X = X,
      backend = "spliced",
      kernel = "gamma",
      GPD = TRUE,
      components = 3,
      param_specs = list(
        gpd = list(
          threshold = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
          tail_scale = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
          tail_shape = list(mode = "fixed", value = 0.1)
        )
      ),
      mcmc = .coverage_mcmc(seed = 13L)
    )
  })
}

.coverage_causal_fit <- function() {
  .coverage_cached("causal_fit", {
    sim <- sim_causal_qte(n = 26, seed = 29)
    sim$y <- abs(sim$y) + 0.2

    fit <- dpmix.causal(
      y = sim$y,
      X = as.matrix(sim$X),
      treat = sim$t,
      backend = c("sb", "crp"),
      kernel = c("normal", "gamma"),
      components = c(3, 3),
      PS = "logit",
      mcmc = c(.coverage_mcmc(seed = 29L), list(show_progress = FALSE))
    )

    list(fit = fit, sim = sim)
  })
}

.coverage_cluster_fit <- function() {
  .coverage_cached("cluster_fit", {
    set.seed(53)
    dat <- data.frame(
      y = abs(stats::rnorm(20)) + 0.2,
      x1 = stats::rnorm(20),
      x2 = stats::runif(20)
    )

    fit <- dpmix.cluster(
      y ~ x1 + x2,
      data = dat,
      kernel = "normal",
      components = 4,
      type = "weights",
      mcmc = .coverage_mcmc(seed = 53L)
    )

    list(fit = fit, data = dat)
  })
}

.coverage_ns_fun <- function(name) {
  getFromNamespace(name, "CausalMixGPD")
}

.coverage_mock_mixgpd_params <- function(empty = FALSE) {
  out <- if (empty) {
    list()
  } else {
    list(
      alpha = 1.5,
      w = c(0.6, 0.3, 0.1),
      mu = c(1, 2, 3),
      sigma = c(0.5, 1, 1.5),
      beta_mu = matrix(
        c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6),
        nrow = 3,
        ncol = 2,
        dimnames = list(paste0("comp", 1:3), c("x1", "x2"))
      )
    )
  }
  class(out) <- "mixgpd_params"
  out
}

.coverage_mock_mixgpd_params_pair <- function() {
  out <- list(
    treated = .coverage_mock_mixgpd_params(),
    control = .coverage_mock_mixgpd_params()
  )
  class(out) <- "mixgpd_params_pair"
  out
}

.coverage_mock_qte <- function() {
  out <- list(
    type = "qte",
    probs = c(0.25, 0.5, 0.75),
    grid = c(0.25, 0.5, 0.75),
    n_pred = 5,
    level = 0.95,
    interval = "credible",
    meta = list(
      backend = list(trt = "sb", con = "sb"),
      kernel = list(trt = "normal", con = "normal"),
      GPD = list(trt = FALSE, con = FALSE)
    ),
    ps = stats::runif(5),
    x = matrix(stats::rnorm(10), nrow = 5, ncol = 2),
    qte = list(
      fit = data.frame(
        id = rep(1:5, each = 3),
        index = rep(c(0.25, 0.5, 0.75), times = 5),
        estimate = stats::rnorm(15),
        lower = stats::rnorm(15) - 0.5,
        upper = stats::rnorm(15) + 0.5
      )
    ),
    fit = matrix(stats::rnorm(15), nrow = 5, ncol = 3),
    lower = matrix(stats::rnorm(15) - 0.5, nrow = 5, ncol = 3),
    upper = matrix(stats::rnorm(15) + 0.5, nrow = 5, ncol = 3),
    trt = list(
      fit = data.frame(
        estimate = stats::rnorm(15),
        id = rep(1:5, 3),
        index = rep(c(0.25, 0.5, 0.75), each = 5)
      )
    ),
    con = list(
      fit = data.frame(
        estimate = stats::rnorm(15),
        id = rep(1:5, 3),
        index = rep(c(0.25, 0.5, 0.75), each = 5)
      )
    )
  )
  class(out) <- "causalmixgpd_qte"
  out
}

.coverage_mock_ate <- function() {
  out <- list(
    type = "ate",
    n_pred = 5,
    level = 0.95,
    interval = "credible",
    nsim_mean = 200,
    meta = list(
      backend = list(trt = "sb", con = "sb"),
      kernel = list(trt = "normal", con = "normal"),
      GPD = list(trt = FALSE, con = FALSE)
    ),
    ps = stats::runif(5),
    x = matrix(stats::rnorm(10), nrow = 5, ncol = 2),
    ate = list(
      fit = data.frame(
        id = 1:5,
        estimate = stats::rnorm(5),
        lower = stats::rnorm(5) - 0.5,
        upper = stats::rnorm(5) + 0.5
      )
    ),
    fit = stats::rnorm(5),
    lower = stats::rnorm(5) - 0.5,
    upper = stats::rnorm(5) + 0.5,
    trt = list(fit = data.frame(estimate = stats::rnorm(5), id = 1:5)),
    con = list(fit = data.frame(estimate = stats::rnorm(5), id = 1:5))
  )
  class(out) <- "causalmixgpd_ate"
  out
}

.coverage_mock_mixgpd_predict <- function(type = "quantile") {
  out <- switch(
    type,
    quantile = list(
      fit = data.frame(
        estimate = c(1, 2, 3),
        index = c(0.25, 0.5, 0.75),
        lower = c(0.5, 1.5, 2.5),
        upper = c(1.5, 2.5, 3.5)
      ),
      type = "quantile",
      grid = c(0.25, 0.5, 0.75)
    ),
    sample = list(fit = stats::rnorm(100), type = "sample", grid = NULL),
    mean = list(fit = data.frame(estimate = 5.2, lower = 4.8, upper = 5.6), type = "mean", grid = NULL),
    density = {
      y_grid <- seq(0, 5, length.out = 20)
      list(
        fit = data.frame(
          id = rep(1L, 20),
          y = y_grid,
          density = stats::dnorm(y_grid, mean = 2.5, sd = 1),
          lower = stats::dnorm(y_grid, mean = 2.5, sd = 1) - 0.05,
          upper = stats::dnorm(y_grid, mean = 2.5, sd = 1) + 0.05
        ),
        type = "density",
        grid = y_grid
      )
    },
    survival = {
      y_grid <- seq(0, 5, length.out = 20)
      list(
        fit = data.frame(
          id = rep(1L, 20),
          y = y_grid,
          survival = 1 - stats::pnorm(y_grid, mean = 2.5, sd = 1),
          lower = pmax(0, 1 - stats::pnorm(y_grid, mean = 2.5, sd = 1) - 0.05),
          upper = pmin(1, 1 - stats::pnorm(y_grid, mean = 2.5, sd = 1) + 0.05)
        ),
        type = "survival",
        grid = y_grid
      )
    },
    location = list(
      fit = data.frame(
        mean = 5.0, mean_lower = 4.5, mean_upper = 5.5,
        median = 4.9, median_lower = 4.4, median_upper = 5.4
      ),
      type = "location",
      grid = NULL
    )
  )
  class(out) <- "mixgpd_predict"
  out
}

.coverage_mock_causal_predict_plots <- function() {
  out <- list(
    trt_control = ggplot2::ggplot() + ggplot2::ggtitle("trt_control"),
    treatment_effect = ggplot2::ggplot() + ggplot2::ggtitle("treatment_effect")
  )
  class(out) <- c("causalmixgpd_causal_predict_plots", "list")
  out
}

.coverage_mock_mixgpd_predict_plots <- function() {
  p <- ggplot2::ggplot() + ggplot2::ggtitle("predict_plot")
  class(p) <- c("mixgpd_predict_plots", class(p))
  p
}

.coverage_mock_mixgpd_fit_plots <- function() {
  out <- list(
    traceplot = ggplot2::ggplot() + ggplot2::ggtitle("traceplot"),
    density = ggplot2::ggplot() + ggplot2::ggtitle("density")
  )
  class(out) <- c("mixgpd_fit_plots", "list")
  out
}

.coverage_mock_mixgpd_fitted <- function() {
  n <- 20L
  fit_vals <- stats::rnorm(n, mean = 5)
  y_obs <- fit_vals + stats::rnorm(n, sd = 0.5)
  out <- data.frame(
    fit = fit_vals,
    lower = fit_vals - 0.5,
    upper = fit_vals + 0.5,
    residuals = y_obs - fit_vals
  )
  class(out) <- c("mixgpd_fitted", "data.frame")
  attr(out, "object") <- list(
    data = list(y = y_obs),
    spec = list(meta = list(backend = "sb", kernel = "normal", GPD = FALSE))
  )
  attr(out, "level") <- 0.95
  attr(out, "interval") <- "credible"
  out
}

.coverage_mock_causal_fit_plots <- function() {
  treated <- .coverage_mock_mixgpd_fit_plots()
  control <- .coverage_mock_mixgpd_fit_plots()
  out <- list(treated = treated, control = control)
  class(out) <- c("causalmixgpd_causal_fit_plots", "list")
  out
}

test_that("coverage-only suite exercises internal summary and visualization helpers directly", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("ggplot2")
  if (identical(Sys.getenv("DPMIXGPD_SKIP_COVR_HELPERS_BLOCK"), "1")) {
    skip("Helper coverage block disabled for isolation.")
  }

  validate_reserved <- .coverage_ns_fun(".validate_nimble_reserved_names")
  coerce_fit_df <- .coverage_ns_fun(".coerce_fit_df")
  compute_interval <- .coverage_ns_fun(".compute_interval")
  posterior_summarize <- .coverage_ns_fun(".posterior_summarize")
  truncate_one_draw <- .coverage_ns_fun(".truncate_components_one_draw")
  wrap_plotly <- .coverage_ns_fun(".wrap_plotly")
  plot_palette <- .coverage_ns_fun(".plot_palette")
  plot_theme <- .coverage_ns_fun(".plot_theme")
  extract_nimble_code <- .coverage_ns_fun(".extract_nimble_code")
  wrap_nimble_code <- .coverage_ns_fun(".wrap_nimble_code")
  plot_quantile_pred <- .coverage_ns_fun(".plot_quantile_pred")
  plot_sample_pred <- .coverage_ns_fun(".plot_sample_pred")
  plot_mean_pred <- .coverage_ns_fun(".plot_mean_pred")
  plot_density_pred <- .coverage_ns_fun(".plot_density_pred")
  plot_survival_pred <- .coverage_ns_fun(".plot_survival_pred")
  plot_location_pred <- .coverage_ns_fun(".plot_location_pred")

  expect_invisible(validate_reserved(c("alpha", "beta")))
  expect_error(validate_reserved(c("if", "alpha")), "reserved NIMBLE keywords")

  df <- data.frame(estimate = 1:3, lower = 0:2, upper = 2:4)
  expect_true(all(c("estimate", "lower", "upper", "id") %in% names(coerce_fit_df(df))))
  expect_equal(coerce_fit_df(c(1, 2, 3))$estimate, c(1, 2, 3))

  iv_cred <- compute_interval(stats::rnorm(1000), level = 0.95, type = "credible")
  expect_named(iv_cred, c("lower", "upper"))
  expect_true(iv_cred["lower"] < iv_cred["upper"])

  if (requireNamespace("coda", quietly = TRUE)) {
    iv_hpd <- compute_interval(stats::rnorm(1000), level = 0.95, type = "hpd")
    expect_true(iv_hpd["lower"] < iv_hpd["upper"])
  }

  summ_vec <- posterior_summarize(stats::rnorm(100), interval = "credible")
  summ_mat <- posterior_summarize(matrix(stats::rnorm(300), nrow = 3), interval = NULL)
  expect_named(summ_vec, c("estimate", "lower", "upper", "q"))
  expect_true(all(is.na(summ_mat$lower)))

  trunc_out <- truncate_one_draw(c(0.1, 0.6, 0.3), list(mu = c(1, 2, 3)), epsilon = 0.01)
  expect_equal(trunc_out$ord, c(2, 3, 1))

  wrapped <- wrap_nimble_code(quote({ a <- 1 }))
  expect_true(is.list(wrapped))
  expect_equal(deparse(extract_nimble_code(wrapped)), deparse(quote({ a <- 1 })))

  expect_length(plot_palette(12), 12)
  expect_s3_class(plot_theme(), "theme")

  p <- ggplot2::ggplot() + ggplot2::geom_point(ggplot2::aes(1, 1))
  expect_true(inherits(wrap_plotly(p), "ggplot") || is.list(wrap_plotly(p)))

  pred_quant <- list(fit = data.frame(index = c(0.25, 0.5, 0.75), estimate = c(1, 2, 3), lower = c(0.5, 1.5, 2.5), upper = c(1.5, 2.5, 3.5)))
  pred_sample <- list(fit = stats::rnorm(100))
  pred_mean <- list(fit = data.frame(estimate = 5, lower = 4.5, upper = 5.5), draws = stats::rnorm(100, mean = 5))
  pred_density <- list(fit = data.frame(y = seq(0, 5, length.out = 20), density = stats::dnorm(seq(0, 5, length.out = 20), mean = 2.5)))
  pred_survival <- list(fit = data.frame(y = seq(0, 5, length.out = 20), survival = 1 - stats::pnorm(seq(0, 5, length.out = 20), mean = 2.5)))
  pred_location <- list(fit = data.frame(mean = 5, mean_lower = 4.5, mean_upper = 5.5, median = 4.9, median_lower = 4.4, median_upper = 5.4))

  expect_true(inherits(plot_quantile_pred(pred_quant), "gg"))
  expect_true(inherits(plot_sample_pred(pred_sample), "gg"))
  expect_true(inherits(plot_mean_pred(pred_mean), "gg"))
  expect_true(inherits(plot_density_pred(pred_density), "gg"))
  expect_true(inherits(plot_survival_pred(pred_survival), "gg"))
  expect_true(inherits(plot_location_pred(pred_location), "gg"))
})

test_that("coverage-only suite exercises causal bundle and parameter methods directly", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("nimble")

  print_causal_bundle <- .coverage_ns_fun("print.causalmixgpd_causal_bundle")
  summary_causal_bundle <- .coverage_ns_fun("summary.causalmixgpd_causal_bundle")
  print_params <- .coverage_ns_fun("print.mixgpd_params")
  print_params_pair <- .coverage_ns_fun("print.mixgpd_params_pair")

  causal <- .coverage_causal_fit()
  causal_bundle <- build_causal_bundle(
    y = causal$sim$y,
    X = as.matrix(causal$sim$X),
    A = causal$sim$t,
    backend = c("sb", "sb"),
    kernel = c("normal", "gamma"),
    GPD = FALSE,
    components = c(3, 3),
    PS = FALSE,
    mcmc_outcome = .coverage_mcmc(seed = 111L)
  )

  expect_output(print_causal_bundle(causal_bundle), "CausalMixGPD causal bundle")
  expect_output(summary_causal_bundle(causal_bundle), "CausalMixGPD causal bundle summary")
  expect_output(print_params(.coverage_mock_mixgpd_params()), "Posterior mean parameters")
  expect_output(print_params_pair(params(causal$fit)), "Posterior mean parameters")
  expect_output(print_params(.coverage_mock_mixgpd_params(empty = TRUE)), "empty")
})

test_that("coverage-only suite exercises mocked effect and plot methods directly", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("ggplot2")

  print_qte <- .coverage_ns_fun("print.causalmixgpd_qte")
  print_ate <- .coverage_ns_fun("print.causalmixgpd_ate")
  summary_qte <- .coverage_ns_fun("summary.causalmixgpd_qte")
  summary_ate <- .coverage_ns_fun("summary.causalmixgpd_ate")
  print_summary_qte <- .coverage_ns_fun("print.summary.causalmixgpd_qte")
  print_summary_ate <- .coverage_ns_fun("print.summary.causalmixgpd_ate")
  plot_predict <- .coverage_ns_fun("plot.mixgpd_predict")
  print_predict_plots <- .coverage_ns_fun("print.mixgpd_predict_plots")
  print_causal_predict_plots <- .coverage_ns_fun("print.causalmixgpd_causal_predict_plots")
  print_fit_plots <- .coverage_ns_fun("print.mixgpd_fit_plots")
  print_causal_fit_plots <- .coverage_ns_fun("print.causalmixgpd_causal_fit_plots")
  plot_fitted <- .coverage_ns_fun("plot.mixgpd_fitted")
  print_fitted_plots <- .coverage_ns_fun("print.mixgpd_fitted_plots")
  plot_qte <- .coverage_ns_fun("plot.causalmixgpd_qte")
  plot_ate <- .coverage_ns_fun("plot.causalmixgpd_ate")

  qte_obj <- .coverage_mock_qte()
  ate_obj <- .coverage_mock_ate()
  expect_output(print_qte(qte_obj), "QTE")
  expect_output(print_ate(ate_obj), "ATE")
  expect_s3_class(summary_qte(qte_obj), "summary.causalmixgpd_qte")
  expect_s3_class(summary_ate(ate_obj), "summary.causalmixgpd_ate")
  expect_output(print_summary_qte(summary_qte(qte_obj)), "QTE Summary")
  expect_output(print_summary_ate(summary_ate(ate_obj)), "ATE Summary")

  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("quantile")), "gg"))
  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("sample")), "gg"))
  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("mean")), "gg"))
  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("density")), "gg"))
  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("survival")), "gg"))
  expect_true(inherits(plot_predict(.coverage_mock_mixgpd_predict("location")), "gg") || is.list(plot_predict(.coverage_mock_mixgpd_predict("location"))))
  expect_silent(invisible(utils::capture.output(print_predict_plots(.coverage_mock_mixgpd_predict_plots()))))
  expect_output(print_causal_predict_plots(.coverage_mock_causal_predict_plots()), ".")
  expect_output(print_fit_plots(.coverage_mock_mixgpd_fit_plots()), "traceplot")
  expect_output(print_causal_fit_plots(.coverage_mock_causal_fit_plots()), "treated")
  expect_s3_class(plot_fitted(.coverage_mock_mixgpd_fitted()), "mixgpd_fitted_plots")
  expect_silent(invisible(utils::capture.output(print_fitted_plots(plot_fitted(.coverage_mock_mixgpd_fitted())))))

  expect_true(inherits(plot_qte(qte_obj, type = "effect"), "gg"))
  expect_true(inherits(plot_ate(ate_obj, type = "effect"), "gg"))
})

test_that("coverage-only suite exercises cluster S3 methods directly", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("nimble")

  print_cluster_bundle <- .coverage_ns_fun("print.dpmixgpd_cluster_bundle")
  summary_cluster_bundle <- .coverage_ns_fun("summary.dpmixgpd_cluster_bundle")
  plot_cluster_bundle <- .coverage_ns_fun("plot.dpmixgpd_cluster_bundle")
  print_cluster_fit <- .coverage_ns_fun("print.dpmixgpd_cluster_fit")
  summary_cluster_fit <- .coverage_ns_fun("summary.dpmixgpd_cluster_fit")
  plot_cluster_fit <- .coverage_ns_fun("plot.dpmixgpd_cluster_fit")
  print_cluster_labels <- .coverage_ns_fun("print.dpmixgpd_cluster_labels")
  summary_cluster_labels <- .coverage_ns_fun("summary.dpmixgpd_cluster_labels")
  plot_cluster_labels <- .coverage_ns_fun("plot.dpmixgpd_cluster_labels")
  print_cluster_psm <- .coverage_ns_fun("print.dpmixgpd_cluster_psm")
  summary_cluster_psm <- .coverage_ns_fun("summary.dpmixgpd_cluster_psm")
  plot_cluster_psm <- .coverage_ns_fun("plot.dpmixgpd_cluster_psm")

  cluster <- .coverage_cluster_fit()
  lbl <- predict(cluster$fit, type = "label", return_scores = TRUE)
  psm <- predict(cluster$fit, type = "psm")

  expect_output(print_cluster_bundle(cluster$fit$bundle), "Cluster bundle")
  expect_s3_class(summary_cluster_bundle(cluster$fit$bundle), "summary.dpmixgpd_cluster_bundle")
  expect_silent(plot_cluster_bundle(cluster$fit$bundle))
  expect_output(print_cluster_fit(cluster$fit), "Cluster fit")
  expect_s3_class(summary_cluster_fit(cluster$fit), "summary.dpmixgpd_cluster_fit")
  expect_silent(plot_cluster_fit(cluster$fit, which = "sizes"))
  expect_silent(plot_cluster_fit(cluster$fit, which = "summary"))
  expect_output(print_cluster_labels(lbl), "Cluster labels")
  expect_s3_class(summary_cluster_labels(lbl), "summary.dpmixgpd_cluster_labels")
  expect_silent(plot_cluster_labels(lbl, type = "certainty"))
  expect_silent(plot_cluster_labels(lbl, type = "summary"))
  expect_output(print_cluster_psm(psm), "Cluster PSM")
  expect_s3_class(summary_cluster_psm(psm), "summary.dpmixgpd_cluster_psm")
  expect_silent(plot_cluster_psm(psm, psm_max_n = nrow(psm$psm)))
})

test_that("coverage-only suite exercises registries, simulations, and distribution wrappers", {
  skip_if_not_test_level("ci")

  init_kernel_registry()
  reg <- get_kernel_registry()
  tail_reg <- get_tail_registry()
  support <- kernel_support_table(round = FALSE)

  expect_true(is.list(reg))
  expect_true(length(reg) >= 7L)
  expect_true(is.list(tail_reg))
  expect_true(is.data.frame(support))

  .coverage_numeric(dgpd(c(1.2, 1.8), threshold = 1, scale = 0.8, shape = 0.2), 2L)
  .coverage_numeric(pgpd(c(1.2, 1.8), threshold = 1, scale = 0.8, shape = 0.2), 2L)
  .coverage_numeric(qgpd(c(0.25, 0.75), threshold = 1, scale = 0.8, shape = 0.2), 2L)
  .coverage_numeric(rgpd(4L, threshold = 1, scale = 0.8, shape = 0.2), 4L)

  .coverage_numeric(dinvgauss(c(0.8, 1.4), mean = 1.5, shape = 5), 2L)
  .coverage_numeric(pinvgauss(c(0.8, 1.4), mean = 1.5, shape = 5), 2L)
  .coverage_numeric(qinvgauss(c(0.3, 0.7), mean = 1.5, shape = 5), 2L)
  .coverage_numeric(rinvgauss(4L, mean = 1.5, shape = 5), 4L)

  .coverage_numeric(damoroso(c(0.5, 1.5), loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), 2L)
  .coverage_numeric(pamoroso(c(0.5, 1.5), loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), 2L)
  .coverage_numeric(qamoroso(c(0.25, 0.75), loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), 2L)
  .coverage_numeric(ramoroso(4L, loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), 4L)

  .coverage_numeric(dcauchy_vec(c(-1, 0, 1), location = 0, scale = 1.2), 3L)
  .coverage_numeric(pcauchy_vec(c(-1, 0, 1), location = 0, scale = 1.2), 3L)
  .coverage_numeric(qcauchy_vec(c(0.25, 0.75), location = 0, scale = 1.2), 2L)
  .coverage_numeric(rcauchy_vec(4L, location = 0, scale = 1.2), 4L)

  for (kernel in names(reg)) {
    .coverage_run_kernel_wrappers(kernel)
  }

  sim_bulk <- sim_bulk_tail(n = 30, seed = 1)
  sim_causal <- sim_causal_qte(n = 20, seed = 2)
  sim_surv <- sim_survival_tail(n = 20, seed = 3)

  expect_equal(length(sim_bulk), 30L)
  expect_named(sim_causal, c("y", "t", "X", "A"))
  expect_true(is.data.frame(sim_surv))
  expect_true(all(c("time", "status", "x1", "x2") %in% names(sim_surv)))
})

test_that("coverage-only suite exercises non-causal build, fit, predict, and glue paths", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  uncond <- .coverage_unconditional_fit()
  fit_u <- uncond$fit
  cond <- .coverage_conditional_fit()
  fit_c <- cond$fit
  y <- cond$y
  X <- cond$X

  expect_s3_class(bundle(y = y, X = X, backend = "sb", kernel = "normal", components = 3, GPD = FALSE), "causalmixgpd_bundle")
  expect_s3_class(.coverage_spliced_bundle(), "causalmixgpd_bundle")

  expect_output(print(fit_u))
  expect_s3_class(summary(fit_u), "mixgpd_summary")
  expect_s3_class(params(fit_u), "mixgpd_params")

  pred_u_q <- predict(fit_u, type = "quantile", index = c(0.25, 0.75))
  pred_u_d <- predict(fit_u, y = uncond$y[1:5], type = "density")
  pred_u_s <- predict(fit_u, y = uncond$y[1:5], type = "survival")

  expect_s3_class(pred_u_q, "mixgpd_predict")
  expect_s3_class(pred_u_d, "mixgpd_predict")
  expect_s3_class(pred_u_s, "mixgpd_predict")

  expect_output(print(fit_c))
  expect_s3_class(summary(fit_c), "mixgpd_summary")
  expect_s3_class(params(fit_c), "mixgpd_params")

  pred_c_mean <- predict(fit_c, newdata =X[1:4, , drop = FALSE], type = "mean", nsim_mean = 20L)
  pred_c_q <- predict(fit_c, newdata =X[1:4, , drop = FALSE], type = "quantile", index = c(0.25, 0.75))
  pred_c_d <- predict(fit_c, newdata =X[1:2, , drop = FALSE], y = y[1:2], type = "density")
  pred_c_s <- predict(fit_c, newdata =X[1:2, , drop = FALSE], y = y[1:2], type = "survival")

  expect_s3_class(pred_c_mean, "mixgpd_predict")
  expect_s3_class(pred_c_q, "mixgpd_predict")
  expect_s3_class(pred_c_d, "mixgpd_predict")
  expect_s3_class(pred_c_s, "mixgpd_predict")
  expect_true(is.numeric(residuals(fit_c)))
  expect_true(is.data.frame(fitted(fit_c)))

  glue_check <- check_glue_validity(
    fit_c,
    grid = seq(min(y), stats::quantile(y, 0.9), length.out = 16L),
    n_draws = 5L,
    check_continuity = FALSE
  )
  expect_true(is.list(glue_check))
  expect_true(all(c("pass", "violations", "n_checked_draws") %in% names(glue_check)))

  if (requireNamespace("coda", quietly = TRUE)) {
    ess <- tryCatch(ess_summary(fit_c), error = function(e) NULL)
    if (!is.null(ess)) {
      expect_s3_class(ess, "mixgpd_ess_summary")
    }
  }
})

test_that("coverage-only suite exercises direct bundle and mcmc wrappers", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  dat <- .coverage_cached("bundle_dat", {
    set.seed(67)
    data.frame(
      y = abs(stats::rnorm(18)) + 0.25,
      x1 = stats::rnorm(18),
      x2 = stats::runif(18)
    )
  })

  b_formula <- bundle(
    formula = y ~ x1 + x2,
    data = dat,
    backend = "sb",
    kernel = "normal",
    components = 3,
    GPD = FALSE
  )
  expect_s3_class(b_formula, "causalmixgpd_bundle")

  fit_formula <- .coverage_quiet(mcmc(b_formula, niter = 20L, nburnin = 5L, thin = 1L, nchains = 1L, seed = 67L, show_progress = FALSE))
  expect_s3_class(fit_formula, "mixgpd_fit")

  b_causal <- bundle(
    formula = y ~ x1 + x2,
    data = transform(dat, A = rep(c(0L, 1L), length.out = nrow(dat))),
    treat = "A",
    backend = c("sb", "sb"),
    kernel = c("normal", "gamma"),
    components = c(3, 3),
    GPD = FALSE
  )
  expect_s3_class(b_causal, "causalmixgpd_causal_bundle")
})

test_that("coverage-only suite exercises causal workflows and treatment-effect methods", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  obj <- .coverage_causal_fit()
  fit <- obj$fit
  X <- as.matrix(obj$sim$X)

  expect_output(print(fit))
  expect_output(print(summary(fit)), "Outcome fits")
  expect_s3_class(params(fit), "mixgpd_params_pair")

  pred_mean <- predict(fit, newdata =X[1:3, , drop = FALSE], type = "mean", nsim_mean = 20L)
  pred_quant <- predict(fit, newdata =X[1:3, , drop = FALSE], type = "quantile", p = c(0.25, 0.75))
  pred_density <- predict(fit, newdata =X[1:2, , drop = FALSE], y = obj$sim$y[1:2], type = "density")
  pred_survival <- predict(fit, newdata =X[1:2, , drop = FALSE], y = obj$sim$y[1:2], type = "survival")

  expect_s3_class(pred_mean, "causalmixgpd_causal_predict")
  expect_s3_class(pred_quant, "causalmixgpd_causal_predict")
  expect_s3_class(pred_density, "causalmixgpd_causal_predict")
  expect_s3_class(pred_survival, "causalmixgpd_causal_predict")

  cate_res <- cate(fit, newdata = X[1:3, , drop = FALSE], nsim_mean = 20L, interval = "credible")
  cqte_res <- cqte(fit, probs = c(0.25, 0.75), newdata = X[1:3, , drop = FALSE], interval = "credible")
  ate_res <- ate(fit, nsim_mean = 20L, interval = "credible")
  att_res <- att(fit, nsim_mean = 20L, interval = "credible")
  qte_res <- qte(fit, probs = c(0.25, 0.75), interval = "credible")
  qtt_res <- qtt(fit, probs = c(0.25, 0.75), interval = "credible")
  rmean_res <- ate_rmean(fit, newdata = X[1:3, , drop = FALSE], cutoff = 10, nsim_mean = 20L, interval = "credible")

  expect_s3_class(cate_res, "causalmixgpd_ate")
  expect_s3_class(cqte_res, "causalmixgpd_qte")
  expect_s3_class(ate_res, "causalmixgpd_ate")
  expect_s3_class(att_res, "causalmixgpd_ate")
  expect_s3_class(qte_res, "causalmixgpd_qte")
  expect_s3_class(qtt_res, "causalmixgpd_qte")
  expect_s3_class(rmean_res, "causalmixgpd_ate")

  expect_silent(summary(ate_res))
  expect_silent(summary(qte_res))
  expect_type(plot(ate_res), "list")
  expect_type(plot(qte_res), "list")
  expect_type(plot(pred_mean), "list")
  expect_type(plot(fit, arm = "both"), "list")
})

test_that("coverage-only suite exercises causal GPD wrapper path", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  fit <- .coverage_cached("causal_gpd_fit", {
    sim <- sim_causal_qte(n = 20, seed = 71)
    sim$y <- abs(sim$y) + 0.2

    dpmgpd.causal(
      y = sim$y,
      X = as.matrix(sim$X[, 1:2, drop = FALSE]),
      treat = sim$t,
      backend = c("sb", "sb"),
      kernel = c("gamma", "gamma"),
      components = c(3, 3),
      PS = FALSE,
      mcmc = c(.coverage_mcmc(seed = 71L), list(show_progress = FALSE))
    )
  })

  expect_s3_class(fit, "causalmixgpd_causal_fit")
})

test_that("coverage-only suite exercises cluster builders, predictors, and S3 methods", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  obj <- .coverage_cluster_fit()
  fit <- obj$fit
  dat <- obj$data

  psm <- predict(fit, type = "psm")
  lbl_train <- predict(fit, type = "label", return_scores = TRUE)
  lbl_new <- predict(fit, newdata = dat[1:4, , drop = FALSE], type = "label", return_scores = TRUE)

  expect_s3_class(fit, "dpmixgpd_cluster_fit")
  expect_s3_class(psm, "dpmixgpd_cluster_psm")
  expect_s3_class(lbl_train, "dpmixgpd_cluster_labels")
  expect_s3_class(lbl_new, "dpmixgpd_cluster_labels")

  expect_output(print(fit))
  expect_silent(summary(fit))
  expect_silent(plot(fit, which = "psm"))
  expect_silent(plot(fit, which = "k"))
  expect_silent(plot(fit, which = "sizes"))

  expect_output(print(lbl_train))
  expect_silent(summary(lbl_train))
  expect_silent(plot(lbl_train, type = "sizes"))
  expect_silent(plot(lbl_train, type = "certainty"))

  expect_output(print(psm))
  expect_silent(summary(psm))
  expect_silent(plot(psm, psm_max_n = nrow(psm$psm)))
})

#' Inverse Gaussian mixture distribution
#'
#' Finite mixture of inverse Gaussian components for positive-support bulk modeling. Each component
#' is parameterized by `mean[j]` and `shape[j]`.
#'
#' The scalar functions in this topic are the compiled building blocks for inverse-Gaussian bulk
#' kernels. For vectorized R usage, use [invgauss_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param mean,shape Numeric vectors of length \eqn{K} giving component means and shapes.
#' @param log Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qInvGaussMix()` returns a numeric
#'   vector with the same length as `p`.
#'
#' @details
#' The mixture distribution is
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k F_{IG}(x \mid \mu_k,\lambda_k),
#' }
#' where each inverse Gaussian component has mean \eqn{\mu_k} and variance \eqn{\mu_k^3/\lambda_k}.
#' Random generation selects a component using the normalized weights and then generates from the
#' corresponding inverse Gaussian law. Quantiles are computed numerically because the finite-mixture
#' inverse CDF is not available in closed form.
#'
#' The analytical mixture mean is
#' \deqn{
#' E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k.
#' }
#' That expression is used by the package whenever inverse-Gaussian mixtures contribute to
#' posterior predictive means.
#'
#' @seealso [InvGauss_mixgpd()], [InvGauss_gpd()], [invgauss_lowercase()],
#'   [build_nimble_bundle()], [kernel_support_table()].
#' @family inverse-gaussian kernel families
#'
#' @examples
#' w <- c(0.55, 0.30, 0.15)
#' mean <- c(1.0, 2.5, 5.0)
#' shape <- c(2, 4, 8)
#'
#' dInvGaussMix(2.0, w = w, mean = mean, shape = shape, log = 0)
#' pInvGaussMix(2.0, w = w, mean = mean, shape = shape,
#'             lower.tail = 1, log.p = 0)
#' qInvGaussMix(0.50, w = w, mean = mean, shape = shape)
#' qInvGaussMix(0.95, w = w, mean = mean, shape = shape)
#' replicate(10, rInvGaussMix(1, w = w, mean = mean, shape = shape))
#' @rdname InvGauss_mix
#' @name InvGauss_mix
#' @aliases dInvGaussMix pInvGaussMix rInvGaussMix qInvGaussMix
#' @importFrom stats uniroot
NULL

#' @describeIn InvGauss_mix Inverse Gaussian mixture density
#' @export
dInvGaussMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    K <- length(w)

    # wsum = sum(w)
    wsum <- 0.0
    for (j in 1:K) {
      wsum <- wsum + w[j]
    }

    if (wsum <= 0.0) {
      if (log == 1L) return(log(eps)) else return(0.0)
    }

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + (w[j] / wsum) * dInvGauss(x, mean[j], shape[j], 0L)
    }

    if (s0 < eps) s0 <- eps
    if (log == 1L) return(log(s0)) else return(s0)
  }
)


#' @describeIn InvGauss_mix Inverse Gaussian mixture distribution function
#' @export
pInvGaussMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    eps <- 1e-300
    K <- length(w)

    # wsum = sum(w)
    wsum <- 0.0
    for (j in 1:K) {
      wsum <- wsum + w[j]
    }

    if (wsum <= 0.0) {
      if (log.p != 0L) return(log(eps)) else return(eps)
    }

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + (w[j] / wsum) * pInvGauss(q, mean[j], shape[j], 1L, 0L)
    }

    # clamp to [0,1]
    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0L) cdf <- 1.0 - cdf

    if (log.p != 0L) {
      if (cdf < eps) cdf <- eps
      return(log(cdf))
    }

    return(cdf)
  }
)


#' @describeIn InvGauss_mix Inverse Gaussian mixture random generation
#' @export
rInvGaussMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1)) {
    returnType(double(0))

    if (n != 1L) return(0.0)

    K <- length(w)

    # wsum = sum(w)
    wsum <- 0.0
    for (j in 1:K) {
      wsum <- wsum + w[j]
    }
    if (wsum <= 0.0) return(0.0)

    u <- runif(1, 0.0, wsum)

    cw <- 0.0
    idx <- 1
    found <- 0L

    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0L) {
        if (u <= cw) {
          idx <- j
          found <- 1L
        }
      }
    }

    return(rInvGauss(1L, mean[idx], shape[idx]))
  }
)


#' @describeIn InvGauss_mix Inverse Gaussian mixture quantile function
#' @export
qInvGaussMix <- function(p, w, mean, shape,
                         lower.tail = TRUE, log.p = FALSE,
                         tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)
  w <- as.numeric(w)
  wsum <- sum(w)
  if (!is.finite(wsum) || wsum <= 0) return(rep(0, length(p)))
  w <- w / wsum
  mean_mix <- sum(w * mean)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- 0; next }
    if (pi >= 1) { out[i] <- Inf; next }
    p0 <- as.numeric(pInvGaussMix(0, w, mean, shape, 1, 0))
    if (!is.finite(p0)) p0 <- 0
    if (p0 >= pi) { out[i] <- 0; next }

    hi <- max(1, mean_mix * 10)
    phi <- as.numeric(pInvGaussMix(hi, w, mean, shape, 1, 0))
    iter <- 0L
    while (is.finite(phi) && phi < pi && hi < 1e20 && iter < 60L) {
      hi <- hi * 2
      phi <- as.numeric(pInvGaussMix(hi, w, mean, shape, 1, 0))
      iter <- iter + 1L
    }

    if (!is.finite(phi) || phi < pi) { out[i] <- Inf; next }

    f0 <- as.numeric(pInvGaussMix(0, w, mean, shape, 1, 0) - pi)
    fhi <- as.numeric(pInvGaussMix(hi, w, mean, shape, 1, 0) - pi)
    if (!is.finite(fhi) || f0 * fhi > 0) {
      out[i] <- Inf
    } else {
      out[i] <- stats::uniroot(function(q) as.numeric(pInvGaussMix(q, w, mean, shape, 1, 0)) - pi,
                               interval = c(0, hi),
                               tol = tol, maxiter = maxiter)$root
    }
  }
  out
}

meanInvGaussMix <- function(w, mean, shape) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(mean)
  lam <- as.numeric(shape)
  if (length(mu) != length(ww) || length(lam) != length(ww)) return(NA_real_)
  if (any(!is.finite(mu)) || any(!is.finite(lam)) || any(mu <= 0) || any(lam <= 0)) return(NA_real_)
  sum(ww * mu)
}

meanInvGaussMixTrunc <- function(w, mean, shape, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(mean)
  lam <- as.numeric(shape)
  u <- as.numeric(threshold)[1]
  if (length(mu) != length(ww) || length(lam) != length(ww)) return(NA_real_)
  if (any(!is.finite(mu)) || any(!is.finite(lam)) || any(mu <= 0) || any(lam <= 0) || is.na(u)) return(NA_real_)
  if (is.infinite(u) && u > 0) return(meanInvGaussMix(w = ww, mean = mu, shape = lam))
  if (u <= 0) return(0)
  sqrt_term <- sqrt(lam / u)
  a <- sqrt_term * (u / mu - 1)
  b <- -sqrt_term * (u / mu + 1)
  comp <- mu * (stats::pnorm(a) - exp(2 * lam / mu) * stats::pnorm(b))
  sum(ww * comp)
}


#' Inverse Gaussian mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to an inverse Gaussian
#' mixture bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}.
#' @param mean,shape Numeric vectors of length \eqn{K} giving component means and shapes.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars.
#'   `qInvGaussMixGpd()` returns a numeric vector with the same length as `p`.
#'
#' @details
#' This family keeps the inverse-Gaussian mixture body below the threshold \eqn{u} and attaches a
#' generalized Pareto exceedance law to the residual survival probability above \eqn{u}. If
#' \eqn{F_{mix}(u)=p_u}, then the tail density is \eqn{(1-p_u)g_{GPD}(x \mid u,\sigma_u,\xi)}.
#'
#' Quantile evaluation is piecewise. For probabilities at or below \eqn{p_u}, the function solves
#' the mixture inverse numerically; above \eqn{p_u}, it rescales the upper-tail probability and
#' applies the GPD inverse directly.
#'
#' @seealso [InvGauss_mix()], [InvGauss_gpd()], [gpd()], [invgauss_lowercase()], [dpmgpd()].
#' @family inverse-gaussian kernel families
#'
#' @examples
#' w <- c(0.55, 0.30, 0.15)
#' mean <- c(1.0, 2.5, 5.0)
#' shape <- c(2, 4, 8)
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dInvGaussMixGpd(4.0, w = w, mean = mean, shape = shape,
#'                threshold = threshold, tail_scale = tail_scale,
#'                tail_shape = tail_shape, log = 0)
#' pInvGaussMixGpd(4.0, w = w, mean = mean, shape = shape,
#'                threshold = threshold, tail_scale = tail_scale,
#'                tail_shape = tail_shape, lower.tail = 1, log.p = 0)
#' qInvGaussMixGpd(0.50, w = w, mean = mean, shape = shape,
#'                threshold = threshold, tail_scale = tail_scale,
#'                tail_shape = tail_shape)
#' qInvGaussMixGpd(0.95, w = w, mean = mean, shape = shape,
#'                threshold = threshold, tail_scale = tail_scale,
#'                tail_shape = tail_shape)
#' replicate(10, rInvGaussMixGpd(1, w = w, mean = mean, shape = shape,
#'                              threshold = threshold,
#'                              tail_scale = tail_scale,
#'                              tail_shape = tail_shape))
#' @rdname InvGauss_mixgpd
#' @name InvGauss_mixgpd
#' @aliases dInvGaussMixGpd pInvGaussMixGpd rInvGaussMixGpd qInvGaussMixGpd
NULL


#' @describeIn InvGauss_mixgpd Inverse Gaussian mixture + GPD tail density
#' @export
dInvGaussMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) return(dInvGaussMix(x, w, mean, shape, log))

    Fu <- pInvGaussMix(threshold, w, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)


#' @describeIn InvGauss_mixgpd Inverse Gaussian mixture + GPD tail distribution function
#' @export
pInvGaussMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) return(pInvGaussMix(q, w, mean, shape, lower.tail, log.p))

    Fu <- pInvGaussMix(threshold, w, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G
    cdf <- max(min(cdf, 1.0), 0.0)

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)


#' @describeIn InvGauss_mixgpd Inverse Gaussian mixture + GPD tail random generation
#' @export
rInvGaussMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 mean = double(1),
                 shape = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pInvGaussMix(threshold, w, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rInvGaussMix(1, w, mean, shape))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)


#' @describeIn InvGauss_mixgpd Inverse Gaussian mixture + GPD tail quantile function
#' @export
qInvGaussMixGpd <- function(p, w, mean, shape, threshold, tail_scale, tail_shape,
                            lower.tail = TRUE, log.p = FALSE,
                            tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)
  w <- as.numeric(w); w <- w / sum(w)

  Fu <- as.numeric(pInvGaussMix(threshold, w, mean, shape, 1, 0))
  Fu <- max(min(Fu, 1.0), 0.0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qInvGaussMix(pi, w, mean, shape,
                             lower.tail = TRUE, log.p = FALSE,
                             tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}



#' Inverse Gaussian with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' inverse Gaussian bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. For portability inside NIMBLE,
#'   the RNG implementation supports \code{n = 1}.
#' @param mean Numeric scalar mean parameter \eqn{\mu>0}.
#' @param shape Numeric scalar shape parameter \eqn{\lambda>0}.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars.
#'   `qInvGaussGpd()` returns a numeric vector with the same length as `p`.
#'
#' @details
#' This is the one-component version of [InvGauss_mixgpd()]. The inverse Gaussian governs the bulk
#' region and the generalized Pareto governs exceedances over the threshold. The splice is
#' continuous at \eqn{u} because the GPD is scaled by the inverse-Gaussian survival probability at
#' the threshold.
#'
#' The ordinary mean of the spliced law exists only when the GPD tail has \eqn{\xi < 1}. When that
#' condition fails, the package uses restricted means or quantile-based summaries instead of an
#' ordinary mean.
#'
#' @seealso [InvGauss_mix()], [InvGauss_mixgpd()], [gpd()], [invgauss_lowercase()].
#' @family inverse-gaussian kernel families
#'
#' @examples
#' mean <- 2.5
#' shape <- 6
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dInvGaussGpd(4.0, mean = mean, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape, log = 0)
#' pInvGaussGpd(4.0, mean = mean, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape, lower.tail = 1, log.p = 0)
#' qInvGaussGpd(0.50, mean = mean, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape)
#' qInvGaussGpd(0.95, mean = mean, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape)
#' replicate(10, rInvGaussGpd(1, mean = mean, shape = shape,
#'                           threshold = threshold,
#'                           tail_scale = tail_scale,
#'                           tail_shape = tail_shape))
#' @rdname InvGauss_gpd
#' @name InvGauss_gpd
#' @aliases dInvGaussGpd pInvGaussGpd rInvGaussGpd qInvGaussGpd
NULL


#' @describeIn InvGauss_gpd Inverse Gaussian + GPD tail density
#' @export
dInvGaussGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 mean = double(0),
                 shape = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) return(dInvGauss(x, mean, shape, log))

    Fu <- pInvGauss(threshold, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)


#' @describeIn InvGauss_gpd Inverse Gaussian + GPD tail distribution function
#' @export
pInvGaussGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 mean = double(0),
                 shape = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) return(pInvGauss(q, mean, shape, lower.tail, log.p))

    Fu <- pInvGauss(threshold, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G
    cdf <- max(min(cdf, 1.0), 0.0)

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p != 0) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)


#' @describeIn InvGauss_gpd Inverse Gaussian + GPD tail random generation
#' @export
rInvGaussGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 mean = double(0),
                 shape = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pInvGauss(threshold, mean, shape, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rInvGauss(1, mean, shape))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)


#' @describeIn InvGauss_gpd Inverse Gaussian + GPD tail quantile function
#' @export
#' @param tol Numeric tolerance for numerical inversion in \code{qInvGaussGpd}.
#' @param maxiter Maximum iterations for numerical inversion in \code{qInvGaussGpd}.
qInvGaussGpd <- function(p, mean, shape, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE,
                         tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- as.numeric(pInvGauss(threshold, mean, shape, 1, 0))
  Fu <- max(min(Fu, 1.0), 0.0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qInvGauss(pi, mean = mean, shape = shape,
                          lower.tail = TRUE, log.p = FALSE,
                          tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


# ==========================================================
# Lowercase vectorized R wrappers for Inverse Gaussian kernels
# ==========================================================

#' Lowercase vectorized inverse Gaussian distribution functions
#'
#' Vectorized R wrappers for the scalar inverse-Gaussian-kernel topics in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param mean,shape Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These functions are vectorized R front ends to the scalar inverse-Gaussian
#' and splice routines. They retain the \eqn{(\mu,\lambda)} parameterization
#' used everywhere else in the package and apply the scalar evaluator
#' repeatedly over the supplied input vector or draw index.
#'
#' @seealso [InvGauss_mix()], [InvGauss_mixgpd()], [InvGauss_gpd()], [bundle()],
#'   [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.3, 0.1)
#' mu <- c(1, 1.5, 2)
#' lam <- c(2, 3, 4)
#'
#' # Inverse Gaussian mixture
#' dinvgaussmix(c(1, 2, 3), w = w, mean = mu, shape = lam)
#' rinvgaussmix(5, w = w, mean = mu, shape = lam)
#'
#' @name invgauss_lowercase
#' @rdname invgauss_lowercase
NULL

# ---- InvGauss Mix lowercase wrappers ----

#' @describeIn invgauss_lowercase Inverse Gaussian mixture density (vectorized)
#' @export
dinvgaussmix <- function(x, w, mean, shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dInvGaussMix(xi, w = w, mean = mean, shape = shape, log = log_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture distribution function (vectorized)
#' @export
pinvgaussmix <- function(q, w, mean, shape, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pInvGaussMix(qi, w = w, mean = mean, shape = shape,
                                                  lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture quantile function (vectorized)
#' @export
qinvgaussmix <- function(p, w, mean, shape, lower.tail = TRUE, log.p = FALSE,
                         tol = 1e-10, maxiter = 200) {
  qInvGaussMix(p, w = w, mean = mean, shape = shape, lower.tail = lower.tail,
               log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture random generation (vectorized)
#' @export
rinvgaussmix <- function(n, w, mean, shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rInvGaussMix(1L, w = w, mean = mean, shape = shape)),
         numeric(1L))
}

# ---- InvGauss Mix + GPD lowercase wrappers ----

#' @describeIn invgauss_lowercase Inverse Gaussian mixture + GPD density (vectorized)
#' @export
dinvgaussmixgpd <- function(x, w, mean, shape, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dInvGaussMixGpd(xi, w = w, mean = mean, shape = shape,
                                                     threshold = threshold, tail_scale = tail_scale,
                                                     tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture + GPD distribution function (vectorized)
#' @export
pinvgaussmixgpd <- function(q, w, mean, shape, threshold, tail_scale, tail_shape,
                            lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pInvGaussMixGpd(qi, w = w, mean = mean, shape = shape,
                                                     threshold = threshold, tail_scale = tail_scale,
                                                     tail_shape = tail_shape,
                                                     lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture + GPD quantile function (vectorized)
#' @export
qinvgaussmixgpd <- function(p, w, mean, shape, threshold, tail_scale, tail_shape,
                            lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qInvGaussMixGpd(p, w = w, mean = mean, shape = shape, threshold = threshold,
                  tail_scale = tail_scale, tail_shape = tail_shape,
                  lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn invgauss_lowercase Inverse Gaussian mixture + GPD random generation (vectorized)
#' @export
rinvgaussmixgpd <- function(n, w, mean, shape, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rInvGaussMixGpd(1L, w = w, mean = mean, shape = shape,
                                                            threshold = threshold, tail_scale = tail_scale,
                                                            tail_shape = tail_shape)),
         numeric(1L))
}

# ---- InvGauss + GPD lowercase wrappers ----

#' @describeIn invgauss_lowercase Inverse Gaussian + GPD density (vectorized)
#' @export
dinvgaussgpd <- function(x, mean, shape, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dInvGaussGpd(xi, mean = mean, shape = shape,
                                                  threshold = threshold, tail_scale = tail_scale,
                                                  tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian + GPD distribution function (vectorized)
#' @export
pinvgaussgpd <- function(q, mean, shape, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pInvGaussGpd(qi, mean = mean, shape = shape,
                                                  threshold = threshold, tail_scale = tail_scale,
                                                  tail_shape = tail_shape,
                                                  lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn invgauss_lowercase Inverse Gaussian + GPD quantile function (vectorized)
#' @export
qinvgaussgpd <- function(p, mean, shape, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qInvGaussGpd(p, mean = mean, shape = shape, threshold = threshold,
               tail_scale = tail_scale, tail_shape = tail_shape,
               lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn invgauss_lowercase Inverse Gaussian + GPD random generation (vectorized)
#' @export
rinvgaussgpd <- function(n, mean, shape, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rInvGaussGpd(1L, mean = mean, shape = shape,
                                                          threshold = threshold, tail_scale = tail_scale,
                                                          tail_shape = tail_shape)),
         numeric(1L))
}

#+ filtered ###################################################################
#' Global contracts for CausalMixGPD
#'
#' These lists and helpers capture the modeling rules used throughout the
#' package: backends, kernels, GPD usage, and mixture sizes.
#' They are intentionally loaded before the rest of the core code so downstream
#' builders can reuse the same constants.
#'
#' Backend summary:
#' - "sb": Stick-breaking (truncated Dirichlet process) with mixture evaluation
#' - "crp": Chinese Restaurant Process with component-indexed parameters
#' - "spliced": CRP variant enforcing component-level GPD parameterization with
#'              flexible modes (fixed/dist/link) for threshold, tail_scale, tail_shape.
#'              Uses CRP signatures but allows link mode for tail params.
#'
#' @keywords internal
#' @noRd
allowed_backends <- c("crp", "sb", "spliced")
allowed_kernels <- c("gamma", "lognormal", "invgauss", "normal", "laplace", "cauchy", "amoroso")
positive_support_kernels <- c("gamma", "lognormal", "invgauss", "amoroso")
real_support_kernels <- c("normal", "laplace", "cauchy")

is_allowed_kernel <- function(x) {
  as.character(x) %in% allowed_kernels
}

check_gpd_contract <- function(GPD, kernel) {
  if (isTRUE(GPD) && identical(as.character(kernel), "cauchy")) {
    stop("Cauchy kernels are never paired with GPD tails.", call. = FALSE)
  }
  invisible(NULL)
}




#' Initialize kernel registries
#'
#' Creates/refreshes registries used by the model specification compiler and
#' code generators. Each kernel entry stores bulk parameters, supports, default
#' regression/link behavior, and distribution signatures for SB/CRP backends.
#'
#' @details
#' The kernel registry is the package-level contract that keeps model building,
#' prediction, and documentation aligned. Each entry records the natural bulk
#' parameters for one kernel, the support constraints they must satisfy, the
#' default covariate-link strategy, and the backend-specific distribution names
#' used when generating NIMBLE code.
#'
#' The companion tail registry records the generalized Pareto tail parameters
#' \eqn{u} (threshold), \eqn{\sigma_u} (tail scale), and \eqn{\xi_u} (tail
#' shape) together with their support and allowed modeling modes. Calling
#' `init_kernel_registry()` makes those contracts available in the package
#' namespace so later builders can validate requests without duplicating lookup
#' logic.
#'
#' @return Invisibly returns TRUE.
#' @examples
#' init_kernel_registry()
#' reg <- get_kernel_registry()
#' names(reg)
#' tail <- get_tail_registry()
#' tail$params
#' @export
init_kernel_registry <- function() {
  ns <- getNamespace("CausalMixGPD")

  has_ns <- exists("kernel_registry", envir = ns, inherits = FALSE) &&
    exists("tail_registry", envir = ns, inherits = FALSE)

  if (has_ns) {
    return(invisible(TRUE))
  }

  kernel_registry <- list(

    normal = list(
      key = "normal",
      bulk_params = c("mean", "sd"),
      bulk_support = c(mean = "real", sd = "positive_sd"),
      param_types = c(mean = "location", sd = "sd"),
      allow_gpd = TRUE,
      defaults_X = list(mean = list(mode = "link", link = "identity"),
                        sd   = list(mode = "dist")),
      sb = list(
        d = "dNormMix",
        d_gpd = "dNormMixGpd",
        args = c("w", "mean", "sd"),
        mean = "meanNormMix",
        mean_trunc = "meanNormMixTrunc",
        args_gpd = c("w", "mean", "sd", "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "dnorm",
        mean_base = "meanNormMix",
        mean_trunc_base = "meanNormMixTrunc",
        d_gpd = "dNormGpd",
        args_gpd = c("mean", "sd", "threshold", "tail_scale", "tail_shape")
      )
    ),

    lognormal = list(
      key = "lognormal",
      bulk_params = c("meanlog", "sdlog"),
      bulk_support = c(meanlog = "real", sdlog = "positive_sd"),
      param_types = c(meanlog = "location", sdlog = "sd"),
      allow_gpd = TRUE,
      defaults_X = list(meanlog = list(mode = "link", link = "identity"),
                        sdlog   = list(mode = "dist")),
      sb = list(
        d = "dLognormalMix",
        d_gpd = "dLognormalMixGpd",
        args = c("w", "meanlog", "sdlog"),
        mean = "meanLognormalMix",
        mean_trunc = "meanLognormalMixTrunc",
        args_gpd = c("w", "meanlog", "sdlog", "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "dlnorm",
        mean_base = "meanLognormalMix",
        mean_trunc_base = "meanLognormalMixTrunc",
        d_gpd = "dLognormalGpd",
        args_gpd = c("meanlog", "sdlog", "threshold", "tail_scale", "tail_shape")
      )
    ),

    invgauss = list(
      key = "invgauss",
      bulk_params = c("mean", "shape"),
      bulk_support = c(mean = "positive_location", shape = "positive_shape"),
      param_types = c(mean = "location", shape = "shape"),
      allow_gpd = TRUE,
      defaults_X = list(mean  = list(mode = "link", link = "exp"),
                        shape = list(mode = "dist")),
      sb = list(
        d = "dInvGaussMix",
        d_gpd = "dInvGaussMixGpd",
        args = c("w", "mean", "shape"),
        mean = "meanInvGaussMix",
        mean_trunc = "meanInvGaussMixTrunc",
        args_gpd = c("w", "mean", "shape", "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "dInvGauss",
        mean_base = "meanInvGaussMix",
        mean_trunc_base = "meanInvGaussMixTrunc",
        d_gpd = "dInvGaussGpd",
        args_gpd = c("mean", "shape", "threshold", "tail_scale", "tail_shape")
      )
    ),

    gamma = list(
      key = "gamma",
      bulk_params = c("shape", "scale"),
      bulk_support = c(shape = "positive_shape", scale = "positive_scale"),
      param_types = c(shape = "shape", scale = "scale"),
      allow_gpd = TRUE,
      defaults_X = list(shape = list(mode = "dist"),
                        scale = list(mode = "link", link = "exp")),
      sb = list(
        d = "dGammaMix",
        d_gpd = "dGammaMixGpd",
        args = c("w", "shape", "scale"),
        mean = "meanGammaMix",
        mean_trunc = "meanGammaMixTrunc",
        args_gpd = c("w", "shape", "scale", "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "dgamma",
        mean_base = "meanGammaMix",
        mean_trunc_base = "meanGammaMixTrunc",
        d_gpd = "dGammaGpd",
        args_gpd = c("shape", "scale", "threshold", "tail_scale", "tail_shape")
      )
    ),

    laplace = list(
      key = "laplace",
      bulk_params = c("location", "scale"),
      bulk_support = c(location = "real", scale = "positive_scale"),
      param_types = c(location = "location", scale = "scale"),
      allow_gpd = TRUE,
      defaults_X = list(location = list(mode = "link", link = "identity"),
                        scale    = list(mode = "dist")),
      sb = list(
        d = "dLaplaceMix",
        d_gpd = "dLaplaceMixGpd",
        args = c("w", "location", "scale"),
        mean = "meanLaplaceMix",
        mean_trunc = "meanLaplaceMixTrunc",
        args_gpd = c("w", "location", "scale", "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "ddexp",
        mean_base = "meanLaplaceMix",
        mean_trunc_base = "meanLaplaceMixTrunc",
        d_gpd = "dLaplaceGpd",
        args_gpd = c("location", "scale", "threshold", "tail_scale", "tail_shape")
      )
    ),

    amoroso = list(
      key = "amoroso",
      bulk_params = c("loc", "scale", "shape1", "shape2"),
      bulk_support = c(loc = "real", scale = "positive_scale",
                       shape1 = "positive_shape", shape2 = "positive_shape"),
      param_types = c(loc = "location", scale = "scale",
                      shape1 = "shape", shape2 = "shape"),
      allow_gpd = TRUE,
      defaults_X = list(
        loc    = list(mode = "link", link = "identity"),
        scale  = list(mode = "link", link = "exp"),
        shape1 = list(mode = "fixed", value = 1),
        shape2 = list(mode = "dist")
      ),
      sb = list(
        d = "dAmorosoMix",
        d_gpd = "dAmorosoMixGpd",
        args = c("w", "loc", "scale", "shape1", "shape2"),
        mean = "meanAmorosoMix",
        mean_trunc = "meanAmorosoMixTrunc",
        args_gpd = c("w", "loc", "scale", "shape1", "shape2",
                     "threshold", "tail_scale", "tail_shape")
      ),
      crp = list(
        d_base = "dAmoroso",
        mean_base = "meanAmorosoMix",
        mean_trunc_base = "meanAmorosoMixTrunc",
        d_gpd = "dAmorosoGpd",
        args_gpd = c("loc", "scale", "shape1", "shape2",
                     "threshold", "tail_scale", "tail_shape")
      )
    ),

    cauchy = list(
      key = "cauchy",
      bulk_params = c("location", "scale"),
      bulk_support = c(location = "real", scale = "positive_scale"),
      param_types = c(location = "location", scale = "scale"),
      allow_gpd = FALSE,
      defaults_X = list(location = list(mode = "link", link = "identity"),
                        scale    = list(mode = "dist")),
      sb = list(
        d = "dCauchyMix",
        d_gpd = NA_character_,
        args = c("w", "location", "scale"),
        mean = "meanCauchyMix",
        args_gpd = NA_character_
      ),
      crp = list(
        d_base = "dCauchy",
        mean_base = "meanCauchyMix",
        d_gpd = NA_character_,
        args_gpd = NA_character_
      )
    )
  )

  # ---- add signatures derived from sb/crp blocks ----
  for (k in names(kernel_registry)) {
    ki <- kernel_registry[[k]]

    sigs <- list()

    # SB signatures
    if (!is.null(ki$sb)) {
      sb_bulk_ok <- !is.null(ki$sb$d) && !is.null(ki$sb$args)
      sb_gpd_ok  <- !is.null(ki$sb$d_gpd) && !is.null(ki$sb$args_gpd) &&
        !isTRUE(is.na(ki$sb$d_gpd)) && !isTRUE(is.na(ki$sb$args_gpd))

      sigs$sb <- list(
        bulk = if (sb_bulk_ok) list(dist_name = ki$sb$d, args = ki$sb$args) else NULL,
        gpd  = if (sb_gpd_ok)  list(dist_name = ki$sb$d_gpd, args = ki$sb$args_gpd) else NULL
      )
    }

    # CRP signatures
    if (!is.null(ki$crp)) {
      crp_bulk_ok <- !is.null(ki$crp$d_base) && !isTRUE(is.na(ki$crp$d_base))
      crp_gpd_ok  <- !is.null(ki$crp$d_gpd) && !is.null(ki$crp$args_gpd) &&
        !isTRUE(is.na(ki$crp$d_gpd)) && !isTRUE(is.na(ki$crp$args_gpd))

      sigs$crp <- list(
        bulk = if (crp_bulk_ok) list(dist_name = ki$crp$d_base, args = ki$bulk_params) else NULL,
        gpd  = if (crp_gpd_ok)  list(dist_name = ki$crp$d_gpd, args = ki$crp$args_gpd) else NULL
      )
    }

    # Spliced signatures (identical to CRP, since spliced is a CRP variant)
    if (!is.null(ki$crp)) {
      sigs$spliced <- sigs$crp
    }

    kernel_registry[[k]]$signatures <- sigs
  }

  # Tail registry: GPD parameter metadata
  # indexed_by_cluster_in_crp=FALSE allows link mode for tail params in CRP/spliced
  # backends without NIMBLE sampler conflicts (tail params handled separately from
  # cluster-indexed bulk params that cause deterministic node issues with dCRP).
  tail_registry <- list(
    params = c("threshold", "tail_scale", "tail_shape"),
    support = c(threshold = "real", tail_scale = "positive_scale", tail_shape = "real"),
    indexed_by_cluster_in_crp = FALSE
  )

  ns_assign_ok <- TRUE
  tryCatch({
    assign("kernel_registry", kernel_registry, envir = ns)
    assign("tail_registry", tail_registry, envir = ns)
  }, error = function(e) {
    ns_assign_ok <<- FALSE
  })

  if (!ns_assign_ok ||
      !exists("kernel_registry", envir = ns, inherits = FALSE) ||
      !exists("tail_registry", envir = ns, inherits = FALSE)) {
    stop("Failed to initialize kernel/tail registry in package namespace.", call. = FALSE)
  }

  invisible(TRUE)
}

#' Get kernel registry
#'
#' @details
#' This accessor returns the registry created by [init_kernel_registry()]. The
#' returned object is a named list keyed by kernel name. Each kernel definition
#' describes which bulk parameters are present, how those parameters may depend
#' on covariates, whether a GPD tail is allowed, and which density or mean
#' functions should be dispatched for the supported backends.
#'
#' Downstream builders use this registry as the package-level reference for
#' kernel-specific implementation metadata. Reading it is appropriate when you
#' need to inspect what the package believes a kernel can do before constructing
#' or debugging a model specification.
#'
#' @return A list of kernel metadata.
#' @examples
#' init_kernel_registry()
#' reg <- get_kernel_registry()
#' reg$normal$bulk_params
#' @export
get_kernel_registry <- function() {
  ns <- getNamespace("CausalMixGPD")
  if (exists("kernel_registry", envir = ns, inherits = FALSE)) {
    return(get("kernel_registry", envir = ns, inherits = FALSE))
  }
  init_kernel_registry()
  if (exists("kernel_registry", envir = ns, inherits = FALSE)) {
    return(get("kernel_registry", envir = ns, inherits = FALSE))
  }
  stop("kernel_registry is not initialized.", call. = FALSE)
}

#' Get tail registry
#'
#' @details
#' The tail registry records the generalized Pareto splice used by bulk-tail
#' models. It stores the tail parameter names
#' `threshold`, `tail_scale`, and `tail_shape`, together with the support each
#' parameter must satisfy and the modeling modes the builders may assign to
#' them.
#'
#' In mathematical terms, for a threshold \eqn{u} the upper tail is represented
#' with a generalized Pareto law for excesses above \eqn{u}. Accessing this
#' registry is useful when inspecting how the package encodes those tail
#' parameters before model compilation.
#'
#' @return A list of tail metadata.
#' @examples
#' init_kernel_registry()
#' tail <- get_tail_registry()
#' tail$params
#' @export
get_tail_registry <- function() {
  ns <- getNamespace("CausalMixGPD")
  if (exists("tail_registry", envir = ns, inherits = FALSE)) {
    return(get("tail_registry", envir = ns, inherits = FALSE))
  }
  init_kernel_registry()
  if (exists("tail_registry", envir = ns, inherits = FALSE)) {
    return(get("tail_registry", envir = ns, inherits = FALSE))
  }
  stop("tail_registry is not initialized.", call. = FALSE)
}



NULL

#' Kernel support matrix
#'
#' Returns a data frame summarizing each kernel's supported features.
#'
#' @details
#' The returned table is a compact view of the registry contracts. Each row
#' corresponds to one kernel, while the logical columns indicate whether that
#' kernel can be paired with a GPD tail, whether covariate-linked parameter
#' models are defined, and whether the stick-breaking (`sb`) and
#' Chinese-restaurant (`crp`) backends are implemented.
#'
#' This helper is intended for inspection and reporting rather than model
#' fitting. It is a quick way to verify that a requested combination of kernel,
#' tail, and backend is supported before calling higher-level workflow
#' constructors.
#'
#' @param round Logical; `TRUE` to replace logical values with symbols.
#' @return data.frame with columns `kernel`, `gpd`, `covariates`, `sb`, `crp`.
#' @export
kernel_support_table <- function(round = TRUE) {
  registry <- get_kernel_registry()
  rows <- lapply(registry, function(def) {
    has_gpd <- isTRUE(def$allow_gpd) && (
      (!is.null(def$sb) && !is.null(def$sb$d_gpd) && !isTRUE(is.na(def$sb$d_gpd))) ||
        (!is.null(def$crp) && !is.null(def$crp$d_gpd) && !isTRUE(is.na(def$crp$d_gpd)))
    )
    has_cov <- length(def$defaults_X) > 0 &&
      any(vapply(def$defaults_X, function(x) identical(x$mode, "link"), logical(1)))
    has_sb <- !is.null(def$sb) && !is.null(def$sb$d)
    has_crp <- !is.null(def$crp) && !is.null(def$crp$d_base)

    data.frame(
      kernel = def$key,
      gpd = has_gpd,
      covariates = has_cov,
      sb = has_sb,
      crp = has_crp,
      stringsAsFactors = FALSE
    )
  })

  out <- do.call(rbind, rows)
  if (round) {
    fmt <- function(ok) vapply(ok, function(sf) if (isTRUE(sf)) "\u2714" else "\u274C", character(1))
    out$gpd <- fmt(out$gpd)
    out$covariates <- fmt(out$covariates)
    out$sb <- fmt(out$sb)
    out$crp <- fmt(out$crp)
  }
  out
}

# =========================
# 02-build-and-run.R
# =========================

#' Build the explicit one-arm NIMBLE bundle
#'
#' \code{build_nimble_bundle()} is the detailed constructor behind
#' \code{\link{bundle}} for one-arm models. It compiles the modeling plan into a
#' self-contained object holding code-generation inputs, initialization rules,
#' monitor policy, and stored MCMC defaults.
#'
#' @details
#' The returned bundle encodes a finite approximation to a Dirichlet process
#' mixture using either a stick-breaking (\code{"sb"}) or Chinese restaurant
#' process / spliced (\code{"crp"} / \code{"spliced"}) representation.
#'
#' For the bulk-only model, the target likelihood is the DPM predictive law
#' \deqn{f(y \mid x) = \sum_{k=1}^{K} w_k(x) f_k(y \mid x, \theta_k).}
#' When \code{GPD = TRUE}, the bundle augments the bulk model with a threshold
#' \eqn{u(x)} and generalized Pareto tail above that threshold, producing the
#' spliced predictive distribution described in the manuscript vignette.
#'
#' This function intentionally stops before model compilation and sampling.
#' Use \code{\link{run_mcmc_bundle_manual}} or \code{\link{mcmc}} to execute the
#' stored model definition.
#'
#' The object contains:
#' \itemize{
#'   \item compiled model \code{spec}
#'   \item \code{nimbleCode} model code
#'   \item \code{constants}, \code{data}, explicit \code{dimensions}
#'   \item initialization function \code{inits} (stored as a function)
#'   \item monitor specification
#'   \item MCMC settings list (stored but not used for code generation)
#' }
#'
#' @param y Numeric outcome vector.
#' @param X Optional design matrix/data.frame (N x p) for conditional variants.
#' @param ps Optional numeric vector (length N) of propensity scores. When provided,
#'   augments the design matrix for PS-adjusted outcome modeling.
#' @param backend Character; the Dirichlet process representation:
#'   \itemize{
#'     \item \code{"sb"}: stick-breaking truncation
#'     \item \code{"crp"}: Chinese Restaurant Process
#'   }
#' @param kernel Character kernel name (must exist in \code{get_kernel_registry()}).
#' @param GPD Logical; whether a GPD tail is requested.
#' @param components Integer >= 2. Single user-facing truncation parameter:
#'   \itemize{
#'     \item SB: number of mixture components used in stick-breaking truncation
#'     \item CRP: maximum number of clusters represented in the finite NIMBLE model
#'   }
#' @param param_specs Optional list with entries \code{bulk} and \code{tail} to override defaults.
#' @param mcmc Named list of MCMC settings (niter, nburnin, thin, nchains, seed). Stored in bundle.
#' @param epsilon Numeric in [0,1). For downstream summaries/plots/prediction we keep the
#'   smaller k defined by either (i) cumulative mass >= 1 - epsilon or (ii) per-component
#'   weights >= epsilon, then renormalize.
#' @param alpha_random Logical; whether the DP concentration parameter \eqn{\kappa} is stochastic.
#' @param monitor Character monitor profile:
#'   \itemize{
#'     \item \code{"core"} (default): monitors only the essential model parameters
#'     \item \code{"full"}: monitors all model nodes
#'   }
#' @param monitor_latent Logical; if TRUE, include latent cluster labels (\code{z}) in monitors.
#' @param monitor_v Logical; if TRUE and backend is SB, include stick breaks (\code{v}) in monitors.
#' @return A named list of class \code{"causalmixgpd_bundle"}. Its primary
#'   components are \code{spec}, \code{code}, \code{constants},
#'   \code{dimensions}, \code{data}, \code{inits}, \code{monitors}, and stored
#'   \code{mcmc} settings.
#' @seealso \code{\link{bundle}}, \code{\link{run_mcmc_bundle_manual}},
#'   \code{\link{predict.mixgpd_fit}}, \code{\link{kernel_support_table}},
#'   \code{\link{get_kernel_registry}}.
#' @examples
#' \donttest{
#' y <- abs(rnorm(25)) + 0.1
#' bundle <- build_nimble_bundle(
#'   y = y,
#'   backend = "sb",
#'   kernel = "normal",
#'   GPD = FALSE,
#'   components = 3,
#'   mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' )
#' bundle
#' }
#' @export
build_nimble_bundle <- function(
    y,
    X = NULL,
    ps = NULL,
    backend = c("sb", "crp", "spliced"),
    kernel,
    GPD = FALSE,
    components = 10L,
    param_specs = NULL,
    mcmc = list(niter = 2000, nburnin = 500, thin = 1, nchains = 1, seed = 1),
    epsilon = 0.025,
    alpha_random = TRUE,
    monitor = c("core", "full"),
    monitor_latent = FALSE,
    monitor_v = FALSE
) {

  requested_backend <- match.arg(backend, choices = allowed_backends)
  backend <- if (identical(requested_backend, "spliced") && !isTRUE(GPD)) "crp" else requested_backend
  monitor <- match.arg(monitor)

  if (identical(monitor, "full")) {
    monitor_latent <- TRUE
    if (identical(backend, "sb")) monitor_v <- TRUE
  }
  if (backend %in% c("crp", "spliced") && !isTRUE(monitor_latent)) {
    monitor_latent <- TRUE
  }

  y <- as.numeric(y)
  if (!length(y)) stop("y must be a non-empty numeric vector.", call. = FALSE)

  if (!is.null(X) && !is.matrix(X)) X <- as.matrix(X)
  if (!is.null(ps)) {
    ps <- as.numeric(ps)
    if (length(ps) != length(y)) stop("ps must have the same length as y.", call. = FALSE)
  }

  # Single truncation parameter for both backends
  if (is.null(components)) components <- 10L
  if (length(components) != 1L) {
    stop("components must be a single integer >= 2.", call. = FALSE)
  }
  components <- as.integer(components)
  if (!is.finite(components) || components < 2L) {
    stop("components must be an integer >= 2.", call. = FALSE)
  }

  # Basic epsilon validation (stored; used later by fit-level methods)
  if (!is.numeric(epsilon) || length(epsilon) != 1L || is.na(epsilon) || epsilon < 0 || epsilon >= 1) {
    stop("epsilon must be a single numeric value in [0, 1).", call. = FALSE)
  }

  # Compile spec (DO NOT pass mcmc here; spec/codegen is structural)
  spec <- compile_model_spec(
    y = y,
    X = X,
    ps = ps,
    backend = backend,
    kernel = kernel,
    GPD = GPD,
    components = components,
    param_specs = param_specs,
    alpha_random = alpha_random
  )
  if (identical(requested_backend, "spliced")) {
    spec$meta$backend <- "spliced"
    spec$dispatch$backend <- "spliced"
  }

  code <- .wrap_nimble_code(build_code_from_spec(spec))

  bundle <- list(
    spec       = spec,
    code       = code,
    constants  = build_constants_from_spec(spec),
    dimensions = build_dimensions_from_spec(spec),
    data       = build_data_from_inputs(y = y, X = X, ps = ps),
    inits      = build_inits_from_spec(spec, y = y, X = X),
    monitors   = build_monitors_from_spec(spec, monitor_v = monitor_v, monitor_latent = monitor_latent),
    monitor_policy = list(
      monitor = monitor,
      monitor_latent = isTRUE(monitor_latent),
      monitor_v = isTRUE(monitor_v)
    ),
    mcmc       = mcmc,
    epsilon    = epsilon
  )
  class(bundle) <- "causalmixgpd_bundle"
  bundle
}


# Internal helpers shared by SB and CRP code generators.
.codegen_prior_call <- function(dist, args, backend = "<codegen>") {
  dist <- as.character(dist)
  args <- args %||% list()
  backend <- as.character(backend %||% "<codegen>")

  if (dist == "normal") {
    m <- args$mean %||% 0
    s <- args$sd %||% 1
    return(sprintf("dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
  }
  if (dist == "gamma") {
    sh <- args$shape %||% 1
    rt <- args$rate %||% 1
    return(sprintf("dgamma(%s, %s)", deparse1(sh), deparse1(rt)))
  }
  if (dist == "invgamma") {
    sh <- args$shape %||% 1
    sc <- args$scale %||% 1
    return(sprintf("dinvgamma(%s, %s)", deparse1(sh), deparse1(sc)))
  }
  if (dist == "lognormal") {
    ml <- args$meanlog %||% 0
    sl <- args$sdlog %||% 1
    return(sprintf("dlnorm(meanlog = %s, sdlog = %s)", deparse1(ml), deparse1(sl)))
  }

  stop(sprintf("Unsupported prior dist '%s' in %s codegen.", dist, backend), call. = FALSE)
}

.codegen_link_expr <- function(eta, link, link_power = NULL) {
  link <- as.character(link %||% "identity")
  if (link == "identity") return(eta)
  if (link == "exp") return(sprintf("exp(%s)", eta))
  if (link == "log") return(sprintf("log(%s)", eta))
  if (link == "softplus") return(sprintf("log(1 + exp(%s))", eta))
  if (link == "power") {
    if (is.null(link_power) || length(link_power) != 1L || !is.finite(as.numeric(link_power))) {
      stop("power link requires numeric link_power.", call. = FALSE)
    }
    pw <- as.numeric(link_power)
    return(sprintf("pow(%s, %s)", eta, deparse1(pw)))
  }
  stop(sprintf("Unsupported link '%s'.", link), call. = FALSE)
}


#' Determine whether a compiled spec is conditional on covariates
#'
#' A spec is "conditional" if it uses covariates \code{X} (i.e., \code{has_X=TRUE})
#' and at least one parameter is specified in \code{link} mode.
#' Validate that code generation is supported for a compiled spec
#'
#' Checks that the kernel registry contains the required likelihood signatures
#' for the chosen backend and whether GPD is requested, and validates that any
#' link functions requested in \code{spec$plan} are supported by the code generator.
#' Build NIMBLE data list from user inputs
#'
#' Converts user-provided outcome vector \code{y} and optional covariates \code{X}
#' into a NIMBLE-ready data list.
#'
#' Rules:
#' \itemize{
#'   \item \code{y} is always returned as a numeric vector.
#'   \item \code{X} is returned only if provided; it is coerced to a numeric matrix.
#'   \item No constants (N/P/components) are included here; those belong in \code{constants}.
#' }
#'
#' @param y Numeric outcome vector (length N).
#' @param X Optional covariate matrix/data.frame (N x P).
#' @param ps Optional numeric vector (length N) containing propensity scores.
#' @return Named list suitable to pass as \code{data} into \code{nimbleModel()}.
#' @keywords internal
#' @noRd
build_data_from_inputs <- function(y, X = NULL, ps = NULL) {
  y <- as.numeric(y)
  if (!length(y)) stop("y must be a non-empty numeric vector.", call. = FALSE)

  out <- list(y = y)

  if (!is.null(X)) {
    if (!is.matrix(X)) X <- as.matrix(X)
    # Coerce to numeric matrix (nimble expects numeric)
    storage.mode(X) <- "double"
    if (nrow(X) != length(y)) stop("X must have nrow(X) == length(y).", call. = FALSE)
    if (ncol(X) < 1L) stop("X must have at least one column.", call. = FALSE)
    out$X <- X
  }

  if (!is.null(ps)) {
    ps <- as.numeric(ps)
    if (length(ps) != length(y)) stop("ps must have the same length as y.", call. = FALSE)
    out$ps <- ps
  }

  out
}



#' Build default monitors from a compiled model spec
#'
#' Returns the character vector of node names to monitor in MCMC.
#' This is a pre-run builder used by \code{build_nimble_bundle()}.
#'
#' Monitoring follows these rules:
#' \itemize{
#'   \item Always monitor concentration \code{kappa} (whether fixed or stochastic).
#'   \item SB: monitor \code{w[1:components]} and optionally \code{v[1:(components-1)]}.
#'   \item CRP: monitor \code{z[1:N]}.
#'   \item Bulk parameters:
#'     \itemize{
#'       \item dist/fixed: monitor \code{<param>[1:components]}
#'       \item link: monitor \code{beta_<param>[1:components, 1:P]}
#'     }
#'   \item GPD (if enabled):
#'     \itemize{
#'       \item threshold: monitor scalar \code{threshold} when not link-mode; \code{threshold[1:N]} for link-mode
#'       \item if threshold is link-mode: monitor \code{beta_threshold[1:P]}
#'       \item if threshold uses LN link-dist default: monitor \code{sdlog_u}
#'       \item tail_scale: if link-mode, monitor \code{beta_tail_scale[1:P]}
#'       \item tail_shape: monitor scalar \code{tail_shape} (fixed or dist)
#'     }
#' }
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @param monitor_v Logical; for SB, whether to also monitor stick breaks \code{v}.
#' @param monitor_latent Logical; whether to monitor latent cluster labels \code{z}.
#' @return Character vector of node names to monitor.
#' @keywords internal
build_monitors_from_spec <- function(spec, monitor_v = FALSE, monitor_latent = FALSE) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  meta <- spec$meta
  plan <- spec$plan

  backend <- meta$backend
  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  # PS is optional: check if it's in the plan
  has_ps <- !is.null(plan$ps)
  K <- as.integer(meta$components)
  cluster_gating <- isTRUE((spec$cluster %||% list())$gating)

  mons <- character()

  # Always monitor alpha (fixed alpha is still useful to carry in samples/prints)
  mons <- c(mons, "alpha")

  # BNP backbone
  if (identical(backend, "sb")) {
    if (isTRUE(cluster_gating)) {
      if (P < 1L) stop("Cluster gating requires P > 0.", call. = FALSE)
      mons <- c(mons, sprintf("eta[1:%d]", K - 1L))
      mons <- c(mons, sprintf("B[1:%d,1:%d]", K - 1L, P))
      if (isTRUE(monitor_latent)) mons <- c(mons, sprintf("z[1:%d]", N))
    } else {
      mons <- c(mons, sprintf("w[1:%d]", K))
      if (isTRUE(monitor_latent)) mons <- c(mons, sprintf("z[1:%d]", N))
      if (isTRUE(monitor_v)) {
        mons <- c(mons, sprintf("v[1:%d]", K - 1L))
      }
    }
  } else if (backend %in% c("crp", "spliced")) {
    if (isTRUE(monitor_latent)) mons <- c(mons, sprintf("z[1:%d]", N))
  } else {
    stop("Unknown backend in spec$meta$backend.", call. = FALSE)
  }

  # Bulk parameters
  bulk <- plan$bulk %||% list()
  for (nm in names(bulk)) {
    ent <- bulk[[nm]]
    mode <- ent$mode %||% NA_character_

    if (mode %in% c("fixed", "dist")) {
      mons <- c(mons, sprintf("%s[1:%d]", nm, K))
    } else if (identical(mode, "link")) {
      if (P < 1L) stop(sprintf("bulk[%s] is link-mode but P=0.", nm), call. = FALSE)
      mons <- c(mons, sprintf("beta_%s[1:%d,1:%d]", nm, K, P))
      if (has_ps) {
        mons <- c(mons, sprintf("beta_ps_%s[1:%d]", nm, K))
      }
    } else {
      stop(sprintf("Invalid bulk plan mode for '%s'.", nm), call. = FALSE)
    }
  }

  # GPD
  if (isTRUE(meta$GPD)) {
    gpd <- plan$gpd %||% list()
    is_spliced <- identical(backend, "spliced")

    if (is_spliced) {
      # ======== SPLICED BACKEND: Component-level GPD parameterization ========
      
      # threshold
      if (!is.null(gpd$threshold)) {
        thr_mode <- gpd$threshold$mode %||% NA_character_
        if (identical(thr_mode, "link")) {
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("beta_threshold[1:%d,1:%d]", K, P))
          if (!is.null(gpd$threshold$link_dist) &&
              identical(gpd$threshold$link_dist$dist, "lognormal")) {
            mons <- c(mons, sprintf("threshold_i[1:%d,1:%d]", N, K))
            mons <- c(mons, "sdlog_u")
          }
          # Do NOT monitor deterministic threshold_i[i] by default.
        } else if (thr_mode %in% c("fixed", "dist")) {
          mons <- c(mons, sprintf("threshold[1:%d]", K))
        } else {
          stop("Invalid gpd$threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      if (!is.null(gpd$tail_scale)) {
        ts_mode <- gpd$tail_scale$mode %||% NA_character_
        if (identical(ts_mode, "link")) {
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("beta_tail_scale[1:%d,1:%d]", K, P))
        } else if (ts_mode %in% c("fixed", "dist")) {
          mons <- c(mons, sprintf("tail_scale[1:%d]", K))
        } else {
          stop("Invalid gpd$tail_scale mode.", call. = FALSE)
        }
      }

      # tail_shape
      if (!is.null(gpd$tail_shape)) {
        tsh_mode <- gpd$tail_shape$mode %||% NA_character_
        if (identical(tsh_mode, "link")) {
          if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("beta_tail_shape[1:%d,1:%d]", K, P))
        } else if (tsh_mode %in% c("fixed", "dist")) {
          mons <- c(mons, sprintf("tail_shape[1:%d]", K))
        } else {
          stop("Invalid gpd$tail_shape mode.", call. = FALSE)
        }
      }

    } else {
      # ======== STANDARD CRP/SB BACKEND: Original behavior ========
      
      # threshold
      if (!is.null(gpd$threshold)) {
        thr_mode <- gpd$threshold$mode %||% NA_character_
        if (identical(thr_mode, "link")) {
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("threshold[1:%d]", N))
          mons <- c(mons, sprintf("beta_threshold[1:%d]", P))

          # LN around-link default: monitor sdlog_u if present in plan
          if (!is.null(gpd$threshold$link_dist) &&
              identical(gpd$threshold$link_dist$dist, "lognormal")) {
            mons <- c(mons, "sdlog_u")
          }
        } else if (thr_mode %in% c("fixed", "dist")) {
          mons <- c(mons, "threshold")
        } else {
          stop("Invalid gpd$threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      if (!is.null(gpd$tail_scale)) {
        ts_mode <- gpd$tail_scale$mode %||% NA_character_
        if (identical(ts_mode, "link")) {
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("beta_tail_scale[1:%d]", P))
        } else if (ts_mode %in% c("fixed", "dist")) {
          mons <- c(mons, "tail_scale")
        } else {
          stop("Invalid gpd$tail_scale mode.", call. = FALSE)
        }
      }

      # tail_shape
      if (!is.null(gpd$tail_shape)) {
        tsh_mode <- gpd$tail_shape$mode %||% NA_character_
        if (identical(tsh_mode, "link")) {
          if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
          mons <- c(mons, sprintf("beta_tail_shape[1:%d]", P))
        } else {
          # fixed or dist: monitor scalar
          mons <- c(mons, "tail_shape")
        }
      }
    }
  }

  unique(mons)
}

#' Build initial values from a compiled model spec
#'
#' Produces a list of initial values suitable for passing to \code{nimbleModel}.
#' The initial values are derived from \code{spec$plan} and are intended to be
#' stable and support-respecting (e.g., positive parameters start positive).
#'
#' Notes:
#' \itemize{
#'   \item Uses only \code{components} as the model size parameter.
#'   \item SB: initializes stick breaks \code{v}; weights \code{w} are deterministic.
#'   \item CRP: initializes memberships \code{z} in \code{1:components}.
#'   \item Link-mode parameters initialize regression coefficients \code{beta_<param>}
#'         with shape \code{components x P}.
#'   \item Default GPD threshold under X is stochastic lognormal:
#'         initializes \code{threshold[1:N]} and scalar \code{sdlog_u} (non-link thresholds are scalar).
#' }
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @param seed Optional seed (single integer or vector). If provided, the first element is used.
#' @param y Optional numeric vector of observed outcomes used for heuristic initializations.
#' @param X Optional numeric matrix of covariates used for link-mode parameter initializations.
#' @return Named list of initial values.
#' @keywords internal
build_inits_from_spec <- function(spec, seed = NULL, y = NULL, X = NULL) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  if (!is.null(seed)) {
    seed <- as.integer(seed)
    if (length(seed) >= 1L && is.finite(seed[1L])) {
      set.seed(seed[1L])
    }
  }

  meta <- spec$meta
  plan <- spec$plan

  backend <- meta$backend
  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  has_ps <- !is.null(plan$ps)
  K <- as.integer(meta$components)
  y_obs <- if (!is.null(y)) as.numeric(y) else numeric()
  X_obs <- if (!is.null(X)) {
    if (!is.matrix(X)) X <- as.matrix(X)
    storage.mode(X) <- "double"
    X
  } else {
    NULL
  }
  cluster_gating <- isTRUE((spec$cluster %||% list())$gating)

  inits <- list()

  threshold_seed_value <- function() {
    q <- if (length(y_obs)) {
      suppressWarnings(stats::quantile(y_obs, probs = 0.8, na.rm = TRUE, names = FALSE))
    } else {
      NA_real_
    }
    if (!is.finite(q) || length(q) != 1L) q <- 1
    max(as.numeric(q), .Machine$double.eps)
  }

  threshold_link_seed <- function() {
    q <- threshold_seed_value()
    if (P < 1L) {
      return(numeric())
    }
    target <- rep(log(q), N)
    beta <- tryCatch(
      as.numeric(stats::coef(stats::lm.fit(x = X_obs, y = target))),
      error = function(e) rep(NA_real_, P)
    )
    if (length(beta) != P || anyNA(beta) || any(!is.finite(beta))) {
      beta <- rep(0, P)
      if (!is.null(X_obs) && is.matrix(X_obs) && ncol(X_obs) >= 1L &&
          all(is.finite(X_obs[, 1L])) && stats::var(X_obs[, 1L]) < .Machine$double.eps) {
        beta[1L] <- log(q)
      }
    }
    beta
  }

  latent_label_seed <- function(K_init) {
    if (length(y_obs) == N && N >= 1L && sum(is.finite(y_obs)) == N) {
      ord <- order(y_obs)
      z_init <- integer(N)
      grp <- floor(((seq_len(N) - 1L) * K_init) / N) + 1L
      grp_labels <- sample.int(K_init, K_init)
      z_init[ord] <- grp_labels[pmin.int(K_init, grp)]
      return(z_init)
    }
    sample.int(K_init, size = N, replace = TRUE)
  }

  # ---- concentration alpha ----
  conc <- plan$concentration %||% list()
  if (identical(conc$mode, "dist")) {
    # alpha is stochastic -> needs init
    inits$alpha <- 1
  } else if (identical(conc$mode, "fixed")) {
    # alpha deterministic in code: alpha <- value
    # do not set inits$alpha to avoid conflicts
  } else {
    stop("Invalid plan$concentration$mode.", call. = FALSE)
  }

  # ---- BNP backbone ----
  if (identical(backend, "sb")) {
    if (isTRUE(cluster_gating)) {
      if (P < 1L) stop("Cluster gating requires P > 0.", call. = FALSE)
      inits$eta <- stats::rnorm(K - 1L, mean = 0, sd = 0.1)
      inits$B <- matrix(
        stats::rnorm((K - 1L) * P, mean = 0, sd = 0.1),
        nrow = K - 1L,
        ncol = P
      )
    } else {
      # v[j] ~ dbeta(1, alpha); initialize in (0,1)
      if (K <= 2L) {
        inits$v <- runif(1L, 0.2, 0.8)
      } else {
        inits$v <- runif(K - 1L, 0.2, 0.8)
      }
    }
    # w/w_x are deterministic from latent gating/stick blocks; do not init
    K_init <- max(2L, min(K, 5L))
    inits$z <- latent_label_seed(K_init)
  } else if (backend %in% c("crp", "spliced")) {
    # z[1:N] ~ dCRP(...); init in 1:K, avoid all unique labels for stability
    K_init <- max(2L, min(K, 5L))
    inits$z <- latent_label_seed(K_init)
  } else {
    stop("Unknown backend in spec$meta$backend.", call. = FALSE)
  }

  # ---- bulk parameters ----
  bulk <- plan$bulk %||% list()
  kinfo <- spec$kernel_info %||% list()
  ptypes <- kinfo$param_types %||% list()
  psupport <- kinfo$bulk_support %||% list()

  init_by_type <- function(type, support = NULL) {
    type <- as.character(type %||% "location")
    support <- as.character(support %||% "")
    if (support %in% c("positive_location", "positive_scale", "positive_shape", "positive_sd")) return(1)
    if (type == "location") return(0)
    if (type %in% c("scale", "shape", "sd")) return(1)
    0
  }

  for (nm in names(bulk)) {
    ent <- bulk[[nm]]
    mode <- ent$mode %||% NA_character_

    if (identical(mode, "fixed")) {
      # deterministic; no init
      next
    }

    if (identical(mode, "dist")) {
      val <- init_by_type(ptypes[[nm]], psupport[[nm]])
      inits[[nm]] <- rep(val, K)
      next
    }

    if (identical(mode, "link")) {
      if (P < 1L) stop(sprintf("bulk[%s] is link-mode but P=0.", nm), call. = FALSE)
      # beta_<nm>[1:K,1:P]
      inits[[paste0("beta_", nm)]] <- matrix(0, nrow = K, ncol = P)
      if (has_ps) {
        inits[[paste0("beta_ps_", nm)]] <- rep(0, K)
      }
      next
    }

    stop(sprintf("Invalid bulk plan mode for '%s'.", nm), call. = FALSE)
  }

  # ---- GPD parameters ----
  if (isTRUE(meta$GPD)) {
    gpd <- plan$gpd %||% list()
    is_spliced <- identical(backend, "spliced")

    if (is_spliced) {
      # ======== SPLICED BACKEND: Component-level GPD parameterization ========

      # threshold
      if (!is.null(gpd$threshold)) {
        thr_mode <- gpd$threshold$mode %||% NA_character_

        if (identical(thr_mode, "dist")) {
          # Stochastic component-level vector
          q <- if (length(y_obs)) {
            suppressWarnings(stats::quantile(y_obs, probs = 0.8, na.rm = TRUE, names = FALSE))
          } else {
            NA_real_
          }
          if (!is.finite(q) || length(q) != 1L) q <- 1
          q <- max(as.numeric(q), .Machine$double.eps)
          inits$threshold <- rep(q, K)
        } else if (identical(thr_mode, "link")) {
          # Component-specific beta coefficients
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          inits$beta_threshold <- matrix(0, nrow = K, ncol = P)
          if (!is.null(gpd$threshold$link_dist) &&
              identical(gpd$threshold$link_dist$dist, "lognormal")) {
            inits$threshold_i <- matrix(threshold_seed_value(), nrow = N, ncol = K)
            inits$sdlog_u <- 0.2
          }
        } else if (identical(thr_mode, "fixed")) {
          # Deterministic; no init
        } else {
          stop("Invalid gpd$threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      if (!is.null(gpd$tail_scale)) {
        ts_mode <- gpd$tail_scale$mode %||% NA_character_

        if (identical(ts_mode, "dist")) {
          # Stochastic component-level vector
          inits$tail_scale <- rep(1, K)
        } else if (identical(ts_mode, "link")) {
          # Component-specific beta coefficients
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          inits$beta_tail_scale <- matrix(0, nrow = K, ncol = P)
          # tail_scale_i[i] is deterministic; do not init
        } else if (identical(ts_mode, "fixed")) {
          # Deterministic; no init
        } else {
          stop("Invalid gpd$tail_scale mode.", call. = FALSE)
        }
      }

      # tail_shape
      if (!is.null(gpd$tail_shape)) {
        tsh_mode <- gpd$tail_shape$mode %||% NA_character_

        if (identical(tsh_mode, "dist")) {
          # Stochastic component-level vector
          inits$tail_shape <- rep(0, K)
        } else if (identical(tsh_mode, "link")) {
          # Component-specific beta coefficients
          if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
          inits$beta_tail_shape <- matrix(0, nrow = K, ncol = P)
          # tail_shape_i[i] is deterministic; do not init
        } else if (identical(tsh_mode, "fixed")) {
          # Deterministic; no init
        } else {
          stop("Invalid gpd$tail_shape mode.", call. = FALSE)
        }
      }

    } else {
      # ======== STANDARD CRP/SB BACKEND: Original behavior ========

      # tail_shape (stochastic only)
      if (!is.null(gpd$tail_shape) && identical(gpd$tail_shape$mode, "dist")) {
        inits$tail_shape <- 0
      } else if (!is.null(gpd$tail_shape) && identical(gpd$tail_shape$mode, "link")) {
        # Standard CRP with link mode for tail_shape (newly supported)
        if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
        inits$beta_tail_shape <- rep(0, P)
        # tail_shape[i] deterministic; do not init
      }

      # threshold
      if (!is.null(gpd$threshold)) {
        thr_mode <- gpd$threshold$mode %||% NA_character_

        if (thr_mode %in% c("fixed", "dist")) {
          inits$threshold <- threshold_seed_value()
        } else if (identical(thr_mode, "link")) {
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          inits$beta_threshold <- threshold_link_seed()

          # if LN around-link: threshold[i] is stochastic lognormal and sdlog_u exists
          if (!is.null(gpd$threshold$link_dist) &&
              identical(gpd$threshold$link_dist$dist, "lognormal")) {
            # positive threshold init; 0.8-quantile is usually safe and data-informed
            q <- threshold_seed_value()
            inits$threshold <- rep(q, N)
            inits$sdlog_u <- 0.2
          } else {
            # link-mode without link_dist: treat threshold deterministic from link later;
            # but our code currently represents threshold[i] as stochastic only in LN default.
            # Still initialize threshold to be safe if node exists.
            inits$threshold <- rep(1, N)
          }
        } else {
          stop("Invalid gpd$threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      if (!is.null(gpd$tail_scale)) {
        ts_mode <- gpd$tail_scale$mode %||% NA_character_
        if (identical(ts_mode, "link")) {
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          inits$beta_tail_scale <- rep(0, P)
          # tail_scale[i] deterministic from beta_tail_scale; do not init tail_scale
        } else if (identical(ts_mode, "dist")) {
          # scalar stochastic tail_scale; init positive if node exists in code
          inits$tail_scale <- 1
        } else if (identical(ts_mode, "fixed")) {
          # deterministic; no init
        } else {
          stop("Invalid gpd$tail_scale mode.", call. = FALSE)
        }
      }
    }
  }

  inits
}


#' Build constants list from a compiled model spec
#'
#' Produces a named list of constants to pass into \code{nimbleModel}.
#' Constants include core sizes (\code{N}, \code{P}, \code{components}) and
#' hyperparameters for priors implied by \code{spec$plan}.
#'
#' This function is pre-run only; it does not compile or execute NIMBLE.
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @return Named list of constants.
#' @keywords internal
build_constants_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  meta <- spec$meta
  plan <- spec$plan
  has_ps <- !is.null(plan$ps)

  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  K <- as.integer(meta$components)
  default_ps_prior <- list(dist = "normal", args = list(mean = 0, sd = 2))
  ps_prior <- (plan$ps %||% list(prior = default_ps_prior))$prior

  const <- list(
    N = N,
    P = P,
    components = K
  )

  # ---- helper: register prior hypers in a uniform way ----
  # Supported priors here (as constants):
  # normal: mean, sd
  # gamma: shape, rate
  # invgamma: shape, scale
  # lognormal: meanlog, sdlog  (rare as prior, but allowed)
  add_prior_constants <- function(prefix, dist, args) {
    dist <- as.character(dist)
    args <- args %||% list()

    if (dist == "normal") {
      const[[paste0(prefix, "_mean")]] <- as.numeric(args$mean %||% 0)
      const[[paste0(prefix, "_sd")]]   <- as.numeric(args$sd %||% 1)
    } else if (dist == "gamma") {
      const[[paste0(prefix, "_shape")]] <- as.numeric(args$shape %||% 1)
      const[[paste0(prefix, "_rate")]]  <- as.numeric(args$rate %||% 1)
    } else if (dist == "invgamma") {
      const[[paste0(prefix, "_shape")]] <- as.numeric(args$shape %||% 1)
      const[[paste0(prefix, "_scale")]] <- as.numeric(args$scale %||% 1)
    } else if (dist == "lognormal") {
      const[[paste0(prefix, "_meanlog")]] <- as.numeric(args$meanlog %||% 0)
      const[[paste0(prefix, "_sdlog")]]   <- as.numeric(args$sdlog %||% 1)
    } else {
      stop(sprintf("Unsupported prior distribution '%s' for constants.", dist), call. = FALSE)
    }

    invisible(NULL)
  }

  # ---- concentration alpha ----
  conc <- plan$concentration %||% list()
  if (identical(conc$mode, "dist")) {
    add_prior_constants("alpha", conc$dist %||% "gamma", conc$args %||% list(shape = 1, rate = 1))
  } else if (identical(conc$mode, "fixed")) {
    # fixed alpha: no hypers; the code will set alpha <- value
  } else {
    stop("Invalid plan$concentration$mode.", call. = FALSE)
  }

  # ---- bulk priors ----
  bulk <- plan$bulk %||% list()
  for (nm in names(bulk)) {
    ent <- bulk[[nm]]
    mode <- ent$mode %||% NA_character_

    if (identical(mode, "dist")) {
      add_prior_constants(paste0("bulk_", nm), ent$dist, ent$args)
    } else if (identical(mode, "link")) {
      bp <- ent$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 2))
      add_prior_constants(paste0("beta_", nm), bp$dist %||% "normal", bp$args %||% list(mean = 0, sd = 2))
      if (has_ps) {
        add_prior_constants(paste0("beta_ps_", nm), ps_prior$dist %||% "normal",
                            ps_prior$args %||% list(mean = 0, sd = 2))
      }
    } else if (identical(mode, "fixed")) {
      # no hypers
    } else {
      stop(sprintf("Invalid bulk plan mode for '%s'.", nm), call. = FALSE)
    }
  }

  # ---- GPD priors ----
  if (isTRUE(meta$GPD)) {
    gpd <- plan$gpd %||% list()

    # threshold
    thr <- gpd$threshold %||% NULL
    if (!is.null(thr)) {
      thr_mode <- thr$mode %||% NA_character_

      if (identical(thr_mode, "dist")) {
        add_prior_constants("gpd_threshold", thr$dist, thr$args)
      } else if (identical(thr_mode, "link")) {
        # beta_threshold prior
        bp <- thr$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
        add_prior_constants("beta_threshold", bp$dist %||% "normal", bp$args %||% list(mean = 0, sd = 0.2))

        # link_dist: if LN around-link default, sdlog_u prior exists in plan as gpd$sdlog_u
        if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
          sdlog_u <- gpd$sdlog_u %||% list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
          if (!identical(sdlog_u$mode, "dist")) {
            stop("sdlog_u must be dist-mode when using lognormal threshold link_dist.", call. = FALSE)
          }
          add_prior_constants("sdlog_u", sdlog_u$dist %||% "invgamma", sdlog_u$args %||% list(shape = 2, scale = 1))
        }
      } else if (identical(thr_mode, "fixed")) {
        # no hypers
      } else {
        stop("Invalid gpd$threshold mode.", call. = FALSE)
      }
    }

    # tail_scale
    ts <- gpd$tail_scale %||% NULL
    if (!is.null(ts)) {
      ts_mode <- ts$mode %||% NA_character_
      if (identical(ts_mode, "dist")) {
        add_prior_constants("gpd_tail_scale", ts$dist, ts$args)
      } else if (identical(ts_mode, "link")) {
        bp <- ts$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.5))
        add_prior_constants("beta_tail_scale", bp$dist %||% "normal", bp$args %||% list(mean = 0, sd = 0.5))
      } else if (identical(ts_mode, "fixed")) {
        # no hypers
      } else {
        stop("Invalid gpd$tail_scale mode.", call. = FALSE)
      }
    }

    # tail_shape
    tsh <- gpd$tail_shape %||% NULL
    if (!is.null(tsh)) {
      tsh_mode <- tsh$mode %||% NA_character_
      if (identical(tsh_mode, "dist")) {
        add_prior_constants("gpd_tail_shape", tsh$dist, tsh$args)
      } else if (identical(tsh_mode, "link")) {
        bp <- tsh$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.3))
        add_prior_constants("beta_tail_shape", bp$dist %||% "normal", bp$args %||% list(mean = 0, sd = 0.3))
      } else if (identical(tsh_mode, "fixed")) {
        # no hypers
      } else {
        stop("Invalid gpd$tail_shape mode.", call. = FALSE)
      }
    }
  }

  const
}

#' Build dimension declarations from a compiled model spec
#'
#' Returns a named list of array dimensions used by downstream builders
#' (inits/monitors/code generation). Dimensions are derived solely from
#' \code{spec$meta} and \code{spec$plan}. This function does not inspect data.
#'
#' The model size is controlled by a single parameter: \code{components}.
#' For SB this is the truncation level of the stick-breaking mixture.
#' For CRP this is the maximum number of clusters represented in the finite model.
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @return Named list of dimensions (integer vectors). Scalars are omitted.
#' @keywords internal
build_dimensions_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))

  meta <- spec$meta
  plan <- spec$plan

  backend <- meta$backend
  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  has_ps <- !is.null(plan$ps)
  K <- as.integer(meta$components)

  cluster_gating <- isTRUE((spec$cluster %||% list())$gating)

  dims <- list()

  # --- BNP backbone dims ---
  if (identical(backend, "sb")) {
    if (isTRUE(cluster_gating)) {
      if (P < 1L) stop("Cluster gating requires P > 0.", call. = FALSE)
      dims$eta <- c(K - 1L)
      dims$B <- c(K - 1L, P)
      dims$logit_ij <- c(N, K)
      dims$logit_max <- c(N)
      dims$logit_denom <- c(N)
      dims$exp_logit_ij <- c(N, K)
      dims$w_x <- c(N, K)
    } else {
      # SB breaks v[1:(K-1)] and weights w[1:K]
      dims$v <- c(K - 1L)
      dims$w <- c(K)
    }
    dims$z <- c(N)
  } else if (backend %in% c("crp", "spliced")) {
    # CRP/spliced memberships z[1:N]
    dims$z <- c(N)
  } else {
    stop("Unknown backend in spec$meta$backend.", call. = FALSE)
  }

  # --- Bulk parameter dims ---
  bulk_plan <- plan$bulk %||% list()
  bulk_names <- names(bulk_plan)

  for (nm in bulk_names) {
    entry <- bulk_plan[[nm]]
    mode <- entry$mode %||% NA_character_

    # Component-level parameter vectors exist for fixed/dist modes.
    # For link mode, the component-level parameter is represented via beta_<nm>,
    # and per-(i,component) derived arrays are deterministic and not dimension-declared here.
    if (mode %in% c("fixed", "dist")) {
      dims[[nm]] <- c(K)
    }

    if (identical(mode, "link")) {
      # regression coefficients: beta_<param>[1:K, 1:P]
      if (P < 1L) stop(sprintf("Parameter '%s' is link-mode but P=0.", nm), call. = FALSE)
      dims[[paste0("beta_", nm)]] <- c(K, P)
      if (has_ps) {
        dims[[paste0("beta_ps_", nm)]] <- c(K)
      }
    }
  }

  # --- GPD dims (if enabled) ---
  if (isTRUE(meta$GPD)) {
    gpd_plan <- plan$gpd %||% list()
    is_spliced <- identical(backend, "spliced")

    if (is_spliced) {
      # ======== SPLICED BACKEND: Component-level GPD parameterization ========
      
      # threshold
      thr <- gpd_plan$threshold %||% NULL
      if (!is.null(thr)) {
        thr_mode <- thr$mode %||% NA_character_

        if (thr_mode %in% c("fixed", "dist")) {
          dims$threshold <- c(K)  # component-level vector
        } else if (identical(thr_mode, "link")) {
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          dims$beta_threshold <- c(K, P)      # component-specific coefficients
          if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
            dims$eta_threshold <- c(N, K)
            dims$threshold_i <- c(N, K)
          } else {
            dims$eta_threshold <- c(N)        # linear predictor (optional, but harmless)
            dims$threshold_i <- c(N)          # transformed parameter
          }
        } else {
          stop("Invalid gpd$threshold mode in plan.", call. = FALSE)
        }
      }

      # tail_scale
      ts <- gpd_plan$tail_scale %||% NULL
      if (!is.null(ts)) {
        ts_mode <- ts$mode %||% NA_character_

        if (ts_mode %in% c("fixed", "dist")) {
          dims$tail_scale <- c(K)  # component-level vector
        } else if (identical(ts_mode, "link")) {
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          dims$beta_tail_scale <- c(K, P)
          dims$eta_tail_scale <- c(N)
          dims$tail_scale_i <- c(N)
        } else {
          stop("Invalid gpd$tail_scale mode in plan.", call. = FALSE)
        }
      }

      # tail_shape
      tsh <- gpd_plan$tail_shape %||% NULL
      if (!is.null(tsh)) {
        tsh_mode <- tsh$mode %||% NA_character_

        if (tsh_mode %in% c("fixed", "dist")) {
          dims$tail_shape <- c(K)  # component-level vector
        } else if (identical(tsh_mode, "link")) {
          if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
          dims$beta_tail_shape <- c(K, P)
          dims$eta_tail_shape <- c(N)
          dims$tail_shape_i <- c(N)
        } else {
          stop("Invalid gpd$tail_shape mode in plan.", call. = FALSE)
        }
      }

    } else {
      # ======== STANDARD CRP/SB BACKEND: Original behavior ========
      
      # threshold
      thr <- gpd_plan$threshold %||% NULL
      if (!is.null(thr)) {
        thr_mode <- thr$mode %||% NA_character_

        if (thr_mode %in% c("fixed", "dist")) {
          # scalar threshold
        } else if (identical(thr_mode, "link")) {
          # threshold[i] stochastic LN around X beta
          dims$threshold <- c(N)
          if (P < 1L) stop("GPD threshold is link-mode but P=0.", call. = FALSE)
          if (P > 1L) dims$beta_threshold <- c(P)

          # if link_dist exists and uses sdlog_u, include its scalar node (no dims entry)
          # but if user chooses to model sdlog_u as a vector later, this would change.
          # For now: sdlog_u is scalar -> omitted from dims.
        } else {
          stop("Invalid gpd$threshold mode in plan.", call. = FALSE)
        }
      }

      # tail_scale
      ts <- gpd_plan$tail_scale %||% NULL
      if (!is.null(ts)) {
        ts_mode <- ts$mode %||% NA_character_

        if (ts_mode %in% c("fixed", "dist")) {
          # scalar tail_scale when not linked (most common non-X default)
          # (no dims entry for scalar)
        } else if (identical(ts_mode, "link")) {
          # tail_scale[i] is deterministic from X beta
          if (P < 1L) stop("GPD tail_scale is link-mode but P=0.", call. = FALSE)
          if (P > 1L) dims$beta_tail_scale <- c(P)
          # tail_scale[i] deterministic -> not dimension-declared
        } else {
          stop("Invalid gpd$tail_scale mode in plan.", call. = FALSE)
        }
      }

      # tail_shape
      tsh <- gpd_plan$tail_shape %||% NULL
      if (!is.null(tsh)) {
        tsh_mode <- tsh$mode %||% NA_character_
        if (!tsh_mode %in% c("fixed", "dist", "link")) stop("Invalid gpd$tail_shape mode in plan.", call. = FALSE)
        # For standard CRP: tail_shape scalar for fixed/dist, or observation-level for link
        if (identical(tsh_mode, "link")) {
          if (P < 1L) stop("GPD tail_shape is link-mode but P=0.", call. = FALSE)
          if (P > 1L) dims$beta_tail_shape <- c(P)
          # tail_shape[i] deterministic -> not dimension-declared
        }
        # tail_shape scalar (fixed/dist) -> no dims entry
      }

      # sdlog_u (scalar by design; omitted from dims)
      # If you later allow sdlog_u[i], you'd add dims here.
    }
  }

  dims
}

#' Build NIMBLE model code from a compiled model spec
#'
#' Dispatches to the backend-specific code generators:
#' \itemize{
#'   \item \code{build_code_sb_from_spec()} for stick-breaking (\code{"sb"})
#'   \item \code{build_code_crp_from_spec()} for CRP and spliced (\code{"crp"}, \code{"spliced"})
#' }
#'
#' The model size is controlled by \code{spec$meta$components} only.
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @return A \code{nimbleCode} object.
#' @keywords internal
build_code_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$meta$backend))

  backend <- spec$meta$backend
  if (identical(backend, "sb")) {
    return(build_code_sb_from_spec(spec))
  }
  if (backend %in% c("crp", "spliced")) {
    return(build_code_crp_from_spec(spec))
  }

  stop(sprintf("Unknown backend '%s' in spec$meta$backend.", as.character(backend)), call. = FALSE)
}


#' Build NIMBLE code for SB backend from a compiled spec
#'
#' Generates \code{nimbleCode} for the stick-breaking (SB) backend using native
#' NIMBLE BNP utilities:
#' \itemize{
#'   \item stick breaks \code{v[j] ~ dbeta(1, alpha)}
#'   \item weights computed via \code{v} stick multiplications (no call to \code{stick_breaking})
#' }
#'
#' Likelihood calls are emitted using positional arguments only (no names).
#'
#' @param spec A compiled model specification from \code{compile_model_spec()}.
#' @return A \code{nimbleCode} object.
#' @keywords internal
#' @noRd
# nocov start
.strip_covr_counts <- function(txt) {
  lines <- strsplit(txt, "\n", fixed = TRUE)[[1]]
  if (!length(lines)) return("")

  # Remove standalone and inline covr counter calls inserted by instrumentation.
  lines <- gsub("\\bcovr:::count\\([^)]*\\)\\s*;?", "", lines, perl = TRUE)

  # Drop covr-only conditional wrappers that can leave bare `if` tokens in parsed code.
  covr_if <- grepl("^\\s*if\\s*\\([^)]*covr:::[^)]*\\)\\s*\\{?\\s*$", lines)
  lines[covr_if] <- ""

  # Remove braces that only closed an instrumentation-only `if` block.
  orphan_closing <- logical(length(lines))
  for (i in seq_along(lines)) {
    if (grepl("^\\s*}\\s*$", lines[i])) {
      prev_nonempty <- i - 1L
      while (prev_nonempty >= 1L && !nzchar(trimws(lines[prev_nonempty]))) {
        prev_nonempty <- prev_nonempty - 1L
      }
      if (prev_nonempty >= 1L && grepl("^\\s*if\\s*\\([^)]*covr:::[^)]*\\)\\s*\\{?\\s*$", lines[prev_nonempty])) {
        orphan_closing[i] <- TRUE
      }
    }
  }
  lines[orphan_closing] <- ""

  lines <- trimws(lines, which = "right")
  lines <- lines[nzchar(trimws(lines))]
  paste(lines, collapse = "\n")
}

.deparse_without_covr <- function(expr) {
  .strip_covr_counts(paste(deparse(expr), collapse = "\n"))
}

build_code_sb_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  meta <- spec$meta
  plan <- spec$plan
  kinfo <- spec$kernel_info %||% list()
  sigs <- spec$signatures %||% list()

  if (!identical(meta$backend, "sb")) stop("spec backend is not 'sb'.", call. = FALSE)

  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  K <- as.integer(meta$components)
  has_X <- isTRUE(meta$has_X)
  has_ps <- !is.null(plan$ps)
  cluster_gating <- isTRUE((spec$cluster %||% list())$gating)
  default_ps_prior <- list(dist = "normal", args = list(mean = 0, sd = 2))
  ps_prior <- (plan$ps %||% list(prior = default_ps_prior))$prior
  if (isTRUE(cluster_gating) && !isTRUE(has_X)) {
    stop("Cluster gating for SB requires X covariates.", call. = FALSE)
  }
  if (isTRUE(cluster_gating) && P < 1L) {
    stop("Cluster gating for SB requires P > 0.", call. = FALSE)
  }

  # ---- resolve single-component likelihood signature (uses latent z) ----
  dist_name <- NULL
  arg_order <- NULL
  if (isTRUE(meta$GPD)) {
    dist_name <- kinfo$crp$d_gpd %||% NULL
    arg_order <- kinfo$crp$args_gpd %||% NULL
    if (is.null(dist_name) || isTRUE(is.na(dist_name))) {
      dist_name <- kinfo$sb$d_gpd %||% NULL
      if (!is.null(dist_name)) dist_name <- sub("Mix", "", dist_name, fixed = TRUE)
    }
    if (is.null(arg_order) || anyNA(arg_order)) {
      arg_order <- kinfo$sb$args_gpd %||% NULL
    }
  } else {
    dist_name <- kinfo$crp$d_base %||% NULL
    arg_order <- kinfo$bulk_params %||% NULL
  }
  if (is.null(dist_name) || isTRUE(is.na(dist_name))) {
    stop("Missing SB single-component likelihood in kernel registry.", call. = FALSE)
  }
  if (is.null(arg_order) || !length(arg_order)) {
    stop("Missing SB likelihood argument order.", call. = FALSE)
  }
  if (any(arg_order == "w")) arg_order <- setdiff(arg_order, "w")

  bulk_plan <- plan$bulk %||% list()
  bulk_params <- kinfo$bulk_params %||% names(bulk_plan)
  bulk_link <- any(vapply(bulk_params, function(nm) {
    ent <- bulk_plan[[nm]]
    identical(ent$mode %||% NA_character_, "link")
  }, logical(1)))
  default_ps_prior <- list(dist = "normal", args = list(mean = 0, sd = 2))
  ps_prior <- (plan$ps %||% list(prior = default_ps_prior))$prior

  # ---- build nimbleCode ----
  nimble::nimbleCode({

    # --- concentration ---
    CONC_PLACEHOLDER()

    # --- mixing weights ---
    MIXING_BLOCK()

    # --- bulk parameters (component-level + betas) ---
    BULK_BLOCK()

    # --- beta blocks for link-mode bulk params ---
    HASX_BETA_BLOCK()

    # --- build per-(i,j) linked bulk params as deterministic arrays ---
    HASX_DET_BLOCK()

    # --- GPD tail nodes (if requested) ---
    GPD_BLOCK()

    # --- likelihood ---
    LIKELIHOOD_BLOCK()
  }) -> code

  # ---- Now patch the code body programmatically (no placeholders left) ----
  # Convert nimbleCode to text, inject lines, then re-parse into nimbleCode.
  txt <- .deparse_without_covr(code)
  txt <- .strip_covr_counts(txt)

  inject <- function(pattern, replacement) {
    txt <<- sub(pattern, replacement, txt)
  }

  # (0) concentration
  conc <- plan$concentration %||% list()
  conc_line <- if (identical(conc$mode, "fixed")) {
    sprintf("alpha <- %s", deparse1(conc$value))
  } else if (identical(conc$mode, "dist")) {
    sprintf("alpha ~ %s", .codegen_prior_call(conc$dist, conc$args, backend = "SB"))
  } else {
    stop("Invalid plan$concentration$mode.", call. = FALSE)
  }
  inject("CONC_PLACEHOLDER\\(\\)", paste0(conc_line, "\n"))

  # (0b) SB global weights vs cluster-gating weights
  mixing_block <- if (isTRUE(cluster_gating)) {
    gating_eta <- if (P == 1L) {
      "eta[j] + X[i, 1] * B[j, 1]"
    } else {
      "eta[j] + inprod(X[i, 1:P], B[j, 1:P])"
    }
    paste0(
      "for (j in 1:(components - 1)) {\n",
      "      eta[j] ~ dnorm(0, sd = 2)\n",
      if (P == 1L) {
        "      B[j, 1] ~ dnorm(0, sd = 1)\n"
      } else {
        "      for (p in 1:P) B[j, p] ~ dnorm(0, sd = 1)\n"
      },
      "    }\n",
      "    for (i in 1:N) {\n",
      "      for (j in 1:(components - 1)) {\n",
      "        logit_ij[i, j] <- ", gating_eta, "\n",
      "      }\n",
      "      logit_ij[i, components] <- 0\n",
      "      logit_max[i] <- max(logit_ij[i, 1:components])\n",
      "      for (j in 1:components) {\n",
      "        exp_logit_ij[i, j] <- exp(logit_ij[i, j] - logit_max[i])\n",
      "      }\n",
      "      logit_denom[i] <- sum(exp_logit_ij[i, 1:components])\n",
      "      for (j in 1:components) {\n",
      "        w_x[i, j] <- exp_logit_ij[i, j] / logit_denom[i]\n",
      "      }\n",
      "    }\n"
    )
  } else {
    paste0(
      "for (j in 1:(components - 1)) {\n",
      "      v[j] ~ dbeta(1, alpha)\n",
      "    }\n",
      "    stick_mass[1] <- 1\n",
      "    for (j in 2:components) {\n",
      "      stick_mass[j] <- stick_mass[j - 1] * (1 - v[j - 1])\n",
      "    }\n",
      "    for (j in 1:(components - 1)) {\n",
      "      w[j] <- v[j] * stick_mass[j]\n",
      "    }\n",
      "    w[components] <- stick_mass[components]\n"
    )
  }
  inject("MIXING_BLOCK\\(\\)", mixing_block)

  # (1) Bulk param declarations (component-level)
  bulk_decl_lines <- character()
  for (nm in bulk_params) {
    ent <- bulk_plan[[nm]]
    mode <- ent$mode %||% NA_character_
    if (mode == "fixed") {
      bulk_decl_lines <- c(bulk_decl_lines, sprintf("%s[j] <- %s", nm, deparse1(ent$value)))
    } else if (mode == "dist") {
      bulk_decl_lines <- c(bulk_decl_lines, sprintf("%s[j] ~ %s", nm,
                                                    .codegen_prior_call(ent$dist, ent$args, backend = "SB")))
    } else if (mode == "link") {
      # link-mode params are represented via beta_<param>; no component prior node.
    } else {
      stop(sprintf("Invalid bulk mode for '%s'.", nm), call. = FALSE)
    }
  }
  bulk_block <- if (length(bulk_decl_lines)) {
    paste0(
      "for (j in 1:components) {\n",
      "      ", paste(bulk_decl_lines, collapse = "\n      "), "\n",
      "    }\n"
    )
  } else {
    ""
  }
  inject("BULK_BLOCK\\(\\)", bulk_block)

  # (2) Beta priors for link-mode bulk params
  beta_lines <- character()
  if (has_X) {
    for (nm in bulk_params) {
      ent <- bulk_plan[[nm]]
      if (identical(ent$mode, "link")) {
        bp <- ent$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 2))
        if (bp$dist != "normal") stop("Only normal beta priors are supported by default for link-mode betas.", call. = FALSE)
        m <- bp$args$mean %||% 0
        s <- bp$args$sd %||% 2
        beta_lines <- c(beta_lines, sprintf("for (p in 1:P) beta_%s[j, p] ~ dnorm(%s, sd = %s)", nm, deparse1(m), deparse1(s)))
        if (has_ps) {
          if (!identical(ps_prior$dist, "normal")) {
            stop("beta_ps priors must be normal for SB codegen.", call. = FALSE)
          }
          m_ps <- ps_prior$args$mean %||% 0
          s_ps <- ps_prior$args$sd %||% 2
          beta_lines <- c(beta_lines, sprintf("beta_ps_%s[j] ~ dnorm(%s, sd = %s)", nm, deparse1(m_ps), deparse1(s_ps)))
        }
      }
    }
  }
  beta_block <- if (has_X && length(beta_lines)) {
    paste0("for (j in 1:components) {\n",
           "      ", paste(beta_lines, collapse = "\n      "), "\n",
           "    }\n")
  } else {
    ""
  }
  inject("HASX_BETA_BLOCK\\(\\)", beta_block)

  # (3) Deterministic linked bulk param_ij
  det_lines <- character()
  if (has_X) {
    for (nm in bulk_params) {
      ent <- bulk_plan[[nm]]
      if (identical(ent$mode, "link")) {
        eta_terms <- character()
        if (P == 1L) {
          eta_terms <- c(eta_terms, sprintf("X[i, 1] * beta_%s[j, 1]", nm))
        } else {
          eta_terms <- c(eta_terms, sprintf("inprod(X[i, 1:P], beta_%s[j, 1:P])", nm))
        }
        if (has_ps) {
          eta_terms <- c(eta_terms, sprintf("ps[i] * beta_ps_%s[j]", nm))
        }
        if (!length(eta_terms)) stop(sprintf("Unable to build eta for '%s'.", nm), call. = FALSE)
        eta <- paste(eta_terms, collapse = " + ")
        expr <- .codegen_link_expr(eta, ent$link, ent$link_power)
        det_lines <- c(det_lines, sprintf("%s_ij[i, j] <- %s", nm, expr))
      }
    }
  }
  det_block <- if (has_X && length(det_lines)) {
    paste0("for (i in 1:N) {\n",
           "      for (j in 1:components) {\n",
           "        ", paste(det_lines, collapse = "\n        "), "\n",
           "      }\n",
           "    }\n")
  } else {
    ""
  }
  inject("HASX_DET_BLOCK\\(\\)", det_block)

  # (4) GPD blocks
  gpd_lines <- character()
  if (isTRUE(meta$GPD)) {
    gpd <- plan$gpd %||% list()

    # threshold
    thr <- gpd$threshold %||% NULL
    if (!is.null(thr)) {
      thr_scalar <- thr$mode %in% c("fixed", "dist")
      if (thr$mode == "fixed") {
        if (thr_scalar) {
          gpd_lines <- c(gpd_lines, sprintf("threshold <- %s", deparse1(thr$value)))
        } else {
          gpd_lines <- c(gpd_lines, sprintf("for (i in 1:N) threshold[i] <- %s", deparse1(thr$value)))
        }
      } else if (thr$mode == "dist") {
        if (thr_scalar) {
          gpd_lines <- c(gpd_lines, sprintf("threshold ~ %s",
                                            .codegen_prior_call(thr$dist, thr$args, backend = "SB")))
        } else {
          gpd_lines <- c(gpd_lines, sprintf("for (i in 1:N) threshold[i] ~ %s",
                                            .codegen_prior_call(thr$dist, thr$args, backend = "SB")))
        }
      } else if (thr$mode == "link") {
        if (!has_X) stop("threshold link-mode requires X.", call. = FALSE)
        bp <- thr$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
        m <- bp$args$mean %||% 0
        s <- bp$args$sd %||% 0.2
        gpd_lines <- c(gpd_lines, sprintf("for (p in 1:P) beta_threshold[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))

        if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
          sdlog_u <- gpd$sdlog_u %||% list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
          if (!identical(sdlog_u$mode, "dist")) stop("sdlog_u must be dist-mode under lognormal threshold.", call. = FALSE)
          gpd_lines <- c(gpd_lines, sprintf("sdlog_u ~ %s",
                                            .codegen_prior_call(sdlog_u$dist, sdlog_u$args, backend = "SB")))
          eta_u_line <- if (P == 1L) "  eta_u[i] <- X[i, 1] * beta_threshold[1]" else
            "  eta_u[i] <- inprod(X[i, 1:P], beta_threshold[1:P])"
          gpd_lines <- c(gpd_lines, "for (i in 1:N) {",
                         eta_u_line,
                         "  threshold[i] ~ dlnorm(meanlog = eta_u[i], sdlog = sdlog_u)",
                         "}")
        } else {
          eta_u_line <- if (P == 1L) "  eta_u[i] <- X[i, 1] * beta_threshold[1]" else
            "  eta_u[i] <- inprod(X[i, 1:P], beta_threshold[1:P])"
          gpd_lines <- c(gpd_lines, "for (i in 1:N) {",
                         eta_u_line,
                         sprintf("  threshold[i] <- %s",
                                 .codegen_link_expr("eta_u[i]", thr$link, thr$link_power)),
                         "}")
        }
      } else {
        stop("Invalid gpd threshold mode.", call. = FALSE)
      }
    }

    # tail_scale
    ts <- gpd$tail_scale %||% NULL
    if (!is.null(ts)) {
      if (ts$mode == "fixed") {
        gpd_lines <- c(gpd_lines, sprintf("tail_scale <- %s", deparse1(ts$value)))
      } else if (ts$mode == "dist") {
        gpd_lines <- c(gpd_lines, sprintf("tail_scale ~ %s",
                                          .codegen_prior_call(ts$dist, ts$args, backend = "SB")))
      } else if (ts$mode == "link") {
        if (!has_X) stop("tail_scale link-mode requires X.", call. = FALSE)
        bp <- ts$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.5))
        m <- bp$args$mean %||% 0
        s <- bp$args$sd %||% 0.5
        gpd_lines <- c(gpd_lines, sprintf("for (p in 1:P) beta_tail_scale[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
        eta_ts_line <- if (P == 1L) "  eta_ts[i] <- X[i, 1] * beta_tail_scale[1]" else
          "  eta_ts[i] <- inprod(X[i, 1:P], beta_tail_scale[1:P])"
        gpd_lines <- c(gpd_lines, "for (i in 1:N) {",
                       eta_ts_line,
                       "  tail_scale[i] <- exp(eta_ts[i])",
                       "}")
      } else {
        stop("Invalid gpd tail_scale mode.", call. = FALSE)
      }
    }

    # tail_shape
    tsh <- gpd$tail_shape %||% NULL
    if (!is.null(tsh)) {
      if (tsh$mode == "fixed") {
        gpd_lines <- c(gpd_lines, sprintf("tail_shape <- %s", deparse1(tsh$value)))
      } else if (tsh$mode == "dist") {
        gpd_lines <- c(gpd_lines, sprintf("tail_shape ~ %s",
                                          .codegen_prior_call(tsh$dist, tsh$args, backend = "SB")))
      } else if (tsh$mode == "link") {
        if (!has_X) stop("tail_shape link-mode requires X.", call. = FALSE)
        bp <- tsh$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.3))
        m <- bp$args$mean %||% 0
        s <- bp$args$sd %||% 0.3
        gpd_lines <- c(gpd_lines, sprintf("for (p in 1:P) beta_tail_shape[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
        eta_tsh_line <- if (P == 1L) "  eta_tsh[i] <- X[i, 1] * beta_tail_shape[1]" else
          "  eta_tsh[i] <- inprod(X[i, 1:P], beta_tail_shape[1:P])"
        link_expr <- .codegen_link_expr("eta_tsh[i]", tsh$link %||% "identity", tsh$link_power)
        gpd_lines <- c(gpd_lines, "for (i in 1:N) {",
                       eta_tsh_line,
                       sprintf("  tail_shape[i] <- %s", link_expr),
                       "}")
      } else {
        stop("Invalid gpd tail_shape mode.", call. = FALSE)
      }
    }
  }

  inject("GPD_BLOCK\\(\\)",
         if (length(gpd_lines)) paste0(paste(gpd_lines, collapse = "\n    "), "\n") else "")

  # (5) Likelihood call
  like_lines <- character()
  gpd_for_args <- plan$gpd %||% list()
  thr_for_args <- gpd_for_args$threshold %||% NULL
  thr_scalar <- !is.null(thr_for_args) && thr_for_args$mode %in% c("fixed", "dist")
  args_expr <- character()
  for (a in arg_order) {
    if (a %in% bulk_params) {
      ent <- bulk_plan[[a]]
      if (identical(ent$mode, "link")) {
        args_expr <- c(args_expr, sprintf("%s_ij[i, z[i]]", a))
      } else {
        args_expr <- c(args_expr, sprintf("%s[z[i]]", a))
      }
    } else if (a == "threshold") {
      args_expr <- c(args_expr, if (thr_scalar) "threshold" else "threshold[i]")
    } else if (a == "tail_scale") {
      ts <- plan$gpd$tail_scale %||% NULL
      if (!is.null(ts) && identical(ts$mode, "link")) args_expr <- c(args_expr, "tail_scale[i]") else args_expr <- c(args_expr, "tail_scale")
    } else if (a == "tail_shape") {
      tsh <- plan$gpd$tail_shape %||% NULL
      if (!is.null(tsh) && identical(tsh$mode, "link")) args_expr <- c(args_expr, "tail_shape[i]") else args_expr <- c(args_expr, "tail_shape")
    } else {
      stop(sprintf("Unknown argument '%s' in SB signature for kernel '%s'.", a, meta$kernel), call. = FALSE)
    }
  }

  alloc_prob <- if (isTRUE(cluster_gating)) "w_x[i, 1:components]" else "w[1:components]"
  like_lines <- c(
    sprintf("z[i] ~ dcat(prob = %s)", alloc_prob),
    sprintf("y[i] ~ %s(%s)", dist_name, paste(args_expr, collapse = ", "))
  )
  like_block <- paste0("for (i in 1:N) {\n",
                       "      ", paste(like_lines, collapse = "\n      "), "\n",
                       "    }\n")
  inject("LIKELIHOOD_BLOCK\\(\\)", like_block)

  # Rebuild nimbleCode from modified text without evaluating model symbols.
  expr <- parse(text = txt)[[1]]

  # IMPORTANT: nimbleCode uses NSE; use do.call to pass evaluated expression, not a captured call.
  do.call(nimble::nimbleCode, list(expr))

}


#' Build NIMBLE code for CRP backend from a compiled spec
#'
#' Generates \code{nimbleCode} for the Chinese Restaurant Process (CRP) backend
#' using native NIMBLE BNP distribution:
#' \itemize{
#'   \item memberships: \code{z[1:N] ~ dCRP(conc = alpha, size = N)}
#' }
#'
#' The finite represented number of clusters is controlled by a single parameter
#' \code{components}. Component-specific parameters are declared for
#' \code{k = 1:components}.
#'
#' Bulk parameters follow the tri-mode plan:
#' \itemize{
#'   \item fixed: \code{param[k] <- value}
#'   \item dist:  \code{param[k] ~ d<prior>(...)}
#'   \item link:  \code{beta_param[k,p] ~ dnorm(...)} and
#'         \code{param_ik[i,k] <- g(inprod(X[i,], beta_param[k,]))}
#' }
#'
#' If \code{GPD=TRUE}, the default tail under \code{X} is:
#' \itemize{
#'   \item \code{threshold[i] ~ dlnorm(meanlog = inprod(X[i,], beta_threshold), sdlog = sdlog_u)}
#'   \item \code{tail_scale[i] <- exp(inprod(X[i,], beta_tail_scale))}
#'   \item \code{tail_shape ~ dnorm(0, sd = 0.2)} (unless fixed)
#' }
#'
#' Likelihood calls are emitted using positional arguments only (no names).
#'
#' @param spec A compiled model specification from \code{compile_model_spec()}.
#' @return A \code{nimbleCode} object.
#' @keywords internal
#' @noRd
build_code_crp_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  meta <- spec$meta
  plan <- spec$plan
  kinfo <- spec$kernel_info %||% list()
  sigs <- spec$signatures %||% list()

  if (!meta$backend %in% c("crp", "spliced")) {
    stop("spec backend must be 'crp' or 'spliced'.", call. = FALSE)
  }
  is_spliced <- identical(meta$backend, "spliced")

  N <- as.integer(meta$N)
  P <- as.integer(meta$P %||% 0L)
  K <- as.integer(meta$components)
  has_X <- isTRUE(meta$has_X)
  has_ps <- !is.null(plan$ps)
  default_ps_prior <- list(dist = "normal", args = list(mean = 0, sd = 2))
  ps_prior <- (plan$ps %||% list(prior = default_ps_prior))$prior

  # ---- resolve likelihood signature ----
  dist_name <- NULL
  arg_order <- NULL
  if (isTRUE(meta$GPD)) {
    dist_name <- sigs$gpd$dist_name %||% NULL
    arg_order <- sigs$gpd$args %||% NULL
  } else {
    dist_name <- sigs$bulk$dist_name %||% NULL
    arg_order <- sigs$bulk$args %||% NULL
  }
  if (is.null(dist_name) || is.null(arg_order) || !length(arg_order)) {
    stop("Missing CRP likelihood signature in spec$signatures.", call. = FALSE)
  }

  bulk_plan <- plan$bulk %||% list()
  bulk_params <- kinfo$bulk_params %||% names(bulk_plan)
  bulk_link <- any(vapply(bulk_params, function(nm) {
    ent <- bulk_plan[[nm]]
    identical(ent$mode %||% NA_character_, "link")
  }, logical(1)))

  # ---- assemble model code as text ----
  lines <- character()
  add <- function(...) lines <<- c(lines, ...)

  # concentration
  conc <- plan$concentration %||% list()
  if (identical(conc$mode, "fixed")) {
    add(sprintf("  alpha <- %s", deparse1(conc$value)))
  } else if (identical(conc$mode, "dist")) {
    add(sprintf("  alpha ~ %s", .codegen_prior_call(conc$dist, conc$args, backend = "CRP")))
  } else {
    stop("Invalid plan$concentration$mode.", call. = FALSE)
  }

  # CRP memberships
  add("  z[1:N] ~ dCRP(conc = alpha, size = N)")

  # bulk component-level declarations
  bulk_decl_lines <- character()
  for (nm in bulk_params) {
    ent <- bulk_plan[[nm]]
    mode <- ent$mode %||% NA_character_

    if (mode == "fixed") {
      bulk_decl_lines <- c(bulk_decl_lines, sprintf("    %s[k] <- %s", nm, deparse1(ent$value)))
    } else if (mode == "dist") {
      bulk_decl_lines <- c(
        bulk_decl_lines,
        sprintf("    %s[k] ~ %s", nm, .codegen_prior_call(ent$dist, ent$args, backend = "CRP"))
      )
    } else if (mode == "link") {
      # link-mode params are represented via beta_<param>; no component prior node.
    } else {
      stop(sprintf("Invalid bulk mode for '%s'.", nm), call. = FALSE)
    }
  }
  if (length(bulk_decl_lines)) {
    add("  for (k in 1:components) {")
    add(bulk_decl_lines)
    add("  }")
  }

  # beta priors for link-mode bulk params
  if (has_X) {
    for (nm in bulk_params) {
      ent <- bulk_plan[[nm]]
      if (identical(ent$mode, "link")) {
        bp <- ent$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 2))
        if (!identical(bp$dist, "normal")) {
          stop("Only normal beta priors are supported by default for link-mode betas.", call. = FALSE)
        }
        m <- bp$args$mean %||% 0
        s <- bp$args$sd %||% 2
        add("  for (k in 1:components) {")
        add(sprintf("    for (p in 1:P) beta_%s[k, p] ~ dnorm(%s, sd = %s)", nm, deparse1(m), deparse1(s)))
        if (has_ps) {
          if (!identical(ps_prior$dist, "normal")) {
            stop("beta_ps priors must be normal for CRP codegen.", call. = FALSE)
          }
          m_ps <- ps_prior$args$mean %||% 0
          s_ps <- ps_prior$args$sd %||% 2
          add(sprintf("    beta_ps_%s[k] ~ dnorm(%s, sd = %s)", nm, deparse1(m_ps), deparse1(s_ps)))
        }
        add("  }")
      }
    }

    # CRP link-mode parameters are applied directly in the likelihood expression.
  }

  # ---- GPD tail blocks ----
  if (isTRUE(meta$GPD)) {
    gpd <- plan$gpd %||% list()

    if (is_spliced) {
      # ======== SPLICED BACKEND: Component-level GPD parameterization ========
      # Generate component-level nodes for fixed/dist modes, or component-specific
      # beta coefficients + observation-level deterministic nodes for link mode.

      # threshold
      thr <- gpd$threshold %||% NULL
      if (!is.null(thr)) {
        if (thr$mode == "fixed") {
          add("  for (k in 1:components) {")
          add(sprintf("    threshold[k] <- %s", deparse1(thr$value)))
          add("  }")
        } else if (thr$mode == "dist") {
          add("  for (k in 1:components) {")
          add(sprintf("    threshold[k] ~ %s",
                      .codegen_prior_call(thr$dist, thr$args, backend = "CRP")))
          add("  }")
        } else if (thr$mode == "link") {
          if (!has_X) stop("threshold link-mode requires X.", call. = FALSE)
          bp <- thr$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
          if (!identical(bp$dist, "normal")) stop("beta_threshold prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.2
          add("  for (k in 1:components) {")
          add(sprintf("    for (p in 1:P) beta_threshold[k, p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
          add("  }")
          if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
            sdlog_u <- gpd$sdlog_u %||% list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
            if (!identical(sdlog_u$mode, "dist")) stop("sdlog_u must be dist-mode under lognormal threshold.", call. = FALSE)
            add(sprintf("  sdlog_u ~ %s",
                        .codegen_prior_call(sdlog_u$dist, sdlog_u$args, backend = "CRP")))
            add("  for (i in 1:N) {")
            add("    for (k in 1:components) {")
            if (P == 1L) {
              add("      eta_threshold[i, k] <- X[i, 1] * beta_threshold[k, 1]")
            } else {
              add("      eta_threshold[i, k] <- inprod(X[i, 1:P], beta_threshold[k, 1:P])")
            }
            add("      threshold_i[i, k] ~ dlnorm(meanlog = eta_threshold[i, k], sdlog = sdlog_u)")
            add("    }")
            add("  }")
          } else {
            add("  for (i in 1:N) {")
            if (P == 1L) {
              add("    eta_threshold[i] <- X[i, 1] * beta_threshold[z[i], 1]")
            } else {
              add("    eta_threshold[i] <- inprod(X[i, 1:P], beta_threshold[z[i], 1:P])")
            }
            add(sprintf("    threshold_i[i] <- %s", .codegen_link_expr("eta_threshold[i]", thr$link, thr$link_power)))
            add("  }")
          }
        } else {
          stop("Invalid gpd threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      ts <- gpd$tail_scale %||% NULL
      if (!is.null(ts)) {
        if (ts$mode == "fixed") {
          add("  for (k in 1:components) {")
          add(sprintf("    tail_scale[k] <- %s", deparse1(ts$value)))
          add("  }")
        } else if (ts$mode == "dist") {
          add("  for (k in 1:components) {")
          add(sprintf("    tail_scale[k] ~ %s",
                      .codegen_prior_call(ts$dist, ts$args, backend = "CRP")))
          add("  }")
        } else if (ts$mode == "link") {
          if (!has_X) stop("tail_scale link-mode requires X.", call. = FALSE)
          bp <- ts$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.5))
          if (!identical(bp$dist, "normal")) stop("beta_tail_scale prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.5
          add("  for (k in 1:components) {")
          add(sprintf("    for (p in 1:P) beta_tail_scale[k, p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
          add("  }")
          add("  for (i in 1:N) {")
          if (P == 1L) {
            add("    eta_tail_scale[i] <- X[i, 1] * beta_tail_scale[z[i], 1]")
          } else {
            add("    eta_tail_scale[i] <- inprod(X[i, 1:P], beta_tail_scale[z[i], 1:P])")
          }
          add(sprintf("    tail_scale_i[i] <- %s", .codegen_link_expr("eta_tail_scale[i]", ts$link, ts$link_power)))
          add("  }")
        } else {
          stop("Invalid gpd tail_scale mode.", call. = FALSE)
        }
      }

      # tail_shape
      tsh <- gpd$tail_shape %||% NULL
      if (!is.null(tsh)) {
        if (tsh$mode == "fixed") {
          add("  for (k in 1:components) {")
          add(sprintf("    tail_shape[k] <- %s", deparse1(tsh$value)))
          add("  }")
        } else if (tsh$mode == "dist") {
          add("  for (k in 1:components) {")
          add(sprintf("    tail_shape[k] ~ %s",
                      .codegen_prior_call(tsh$dist, tsh$args, backend = "CRP")))
          add("  }")
        } else if (tsh$mode == "link") {
          if (!has_X) stop("tail_shape link-mode requires X.", call. = FALSE)
          bp <- tsh$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
          if (!identical(bp$dist, "normal")) stop("beta_tail_shape prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.2
          add("  for (k in 1:components) {")
          add(sprintf("    for (p in 1:P) beta_tail_shape[k, p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
          add("  }")
          add("  for (i in 1:N) {")
          if (P == 1L) {
            add("    eta_tail_shape[i] <- X[i, 1] * beta_tail_shape[z[i], 1]")
          } else {
            add("    eta_tail_shape[i] <- inprod(X[i, 1:P], beta_tail_shape[z[i], 1:P])")
          }
          add(sprintf("    tail_shape_i[i] <- %s", .codegen_link_expr("eta_tail_shape[i]", tsh$link, tsh$link_power)))
          add("  }")
        } else {
          stop("Invalid gpd tail_shape mode.", call. = FALSE)
        }
      }

    } else {
      # ======== STANDARD CRP BACKEND: Observation-level GPD parameterization ========
      # (Original behavior preserved for backward compatibility)

      # threshold
      thr <- gpd$threshold %||% NULL
      if (!is.null(thr)) {
        thr_scalar <- thr$mode %in% c("fixed", "dist")
        if (thr$mode == "fixed") {
          if (thr_scalar) {
            add(sprintf("  threshold <- %s", deparse1(thr$value)))
          } else {
            add(sprintf("  for (i in 1:N) threshold[i] <- %s", deparse1(thr$value)))
          }
        } else if (thr$mode == "dist") {
          if (thr_scalar) {
            add(sprintf("  threshold ~ %s",
                        .codegen_prior_call(thr$dist, thr$args, backend = "CRP")))
          } else {
            add(sprintf("  for (i in 1:N) threshold[i] ~ %s",
                        .codegen_prior_call(thr$dist, thr$args, backend = "CRP")))
          }
        } else if (thr$mode == "link") {
          if (!has_X) stop("threshold link-mode requires X.", call. = FALSE)
          bp <- thr$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
          if (!identical(bp$dist, "normal")) stop("beta_threshold prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.2
          add(sprintf("  for (p in 1:P) beta_threshold[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))

          if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
            sdlog_u <- gpd$sdlog_u %||% list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
            if (!identical(sdlog_u$mode, "dist")) stop("sdlog_u must be dist-mode under lognormal threshold.", call. = FALSE)
            add(sprintf("  sdlog_u ~ %s",
                        .codegen_prior_call(sdlog_u$dist, sdlog_u$args, backend = "CRP")))
            add("  for (i in 1:N) {")
            if (P == 1L) {
              add("    eta_u[i] <- X[i, 1] * beta_threshold[1]")
            } else {
              add("    eta_u[i] <- inprod(X[i, 1:P], beta_threshold[1:P])")
            }
            add("    threshold[i] ~ dlnorm(meanlog = eta_u[i], sdlog = sdlog_u)")
            add("  }")
          } else {
            add("  for (i in 1:N) {")
            if (P == 1L) {
              add("    eta_u[i] <- X[i, 1] * beta_threshold[1]")
            } else {
              add("    eta_u[i] <- inprod(X[i, 1:P], beta_threshold[1:P])")
            }
            add(sprintf("    threshold[i] <- %s", .codegen_link_expr("eta_u[i]", thr$link, thr$link_power)))
            add("  }")
          }
        } else {
          stop("Invalid gpd threshold mode.", call. = FALSE)
        }
      }

      # tail_scale
      ts <- gpd$tail_scale %||% NULL
      if (!is.null(ts)) {
        if (ts$mode == "fixed") {
          add(sprintf("  tail_scale <- %s", deparse1(ts$value)))
        } else if (ts$mode == "dist") {
          add(sprintf("  tail_scale ~ %s",
                      .codegen_prior_call(ts$dist, ts$args, backend = "CRP")))
        } else if (ts$mode == "link") {
          if (!has_X) stop("tail_scale link-mode requires X.", call. = FALSE)
          bp <- ts$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.5))
          if (!identical(bp$dist, "normal")) stop("beta_tail_scale prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.5
          add(sprintf("  for (p in 1:P) beta_tail_scale[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
          add("  for (i in 1:N) {")
          if (P == 1L) {
            add("    eta_ts[i] <- X[i, 1] * beta_tail_scale[1]")
          } else {
            add("    eta_ts[i] <- inprod(X[i, 1:P], beta_tail_scale[1:P])")
          }
          add("    tail_scale[i] <- exp(eta_ts[i])")
          add("  }")
        } else {
          stop("Invalid gpd tail_scale mode.", call. = FALSE)
        }
      }

      # tail_shape
      tsh <- gpd$tail_shape %||% NULL
      if (!is.null(tsh)) {
        if (tsh$mode == "fixed") {
          add(sprintf("  tail_shape <- %s", deparse1(tsh$value)))
        } else if (tsh$mode == "dist") {
          add(sprintf("  tail_shape ~ %s",
                      .codegen_prior_call(tsh$dist, tsh$args, backend = "CRP")))
        } else if (tsh$mode == "link") {
          if (!has_X) stop("tail_shape link-mode requires X.", call. = FALSE)
          bp <- tsh$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
          if (!identical(bp$dist, "normal")) stop("beta_tail_shape prior must be normal.", call. = FALSE)
          m <- bp$args$mean %||% 0
          s <- bp$args$sd %||% 0.2
          add(sprintf("  for (p in 1:P) beta_tail_shape[p] ~ dnorm(%s, sd = %s)", deparse1(m), deparse1(s)))
          add("  for (i in 1:N) {")
          if (P == 1L) {
            add("    eta_ts[i] <- X[i, 1] * beta_tail_shape[1]")
          } else {
            add("    eta_ts[i] <- inprod(X[i, 1:P], beta_tail_shape[1:P])")
          }
          add(sprintf("    tail_shape[i] <- %s", .codegen_link_expr("eta_ts[i]", tsh$link, tsh$link_power)))
          add("  }")
        } else {
          stop("Invalid gpd tail_shape mode.", call. = FALSE)
        }
      }
    }
  }

  # ---- likelihood ----
  add("  for (i in 1:N) {")

  # build arg expressions in signature order
  gpd_for_args <- plan$gpd %||% list()
  thr_for_args <- gpd_for_args$threshold %||% NULL
  ts_for_args <- gpd_for_args$tail_scale %||% NULL
  tsh_for_args <- gpd_for_args$tail_shape %||% NULL
  
  args_expr <- character()
  for (a in arg_order) {
    if (a %in% bulk_params) {
      ent <- bulk_plan[[a]]
      if (identical(ent$mode, "link")) {
        eta_terms <- character()
        if (P == 1L) {
          eta_terms <- c(eta_terms, sprintf("X[i, 1] * beta_%s[z[i], 1]", a))
        } else {
          eta_terms <- c(eta_terms, sprintf("inprod(X[i, 1:P], beta_%s[z[i], 1:P])", a))
        }
        if (has_ps) {
          eta_terms <- c(eta_terms, sprintf("ps[i] * beta_ps_%s[z[i]]", a))
        }
        if (!length(eta_terms)) {
          stop(sprintf("Unable to build eta for '%s'.", a), call. = FALSE)
        }
        eta <- paste(eta_terms, collapse = " + ")
        args_expr <- c(args_expr, .codegen_link_expr(eta, ent$link, ent$link_power))
      } else {
        args_expr <- c(args_expr, sprintf("%s[z[i]]", a))
      }
    } else if (a == "threshold") {
      if (is_spliced) {
        # Spliced backend: link mode uses threshold_i[i], others use threshold[z[i]]
        if (!is.null(thr_for_args) && identical(thr_for_args$mode, "link")) {
          if (!is.null(thr_for_args$link_dist) &&
              identical(thr_for_args$link_dist$dist, "lognormal")) {
            args_expr <- c(args_expr, "threshold_i[i, z[i]]")
          } else {
            args_expr <- c(args_expr, "threshold_i[i]")
          }
        } else {
          args_expr <- c(args_expr, "threshold[z[i]]")
        }
      } else {
        # Standard CRP backend: scalar or observation-level
        thr_scalar <- !is.null(thr_for_args) && thr_for_args$mode %in% c("fixed", "dist")
        args_expr <- c(args_expr, if (thr_scalar) "threshold" else "threshold[i]")
      }
    } else if (a == "tail_scale") {
      if (is_spliced) {
        # Spliced backend: link mode uses tail_scale_i[i], others use tail_scale[z[i]]
        if (!is.null(ts_for_args) && identical(ts_for_args$mode, "link")) {
          args_expr <- c(args_expr, "tail_scale_i[i]")
        } else {
          args_expr <- c(args_expr, "tail_scale[z[i]]")
        }
      } else {
        # Standard CRP backend: scalar or observation-level
        if (!is.null(ts_for_args) && identical(ts_for_args$mode, "link")) {
          args_expr <- c(args_expr, "tail_scale[i]")
        } else {
          args_expr <- c(args_expr, "tail_scale")
        }
      }
    } else if (a == "tail_shape") {
      if (is_spliced) {
        # Spliced backend: link mode uses tail_shape_i[i], others use tail_shape[z[i]]
        if (!is.null(tsh_for_args) && identical(tsh_for_args$mode, "link")) {
          args_expr <- c(args_expr, "tail_shape_i[i]")
        } else {
          args_expr <- c(args_expr, "tail_shape[z[i]]")
        }
      } else {
        # Standard CRP backend: scalar only (no link mode support in original)
        args_expr <- c(args_expr, "tail_shape")
      }
    } else {
      stop(sprintf("Unknown argument '%s' in CRP signature for kernel '%s'.", a, meta$kernel), call. = FALSE)
    }
  }

  add(sprintf("    y[i] ~ %s(%s)", dist_name, paste(args_expr, collapse = ", ")))
  add("  }")

  # parse into nimbleCode
  code_str <- paste(lines, collapse = "\n")
  expr <- parse(text = paste0("{\n", code_str, "\n}"))[[1]]
  nimble::nimbleCode(expr)
}
# nocov end


#' Build a prior/parameter specification table from a compiled model spec
#'
#' Creates a human-readable table describing how each parameter is modeled:
#' fixed value, prior distribution (no regression), or regression/link (with beta prior),
#' including the special case where a linked parameter is stochastic around the link
#' (e.g., `threshold[i] ~ Lognormal(meanlog = X %*% beta, sdlog = sdlog_u)`).
#'
#' This is purely descriptive and is used by bundle-level summaries.
#'
#' @param spec A compiled model specification produced by \code{compile_model_spec()}.
#' @return A data.frame with columns describing each parameter block.
#' @keywords internal
#' @noRd
build_prior_table_from_spec <- function(spec) {
  stopifnot(is.list(spec), !is.null(spec$meta), !is.null(spec$plan))


  meta <- spec$meta
  plan <- spec$plan

  backend <- meta$backend
  kernel  <- meta$kernel
  GPD     <- isTRUE(meta$GPD)
  has_X   <- isTRUE(meta$has_X)
  N       <- as.integer(meta$N)
  P       <- as.integer(meta$P %||% 0L)
  K       <- as.integer(meta$components)

  fmt_args <- function(args) {
    if (is.null(args) || !length(args)) return("")
    paste(sprintf("%s=%s", names(args), vapply(args, deparse1, character(1))), collapse = ", ")
  }

  add_row <- function(block, param, mode, level, prior, link, notes) {
    data.frame(
      block = block,
      parameter = param,
      mode = mode,
      level = level,
      prior = prior,
      link = link,
      notes = notes,
      stringsAsFactors = FALSE
    )
  }

  rows <- list()

  # --- meta header-ish rows (not priors, but useful) ---
  rows[[length(rows) + 1L]] <- add_row("meta", "backend", "info", "model", backend, "", "")
  rows[[length(rows) + 1L]] <- add_row("meta", "kernel", "info", "model", kernel, "", "")
  rows[[length(rows) + 1L]] <- add_row("meta", "components", "info", "model", as.character(K), "", "")
  rows[[length(rows) + 1L]] <- add_row("meta", "N", "info", "model", as.character(N), "", "")
  rows[[length(rows) + 1L]] <- add_row("meta", "P", "info", "model", as.character(P), "", "")

  # --- concentration ---
  conc <- plan$concentration %||% list()
  if (identical(conc$mode, "fixed")) {
    rows[[length(rows) + 1L]] <- add_row("concentration", "alpha", "fixed", "scalar",
                                         prior = deparse1(conc$value), link = "", notes = "fixed concentration")
  } else if (identical(conc$mode, "dist")) {
    rows[[length(rows) + 1L]] <- add_row("concentration", "alpha", "dist", "scalar",
                                         prior = sprintf("%s(%s)", conc$dist, fmt_args(conc$args)),
                                         link = "", notes = "stochastic concentration")
  } else {
    stop("Invalid plan$concentration$mode.", call. = FALSE)
  }

  # --- bulk ---
  bulk <- plan$bulk %||% list()
  for (nm in names(bulk)) {
    ent <- bulk[[nm]]
    mode <- ent$mode %||% NA_character_

    if (mode == "fixed") {
      rows[[length(rows) + 1L]] <- add_row("bulk", nm, "fixed", sprintf("component (1:%d)", K),
                                           prior = deparse1(ent$value), link = "", notes = "")
    } else if (mode == "dist") {
      rows[[length(rows) + 1L]] <- add_row("bulk", nm, "dist", sprintf("component (1:%d)", K),
                                           prior = sprintf("%s(%s)", ent$dist, fmt_args(ent$args)),
                                           link = "", notes = "iid across components")
    } else if (mode == "link") {
      bp <- ent$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 2))
      lk <- ent$link %||% "identity"
      note <- sprintf("beta_%s is %d x %d (components x P)", nm, K, P)
      if (lk == "power") note <- paste0(note, sprintf("; power=%s", deparse1(ent$link_power)))
      rows[[length(rows) + 1L]] <- add_row("bulk", nm, "link", "regression",
                                           prior = sprintf("beta_%s ~ %s(%s)", nm, bp$dist, fmt_args(bp$args)),
                                           link = lk, notes = note)
    } else {
      stop(sprintf("Invalid bulk mode for '%s'.", nm), call. = FALSE)
    }
  }

  # --- GPD ---
  if (GPD) {
    gpd <- plan$gpd %||% list()

    # threshold
    thr <- gpd$threshold %||% NULL
    if (!is.null(thr)) {
      if (thr$mode == "fixed") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "threshold", "fixed", "scalar",
                                             prior = deparse1(thr$value),
                                             link = "", notes = "scalar threshold")
      } else if (thr$mode == "dist") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "threshold", "dist", "scalar",
                                             prior = sprintf("%s(%s)", thr$dist, fmt_args(thr$args)),
                                             link = "", notes = "scalar threshold")
      } else if (thr$mode == "link") {
        bp <- thr$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.2))
        lk <- thr$link %||% "identity"
        note <- sprintf("beta_threshold is length P=%d", P)

        # link_dist case (LN around link)
        if (!is.null(thr$link_dist) && identical(thr$link_dist$dist, "lognormal")) {
          sdlog_u <- gpd$sdlog_u %||% list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
          note <- paste0(note, "; threshold[i] ~ Lognormal(meanlog = X beta, sdlog = sdlog_u)")
          rows[[length(rows) + 1L]] <- add_row("gpd", "threshold", "link+dist", "observation (1:N)",
                                               prior = sprintf("beta_threshold ~ %s(%s); sdlog_u ~ %s(%s)",
                                                               bp$dist, fmt_args(bp$args),
                                                               sdlog_u$dist, fmt_args(sdlog_u$args)),
                                               link = lk, notes = note)
        } else {
          if (lk == "power") note <- paste0(note, sprintf("; power=%s", deparse1(thr$link_power)))
          rows[[length(rows) + 1L]] <- add_row("gpd", "threshold", "link", "observation (1:N)",
                                               prior = sprintf("beta_threshold ~ %s(%s)", bp$dist, fmt_args(bp$args)),
                                               link = lk, notes = note)
        }
      } else {
        stop("Invalid gpd$threshold mode.", call. = FALSE)
      }
    }

    # tail_scale
    ts <- gpd$tail_scale %||% NULL
    if (!is.null(ts)) {
      if (ts$mode == "fixed") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_scale", "fixed", "scalar",
                                             prior = deparse1(ts$value), link = "", notes = "")
      } else if (ts$mode == "dist") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_scale", "dist", "scalar",
                                             prior = sprintf("%s(%s)", ts$dist, fmt_args(ts$args)),
                                             link = "", notes = "")
      } else if (ts$mode == "link") {
        bp <- ts$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.5))
        lk <- ts$link %||% "exp"
        note <- sprintf("beta_tail_scale is length P=%d; tail_scale[i] deterministic", P)
        if (lk == "power") note <- paste0(note, sprintf("; power=%s", deparse1(ts$link_power)))
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_scale", "link", "observation (1:N)",
                                             prior = sprintf("beta_tail_scale ~ %s(%s)", bp$dist, fmt_args(bp$args)),
                                             link = lk, notes = note)
      } else {
        stop("Invalid gpd$tail_scale mode.", call. = FALSE)
      }
    }

    # tail_shape
    tsh <- gpd$tail_shape %||% NULL
    if (!is.null(tsh)) {
      if (tsh$mode == "fixed") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_shape", "fixed", "scalar",
                                             prior = deparse1(tsh$value), link = "", notes = "")
      } else if (tsh$mode == "dist") {
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_shape", "dist", "scalar",
                                             prior = sprintf("%s(%s)", tsh$dist, fmt_args(tsh$args)),
                                             link = "", notes = "")
      } else if (tsh$mode == "link") {
        bp <- tsh$beta_prior %||% list(dist = "normal", args = list(mean = 0, sd = 0.3))
        lk <- tsh$link %||% "identity"
        note <- sprintf("beta_tail_shape is length P=%d; tail_shape[i] deterministic", P)
        if (lk == "power") note <- paste0(note, sprintf("; power=%s", deparse1(tsh$link_power)))
        rows[[length(rows) + 1L]] <- add_row("gpd", "tail_shape", "link", "observation (1:N)",
                                            prior = sprintf("beta_tail_shape ~ %s(%s)", bp$dist, fmt_args(bp$args)),
                                             link = lk, notes = note)
      } else {
        stop("Invalid gpd$tail_shape mode.", call. = FALSE)
      }
    }
  }

  out <- do.call(rbind, rows)
  rownames(out) <- NULL
  out
}

.mcmc_compile_cache_env <- local({
  env <- new.env(parent = emptyenv())
  env$entries <- new.env(parent = emptyenv())
  env
})

.mcmc_cache_key <- function(code, constants, data, dims, monitors, waic_enabled) {
  parts <- list(
    code = .deparse_without_covr(code),
    constants = constants,
    data = data,
    dims = dims,
    monitors = sort(unique(as.character(monitors %||% character(0)))),
    waic_enabled = isTRUE(waic_enabled)
  )
  tf <- tempfile(fileext = ".rds")
  on.exit(unlink(tf), add = TRUE)
  saveRDS(parts, tf)
  unname(tools::md5sum(tf))
}

.mcmc_cache_get <- function(key) {
  if (exists(key, envir = .mcmc_compile_cache_env$entries, inherits = FALSE)) {
    return(get(key, envir = .mcmc_compile_cache_env$entries, inherits = FALSE))
  }
  NULL
}

.mcmc_cache_set <- function(key, value) {
  assign(key, value, envir = .mcmc_compile_cache_env$entries)
  invisible(value)
}

.configure_samplers <- function(conf, spec, data_info = list(), z_update_every = 1L) {
  if (is.null(conf$samplerConfs) || !length(conf$samplerConfs)) return(conf)

  model <- tryCatch(conf$getModel(), error = function(e) NULL)
  if (is.null(model)) {
    model <- tryCatch(conf$model, error = function(e) NULL)
  }
  stoch_nodes <- tryCatch(model$getNodeNames(stochOnly = TRUE, includeData = FALSE), error = function(e) character(0))
  z_nodes <- stoch_nodes[grepl("^z\\[", stoch_nodes)]

  # Normalize existing sampler controls.
  for (i in seq_along(conf$samplerConfs)) {
    ctl <- conf$samplerConfs[[i]]$control %||% list()
    if (is.null(ctl$checkConjugacy)) ctl$checkConjugacy <- FALSE

    # Reduce adaptation overhead for RW-family samplers.
    nm <- conf$samplerConfs[[i]]$name %||% ""
    if (nm %in% c("RW", "RW_block")) {
      if (is.null(ctl$adaptInterval)) ctl$adaptInterval <- 200L
      if (is.null(ctl$adaptScaleOnly)) ctl$adaptScaleOnly <- TRUE
    }

    # Keep any CRP-related cluster nodes restricted to stochastic nodes.
    if (!is.null(ctl$clusterVarInfo) && length(stoch_nodes) > 0) {
      cvi <- ctl$clusterVarInfo
      cn <- cvi$clusterNodes
      if (is.character(cn)) cn <- list(cn)
      if (is.list(cn) && length(cn) > 0) {
        cn_filtered <- lapply(cn, function(nodes) nodes[nodes %in% stoch_nodes])
        if (any(lengths(cn_filtered) != lengths(cn))) {
          cvi$clusterNodes <- cn_filtered
          if (!is.null(cvi$numNodesPerCluster)) cvi$numNodesPerCluster <- lengths(cn_filtered)
          ctl$clusterVarInfo <- cvi
        }
      }
    }

    conf$samplerConfs[[i]]$control <- ctl
  }

  # Only re-block beta nodes that already use RW-family samplers.
  # Leave conjugate and CRP-cluster wrapper samplers intact; removing them can
  # strand nodes like beta_threshold without any sampler at all.
  beta_block_nodes <- unique(unlist(lapply(conf$samplerConfs, function(sc) {
    nm <- sc$name %||% ""
    tgt <- sc$target %||% character(0)
    if (!nm %in% c("RW", "RW_block") || !length(tgt) || !all(grepl("^beta", tgt))) {
      return(character(0))
    }
    tgt
  }), use.names = FALSE))
  if (length(beta_block_nodes) >= 2L) {
    if (all(vapply(beta_block_nodes, function(nn) {
      any(vapply(conf$getSamplers(), function(ss) nn %in% ss$target, logical(1)))
    }, logical(1)))) {
      suppressWarnings(try(conf$removeSamplers(beta_block_nodes), silent = TRUE))
      add_beta_block <- try(
        conf$addSampler(
          target = beta_block_nodes,
          type = "RW_block",
          control = list(adaptInterval = 200L, adaptive = TRUE)
        ),
        silent = TRUE
      )
      if (inherits(add_beta_block, "try-error")) {
        for (nn in beta_block_nodes) {
          suppressWarnings(try(conf$addSampler(
            target = nn,
            type = "RW",
            control = list(adaptInterval = 200L, adaptive = TRUE)
          ), silent = TRUE))
        }
      }
    }
  }

  # Prefer slice for strictly positive scale/rate style parameters.
  pos_nodes <- stoch_nodes[grepl("(sigma|scale|rate|sdlog|tail_scale)", stoch_nodes)]
  if (length(pos_nodes)) {
    for (nn in unique(pos_nodes)) {
      suppressWarnings(try(conf$removeSamplers(nn), silent = TRUE))
      suppressWarnings(try(conf$addSampler(target = nn, type = "slice"), silent = TRUE))
    }
  }

  # z_update_every > 1 retains exact target but lowers frequency of z updates.
  # NIMBLE lacks built-in periodic execution per-sampler here; we tune only adaptation
  # and retain z samplers each iteration for correctness.
  if (length(z_nodes) && isTRUE(z_update_every > 1L)) {
    for (i in seq_along(conf$samplerConfs)) {
      tgt <- conf$samplerConfs[[i]]$target %||% character(0)
      if (!length(tgt) || !all(grepl("^z\\[", tgt))) next
      ctl <- conf$samplerConfs[[i]]$control %||% list()
      if (is.null(ctl$adaptInterval)) ctl$adaptInterval <- max(200L, as.integer(200L * z_update_every))
      conf$samplerConfs[[i]]$control <- ctl
    }
  }

  conf
}


#' Run posterior sampling for a prepared one-arm bundle
#'
#' \code{run_mcmc_bundle_manual()} is the explicit runner for objects created by
#' \code{\link{build_nimble_bundle}}. It compiles the stored NIMBLE code,
#' executes MCMC, and returns a \code{"mixgpd_fit"} object.
#'
#' @details
#' The resulting fit supports posterior summaries of the model parameters as
#' well as posterior predictive functionals such as
#' \eqn{f(y \mid x)}, \eqn{S(y \mid x)},
#' \eqn{Q(\tau \mid x)}, and restricted means.
#'
#' If \code{parallel_chains = TRUE}, chains are run concurrently when the stored
#' MCMC configuration uses more than one chain. If the bundle was built with
#' latent cluster labels monitored, the \code{z_update_every} argument controls how
#' frequently those latent indicators are refreshed during sampling.
#'
#' @param bundle A \code{causalmixgpd_bundle} from \code{build_nimble_bundle()}.
#' @param show_progress Logical; if TRUE, print step messages and render progress where supported.
#' @param quiet Logical; if TRUE, suppress console status messages.
#'   Set to FALSE to see progress messages during MCMC setup and execution.
#' @param parallel_chains Logical; run chains concurrently when \code{nchains > 1}.
#' @param workers Optional integer number of workers for parallel execution.
#' @param timing Logical; if TRUE, include stage timings (\code{build}, \code{compile}, \code{mcmc})
#'   in \code{fit$timing}.
#' @param z_update_every Integer >= 1 controlling latent cluster-label update cadence.
#' @return A fitted object of class \code{"mixgpd_fit"} containing posterior
#'   draws, model metadata, and cached objects used by downstream S3 methods.
#' @seealso \code{\link{build_nimble_bundle}}, \code{\link{mcmc}},
#'   \code{\link{summary.mixgpd_fit}}, \code{\link{predict.mixgpd_fit}}.
#' @examples
#' \donttest{
#' library(nimble)
#' y <- abs(rnorm(25)) + 0.1
#' bundle <- build_nimble_bundle(
#'   y = y,
#'   backend = "sb",
#'   kernel = "normal",
#'   GPD = FALSE,
#'   components = 3,
#'   mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
#' )
#' fit <- run_mcmc_bundle_manual(bundle, show_progress = FALSE)
#' fit
#' }
#' @export
run_mcmc_bundle_manual <- function(bundle, show_progress = TRUE, quiet = FALSE,
                                   parallel_chains = FALSE, workers = NULL, timing = FALSE,
                                   z_update_every = NULL) {
  suppressPackageStartupMessages(base::require("nimble", quietly = TRUE, warn.conflicts = FALSE))
  stopifnot(inherits(bundle, "causalmixgpd_bundle"))
  progress_label <- if (inherits(bundle, "dpmixgpd_cluster_bundle")) "cluster" else "mixgpd"

  progress_ctx <- .cmgpd_progress_start(
    total_steps = 8L,
    enabled = isTRUE(show_progress),
    quiet = isTRUE(quiet),
    label = progress_label
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Validating configuration")

  spec <- bundle$spec
  m <- bundle$mcmc %||% list()
  waic_enabled <- !isFALSE(m$waic)

  code <- .extract_nimble_code(bundle$code)
  constants <- bundle$constants %||% list()
  data <- bundle$data %||% list()
  dims <- bundle$dimensions %||% list()
  monitors <- bundle$monitors %||% character(0)

  inits_obj <- bundle$inits_fun %||% bundle$inits %||% function() list()
  inits_fun <- if (is.function(inits_obj)) inits_obj else if (is.list(inits_obj)) function() inits_obj else function() list()

  niter <- as.integer(m$niter %||% 2000)
  nburnin <- as.integer(m$nburnin %||% 500)
  thin <- as.integer(m$thin %||% 1)
  nchains <- as.integer(m$nchains %||% 1)
  timing <- isTRUE(timing %||% m$timing %||% FALSE)
  parallel_chains <- isTRUE(parallel_chains %||% m$parallel_chains %||% FALSE)
  nimble_quiet <- isTRUE(show_progress) || isTRUE(quiet)
  z_update_every <- as.integer(z_update_every %||% m$z_update_every %||% 1L)
  if (!is.finite(z_update_every) || z_update_every < 1L) {
    stop("'z_update_every' must be an integer >= 1.", call. = FALSE)
  }
  if (z_update_every > 1L && !isTRUE(quiet) && !isTRUE(show_progress)) {
    .cmgpd_message(sprintf("Using z_update_every = %d; this may reduce compute but can slow mixing.", z_update_every))
  }
  workers <- as.integer(workers %||% m$workers %||% max(1L, min(nchains, parallel::detectCores(logical = FALSE))))
  if (!is.finite(workers) || workers < 1L) workers <- 1L

  seed <- m$seed %||% NULL
  if (is.null(seed) || identical(seed, FALSE)) seed <- as.integer(Sys.time()) + Sys.getpid()
  seed <- as.integer(seed)
  if (length(seed) == 1L && nchains > 1L) seed <- seed + seq_len(nchains) - 1L
  if (length(seed) != nchains) stop("mcmc$seed must be length 1 or length nchains.", call. = FALSE)

  tic <- function() proc.time()[["elapsed"]]
  timing_info <- list(build = 0, compile = 0, mcmc = 0, cache_hit = FALSE)

  if (parallel_chains && nchains > 1L &&
      requireNamespace("future", quietly = TRUE) &&
      requireNamespace("future.apply", quietly = TRUE)) {
    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = min(workers, nchains))
    t0 <- tic()
    chain_fits <- future.apply::future_lapply(seq_len(nchains), function(ch) {
      b <- bundle
      b$mcmc <- utils::modifyList(
        b$mcmc %||% list(),
        list(nchains = 1L, seed = seed[ch], parallel_chains = FALSE, z_update_every = z_update_every)
      )
      run_mcmc_bundle_manual(
        b,
        show_progress = FALSE,
        quiet = TRUE,
        parallel_chains = FALSE,
        workers = 1L,
        timing = timing,
        z_update_every = z_update_every
      )
    }, future.seed = TRUE)
    timing_info$mcmc <- tic() - t0

    sample_list <- lapply(chain_fits, function(fit_i) {
      smp <- fit_i$samples
      if (inherits(smp, "mcmc.list")) smp[[1L]] else smp
    })
    samples <- if (requireNamespace("coda", quietly = TRUE)) coda::mcmc.list(sample_list) else sample_list

    fit <- chain_fits[[1L]]
    fit$mcmc$nchains <- nchains
    fit$mcmc$seed <- seed
    fit$mcmc$samples <- samples
    fit$mcmc$parallel_chains <- TRUE
    fit$samples <- samples
    fit$timing <- timing_info
    class(fit) <- unique(c("mixgpd_fit", "causalmixgpd_fit", "list"))
    return(fit)
  }

  if (parallel_chains && nchains > 1L && !isTRUE(quiet) && !isTRUE(show_progress)) {
    warning("parallel_chains=TRUE requested but 'future'/'future.apply' are unavailable; running sequentially.",
            call. = FALSE)
  }

  .cmgpd_progress_step(progress_ctx, "Checking build/compile cache")
  cache_key <- .mcmc_cache_key(code = code, constants = constants, data = data, dims = dims, monitors = monitors, waic_enabled = waic_enabled)
  cache_entry <- .mcmc_cache_get(cache_key)
  if (!is.null(cache_entry)) {
    timing_info$cache_hit <- TRUE
    if (!isTRUE(quiet) && !isTRUE(show_progress)) .cmgpd_message("[MCMC] Reusing cached build/compile.")
  }

  if (is.null(cache_entry)) {
    cache_entry <- .with_nimble_exports({
      .cmgpd_progress_step(progress_ctx, "Building model and MCMC configuration")
      t0_build <- tic()
      # Generated models are validated upstream; NIMBLE's full check path can be
      # disproportionately expensive for manuscript-scale fits.
      .register_nimble_exports()
      Rmodel <- tryCatch(
        .cmgpd_capture_nimble(
          nimble::nimbleModel(
            code = code, data = data, constants = constants,
            inits = inits_fun(), dimensions = dims, check = FALSE, calculate = FALSE
          ),
          suppress = nimble_quiet
        ),
        error = function(e) {
          msg <- conditionMessage(e)
          if (grepl("keywords:", msg, ignore.case = TRUE) && grepl("Please use a different name", msg, fixed = TRUE)) {
            return(
              .cmgpd_capture_nimble(
                nimble::nimbleModel(
                  code = code, data = data, constants = constants,
                  inits = inits_fun(), dimensions = dims, check = FALSE, calculate = FALSE
                ),
                suppress = nimble_quiet
              )
            )
          }
          stop(e)
        }
      )

      conf <- .cmgpd_capture_nimble(
        nimble::configureMCMC(Rmodel, monitors = monitors, enableWAIC = waic_enabled),
        suppress = nimble_quiet
      )
      conf <- .cmgpd_capture_nimble(
        .configure_samplers(
          conf,
          spec = spec,
          data_info = list(constants = constants, dimensions = dims),
          z_update_every = z_update_every
        ),
        suppress = nimble_quiet
      )

      Rmcmc <- .cmgpd_capture_nimble(nimble::buildMCMC(conf), suppress = nimble_quiet)
      timing_info$build <- tic() - t0_build

      .cmgpd_progress_step(progress_ctx, "Compiling NIMBLE model")
      compiled <- TRUE
      Cmodel <- NULL
      Cmcmc <- NULL
      t0_compile <- tic()
      compile_err <- tryCatch({
        Cmodel <- .cmgpd_capture_nimble(
          nimble::compileNimble(Rmodel, showCompilerOutput = FALSE),
          suppress = nimble_quiet
        )
        Cmcmc <- .cmgpd_capture_nimble(
          nimble::compileNimble(Rmcmc, project = Rmodel, showCompilerOutput = FALSE),
          suppress = nimble_quiet
        )
        NULL
      }, error = function(e) e)
      timing_info$compile <- tic() - t0_compile
      if (inherits(compile_err, "error")) {
        compiled <- FALSE
        if (!isTRUE(quiet) && !isTRUE(show_progress)) {
          warning(paste0("nimble model compilation failed; running uncompiled MCMC for portability: ", conditionMessage(compile_err)), call. = FALSE)
        }
      }

      list(
        compiled = compiled,
        Rmodel = Rmodel,
        Rmcmc = Rmcmc,
        Cmodel = if (compiled) Cmodel else NULL,
        Cmcmc = if (compiled) Cmcmc else NULL,
        conf = conf
      )
    })
    .mcmc_cache_set(cache_key, cache_entry)
  } else {
    .cmgpd_progress_step(progress_ctx, "Building model and MCMC configuration (cached)")
    .cmgpd_progress_step(progress_ctx, "Compiling NIMBLE model (cached)")
  }

  compiled <- isTRUE(cache_entry$compiled)
  Rmodel <- cache_entry$Rmodel
  Rmcmc <- cache_entry$Rmcmc
  Cmodel <- cache_entry$Cmodel
  Cmcmc <- cache_entry$Cmcmc
  conf <- cache_entry$conf
  engine_mcmc <- if (compiled) Cmcmc else Rmcmc

  .cmgpd_progress_step(progress_ctx, "Initializing chains")
  inits_list <- if (nchains > 1L) {
    out <- vector("list", nchains)
    for (ch in seq_len(nchains)) {
      set.seed(seed[ch])
      out[[ch]] <- inits_fun()
    }
    out
  } else {
    set.seed(seed[1L])
    inits_fun()
  }

  .cmgpd_progress_step(progress_ctx, "Running MCMC")
  t0_mcmc <- tic()
  run_mcmc_once <- function(current_inits) {
    tryCatch(
      .cmgpd_capture_nimble(
        nimble::runMCMC(
          engine_mcmc,
          niter = niter,
          nburnin = nburnin,
          thin = thin,
          nchains = nchains,
          inits = current_inits,
          progressBar = FALSE,
          samplesAsCodaMCMC = TRUE
        ),
        suppress = nimble_quiet
      ),
      error = function(e) e
    )
  }
  res <- .with_nimble_exports(run_mcmc_once(inits_list))
  is_crp_init_error <- inherits(res, "error") &&
    grepl(
      "CRP_sampler: sampler encountered case where the log probability density values corresponding to all potential cluster memberships are negative infinity",
      conditionMessage(res),
      fixed = TRUE
    )
  if (is_crp_init_error && identical(spec$meta$backend, "crp")) {
    max_crp_retries <- 5L
    for (attempt in seq_len(max_crp_retries)) {
      retry_seed <- seed + attempt
      retry_inits <- if (nchains > 1L) {
        out <- vector("list", nchains)
        for (ch in seq_len(nchains)) {
          out[[ch]] <- build_inits_from_spec(
            spec,
            seed = retry_seed[ch],
            y = data$y,
            X = data$X %||% NULL
          )
        }
        out
      } else {
        build_inits_from_spec(
          spec,
          seed = retry_seed[1L],
          y = data$y,
          X = data$X %||% NULL
        )
      }
      res <- .with_nimble_exports(run_mcmc_once(retry_inits))
      if (!inherits(res, "error")) {
        inits_list <- retry_inits
        break
      }
    }
  }
  timing_info$mcmc <- tic() - t0_mcmc

  if (inherits(res, "error")) {
    stop(res)
  } else if (is.list(res) && !is.null(res$WAIC)) {
    waic_obj <- res$WAIC
    samples <- res$samples
  } else {
    waic_obj <- NULL
    samples <- res
  }

  .cmgpd_progress_step(progress_ctx, "Finalizing WAIC and diagnostics")
  if (is.null(waic_obj) && waic_enabled) {
    waic_obj <- tryCatch(
      .cmgpd_capture_nimble(
        if (compiled) nimble::calculateWAIC(Cmcmc) else nimble::calculateWAIC(Rmcmc),
        suppress = nimble_quiet
      ),
      error = function(e) NULL
    )
  }

  .cmgpd_progress_step(progress_ctx, "Assembling fit object")
  fit <- list(
    call = match.call(),
    spec = spec,
    data = data,
    model = if (compiled) Cmodel else Rmodel,
    mcmc_conf = conf,
    mcmc = list(
      engine = if (compiled) "compiled" else "uncompiled",
      niter = niter,
      nburnin = nburnin,
      thin = thin,
      nchains = nchains,
      seed = seed,
      z_update_every = z_update_every,
      samples = samples,
      waic = waic_obj
    ),
    code = code,
    constants = constants,
    dimensions = dims,
    monitors = monitors,
    cache = list(predict_env = new.env(parent = emptyenv())),
    epsilon = bundle$epsilon %||% NULL,
    timing = timing_info
  )

  fit$samples <- samples
  fit$waic <- waic_obj

  class(fit) <- unique(c("mixgpd_fit", "causalmixgpd_fit", "list"))
  fit
}

# ============================================================
# Utilities (internal)
# ============================================================



#' Null-coalescing operator
#' @keywords internal
#' @noRd
#' @rdname null_coalescing
#' @name null_coalescing
`%||%` <- function(a, b) if (!is.null(a)) a else b

.cmgpd_message <- function(...) {
  msg <- paste0(..., collapse = "")
  old <- getOption("CausalMixGPD._allow_message", FALSE)
  options(CausalMixGPD._allow_message = TRUE)
  on.exit(options(CausalMixGPD._allow_message = old), add = TRUE)
  message(msg)
  invisible(msg)
}

.cmgpd_progress_colorize <- function(text, step_index, enabled = TRUE) {
  if (!isTRUE(enabled) || !nzchar(text)) return(text)
  if (!requireNamespace("cli", quietly = TRUE)) return(text)
  ncols <- get0("num_ansi_colors", envir = asNamespace("cli"), mode = "function", inherits = FALSE)
  ncols <- if (is.function(ncols)) {
    tryCatch(as.integer(ncols()), error = function(...) 0L)
  } else {
    0L
  }
  idx <- max(1L, as.integer(step_index)[1L])

  if (isTRUE(ncols >= 256L)) {
    # Broad hue spread in 256-color ANSI to keep adjacent steps visually distinct.
    palette_256 <- c(39L, 45L, 51L, 50L, 49L, 48L, 47L, 46L, 82L, 118L,
                     154L, 190L, 226L, 220L, 214L, 208L, 202L, 196L,
                     197L, 198L, 199L, 165L, 129L, 93L, 57L, 63L, 69L, 75L)
    code <- palette_256[((idx - 1L) %% length(palette_256)) + 1L]
    return(paste0("\033[38;5;", code, "m", text, "\033[39m"))
  }

  palette_named <- c(
    "col_blue", "col_cyan", "col_green", "col_yellow", "col_magenta", "col_red",
    "col_br_blue", "col_br_cyan", "col_br_green", "col_br_yellow", "col_br_magenta", "col_br_red",
    "col_silver"
  )
  fn <- get0(
    palette_named[((idx - 1L) %% length(palette_named)) + 1L],
    envir = asNamespace("cli"),
    mode = "function",
    inherits = FALSE
  )
  if (is.null(fn)) return(text)
  tryCatch(fn(text), error = function(...) text)
}

.cmgpd_progress_format <- function(current, total, step_label, label = NULL, width = 12L, color = FALSE) {
  step_label <- trimws(as.character(step_label %||% ""))
  if (!nzchar(step_label)) step_label <- "Working..."
  family_label <- trimws(as.character(label %||% ""))
  msg <- if (nzchar(family_label)) {
    paste0("[", family_label, "] ", step_label)
  } else {
    step_label
  }
  .cmgpd_progress_colorize(msg, step_index = current, enabled = color)
}

.cmgpd_progress_live_format <- function(current, total, step_label, label = NULL, color = FALSE) {
  .cmgpd_progress_format(
    current = current,
    total = total,
    step_label = step_label,
    label = label,
    color = color
  )
}

.cmgpd_progress_bar_line <- function(current, total, width = 28L) {
  current <- max(0L, as.integer(current)[1L])
  total <- max(1L, as.integer(total)[1L])
  width <- max(10L, as.integer(width)[1L])
  current <- min(current, total)
  pct <- as.integer(round(100 * current / total))
  filled <- if (current >= total) width else as.integer(floor(width * current / total))
  empty <- max(0L, width - filled)
  paste0("  [", strrep("=", filled), strrep("-", empty), "] ", sprintf("%3d%%", pct))
}

.cmgpd_progress_write <- function(text) {
  cat(text)
  utils::flush.console()
  invisible(text)
}

.cmgpd_progress_visible_nchar <- function(text) {
  text <- gsub("\033\\[[0-9;]*[A-Za-z]", "", as.character(text), perl = TRUE)
  nchar(text, type = "width")
}

.cmgpd_progress_pad_line <- function(text, width) {
  width <- max(0L, as.integer(width)[1L])
  pad <- max(0L, width - .cmgpd_progress_visible_nchar(text))
  paste0(text, strrep(" ", pad))
}

.cmgpd_progress_render <- function(ctx, status = ctx$status, final = FALSE) {
  if (!is.environment(ctx) || !identical(ctx$live_backend, "inline")) return(invisible(ctx))
  bar <- .cmgpd_progress_bar_line(ctx$current, ctx$total, width = ctx$bar_width)
  ctx$last_status_width <- max(ctx$last_status_width %||% 0L, .cmgpd_progress_visible_nchar(status))
  ctx$last_bar_width <- max(ctx$last_bar_width %||% 0L, .cmgpd_progress_visible_nchar(bar))
  status_out <- .cmgpd_progress_pad_line(status, ctx$last_status_width)
  bar_out <- .cmgpd_progress_pad_line(bar, ctx$last_bar_width)
  prefix <- if (isTRUE(ctx$rendered)) "\r\033[F" else ""
  line2_prefix <- "\n"
  suffix <- if (isTRUE(final)) "\n" else ""
  .cmgpd_progress_write(paste0(prefix, status_out, line2_prefix, bar_out, suffix))
  ctx$rendered <- TRUE
  invisible(ctx)
}

.cmgpd_progress_start <- function(total_steps, enabled = TRUE, quiet = FALSE, label = NULL) {
  total_steps <- as.integer(total_steps)[1L]
  if (!is.finite(total_steps) || total_steps < 1L) total_steps <- 1L
  enabled <- isTRUE(enabled) && !isTRUE(quiet)

  ctx <- new.env(parent = emptyenv())
  ctx$enabled <- enabled
  ctx$total <- total_steps
  ctx$current <- 0L
  ctx$label <- as.character(label %||% "")
  ctx$inline_width <- 12L
  ctx$bar_width <- 28L
  ctx$live_backend <- "none"
  ctx$rendered <- FALSE
  ctx$last_status_width <- 0L
  ctx$last_bar_width <- 0L
  ctx$step_label <- "Starting..."
  ctx$status <- .cmgpd_progress_live_format(
    current = 0L,
    total = total_steps,
    step_label = ctx$step_label,
    label = ctx$label,
    color = FALSE
  )

  has_cli <- requireNamespace("cli", quietly = TRUE)
  ctx$has_cli <- has_cli
  ctx$color_enabled <- FALSE
  if (has_cli) {
    ncols <- get0("num_ansi_colors", envir = asNamespace("cli"), mode = "function", inherits = FALSE)
    if (is.function(ncols)) {
      cols <- tryCatch(as.integer(ncols()), error = function(...) 0L)
      ctx$color_enabled <- isTRUE(cols > 1L)
    }
  }

  # Live 2-line renderer only in interactive terminals; step messages still emitted elsewhere.
  can_bar <- enabled && interactive() && !isTRUE(getOption("knitr.in.progress"))
  if (can_bar) {
    ctx$live_backend <- "inline"
    .cmgpd_progress_render(ctx, status = ctx$status, final = FALSE)
  }
  ctx
}

.cmgpd_progress_step <- function(ctx, step_label) {
  if (!is.environment(ctx) || !isTRUE(ctx$enabled)) return(invisible(ctx))
  prev <- as.integer(ctx$current)
  ctx$current <- min(prev + 1L, ctx$total)
  inc <- ctx$current - prev

  ctx$step_label <- trimws(as.character(step_label %||% ""))
  if (!nzchar(ctx$step_label)) ctx$step_label <- "Working..."
  status <- if (identical(ctx$live_backend, "none")) {
    .cmgpd_progress_format(
      current = ctx$current,
      total = ctx$total,
      step_label = ctx$step_label,
      label = ctx$label,
      width = ctx$inline_width,
      color = ctx$color_enabled
    )
  } else {
    .cmgpd_progress_live_format(
      current = ctx$current,
      total = ctx$total,
      step_label = ctx$step_label,
      label = ctx$label,
      color = ctx$color_enabled
    )
  }
  ctx$status <- status

  if (identical(ctx$live_backend, "inline")) {
    .cmgpd_progress_render(ctx, status = status, final = FALSE)
  } else {
    .cmgpd_message(status)
  }
  invisible(ctx)
}

.cmgpd_progress_done <- function(ctx, final_label = NULL) {
  if (!is.environment(ctx) || !isTRUE(ctx$enabled)) return(invisible(ctx))
  remain <- max(0L, as.integer(ctx$total) - as.integer(ctx$current))
  if (remain > 0L) ctx$current <- as.integer(ctx$total)
  final_status <- if (is.null(final_label)) {
    if (identical(ctx$live_backend, "none")) {
      .cmgpd_progress_format(
        current = ctx$total,
        total = ctx$total,
        step_label = ctx$step_label,
        label = ctx$label,
        width = ctx$inline_width,
        color = ctx$color_enabled
      )
    } else {
      .cmgpd_progress_live_format(
        current = ctx$total,
        total = ctx$total,
        step_label = ctx$step_label,
        label = ctx$label,
        color = ctx$color_enabled
      )
    }
  } else if (identical(ctx$live_backend, "none")) {
    .cmgpd_progress_format(
      current = ctx$total,
      total = ctx$total,
      step_label = as.character(final_label),
      label = ctx$label,
      width = ctx$inline_width,
      color = ctx$color_enabled
    )
  } else {
    .cmgpd_progress_live_format(
      current = ctx$total,
      total = ctx$total,
      step_label = as.character(final_label),
      label = ctx$label,
      color = ctx$color_enabled
    )
  }

  ctx$current <- as.integer(ctx$total)
  ctx$status <- final_status
  if (!is.null(final_label)) {
    ctx$step_label <- as.character(final_label)
  }
  if (identical(ctx$live_backend, "inline")) {
    .cmgpd_progress_render(ctx, status = final_status, final = TRUE)
  } else if (!is.null(final_label)) {
    .cmgpd_message(final_status)
  }
  invisible(ctx)
}

.cmgpd_capture_nimble <- function(expr, suppress = FALSE) {
  if (!isTRUE(suppress)) {
    return(eval.parent(substitute(expr)))
  }
  result <- NULL
  utils::capture.output(
    withCallingHandlers(
      result <- eval.parent(substitute(expr)),
      warning = function(w) invokeRestart("muffleWarning"),
      message = function(m) invokeRestart("muffleMessage")
    ),
    file = if (.Platform$OS.type == "windows") "NUL" else "/dev/null"
  )
  result
}


.silent_wrapper <- function(fun_name, fun, opt_name) {
  wrapper <- function(...) {
    if (isFALSE(getOption(opt_name, TRUE))) {
      return(fun(...))
    }
    call <- match.call(expand.dots = FALSE)
    args <- as.list(call)[-1]
    if (length(args) && !is.null(args$...)) {
      dot_args <- eval(args$..., parent.frame())
      args$... <- NULL
      args <- c(args, dot_args)
    }
    call_env <- new.env(parent = parent.frame())
    assign(fun_name, fun, envir = call_env)
    run <- function() do.call(fun_name, args, envir = call_env)
    vis <- withCallingHandlers(
      withVisible(run()),
      warning = function(w) invokeRestart("muffleWarning"),
      message = function(m) {
        if (isTRUE(getOption("CausalMixGPD._allow_message", FALSE))) return()
        invokeRestart("muffleMessage")
      }
    )
    if (isTRUE(vis$visible)) return(vis$value)
    invisible(vis$value)
  }
  attr(wrapper, "CausalMixGPD.silent_wrapper") <- TRUE
  wrapper
}

.wrap_exported_silent <- function(pkgname, opt_name = "CausalMixGPD.silent") {
  if (isFALSE(getOption(opt_name, TRUE))) return(invisible(FALSE))
  ns <- asNamespace(pkgname)
  exports <- getNamespaceExports(pkgname)
  for (nm in exports) {
    # NIMBLE must see the original exported generator objects for custom
    # distributions and helpers; wrapping them in plain closures breaks type
    # discovery during model build/compile.
    if (grepl("^[dpqr][A-Z]", nm)) next
    obj <- get(nm, envir = ns, inherits = FALSE)
    if (!is.function(obj)) next
    if (isTRUE(attr(obj, "CausalMixGPD.silent_wrapper"))) next
    assign(nm, .silent_wrapper(nm, obj, opt_name), envir = ns)
  }
  invisible(TRUE)
}

.nimble_export_source_env <- function(pkgname = "CausalMixGPD") {
  src_env <- environment(.register_nimble_exports)
  local_exports <- grep("^[dpqr][A-Z]", ls(src_env, all.names = TRUE), value = TRUE)
  if (length(local_exports)) {
    return(src_env)
  }
  asNamespace(pkgname)
}

.nimble_export_names <- function(pkgname = "CausalMixGPD") {
  src_env <- .nimble_export_source_env(pkgname)
  nms <- grep("^[dpqr][A-Z]", ls(src_env, all.names = TRUE), value = TRUE)
  nms[vapply(
    nms,
    function(nm) exists(nm, envir = src_env, inherits = FALSE) &&
      is.function(get(nm, envir = src_env, inherits = FALSE)),
    logical(1)
  )]
}

.nimble_support_names <- function() {
  c(
    "getNimbleOption",
    "messageIfVerbose",
    "nimSwitch",
    "nimbleInternalFunctions"
  )
}

.register_nimble_exports <- function(pkgname = "CausalMixGPD", envir = parent.frame()) {
  src_env <- .nimble_export_source_env(pkgname)
  exports <- .nimble_export_names(pkgname)
  for (nm in exports) {
    if (!exists(nm, envir = src_env, inherits = FALSE)) next
    assign(nm, get(nm, envir = src_env, inherits = FALSE), envir = envir)
  }
  nimble_ns <- asNamespace("nimble")
  support <- .nimble_support_names()
  for (nm in support) {
    if (!exists(nm, envir = nimble_ns, inherits = FALSE)) next
    assign(nm, get(nm, envir = nimble_ns, inherits = FALSE), envir = envir)
  }
  invisible(c(exports, support))
}

.with_nimble_exports <- function(expr, pkgname = "CausalMixGPD", envir = .GlobalEnv) {
  src_env <- .nimble_export_source_env(pkgname)
  custom_exports <- .nimble_export_names(pkgname)
  support_exports <- .nimble_support_names()
  exports <- c(custom_exports, support_exports)
  if (!length(exports)) {
    return(eval.parent(substitute(expr)))
  }

  had_binding <- stats::setNames(logical(length(exports)), exports)
  old_values <- stats::setNames(vector("list", length(exports)), exports)

  for (nm in exports) {
    had_binding[[nm]] <- exists(nm, envir = envir, inherits = FALSE)
    if (isTRUE(had_binding[[nm]])) {
      old_values[[nm]] <- get(nm, envir = envir, inherits = FALSE)
    }
    src <- if (nm %in% support_exports) asNamespace("nimble") else src_env
    if (exists(nm, envir = src, inherits = FALSE)) {
      assign(nm, get(nm, envir = src, inherits = FALSE), envir = envir)
    }
  }

  on.exit({
    for (nm in rev(exports)) {
      if (isTRUE(had_binding[[nm]])) {
        assign(nm, old_values[[nm]], envir = envir)
      } else if (exists(nm, envir = envir, inherits = FALSE)) {
        rm(list = nm, envir = envir)
      }
    }
  }, add = TRUE)

  eval.parent(substitute(expr))
}

#' Extract indexed parameter blocks from a draws matrix
#' @keywords internal
#' @noRd
.indexed_block <- function(mat0, base, K = NULL, allow_missing = FALSE) {
  cn0 <- colnames(mat0)
  pat <- paste0("^", base, "\\[([0-9]+)\\]$")
  hit <- grepl(pat, cn0)
  if (!any(hit)) {
    if (isTRUE(allow_missing)) return(NULL)
    stop(sprintf("No indexed columns found for '%s[i]'.", base), call. = FALSE)
  }

  idx <- as.integer(sub(pat, "\\1", cn0[hit]))
  ord <- order(idx)
  idx <- idx[ord]
  cols <- cn0[hit][ord]

  if (is.null(K)) K <- max(idx, na.rm = TRUE)
  K <- as.integer(K)

  out <- matrix(0.0, nrow = nrow(mat0), ncol = K)
  for (j in seq_along(cols)) {
    k <- idx[j]
    if (!is.na(k) && k >= 1 && k <= K) out[, k] <- mat0[, cols[j]]
  }
  out
}

#' Extract indexed component-by-column parameter blocks from a draws matrix
#' @keywords internal
#' @noRd
.indexed_block_matrix <- function(mat0, base, K = NULL, P = NULL, allow_missing = FALSE) {
  cn0 <- colnames(mat0)
  pat <- paste0("^", base, "\\[([0-9]+),\\s*([0-9]+)\\]$")
  hit <- grepl(pat, cn0)
  if (!any(hit)) {
    if (isTRUE(allow_missing)) return(NULL)
    stop(sprintf("No indexed columns found for '%s[i,j]'.", base), call. = FALSE)
  }

  idx1 <- as.integer(sub(pat, "\\1", cn0[hit]))
  idx2 <- as.integer(sub(pat, "\\2", cn0[hit]))
  cols <- cn0[hit]

  if (is.null(K)) K <- max(idx1, na.rm = TRUE)
  if (is.null(P)) P <- max(idx2, na.rm = TRUE)
  K <- as.integer(K)
  P <- as.integer(P)

  out <- array(0.0, dim = c(nrow(mat0), K, P))
  for (j in seq_along(cols)) {
    k <- idx1[j]
    p <- idx2[j]
    if (!is.na(k) && !is.na(p) && k >= 1L && k <= K && p >= 1L && p <= P) {
      out[, k, p] <- mat0[, cols[j]]
    }
  }
  out
}

#' Extract mixture weights from draws matrix
#' @keywords internal
#' @noRd
.extract_weights <- function(draw_mat, backend = c("sb", "crp")) {
  if (is.null(draw_mat) || !is.matrix(draw_mat)) {
    stop("draw_mat must be a numeric matrix.", call. = FALSE)
  }
  backend <- match.arg(backend)
  cn <- colnames(draw_mat)

  if (identical(backend, "sb")) {
    if (any(grepl("^w\\[[0-9]+\\]$", cn))) {
      return(.indexed_block(draw_mat, "w"))
    }
    if (any(grepl("^weights\\[[0-9]+\\]$", cn))) {
      return(.indexed_block(draw_mat, "weights"))
    }
    stop("Could not find component weights in posterior draws.", call. = FALSE)
  }

  if (!any(grepl("^z\\[[0-9]+\\]$", cn))) {
    stop("Backend requires z[i] in samples to derive weights.", call. = FALSE)
  }

  Z <- .indexed_block(draw_mat, "z")
  K <- max(Z, na.rm = TRUE)
  if (!is.finite(K) || K < 1L) stop("Could not infer K for CRP weights.", call. = FALSE)
  K <- as.integer(K)

  S <- nrow(draw_mat)
  W <- matrix(0.0, nrow = S, ncol = K)
  for (s in seq_len(S)) {
    z_s <- Z[s, ]
    z_s <- z_s[is.finite(z_s)]
    z_s <- z_s[z_s >= 1 & z_s <= K]
    if (length(z_s)) W[s, ] <- tabulate(z_s, nbins = K) / length(z_s)
  }
  W
}

#' Extract bulk parameter blocks from draws matrix
#' @keywords internal
#' @noRd
.extract_bulk_params <- function(draw_mat, bulk_params) {
  if (is.null(draw_mat) || !is.matrix(draw_mat)) {
    stop("draw_mat must be a numeric matrix.", call. = FALSE)
  }
  bulk_params <- as.character(bulk_params %||% character(0))
  if (!length(bulk_params)) return(list())

  cn <- colnames(draw_mat)

  infer_K_from_params <- function() {
    k_vals <- integer(0)
    for (nm in bulk_params) {
      idx <- as.integer(sub(paste0("^", nm, "\\[([0-9]+)\\]$"), "\\1",
                            cn[grepl(paste0("^", nm, "\\[[0-9]+\\]$"), cn)]))
      if (length(idx)) k_vals <- c(k_vals, idx)
    }
    if (!length(k_vals)) return(NA_integer_)
    as.integer(max(k_vals, na.rm = TRUE))
  }

  infer_K_from_weights <- function() {
    idx <- as.integer(sub("^w\\[([0-9]+)\\]$", "\\1", cn[grepl("^w\\[[0-9]+\\]$", cn)]))
    if (length(idx)) return(as.integer(max(idx, na.rm = TRUE)))
    idx <- as.integer(sub("^weights\\[([0-9]+)\\]$", "\\1",
                          cn[grepl("^weights\\[[0-9]+\\]$", cn)]))
    if (length(idx)) return(as.integer(max(idx, na.rm = TRUE)))
    NA_integer_
  }

  K <- infer_K_from_params()
  if (!is.finite(K) || K < 1L) K <- infer_K_from_weights()
  if (!is.finite(K) || K < 1L) stop("Could not infer K for bulk parameter draws.", call. = FALSE)

  out <- list()
  for (nm in bulk_params) {
    blk <- .indexed_block(draw_mat, nm, K = K, allow_missing = TRUE)
    if (!is.null(blk)) out[[nm]] <- blk
  }
  out
}

#' Reserved-name validation for NIMBLE
#' @param names Character vector of names to validate.
#' @param context Human-readable context for error messages.
#' @keywords internal
#' @noRd
.validate_nimble_reserved_names <- function(names, context = "names") {
  if (is.null(names) || !length(names)) return(invisible(TRUE))
  names <- as.character(names)
  names <- names[!is.na(names) & nzchar(names)]
  if (!length(names)) return(invisible(TRUE))

  reserved <- c(
    "if", "else", "for", "while", "repeat", "break", "next", "in",
    "function", "return",
    "true", "false", "null", "na", "nan", "inf",
    "na_integer_", "na_real_", "na_character_", "na_complex_",
    "t", "f"
  )

  bad <- unique(names[tolower(names) %in% reserved])
  if (length(bad)) {
    stop(sprintf(
      "%s include reserved NIMBLE keywords: %s. Rename columns (e.g., if -> x_if).",
      context,
      paste(bad, collapse = ", ")
    ), call. = FALSE)
  }

  invisible(TRUE)
}

#' Extract nimbleCode from bundle code
#'
#' @details
#' Bundle objects may store generated NIMBLE code either directly as a
#' `nimbleCode` object or inside a small wrapper list used for package storage.
#' This helper normalizes those storage conventions so downstream code can work
#' with the underlying code object without repeatedly checking both cases.
#' @keywords internal
.extract_nimble_code <- function(code) {
  .register_nimble_exports(envir = .GlobalEnv)
  if (is.list(code) && !inherits(code, "nimbleCode")) {
    if (!is.null(code$nimble)) return(code$nimble)
    if (!is.null(code$code)) return(code$code)
  }
  code
}

#' Wrap nimbleCode for bundle storage
#'
#' @details
#' This helper is the inverse of `.extract_nimble_code()`. It stores a raw
#' `nimbleCode` object inside a lightweight list so bundle objects can carry code
#' alongside other metadata without ambiguity about the field layout.
#' @keywords internal
.wrap_nimble_code <- function(code) {
  if (is.list(code) && !inherits(code, "nimbleCode")) return(code)
  list(nimble = code)
}

# ---- Plot styling helpers (internal) ----
.plot_palette <- function(n = 8L) {
  base <- c(
    "#0072B2", # blue
    "#D55E00", # vermillion
    "#009E73", # green
    "#CC79A7", # purple
    "#56B4E9", # sky blue
    "#E69F00", # orange
    "#000000", # black
    "#999999"  # gray
  )
  n <- as.integer(n %||% length(base))
  if (n <= length(base)) return(base[seq_len(n)])
  rep_len(base, n)
}

.plot_theme <- function() {
  ggplot2::theme_minimal(base_size = 12) +
    ggplot2::theme(
      panel.grid.minor = ggplot2::element_blank(),
      plot.title = ggplot2::element_text(face = "bold"),
      legend.position = "top"
    )
}

.strip_fill_scales <- function(p) {
  if (!inherits(p, "ggplot")) return(p)
  if (is.null(p$scales) || !length(p$scales$scales)) return(p)
  keep <- vapply(p$scales$scales, function(s) {
    !("fill" %in% (s$aesthetics %||% character()))
  }, logical(1))
  p$scales$scales <- p$scales$scales[keep]
  p
}

.safe_ggplotly <- function(p) {
  tryCatch(
    plotly::ggplotly(p),
    error = function(e) {
      p2 <- .strip_fill_scales(p)
      tryCatch(plotly::ggplotly(p2), error = function(e2) p2)
    }
  )
}

.wrap_plotly <- function(p) {
  # Default to static output; opt in via options(CausalMixGPD.plotly = TRUE).
  if (isTRUE(getOption("CausalMixGPD.plotly", FALSE)) &&
      requireNamespace("plotly", quietly = TRUE)) {
    if (is.list(p) && !inherits(p, "ggplot")) {
      # List of plots - wrap each, preserve class
      result <- lapply(p, function(plt) {
        if (inherits(plt, "ggplot")) .safe_ggplotly(plt) else plt
      })
      # Preserve original class attributes
      class(result) <- class(p)
      result
    } else if (inherits(p, "ggplot")) {
      # Single ggplot - wrap it
      .safe_ggplotly(p)
    } else {
      # Not a ggplot, return as-is
      p
    }
  } else {
    # plotly not available - return original
    p
  }
}

#' Prediction helpers
#'
#' @keywords internal
#' @noRd
.resolve_predict_id <- function(x, id = NULL) {
  id_vec <- NULL
  x_out <- x

  if (is.null(x_out)) {
    if (!is.null(id)) {
      stop("'id' requires 'x'/'newdata' to supply prediction rows.", call. = FALSE)
    }
    return(list(x = x_out, id = id_vec))
  }

  if (is.data.frame(x_out)) {
    if (is.character(id) && length(id) == 1L) {
      if (!id %in% names(x_out)) {
        stop(sprintf("Column '%s' not found in 'x'.", id), call. = FALSE)
      }
      id_vec <- x_out[[id]]
      x_out <- x_out[, names(x_out) != id, drop = FALSE]
    } else {
      if ("id" %in% names(x_out)) {
        x_out <- x_out[, names(x_out) != "id", drop = FALSE]
      }
    }
  } else if (is.character(id) && length(id) == 1L) {
    stop("'id' must reference a column in data.frame 'x'.", call. = FALSE)
  }

  if (!is.null(id) && !is.character(id)) {
    n_x <- nrow(as.matrix(x_out))
    if (length(id) != n_x) {
      stop("Length of 'id' must match nrow(x).", call. = FALSE)
    }
    id_vec <- id
  }

  list(x = x_out, id = id_vec)
}

.reorder_predict_cols <- function(df) {
  if (!is.data.frame(df)) return(df)
  cols <- names(df)
  profile_col <- if ("profile" %in% cols) "profile" else NULL
  id_col <- if ("id" %in% cols) "id" else NULL
  idx_col <- if ("index" %in% cols) "index" else if ("y" %in% cols) "y" else NULL
  est_col <- if ("estimate" %in% cols) {
    "estimate"
  } else if ("density" %in% cols) {
    "density"
  } else if ("survival" %in% cols) {
    "survival"
  } else {
    NULL
  }
  base <- c(profile_col, id_col, idx_col, est_col, intersect(c("lower", "upper"), cols))
  base <- base[!is.na(base) & nzchar(base)]
  rest <- setdiff(cols, base)
  df[, c(base, rest), drop = FALSE]
}

.values_to_long_df <- function(x, id = NULL, value_name = "value") {
  if (is.data.frame(x)) {
    return(.reorder_predict_cols(x))
  }

  if (is.null(dim(x))) {
    out <- data.frame(
      draw = seq_along(x),
      row.names = NULL
    )
    out[[value_name]] <- as.numeric(x)
    return(out)
  }

  mat <- as.matrix(x)
  n_row <- nrow(mat)
  n_col <- ncol(mat)
  id_use <- if (!is.null(id) && length(id) == n_row) id else seq_len(n_row)

  out <- data.frame(
    id = rep(id_use, each = n_col),
    draw = rep(seq_len(n_col), times = n_row),
    row.names = NULL
  )
  out[[value_name]] <- as.vector(t(mat))
  .reorder_predict_cols(out)
}

#' Coerce fit object to standardized data frame
#'
#' Converts various fit object formats (vector, matrix, data.frame) to a

#' standardized data.frame with columns: estimate, lower, upper, and id.
#' Used by plot methods to ensure consistent input handling.
#'
#' @param fit A fit object: vector, matrix, or data.frame.
#' @param n_pred Optional expected number of prediction rows for validation.
#' @param probs Optional vector of probability levels (for QTE-style objects).
#' @return A data.frame with columns: estimate, lower, upper, id (and optionally index).
#' @keywords internal
#' @noRd
.coerce_fit_df <- function(fit, n_pred = NULL, probs = NULL) {

  # Case 1: Already a data.frame

  if (is.data.frame(fit)) {
    df <- fit
    # Ensure required columns exist
    if (!"estimate" %in% names(df)) {
      if ("fit" %in% names(df)) {
        df$estimate <- df$fit
      } else if ("density" %in% names(df)) {
        df$estimate <- df$density
      } else if ("survival" %in% names(df)) {
        df$estimate <- df$survival
      } else {
        num_cols <- names(df)[vapply(df, is.numeric, logical(1))]
        num_cols <- setdiff(num_cols, c("id", "index", "y", "ps", "lower", "upper"))
        if (length(num_cols)) {
          df$estimate <- df[[num_cols[1]]]
        } else {
          df$estimate <- NA_real_
        }
      }
    }
    if (!"lower" %in% names(df)) df$lower <- NA_real_
    if (!"upper" %in% names(df)) df$upper <- NA_real_
    if (!"id" %in% names(df)) {
      df$id <- seq_len(nrow(df))
    }
    return(.reorder_predict_cols(df))
  }

  # Case 2: Matrix (rows = observations, cols = quantiles or estimate/lower/upper)
  if (is.matrix(fit)) {
    nr <- nrow(fit)
    nc <- ncol(fit)
    cn <- colnames(fit)

    # Check if columns are named estimate/lower/upper
    if (!is.null(cn) && all(c("estimate", "lower", "upper") %in% cn)) {
      df <- as.data.frame(fit)
      df$id <- seq_len(nr)
      return(.reorder_predict_cols(df))
    }

    # If probs provided, assume matrix is n_pred x length(probs) of estimates
    if (!is.null(probs) && nc == length(probs)) {
      # Expand to long format: id x index
      df <- data.frame(
        id = rep(seq_len(nr), each = nc),
        index = rep(probs, times = nr),
        estimate = as.vector(t(fit)),
        lower = NA_real_,
        upper = NA_real_
      )
      return(.reorder_predict_cols(df))
    }

    # Default: treat first column as estimate
    df <- data.frame(
      id = seq_len(nr),
      estimate = fit[, 1],
      lower = if (nc >= 2) fit[, 2] else NA_real_,
      upper = if (nc >= 3) fit[, 3] else NA_real_
    )
    return(.reorder_predict_cols(df))
  }

  # Case 3: Numeric vector -> single-row or multi-row df
  if (is.numeric(fit)) {
    n <- length(fit)
    df <- data.frame(
      id = seq_len(n),
      estimate = fit,
      lower = NA_real_,
      upper = NA_real_
    )
    return(.reorder_predict_cols(df))
  }

  # Fallback: error

  stop("Cannot coerce fit to data.frame: unsupported type.", call. = FALSE)
}

#' Nimble helpers
#'
#' @keywords internal
#' @noRd
#' @importFrom nimble nimNumeric
NULL

#' Backend label formatter
#' @param x Backend key.
#' @return Character label.
#' @keywords internal
#' @noRd
.backend_label <- function(x) {
  switch(
    x,
    sb  = "Stick-Breaking Process",
    crp = "Chinese Restaurant Process",
    x
  )
}

# Deterministic stick-breaking map used by NIMBLE code generation.
stick_breaking <- nimble::nimbleFunction(
  run = function(v = double(1)) {
    returnType(double(1))
    K <- length(v) + 1L
    w <- numeric(K)
    remainder <- 1
    for (j in 1:(K - 1L)) {
      w[j] <- v[j] * remainder
      remainder <- remainder * (1 - v[j])
    }
    w[K] <- remainder
    return(w)
  }
)


#' Kernel label formatter
#' @param x Kernel key.
#' @return Character label.
#' @keywords internal
#' @noRd
.kernel_label <- function(x) {
  switch(
    x,
    normal    = "Normal Distribution",
    gamma     = "Gamma Distribution",
    lognormal = "Lognormal Distribution",
    laplace   = "Laplace Distribution",
    invgauss  = "Inverse Gaussian Distribution",
    amoroso   = "Amoroso Distribution",
    cauchy    = "Cauchy Distribution",
    x
  )
}


#'  Get epsilon value from object spec/meta or argument
#'
#' @details
#' Many downstream summaries truncate mixture components according to the bundle
#' or fit-level `epsilon` setting. This helper centralizes that lookup so an
#' explicit function argument overrides the stored fit metadata, and the package
#' fallback is used only when neither source is present.
#'
#' @param object A mixgpd_fit object.
#' @param epsilon Numeric; if provided, overrides object spec/meta.
#' @keywords internal
.get_epsilon <- function(object, epsilon = NULL) {
  if (!is.null(epsilon)) return(as.numeric(epsilon)[1])
  spec <- object$spec %||% list()
  meta <- spec$meta %||% list()
  as.numeric(object$epsilon %||% meta$epsilon %||% 0.025)
}

#' Truncate component draws in a draws matrix
#'
#' @details
#' Posterior draw matrices often contain more components than are effectively
#' needed for reporting. This helper applies the package truncation rule
#' draw-by-draw, keeping the retained component blocks, associated weights, and
#' any linked coefficient matrices aligned after reordering and truncation.
#'
#' The bookkeeping attached to the returned matrix records both the cumulative
#' mass rule and the per-component weight rule so later summaries can report how
#' many components were effectively retained.
#' @param object A mixgpd_fit object.
#' @param mat Numeric matrix of draws (iter x parameters).
#' @param epsilon Numeric in [0,1). Truncation level.
#' @return Numeric matrix with truncated components.
#' @keywords internal
.truncate_draws_matrix_components <- function(object, mat, epsilon) {
  if (!is.numeric(epsilon) || length(epsilon) != 1L || is.na(epsilon) || epsilon < 0 || epsilon >= 1) {
    stop("epsilon must be a single number in [0, 1).", call. = FALSE)
  }

  spec <- object$spec %||% list()
  meta <- spec$meta %||% list()
  plan <- spec$plan %||% list()
  bulk_plan <- plan$bulk %||% list()
  gpd_plan <- plan$gpd %||% list()
  has_ps <- !is.null(plan$ps)
  backend_raw <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  is_spliced <- identical(backend_raw, "spliced")
  backend <- if (is_spliced) "crp" else backend_raw
  kernel  <- meta$kernel  %||% spec$kernel$key %||% "<unknown>"

  kdef <- get_kernel_registry()[[kernel]]
  if (is.null(kdef)) stop("Kernel not found in registry: ", kernel, call. = FALSE)
  bulk_params <- kdef$bulk_params %||% character(0)

  cn <- colnames(mat)
  has_z <- any(grepl("^z\\[[0-9]+\\]$", cn))
  has_w <- any(grepl("^w\\[[0-9]+\\]$", cn))
  has_weights <- any(grepl("^weights\\[[0-9]+\\]$", cn))

  component_matrix_bases <- unique(c(
    unlist(lapply(names(bulk_plan), function(nm) {
      ent <- bulk_plan[[nm]] %||% list()
      if (identical(ent$mode %||% "", "link")) paste0("beta_", nm) else character(0)
    }), use.names = FALSE),
    if (is_spliced) {
      unlist(lapply(c("threshold", "tail_scale", "tail_shape"), function(nm) {
        ent <- gpd_plan[[nm]] %||% list()
        if (identical(ent$mode %||% "", "link")) paste0("beta_", nm) else character(0)
      }), use.names = FALSE)
    } else {
      character(0)
    }
  ))
  component_aux_vector_bases <- unique(c(
    if (isTRUE(has_ps)) {
      unlist(lapply(names(bulk_plan), function(nm) {
        ent <- bulk_plan[[nm]] %||% list()
        if (identical(ent$mode %||% "", "link")) paste0("beta_ps_", nm) else character(0)
      }), use.names = FALSE)
    } else {
      character(0)
    }
  ))

  infer_K_from_bulk <- function() {
    if (length(bulk_params) < 1) return(NA_integer_)
    firstp <- bulk_params[1]
    idx <- as.integer(sub(paste0("^", firstp, "\\[([0-9]+)\\]$"), "\\1",
                          cn[grepl(paste0("^", firstp, "\\[[0-9]+\\]$"), cn)]))
    if (!length(idx)) return(NA_integer_)
    as.integer(max(idx, na.rm = TRUE))
  }

  infer_K_from_w <- function() {
    widx <- as.integer(sub("^w\\[([0-9]+)\\]$", "\\1", cn[grepl("^w\\[[0-9]+\\]$", cn)]))
    if (length(widx)) return(as.integer(max(widx, na.rm = TRUE)))
    widx <- as.integer(sub("^weights\\[([0-9]+)\\]$", "\\1", cn[grepl("^weights\\[[0-9]+\\]$", cn)]))
    if (length(widx)) return(as.integer(max(widx, na.rm = TRUE)))
    NA_integer_
  }

  infer_K_from_z <- function() {
    if (!has_z) return(NA_integer_)
    Z <- .indexed_block(mat, "z")
    zmax <- suppressWarnings(max(Z, na.rm = TRUE))
    if (!is.finite(zmax) || zmax < 1L) return(NA_integer_)
    as.integer(zmax)
  }

  infer_K_from_component_vectors <- function() {
    bases <- unique(c(
      bulk_params,
      if (is_spliced) c("threshold", "tail_scale", "tail_shape") else character(0)
    ))
    idx_all <- integer(0)
    for (nm in bases) {
      idx <- as.integer(sub(paste0("^", nm, "\\[([0-9]+)\\]$"), "\\1",
                            cn[grepl(paste0("^", nm, "\\[[0-9]+\\]$"), cn)]))
      if (length(idx)) idx_all <- c(idx_all, idx)
    }
    if (!length(idx_all)) return(NA_integer_)
    as.integer(max(idx_all, na.rm = TRUE))
  }

  infer_K_from_component_matrices <- function() {
    if (!length(component_matrix_bases)) return(NA_integer_)
    idx_all <- integer(0)
    for (nm in component_matrix_bases) {
      idx <- as.integer(sub(paste0("^", nm, "\\[([0-9]+),\\s*[0-9]+\\]$"), "\\1",
                            cn[grepl(paste0("^", nm, "\\[[0-9]+,\\s*[0-9]+\\]$"), cn)]))
      if (length(idx)) idx_all <- c(idx_all, idx)
    }
    if (!length(idx_all)) return(NA_integer_)
    as.integer(max(idx_all, na.rm = TRUE))
  }

  if (identical(backend, "sb")) {
    K <- infer_K_from_w()
    if (!is.finite(K) || K < 1L) K <- infer_K_from_bulk()
    if (!is.finite(K) || K < 1L) stop("Could not infer K for SB weights.", call. = FALSE)
  } else if (identical(backend, "crp")) {
    K <- infer_K_from_bulk()
    if (!is.finite(K) || K < 1L) K <- infer_K_from_component_vectors()
    if (!is.finite(K) || K < 1L) K <- infer_K_from_component_matrices()
    if (!is.finite(K) || K < 1L) K <- infer_K_from_z()
    if (!is.finite(K) || K < 1L) K <- infer_K_from_w()
    if (!is.finite(K) || K < 1L) stop("Could not infer Kmax from component parameter draws.", call. = FALSE)
  } else {
    stop("Unknown backend: ", backend, call. = FALSE)
  }
  K <- as.integer(K)

  S <- nrow(mat)
  W <- matrix(0.0, nrow = S, ncol = K)
  if (identical(backend, "sb")) {
    if (has_w) {
      W <- .indexed_block(mat, "w", K = K)
    } else if (has_weights) {
      W <- .indexed_block(mat, "weights", K = K)
    } else if (has_z) {
      Z <- .indexed_block(mat, "z")
      storage.mode(Z) <- "integer"
      for (s in 1:S) {
        z_s <- Z[s, ]
        z_s <- z_s[is.finite(z_s)]
        z_s <- z_s[z_s >= 1 & z_s <= K]
        if (length(z_s)) W[s, ] <- tabulate(z_s, nbins = K) / length(z_s)
      }
    } else {
      stop("Could not derive SB weights: expected w[i]/weights[i] or z[i] in samples.", call. = FALSE)
    }
  } else {
    if (has_z) {
      Z <- .indexed_block(mat, "z")
      storage.mode(Z) <- "integer"
      for (s in 1:S) {
        z_s <- Z[s, ]
        z_s <- z_s[is.finite(z_s)]
        z_s <- z_s[z_s >= 1 & z_s <= K]
        if (length(z_s)) W[s, ] <- tabulate(z_s, nbins = K) / length(z_s)
      }
    } else if (has_w) {
      W <- .indexed_block(mat, "w", K = K)
    } else if (has_weights) {
      W <- .indexed_block(mat, "weights", K = K)
    } else {
      stop("Backend requires z[i] in samples to derive weights.", call. = FALSE)
    }
  }

  component_vector_draws <- list()
  for (nm in bulk_params) {
    blk <- .indexed_block(mat, nm, K = K, allow_missing = TRUE)
    if (!is.null(blk)) component_vector_draws[[nm]] <- blk
  }
  if (is_spliced) {
    for (nm in c("threshold", "tail_scale", "tail_shape")) {
      blk <- .indexed_block(mat, nm, K = K, allow_missing = TRUE)
      if (!is.null(blk)) component_vector_draws[[nm]] <- blk
    }
  }
  for (nm in component_aux_vector_bases) {
    blk <- .indexed_block(mat, nm, K = K, allow_missing = TRUE)
    if (!is.null(blk)) component_vector_draws[[nm]] <- blk
  }
  component_matrix_draws <- list()
  for (base in component_matrix_bases) {
    blk <- .indexed_block_matrix(mat, base, K = K, allow_missing = TRUE)
    if (!is.null(blk)) component_matrix_draws[[base]] <- blk
  }
  component_vector_bases <- names(component_vector_draws)

  # ----- per-draw truncation (sorted by weight; keep params aligned) -----
  S <- nrow(mat)
  ks <- integer(S)
  k_weight_vec <- integer(S)
  k_cum_vec <- integer(S)
  ords <- vector("list", S)
  w_list <- vector("list", S)
  p_list <- vector("list", S)

  for (s in 1:S) {
    params_s <- lapply(component_vector_bases, function(nm) as.numeric(component_vector_draws[[nm]][s, ]))
    names(params_s) <- component_vector_bases
    tr <- .truncate_components_one_draw(w = as.numeric(W[s, ]), params = params_s, epsilon = epsilon)

    ks[s] <- tr$k
    k_weight_vec[s] <- tr$k_weight %||% tr$k
    k_cum_vec[s] <- tr$k_cum %||% tr$k
    ords[[s]] <- tr$ord
    w_list[[s]] <- tr$weights
    p_list[[s]] <- tr$params
  }

  # Fixed K across draws: keep only components selected in all draws
  Kt <- min(ks)
  if (!is.finite(Kt) || Kt < 1L) Kt <- 1L

  # ----- build new matrix: keep all NON-component columns + replace component blocks -----
  drop_pat <- c(
    "^weights\\[[0-9]+\\]$",
    "^w\\[[0-9]+\\]$",
    paste0("^", unique(component_vector_bases), "\\[[0-9]+\\]$"),
    paste0("^", unique(names(component_matrix_draws)), "\\[[0-9]+,\\s*[0-9]+\\]$")
  )
  drop_hit <- rep(FALSE, length(cn))
  for (pp in drop_pat) drop_hit <- drop_hit | grepl(pp, cn)
  keep_cn <- cn[!drop_hit]

  out <- mat[, keep_cn, drop = FALSE]

  # add truncated weights + params (ranked by decreasing weight)
  w_out <- matrix(0.0, nrow = S, ncol = Kt)
  colnames(w_out) <- paste0("w[", seq_len(Kt), "]")

  out_params <- list()
  for (nm in component_vector_bases) {
    tmp <- matrix(NA_real_, nrow = S, ncol = Kt)
    colnames(tmp) <- paste0(nm, "[", seq_len(Kt), "]")
    out_params[[nm]] <- tmp
  }
  out_mats <- list()
  for (base in names(component_matrix_draws)) {
    Pj <- dim(component_matrix_draws[[base]])[3L]
    tmp <- matrix(NA_real_, nrow = S, ncol = Kt * Pj)
    colnames(tmp) <- unlist(lapply(seq_len(Kt), function(k) {
      paste0(base, "[", k, ",", seq_len(Pj), "]")
    }), use.names = FALSE)
    out_mats[[base]] <- tmp
  }

  for (s in 1:S) {
    w_s <- w_list[[s]]
    k_s <- min(length(w_s), Kt)
    if (k_s < 1L) next
    w_out[s, seq_len(k_s)] <- w_s[seq_len(k_s)]
    for (nm in component_vector_bases) {
      out_params[[nm]][s, seq_len(k_s)] <- p_list[[s]][[nm]][seq_len(k_s)]
    }
    if (length(out_mats)) {
      ord_s <- ords[[s]]
      for (base in names(out_mats)) {
        mat_s <- component_matrix_draws[[base]][s, , , drop = TRUE]
        if (is.null(dim(mat_s))) {
          mat_s <- matrix(mat_s, nrow = K, ncol = 1L)
        }
        mat_ord <- mat_s[ord_s, , drop = FALSE]
        Pj <- ncol(mat_ord)
        out_mats[[base]][s, seq_len(k_s * Pj)] <- as.vector(t(mat_ord[seq_len(k_s), , drop = FALSE]))
      }
    }
  }

  out <- cbind(out, w_out)
  for (nm in component_vector_bases) out <- cbind(out, out_params[[nm]])
  for (base in names(out_mats)) out <- cbind(out, out_mats[[base]])

  attr(out, "truncation") <- list(
    k = ks,
    k_weight = k_weight_vec,
    k_cum = k_cum_vec,
    Kt = Kt,
    epsilon = epsilon
  )

  out
}

#' Validate a fitted object
#'
#' @details
#' This is a lightweight structural check used by multiple internal helpers. It
#' verifies that the object inherits from `mixgpd_fit` and that posterior draws
#' are available in one of the expected storage locations before later code tries
#' to summarize or predict from the fit.
#' @param object A fitted object.
#' @return Invisibly TRUE, otherwise errors.
#' @keywords internal
.validate_fit <- function(object) {
  if (!inherits(object, "mixgpd_fit")) {
    stop("Object must inherit from class 'mixgpd_fit'.", call. = FALSE)
  }
  smp <- object$mcmc$samples %||% object$samples
  if (is.null(smp)) stop("No samples found in object$mcmc$samples (or object$samples).", call. = FALSE)
  invisible(TRUE)
}


#' Safely coerce MCMC samples to coda::mcmc.list
#'
#' @details
#' Downstream summary and plotting code relies on the `coda` interface. This
#' helper validates the fit, locates the stored posterior draws, and converts a
#' single-chain `mcmc` object into an `mcmc.list` so later code can treat the
#' one-chain and multi-chain cases uniformly.
#' @param object A mixgpd_fit.
#' @return A coda::mcmc.list object.
#' @keywords internal
.get_samples_mcmclist <- function(object) {
  .validate_fit(object)
  smp <- object$mcmc$samples %||% object$samples

  if (!requireNamespace("coda", quietly = TRUE)) {
    stop("Package 'coda' is required for summary/plot. Install it.", call. = FALSE)
  }

  if (inherits(smp, "mcmc")) smp <- coda::mcmc.list(smp)
  if (!inherits(smp, "mcmc.list")) {
    stop("Expected samples to be coda::mcmc or coda::mcmc.list.", call. = FALSE)
  }
  smp
}


#' Extract posterior draws as a numeric matrix (iter x parameters)
#'
#' @details
#' This helper stacks the posterior chains into one numeric matrix, optionally
#' removes stick-breaking `v` variables, and then applies the component
#' truncation rule controlled by `epsilon`. The result is the standardized draw
#' representation used by most internal summary and prediction helpers.
#' @param object A mixgpd_fit.
#' @param drop_v Logical; if TRUE, drop stick-breaking v parameters.
#' @return Numeric matrix of draws.
#' @keywords internal
.extract_draws_matrix <- function(object, drop_v = TRUE, epsilon = NULL) {
  smp <- .get_samples_mcmclist(object)
  mat <- do.call(rbind, lapply(smp, function(ch) as.matrix(ch)))
  if (is.null(colnames(mat))) stop("Draw matrix has no column names.", call. = FALSE)

  if (isTRUE(drop_v)) {
    cn <- colnames(mat)
    keep <- !(grepl("^v\\[", cn) | cn == "v")
    mat <- mat[, keep, drop = FALSE]
  }

  eps <- .get_epsilon(object, epsilon)
  mat <- .truncate_draws_matrix_components(object, mat, eps)

  mat
}


#' Get number of observations used in fitting
#'
#' @details
#' The fitted object may carry the response in slightly different storage slots
#' depending on how it was built. This helper centralizes the lookup and returns
#' the effective training sample size used by summaries and print methods.
#' @param object A mixgpd_fit.
#' @return Integer n.
#' @keywords internal
.get_nobs <- function(object) {
  if (!is.null(object$data) && !is.null(object$data$y)) return(length(object$data$y))
  if (!is.null(object$y)) return(length(object$y))
  NA_integer_
}

#' Safely coerce MCMC samples to a numeric matrix
#'
#' @details
#' This is the matrix-oriented companion to `.get_samples_mcmclist()`. It can
#' either stack all chains or keep only the first chain, always removes
#' stick-breaking `v` variables, applies the standard component truncation rule,
#' and optionally filters to an exact set of parameter names.
#' @param object A mixgpd_fit.
#' @param pars Optional character vector of parameter names to keep (exact match).
#' @return Numeric matrix of draws (iter x parameters).
#' @keywords internal
.extract_draws <- function(object, pars = NULL, chains = c("stack", "first"), epsilon = NULL) {
  .validate_fit(object)
  chains <- match.arg(chains)

  smp <- .get_samples_mcmclist(object)

  mats <- lapply(smp, function(ch) {
    m <- as.matrix(ch)
    storage.mode(m) <- "double"
    m
  })

  if (chains == "first") {
    mat <- mats[[1]]
  } else {
    # common columns only
    cn <- Reduce(intersect, lapply(mats, colnames))
    mats <- lapply(mats, function(m) m[, cn, drop = FALSE])
    mat <- do.call(rbind, mats)
  }

  # drop v's always
  cn0 <- colnames(mat)
  keep0 <- !(grepl("^v\\[", cn0) | cn0 == "v")
  mat <- mat[, keep0, drop = FALSE]

  eps <- .get_epsilon(object, epsilon)
  mat <- .truncate_draws_matrix_components(object, mat, eps)

  if (!is.null(pars)) {
    miss <- setdiff(pars, colnames(mat))
    if (length(miss)) stop("Unknown params: ", paste(miss, collapse = ", "), call. = FALSE)
    mat <- mat[, pars, drop = FALSE]
  }

  mat
}

#' Summarize truncation results from draws
#'
#' @details
#' After `.extract_draws()` applies draw-level component truncation, the chosen
#' number of retained components is stored as an attribute. This helper condenses
#' that bookkeeping into min, median, and max summaries for the effective
#' retained component count under both the cumulative-mass and per-weight
#' criteria.
#' @param object A mixgpd_fit.
#' @param epsilon Numeric; optional override.
#' @return List with k summary.
#' @keywords internal
.truncation_info <- function(object, epsilon = NULL) {
  mat <- .extract_draws(object, pars = NULL, chains = "stack", epsilon = epsilon)
  tr <- attr(mat, "truncation") %||% list()
  k <- tr$k %||% integer(0)
  k_weight <- tr$k_weight %||% integer(0)
  k_cum <- tr$k_cum %||% integer(0)
  if (!length(k)) {
    return(list(
      k_min = NA_integer_, k_median = NA_integer_, k_max = NA_integer_,
      k_weight_min = NA_integer_, k_weight_median = NA_integer_, k_weight_max = NA_integer_,
      k_cum_min = NA_integer_, k_cum_median = NA_integer_, k_cum_max = NA_integer_,
      Kt = NA_integer_
    ))
  }
  list(
    k_min = min(k),
    k_median = as.integer(stats::median(k)),
    k_max = max(k),
    k_weight_min = if (length(k_weight)) min(k_weight) else NA_integer_,
    k_weight_median = if (length(k_weight)) as.integer(stats::median(k_weight)) else NA_integer_,
    k_weight_max = if (length(k_weight)) max(k_weight) else NA_integer_,
    k_cum_min = if (length(k_cum)) min(k_cum) else NA_integer_,
    k_cum_median = if (length(k_cum)) as.integer(stats::median(k_cum)) else NA_integer_,
    k_cum_max = if (length(k_cum)) max(k_cum) else NA_integer_,
    Kt = tr$Kt %||% max(k)
  )
}


#' Format a short header for printing
#'
#' @details
#' This helper builds the short header used by `print.mixgpd_fit()`. It extracts
#' model identity, training size, truncation size, epsilon, and stored MCMC
#' settings into a compact character vector so higher-level print methods do not
#' duplicate formatting logic.
#' @param x A mixgpd_fit.
#' @return Character vector lines.
#' @keywords internal
.format_fit_header <- function(x) {
  spec <- x$spec %||% list()
  meta <- spec$meta %||% list()

  backend <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  kernel  <- meta$kernel  %||% spec$kernel$key %||% "<unknown>"

  gpd_val <- meta$GPD %||% meta$gpd %||% spec$dispatch$GPD
  gpd_txt <- if (isTRUE(gpd_val)) "TRUE" else if (identical(gpd_val, FALSE)) "FALSE" else "<unknown>"

  y <- x$data$y %||% x$y %||% NULL
  n <- if (!is.null(y)) length(y) else (meta$N %||% spec$N %||% NA_integer_)
  Kmax <- meta$Kmax %||% spec$Kmax %||% NA_integer_

  eps <- .get_epsilon(x, epsilon = NULL)
  lines <- c(
    sprintf("MixGPD fit | backend: %s | kernel: %s | GPD tail: %s",
            .backend_label(backend), .kernel_label(kernel), gpd_txt),
    sprintf("n = %s | components = %s | epsilon = %s",
            ifelse(is.na(n), "<unknown>", n),
            ifelse(is.na(meta$components %||% NA_integer_), "<unknown>", meta$components),
            ifelse(is.na(eps), "<unknown>", eps))
  )

  m <- x$mcmc %||% list()
  it <- m$niter %||% NA_integer_
  nb <- m$nburnin %||% NA_integer_
  th <- m$thin %||% NA_integer_
  ch <- m$nchains %||% NA_integer_
  if (!all(is.na(c(it, nb, th, ch)))) {
    lines <- c(lines, sprintf("MCMC: niter=%s, nburnin=%s, thin=%s, nchains=%s",
                              ifelse(is.na(it), "?", it),
                              ifelse(is.na(nb), "?", nb),
                              ifelse(is.na(th), "?", th),
                              ifelse(is.na(ch), "?", ch)))
  }

  lines
}

#' Summarize posterior draws for selected parameters
#'
#' @details
#' This helper powers the one-arm summary methods. It extracts the retained draw
#' matrix, chooses a default set of non-redundant parameters when `pars` is not
#' supplied, and computes posterior means, standard deviations, quantiles,
#' effective sample sizes, and Gelman diagnostics when available.
#'
#' The resulting table is parameter oriented rather than prediction oriented. It
#' is the internal workhorse behind `summary.mixgpd_fit()`.
#' @param object mixgpd_fit
#' @param pars character vector; if NULL uses all non-v parameters
#' @param probs quantiles to report
#' @return data.frame with mean/sd/quantiles + ess/rhat where available
#' @keywords internal
.summarize_posterior <- function(object, pars = NULL, probs = c(0.025, 0.5, 0.975)) {
  stopifnot(inherits(object, "mixgpd_fit"))

  if (!requireNamespace("coda", quietly = TRUE)) stop("Need 'coda'.", call. = FALSE)

  mat <- .extract_draws(object, pars = NULL, chains = "stack", epsilon = NULL)
  eps <- .get_epsilon(object, epsilon = NULL)

  if (is.null(pars)) {
    pars <- colnames(mat)

    spec <- object$spec %||% list()
    plan <- spec$plan %||% list()
    meta <- spec$meta %||% list()
    is_spliced <- identical(meta$backend %||% spec$dispatch$backend %||% "", "spliced")
    bulk <- plan$bulk %||% list()
    gpd <- plan$gpd %||% list()

    cn <- pars
    keep <- cn %in% "alpha"
    keep <- keep | grepl("^w\\[[0-9]+\\]$", cn)

    for (nm in names(bulk)) {
      ent <- bulk[[nm]] %||% list()
      mode <- ent$mode %||% NA_character_
      if (identical(mode, "link")) {
        keep <- keep | grepl(paste0("^beta_", nm, "\\["), cn)
      } else {
        keep <- keep | grepl(paste0("^", nm, "\\[[0-9]+\\]$"), cn)
      }
    }

    if (!is.null(gpd$threshold)) {
      thr_mode <- gpd$threshold$mode %||% NA_character_
      if (identical(thr_mode, "link")) {
        keep <- keep | grepl("^beta_threshold\\[", cn)
        if (!is.null(gpd$threshold$link_dist) &&
            identical(gpd$threshold$link_dist$dist, "lognormal")) {
          keep <- keep | cn == "sdlog_u"
        }
      } else {
        keep <- keep | cn == "threshold" | grepl("^threshold\\[[0-9]+\\]$", cn)
      }
    }

    if (!is.null(gpd$tail_scale)) {
      ts_mode <- gpd$tail_scale$mode %||% NA_character_
      if (identical(ts_mode, "link")) {
        keep <- keep | grepl("^beta_tail_scale\\[", cn)
      } else if (ts_mode %in% c("dist", "fixed")) {
        if (is_spliced) {
          keep <- keep | grepl("^tail_scale\\[[0-9]+\\]$", cn)
        } else {
          keep <- keep | cn == "tail_scale"
        }
      }
    }

    if (!is.null(gpd$tail_shape)) {
      tsh_mode <- gpd$tail_shape$mode %||% NA_character_
      if (identical(tsh_mode, "link")) {
        keep <- keep | grepl("^beta_tail_shape\\[", cn)
      } else if (is_spliced) {
        keep <- keep | grepl("^tail_shape\\[[0-9]+\\]$", cn)
      } else {
        keep <- keep | cn == "tail_shape"
      }
    }

    pars <- cn[keep]
    mat <- mat[, pars, drop = FALSE]
  } else {
    pars <- gsub("^weight\\[", "w[", pars)
    miss <- setdiff(pars, colnames(mat))
    if (length(miss)) stop("Unknown params: ", paste(miss, collapse = ", "), call. = FALSE)
    mat <- mat[, pars, drop = FALSE]
  }
  wpars <- pars[grepl("^w\\[[0-9]+\\]$", pars)]
  if (length(wpars)) {
    pars <- c(wpars, setdiff(pars, wpars))
    mat <- mat[, pars, drop = FALSE]
  }
  if (length(wpars) && is.finite(eps) && eps > 0) {
    wmat <- mat[, wpars, drop = FALSE]
    wmat[wmat < eps] <- NA_real_
    keep_w <- apply(wmat, 2, function(v) any(is.finite(v)))
    w_keep <- wpars[keep_w]
    if (length(w_keep)) {
      mat[, w_keep] <- wmat[, w_keep, drop = FALSE]
      pars <- c(w_keep, setdiff(pars, wpars))
      mat <- mat[, pars, drop = FALSE]
    } else {
      pars <- setdiff(pars, wpars)
      mat <- mat[, pars, drop = FALSE]
    }
  }

  backend <- object$spec$meta$backend %||% object$spec$dispatch$backend %||% ""
  thr_cols <- grep("^threshold\\[[0-9]+\\]$", colnames(mat), value = TRUE)
  if (!identical(backend, "spliced") && length(thr_cols) >= 1) {
    thr_vec <- if (length(thr_cols) == 1) {
      as.numeric(mat[, thr_cols[1]])
    } else {
      rowMeans(mat[, thr_cols, drop = FALSE], na.rm = TRUE)
    }
    mat <- mat[, setdiff(colnames(mat), thr_cols), drop = FALSE]
    mat <- cbind(mat, threshold = thr_vec)
    pars <- c(setdiff(pars, thr_cols), "threshold")
  }

  meanv <- colMeans(mat, na.rm = TRUE)
  sdv   <- apply(mat, 2, stats::sd, na.rm = TRUE)

  qmat <- t(apply(mat, 2, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
  colnames(qmat) <- paste0("q", formatC(probs, format = "f", digits = 3))

  out <- data.frame(
    parameter = pars,
    mean = as.numeric(meanv[pars]),
    sd   = as.numeric(sdv[pars]),
    qmat[pars, , drop = FALSE],
    stringsAsFactors = FALSE
  )

  ess_vec <- rep(NA_real_, ncol(mat))
  for (j in seq_len(ncol(mat))) {
    v <- mat[, j]
    v <- v[is.finite(v)]
    if (length(v) >= 3L) {
      ess_vec[j] <- as.numeric(coda::effectiveSize(coda::mcmc(v)))
    }
  }
  names(ess_vec) <- colnames(mat)
  out$ess <- as.numeric(ess_vec[out$parameter])
  out$parameter <- sub("^w\\[", "weights[", out$parameter)

  rownames(out) <- NULL
  out
}

#' Resolve kernel dispatch functions (scalar)
#' Dispatch returns raw scalar nimbleFunctions for codegen; do not wrap.
#'
#' @details
#' This helper resolves the density, distribution, quantile, random-generation,
#' and mean functions implied by a kernel, backend, and GPD setting. The result
#' is intentionally scalar and wrapper-free because it is used in code-generation
#' contexts where NIMBLE expects raw function objects rather than vectorized R
#' adapters.
#' @param spec_or_fit mixgpd_fit or spec list
#' @return List with d/p/q/r/mean/mean_trunc functions and bulk_params.
#' @keywords internal
.get_dispatch_scalar <- function(spec_or_fit, backend_override = NULL, gpd_override = NULL) {
  spec <- spec_or_fit
  if (inherits(spec_or_fit, "mixgpd_fit")) {
    spec <- spec_or_fit$spec %||% list()
  }

  meta <- spec$meta %||% list()
  backend <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  if (!is.null(backend_override)) backend <- backend_override
  kernel <- meta$kernel %||% spec$kernel$key %||% "<unknown>"
  GPD <- isTRUE(meta$GPD %||% spec$dispatch$GPD)
  if (!is.null(gpd_override)) GPD <- isTRUE(gpd_override)

  kdef <- get_kernel_registry()[[kernel]]
  if (is.null(kdef)) stop(sprintf("Kernel '%s' not found in registry.", kernel), call. = FALSE)
  if (isTRUE(GPD) && isFALSE(kdef$allow_gpd)) stop(sprintf("Kernel '%s' does not allow GPD.", kernel), call. = FALSE)

  backend_key <- match.arg(backend, choices = allowed_backends)
  dispatch <- kdef[[backend_key]]
  if (is.null(dispatch)) {
    stop(sprintf("Missing %s dispatch in kernel registry.", backend_key), call. = FALSE)
  }

  d_name <- if (isTRUE(GPD)) {
    dispatch$d_gpd
  } else {
    dispatch$d %||% dispatch$d_base
  }
  if (is.na(d_name) || !nzchar(d_name)) {
    stop(sprintf("Missing %s dispatch for kernel '%s'.", backend_key, kernel), call. = FALSE)
  }

  p_name <- sub("^d", "p", d_name)
  q_name <- sub("^d", "q", d_name)
  r_name <- sub("^d", "r", d_name)
  mean_name <- if (!isTRUE(GPD)) dispatch$mean %||% dispatch$mean_base %||% NULL else NULL
  mean_trunc_name <- dispatch$mean_trunc %||% dispatch$mean_trunc_base %||% NULL

  ns_pkg <- getNamespace("CausalMixGPD")
  ns_stats <- getNamespace("stats")
  ns_nimble <- getNamespace("nimble")

  .resolve_fun <- function(fname, kernel) {
    # Try PascalCase NIMBLE function first (for NIMBLE model code)
    if (exists(fname, envir = ns_pkg, inherits = FALSE)) {
      return(get(fname, envir = ns_pkg))
    }
    if (exists(fname, envir = ns_stats, inherits = FALSE)) {
      return(get(fname, envir = ns_stats))
    }
    if (exists(fname, envir = ns_nimble, inherits = FALSE)) {
      return(get(fname, envir = ns_nimble))
    }
    # Fallback for predictions: use lowercase R wrappers if PascalCase NIMBLE function
    # not found (e.g., in some build contexts). Lowercase wrappers are vectorized
    # and work for predictions but should not be used in NIMBLE model code.
    fname_lower <- tolower(fname)
    if (exists(fname_lower, envir = ns_pkg, inherits = FALSE)) {
      return(get(fname_lower, envir = ns_pkg))
    }
    stop(sprintf("Missing function '%s' for kernel '%s'.", fname, kernel), call. = FALSE)
  }

  d_fun <- .wrap_density_fun(.resolve_fun(d_name, kernel))
  p_fun <- .wrap_cdf_fun(.resolve_fun(p_name, kernel))
  q_fun <- .wrap_quantile_fun(.resolve_fun(q_name, kernel))
  r_fun <- .wrap_rng_fun(.resolve_fun(r_name, kernel))
  mean_fun <- if (!is.null(mean_name) && nzchar(mean_name)) .resolve_fun(mean_name, kernel) else NULL
  mean_trunc_fun <- if (!is.null(mean_trunc_name) && nzchar(mean_trunc_name)) .resolve_fun(mean_trunc_name, kernel) else NULL

  if (isTRUE(attr(d_fun, "vectorized_wrapper")) ||
      isTRUE(attr(p_fun, "vectorized_wrapper")) ||
      isTRUE(attr(q_fun, "vectorized_wrapper")) ||
      isTRUE(attr(r_fun, "vectorized_wrapper"))) {
    stop("Scalar dispatch unexpectedly received vectorized wrappers.", call. = FALSE)
  }

  list(d = d_fun, p = p_fun, q = q_fun, r = r_fun, mean = mean_fun, mean_trunc = mean_trunc_fun, bulk_params = kdef$bulk_params)
}

#' Resolve kernel dispatch functions
#' Dispatch returns vector-aware d/p/q and n-aware r via wrappers; do not mutate namespace.
#'
#' @details
#' This is the prediction-oriented companion to `.get_dispatch_scalar()`. It
#' starts from the same kernel dispatch lookup, then wraps the scalar functions
#' so they can accept vector inputs and the package's preferred argument naming
#' conventions in ordinary R evaluation.
#' @param spec_or_fit mixgpd_fit or spec list
#' @return List with d/p/q/r/mean/mean_trunc functions and bulk_params.
#' @keywords internal
.get_dispatch <- function(spec_or_fit, backend_override = NULL, gpd_override = NULL) {
  scalar <- .get_dispatch_scalar(spec_or_fit, backend_override = backend_override, gpd_override = gpd_override)
  list(
    d = .wrap_scalar_first_arg(scalar$d, "x"),
    p = .wrap_scalar_p(scalar$p),
    q = .wrap_scalar_first_arg(scalar$q, "p"),
    r = .wrap_scalar_r(scalar$r),
    mean = scalar$mean,
    mean_trunc = scalar$mean_trunc,
    bulk_params = scalar$bulk_params
  )
}

#' Compute credible or HPD interval from posterior draws
#'
#' Dispatches to either equal-tailed quantile intervals or highest posterior
#' density (HPD) intervals using \code{coda::HPDinterval()}.
#'
#' @param draws Numeric vector of posterior draws.
#' @param level Numeric; credible level (e.g., 0.95 for 95 percent interval).
#' @param type Character; interval type:
#'   \itemize{
#'     \item \code{"credible"}: equal-tailed quantile intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @return Named numeric vector with \code{lower} and \code{upper}.
#' @keywords internal
#' @noRd
.compute_interval <- function(draws, level = 0.95, type = c("credible", "hpd")) {
  type <- match.arg(type)
  draws <- draws[is.finite(draws)]
  if (length(draws) < 2L) {
    return(c(lower = NA_real_, upper = NA_real_))
  }

  if (type == "credible") {
    probs <- c((1 - level) / 2, (1 + level) / 2)
    q <- stats::quantile(draws, probs = probs, na.rm = TRUE)
    c(lower = unname(q[1]), upper = unname(q[2]))
  } else {
    # HPD via coda
    if (!requireNamespace("coda", quietly = TRUE)) {
      stop("Package 'coda' is required for HPD intervals.", call. = FALSE)
    }
    hpd <- coda::HPDinterval(coda::as.mcmc(draws), prob = level)
    c(lower = hpd[1, "lower"], upper = hpd[1, "upper"])
  }
}

#' Summarize posterior draws (mean + quantiles)
#'
#' @details
#' The last dimension of `draws` is interpreted as the posterior-draw dimension.
#' This helper collapses that dimension to posterior means and interval summaries,
#' while preserving the leading dimensions of the input object. It is used
#' throughout prediction and treatment-effect code to turn per-draw evaluations
#' into reported posterior summaries.
#' @param draws Numeric vector, matrix, or array with draws in last dimension.
#' @param probs Numeric quantile probs.
#' @param interval Character or NULL; interval type:
#'   \itemize{
#'     \item \code{NULL}: no interval
#'     \item \code{"credible"} (default): equal-tailed quantile
#'       intervals
#'     \item \code{"hpd"}: highest posterior density intervals
#'   }
#' @return List with estimate, lower, upper, and q.
#' @keywords internal
.posterior_summarize <- function(draws, probs = c(0.025, 0.5, 0.975),
                                 interval = "credible") {
  # Handle NULL interval (no interval computation)
  if (is.null(interval)) {
    interval <- "none"
  } else {
    interval <- match.arg(interval, choices = c("credible", "hpd"))
  }
  probs <- as.numeric(probs)

  # Compute credible level from probs (for HPD)
  level <- probs[length(probs)] - probs[1]
  if (!is.finite(level) || level <= 0 || level >= 1) level <- 0.95

  # Helper to compute intervals for a row of draws
  .row_interval <- function(row) {
    if (interval == "none") {
      c(NA_real_, NA_real_)
    } else if (interval == "credible") {
      q <- stats::quantile(row, probs = probs, na.rm = TRUE, names = FALSE)
      c(q[1], q[length(probs)])
    } else {
      iv <- .compute_interval(row, level = level, type = "hpd")
      c(iv["lower"], iv["upper"])
    }
  }

  if (is.null(dim(draws))) {
    mat <- matrix(as.numeric(draws), nrow = 1)
    qmat <- t(apply(mat, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
    iv <- .row_interval(as.numeric(draws))
    return(list(
      estimate = rowMeans(mat, na.rm = TRUE),
      lower = iv[1],
      upper = iv[2],
      q = qmat
    ))
  }

  dims <- dim(draws)
  if (length(dims) == 2L) {
    qmat <- t(apply(draws, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
    ivmat <- t(apply(draws, 1, .row_interval))
    return(list(
      estimate = rowMeans(draws, na.rm = TRUE),
      lower = ivmat[, 1],
      upper = ivmat[, 2],
      q = qmat
    ))
  }

  Sdim <- dims[length(dims)]
  mat <- matrix(draws, nrow = prod(dims[-length(dims)]), ncol = Sdim)
  qmat <- t(apply(mat, 1, stats::quantile, probs = probs, na.rm = TRUE, names = FALSE))
  ivmat <- t(apply(mat, 1, .row_interval))
  estimate <- rowMeans(mat, na.rm = TRUE)
  lower <- ivmat[, 1]
  upper <- ivmat[, 2]
  dim(estimate) <- dims[-length(dims)]
  dim(lower) <- dims[-length(dims)]
  dim(upper) <- dims[-length(dims)]
  list(estimate = estimate, lower = lower, upper = upper, q = qmat)
}

#' Detect the first present argument name in dots.
#'
#' @details
#' Some wrapped scalar functions accept either `q` or `x` as their first numeric
#' argument depending on the original API. This helper inspects `...` and returns
#' the first candidate name that is actually present so wrapper code can map user
#' input onto the target function signature.
#' @keywords internal
.detect_first_present <- function(dots, candidates = c("q", "x")) {
  for (nm in candidates) {
    if (!is.null(dots[[nm]])) return(nm)
  }
  stop("Expected one of: ", paste(candidates, collapse = ", "), call. = FALSE)
}

#' Wrap scalar first-argument functions to handle vector inputs.
#'
#' @details
#' Many low-level distribution helpers are scalar in their first argument. This
#' wrapper lifts such functions to vector inputs by evaluating the scalar
#' function repeatedly and combining the results into either a numeric vector or
#' a matrix, depending on the length of the original return value.
#' @keywords internal
.wrap_scalar_first_arg <- function(fun, first_arg_name) {
  if (isTRUE(attr(fun, "vectorized_wrapper"))) return(fun)
  force(fun)
  force(first_arg_name)
  wrapper <- function(...) {
    dots <- list(...)
    if (!first_arg_name %in% names(dots)) {
      stop("Missing required argument: ", first_arg_name, call. = FALSE)
    }
    vec <- dots[[first_arg_name]]
    if (length(vec) <= 1L) return(do.call(fun, dots))

    dots[[first_arg_name]] <- vec[1]
    one <- do.call(fun, dots)
    if (length(one) <= 1L) {
      return(vapply(vec, function(v) {
        dots[[first_arg_name]] <- v
        do.call(fun, dots)
      }, numeric(1)))
    }

    mat <- vapply(vec, function(v) {
      dots[[first_arg_name]] <- v
      as.numeric(do.call(fun, dots))
    }, numeric(length(one)))
    t(mat)
  }
  attr(wrapper, "vectorized_wrapper") <- TRUE
  wrapper
}

#' Wrap scalar CDF to handle q/x naming and vector inputs.
#'
#' @details
#' Different scalar CDF helpers use either `q` or `x` for their evaluation
#' argument. This wrapper normalizes those naming differences and then applies the
#' same vector-lifting strategy used elsewhere so prediction code can call the
#' resulting function consistently.
#' @keywords internal
.wrap_scalar_p <- function(fun) {
  if (isTRUE(attr(fun, "vectorized_wrapper"))) return(fun)
  force(fun)
  wrapper <- function(...) {
    dots <- list(...)
    given <- .detect_first_present(dots, candidates = c("q", "x"))

    formal_names <- names(formals(fun)) %||% character()
    target <- if ("q" %in% formal_names && !"x" %in% formal_names) {
      "q"
    } else if ("x" %in% formal_names && !"q" %in% formal_names) {
      "x"
    } else {
      given
    }

    if (!identical(given, target)) {
      dots[[target]] <- dots[[given]]
      dots[[given]] <- NULL
    }

    vec <- dots[[target]]
    if (length(vec) <= 1L) return(do.call(fun, dots))

    dots[[target]] <- vec[1]
    one <- do.call(fun, dots)
    if (length(one) <= 1L) {
      return(vapply(vec, function(v) {
        dots[[target]] <- v
        do.call(fun, dots)
      }, numeric(1)))
    }

    mat <- vapply(vec, function(v) {
      dots[[target]] <- v
      as.numeric(do.call(fun, dots))
    }, numeric(length(one)))
    t(mat)
  }
  attr(wrapper, "vectorized_wrapper") <- TRUE
  wrapper
}

#' Wrap scalar RNG to handle n > 1.
#'
#' @details
#' Random-generation helpers in the package are scalar-at-a-time. This wrapper
#' promotes them to the standard `n` interface by repeating the scalar generator
#' and returning either a numeric vector or a matrix of generated values,
#' depending on the length of one draw.
#' @keywords internal
.wrap_scalar_r <- function(fun) {
  if (isTRUE(attr(fun, "vectorized_wrapper"))) return(fun)
  force(fun)
  wrapper <- function(...) {
    dots <- list(...)
    if (!("n" %in% names(dots))) stop("Missing required argument: n", call. = FALSE)
    n <- as.integer(dots$n)
    dots$n <- NULL
    if (is.na(n) || n < 0L) stop("n must be a non-negative integer.", call. = FALSE)
    if (n == 0L) {
      one <- do.call(fun, c(list(n = 1L), dots))
      if (length(one) <= 1L) return(numeric(0))
      return(matrix(numeric(0), nrow = 0, ncol = length(one)))
    }
    if (n == 1L) return(do.call(fun, c(list(n = 1L), dots)))

    one <- do.call(fun, c(list(n = 1L), dots))
    if (length(one) <= 1L) {
      return(vapply(seq_len(n), function(i) {
        do.call(fun, c(list(n = 1L), dots))
      }, numeric(1)))
    }

    mat <- vapply(seq_len(n), function(i) {
      as.numeric(do.call(fun, c(list(n = 1L), dots)))
    }, numeric(length(one)))
    t(mat)
  }
  attr(wrapper, "vectorized_wrapper") <- TRUE
  wrapper
}

#' Truncate and reorder mixture components by cumulative weight mass
#'
#' @details
#' This helper operates on one posterior draw at a time. It first orders mixture
#' components by decreasing weight, then keeps the smallest effective subset of
#' components implied by the package truncation rule, and finally renormalizes
#' the retained weights so they sum to one.
#'
#' The same permutation is applied to every component-specific parameter vector in
#' `params`, which keeps the retained parameter blocks aligned with the retained
#' weights.
#' @param w Numeric vector of component weights (length K).
#' @param params Named list of numeric vectors, each length K (component-specific params).
#' @param epsilon Numeric in [0,1). Keep the smallest k s.t. cumweight >= 1-epsilon.
#' @return A list with reordered+truncated weights/params and bookkeeping.
#' @keywords internal
.truncate_components_one_draw <- function(w, params, epsilon = 0.01) {
  stopifnot(is.numeric(w), length(w) >= 1L)
  if (!is.numeric(epsilon) || length(epsilon) != 1L || is.na(epsilon) || epsilon < 0 || epsilon >= 1) {
    stop("epsilon must be a single number in [0, 1).", call. = FALSE)
  }
  if (!is.list(params) || length(params) == 0L) params <- list()

  K <- length(w)

  # Validate params lengths
  for (nm in names(params)) {
    v <- params[[nm]]
    if (!is.numeric(v) || length(v) != K) {
      stop("params[['", nm, "']] must be numeric and length K (= length(w)).", call. = FALSE)
    }
  }

  # Sort by decreasing weight
  ord <- order(w, decreasing = TRUE)
  w_sorted <- w[ord]
  params_sorted <- lapply(params, function(v) v[ord])

  # Two criteria:
  # (1) per-component minimum weight >= epsilon
  # (2) cumulative mass >= 1 - epsilon
  # Keep the smaller k that satisfies either criterion.
  keep_idx <- which(w_sorted >= epsilon)
  k_weight <- if (length(keep_idx)) max(keep_idx) else 0L

  cw <- cumsum(w_sorted)
  k_cum <- which(cw >= (1 - epsilon))[1]
  if (is.na(k_cum)) k_cum <- K

  k_keep <- min(k_weight, k_cum)
  if (!is.finite(k_keep) || k_keep < 1L) k_keep <- 1L

  keep <- seq_len(k_keep)
  w_keep <- w_sorted[keep]

  # Adjust the smallest kept weight to make the kept weights sum to 1
  if (length(w_keep) > 1L) {
    min_idx <- which.min(w_keep)
    w_keep[min_idx] <- w_keep[min_idx] + (1 - sum(w_keep))
  } else {
    w_keep[1] <- 1
  }

  s <- sum(w_keep)
  if (!is.finite(s) || s <= 0) stop("Invalid weight sum after truncation.", call. = FALSE)

  params_keep <- lapply(params_sorted, function(v) v[keep])

  list(
    k = length(keep),
    k_weight = k_weight,
    k_cum = k_cum,
    ord = ord,
    weights = w_keep,
    params = params_keep
  )
}


#' Internal prediction engine: evaluate per posterior draw, then summarize.
#'
#' Project rules:
#' - density/survival: either provide both (x,y) or neither (defaults to training X and training y).
#' - quantile/sample/mean: y must be NULL; x may be provided (new X) or NULL (defaults to training X).
#' - CRP predictions use posterior weights derived from z for each draw.
#' - Stores per-draw results in object$cache$predict (environment) for reuse in treatment effects.
#'
#' @details
#' This is the main internal workhorse behind `predict.mixgpd_fit()` and the
#' causal effect helpers. It evaluates the requested predictive functional
#' separately for each retained posterior draw, using either explicit SB weights
#' or CRP weights reconstructed from latent cluster labels, and only then
#' collapses the draw-level results into posterior summaries.
#'
#' The helper also manages caching of per-draw predictive quantities because
#' treatment-effect functions repeatedly reuse the same arm-specific predictive
#' draws. That cache avoids recomputation while keeping the public prediction
#' interface simple.
#'
#' @keywords internal
.predict_mixgpd <- function(object,
                            x = NULL, y = NULL, ps = NULL, id = NULL,
                            type = c("density", "survival", "quantile", "sample", "mean", "rmean", "median", "fit"),
                            p = NULL, index = NULL, nsim = NULL,
                            level = 0.95,
                            interval = "credible",
                            probs = c(0.025, 0.5, 0.975),
                            store_draws = TRUE,
                            nsim_mean = 200L,
                            cutoff = NULL,
                            ndraws_pred = NULL,
                            chunk_size = NULL,
                            show_progress = TRUE,
                            ncores = 1L,
                            sample_draw_idx = NULL) {

  .validate_fit(object)
  type <- match.arg(type)

  id_info <- .resolve_predict_id(x, id = id)
  x <- id_info$x
  id_vec <- id_info$id

  # Handle interval: NULL means no interval
  compute_interval <- TRUE
  if (is.null(interval)) {
    compute_interval <- FALSE
    interval <- "credible"
  } else {
    interval <- match.arg(interval, choices = c("credible", "hpd"))
  }

  ncores <- as.integer(ncores)
  if (is.na(ncores) || ncores < 1L) stop("'ncores' must be an integer >= 1.", call. = FALSE)
  if (!is.null(ndraws_pred)) {
    ndraws_pred <- as.integer(ndraws_pred)[1L]
    if (!is.finite(ndraws_pred) || ndraws_pred < 1L) {
      stop("'ndraws_pred' must be NULL or a positive integer.", call. = FALSE)
    }
  }
  if (!is.null(chunk_size)) {
    chunk_size <- as.integer(chunk_size)[1L]
    if (!is.finite(chunk_size) || chunk_size < 1L) {
      stop("'chunk_size' must be NULL or a positive integer.", call. = FALSE)
    }
  }

  # Spec / meta
  spec <- object$spec %||% list()
  meta <- spec$meta %||% list()

  backend <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  kernel  <- meta$kernel  %||% spec$kernel$key %||% "<unknown>"
  GPD     <- isTRUE(meta$GPD %||% spec$dispatch$GPD)

  # Use mixture dispatch for prediction even when backend is CRP/spliced
  # For spliced backend with link-mode GPD params, component-level parameters
  # must be reconstructed from beta coefficients and newdata X values.
  pred_backend <- if (backend %in% c("crp", "spliced")) "sb" else backend
  is_spliced <- identical(backend, "spliced")

  # Training data
  Xtrain <- object$data$X %||% object$X %||% NULL
  ytrain <- object$data$y %||% object$y %||% NULL
  ps_train <- object$data$ps %||% NULL

  has_X <- isTRUE(meta$has_X %||% (!is.null(Xtrain)))

  # Validate X helper
  .validate_X_pred <- function(Xpred, Xtrain) {
    Xpred <- as.matrix(Xpred)
    storage.mode(Xpred) <- "double"
    if (anyNA(Xpred)) stop("Missing values (NA) found in 'x'.", call. = FALSE)

    if (!is.null(Xtrain)) {
      Xtrain <- as.matrix(Xtrain)

      if (!is.null(colnames(Xtrain)) && !is.null(colnames(Xpred))) {
        if (!setequal(colnames(Xpred), colnames(Xtrain))) {
          stop("Column names of 'x' do not match training design matrix.", call. = FALSE)
        }
        Xpred <- Xpred[, colnames(Xtrain), drop = FALSE]
      } else {
        if (ncol(Xpred) != ncol(Xtrain)) {
          stop("Number of columns in 'x' does not match training design matrix.", call. = FALSE)
        }
      }
    }
    Xpred
  }

  # Resolve MIX functions for kernel
  fns <- .get_dispatch(object, backend_override = pred_backend)
  bulk_params <- fns$bulk_params
  d_fun <- fns$d
  p_fun <- fns$p
  q_fun <- fns$q
  r_fun <- fns$r
  mean_fun <- fns$mean %||% NULL
  bulk_scalar <- .get_dispatch_scalar(object, backend_override = pred_backend, gpd_override = FALSE)
  bulk_p_fun <- bulk_scalar$p
  bulk_mean_fun <- bulk_scalar$mean %||% NULL
  bulk_mean_trunc_fun <- bulk_scalar$mean_trunc %||% NULL

  kdef <- get_kernel_registry()[[kernel]] %||% list()
  bulk_support <- kdef$bulk_support %||% list()

  .resolve_pkg_fun_local <- function(fname) {
    if (is.null(fname) || !nzchar(fname)) {
      return(NULL)
    }
    ns_pkg <- getNamespace("CausalMixGPD")
    if (exists(fname, envir = ns_pkg, inherits = FALSE)) {
      return(get(fname, envir = ns_pkg))
    }
    if (exists(fname, envir = .GlobalEnv, inherits = FALSE)) {
      return(get(fname, envir = .GlobalEnv))
    }
    NULL
  }

  if (GPD && (is.null(bulk_mean_trunc_fun) || !is.function(bulk_mean_trunc_fun))) {
    bulk_dispatch <- kdef[[pred_backend]] %||% kdef[[backend]] %||% list()
    bulk_mean_trunc_name <- bulk_dispatch$mean_trunc %||% bulk_dispatch$mean_trunc_base %||% NULL
    bulk_mean_trunc_fun <- .resolve_pkg_fun_local(bulk_mean_trunc_name)
  }

  # Link helper
  .apply_link <- function(eta, link, link_power = NULL) {
    link <- as.character(link %||% "identity")
    if (link == "identity") return(eta)
    if (link == "exp") return(exp(eta))
    if (link == "log") return(log(eta))
    if (link == "softplus") return(log1p(exp(eta)))
    if (link == "power") {
      if (is.null(link_power) || length(link_power) != 1L || !is.finite(as.numeric(link_power))) {
        stop("power link requires numeric link_power.", call. = FALSE)
      }
      pw <- as.numeric(link_power)
      return(eta ^ pw)
    }
    stop(sprintf("Unsupported link '%s'.", link), call. = FALSE)
  }

  # -----------------------------
  # Resolve inputs by type (contract)
  # -----------------------------
  Xpred <- NULL
  ygrid <- NULL
  pgrid <- NULL

  if (type %in% c("density", "survival")) {
    if (has_X) {
      if (is.null(x) && is.null(y)) {
        if (is.null(Xtrain)) stop("Training X not found in fit object.", call. = FALSE)
        if (is.null(ytrain)) stop("Training y not found in fit object.", call. = FALSE)
        Xpred <- Xtrain
        ygrid <- ytrain
      } else if (!is.null(x) && !is.null(y)) {
        Xpred <- x
        ygrid <- y
      } else {
        stop("For type='density'/'survival' with X, provide BOTH 'x' and 'y', or provide NEITHER to use training defaults.",
             call. = FALSE)
      }
    } else {
      if (!is.null(x)) stop("Unconditional model: 'x' is not allowed.", call. = FALSE)
      ygrid <- y %||% ytrain
      if (is.null(ygrid)) stop("No 'y' provided and training y not found.", call. = FALSE)
    }

    if (!is.null(Xpred)) Xpred <- .validate_X_pred(Xpred, Xtrain)
    ygrid <- as.numeric(ygrid)
    if (anyNA(ygrid)) stop("Missing values (NA) found in 'y'.", call. = FALSE)
  }

  if (type %in% c("quantile", "median")) {
    if (!is.null(p)) index <- p
    if (type == "median" && is.null(index)) index <- 0.5
    if (is.null(index)) stop("For type='quantile'/'median', provide 'index' (or 'p').", call. = FALSE)
    pgrid <- as.numeric(index)
    if (anyNA(pgrid) || any(pgrid <= 0 | pgrid >= 1)) stop("'index' must be in (0,1).", call. = FALSE)

    if (has_X) {
      if (is.null(x)) {
        if (is.null(Xtrain)) stop("Training X not found in fit object.", call. = FALSE)
        Xpred <- Xtrain
      } else {
        Xpred <- .validate_X_pred(x, Xtrain)
      }
    } else {
      if (!is.null(x)) stop("Unconditional model: 'x' is not allowed for quantiles.", call. = FALSE)
    }
  }

  if (type %in% c("sample", "fit", "mean", "rmean")) {
    if (has_X) {
      if (is.null(x)) {
        if (is.null(Xtrain)) stop("Training X not found in fit object.", call. = FALSE)
        Xpred <- Xtrain
      } else {
        Xpred <- .validate_X_pred(x, Xtrain)
      }
    } else {
      if (!is.null(x)) stop("Unconditional model: 'x' is not allowed.", call. = FALSE)
    }
  }

  # Prediction dimensions
  n_pred <- if (!is.null(Xpred)) nrow(Xpred) else 1L
  if (!is.null(id_vec) && length(id_vec) != n_pred) {
    stop("Length of 'id' must match the number of prediction rows.", call. = FALSE)
  }
  id_vals <- if (!is.null(id_vec)) id_vec else seq_len(n_pred)

  chunk_starts <- NULL
  if (!is.null(chunk_size) && !is.null(Xpred) && n_pred > chunk_size) {
    chunk_starts <- seq.int(1L, n_pred, by = chunk_size)
  }
  progress_total <- 4L + if (!is.null(chunk_starts)) length(chunk_starts) else 0L
  progress_ctx <- .cmgpd_progress_start(
    total_steps = progress_total,
    enabled = isTRUE(show_progress),
    quiet = FALSE,
    label = "predict_mixgpd"
  )
  on.exit(.cmgpd_progress_done(progress_ctx, final_label = NULL), add = TRUE)
  .cmgpd_progress_step(progress_ctx, "Validating prediction inputs")

  .return_out <- function(obj) {
    .cmgpd_progress_step(progress_ctx, "Assembling prediction output")
    obj
  }

  # -----------------------------
  # Posterior draws extraction
  # -----------------------------
  .cmgpd_progress_step(progress_ctx, "Extracting posterior draws")
  draw_mat <- .extract_draws_matrix(object)
  if (is.null(draw_mat) || !is.matrix(draw_mat) || nrow(draw_mat) < 2L) {
    stop("Posterior draws not found or malformed in fitted object.", call. = FALSE)
  }
  if (!is.null(ndraws_pred) && ndraws_pred < nrow(draw_mat)) {
    keep_idx <- sort(sample.int(nrow(draw_mat), size = ndraws_pred, replace = FALSE))
    draw_mat <- draw_mat[keep_idx, , drop = FALSE]
  }
  S <- nrow(draw_mat)

  # mixture weights + bulk parameter blocks
  # - W_draws: S x K
  # - bulk_draws: list(param -> S x K or S x 1 or S x ?)
  # - base_params: names of bulk params required by kernel
  W_draws <- .extract_weights(draw_mat, backend = pred_backend)
  bulk_draws <- .extract_bulk_params(draw_mat, bulk_params = bulk_params)
  base_params <- names(bulk_draws)

  if (!is.null(chunk_starts)) {
    .cmgpd_progress_step(progress_ctx, "Dispatching chunked prediction")
    parts <- vector("list", length(chunk_starts))
    for (ii in seq_along(chunk_starts)) {
      .cmgpd_progress_step(progress_ctx, sprintf("Computing chunk %d/%d", ii, length(chunk_starts)))
      i0 <- chunk_starts[ii]
      i1 <- min(n_pred, i0 + chunk_size - 1L)
      idx <- i0:i1
      x_chunk <- Xpred[idx, , drop = FALSE]
      y_chunk <- NULL
      if (type %in% c("density", "survival") && has_X) y_chunk <- ygrid[idx]
      ps_chunk <- if (!is.null(ps)) ps[idx] else NULL
      id_chunk <- if (!is.null(id_vals)) id_vals[idx] else NULL
      parts[[ii]] <- .predict_mixgpd(
        object = object,
        x = x_chunk,
        y = y_chunk,
        ps = ps_chunk,
        id = id_chunk,
        type = type,
        p = p,
        index = index,
        nsim = nsim,
        level = level,
        interval = interval,
        probs = probs,
        store_draws = store_draws,
        nsim_mean = nsim_mean,
        cutoff = cutoff,
        ndraws_pred = ndraws_pred,
        chunk_size = NULL,
        show_progress = FALSE,
        ncores = ncores,
        sample_draw_idx = sample_draw_idx
      )
    }
    out <- parts[[1L]]
    if (is.data.frame(out$fit)) {
      out$fit <- do.call(rbind, lapply(parts, function(z) z$fit))
      rownames(out$fit) <- NULL
    }
    if (!is.null(out$diagnostics)) {
      out$diagnostics$n_chunks <- length(parts)
    }
    return(.return_out(out))
  }

  .cmgpd_progress_step(progress_ctx, "Computing posterior summaries")

  # -----------------------------
  # Optional PS covariate (used only if model expects it)
  # -----------------------------
  if (!is.null(ps)) {
    ps <- as.numeric(ps)
    if (anyNA(ps)) stop("Missing values (NA) found in 'ps'.", call. = FALSE)
    if (has_X && length(ps) != n_pred) stop("Length of 'ps' must equal nrow(x).", call. = FALSE)
    if (!has_X && length(ps) != 1L) stop("Unconditional model: 'ps' must be scalar if provided.", call. = FALSE)
  }

  # -----------------------------
  # Link-mode parameter handling (conditional)
  # -----------------------------
  link_plan <- spec$dispatch$link_params %||% meta$link_params %||% list()
  if (!length(link_plan)) {
    bulk_plan <- spec$plan$bulk %||% list()
    for (nm in names(bulk_plan)) {
      ent <- bulk_plan[[nm]] %||% list()
      if (identical(ent$mode %||% "constant", "link")) {
        link_plan[[nm]] <- list(
          mode = "link",
          link = ent$link %||% "identity",
          link_power = ent$link_power %||% NULL
        )
      }
    }
  }
  link_params <- names(link_plan)
  P <- if (!is.null(Xpred)) ncol(Xpred) else 0L

  .compute_link_eta <- function(s) {
    if (!length(link_params)) return(list())
    out <- list()
    for (nm in link_params) {
      plan <- link_plan[[nm]] %||% list()
      mode <- plan$mode %||% "constant"
      if (identical(mode, "link")) {
        if (!has_X) stop("link-mode requires X.", call. = FALSE)
        link <- plan$link %||% "identity"
        pw   <- plan$link_power %||% NULL

        beta_cols <- grep(paste0("^beta_", nm, "\\[[0-9]+,\\s*[0-9]+\\]$"), colnames(draw_mat), value = TRUE)
        if (length(beta_cols)) {
          idx1 <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+),\\s*([0-9]+)\\]$"), "\\1", beta_cols))
          idx2 <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+),\\s*([0-9]+)\\]$"), "\\2", beta_cols))
          Kb <- max(idx1, na.rm = TRUE)
          Pb <- max(idx2, na.rm = TRUE)
          beta_mat <- matrix(NA_real_, nrow = Kb, ncol = Pb)
          for (j in seq_along(beta_cols)) {
            beta_mat[idx1[j], idx2[j]] <- draw_mat[s, beta_cols[j]]
          }
          eta_mat <- tcrossprod(Xpred, beta_mat)
          out[[nm]] <- .apply_link(eta_mat, link, pw)
        } else {
          beta_cols_1d <- grep(paste0("^beta_", nm, "\\[[0-9]+\\]$"), colnames(draw_mat), value = TRUE)
          if (length(beta_cols_1d)) {
            idx <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+)\\]$"), "\\1", beta_cols_1d))
            ord <- order(idx)
            beta_cols_1d <- beta_cols_1d[ord]
            Kb <- ncol(W_draws)
            Pb <- P
            if (length(beta_cols_1d) == Kb * Pb) {
              beta_vec <- as.numeric(draw_mat[s, beta_cols_1d])
              beta_mat <- matrix(beta_vec, nrow = Kb, ncol = Pb)
              eta_mat <- tcrossprod(Xpred, beta_mat)
              out[[nm]] <- .apply_link(eta_mat, link, pw)
            } else {
              beta_nm <- .indexed_block(draw_mat, paste0("beta_", nm), K = P) # S x P
              eta <- as.numeric(tcrossprod(Xpred, beta_nm[s, , drop = FALSE]))
              out[[nm]] <- matrix(as.numeric(.apply_link(eta, link, pw)), nrow = n_pred)
            }
          } else {
            beta_nm <- .indexed_block(draw_mat, paste0("beta_", nm), K = P) # S x P
            eta <- as.numeric(tcrossprod(Xpred, beta_nm[s, , drop = FALSE]))
            out[[nm]] <- matrix(as.numeric(.apply_link(eta, link, pw)), nrow = n_pred)
          }
        }
      } else {
        # constant (scalar per draw)
        if (!(nm %in% colnames(draw_mat))) stop(sprintf("'%s' not found in posterior draws.", nm), call. = FALSE)
        out[[nm]] <- matrix(rep(as.numeric(draw_mat[s, nm]), n_pred), nrow = n_pred)
      }
    }
    out
  }

  # -----------------------------
  # GPD tail plan extraction (if enabled)
  # -----------------------------
  tail_shape <- NULL
  threshold_mat <- NULL
  threshold_scalar <- NULL
  tail_scale <- NULL
  spliced_gpd_draws <- list()
  spliced_gpd_link <- list()
  spliced_gpd_obs <- list()

  if (GPD) {
    gpd_plan <- spec$dispatch$gpd %||% meta$gpd %||% list()

    if (is_spliced) {
      K_sp <- ncol(W_draws)
      for (nm in c("threshold", "tail_scale", "tail_shape")) {
        ent <- gpd_plan[[nm]] %||% list(mode = "dist")
        mode <- ent$mode %||% "dist"
        if (identical(mode, "link")) {
          if (!has_X) stop(sprintf("%s link-mode requires X.", nm), call. = FALSE)
          beta_arr <- .indexed_block_matrix(draw_mat, paste0("beta_", nm), K = K_sp, P = P, allow_missing = TRUE)
          if (is.null(beta_arr)) stop(sprintf("beta_%s not found in posterior draws.", nm), call. = FALSE)
          if (identical(nm, "threshold") &&
              !is.null(ent$link_dist) &&
              identical(ent$link_dist$dist, "lognormal")) {
            obs_arr <- .indexed_block_matrix(draw_mat, "threshold_i", allow_missing = TRUE)
            if (!is.null(obs_arr) &&
                identical(dim(obs_arr)[2L], n_pred) &&
                identical(dim(obs_arr)[3L], K_sp)) {
              spliced_gpd_obs[[nm]] <- obs_arr
            }
          }
          spliced_gpd_link[[nm]] <- list(
            beta = beta_arr,
            link = ent$link %||% if (identical(nm, "tail_shape")) "identity" else "exp",
            link_power = ent$link_power %||% NULL
          )
        } else {
          vals <- .indexed_block(draw_mat, nm, K = K_sp, allow_missing = TRUE)
          if (is.null(vals) && identical(mode, "fixed") && !is.null(ent$value)) {
            vals <- matrix(rep(as.numeric(ent$value), S * K_sp), nrow = S, ncol = K_sp)
          }
          if (is.null(vals)) stop(sprintf("%s not found in posterior draws.", nm), call. = FALSE)
          spliced_gpd_draws[[nm]] <- vals
        }
      }
    } else {
      # threshold
      thr_mode <- gpd_plan$threshold$mode %||% "constant"
      if (identical(thr_mode, "link")) {
        if (!has_X) stop("threshold link-mode requires X.", call. = FALSE)
        beta_thr <- .indexed_block(draw_mat, "beta_threshold", K = P)
        threshold_mat <- matrix(NA_real_, nrow = S, ncol = n_pred)
        thr_link <- gpd_plan$threshold$link %||% "exp"
        thr_power <- gpd_plan$threshold$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(Xpred, beta_thr[s, , drop = FALSE]))
          threshold_mat[s, ] <- as.numeric(.apply_link(eta, thr_link, thr_power))
        }
      } else {
        thr_cols <- grep("^threshold(\\b|_)", colnames(draw_mat), value = TRUE)
        if (length(thr_cols) == 0L && "threshold" %in% colnames(draw_mat)) thr_cols <- "threshold"
        if (length(thr_cols) == 0L) stop("threshold not found in posterior draws.", call. = FALSE)
        if (length(thr_cols) == 1L) {
          threshold_scalar <- as.numeric(draw_mat[, thr_cols])
        } else {
          threshold_scalar <- rowMeans(draw_mat[, thr_cols, drop = FALSE], na.rm = TRUE)
        }
      }

      has_beta_ts <- any(grepl("^beta_tail_scale\\[", colnames(draw_mat)))
      ts_mode <- gpd_plan$tail_scale$mode %||% if (has_beta_ts) "link" else "constant"
      if (identical(ts_mode, "link")) {
        if (!has_X) stop("tail_scale link-mode requires X.", call. = FALSE)
        beta_ts <- .indexed_block(draw_mat, "beta_tail_scale", K = P)
        tail_scale <- matrix(NA_real_, nrow = S, ncol = n_pred)
        ts_link <- gpd_plan$tail_scale$link %||% "exp"
        ts_power <- gpd_plan$tail_scale$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(Xpred, beta_ts[s, , drop = FALSE]))
          tail_scale[s, ] <- as.numeric(.apply_link(eta, ts_link, ts_power))
        }
      } else {
        if ("tail_scale" %in% colnames(draw_mat)) {
          tail_scale <- as.numeric(draw_mat[, "tail_scale"])
        } else if (!is.null(gpd_plan$tail_scale$value)) {
          tail_scale <- rep(as.numeric(gpd_plan$tail_scale$value), S)
        } else {
          stop("tail_scale not found in posterior draws.", call. = FALSE)
        }
      }

      has_beta_tsh <- any(grepl("^beta_tail_shape\\[", colnames(draw_mat)))
      tsh_mode <- gpd_plan$tail_shape$mode %||% if (has_beta_tsh) "link" else "constant"
      if (identical(tsh_mode, "link")) {
        if (!has_X) stop("tail_shape link-mode requires X.", call. = FALSE)
        beta_tsh <- .indexed_block(draw_mat, "beta_tail_shape", K = P)
        tail_shape <- matrix(NA_real_, nrow = S, ncol = n_pred)
        tsh_link <- gpd_plan$tail_shape$link %||% "identity"
        tsh_power <- gpd_plan$tail_shape$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(Xpred, beta_tsh[s, , drop = FALSE]))
          tail_shape[s, ] <- as.numeric(.apply_link(eta, tsh_link, tsh_power))
        }
      } else if ("tail_shape" %in% colnames(draw_mat)) {
        tail_shape <- as.numeric(draw_mat[, "tail_shape"])
      } else if (!is.null(gpd_plan$tail_shape$value)) {
        tail_shape <- rep(as.numeric(gpd_plan$tail_shape$value), S)
      } else {
        stop("tail_shape not found in posterior draws.", call. = FALSE)
      }
    }
  }

  .threshold_at <- function(s, i) {
    if (!is.null(threshold_mat)) return(threshold_mat[s, i])
    threshold_scalar[s]
  }

  .tail_scale_at <- function(s, i) {
    if (is.matrix(tail_scale)) return(tail_scale[s, i])
    tail_scale[s]
  }

  .tail_shape_at <- function(s, i) {
    if (is.matrix(tail_shape)) return(tail_shape[s, i])
    tail_shape[s]
  }

  # -----------------------------
  # Draw validation (NO silent repair)
  # -----------------------------
  .support_ok <- function(nm, v) {
    sup <- as.character(bulk_support[[nm]] %||% "")
    if (sup %in% c("positive_sd", "positive_scale", "positive_shape", "positive_location")) {
      return(all(is.finite(v) & (v > 0)))
    }
    all(is.finite(v))
  }

  .build_args0_or_null <- function(s) {
    w_s <- as.numeric(W_draws[s, ])
    if (!all(is.finite(w_s))) return(NULL)

    args0 <- if (pred_backend == "sb") list(w = w_s) else list()

    for (nm in base_params) {
      v <- as.numeric(bulk_draws[[nm]][s, ])
      if (!.support_ok(nm, v)) return(NULL)
      args0[[nm]] <- v
    }

    if (GPD && !is_spliced && !is.matrix(tail_shape)) {
      xi <- as.numeric(tail_shape[s])
      if (!is.finite(xi)) return(NULL)
      args0$tail_shape <- xi
    }

    args0
  }

  .draw_valid <- logical(S)
  for (s in seq_len(S)) {
    ok <- !is.null(.build_args0_or_null(s))
    if (ok && GPD) {
      if (is_spliced) {
        if (length(spliced_gpd_link)) {
          gpd_eta <- .compute_spliced_gpd_eta(s)
          for (nm in names(gpd_eta)) {
            vals <- as.numeric(gpd_eta[[nm]])
            if (identical(nm, "tail_scale")) {
              ok <- all(is.finite(vals) & (vals > 0))
            } else {
              ok <- all(is.finite(vals))
            }
            if (!ok) break
          }
        }
        if (!ok) {
          .draw_valid[s] <- FALSE
          next
        }
        for (nm in names(spliced_gpd_draws)) {
          vals <- as.numeric(spliced_gpd_draws[[nm]][s, ])
          if (identical(nm, "tail_scale")) {
            ok <- all(is.finite(vals) & (vals > 0))
          } else {
            ok <- all(is.finite(vals))
          }
          if (!ok) break
        }
      } else {
        if (!is.null(threshold_mat)) {
          if (!all(is.finite(threshold_mat[s, ]))) ok <- FALSE
        } else {
          if (!is.finite(threshold_scalar[s])) ok <- FALSE
        }
        if (ok) {
          if (is.matrix(tail_scale)) {
            ok <- all(is.finite(tail_scale[s, ]) & (tail_scale[s, ] > 0))
          } else {
            ok <- is.finite(tail_scale[s]) && (tail_scale[s] > 0)
          }
        }
        if (ok) {
          if (is.matrix(tail_shape)) {
            ok <- all(is.finite(tail_shape[s, ]))
          } else {
            ok <- is.finite(tail_shape[s])
          }
        }
      }
    }
    .draw_valid[s] <- ok
  }

  n_valid <- sum(.draw_valid)
  if (n_valid == 0L) stop("All posterior draws are invalid for prediction (non-finite or out-of-support parameters).", call. = FALSE)

  # Parallel helper
  .lapply_draws <- function(FUN) {
    idx <- seq_len(S)
    if (ncores == 1L) return(lapply(idx, FUN))

    if (!requireNamespace("future.apply", quietly = TRUE) ||
        !requireNamespace("future", quietly = TRUE)) {
      warning("ncores > 1 requested but 'future'/'future.apply' are unavailable; running sequentially.",
              call. = FALSE)
      return(lapply(idx, FUN))
    }

    old_plan <- future::plan()
    on.exit(future::plan(old_plan), add = TRUE)
    future::plan(future::multisession, workers = ncores)

    old_max <- getOption("future.globals.maxSize")
    on.exit(options(future.globals.maxSize = old_max), add = TRUE)
    options(future.globals.maxSize = Inf)

    future.apply::future_lapply(idx, FUN)
  }

  spliced_scalar <- if (is_spliced && GPD) .get_dispatch_scalar(object, backend_override = "crp") else NULL

  .row_component_value <- function(obj, i, k) {
    if (is.null(obj)) return(NA_real_)
    if (is.null(dim(obj))) return(as.numeric(obj[1L]))
    if (length(dim(obj)) != 2L) stop("Expected a scalar or matrix component block.", call. = FALSE)
    j <- if (ncol(obj) >= k && ncol(obj) > 1L) k else 1L
    as.numeric(obj[i, j])
  }

  .compute_spliced_gpd_eta <- function(s) {
    if (!length(spliced_gpd_link)) return(list())
    out <- list()
    for (nm in names(spliced_gpd_link)) {
      if (!is.null(spliced_gpd_obs[[nm]])) {
        out[[nm]] <- spliced_gpd_obs[[nm]][s, , , drop = TRUE]
        next
      }
      ent <- spliced_gpd_link[[nm]]
      beta_mat <- ent$beta[s, , , drop = TRUE]
      if (is.null(dim(beta_mat))) beta_mat <- matrix(beta_mat, nrow = ncol(W_draws), ncol = P)
      eta_mat <- tcrossprod(Xpred, beta_mat)
      out[[nm]] <- .apply_link(eta_mat, ent$link, ent$link_power)
    }
    out
  }

  .spliced_gpd_value_at <- function(nm, s, i, k, eta_cache) {
    if (!is.null(eta_cache[[nm]])) return(.row_component_value(eta_cache[[nm]], i, k))
    vals <- spliced_gpd_draws[[nm]]
    if (is.null(vals)) return(NA_real_)
    as.numeric(vals[s, k])
  }

  .spliced_component_args_or_null <- function(s, i, k, link_eta, gpd_eta) {
    args <- list()
    for (nm in bulk_params) {
      if (nm %in% link_params) {
        vv <- .row_component_value(link_eta[[nm]], i, k)
      } else if (nm %in% base_params) {
        vv <- as.numeric(bulk_draws[[nm]][s, k])
      } else {
        next
      }
      if (!.support_ok(nm, vv)) return(NULL)
      args[[nm]] <- as.numeric(vv)
    }

    args$threshold <- .spliced_gpd_value_at("threshold", s, i, k, gpd_eta)
    args$tail_scale <- .spliced_gpd_value_at("tail_scale", s, i, k, gpd_eta)
    args$tail_shape <- .spliced_gpd_value_at("tail_shape", s, i, k, gpd_eta)
    if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
      return(NULL)
    }
    args
  }

  .spliced_component_args_list_or_null <- function(s, i, link_eta, gpd_eta) {
    K <- ncol(W_draws)
    out <- vector("list", K)
    for (k in seq_len(K)) {
      out[[k]] <- .spliced_component_args_or_null(s, i, k, link_eta = link_eta, gpd_eta = gpd_eta)
      if (is.null(out[[k]])) return(NULL)
    }
    out
  }

  .normalize_weights_or_null <- function(w) {
    w <- as.numeric(w)
    if (!all(is.finite(w))) return(NULL)
    w[w < 0] <- 0
    sw <- sum(w)
    if (!is.finite(sw) || sw <= 0) return(NULL)
    w / sw
  }

  .spliced_density_survival_row <- function(s, i, yvals, type, link_eta, gpd_eta) {
    w_s <- .normalize_weights_or_null(W_draws[s, ])
    if (is.null(w_s)) return(rep(NA_real_, length(yvals)))
    comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
    if (is.null(comp_args)) return(rep(NA_real_, length(yvals)))

    out <- numeric(length(yvals))
    for (j in seq_along(yvals)) {
      yj <- yvals[j]
      vals <- vapply(seq_along(comp_args), function(k) {
        fun_args <- comp_args[[k]]
        if (type == "density") {
          as.numeric(do.call(spliced_scalar$d, c(list(x = yj, log = 0L), fun_args)))[1]
        } else {
          cdfv <- as.numeric(do.call(spliced_scalar$p, c(list(q = yj, lower.tail = 1L, log.p = 0L), fun_args)))[1]
          cdfv <- pmin(pmax(cdfv, 0), 1)
          1 - cdfv
        }
      }, numeric(1))
      out[j] <- sum(w_s * vals)
    }
    out
  }

  .spliced_quantile_one <- function(s, i, p0, link_eta, gpd_eta) {
    w_s <- .normalize_weights_or_null(W_draws[s, ])
    if (is.null(w_s)) return(NA_real_)
    comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
    if (is.null(comp_args)) return(NA_real_)

    q_lo <- vapply(comp_args, function(args) {
      suppressWarnings(tryCatch(as.numeric(do.call(spliced_scalar$q, c(list(p = 1e-8), args)))[1], error = function(e) NA_real_))
    }, numeric(1))
    q_hi <- vapply(comp_args, function(args) {
      suppressWarnings(tryCatch(as.numeric(do.call(spliced_scalar$q, c(list(p = 1 - 1e-8), args)))[1], error = function(e) NA_real_))
    }, numeric(1))
    q_mid <- vapply(comp_args, function(args) {
      suppressWarnings(tryCatch(as.numeric(do.call(spliced_scalar$q, c(list(p = p0), args)))[1], error = function(e) NA_real_))
    }, numeric(1))

    finite_seed <- c(q_lo[is.finite(q_lo)], q_hi[is.finite(q_hi)], q_mid[is.finite(q_mid)])
    if (!length(finite_seed)) return(NA_real_)

    cdf_mix <- function(y) {
      vals <- vapply(comp_args, function(args) {
        as.numeric(do.call(spliced_scalar$p, c(list(q = y, lower.tail = 1L, log.p = 0L), args)))[1]
      }, numeric(1))
      sum(w_s * pmin(pmax(vals, 0), 1))
    }

    lower <- min(finite_seed, na.rm = TRUE)
    upper <- max(finite_seed, na.rm = TRUE)
    step0 <- max(1, diff(range(finite_seed)), abs(lower), abs(upper), na.rm = TRUE)
    if (!is.finite(step0) || step0 <= 0) step0 <- 1

    f_lower <- cdf_mix(lower) - p0
    f_upper <- cdf_mix(upper) - p0

    step <- step0
    iter <- 0L
    while (is.finite(f_lower) && f_lower > 0 && iter < 60L) {
      lower <- lower - step
      f_lower <- cdf_mix(lower) - p0
      step <- step * 2
      iter <- iter + 1L
    }

    step <- step0
    iter <- 0L
    while (is.finite(f_upper) && f_upper < 0 && iter < 60L) {
      upper <- upper + step
      f_upper <- cdf_mix(upper) - p0
      step <- step * 2
      iter <- iter + 1L
    }

    if (!is.finite(f_lower) || !is.finite(f_upper) || f_lower > 0 || f_upper < 0) {
      return(NA_real_)
    }

    suppressWarnings(tryCatch(
      stats::uniroot(function(y) cdf_mix(y) - p0, interval = c(lower, upper), tol = .Machine$double.eps^0.5)$root,
      error = function(e) NA_real_
    ))
  }

  .spliced_sample_values <- function(s, i, n, link_eta, gpd_eta) {
    w_s <- .normalize_weights_or_null(W_draws[s, ])
    if (is.null(w_s)) return(rep(NA_real_, n))
    comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
    if (is.null(comp_args)) return(rep(NA_real_, n))

    kk <- sample.int(length(w_s), size = n, replace = TRUE, prob = w_s)
    out <- rep(NA_real_, n)
    for (ii in seq_len(n)) {
      out[ii] <- as.numeric(do.call(spliced_scalar$r, c(list(n = 1L), comp_args[[kk[ii]]])))[1]
    }
    out
  }

  .spliced_mean_infinite <- function(s, i, gpd_eta) {
    w_s <- as.numeric(W_draws[s, ])
    if (!all(is.finite(w_s))) return(FALSE)
    xi <- vapply(seq_len(ncol(W_draws)), function(k) {
      .spliced_gpd_value_at("tail_shape", s, i, k, gpd_eta)
    }, numeric(1))
    any((w_s > 0) & is.finite(xi) & (xi >= 1))
  }

  .clamp_prob <- function(x) {
    pmin(pmax(as.numeric(x), 0), 1)
  }

  .gpd_tail_mean_or_error <- function(threshold, tail_scale, tail_shape) {
    xi <- as.numeric(tail_shape)[1]
    u <- as.numeric(threshold)[1]
    sigma_u <- as.numeric(tail_scale)[1]
    if (!is.finite(u) || !is.finite(sigma_u) || sigma_u <= 0 || !is.finite(xi)) return(NA_real_)
    if (xi >= 1) {
      stop("Mean is not supported when the GPD tail has tail_shape (xi) >= 1; use type='rmean'.", call. = FALSE)
    }
    u + sigma_u / (1 - xi)
  }

  .analytic_spliced_mean_row <- function(s, i, link_eta, gpd_eta) {
    w_s <- .normalize_weights_or_null(W_draws[s, ])
    if (is.null(w_s)) return(NA_real_)
    comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
    if (is.null(comp_args) || is.null(bulk_mean_trunc_fun) || !is.function(bulk_mean_trunc_fun)) {
      return(NA_real_)
    }

    comp_mean <- vapply(seq_along(comp_args), function(k) {
      if (!is.finite(w_s[k]) || w_s[k] <= 0) return(0)
      args_k <- comp_args[[k]]
      threshold_k <- as.numeric(args_k$threshold)[1]
      tail_scale_k <- as.numeric(args_k$tail_scale)[1]
      tail_shape_k <- as.numeric(args_k$tail_shape)[1]
      bulk_args_k <- c(list(w = 1), args_k[bulk_params])
      bulk_trunc_k <- as.numeric(do.call(bulk_mean_trunc_fun, c(bulk_args_k, list(threshold = threshold_k))))[1]
      Fu_k <- as.numeric(do.call(bulk_p_fun, c(list(q = threshold_k, lower.tail = 1L, log.p = 0L), bulk_args_k)))[1]
      bulk_trunc_k + (1 - .clamp_prob(Fu_k)) * .gpd_tail_mean_or_error(threshold_k, tail_scale_k, tail_shape_k)
    }, numeric(1))

    sum(w_s * comp_mean)
  }

  # -----------------------------
  # density / survival
  # -----------------------------
  if (type %in% c("density", "survival")) {
    G <- length(ygrid)
    ygrid_num <- as.numeric(ygrid)

    .one_draw <- function(s) {
      if (!.draw_valid[s]) return(list(valid = FALSE, out = matrix(NA_real_, nrow = n_pred, ncol = G)))

      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) return(list(valid = FALSE, out = matrix(NA_real_, nrow = n_pred, ncol = G)))

      link_eta <- .compute_link_eta(s)
      gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

      out <- matrix(NA_real_, nrow = n_pred, ncol = G)
      for (i in seq_len(n_pred)) {
        if (is_spliced && GPD) {
          out[i, ] <- .spliced_density_survival_row(s, i, ygrid_num, type = type, link_eta = link_eta, gpd_eta = gpd_eta)
          next
        }

        args <- args0
        if (GPD) {
          args$threshold <- .threshold_at(s, i)
          args$tail_scale <- .tail_scale_at(s, i)
          args$tail_shape <- .tail_shape_at(s, i)
          if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
            out[i, ] <- NA_real_
            next
          }
        }
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!all(is.finite(vv))) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            out[i, ] <- NA_real_
            next
          }
        }

        if (type == "density") {
          v <- as.numeric(do.call(d_fun, c(list(x = ygrid_num, log = 0L), args)))
          out[i, ] <- v
        } else {
          cdfv <- as.numeric(do.call(p_fun, c(list(q = ygrid_num, lower.tail = 1L, log.p = 0L), args)))
          # Clamp to [0,1] before survival transform
          cdfv <- pmin(pmax(cdfv, 0), 1)
          surv <- 1 - cdfv
          surv <- pmin(pmax(surv, 0), 1)
          out[i, ] <- surv
        }
      }
      list(valid = TRUE, out = out)
    }

    res_list <- .lapply_draws(.one_draw)

    draws_arr <- array(NA_real_, dim = c(S, n_pred, G))
    valid_vec <- logical(S)
    for (s in seq_len(S)) {
      valid_vec[s] <- isTRUE(res_list[[s]]$valid)
      draws_arr[s, , ] <- res_list[[s]]$out
    }

    fit <- apply(draws_arr, c(2, 3), mean, na.rm = TRUE)

    lower <- upper <- NULL
    if (compute_interval) {
      lower <- matrix(NA_real_, nrow = n_pred, ncol = G)
      upper <- matrix(NA_real_, nrow = n_pred, ncol = G)
      for (i in seq_len(n_pred)) {
        for (j in seq_len(G)) {
          iv <- .compute_interval(draws_arr[, i, j], level = level, type = interval)
          lower[i, j] <- iv["lower"]
          upper[i, j] <- iv["upper"]
        }
      }
    }

    # Build DF (id varies slowest, y varies fastest)
    fit_df <- data.frame(
      id = rep(id_vals, each = G),
      y = rep(ygrid_num, times = n_pred),
      estimate = as.vector(t(fit)),
      lower = if (!is.null(lower)) as.vector(t(lower)) else NA_real_,
      upper = if (!is.null(upper)) as.vector(t(upper)) else NA_real_,
      row.names = NULL
    )
    colnames(fit_df)[colnames(fit_df) == "estimate"] <- ifelse(type == "density", "density", "survival")
    fit_df <- .reorder_predict_cols(fit_df)

    out <- list(
      fit = fit_df,
      fit_df = fit_df,
      type = type,
      grid = ygrid_num,
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(valid_vec),
        n_draws_dropped = S - sum(valid_vec)
      )
    )
    class(out) <- "mixgpd_predict"
    return(.return_out(out))
  }

  # -----------------------------
  # quantile / median
  # -----------------------------
  if (type %in% c("quantile", "median")) {
    M <- length(pgrid)

    if (!has_X) {
      draws_mat <- matrix(NA_real_, nrow = M, ncol = S)

      for (s in seq_len(S)) {
        if (!.draw_valid[s]) next
        args0 <- .build_args0_or_null(s)
        if (is.null(args0)) next

        if (is_spliced && GPD) {
          draws_mat[, s] <- vapply(pgrid, function(pp) {
            .spliced_quantile_one(s, 1L, pp, link_eta = list(), gpd_eta = list())
          }, numeric(1))
          next
        }

        if (GPD) {
          args0$threshold <- threshold_scalar[s]
          args0$tail_scale <- tail_scale[s]
          args0$tail_shape <- .tail_shape_at(s, 1L)
          if (!is.finite(args0$threshold) || !is.finite(args0$tail_scale) || args0$tail_scale <= 0 || !is.finite(args0$tail_shape)) next
        }

        draws_mat[, s] <- as.numeric(do.call(q_fun, c(list(p = pgrid), args0)))
      }

      summ <- .posterior_summarize(draws_mat, probs = probs, interval = if (compute_interval) interval else NULL)

      fit_df <- data.frame(
        id       = rep(id_vals, each = length(pgrid)),
        index    = pgrid,
        estimate = as.numeric(summ$estimate),
        lower    = if (compute_interval) as.numeric(summ$lower) else NA_real_,
        upper    = if (compute_interval) as.numeric(summ$upper) else NA_real_,
        row.names = NULL
      )
      fit_df <- .reorder_predict_cols(fit_df)

      out <- list(
        fit = fit_df,
        fit_df = fit_df,
        type = type,
        grid = pgrid,
        draws = if (isTRUE(store_draws)) draws_mat else NULL,
        diagnostics = list(
          n_draws_total = S,
          n_draws_valid = sum(.draw_valid),
          n_draws_dropped = S - sum(.draw_valid)
        )
      )
      class(out) <- "mixgpd_predict"
      return(.return_out(out))
    }

    # Conditional quantiles
    draws_arr <- array(NA_real_, dim = c(n_pred, M, S))

    for (s in seq_len(S)) {
      if (!.draw_valid[s]) next
      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) next

      link_eta <- .compute_link_eta(s)
      gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

      for (i in seq_len(n_pred)) {
        if (is_spliced && GPD) {
          draws_arr[i, , s] <- vapply(pgrid, function(pp) {
            .spliced_quantile_one(s, i, pp, link_eta = link_eta, gpd_eta = gpd_eta)
          }, numeric(1))
          next
        }

        args <- args0
        if (GPD) {
          args$threshold <- .threshold_at(s, i)
          args$tail_scale <- .tail_scale_at(s, i)
          args$tail_shape <- .tail_shape_at(s, i)
          if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
            draws_arr[i, , s] <- NA_real_
            next
          }
        }
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!all(is.finite(vv))) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            draws_arr[i, , s] <- NA_real_
            next
          }
        }
        draws_arr[i, , s] <- as.numeric(do.call(q_fun, c(list(p = pgrid), args)))
      }
    }

    summ <- .posterior_summarize(draws_arr, probs = probs, interval = if (compute_interval) interval else NULL)
    estimate <- summ$estimate
    lower <- summ$lower
    upper <- summ$upper

    fit_df <- data.frame(
      id       = rep(id_vals, each = M),
      index    = rep(pgrid, times = n_pred),
      estimate = as.vector(t(estimate)),
      lower    = if (compute_interval) as.vector(t(lower)) else NA_real_,
      upper    = if (compute_interval) as.vector(t(upper)) else NA_real_,
      row.names = NULL
    )
    fit_df <- .reorder_predict_cols(fit_df)

    out <- list(
      fit = fit_df,
      fit_df = fit_df,
      type = type,
      grid = pgrid,
      draws = if (isTRUE(store_draws)) aperm(draws_arr, c(3, 1, 2)) else NULL, # S x n_pred x M
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(.draw_valid),
        n_draws_dropped = S - sum(.draw_valid)
      )
    )
    class(out) <- "mixgpd_predict"
    return(.return_out(out))
  }

  # -----------------------------
  # sample (posterior predictive)
  # -----------------------------
  if (type == "sample") {
    if (!is.null(sample_draw_idx)) {
      idx <- as.integer(sample_draw_idx)
      if (!length(idx) || any(!is.finite(idx)) || any(idx < 1L) || any(idx > S)) {
        stop("'sample_draw_idx' must be a non-empty integer vector indexing posterior draws.", call. = FALSE)
      }
      nsim <- length(idx)
    }

    if (!has_X) {
      if (is.na(nsim) || nsim < 1L) nsim <- length(ytrain)
      if (is.null(sample_draw_idx)) idx <- sample.int(S, size = nsim, replace = TRUE)
      outv <- numeric(nsim)

      for (t in seq_len(nsim)) {
        s <- idx[t]
        if (!.draw_valid[s]) { outv[t] <- NA_real_; next }
        args0 <- .build_args0_or_null(s)
        if (is.null(args0)) { outv[t] <- NA_real_; next }

        if (is_spliced && GPD) {
          outv[t] <- .spliced_sample_values(s, 1L, n = 1L, link_eta = list(), gpd_eta = list())[1]
          next
        }

        if (GPD) {
          args0$threshold <- threshold_scalar[s]
          args0$tail_scale <- tail_scale[s]
          args0$tail_shape <- .tail_shape_at(s, 1L)
          if (!is.finite(args0$threshold) || !is.finite(args0$tail_scale) || args0$tail_scale <= 0 || !is.finite(args0$tail_shape)) {
            outv[t] <- NA_real_
            next
          }
        }
        outv[t] <- as.numeric(do.call(r_fun, c(list(n = 1L), args0)))[1]
      }

      res <- list(
        fit = outv,
        fit_df = .values_to_long_df(outv, value_name = "sample"),
        type = type,
        grid = NULL,
        diagnostics = list(
          n_draws_total = S,
          n_draws_valid = sum(.draw_valid),
          n_draws_dropped = S - sum(.draw_valid)
        )
      )
      class(res) <- "mixgpd_predict"
      return(.return_out(res))
    }

    if (is.na(nsim) || nsim < 1L) nsim <- n_pred
    if (is.null(sample_draw_idx)) idx <- sample.int(S, size = nsim, replace = TRUE)
    outm <- matrix(NA_real_, nrow = n_pred, ncol = nsim)

    for (t in seq_len(nsim)) {
      s <- idx[t]
      if (!.draw_valid[s]) next
      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) next

      link_eta <- .compute_link_eta(s)
      gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

      for (i in seq_len(n_pred)) {
        if (is_spliced && GPD) {
          outm[i, t] <- .spliced_sample_values(s, i, n = 1L, link_eta = link_eta, gpd_eta = gpd_eta)[1]
          next
        }

        args <- args0
        if (GPD) {
          args$threshold <- .threshold_at(s, i)
          args$tail_scale <- .tail_scale_at(s, i)
          args$tail_shape <- .tail_shape_at(s, i)
          if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
            outm[i, t] <- NA_real_
            next
          }
        }
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!all(is.finite(vv))) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            outm[i, t] <- NA_real_
            next
          }
        }
        outm[i, t] <- as.numeric(do.call(r_fun, c(list(n = 1L), args)))[1]
      }
    }

    res <- list(
      fit = outm,
      fit_df = .values_to_long_df(outm, id = id_vals, value_name = "sample"),
      type = type,
      grid = NULL,
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(.draw_valid),
        n_draws_dropped = S - sum(.draw_valid)
      )
    )
    class(res) <- "mixgpd_predict"
    return(.return_out(res))
  }

  # -----------------------------
  # fit (one posterior predictive draw per observation per posterior draw)
  # -----------------------------
  if (type == "fit") {
    n_obs <- if (!has_X) length(ytrain) else n_pred
    if (!is.numeric(n_obs) || n_obs < 1L) stop("Could not determine n_obs for type='fit'.", call. = FALSE)

    samples_mat <- matrix(NA_real_, nrow = S, ncol = n_obs)

    for (s in seq_len(S)) {
      if (!.draw_valid[s]) next
      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) next

      # Conditional: one sample per observation
      if (has_X) {
        link_eta <- .compute_link_eta(s)
        gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()
        for (i in seq_len(n_obs)) {
          if (is_spliced && GPD) {
            samples_mat[s, i] <- .spliced_sample_values(s, i, n = 1L, link_eta = link_eta, gpd_eta = gpd_eta)[1]
            next
          }

          args <- args0
          if (GPD) {
            args$threshold <- .threshold_at(s, i)
            args$tail_scale <- .tail_scale_at(s, i)
            args$tail_shape <- .tail_shape_at(s, i)
            if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
              samples_mat[s, i] <- NA_real_
              next
            }
          }
          if (length(link_params)) {
            bad <- FALSE
            for (nm in link_params) {
              vv <- as.numeric(link_eta[[nm]][i, ])
              if (!all(is.finite(vv))) {
                bad <- TRUE
                break
              }
              args[[nm]] <- vv
            }
            if (bad) {
              samples_mat[s, i] <- NA_real_
              next
            }
          }
          samples_mat[s, i] <- as.numeric(do.call(r_fun, c(list(n = 1L), args)))[1]
        }
      } else {
        # Unconditional: draw n_obs from marginal predictive for draw s
        if (is_spliced && GPD) {
          samples_mat[s, ] <- .spliced_sample_values(s, 1L, n = n_obs, link_eta = list(), gpd_eta = list())
          next
        }

        if (GPD) {
          args0$threshold <- threshold_scalar[s]
          args0$tail_scale <- tail_scale[s]
          args0$tail_shape <- .tail_shape_at(s, 1L)
          if (!is.finite(args0$threshold) || !is.finite(args0$tail_scale) || args0$tail_scale <= 0 || !is.finite(args0$tail_shape)) next
        }
        samples_mat[s, ] <- as.numeric(do.call(r_fun, c(list(n = n_obs), args0)))
      }
    }

    estimate <- as.numeric(colMeans(samples_mat, na.rm = TRUE))

    if (compute_interval) {
      lower <- apply(samples_mat, 2, stats::quantile, probs = probs[1], na.rm = TRUE)
      upper <- apply(samples_mat, 2, stats::quantile, probs = probs[length(probs)], na.rm = TRUE)
    } else {
      lower <- rep(NA_real_, n_obs)
      upper <- rep(NA_real_, n_obs)
    }

    id_use <- if (length(id_vals) == n_obs) id_vals else seq_len(n_obs)
    fit_df <- data.frame(
      id = id_use,
      estimate = estimate,
      lower = as.numeric(lower),
      upper = as.numeric(upper),
      row.names = NULL
    )
    fit_df <- .reorder_predict_cols(fit_df)

    out <- list(
      fit = fit_df,
      fit_df = fit_df,
      type = type,
      draws = if (isTRUE(store_draws)) samples_mat else NULL,
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(.draw_valid),
        n_draws_dropped = S - sum(.draw_valid)
      )
    )
    class(out) <- "mixgpd_predict"
    return(.return_out(out))
  }

  # -----------------------------
  # mean (posterior mean of predictive distribution)
  # -----------------------------
  if (type == "mean") {
    has_analytic_mean <- if (!GPD) (!is.null(bulk_mean_fun) && is.function(bulk_mean_fun)) else
      (!is.null(bulk_mean_trunc_fun) && is.function(bulk_mean_trunc_fun))

    nsim_inner_mean <- as.integer(nsim_mean)
    if (is.na(nsim_inner_mean) || nsim_inner_mean < 10L) nsim_inner_mean <- NA_integer_
    use_sim_mean <- !has_analytic_mean && !is.na(nsim_inner_mean)

    if (!has_analytic_mean && !use_sim_mean) {
      if (!GPD) {
        stop(sprintf("Analytical mean is not implemented for kernel '%s'. Supply nsim_mean >= 10 for simulation-based mean.", kernel), call. = FALSE)
      } else {
        stop(sprintf("Analytical GPD mean is not implemented for kernel '%s'. Supply nsim_mean >= 10 for simulation-based mean.", kernel), call. = FALSE)
      }
    }

    if (!has_X) {
      draw_means <- rep(NA_real_, S)
      for (s in seq_len(S)) {
        if (!.draw_valid[s]) next
        args0 <- .build_args0_or_null(s)
        if (is.null(args0)) next

        if (use_sim_mean) {
          if (is_spliced && GPD) {
            yy <- .spliced_sample_values(s, 1L, n = nsim_inner_mean, link_eta = list(), gpd_eta = list())
            draw_means[s] <- mean(yy, na.rm = TRUE)
            next
          }
          if (GPD) {
            args0$threshold <- .threshold_at(s, 1L)
            args0$tail_scale <- .tail_scale_at(s, 1L)
            args0$tail_shape <- .tail_shape_at(s, 1L)
            if (!is.finite(args0$threshold) || !is.finite(args0$tail_scale) || args0$tail_scale <= 0 || !is.finite(args0$tail_shape)) next
          }
          yy <- as.numeric(do.call(r_fun, c(list(n = nsim_inner_mean), args0)))
          draw_means[s] <- mean(yy, na.rm = TRUE)
          next
        }

        if (!GPD) {
          draw_means[s] <- as.numeric(do.call(bulk_mean_fun, args0))[1]
          next
        }

        if (is_spliced) {
          draw_means[s] <- .analytic_spliced_mean_row(s, 1L, link_eta = list(), gpd_eta = list())
          next
        }

        threshold_s <- .threshold_at(s, 1L)
        tail_scale_s <- .tail_scale_at(s, 1L)
        tail_shape_s <- .tail_shape_at(s, 1L)
        if (!is.finite(threshold_s) || !is.finite(tail_scale_s) || tail_scale_s <= 0 || !is.finite(tail_shape_s)) next
        bulk_args_s <- c(if ("w" %in% names(args0)) list(w = args0$w) else list(), args0[bulk_params])
        bulk_trunc_s <- as.numeric(do.call(bulk_mean_trunc_fun, c(bulk_args_s, list(threshold = threshold_s))))[1]
        Fu_s <- as.numeric(do.call(bulk_p_fun, c(list(q = threshold_s, lower.tail = 1L, log.p = 0L), bulk_args_s)))[1]
        draw_means[s] <- bulk_trunc_s + (1 - .clamp_prob(Fu_s)) *
          .gpd_tail_mean_or_error(threshold_s, tail_scale_s, tail_shape_s)
      }

      summ <- .posterior_summarize(draw_means, probs = probs, interval = if (compute_interval) interval else NULL)

      fit_df <- data.frame(
        id = if (length(id_vals) >= 1L) id_vals[1] else 1L,
        estimate = as.numeric(summ$estimate)[1],
        lower = if (compute_interval) as.numeric(summ$lower)[1] else NA_real_,
        upper = if (compute_interval) as.numeric(summ$upper)[1] else NA_real_,
        row.names = NULL
      )
      fit_df <- .reorder_predict_cols(fit_df)

      out <- list(
        fit = fit_df,
        fit_df = fit_df,
        type = type,
        grid = NULL,
        draws = if (isTRUE(store_draws)) draw_means else NULL,
        diagnostics = list(
          n_draws_total = S,
          n_draws_valid = sum(.draw_valid),
          n_draws_dropped = S - sum(.draw_valid),
          mean_method = if (use_sim_mean) "simulation" else "analytic",
          nsim_mean = if (use_sim_mean) nsim_inner_mean else NULL
        )
      )
      class(out) <- "mixgpd_predict"
      return(.return_out(out))
    }

    draw_means_mat <- matrix(NA_real_, nrow = S, ncol = n_pred)
    for (s in seq_len(S)) {
      if (!.draw_valid[s]) next
      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) next

      link_eta <- .compute_link_eta(s)
      gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

      for (i in seq_len(n_pred)) {
        if (use_sim_mean) {
          if (is_spliced && GPD) {
            yy <- .spliced_sample_values(s, i, n = nsim_inner_mean, link_eta = link_eta, gpd_eta = gpd_eta)
            draw_means_mat[s, i] <- mean(yy, na.rm = TRUE)
            next
          }
          args_sim <- args0
          if (GPD) {
            args_sim$threshold <- .threshold_at(s, i)
            args_sim$tail_scale <- .tail_scale_at(s, i)
            args_sim$tail_shape <- .tail_shape_at(s, i)
            if (!is.finite(args_sim$threshold) || !is.finite(args_sim$tail_scale) || args_sim$tail_scale <= 0 || !is.finite(args_sim$tail_shape)) {
              draw_means_mat[s, i] <- NA_real_
              next
            }
          }
          if (length(link_params)) {
            bad <- FALSE
            for (nm in link_params) {
              vv <- as.numeric(link_eta[[nm]][i, ])
              if (!all(is.finite(vv))) { bad <- TRUE; break }
              args_sim[[nm]] <- vv
            }
            if (bad) { draw_means_mat[s, i] <- NA_real_; next }
          }
          yy <- as.numeric(do.call(r_fun, c(list(n = nsim_inner_mean), args_sim)))
          draw_means_mat[s, i] <- mean(yy, na.rm = TRUE)
          next
        }

        if (is_spliced && GPD) {
          draw_means_mat[s, i] <- .analytic_spliced_mean_row(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
          next
        }

        args <- args0
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!.support_ok(nm, vv)) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            draw_means_mat[s, i] <- NA_real_
            next
          }
        }

        if (!GPD) {
          draw_means_mat[s, i] <- as.numeric(do.call(bulk_mean_fun, args))[1]
          next
        }

        threshold_i <- .threshold_at(s, i)
        tail_scale_i <- .tail_scale_at(s, i)
        tail_shape_i <- .tail_shape_at(s, i)
        if (!is.finite(threshold_i) || !is.finite(tail_scale_i) || tail_scale_i <= 0 || !is.finite(tail_shape_i)) {
          draw_means_mat[s, i] <- NA_real_
          next
        }
        bulk_args_i <- c(if ("w" %in% names(args)) list(w = args$w) else list(), args[bulk_params])
        bulk_trunc_i <- as.numeric(do.call(bulk_mean_trunc_fun, c(bulk_args_i, list(threshold = threshold_i))))[1]
        Fu_i <- as.numeric(do.call(bulk_p_fun, c(list(q = threshold_i, lower.tail = 1L, log.p = 0L), bulk_args_i)))[1]
        draw_means_mat[s, i] <- bulk_trunc_i + (1 - .clamp_prob(Fu_i)) *
          .gpd_tail_mean_or_error(threshold_i, tail_scale_i, tail_shape_i)
      }
    }

    summ <- .posterior_summarize(t(draw_means_mat), probs = probs, interval = if (compute_interval) interval else NULL)
    fit_df <- data.frame(
      id = id_vals,
      estimate = as.numeric(summ$estimate),
      lower = if (compute_interval) as.numeric(summ$lower) else NA_real_,
      upper = if (compute_interval) as.numeric(summ$upper) else NA_real_,
      row.names = NULL
    )
    fit_df <- .reorder_predict_cols(fit_df)

    out <- list(
      fit = fit_df,
      fit_df = fit_df,
      type = type,
      grid = NULL,
      draws = if (isTRUE(store_draws)) draw_means_mat else NULL,
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(.draw_valid),
        n_draws_dropped = S - sum(.draw_valid),
        mean_method = if (use_sim_mean) "simulation" else "analytic",
        nsim_mean = if (use_sim_mean) nsim_inner_mean else NULL
      )
    )
    class(out) <- "mixgpd_predict"
    return(.return_out(out))
  }


  # -----------------------------
  # rmean (restricted mean E[min(Y, cutoff)])
  # -----------------------------
  if (type == "rmean") {
    if (is.null(cutoff) || length(cutoff) != 1L || !is.finite(as.numeric(cutoff))) {
      stop("For type='rmean', provide a finite numeric 'cutoff'.", call. = FALSE)
    }
    cutoff <- as.numeric(cutoff)

    nsim_inner <- as.integer(nsim_mean)
    if (is.na(nsim_inner) || nsim_inner < 10L) nsim_inner <- 200L

    if (!has_X) {
      draw_rmeans <- rep(NA_real_, S)
      for (s in seq_len(S)) {
        if (!.draw_valid[s]) next
        args0 <- .build_args0_or_null(s)
        if (is.null(args0)) next
        if (is_spliced && GPD) {
          yy <- .spliced_sample_values(s, 1L, n = nsim_inner, link_eta = list(), gpd_eta = list())
          draw_rmeans[s] <- mean(pmin(yy, cutoff), na.rm = TRUE)
          next
        }
        if (GPD) {
          args0$threshold <- threshold_scalar[s]
          args0$tail_scale <- tail_scale[s]
          args0$tail_shape <- .tail_shape_at(s, 1L)
          if (!is.finite(args0$threshold) || !is.finite(args0$tail_scale) || args0$tail_scale <= 0 || !is.finite(args0$tail_shape)) next
        }
        yy <- as.numeric(do.call(r_fun, c(list(n = nsim_inner), args0)))
        draw_rmeans[s] <- mean(pmin(yy, cutoff), na.rm = TRUE)
      }

      summ <- .posterior_summarize(draw_rmeans, probs = probs, interval = if (compute_interval) interval else NULL)

      fit_df <- data.frame(
        id = if (length(id_vals) >= 1L) id_vals[1] else 1L,
        estimate = as.numeric(summ$estimate)[1],
        lower = if (compute_interval) as.numeric(summ$lower)[1] else NA_real_,
        upper = if (compute_interval) as.numeric(summ$upper)[1] else NA_real_,
        row.names = NULL
      )
      fit_df <- .reorder_predict_cols(fit_df)

      out <- list(
        fit = fit_df,
        fit_df = fit_df,
        type = type,
        grid = NULL,
        cutoff = cutoff,
        draws = if (isTRUE(store_draws)) draw_rmeans else NULL,
        diagnostics = list(
          n_draws_total = S,
          n_draws_valid = sum(.draw_valid),
          n_draws_dropped = S - sum(.draw_valid),
          nsim_mean = nsim_inner,
          mean_infinite = any(!is.finite(draw_rmeans))
        )
      )
      class(out) <- "mixgpd_predict"
      return(.return_out(out))
    }

    draw_rmeans_mat <- matrix(NA_real_, nrow = S, ncol = n_pred)

    for (s in seq_len(S)) {
      if (!.draw_valid[s]) next
      args0 <- .build_args0_or_null(s)
      if (is.null(args0)) next

      link_eta <- .compute_link_eta(s)
      gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

      for (i in seq_len(n_pred)) {
        if (is_spliced && GPD) {
          yy <- .spliced_sample_values(s, i, n = nsim_inner, link_eta = link_eta, gpd_eta = gpd_eta)
          draw_rmeans_mat[s, i] <- mean(pmin(yy, cutoff), na.rm = TRUE)
          next
        }

        args <- args0
        if (GPD) {
          args$threshold <- .threshold_at(s, i)
          args$tail_scale <- .tail_scale_at(s, i)
          args$tail_shape <- .tail_shape_at(s, i)
          if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
            draw_rmeans_mat[s, i] <- NA_real_
            next
          }
        }
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!all(is.finite(vv))) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            draw_rmeans_mat[s, i] <- NA_real_
            next
          }
        }

        yy <- as.numeric(do.call(r_fun, c(list(n = nsim_inner), args)))
        draw_rmeans_mat[s, i] <- mean(pmin(yy, cutoff), na.rm = TRUE)
      }
    }

    estimate <- apply(draw_rmeans_mat, 2, mean, na.rm = TRUE)

    lower <- upper <- NULL
    if (compute_interval) {
      lower <- upper <- rep(NA_real_, n_pred)
      for (i in seq_len(n_pred)) {
        iv <- .compute_interval(draw_rmeans_mat[, i], level = level, type = interval)
        lower[i] <- iv["lower"]
        upper[i] <- iv["upper"]
      }
    }

    fit_df <- data.frame(
      id = id_vals,
      estimate = as.numeric(estimate),
      lower = if (compute_interval) as.numeric(lower) else NA_real_,
      upper = if (compute_interval) as.numeric(upper) else NA_real_,
      row.names = NULL
    )
    fit_df <- .reorder_predict_cols(fit_df)

    out <- list(
      fit = fit_df,
      fit_df = fit_df,
      type = type,
      grid = NULL,
      cutoff = cutoff,
      draws = if (isTRUE(store_draws)) draw_rmeans_mat else NULL,
      diagnostics = list(
        n_draws_total = S,
        n_draws_valid = sum(.draw_valid),
        n_draws_dropped = S - sum(.draw_valid),
        nsim_mean = nsim_inner
      )
    )
    class(out) <- "mixgpd_predict"
    return(.return_out(out))
  }

stop(sprintf("Unsupported prediction type '%s'.", type), call. = FALSE)
}





.wrap_density_fun <- function(fun) {
  function(x, ...) {
    if (length(x) == 0) return(numeric(0))
    if (length(x) == 1) return(as.numeric(fun(x, ...)))
    vapply(x, function(xx) as.numeric(fun(xx, ...)), numeric(1))
  }
}

.wrap_cdf_fun <- function(fun) {
  function(q, ...) {
    if (length(q) == 0) return(numeric(0))
    if (length(q) == 1) return(as.numeric(fun(q, ...)))
    vapply(q, function(qq) as.numeric(fun(qq, ...)), numeric(1))
  }
}

.wrap_quantile_fun <- function(fun) {
  function(p, ...) {
    if (length(p) == 0) return(numeric(0))
    fun(p, ...)
  }
}

.wrap_rng_fun <- function(fun) {
  function(n, ...) {
    if (n <= 0) return(numeric(0))
    if (n == 1) return(as.numeric(fun(1, ...)))
    vapply(seq_len(n), function(i) as.numeric(fun(1, ...)), numeric(1))
  }
}


# ==============================================================================
# Internal helpers for cluster-label summaries (PSM, Dahl's method)
# ==============================================================================

#' Extract cluster assignment matrix from MCMC samples
#'
#' Extracts the posterior draws of cluster assignments \code{z[1:N]} from a fitted
#' mixgpd_fit object and returns them as an integer matrix (iterations x N).
#'
#' @details
#' Cluster samplers store latent labels as separate monitored nodes `z[i]`. This
#' helper locates those nodes in every retained chain, orders them by observation
#' index, stacks the chains, and returns the result as one integer matrix that is
#' ready for PSM and representative-partition calculations.
#'
#' @param object A \code{mixgpd_fit} object.
#' @return Integer matrix with rows = posterior draws, cols = observations.
#' @keywords internal
.extract_z_matrix <- function(object) {
  smp <- .get_samples_mcmclist(object)
  if (is.null(smp)) {
    stop("No MCMC samples found in object.", call. = FALSE)
  }

  # Stack all chains
  zmat_list <- vector("list", length(smp))
  for (i in seq_along(smp)) {
    M <- as.matrix(smp[[i]])
    z_cols <- grep("^z\\[[0-9]+\\]$", colnames(M))
    if (length(z_cols) == 0) {
      stop("No cluster assignment variables 'z[i]' found in MCMC samples.", call. = FALSE)
    }
    # Sort z columns numerically by index
    z_colnames <- colnames(M)[z_cols]
    z_indices <- as.integer(sub("^z\\[([0-9]+)\\]$", "\\1", z_colnames))
    z_cols <- z_cols[order(z_indices)]
    zmat_list[[i]] <- M[, z_cols, drop = FALSE]
  }

  zmat <- do.call(rbind, zmat_list)
  # Convert to integer matrix
  mode(zmat) <- "integer"
  return(zmat)
}

#' Compute posterior similarity matrix
#'
#' Computes the posterior similarity matrix (PSM) from a matrix of cluster
#' assignments. \code{PSM[i,j]} = probability that observations i and j are in the
#' same cluster.
#'
#' @details
#' If \eqn{z_i^{(s)}} denotes the cluster label of observation \eqn{i} at draw
#' \eqn{s}, then this helper computes
#' \deqn{\mathrm{PSM}_{ij} \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}).}
#' The resulting matrix is the basic posterior co-clustering summary used by the
#' Dahl representative partition and several cluster diagnostics.
#'
#' @param z_matrix Integer matrix (iterations x N) of cluster assignments.
#' @return Symmetric N x N matrix of co-clustering probabilities.
#' @keywords internal
.compute_psm <- function(z_matrix) {
  n_iter <- nrow(z_matrix)
  n_obs <- ncol(z_matrix)
  PSM <- matrix(0, n_obs, n_obs)

  for (s in seq_len(n_iter)) {
    z <- z_matrix[s, ]
    # Indicator matrix: same cluster
    A <- outer(z, z, "==") * 1.0
    PSM <- PSM + A
  }
  PSM <- PSM / n_iter
  return(PSM)
}

#' Find Dahl representative clustering
#'
#' Identifies the posterior draw that minimizes squared distance to the
#' posterior similarity matrix, following Dahl (2006). Returns relabeled
#' cluster assignments as consecutive integers 1, 2, ..., K.
#'
#' @details
#' For each posterior draw, the helper forms its adjacency matrix and computes
#' the squared Frobenius distance to the PSM. The selected representative draw is
#' the one that minimizes that loss, which is Dahl's least-squares rule for
#' choosing one clustering from the posterior sample.
#'
#' @param z_matrix Integer matrix (iterations x N) of cluster assignments.
#' @param PSM Posterior similarity matrix (N x N).
#' @return List with components: draw_index (integer), labels (integer vector),
#'   K (number of clusters).
#' @references Dahl, D. B. (2006). Model-based clustering for expression data
#'   via a Dirichlet process mixture model. In M. Vannucci, et al. (Eds.),
#'   Bayesian Inference for Gene Expression and Proteomics (pp. 201-218).
#'   Cambridge University Press.
#' @keywords internal
.dahl_representative <- function(z_matrix, PSM) {
  n_iter <- nrow(z_matrix)
  ssq <- numeric(n_iter)

  for (s in seq_len(n_iter)) {
    z <- z_matrix[s, ]
    A <- outer(z, z, "==") * 1.0
    ssq[s] <- sum((A - PSM)^2)
  }

  s_star <- which.min(ssq)
  z_hat <- as.integer(z_matrix[s_star, ])

  # Relabel to consecutive integers
  labels <- match(z_hat, unique(z_hat))
  K <- length(unique(labels))

  return(list(
    draw_index = s_star,
    labels = labels,
    K = K
  ))
}

#' Compute cluster membership probabilities from PSM
#'
#' For each observation, computes the probability of membership in each cluster
#' defined by the representative clustering, derived from the posterior
#' similarity matrix.
#'
#' @details
#' The representative labels define a reference partition with clusters
#' \eqn{C_1, \dots, C_K}. For each observation \eqn{i}, this helper averages the
#' posterior similarity scores \eqn{\mathrm{PSM}_{ij}} over members
#' \eqn{j \in C_k} to obtain a cluster-membership score for cluster \eqn{k}, and
#' then normalizes those scores to sum to one across clusters.
#'
#' @param z_matrix Integer matrix (iterations x N) of cluster assignments.
#' @param labels_representative Integer vector of representative cluster labels.
#' @param PSM Posterior similarity matrix (N x N).
#' @return N x K matrix of cluster membership probabilities.
#' @keywords internal
.compute_cluster_probs <- function(z_matrix, labels_representative, PSM) {
  n_obs <- length(labels_representative)
  K <- length(unique(labels_representative))

  probs <- matrix(0, nrow = n_obs, ncol = K)

  for (k in seq_len(K)) {
    # Observations in cluster k in the representative partition
    idx_k <- which(labels_representative == k)
    # For each observation i, average PSM with all members of cluster k
    for (i in seq_len(n_obs)) {
      probs[i, k] <- mean(PSM[i, idx_k])
    }
  }

  # Normalize rows to sum to 1
  row_sums <- rowSums(probs)
  for (i in seq_len(n_obs)) {
    if (row_sums[i] > 0) {
      probs[i, ] <- probs[i, ] / row_sums[i]
    } else {
      # Fallback: uniform distribution
      probs[i, ] <- 1 / K
    }
  }

  return(probs)
}

# Coverage expansion tests targeting low-coverage helper and S3 files.

skip_if_not_test_level("ci")

.coverage_heavy_cached <- function(key, expr) {
  hit <- .cache_get(key)
  if (!is.null(hit)) return(hit)
  val <- force(expr)
  .cache_set(key, val)
  val
}

.coverage_heavy_fit <- function() {
  .coverage_heavy_cached("coverage-heavy-fit", {
    set.seed(901)
    n <- 24L
    X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
    y <- abs(0.6 + 0.4 * X[, 1] + stats::rnorm(n)) + 0.2
    bundle <- build_nimble_bundle(
      y = y,
      X = X,
      backend = "sb",
      kernel = "normal",
      GPD = TRUE,
      components = 4L,
      mcmc = list(niter = 60L, nburnin = 20L, thin = 1L, nchains = 1L, seed = 901L, waic = TRUE)
    )
    run_mcmc_bundle_manual(bundle, show_progress = FALSE, quiet = TRUE, timing = TRUE, z_update_every = 2L)
  })
}

.coverage_heavy_uncond_fit <- function() {
  .coverage_heavy_cached("coverage-heavy-fit-uncond", {
    set.seed(902)
    y <- abs(stats::rnorm(20L)) + 0.2
    bundle <- build_nimble_bundle(
      y = y,
      backend = "crp",
      kernel = "gamma",
      GPD = FALSE,
      components = 4L,
      mcmc = list(niter = 50L, nburnin = 15L, thin = 1L, nchains = 1L, seed = 902L, waic = FALSE)
    )
    run_mcmc_bundle_manual(bundle, show_progress = FALSE, quiet = TRUE, timing = TRUE)
  })
}

.coverage_heavy_cluster_fit <- function() {
  .coverage_heavy_cached("coverage-heavy-fit-cluster", {
    set.seed(903)
    dat <- data.frame(
      y = abs(stats::rnorm(18L)) + 0.2,
      x1 = stats::rnorm(18L),
      x2 = stats::runif(18L)
    )
    fit <- dpmix.cluster(
      y ~ x1 + x2,
      data = dat,
      kernel = "normal",
      components = 4L,
      type = "weights",
      mcmc = list(niter = 50L, nburnin = 15L, thin = 1L, nchains = 1L, seed = 903L, waic = FALSE)
    )
    list(fit = fit, data = dat)
  })
}

.coverage_heavy_fit_2chain <- function() {
  .coverage_heavy_cached("coverage-heavy-fit-2chain", {
    set.seed(906)
    n <- 18L
    X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
    y <- abs(0.4 + 0.3 * X[, 1] + stats::rnorm(n, sd = 0.4)) + 0.2
    bundle <- build_nimble_bundle(
      y = y,
      X = X,
      backend = "sb",
      kernel = "normal",
      GPD = TRUE,
      components = 4L,
      mcmc = list(niter = 45L, nburnin = 10L, thin = 1L, nchains = 2L, seed = 906L, waic = FALSE)
    )
    run_mcmc_bundle_manual(bundle, show_progress = FALSE, quiet = TRUE, timing = TRUE)
  })
}

.coverage_heavy_causal_fit <- function() {
  .coverage_heavy_cached("coverage-heavy-causal-fit", {
    set.seed(903)
    n <- 24L
    X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
    A <- stats::rbinom(n, 1L, plogis(0.2 + 0.5 * X[, 1]))
    y <- abs(0.5 + A + 0.4 * X[, 1] + stats::rnorm(n)) + 0.2
    bundle <- build_causal_bundle(
      y = y,
      X = X,
      A = A,
      backend = c("sb", "sb"),
      kernel = c("normal", "normal"),
      GPD = c(FALSE, FALSE),
      components = c(4L, 4L),
      PS = "logit",
      mcmc_outcome = list(niter = 50L, nburnin = 15L, thin = 1L, nchains = 1L, seed = 903L),
      mcmc_ps = list(niter = 40L, nburnin = 10L, thin = 1L, nchains = 1L, seed = 904L)
    )
    run_mcmc_causal(bundle, show_progress = FALSE, quiet = TRUE, timing = TRUE)
  })
}

.fake_sampler_conf <- function() {
  conf <- new.env(parent = emptyenv())
  conf$samplerConfs <- list(
    list(name = "RW", target = "beta[1]", control = list()),
    list(
      name = "RW",
      target = "beta[2]",
      control = list(
        clusterVarInfo = list(
          clusterNodes = list(c("beta[1]", "bad_node"), c("z[1]", "bad_z")),
          numNodesPerCluster = c(2L, 2L)
        )
      )
    ),
    list(name = "slice", target = "z[1]", control = list()),
    list(name = "RW", target = "tail_scale[1]", control = list())
  )
  model <- list(
    getNodeNames = function(stochOnly = TRUE, includeData = FALSE) {
      c("beta[1]", "beta[2]", "z[1]", "tail_scale[1]")
    }
  )
  conf$getModel <- function() model
  conf$getSamplers <- function() {
    lapply(conf$samplerConfs, function(x) list(target = x$target))
  }
  conf$removed <- list()
  conf$added <- list()
  conf$removeSamplers <- function(target) {
    conf$removed[[length(conf$removed) + 1L]] <<- target
    invisible(NULL)
  }
  conf$addSampler <- function(target, type, control = list()) {
    conf$added[[length(conf$added) + 1L]] <<- list(target = target, type = type, control = control)
    invisible(NULL)
  }
  conf
}

test_that("coverage-heavy formatting helpers cover knitr dt and print branches", {
  expect_s3_class(.kable_table(data.frame(a = 1:2), row.names = FALSE), "knitr_kable")
  expect_match(as.character(.knitr_asis("alpha", c("beta", "gamma"))), "alpha")
  expect_true(grepl("e", fmt3_sci(c(1, 1e5), big = 1000)[2], fixed = TRUE))
  expect_identical(format_df3("x"), "x")
  expect_identical(format_mat3("x"), "x")

  dt_fun <- .dt_view_table
  environment(dt_fun) <- list2env(
    list(
      interactive = function() TRUE,
      print = function(x, ...) invisible(x),
      .cmgpd_message = function(...) invisible(NULL)
    ),
    parent = environment(.dt_view_table)
  )
  expect_silent(dt_fun(as.data.frame(matrix(runif(120), nrow = 12, ncol = 10)), row.names = FALSE))

  old_knitr <- getOption("knitr.in.progress")
  old_kable <- getOption("causalmixgpd.knitr.kable")
  options(knitr.in.progress = TRUE, causalmixgpd.knitr.kable = TRUE)
  on.exit(options(knitr.in.progress = old_knitr, causalmixgpd.knitr.kable = old_kable), add = TRUE)

  expect_gt(length(utils::capture.output(print_fmt3(
    data.frame(a = c(1.2345, 2.3456), b = c("x", "y")),
    row.names = FALSE
  ))), 0L)
  expect_gt(length(utils::capture.output(print_fmt3(matrix(c(1.2, 3.4), nrow = 1)))), 0L)
  expect_gt(length(utils::capture.output(print_fmt3(1.234))), 0L)
  expect_gt(length(utils::capture.output(print_fmt3_sci(
    data.frame(a = c(1, 100000)),
    row.names = FALSE
  ))), 0L)
  expect_gt(length(utils::capture.output(print_fmt3_sci(
    matrix(c(1, 200000), nrow = 1),
    big = 1000
  ))), 0L)
  expect_gt(length(utils::capture.output(print_fmt3_sci(100000, big = 1000))), 0L)
})

test_that("coverage-heavy wrapper helpers cover GPD stripping and wrapper dispatch", {
  bundle_one <- structure(
    list(
      spec = list(meta = list(GPD = TRUE), plan = list(GPD = TRUE, gpd = list())),
      data = list(y = c(1, 2, 3)),
      monitor_policy = list(monitor_v = TRUE, monitor_latent = TRUE),
      mcmc = list(niter = 20L)
    ),
    class = "causalmixgpd_bundle"
  )
  causal_bundle <- structure(
    list(
      outcome = list(con = bundle_one, trt = bundle_one),
      meta = list(GPD = list(con = TRUE, trt = FALSE))
    ),
    class = "causalmixgpd_causal_bundle"
  )

  testthat::local_mocked_bindings(
    build_code_from_spec = function(spec) list(code = "fake"),
    build_constants_from_spec = function(spec) list(N = 3L),
    build_dimensions_from_spec = function(spec) list(z = 3L),
    build_inits_from_spec = function(spec, y = NULL) list(z = c(1L, 1L, 1L)),
    build_monitors_from_spec = function(spec, ...) c("alpha", "z[1:3]"),
    .package = "CausalMixGPD"
  )

  stripped <- .strip_gpd_single_bundle(bundle_one)
  expect_false(isTRUE(stripped$spec$meta$GPD))
  expect_equal(stripped$spec$plan$gpd, list())
  expect_true(.bundle_has_any_gpd(bundle_one))
  expect_true(.bundle_has_any_gpd(causal_bundle))
  expect_false(.bundle_all_gpd(causal_bundle))

  stripped_causal <- .bundle_strip_gpd(causal_bundle)
  expect_false(isTRUE(stripped_causal$meta$GPD$trt))
  expect_false(isTRUE(stripped_causal$meta$GPD$con))

  testthat::local_mocked_bindings(
    bundle = function(...) structure(list(tag = "built", spec = list(meta = list(GPD = TRUE))), class = "causalmixgpd_bundle"),
    .run_bundle_mcmc = function(b, mcmc_args = list()) structure(list(bundle = b, args = mcmc_args), class = "wrapped_fit"),
    .package = "CausalMixGPD"
  )
  expect_s3_class(dpmix(y = c(1, 2, 3), kernel = "normal", components = 3L, mcmc = list(seed = 1L)), "wrapped_fit")
  expect_s3_class(dpmgpd(y = c(1, 2, 3), kernel = "normal", components = 3L, mcmc = list(seed = 1L)), "wrapped_fit")
  expect_s3_class(dpmix.causal(y = c(1, 2, 3), X = cbind(x1 = c(0, 1, 0)), treat = c(0L, 1L, 0L), kernel = "normal", components = c(3L, 3L), mcmc = list(seed = 1L)), "wrapped_fit")
  expect_s3_class(dpmgpd.causal(y = c(1, 2, 3), X = cbind(x1 = c(0, 1, 0)), treat = c(0L, 1L, 0L), kernel = "normal", components = c(3L, 3L), mcmc = list(seed = 1L)), "wrapped_fit")
  fit_inline <- dpmgpd.causal(
    y = c(1, 2, 3),
    X = cbind(x1 = c(0, 1, 0)),
    treat = c(0L, 1L, 0L),
    kernel = "normal",
    components = c(3L, 3L),
    mcmc = list(seed = 1L),
    parallel_arms = TRUE,
    workers = 2L
  )
  expect_s3_class(fit_inline, "wrapped_fit")
  expect_true(isTRUE(fit_inline$args$parallel_arms))
  expect_equal(fit_inline$args$workers, 2L)
})

test_that("coverage-heavy build-run helpers cover codegen priors dimensions and sampler tuning", {
  expect_match(.codegen_prior_call("normal", list(mean = 0, sd = 1)), "dnorm")
  expect_match(.codegen_prior_call("gamma", list(shape = 2, rate = 1)), "dgamma")
  expect_match(.codegen_prior_call("invgamma", list(shape = 2, scale = 1)), "dinvgamma")
  expect_match(.codegen_prior_call("lognormal", list(meanlog = 0, sdlog = 1)), "dlnorm")
  expect_error(.codegen_prior_call("bad", list()), "Unsupported prior dist")

  expect_equal(.codegen_link_expr("eta", "identity"), "eta")
  expect_equal(.codegen_link_expr("eta", "exp"), "exp(eta)")
  expect_equal(.codegen_link_expr("eta", "log"), "log(eta)")
  expect_equal(.codegen_link_expr("eta", "softplus"), "log(1 + exp(eta))")
  expect_match(.codegen_link_expr("eta", "power", link_power = 2), "pow")
  expect_error(.codegen_link_expr("eta", "power"), "link requires numeric link_power")
  expect_error(.codegen_link_expr("eta", "bad"), "Unsupported link")

  y <- abs(stats::rnorm(8)) + 0.1
  X <- cbind(x1 = stats::rnorm(8), x2 = stats::runif(8))
  spec_link <- compile_model_spec(
    y = y,
    X = X,
    ps = rep(0.5, length(y)),
    backend = "spliced",
    kernel = "normal",
    GPD = TRUE,
    components = 4L,
    param_specs = list(
      bulk = list(mean = list(mode = "link", link = "identity")),
      gpd = list(
        threshold = list(mode = "link", link = "identity"),
        tail_scale = list(mode = "link", link = "exp"),
        tail_shape = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 0.3))
      )
    )
  )

  dims <- build_dimensions_from_spec(spec_link)
  mons <- build_monitors_from_spec(spec_link, monitor_v = TRUE, monitor_latent = TRUE)
  inits <- build_inits_from_spec(spec_link, seed = 1L, y = y)
  priors <- build_prior_table_from_spec(spec_link)

  expect_true(all(c("v", "w", "z", "beta_mean", "beta_ps_mean") %in% names(build_dimensions_from_spec(
    compile_model_spec(y = y, X = X, ps = rep(0.5, length(y)), backend = "sb", kernel = "normal", GPD = FALSE, components = 4L,
                       param_specs = list(bulk = list(mean = list(mode = "link", link = "identity"))))
  ))))
  expect_true(all(c("beta_threshold", "threshold_i", "beta_tail_scale", "tail_shape") %in% names(dims)))
  expect_true(any(grepl("^beta_mean", mons)))
  expect_true(any(grepl("^beta_threshold", mons)))
  expect_true(all(c("alpha", "z", "beta_mean", "beta_ps_mean", "beta_threshold", "beta_tail_scale", "tail_shape") %in% names(inits)))
  expect_true(any(priors$parameter == "tail_scale" & priors$mode == "link"))

  key <- .mcmc_cache_key(list(code = "x"), list(N = 1L), list(y = 1), list(), "alpha", TRUE)
  expect_null(.mcmc_cache_get(key))
  .mcmc_cache_set(key, list(answer = 42))
  expect_equal(.mcmc_cache_get(key)$answer, 42)

  conf <- .fake_sampler_conf()
  tuned <- .configure_samplers(conf, spec = spec_link, z_update_every = 3L)
  expect_identical(tuned, conf)
  expect_true(length(conf$removed) >= 2L)
  expect_true(any(vapply(conf$added, function(x) identical(x$type, "RW_block"), logical(1))))
  expect_true(any(vapply(conf$added, function(x) identical(x$type, "slice"), logical(1))))
  expect_true(all(vapply(conf$samplerConfs, function(x) isFALSE(is.null(x$control$checkConjugacy)), logical(1))))
})

test_that("coverage-heavy internal helpers cover silent wrappers and capture helpers", {
  wrapped <- .silent_wrapper(
    "demo_fun",
    function(x) {
      warning("suppressed warning", call. = FALSE)
      message("suppressed message")
      x + 1
    },
    "CausalMixGPD.silent"
  )
  options(CausalMixGPD.silent = TRUE)
  on.exit(options(CausalMixGPD.silent = NULL), add = TRUE)
  expect_equal(wrapped(1), 2)

  expect_equal(.cmgpd_capture_nimble({ 1 + 1 }, suppress = FALSE), 2)
  expect_equal(.cmgpd_capture_nimble({ message("hidden"); 3 }, suppress = TRUE), 3)
})

test_that("coverage-heavy runner and predictive methods cover build-run methods and internal branches", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("ggmcmc")
  skip_if_not_installed("coda")

  fit <- .coverage_heavy_fit()
  bundle_cached <- build_nimble_bundle(
    y = fit$data$y,
    X = fit$data$X,
    backend = "sb",
    kernel = "normal",
    GPD = TRUE,
    components = 4L,
    mcmc = list(niter = 60L, nburnin = 20L, thin = 1L, nchains = 1L, seed = 901L, waic = TRUE)
  )
  fit_cached <- run_mcmc_bundle_manual(bundle_cached, show_progress = FALSE, quiet = TRUE)
  expect_s3_class(fit, "mixgpd_fit")
  expect_s3_class(fit_cached, "mixgpd_fit")
  expect_true(is.list(fit$timing))
  expect_warning(
    run_mcmc_bundle_manual(
      build_nimble_bundle(
        y = fit$data$y,
        X = fit$data$X,
        backend = "sb",
        kernel = "normal",
        GPD = TRUE,
        components = 4L,
        mcmc = list(niter = 30L, nburnin = 10L, thin = 1L, nchains = 2L, seed = 910L, waic = FALSE)
      ),
      show_progress = FALSE,
      quiet = TRUE,
      parallel_chains = TRUE
    ),
    "falls back to sequential execution"
  )

  Xp <- fit$data$X[1:4, , drop = FALSE]
  yp <- fit$data$y[1:4]
  pred_mean <- predict(fit, newdata =Xp, type = "mean", nsim_mean = 20L, interval = "credible", show_progress = FALSE)
  pred_location <- predict(fit, newdata =Xp, type = "location", interval = "hpd", show_progress = FALSE)
  pred_quant <- predict(fit, newdata =Xp, type = "quantile", index = c(0.25, 0.75), interval = "credible", show_progress = FALSE)
  pred_median <- predict(fit, newdata =Xp, type = "median", show_progress = FALSE)
  pred_rmean <- predict(fit, newdata =Xp, type = "rmean", cutoff = 3, show_progress = FALSE)
  pred_density <- predict(fit, newdata =Xp[1:2, , drop = FALSE], y = yp[1:2], type = "density", interval = NULL, show_progress = FALSE)
  pred_survival <- predict(fit, newdata =Xp[1:2, , drop = FALSE], y = yp[1:2], type = "survival", interval = NULL, show_progress = FALSE)
  pred_sample <- predict(fit, newdata =Xp, type = "sample", nsim = 5L, store_draws = FALSE, show_progress = FALSE)

  expect_s3_class(pred_mean, "mixgpd_predict")
  expect_s3_class(pred_location, "mixgpd_predict")
  expect_s3_class(pred_quant, "mixgpd_predict")
  expect_s3_class(pred_median, "mixgpd_predict")
  expect_s3_class(pred_rmean, "mixgpd_predict")
  expect_s3_class(pred_density, "mixgpd_predict")
  expect_s3_class(pred_survival, "mixgpd_predict")
  expect_s3_class(pred_sample, "mixgpd_predict")

  fit_loc <- fitted(fit, type = "location", interval = "credible")
  fit_q <- fitted(fit, type = "quantile", p = 0.75, interval = NULL)
  expect_s3_class(fit_loc, "mixgpd_fitted")
  expect_s3_class(fit_q, "mixgpd_fitted")

  res_raw <- residuals(fit, type = "raw", fitted_type = "median")
  res_pit_plugin <- residuals(fit, type = "pit", pit = "plugin")
  expect_length(res_raw, nrow(fit$data$X))
  expect_length(res_pit_plugin, nrow(fit$data$X))
  expect_error(residuals(fit, type = "pit", pit = "bayes_mean", pit_seed = 1L), "argument \"mean\" is missing")
  expect_error(residuals(fit, type = "pit", pit = "bayes_draw", pit_seed = 1L), "argument \"mean\" is missing")

  pars <- params(fit)
  fit_sum <- summary(fit)
  ess <- ess_summary(fit, per_chain = TRUE)
  expect_s3_class(pars, "mixgpd_params")
  expect_s3_class(fit_sum, "mixgpd_summary")
  expect_s3_class(ess, "mixgpd_ess_summary")
  expect_output(print(pars))
  expect_output(print(fit_sum))
  expect_output(print(ess))
  expect_true(is.data.frame(summary(ess)))

  expect_s3_class(plot(fit, family = c("traceplot", "density"), params = "alpha"), "mixgpd_fit_plots")
  expect_s3_class(plot(pred_mean), "mixgpd_predict_plots")
  expect_s3_class(plot(fit_loc), "mixgpd_fitted_plots")
})

test_that("coverage-heavy causal methods and summaries cover methods branches", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if_not_installed("ggplot2")

  fit <- .coverage_heavy_causal_fit()
  Xp <- fit$bundle$data$X[1:4, , drop = FALSE]
  yp <- fit$bundle$data$y[1:2]

  expect_output(print(fit), "CausalMixGPD causal fit")
  expect_output(print(summary(fit)), "Outcome fits")

  pars <- params(fit)
  expect_s3_class(pars, "mixgpd_params_pair")

  pred_mean <- predict(fit, newdata =Xp, type = "mean", nsim_mean = 20L, interval = "credible", show_progress = FALSE)
  pred_quant <- predict(fit, newdata =Xp, type = "quantile", p = c(0.25, 0.75), interval = "credible", show_progress = FALSE)
  pred_density <- predict(fit, newdata =Xp[1:2, , drop = FALSE], y = yp, type = "density", interval = NULL, show_progress = FALSE)
  expect_s3_class(pred_mean, "causalmixgpd_causal_predict")
  expect_s3_class(pred_quant, "causalmixgpd_causal_predict")
  expect_s3_class(pred_density, "causalmixgpd_causal_predict")

  qte_obj <- qte(fit, probs = c(0.25, 0.75), interval = "credible", show_progress = FALSE)
  ate_obj <- ate(fit, interval = "credible", nsim_mean = 20L, show_progress = FALSE)
  cqte_obj <- cqte(fit, probs = c(0.25, 0.75), newdata = Xp, interval = "credible", show_progress = FALSE)
  cate_obj <- cate(fit, newdata = Xp, interval = "credible", nsim_mean = 20L, show_progress = FALSE)
  expect_output(print(qte_obj), "QTE")
  expect_output(print(ate_obj), "ATE")
  expect_output(print(cqte_obj), "CQTE")
  expect_output(print(cate_obj), "CATE")
  expect_output(print(summary(qte_obj)), "QTE Summary")
  expect_output(print(summary(ate_obj)), "ATE Summary")

  pred_plot <- plot(pred_mean)
  fit_plot <- plot(fit, arm = "both")
  qte_plot <- plot(qte_obj)
  ate_plot <- plot(ate_obj)
  qte_effect_plot <- plot(qte_obj, type = "effect")
  qte_arms_plot <- plot(qte_obj, type = "arms")
  ate_effect_plot <- plot(ate_obj, type = "effect")
  ate_arms_plot <- plot(ate_obj, type = "arms")
  cqte_plot <- plot(cqte_obj, type = "both")
  cate_plot <- plot(cate_obj, type = "both")
  expect_s3_class(pred_plot, "causalmixgpd_causal_predict_plots")
  expect_s3_class(fit_plot, "causalmixgpd_causal_fit_plots")
  expect_true(is.list(qte_plot))
  expect_true(is.list(ate_plot))
  expect_s3_class(qte_effect_plot, "ggplot")
  expect_s3_class(qte_arms_plot, "ggplot")
  expect_s3_class(ate_effect_plot, "ggplot")
  expect_s3_class(ate_arms_plot, "ggplot")
  expect_true(is.list(cqte_plot))
  expect_true(is.list(cate_plot))
})

test_that("coverage-heavy bundle methods cover knitr presentation branches", {
  skip_if_not_installed("knitr")
  skip_if_not_installed("kableExtra")

  old_knitr <- getOption("knitr.in.progress")
  old_kable <- getOption("causalmixgpd.knitr.kable")
  options(knitr.in.progress = TRUE, causalmixgpd.knitr.kable = TRUE)
  on.exit(options(knitr.in.progress = old_knitr, causalmixgpd.knitr.kable = old_kable), add = TRUE)

  set.seed(905)
  y <- abs(stats::rnorm(12L)) + 0.1
  X <- cbind(x1 = stats::rnorm(12L), x2 = stats::runif(12L))
  A <- rep(c(0L, 1L), length.out = 12L)
  b <- build_nimble_bundle(y = y, X = X, backend = "sb", kernel = "normal", GPD = FALSE, components = 3L)
  cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal", GPD = FALSE, components = 3L)

  expect_silent(print(b))
  expect_silent(summary(b))
  expect_s3_class(print(cb), "knit_asis")
  expect_s3_class(summary(cb, code = TRUE, max_code_lines = 5L), "knit_asis")
})

test_that("coverage-heavy wrappers cover parsing treatment normalization and mcmc dispatch", {
  dat <- data.frame(
    y = c(1, 2, 3, 4, 5),
    x1 = c(0.1, 0.2, NA, 0.4, 0.5),
    id = 11:15,
    trt = factor(c("control", "treated", "control", "treated", "control"))
  )

  parsed <- .parse_formula_yX(y ~ x1 + id, data = dat)
  parsed_uncond <- .parse_formula_yX(y ~ 1, data = dat)
  expect_equal(colnames(parsed$X), "x1")
  expect_true(parsed_uncond$is_unconditional)

  expect_equal(.coerce_treat(factor(c("control", "treated", "control"))), c(0L, 1L, 0L))
  expect_equal(.coerce_treat(c(FALSE, TRUE, FALSE)), c(0L, 1L, 0L))
  expect_equal(.coerce_treat(c("control", "treated", "treated")), c(0L, 1L, 1L))
  expect_error(.coerce_treat(c(0, 2)), "binary")
  expect_error(.coerce_treat(c(0, NA)), "cannot contain NA")

  expect_equal(.extract_treat_from_data(parsed$mf, data = dat, treat = "trt"), c(0L, 1L, 1L, 0L))
  expect_equal(
    .extract_treat_from_data(parsed$mf, data = dat, treat = c(0L, 1L, 0L, 1L, 0L)),
    c(0L, 1L, 1L, 0L)
  )
  expect_error(.extract_treat_from_data(parsed$mf, data = dat, treat = 0:1), "length does not match")

  parsed_mcmc <- .normalize_mcmc_inputs(list(niter = 10L, nburn = 3L, quiet = TRUE, timing = TRUE))
  expect_equal(parsed_mcmc$overrides$nburnin, 3L)
  expect_true(isTRUE(parsed_mcmc$runner$quiet))
  expect_error(.normalize_mcmc_inputs(list(1L)), "must be named")
  expect_error(.normalize_mcmc_inputs(list(bad = 1L)), "Unknown mcmc argument")

  b <- structure(list(mcmc = list(niter = 10L), spec = list(meta = list(GPD = FALSE))), class = "causalmixgpd_bundle")
  cb <- structure(
    list(
      outcome = list(con = b, trt = b),
      design = structure(list(mcmc = list(niter = 5L)), class = "causalmixgpd_ps_bundle"),
      meta = list(GPD = list(con = FALSE, trt = FALSE))
    ),
    class = "causalmixgpd_causal_bundle"
  )

  expect_equal(.apply_mcmc_overrides(b, list(seed = 3L))$mcmc$seed, 3L)
  cb_over <- .apply_mcmc_overrides(cb, list(seed = 9L))
  expect_equal(cb_over$outcome$con$mcmc$seed, 9L)
  expect_equal(cb_over$design$mcmc$seed, 9L)

  testthat::local_mocked_bindings(
    build_nimble_bundle = function(y, X = NULL, GPD = FALSE, ...) {
      structure(list(kind = "one-arm", y = y, X = X, GPD = GPD), class = "causalmixgpd_bundle")
    },
    build_causal_bundle = function(y, X = NULL, A = NULL, GPD = FALSE, ...) {
      structure(list(kind = "causal", y = y, X = X, A = A, GPD = GPD), class = "causalmixgpd_causal_bundle")
    },
    run_mcmc_bundle_manual = function(bundle, ...) structure(list(bundle = bundle), class = "mixgpd_fit"),
    run_mcmc_causal = function(bundle, ...) structure(list(bundle = bundle), class = "causalmixgpd_causal_fit"),
    .package = "CausalMixGPD"
  )

  built_one <- bundle(formula = y ~ x1 + id, data = dat, backend = "sb", kernel = "normal", components = 3L)
  built_causal <- bundle(
    formula = y ~ x1 + id,
    data = dat,
    treat = "trt",
    backend = "sb",
    kernel = "normal",
    components = 3L
  )
  built_causal_symbol <- bundle(
    formula = y ~ x1 + id,
    data = dat,
    treat = trt,
    backend = "sb",
    kernel = "normal",
    components = 3L
  )
  expect_s3_class(built_one, "causalmixgpd_bundle")
  expect_s3_class(built_causal, "causalmixgpd_causal_bundle")
  expect_s3_class(built_causal_symbol, "causalmixgpd_causal_bundle")
  expect_s3_class(mcmc(b, quiet = TRUE), "mixgpd_fit")
  expect_s3_class(mcmc(cb, quiet = TRUE), "causalmixgpd_causal_fit")
  expect_s3_class(
    dpmix.causal(
      formula = y ~ x1 + id,
      data = dat,
      treat = trt,
      backend = "sb",
      kernel = "normal",
      components = 3L,
      mcmc = list(quiet = TRUE)
    ),
    "causalmixgpd_causal_fit"
  )
  expect_s3_class(
    dpmgpd.causal(
      formula = y ~ x1 + id,
      data = dat,
      treat = trt,
      backend = "sb",
      kernel = "normal",
      components = 3L,
      mcmc = list(quiet = TRUE)
    ),
    "causalmixgpd_causal_fit"
  )
  expect_error(
    dpmix(
      formula = y ~ x1 + id,
      data = dat,
      treat = trt,
      backend = "sb",
      kernel = "normal",
      components = 3L,
      mcmc = list()
    ),
    "Use dpmix.causal"
  )
  expect_error(
    dpmgpd(
      formula = y ~ x1 + id,
      data = dat,
      treat = trt,
      backend = "sb",
      kernel = "normal",
      components = 3L,
      mcmc = list()
    ),
    "Use dpmgpd.causal"
  )
  expect_error(mcmc(b, parallel_arms = TRUE), "Unsupported runner argument")
  expect_error(dpmix.causal(y = c(1, 2, 3), kernel = "normal", components = 3L, mcmc = list()), "requires 'treat'")
  expect_error(dpmgpd.causal(y = b, mcmc = list()), "requires a causal bundle")
  expect_error(dpmgpd(y = b, mcmc = list()), "requires a bundle with GPD enabled")
})

test_that("coverage-heavy internal helpers cover draw coercion id handling and plotting helpers", {
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("coda")

  fit <- .coverage_heavy_fit()
  fit_u <- .coverage_heavy_uncond_fit()

  expect_silent(.validate_nimble_reserved_names(c("alpha", "beta"), context = "columns"))
  expect_error(.validate_nimble_reserved_names(c("if", "alpha"), context = "columns"), "reserved NIMBLE keywords")

  wrapped_code <- .wrap_nimble_code(quote(alpha <- 1))
  expect_true(is.list(wrapped_code))
  expect_identical(.extract_nimble_code(list(code = quote(alpha <- 2))), quote(alpha <- 2))

  expect_length(.plot_palette(12L), 12L)
  p_fill <- ggplot2::ggplot(data.frame(x = c("a", "b"), y = c(1, 2), g = c(1, 2)),
                            ggplot2::aes(x = x, y = y, fill = g)) +
    ggplot2::geom_col() +
    ggplot2::scale_fill_viridis_c()
  p_stripped <- .strip_fill_scales(p_fill)
  expect_false(any(vapply(p_stripped$scales$scales, function(s) "fill" %in% (s$aesthetics %||% character(0)), logical(1))))
  expect_identical(.wrap_plotly("plain-object"), "plain-object")

  id_info <- .resolve_predict_id(data.frame(id = letters[1:2], x1 = 1:2), id = "id")
  expect_equal(id_info$id, letters[1:2])
  expect_false("id" %in% names(id_info$x))
  expect_error(.resolve_predict_id(matrix(1:4, ncol = 2), id = "id"), "data.frame")
  expect_error(.resolve_predict_id(NULL, id = 1:2), "requires 'x'/'newdata'")

  reordered <- .reorder_predict_cols(data.frame(lower = 1, estimate = 2, id = 3, upper = 4, y = 5, misc = 6))
  expect_identical(names(reordered), c("id", "y", "estimate", "lower", "upper", "misc"))

  fit_df_matrix <- .coerce_fit_df(matrix(c(1, 2, 3, 4), nrow = 2), probs = c(0.25, 0.75))
  fit_df_vector <- .coerce_fit_df(c(1, 2, 3))
  expect_true(all(c("id", "index", "estimate", "lower", "upper") %in% names(fit_df_matrix)))
  expect_true(all(c("id", "estimate", "lower", "upper") %in% names(fit_df_vector)))
  expect_error(.coerce_fit_df(list(a = 1)), "unsupported type")

  sb_weights <- matrix(c(0.6, 0.4, 0.7, 0.3), nrow = 2, byrow = TRUE,
                       dimnames = list(NULL, c("w[1]", "w[2]")))
  crp_weights <- matrix(c(1, 2, 1, 2, 2, 1), nrow = 2, byrow = TRUE,
                        dimnames = list(NULL, c("z[1]", "z[2]", "z[3]")))
  bulk_draws <- cbind(
    "mu[1]" = c(1, 2),
    "mu[2]" = c(3, 4),
    "sigma[1]" = c(0.5, 0.6)
  )
  expect_equal(.extract_weights(sb_weights, backend = "sb")[1, 1], 0.6)
  expect_equal(ncol(.extract_weights(crp_weights, backend = "crp")), 2L)
  expect_true(all(c("mu", "sigma") %in% names(.extract_bulk_params(bulk_draws, bulk_params = c("mu", "sigma")))))

  expect_equal(.get_epsilon(fit), fit$epsilon)
  expect_true(isTRUE(.validate_fit(fit)))
  expect_s3_class(.get_samples_mcmclist(fit), "mcmc.list")
  expect_equal(.get_nobs(fit_u), length(fit_u$data$y))

  post_vec <- .posterior_summarize(c(1, 2, 3), interval = NULL)
  post_mat <- .posterior_summarize(matrix(1:6, nrow = 2), interval = "credible")
  expect_equal(post_vec$estimate, 2)
  expect_length(post_mat$estimate, 2L)
})

test_that("coverage-heavy internal advanced helpers cover dispatch truncation and scalar wrappers", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("coda")

  fit <- .coverage_heavy_fit()

  add_one <- .wrap_scalar_first_arg(function(x, offset = 0) x + offset, "x")
  cdf_wrap <- .wrap_scalar_p(function(q, shift = 0) q + shift)
  rng_wrap <- .wrap_scalar_r(function(n, mu = 1) rep(mu, 1L))

  expect_equal(add_one(x = c(1, 2), offset = 2), c(3, 4))
  expect_equal(cdf_wrap(x = c(1, 2), shift = 1), c(2, 3))
  expect_equal(rng_wrap(n = 3L, mu = 2), c(2, 2, 2))
  expect_equal(length(rng_wrap(n = 0L, mu = 2)), 0L)

  expect_equal(.detect_first_present(list(q = 1), candidates = c("q", "x")), "q")
  expect_error(.detect_first_present(list(), candidates = c("q", "x")), "Expected one of")

  trunc <- .truncate_components_one_draw(
    w = c(0.6, 0.3, 0.1),
    params = list(mu = c(1, 2, 3)),
    epsilon = 0.2
  )
  expect_equal(trunc$k, 2L)
  expect_error(.truncate_components_one_draw(w = c(0.6, 0.4), params = list(mu = 1), epsilon = 0.2), "length K")

  ci <- .compute_interval(c(1, 2, 3, 4), level = 0.5, type = "credible")
  expect_true(all(c("lower", "upper") %in% names(ci)))

  draws_first <- .extract_draws(fit, chains = "first")
  trunc_info <- .truncation_info(fit)
  fit_header <- .format_fit_header(fit)
  post_sum <- .summarize_posterior(fit, pars = "alpha")
  dispatch_scalar <- .get_dispatch_scalar(fit)
  dispatch <- .get_dispatch(fit)

  expect_true(is.matrix(draws_first))
  expect_true(is.list(trunc_info))
  expect_true(length(fit_header) >= 2L)
  expect_true(is.data.frame(post_sum))
  expect_true(is.function(dispatch_scalar$d))
  expect_true(is.function(dispatch$d))
  expect_error(.extract_draws(fit, pars = "missing_param"), "Unknown params")
  expect_error(.summarize_posterior(fit, pars = "missing_param"), "Unknown params")
})

test_that("coverage-heavy build-run helpers cover data constants priors and error branches", {
  y <- abs(stats::rnorm(8L)) + 0.1
  X <- cbind(x1 = stats::rnorm(8L), x2 = stats::runif(8L))
  ps <- rep(0.5, length(y))

  data_obj <- build_data_from_inputs(y = y, X = X, ps = ps)
  expect_true(all(c("y", "X", "ps") %in% names(data_obj)))
  expect_error(build_data_from_inputs(y = numeric(0)), "non-empty")
  expect_error(build_data_from_inputs(y = y, X = matrix(1, nrow = 7, ncol = 1)), "nrow\\(X\\)")
  expect_error(build_data_from_inputs(y = y, X = matrix(numeric(0), nrow = 8, ncol = 0)), "at least one column")
  expect_error(build_data_from_inputs(y = y, ps = 1:3), "same length")

  spec_sb <- compile_model_spec(
    y = y,
    X = X,
    ps = ps,
    backend = "sb",
    kernel = "normal",
    GPD = TRUE,
    components = 4L,
    param_specs = list(
      bulk = list(
        mean = list(mode = "link", link = "identity"),
        sd = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
      ),
      gpd = list(
        threshold = list(mode = "link", link = "identity", link_dist = list(dist = "lognormal")),
        tail_scale = list(mode = "link", link = "exp"),
        tail_shape = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 0.2))
      )
    )
  )
  spec_spliced <- compile_model_spec(
    y = y,
    X = X,
    backend = "spliced",
    kernel = "normal",
    GPD = TRUE,
    components = 4L,
    param_specs = list(
      bulk = list(mean = list(mode = "link", link = "identity")),
      gpd = list(
        threshold = list(mode = "link", link = "identity"),
        tail_scale = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_shape = list(mode = "link", link = "identity")
      )
    )
  )

  const_spec <- spec_sb
  const_spec$plan$concentration <- list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
  const_spec$plan$bulk$mean <- list(
    mode = "link",
    beta_prior = list(dist = "normal", args = list(mean = 0, sd = 2))
  )
  const_spec$plan$ps <- list(prior = list(dist = "normal", args = list(mean = 0, sd = 2)))
  const_spec$plan$gpd$threshold <- list(
    mode = "link",
    beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.2)),
    link_dist = list(dist = "lognormal")
  )
  const_spec$plan$gpd$sdlog_u <- list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1))
  const_spec$plan$gpd$tail_scale <- list(
    mode = "link",
    beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.5))
  )
  const_sb <- build_constants_from_spec(const_spec)
  dims_sp <- build_dimensions_from_spec(spec_spliced)
  mons_sp <- build_monitors_from_spec(spec_spliced, monitor_v = TRUE, monitor_latent = TRUE)
  inits_sp <- build_inits_from_spec(spec_spliced, seed = 2L, y = y)
  prior_sp <- build_prior_table_from_spec(spec_spliced)

  expect_true(all(c("N", "P", "components") %in% names(const_sb)))
  expect_true(all(c("beta_threshold", "tail_scale", "beta_tail_shape") %in% names(dims_sp)))
  expect_true(any(grepl("^beta_tail_shape", mons_sp)))
  expect_true(all(c("beta_threshold", "tail_scale", "beta_tail_shape") %in% names(inits_sp)))
  expect_true(any(prior_sp$parameter == "tail_shape" & prior_sp$mode == "link"))

  bad_const <- spec_sb
  bad_const$plan$bulk$mean$beta_prior$dist <- "bad"
  expect_error(build_constants_from_spec(bad_const), "Unsupported prior distribution")

  bad_mon <- spec_spliced
  bad_mon$plan$bulk$mean$mode <- "bad"
  expect_error(build_monitors_from_spec(bad_mon), "Invalid bulk plan mode")

  bad_inits <- spec_spliced
  bad_inits$plan$gpd$tail_scale$mode <- "bad"
  expect_error(build_inits_from_spec(bad_inits, y = y), "Invalid gpd\\$tail_scale mode")
})

test_that("coverage-heavy methods cover bundle ps summary and cluster printers", {
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("nimble")

  set.seed(907)
  y <- abs(stats::rnorm(10L)) + 0.1
  X <- cbind(x1 = stats::rnorm(10L), x2 = stats::runif(10L))
  bundle_obj <- build_nimble_bundle(y = y, X = X, backend = "sb", kernel = "normal", GPD = FALSE, components = 3L)
  ps_bundle <- structure(
    list(
      spec = list(meta = list(type = "ps_logit", include_intercept = TRUE)),
      code = quote({
        beta0 ~ dnorm(0, sd = 1)
      })
    ),
    class = "causalmixgpd_ps_bundle"
  )
  ps_fit <- structure(list(bundle = ps_bundle), class = "causalmixgpd_ps_fit")
  empty_params <- structure(list(), class = "mixgpd_params")
  param_pair <- structure(
    list(
      treated = structure(list(alpha = 1, w = c(0.6, 0.4)), class = "mixgpd_params"),
      control = structure(list(alpha = 2, w = c(0.5, 0.5)), class = "mixgpd_params")
    ),
    class = "mixgpd_params_pair"
  )
  summary_obj <- structure(
    list(
      model = list(
        backend = "sb",
        kernel = "normal",
        gpd = FALSE,
        epsilon = 0.1,
        truncation = list(Kt = 2L),
        n = 10L,
        components = 3L
      ),
      waic = list(WAIC = 1.23, lppd = 0.4, pWAIC = 0.2),
      table = data.frame(
        parameter = paste0("p", 1:3),
        mean = 1:3,
        sd = rep(0.1, 3),
        q0.025 = rep(0.5, 3),
        q0.500 = 1:3,
        q0.975 = rep(3.5, 3)
      )
    ),
    class = "mixgpd_summary"
  )
  cluster_obj <- .coverage_heavy_cluster_fit()
  cluster_fit <- cluster_obj$fit
  cluster_lbl <- predict(cluster_fit, type = "label", return_scores = TRUE)
  cluster_psm <- predict(cluster_fit, type = "psm")

  expect_output(print(bundle_obj, code = TRUE, max_code_lines = 3L), "CausalMixGPD bundle")
  expect_output(summary(bundle_obj), "Parameter specification")
  expect_output(print(ps_bundle, code = TRUE, max_code_lines = 2L), "PS bundle")
  expect_output(summary(ps_bundle, code = TRUE), "PS bundle")
  expect_output(print(ps_fit), "CausalMixGPD PS fit")
  expect_output(summary(ps_fit), "CausalMixGPD PS fit")
  expect_output(print(empty_params), "<empty>")
  expect_output(print(param_pair), "treated")
  expect_output(print(summary_obj, max_rows = 2L), "Showing first 2")
  expect_output(print(structure(list(table = data.frame(), overall = data.frame(), meta = list()), class = "mixgpd_ess_summary")), "No matched parameters")
  expect_output(print(.coverage_heavy_fit()), "MixGPD fit")
  expect_output(print(cluster_fit), "Cluster fit")
  expect_output(print(cluster_lbl), "Cluster labels")
  expect_output(print(cluster_psm), "Cluster PSM")
})

test_that("coverage-heavy methods cover plotting families causal prediction and cluster branches", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if_not_installed("ggplot2")
  skip_if_not_installed("ggmcmc")
  skip_if_not_installed("coda")

  fit <- .coverage_heavy_fit()
  fit2 <- .coverage_heavy_fit_2chain()
  fit_loc <- fitted(fit, type = "mean", interval = "credible")
  fit_q <- fitted(fit, type = "quantile", p = 0.5, interval = "credible")
  pred_density <- predict(
    fit,
    x = fit$data$X[1:2, , drop = FALSE],
    y = fit$data$y[1:2],
    type = "density",
    interval = "credible",
    show_progress = FALSE
  )
  pred_sample <- predict(fit, newdata =fit$data$X[1:3, , drop = FALSE], type = "sample", nsim = 3L, show_progress = FALSE)
  pred_survival <- predict(
    fit,
    x = fit$data$X[1:2, , drop = FALSE],
    y = fit$data$y[1:2],
    type = "survival",
    interval = "credible",
    show_progress = FALSE
  )
  fit_summary <- summary(fit, pars = "alpha")
  fit_params <- params(.coverage_heavy_uncond_fit())
  cluster_obj <- .coverage_heavy_cluster_fit()
  cluster_fit <- cluster_obj$fit
  cluster_lbl <- predict(cluster_fit, type = "label", return_scores = TRUE)
  cluster_psm <- predict(cluster_fit, type = "psm")

  expect_s3_class(
    suppressWarnings(plot(fit2, family = c("histogram", "running", "compare_partial", "autocorrelation", "geweke", "caterpillar"), params = "alpha")),
    "mixgpd_fit_plots"
  )
  expect_s3_class(
    suppressWarnings(plot(fit2, family = c("crosscorrelation", "Rhat", "effective"), params = c("alpha", "w\\[1\\]"))),
    "mixgpd_fit_plots"
  )
  expect_output(print(fit_summary), "MixGPD summary")
  expect_output(print(fit_params), "Posterior mean parameters")
  expect_s3_class(plot(fit_loc), "mixgpd_fitted_plots")
  expect_silent(print(plot(fit_loc)))
  expect_s3_class(plot(fit_q), "mixgpd_fitted_plots")
  expect_s3_class(plot(pred_density), "mixgpd_predict_plots")
  expect_silent(print(plot(pred_density)))
  expect_s3_class(plot(pred_sample), "mixgpd_predict_plots")
  expect_s3_class(plot(pred_survival), "mixgpd_predict_plots")

  expect_silent(summary(cluster_fit))
  expect_s3_class(plot(cluster_fit, which = "psm", plotly = FALSE), "ggplot")
  expect_s3_class(plot(cluster_fit, which = "k", plotly = FALSE), "ggplot")
  expect_s3_class(plot(cluster_fit, which = "sizes", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")
  expect_silent(summary(cluster_lbl))
  expect_s3_class(plot(cluster_lbl, type = "sizes", top_n = 2L, order_by = "label", plotly = FALSE), "ggplot")
  expect_silent(plot(cluster_lbl, type = "certainty"))
  expect_silent(summary(cluster_psm))
  expect_silent(plot(cluster_psm, psm_max_n = nrow(cluster_psm$psm)))

  cp_mean <- structure(
    data.frame(ps = c(0.2, 0.8), estimate = c(1, 2), lower = c(0.8, 1.8), upper = c(1.2, 2.2)),
    class = c("causalmixgpd_causal_predict", "data.frame")
  )
  attr(cp_mean, "type") <- "mean"
  attr(cp_mean, "trt") <- list(fit = data.frame(id = 1:2, estimate = c(2, 3), lower = c(1.5, 2.5), upper = c(2.5, 3.5)))
  attr(cp_mean, "con") <- list(fit = data.frame(id = 1:2, estimate = c(1, 1.5), lower = c(0.5, 1.0), upper = c(1.5, 2.0)))
  cp_density <- structure(
    data.frame(
      y = c(0.2, 0.4),
      trt_estimate = c(0.4, 0.3),
      trt_lower = c(0.3, 0.2),
      trt_upper = c(0.5, 0.4),
      con_estimate = c(0.2, 0.15),
      con_lower = c(0.1, 0.05),
      con_upper = c(0.3, 0.25)
    ),
    class = c("causalmixgpd_causal_predict", "data.frame")
  )
  attr(cp_density, "type") <- "density"

  expect_s3_class(plot(cp_mean), "causalmixgpd_causal_predict_plots")
  expect_s3_class(plot(cp_density), "ggplot")
  expect_output(print(plot(cp_mean)))
  expect_s3_class(plot(.coverage_heavy_causal_fit(), arm = 1L), "mixgpd_fit_plots")
  expect_s3_class(plot(.coverage_heavy_causal_fit(), arm = 0L), "mixgpd_fit_plots")
})

test_that("coverage-heavy glue diagnostics cover link-mode, unconditional, and validation branches", {
  check_glue_validity <- getFromNamespace("check_glue_validity", "CausalMixGPD")

  dispatch_stub <- list(
    bulk_params = c("mean", "sd"),
    mean = function(w, mean, sd) sum((w / sum(w)) * mean),
    d = function(x, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL, log = 0L) {
      dens <- stats::dnorm(x, mean = mean[1], sd = max(sd[1], 1e-6))
      if (log == 1L) log(pmax(dens, 1e-300)) else pmax(dens, 1e-300)
    },
    p = function(q, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL,
                 lower.tail = 1L, log.p = 0L) {
      cdf <- stats::pnorm(q, mean = mean[1], sd = max(sd[1], 1e-6))
      if (lower.tail == 0L) cdf <- 1 - cdf
      if (log.p == 1L) log(pmax(cdf, 1e-300)) else cdf
    }
  )

  fake_fit <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", GPD = TRUE, has_X = TRUE),
        dispatch = list(
          backend = "sb",
          GPD = TRUE,
          gpd = list(
            threshold = list(mode = "link", link = "identity"),
            tail_scale = list(mode = "link", link = "exp")
          )
        )
      ),
      data = list(
        X = matrix(c(0, 1, 1, 0), nrow = 2, byrow = TRUE, dimnames = list(NULL, c("x1", "x2"))),
        y = c(0.25, 1.1)
      )
    ),
    class = "mixgpd_fit"
  )
  fake_uncond <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", GPD = FALSE, has_X = FALSE),
        dispatch = list(backend = "sb", GPD = FALSE)
      ),
      data = list(y = c(0.2, 0.6, 1.0))
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .get_dispatch = function(...) dispatch_stub,
    .extract_draws_matrix = function(fit) {
      if (isTRUE(fit$spec$meta$GPD)) {
        matrix(c(0.15, 0.2, 0.25), ncol = 1, dimnames = list(NULL, "tail_shape"))
      } else {
        matrix(1:6, nrow = 3, dimnames = list(NULL, c("alpha", "beta")))
      }
    },
    .extract_weights = function(draw_mat, backend) {
      matrix(c(0.7, 0.3, 0.6, 0.4, 0.55, 0.45), nrow = nrow(draw_mat), byrow = TRUE)
    },
    .extract_bulk_params = function(draw_mat, bulk_params) {
      list(
        mean = matrix(c(0.0, 0.8, 0.2, 1.0, 0.4, 1.2), nrow = nrow(draw_mat), byrow = TRUE),
        sd = matrix(c(0.9, 1.1, 1.0, 1.2, 1.1, 1.3), nrow = nrow(draw_mat), byrow = TRUE)
      )
    },
    .indexed_block = function(draw_mat, name, K) {
      if (identical(name, "beta_threshold")) {
        return(matrix(c(0.4, 0.1, 0.5, 0.15, 0.6, 0.2), nrow = nrow(draw_mat), byrow = TRUE))
      }
      if (identical(name, "beta_tail_scale")) {
        return(matrix(log(c(0.9, 1.1, 1.0, 1.2, 1.1, 1.3)), nrow = nrow(draw_mat), byrow = TRUE))
      }
      stop("unexpected indexed block request")
    },
    .package = "CausalMixGPD"
  )

  set.seed(11)
  res <- check_glue_validity(fake_fit, grid = c(-1, 0, 1, 2), n_draws = 2L, check_continuity = TRUE)
  expect_true(all(unlist(res$pass), na.rm = TRUE))
  expect_equal(res$n_checked_draws, 2L)
  expect_equal(res$n_x, 2L)

  res_uncond <- check_glue_validity(fake_uncond, grid = c(0.1, 0.5, 0.9), n_draws = 1L, check_continuity = FALSE)
  expect_true(all(unlist(res_uncond$pass[1:3]), na.rm = TRUE))
  expect_true(is.na(res_uncond$pass$continuity))
  expect_equal(res_uncond$n_x, 1L)

  expect_error(check_glue_validity(fake_uncond, newdata =matrix(1, ncol = 1), grid = c(0.1, 0.5)), "Unconditional model")
  expect_error(check_glue_validity(fake_uncond, grid = c(0.1, NA_real_)), "finite numeric")
  expect_error(check_glue_validity(fake_uncond, grid = c(0.1, 0.5), n_draws = 0L), ">= 1")
})

test_that("coverage-heavy runner mocks cover cache, compile fallback, and validation branches", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")
  skip_if(nzchar(Sys.getenv("COVERAGE")), "Runner mock coverage block is unstable under covr instrumentation")

  cache_store <- new.env(parent = emptyenv())
  run_mode <- "fallback"

  testthat::local_mocked_bindings(
    .cmgpd_progress_start = function(...) list(),
    .cmgpd_progress_step = function(...) invisible(NULL),
    .cmgpd_progress_done = function(...) invisible(NULL),
    .cmgpd_message = function(...) invisible(NULL),
    .cmgpd_capture_nimble = function(x, suppress = FALSE) x,
    .extract_nimble_code = function(code) code,
    .mcmc_cache_key = function(...) "coverage-heavy-runner-key",
    .mcmc_cache_get = function(key) {
      if (exists(key, envir = cache_store, inherits = FALSE)) {
        get(key, envir = cache_store, inherits = FALSE)
      } else {
        NULL
      }
    },
    .mcmc_cache_set = function(key, value) {
      assign(key, value, envir = cache_store)
      invisible(value)
    },
    .configure_samplers = function(conf, ...) {
      conf$configured <- TRUE
      conf
    },
    .package = "CausalMixGPD"
  )
  testthat::local_mocked_bindings(
    nimbleModel = function(...) structure(list(kind = "Rmodel"), class = "fake_nimble_model"),
    configureMCMC = function(...) list(kind = "conf"),
    buildMCMC = function(conf) list(kind = "Rmcmc", conf = conf),
    compileNimble = function(...) stop("mock compile failure"),
    runMCMC = function(...) {
      if (identical(run_mode, "fallback")) {
        run_mode <<- "matrix"
        stop("mock run failure")
      }
      if (identical(run_mode, "list")) {
        return(list(
          samples = matrix(c(5, 6, 7), ncol = 1, dimnames = list(NULL, "alpha")),
          WAIC = list(WAIC = 2.34, lppd = 0.4, pWAIC = 0.2)
        ))
      }
      matrix(c(1, 2, 3), ncol = 1, dimnames = list(NULL, "alpha"))
    },
    calculateWAIC = function(...) list(WAIC = 1.23, lppd = 0.3, pWAIC = 0.1),
    .package = "nimble"
  )

  bundle_stub <- structure(
    list(
      spec = list(meta = list()),
      mcmc = list(niter = 8L, nburnin = 2L, thin = 1L, nchains = 1L, seed = 101L, waic = TRUE),
      code = quote(alpha ~ dnorm(0, 1)),
      constants = list(N = 2L),
      data = list(y = c(1, 2)),
      dimensions = list(),
      monitors = "alpha",
      inits = list(alpha = 0),
      epsilon = 0.05
    ),
    class = "causalmixgpd_bundle"
  )

  expect_warning(
    fit1 <- run_mcmc_bundle_manual(bundle_stub, show_progress = FALSE, quiet = TRUE, timing = TRUE),
    "compilation failed"
  )
  expect_s3_class(fit1, "mixgpd_fit")
  expect_identical(fit1$mcmc$engine, "uncompiled")
  expect_equal(fit1$waic$WAIC, 1.23)
  expect_true(isFALSE(fit1$timing$cache_hit))

  run_mode <- "list"
  fit2 <- run_mcmc_bundle_manual(bundle_stub, show_progress = FALSE, quiet = TRUE, timing = TRUE)
  expect_s3_class(fit2, "mixgpd_fit")
  expect_true(isTRUE(fit2$timing$cache_hit))
  expect_equal(fit2$waic$WAIC, 2.34)

  bad_seed_bundle <- bundle_stub
  bad_seed_bundle$mcmc$nchains <- 2L
  bad_seed_bundle$mcmc$seed <- 1:3
  expect_error(run_mcmc_bundle_manual(bad_seed_bundle, show_progress = FALSE, quiet = TRUE), "length 1 or length nchains")
  expect_error(run_mcmc_bundle_manual(bundle_stub, show_progress = FALSE, quiet = TRUE, z_update_every = 0L), ">= 1")
})

test_that("coverage-heavy prediction internals cover chunking unconditional summaries and gpd guards", {
  predict_impl <- getFromNamespace(".predict_mixgpd", "CausalMixGPD")

  posterior_summary_mock <- function(x, probs = c(0.025, 0.5, 0.975), interval = "credible") {
    dims <- dim(x)
    if (is.null(dims)) {
      est <- mean(x, na.rm = TRUE)
      low <- min(x, na.rm = TRUE)
      up <- max(x, na.rm = TRUE)
    } else if (length(dims) == 2L) {
      est <- rowMeans(x, na.rm = TRUE)
      low <- apply(x, 1, min, na.rm = TRUE)
      up <- apply(x, 1, max, na.rm = TRUE)
    } else {
      est <- apply(x, c(1, 2), mean, na.rm = TRUE)
      low <- apply(x, c(1, 2), min, na.rm = TRUE)
      up <- apply(x, c(1, 2), max, na.rm = TRUE)
    }
    if (is.null(interval)) {
      low[] <- NA_real_
      up[] <- NA_real_
    }
    list(estimate = est, lower = low, upper = up, q = probs)
  }

  dispatch_stub <- list(
    bulk_params = c("mean", "sd"),
    d = function(x, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL, log = 0L) {
      dens <- stats::dnorm(x, mean = mean[1], sd = max(sd[1], 1e-6))
      if (log == 1L) log(pmax(dens, 1e-300)) else pmax(dens, 1e-300)
    },
    p = function(q, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL,
                 lower.tail = 1L, log.p = 0L) {
      cdf <- stats::pnorm(q, mean = mean[1], sd = max(sd[1], 1e-6))
      if (lower.tail == 0L) cdf <- 1 - cdf
      if (log.p == 1L) log(pmax(cdf, 1e-300)) else cdf
    },
    q = function(p, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL) {
      stats::qnorm(p, mean = mean[1], sd = max(sd[1], 1e-6))
    },
    r = function(n, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL) {
      rep(mean[1], n)
    }
  )

  fake_fit <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", kernel = "normal", GPD = FALSE, has_X = TRUE),
        dispatch = list(backend = "sb", GPD = FALSE, link_params = list())
      ),
      data = list(
        X = matrix(c(0, 1, 1, 0), nrow = 2, byrow = TRUE, dimnames = list(NULL, c("x1", "x2"))),
        y = c(0.25, 1.25)
      )
    ),
    class = "mixgpd_fit"
  )
  fake_uncond <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", kernel = "normal", GPD = FALSE, has_X = FALSE),
        dispatch = list(backend = "sb", GPD = FALSE, link_params = list())
      ),
      data = list(y = c(0.25, 1.25))
    ),
    class = "mixgpd_fit"
  )
  fake_gpd <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", kernel = "normal", GPD = TRUE, has_X = FALSE),
        dispatch = list(
          backend = "sb",
          GPD = TRUE,
          link_params = list(),
          gpd = list(
            threshold = list(mode = "constant"),
            tail_scale = list(mode = "constant"),
            tail_shape = list(mode = "dist")
          )
        )
      ),
      data = list(y = c(0.25, 1.25))
    ),
    class = "mixgpd_fit"
  )
  fake_spliced <- structure(
    list(
      spec = list(
        meta = list(backend = "spliced", kernel = "normal", GPD = TRUE, has_X = TRUE),
        dispatch = list(
          backend = "spliced",
          GPD = TRUE,
          link_params = list(),
          gpd = list(
            threshold = list(mode = "link", link = "identity"),
            tail_scale = list(mode = "dist"),
            tail_shape = list(mode = "dist")
          )
        )
      ),
      data = list(
        X = fake_fit$data$X,
        y = fake_fit$data$y
      )
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .validate_fit = function(object) TRUE,
    .cmgpd_progress_start = function(...) list(),
    .cmgpd_progress_step = function(...) invisible(NULL),
    .cmgpd_progress_done = function(...) invisible(NULL),
    .extract_draws_matrix = function(object) {
      if (isTRUE(object$spec$meta$GPD)) {
        matrix(
          c(0.4, 0.5, 0.9,
            1.1, 0.6, 1.0,
            0.3, 0.7, 1.1),
          ncol = 3,
          byrow = TRUE,
          dimnames = list(NULL, c("tail_shape", "threshold", "tail_scale"))
        )
      } else {
        matrix(1:6, ncol = 2, dimnames = list(NULL, c("draw1", "draw2")))
      }
    },
    .extract_weights = function(draw_mat, backend) {
      matrix(c(0.7, 0.3, 0.6, 0.4, 0.5, 0.5), nrow = nrow(draw_mat), byrow = TRUE)
    },
    .extract_bulk_params = function(draw_mat, bulk_params) {
      list(
        mean = matrix(c(0.0, 0.8, 0.2, 1.0, 0.4, 1.2), nrow = nrow(draw_mat), byrow = TRUE),
        sd = matrix(c(1.0, 1.2, 1.1, 1.3, 1.2, 1.4), nrow = nrow(draw_mat), byrow = TRUE)
      )
    },
    .get_dispatch = function(...) dispatch_stub,
    get_kernel_registry = function() list(normal = list(bulk_support = list(mean = "", sd = "positive_sd"))),
    .posterior_summarize = posterior_summary_mock,
    .compute_interval = function(x, level, type) c(lower = min(x, na.rm = TRUE), upper = max(x, na.rm = TRUE)),
    .package = "CausalMixGPD"
  )

  pred_density <- predict_impl(
    fake_fit,
    x = fake_fit$data$X,
    y = c(0.2, 0.9),
    id = 1:2,
    type = "density",
    interval = NULL,
    chunk_size = 1L,
    show_progress = FALSE
  )
  pred_survival <- predict_impl(fake_fit, newdata =fake_fit$data$X, y = c(0.2, 0.9), type = "survival", show_progress = FALSE)
  pred_quant <- predict_impl(fake_fit, newdata =fake_fit$data$X, type = "quantile", index = c(0.25, 0.75), store_draws = TRUE, show_progress = FALSE)
  pred_sample <- predict_impl(fake_fit, newdata =fake_fit$data$X, type = "sample", nsim = 3L, show_progress = FALSE)
  pred_fit <- predict_impl(fake_fit, newdata =fake_fit$data$X, type = "fit", show_progress = FALSE)
  pred_mean <- predict_impl(fake_fit, newdata =fake_fit$data$X, type = "mean", nsim_mean = 12L, show_progress = FALSE)
  pred_rmean <- predict_impl(fake_fit, newdata =fake_fit$data$X, type = "rmean", cutoff = 1.1, nsim_mean = 12L, show_progress = FALSE)

  pred_quant_u <- predict_impl(fake_uncond, type = "quantile", index = c(0.25, 0.75), show_progress = FALSE)
  pred_sample_u <- predict_impl(fake_uncond, type = "sample", nsim = 4L, show_progress = FALSE)
  pred_mean_u <- predict_impl(fake_uncond, type = "mean", nsim_mean = 12L, show_progress = FALSE)
  pred_rmean_u <- predict_impl(fake_uncond, type = "rmean", cutoff = 1.1, nsim_mean = 12L, show_progress = FALSE)

  expect_s3_class(pred_density, "mixgpd_predict")
  expect_s3_class(pred_survival, "mixgpd_predict")
  expect_s3_class(pred_quant, "mixgpd_predict")
  expect_s3_class(pred_sample, "mixgpd_predict")
  expect_s3_class(pred_fit, "mixgpd_predict")
  expect_s3_class(pred_mean, "mixgpd_predict")
  expect_s3_class(pred_rmean, "mixgpd_predict")
  expect_s3_class(pred_quant_u, "mixgpd_predict")
  expect_s3_class(pred_sample_u, "mixgpd_predict")
  expect_s3_class(pred_mean_u, "mixgpd_predict")
  expect_s3_class(pred_rmean_u, "mixgpd_predict")
  expect_equal(pred_density$diagnostics$n_chunks, 2L)
  expect_true(all(c("id", "density") %in% names(pred_density$fit)))
  expect_true(all(c("id", "survival") %in% names(pred_survival$fit)))
  expect_true(all(c("id", "index", "estimate") %in% names(pred_quant$fit)))
  expect_s3_class(pred_sample$fit_df, "data.frame")
  expect_s3_class(pred_sample_u$fit_df, "data.frame")
  expect_true(all(c("draw", "sample") %in% names(pred_sample_u$fit_df)))
  expect_true(all(c("id", "draw", "sample") %in% names(pred_sample$fit_df)))
  expect_equal(pred_mean$diagnostics$mean_method, "analytic")
  expect_equal(pred_mean$fit$estimate, c(0.52, 0.52))
  expect_null(pred_mean$diagnostics$nsim_mean)
  expect_equal(pred_mean_u$fit$estimate, 0.52)
  expect_equal(pred_rmean$diagnostics$nsim_mean, 12L)

  expect_warning(
    pred_mean_inf <- predict_impl(fake_gpd, type = "mean", nsim_mean = 12L, show_progress = FALSE),
    "Posterior mean is infinite"
  )
  expect_true(all(is.infinite(pred_mean_inf$fit$estimate)))

  expect_error(
    predict_impl(fake_spliced, newdata =fake_spliced$data$X, type = "mean", nsim_mean = 12L, show_progress = FALSE),
    "not yet fully implemented"
  )
})

test_that("coverage-heavy predict and residual methods cover wrapper branches with mocked internals", {
  predict_method <- getFromNamespace("predict.mixgpd_fit", "CausalMixGPD")
  residuals_method <- getFromNamespace("residuals.mixgpd_fit", "CausalMixGPD")
  fake_fit <- structure(
    list(
      spec = list(meta = list(backend = "sb", kernel = "normal", GPD = FALSE)),
      data = list(
        X = matrix(c(0, 1, 1, 0), nrow = 2, byrow = TRUE, dimnames = list(NULL, c("x1", "x2"))),
        y = c(0.2, 0.8)
      )
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .validate_fit = function(object) TRUE,
    .predict_mixgpd = function(object, newdata =NULL, y = NULL, ps = NULL, id = NULL, type = "mean", ...) {
      if (identical(type, "mean")) {
        return(list(fit = data.frame(id = 1:2, estimate = c(1.0, 2.0), lower = c(0.8, 1.8), upper = c(1.2, 2.2)), type = type))
      }
      if (identical(type, "median")) {
        return(list(fit = data.frame(id = 1:2, estimate = c(0.9, 1.9), lower = c(0.7, 1.7), upper = c(1.1, 2.1)), type = type))
      }
      list(fit = data.frame(id = 1:2, estimate = c(1.0, 2.0), lower = c(0.8, 1.8), upper = c(1.2, 2.2)), type = type)
    },
    .package = "CausalMixGPD"
  )

  loc <- predict_method(
    fake_fit,
    x = fake_fit$data$X,
    id = 1:2,
    type = "location",
    interval = "none"
  )
  expect_s3_class(loc, "mixgpd_predict")
  expect_true(all(c("mean", "median") %in% names(loc$fit)))

  expect_warning(predict_method(fake_fit, type = "mean", p = 0.2), "ignoring")
  expect_error(predict_method(fake_fit, newdata =fake_fit$data$X, newdata = fake_fit$data$X, type = "mean"), "Provide only one")
  expect_error(predict_method(fake_fit, type = "median", index = 0.25), "index = 0.5")
  expect_error(predict_method(fake_fit, type = "mean", level = 1.5), "between 0 and 1")
  expect_error(predict_method(fake_fit, type = "mean", ncores = 0L), ">= 1")

  testthat::local_mocked_bindings(
    fitted.mixgpd_fit = function(object, type = "mean", ...) {
      structure(data.frame(residuals = c(0.1, -0.2)), class = c("mixgpd_fitted", "data.frame"))
    },
    predict.mixgpd_fit = function(object, newdata =NULL, y = NULL, type = c("density", "survival", "quantile", "sample", "mean", "rmean", "median", "location", "fit"), ...) {
      list(fit = data.frame(survival = c(0.8, 0.4)))
    },
    .get_dispatch = function(...) {
      list(
        p = function(q, w, mean, sd, lower.tail = 1L, log.p = 0L) stats::pnorm(q, mean = mean[1], sd = sd[1]),
        bulk_params = c("mean", "sd")
      )
    },
    .extract_draws_matrix = function(object) matrix(1:6, ncol = 2, dimnames = list(NULL, c("draw1", "draw2"))),
    .extract_weights = function(draw_mat, backend) matrix(c(0.7, 0.3, 0.6, 0.4, 0.5, 0.5), nrow = nrow(draw_mat), byrow = TRUE),
    .extract_bulk_params = function(draw_mat, bulk_params) {
      list(
        mean = matrix(c(0.0, 0.8, 0.2, 1.0, 0.4, 1.2), nrow = nrow(draw_mat), byrow = TRUE),
        sd = matrix(c(1.0, 1.2, 1.1, 1.3, 1.2, 1.4), nrow = nrow(draw_mat), byrow = TRUE)
      )
    },
    get_kernel_registry = function() list(normal = list(bulk_support = list(mean = "", sd = "positive_sd"))),
    .package = "CausalMixGPD"
  )

  expect_equal(residuals_method(fake_fit, type = "raw", fitted_type = "median"), c(0.1, -0.2))
  expect_equal(as.numeric(residuals_method(fake_fit, type = "pit", pit = "plugin")), c(0.2, 0.6))

  pit_mean <- residuals_method(fake_fit, type = "pit", pit = "bayes_mean", pit_seed = 7L)
  pit_draw <- residuals_method(fake_fit, type = "pit", pit = "bayes_draw", pit_seed = 7L)
  expect_length(pit_mean, 2L)
  expect_length(pit_draw, 2L)
  expect_true(!is.null(attr(pit_mean, "pit_diagnostics")))
  expect_true(!is.null(attr(pit_draw, "pit_diagnostics")))
})

test_that("coverage-heavy kernel quantiles and lowercase wrappers cover boundary branches", {
  w <- c(0.6, 0.4)
  norm_mean <- c(-0.5, 1.0)
  norm_sd <- c(0.9, 1.2)

  expect_equal(qNormMix(c(0, 1), w = w, mean = norm_mean, sd = norm_sd), c(-Inf, Inf))
  expect_true(all(is.finite(qNormMix(log(c(0.2, 0.8)), w = w, mean = norm_mean, sd = norm_sd, log.p = TRUE))))
  expect_true(all(is.finite(qNormMixGpd(c(0.2, 0.95), w = w, mean = norm_mean, sd = norm_sd,
                                        threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1))))
  expect_true(all(is.finite(qNormGpd(log(c(0.2, 0.8)), mean = 0.3, sd = 1.1,
                                     threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1,
                                     log.p = TRUE))))
  expect_identical(dnormmix(numeric(0), w = w, mean = norm_mean, sd = norm_sd), numeric(0))
  expect_identical(dnormmixgpd(numeric(0), w = w, mean = norm_mean, sd = norm_sd,
                               threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1), numeric(0))
  expect_identical(dnormgpd(numeric(0), mean = 0.3, sd = 1.1,
                            threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1), numeric(0))
  expect_identical(rnormmix(0L, w = w, mean = norm_mean, sd = norm_sd), numeric(0))
  expect_identical(rnormmixgpd(0L, w = w, mean = norm_mean, sd = norm_sd,
                               threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1), numeric(0))
  expect_identical(rnormgpd(0L, mean = 0.3, sd = 1.1,
                            threshold = 1.5, tail_scale = 0.8, tail_shape = 0.1), numeric(0))
  expect_error(rnormmix(NA_integer_, w = w, mean = norm_mean, sd = norm_sd), "single integer")

  expect_equal(qGpd(c(0, 1), threshold = 1, scale = 0.8, shape = -0.2), c(1, 5))
  expect_true(is.finite(qGpd(log(0.25), threshold = 1, scale = 0.8, shape = 0.2, log.p = TRUE)))
  expect_equal(qInvGauss(c(0, 1), mean = 1.5, shape = 5), c(0, Inf))
  expect_true(is.finite(qInvGauss(log(0.25), mean = 1.5, shape = 5, log.p = TRUE)))
  expect_equal(qAmoroso(c(0, 1), loc = c(0, 0), scale = c(1, -1), shape1 = c(2, 2), shape2 = c(1.2, 1.2)), c(0, -Inf))
  expect_equal(qCauchy(c(0, 1), location = 0, scale = 1), c(-Inf, Inf))
  expect_identical(dgpd(numeric(0), threshold = 1, scale = 0.8, shape = 0.2), numeric(0))
  expect_identical(dinvgauss(numeric(0), mean = 1.5, shape = 5), numeric(0))
  expect_identical(damoroso(numeric(0), loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), numeric(0))
  expect_identical(dcauchy_vec(numeric(0), location = 0, scale = 1), numeric(0))
  expect_identical(rgpd(0L, threshold = 1, scale = 0.8, shape = 0.2), numeric(0))
  expect_identical(rinvgauss(0L, mean = 1.5, shape = 5), numeric(0))
  expect_identical(ramoroso(0L, loc = 0, scale = 1.4, shape1 = 2, shape2 = 1.2), numeric(0))
  expect_identical(rcauchy_vec(0L, location = 0, scale = 1), numeric(0))
  expect_error(rgpd(NA_integer_, threshold = 1, scale = 0.8, shape = 0.2), "single integer")

  amoroso_loc <- c(0.0, 0.2)
  amoroso_scale <- c(1.0, 1.5)
  amoroso_shape1 <- c(2.0, 2.5)
  amoroso_shape2 <- c(1.1, 1.3)
  amoroso_mix <- qAmorosoMix(c(0, 1), w = w, loc = amoroso_loc, scale = amoroso_scale,
                             shape1 = amoroso_shape1, shape2 = amoroso_shape2)
  amoroso_mix_gpd <- qAmorosoMixGpd(c(0.2, 0.95), w = w, loc = amoroso_loc, scale = amoroso_scale,
                                    shape1 = amoroso_shape1, shape2 = amoroso_shape2,
                                    threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1)
  amoroso_gpd <- qAmorosoGpd(c(0.2, 0.95), loc = 0.1, scale = 1.3, shape1 = 2.2, shape2 = 1.1,
                             threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1)

  expect_equal(amoroso_mix, c(-Inf, Inf))
  expect_true(all(is.finite(amoroso_mix_gpd)))
  expect_true(all(is.finite(amoroso_gpd)))
  expect_identical(damorosomix(numeric(0), w = w, loc = amoroso_loc, scale = amoroso_scale,
                               shape1 = amoroso_shape1, shape2 = amoroso_shape2), numeric(0))
  expect_identical(damorosomixgpd(numeric(0), w = w, loc = amoroso_loc, scale = amoroso_scale,
                                  shape1 = amoroso_shape1, shape2 = amoroso_shape2,
                                  threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1), numeric(0))
  expect_identical(damorosogpd(numeric(0), loc = 0.1, scale = 1.3, shape1 = 2.2, shape2 = 1.1,
                               threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1), numeric(0))
  expect_identical(ramorosomix(0L, w = w, loc = amoroso_loc, scale = amoroso_scale,
                               shape1 = amoroso_shape1, shape2 = amoroso_shape2), numeric(0))
  expect_identical(ramorosomixgpd(0L, w = w, loc = amoroso_loc, scale = amoroso_scale,
                                  shape1 = amoroso_shape1, shape2 = amoroso_shape2,
                                  threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1), numeric(0))
  expect_identical(ramorosogpd(0L, loc = 0.1, scale = 1.3, shape1 = 2.2, shape2 = 1.1,
                               threshold = 2.5, tail_scale = 0.9, tail_shape = 0.1), numeric(0))
  expect_error(ramorosomix(NA_integer_, w = w, loc = amoroso_loc, scale = amoroso_scale,
                           shape1 = amoroso_shape1, shape2 = amoroso_shape2), "single integer")
})

test_that("coverage-heavy build-run advanced helpers cover gating and standard gpd branches", {
  skip_if_not_test_level("ci")

  y <- abs(stats::rnorm(10L)) + 0.1
  X <- cbind(x1 = stats::rnorm(10L), x2 = stats::runif(10L))
  ps <- rep(0.5, length(y))
  dat <- data.frame(y = y, x1 = X[, 1], x2 = X[, 2])

  expect_error(build_code_from_spec(list(meta = list(backend = "bad"))), "Unknown backend")
  expect_error(
    build_nimble_bundle(y = y, backend = "sb", kernel = "normal", GPD = FALSE, components = c(2L, 3L)),
    "single integer"
  )

  auto_bundle <- build_nimble_bundle(
    y = y,
    X = X,
    backend = "spliced",
    kernel = "normal",
    GPD = TRUE,
    param_specs = list(
      bulk = list(mean = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 1))),
      gpd = list(
        threshold = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_scale = list(mode = "fixed", value = 1.2),
        tail_shape = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 0.2))
      )
    ),
    monitor = "full"
  )
  expect_equal(auto_bundle$spec$meta$components, length(y))
  expect_identical(auto_bundle$spec$dispatch$backend, "spliced")
  expect_true(isTRUE(auto_bundle$monitor_policy$monitor_latent))

  spec_std <- compile_model_spec(
    y = y,
    X = X,
    ps = ps,
    backend = "sb",
    kernel = "normal",
    GPD = TRUE,
    components = 4L,
    param_specs = list(
      bulk = list(
        mean = list(
          mode = "link",
          link = "identity",
          beta_prior = list(dist = "normal", args = list(mean = 0, sd = 1.5))
        ),
        sd = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
      ),
      gpd = list(
        threshold = list(
          mode = "link",
          link = "identity",
          beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.2)),
          link_dist = list(dist = "lognormal")
        ),
        sdlog_u = list(mode = "dist", dist = "invgamma", args = list(shape = 2, scale = 1)),
        tail_scale = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_shape = list(
          mode = "link",
          link = "identity",
          beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.3))
        )
      )
    )
  )

  const_std <- build_constants_from_spec(spec_std)
  dims_std <- build_dimensions_from_spec(spec_std)
  mons_std <- build_monitors_from_spec(spec_std, monitor_v = TRUE, monitor_latent = TRUE)
  inits_std <- build_inits_from_spec(spec_std, y = y)
  prior_std <- build_prior_table_from_spec(spec_std)

  expect_true(all(c("N", "P", "components") %in% names(const_std)))
  expect_true(all(c("beta_mean", "beta_ps_mean", "threshold", "beta_threshold", "beta_tail_shape") %in% names(dims_std)))
  expect_true(any(grepl("^beta_tail_shape", mons_std)))
  expect_true(all(c("beta_mean", "beta_ps_mean", "beta_threshold", "threshold", "sdlog_u", "tail_scale", "beta_tail_shape") %in% names(inits_std)))
  expect_true(any(prior_std$parameter == "threshold" & prior_std$mode == "link+dist"))
  expect_true(any(prior_std$parameter == "tail_shape" & prior_std$mode == "link"))

  cluster_bundle <- build_cluster_bundle(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    GPD = TRUE,
    type = "weights",
    components = 4L,
    link = list(
      bulk = list(mean = "identity"),
      gpd = list(
        threshold = list(link = "identity"),
        tail_shape = list(link = "identity")
      )
    ),
    priors = list(
      bulk = list(sd = list(dist = "gamma", args = list(shape = 2, rate = 1))),
      gpd = list(
        tail_scale = list(dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_shape = list(dist = "normal", args = list(mean = 0, sd = 0.3))
      ),
      concentration = list(dist = "gamma", args = list(shape = 3, rate = 2))
    ),
    mcmc = mcmc_fast(seed = 17L)
  )
  gating_spec <- cluster_bundle$spec
  dims_gating <- build_dimensions_from_spec(gating_spec)
  mons_gating <- build_monitors_from_spec(gating_spec, monitor_latent = TRUE)
  inits_gating <- build_inits_from_spec(gating_spec, y = y)
  const_gating <- build_constants_from_spec(gating_spec)
  prior_gating <- build_prior_table_from_spec(gating_spec)

  expect_true(all(c("eta", "B", "logit_ij", "w_x") %in% names(dims_gating)))
  expect_true(any(grepl("^eta", mons_gating)))
  expect_true(any(grepl("^B", mons_gating)))
  expect_true(all(c("eta", "B", "z") %in% names(inits_gating)))
  expect_true(any(prior_gating$parameter == "threshold"))
  expect_true(all(c("N", "P", "components") %in% names(const_gating)))
})

test_that("coverage-heavy methods cover params and ESS summary branches with mocked draws", {
  skip_if_not_installed("coda")

  draw_mat_params <- matrix(
    c(
      1.0, 0.7, 0.3, 0.1, 0.2, -0.1, 0.0, 0.05, -0.05, 0.2, 0.3, 0.4, 1.1, 0.1, 0.2,
      1.2, 0.6, 0.4, 0.2, 0.1, -0.2, 0.1, 0.00, -0.02, 0.1, 0.2, 0.5, 1.3, 0.2, 0.1,
      0.8, 0.5, 0.5, 0.3, 0.0, -0.3, 0.2, -0.01, 0.01, 0.3, 0.1, 0.6, 1.2, 0.15, 0.05
    ),
    nrow = 3L,
    byrow = TRUE,
    dimnames = list(
      NULL,
      c(
        "alpha", "w[1]", "w[2]",
        "beta_mean[1,1]", "beta_mean[1,2]", "beta_mean[2,1]", "beta_mean[2,2]",
        "beta_ps_mean[1]", "beta_ps_mean[2]",
        "beta_threshold[1]", "beta_threshold[2]", "sdlog_u",
        "tail_scale",
        "beta_tail_shape[1]", "beta_tail_shape[2]"
      )
    )
  )

  fit_params <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", kernel = "normal", GPD = TRUE, has_X = TRUE),
        plan = list(
          bulk = list(
            mean = list(mode = "link"),
            sd = list(mode = "fixed", value = 1)
          ),
          ps = list(prior = list(dist = "normal", args = list(mean = 0, sd = 1))),
          gpd = list(
            threshold = list(mode = "link", link_dist = list(dist = "lognormal")),
            tail_scale = list(mode = "dist"),
            tail_shape = list(mode = "link")
          )
        )
      ),
      data = list(X = structure(matrix(c(1, 0, 0, 1), nrow = 2L), dimnames = list(NULL, c("x1", "x2"))))
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .extract_draws_matrix = function(...) draw_mat_params,
    .package = "CausalMixGPD"
  )

  param_obj <- params(fit_params)
  expect_s3_class(param_obj, "mixgpd_params")
  expect_true(all(c("alpha", "w", "beta_mean", "beta_threshold", "sdlog_u", "tail_scale", "beta_tail_shape") %in% names(param_obj)))
  expect_false("beta_ps_mean" %in% names(param_obj))
  expect_equal(dim(param_obj$beta_mean), c(2L, 3L))
  expect_identical(colnames(param_obj$beta_mean), c("PropScore", "x1", "x2"))
  expect_equal(rownames(param_obj$beta_mean), c("comp1", "comp2"))
  expect_equal(dim(param_obj$beta_threshold), c(1L, 2L))
  expect_identical(colnames(param_obj$beta_threshold), c("x1", "x2"))
  expect_equal(dim(param_obj$beta_tail_shape), c(1L, 2L))
  expect_identical(colnames(param_obj$beta_tail_shape), c("x1", "x2"))

  chain1 <- coda::mcmc(
    matrix(
      c(
        1.0, 0.2, 0.5, 1.2, 0.1,
        1.1, 0.3, 0.6, 1.1, 0.2,
        0.9, 0.4, 0.7, 1.0, 0.3,
        1.2, 0.5, 0.8, 1.3, 0.4
      ),
      nrow = 4L,
      byrow = TRUE,
      dimnames = list(NULL, c("alpha", "tail_shape", "mu[1]", "sigma[1]", "beta[1]"))
    )
  )
  chain2 <- coda::mcmc(
    matrix(
      c(
        0.8, 0.1, 0.4, 1.0, 0.0,
        0.9, 0.2, 0.5, 1.1, 0.1,
        1.0, 0.3, 0.6, 1.2, 0.2,
        1.1, 0.4, 0.7, 1.3, 0.3
      ),
      nrow = 4L,
      byrow = TRUE,
      dimnames = list(NULL, c("alpha", "tail_shape", "mu[1]", "sigma[1]", "beta[1]"))
    )
  )
  fit_ess <- structure(
    list(
      mcmc = list(samples = coda::mcmc.list(chain1, chain2)),
      timing = list(mcmc = 6)
    ),
    class = "mixgpd_fit"
  )

  ess_obj <- ess_summary(fit_ess, per_chain = TRUE)
  expect_s3_class(ess_obj, "mixgpd_ess_summary")
  expect_true(nrow(ess_obj$table) >= 10L)
  expect_true(nrow(summary(ess_obj)) >= 1L)
  expect_output(print(ess_obj, max_rows = 2L), "ESS summary")
  expect_error(ess_summary(fit_ess, params = "missing_param", robust = FALSE), "No parameters matched")

  fit_ess_no_time <- fit_ess
  fit_ess_no_time$timing <- list()
  expect_warning(no_time_res <- try(ess_summary(fit_ess_no_time, params = "alpha", per_chain = FALSE), silent = TRUE), "No wall-time available")
  expect_true(inherits(no_time_res, "try-error"))

  causal_ess <- structure(list(outcome_fit = list(con = fit_ess, trt = fit_ess)), class = "causalmixgpd_causal_fit")
  ess_causal <- ess_summary(causal_ess, params = c("alpha", "mu\\[1\\]"), wall_time = 8, per_chain = FALSE)
  expect_s3_class(ess_causal, "mixgpd_ess_summary")
  expect_true(all(c("control", "treated") %in% ess_causal$table$arm))
})

test_that("coverage-heavy methods cover predict wrappers fitted and residual PIT branches", {
  fit_stub <- structure(list(), class = "mixgpd_fit")
  x_new <- matrix(c(1, 2, 3, 4), ncol = 2L)
  predict_calls <- character(0)

  testthat::local_mocked_bindings(
    .validate_fit = function(...) invisible(TRUE),
    .predict_mixgpd = function(object, newdata =NULL, type = c("mean", "median", "quantile"), ...) {
      type <- match.arg(type)
      predict_calls <<- c(predict_calls, type)
      if (type == "mean") {
        return(list(fit = data.frame(estimate = c(1.0, 2.0), lower = c(0.5, 1.5), upper = c(1.5, 2.5))))
      }
      if (type == "median") {
        return(list(fit = data.frame(estimate = c(0.8, 1.8), lower = c(0.3, 1.3), upper = c(1.3, 2.3))))
      }
      list(fit = data.frame(id = c(2L, 1L), estimate = c(1.5, 0.5), lower = c(1.0, 0.1), upper = c(2.0, 0.9)))
    },
    .package = "CausalMixGPD"
  )

  loc_obj <- predict(fit_stub, newdata = x_new, type = "location", workers = 2L, parallel = TRUE)
  expect_true(is.list(loc_obj) || is.data.frame(loc_obj))
  expect_equal(predict_calls, c("mean", "median"))
  expect_true(all(c("mean", "median") %in% names(loc_obj$fit)))
  qpred <- predict(fit_stub, newdata = x_new, type = "quantile", p = c(0.2, 0.8))
  qfit <- if (is.list(qpred) && !is.data.frame(qpred)) qpred$fit else qpred
  expect_true(is.data.frame(qfit))
  expect_error(predict(fit_stub, newdata = x_new, type = "quantile", p = 0.2, index = 0.3), "Provide only one")
  expect_error(predict(fit_stub, newdata = x_new, type = "median", index = 0.2), "index = 0.5")
  expect_error(predict(fit_stub, newdata = x_new, type = "mean", cred.level = 0.9), "cred.level")
  expect_warning(predict(fit_stub, newdata = x_new, type = "mean", p = 0.25), "only used for type = 'quantile'")

  plugin_with_id <- FALSE
  fit_cond <- structure(
    list(
      data = list(
        y = c(2.0, 3.0),
        X = structure(matrix(c(1, 0, 0, 1), nrow = 2L), dimnames = list(NULL, c("x1", "x2")))
      ),
      spec = list(
        meta = list(backend = "sb", kernel = "normal", GPD = TRUE, has_X = TRUE),
        dispatch = list(
          backend = "sb",
          GPD = TRUE,
          gpd = list(
            threshold = list(mode = "dist"),
            tail_scale = list(mode = "dist")
          )
        )
      )
    ),
    class = "mixgpd_fit"
  )

  draw_mat_pit <- matrix(
    c(
      0.7, 0.3, 1.0, 2.0, 1.0, 1.2, 0.1, 1.0, 1.0,
      0.6, 0.4, 1.2, 2.2, 1.0, 1.3, 0.2, 1.1, -1.0,
      0.5, 0.5, 1.4, 2.4, 1.0, 1.4, 0.3, 1.2, 1.1
    ),
    nrow = 3L,
    byrow = TRUE,
    dimnames = list(NULL, c("w[1]", "w[2]", "mean[1]", "mean[2]", "sd[1]", "sd[2]", "tail_shape", "threshold", "tail_scale"))
  )

  testthat::local_mocked_bindings(
    predict.mixgpd_fit = function(object, newdata =NULL, y = NULL, type = c("mean", "median", "quantile", "survival"), ...) {
      type <- match.arg(type)
      if (type == "mean") {
        return(list(fit = data.frame(id = c(2L, 1L), estimate = c(3.0, 2.0), lower = c(2.5, 1.5), upper = c(3.5, 2.5))))
      }
      if (type == "median") {
        return(list(fit = data.frame(id = c(2L, 1L), estimate = c(2.8, 1.8), lower = c(2.3, 1.3), upper = c(3.3, 2.3))))
      }
      if (type == "quantile") {
        return(list(fit = data.frame(id = c(2L, 1L), estimate = c(2.6, 1.6), lower = c(2.1, 1.1), upper = c(3.1, 2.1))))
      }
      if (isTRUE(plugin_with_id)) {
        return(list(fit = data.frame(id = c(1L, 1L, 2L, 2L), survival = c(0.8, 0.7, 0.6, 0.5))))
      }
      list(fit = data.frame(survival = c(0.8, 0.6)))
    },
    .get_dispatch = function(...) {
      list(
        p = function(q, w, mean, sd, threshold = NULL, tail_scale = NULL, tail_shape = NULL, lower.tail = 1L, log.p = 0L) {
          cdf <- stats::pnorm(q, mean = mean[1], sd = max(sd[1], 1e-6))
          if (lower.tail == 0L) cdf <- 1 - cdf
          if (log.p == 1L) log(pmax(cdf, 1e-12)) else cdf
        },
        bulk_params = c("mean", "sd")
      )
    },
    .extract_draws_matrix = function(...) draw_mat_pit,
    .extract_weights = function(...) matrix(c(0.7, 0.3, 0.6, 0.4, 0.5, 0.5), nrow = 3L, byrow = TRUE),
    .extract_bulk_params = function(...) {
      list(
        mean = matrix(c(1.0, 2.0, 1.2, 2.2, 1.4, 2.4), nrow = 3L, byrow = TRUE),
        sd = matrix(c(1.0, 1.2, 1.0, 1.3, 1.0, 1.4), nrow = 3L, byrow = TRUE)
      )
    },
    get_kernel_registry = function() list(normal = list(bulk_support = list(mean = "", sd = "positive_sd"))),
    .package = "CausalMixGPD"
  )

  fitted_loc <- fitted(fit_cond, type = "location", interval = "credible", seed = 1L)
  expect_s3_class(fitted_loc, "mixgpd_fitted")
  expect_true(all(c("mean", "median", "mean_lower", "median_upper") %in% names(fitted_loc)))
  expect_s3_class(fitted(fit_cond, type = "quantile", p = 0.4, interval = "credible", seed = 1L), "mixgpd_fitted")
  expect_equal(residuals(fit_cond, type = "raw", fitted_type = "median"), fit_cond$data$y - c(1.8, 2.8))

  plugin_plain <- residuals(fit_cond, type = "pit", pit = "plugin")
  expect_equal(as.numeric(plugin_plain), c(0.2, 0.4))
  plugin_with_id <- TRUE
  plugin_diag <- residuals(fit_cond, type = "pit", pit = "plugin")
  expect_equal(as.numeric(plugin_diag), c(0.2, 0.5))

  bayes_mean <- residuals(fit_cond, type = "pit", pit = "bayes_mean")
  expect_equal(attr(bayes_mean, "pit_type"), "bayes_mean")
  expect_equal(attr(bayes_mean, "pit_diagnostics")$n_draws_dropped, 1L)
  bayes_draw <- residuals(fit_cond, type = "pit", pit = "bayes_draw", pit_seed = 1L)
  expect_equal(attr(bayes_draw, "pit_type"), "bayes_draw")
  expect_equal(length(bayes_draw), 2L)

  fit_uncond <- fit_cond
  fit_uncond$data$X <- NULL
  fit_uncond$spec$meta$has_X <- FALSE
  expect_error(fitted(fit_uncond), "not supported for unconditional models")
  expect_error(residuals(fit_uncond), "not supported for unconditional models")
})

test_that("coverage-heavy cluster helper internals cover density scoring and design branches", {
  skip_if_not_installed("coda")

  ns_fun <- function(name) getFromNamespace(name, "CausalMixGPD")
  cluster_default_link <- ns_fun(".cluster_default_link")
  cluster_link_bulk_specs <- ns_fun(".cluster_link_bulk_specs")
  cluster_default_beta_prior <- ns_fun(".cluster_default_beta_prior")
  cluster_split_overrides <- ns_fun(".cluster_split_overrides")
  cluster_normalize_link_entry <- ns_fun(".cluster_normalize_link_entry")
  cluster_normalize_prior_entry <- ns_fun(".cluster_normalize_prior_entry")
  cluster_parse_formula <- ns_fun(".cluster_parse_formula")
  cluster_check_new_factor_levels <- ns_fun(".cluster_check_new_factor_levels")
  cluster_build_design <- ns_fun(".cluster_build_design")
  cluster_link_apply <- ns_fun(".cluster_link_apply")
  cluster_extract_beta_component <- ns_fun(".cluster_extract_beta_component")
  cluster_extract_beta_global <- ns_fun(".cluster_extract_beta_global")
  cluster_extract_beta_auto <- ns_fun(".cluster_extract_beta_auto")
  cluster_softmax <- ns_fun(".cluster_softmax")
  cluster_extract_gating_draw <- ns_fun(".cluster_extract_gating_draw")
  cluster_gating_weights <- ns_fun(".cluster_gating_weights")
  cluster_resolve_density_fun <- ns_fun(".cluster_resolve_density_fun")
  cluster_component_density <- ns_fun(".cluster_component_density")
  cluster_compute_scores <- ns_fun(".cluster_compute_scores")

  expect_equal(cluster_default_beta_prior("threshold")$args$sd, 0.2)
  expect_equal(cluster_default_link(get_kernel_registry()[["gamma"]], "scale"), "exp")
  expect_true(is.list(cluster_link_bulk_specs("normal", beta_sd = 1.5)))

  split1 <- cluster_split_overrides(list(mean = list(mode = "link"), tail_shape = list(mode = "dist")), c("mean"), c("tail_shape"))
  expect_true("mean" %in% names(split1$bulk))
  expect_true("tail_shape" %in% names(split1$gpd))
  expect_equal(cluster_normalize_link_entry("exp")$link, "exp")
  expect_error(cluster_normalize_link_entry(list(mode = "dist")), "mode='link'")

  expect_equal(cluster_normalize_prior_entry(1.2, current_mode = "dist")$mode, "fixed")
  expect_equal(cluster_normalize_prior_entry("gamma", current_mode = "dist")$dist, "gamma")
  expect_equal(cluster_normalize_prior_entry(list(beta_prior = list(dist = "normal", args = list(mean = 0, sd = 1))), current_mode = "link")$mode, "link")
  expect_error(cluster_normalize_prior_entry(list(foo = 1), current_mode = "dist"), "Could not interpret")

  dat <- data.frame(
    y = c(1, 2, 3, 4),
    x1 = c(0.1, 0.2, 0.3, 0.4),
    grp = factor(c("a", "b", "a", "b")),
    id = 1:4
  )
  parsed <- cluster_parse_formula(y ~ x1 + grp + id, data = dat)
  expect_equal(parsed$response, "y")
  expect_true(isTRUE(parsed$has_X))
  expect_true(length(parsed$X_cols) >= 2L)

  meta <- list(
    terms = stats::terms(y ~ grp + x1, data = dat),
    xlevels = list(grp = c("a", "b")),
    contrasts = attr(stats::model.matrix(y ~ grp + x1, data = dat), "contrasts"),
    X_cols = c("grpb", "x1"),
    response = "y"
  )
  expect_silent(cluster_check_new_factor_levels(meta, data.frame(y = 1, grp = factor("a", levels = c("a", "b")), x1 = 0.5)))
  expect_error(cluster_check_new_factor_levels(meta, data.frame(y = 1, grp = factor("c", levels = c("a", "b", "c")), x1 = 0.5)), "unseen factor levels")
  design_ok <- cluster_build_design(meta, data.frame(y = 5, grp = factor("b", levels = c("a", "b")), x1 = 0.8))
  expect_true(is.matrix(design_ok$X))
  expect_error(cluster_build_design(meta, data.frame(grp = factor("a", levels = c("a", "b")), x1 = 0.1)), "response column")
  expect_error(cluster_build_design(meta, data.frame(y = 1, grp = factor("c", levels = c("a", "b", "c")), x1 = 0.1)), "unseen factor levels")

  expect_equal(cluster_link_apply(1, "identity"), 1)
  expect_equal(cluster_link_apply(0, "exp"), 1)
  expect_equal(cluster_link_apply(2, "power", link_power = 2), 4)
  expect_error(cluster_link_apply(2, "power"), "Invalid power link exponent")

  draw_row <- c(
    "beta_mean[1,1]" = 0.3,
    "beta_mean[1,2]" = -0.2,
    "beta_mean[2,1]" = 0.1,
    "beta_mean[2,2]" = 0.4,
    "beta_threshold[1]" = 0.5,
    "beta_threshold[2]" = -0.1,
    "eta[1]" = 0.2,
    "B[1,1]" = 0.1,
    "B[1,2]" = -0.2,
    "mean[1]" = 1.5,
    "mean[2]" = 2.5,
    "sd[1]" = 1.0,
    "sd[2]" = 1.2
  )
  expect_equal(cluster_extract_beta_component(draw_row, "beta_mean", 1L, 2L), c(0.3, -0.2))
  expect_equal(cluster_extract_beta_global(draw_row, "beta_threshold", 2L), c(0.5, -0.1))
  expect_equal(cluster_extract_beta_auto(draw_row, "beta_mean", 2L, 2L), c(0.1, 0.4))
  expect_equal(cluster_softmax(numeric(0)), numeric(0))
  expect_equal(round(sum(cluster_softmax(c(1, 2, 3))), 8), 1)

  gating_draw <- cluster_extract_gating_draw(draw_row, K = 2L, P = 2L)
  expect_true(is.list(gating_draw))
  gating_w <- cluster_gating_weights(gating_draw, x_row = c(1, 2))
  expect_equal(round(sum(gating_w), 8), 1)

  cluster_bundle <- build_cluster_bundle(
    y ~ x1 + grp,
    data = transform(dat, grp = factor(grp)),
    kernel = "normal",
    GPD = FALSE,
    type = "param",
    components = 3L,
    mcmc = mcmc_fast(seed = 19L)
  )
  density_fun <- cluster_resolve_density_fun(cluster_bundle$spec)
  dens_val <- cluster_component_density(
    spec = cluster_bundle$spec,
    draw_row = draw_row,
    k = 1L,
    x_row = c(1, 0.5),
    y_val = 1.2,
    density_fun = density_fun
  )
  expect_true(is.numeric(dens_val) && dens_val >= 0)

  z_draws <- matrix(c(1, 1, 2, 2, 1, 2, 1, 2), nrow = 2L, byrow = TRUE)
  psm <- compute_psm(z_draws)
  dahl <- dahl_labels(z_draws, psm)
  scores <- cluster_compute_scores(z_draws, dahl$labels, psm)
  expect_true(is.matrix(scores))
  expect_equal(round(rowSums(scores), 8), rep(1, nrow(scores)))

  z_mcmc <- coda::mcmc.list(coda::mcmc(matrix(c(1, 1, 2, 2, 1, 2), ncol = 2L, byrow = TRUE, dimnames = list(NULL, c("z[1]", "z[2]")))))
  expect_true(is.matrix(extract_z_draws(z_mcmc, burnin = 1L, thin = 1L)))
})

test_that("coverage-heavy glue diagnostics cover missing-data and gpd failure branches", {
  check_glue_validity <- getFromNamespace("check_glue_validity", "CausalMixGPD")

  bad_fit <- structure(
    list(
      spec = list(
        meta = list(backend = "sb", GPD = TRUE, has_X = TRUE),
        dispatch = list(
          backend = "sb",
          GPD = TRUE,
          gpd = list(
            threshold = list(mode = "link", link = "power"),
            tail_scale = list(mode = "constant")
          )
        )
      ),
      data = list(X = NULL, y = NULL)
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .get_dispatch = function(...) list(
      d = function(x, ...) rep(0.1, length(x)),
      p = function(q, ...) rep(0.5, length(q)),
      bulk_params = c("mean")
    ),
    .extract_draws_matrix = function(...) {
      matrix(c(1, 0.5, 0.1, 1, 0.6, 0.2), nrow = 2L, byrow = TRUE,
             dimnames = list(NULL, c("w[1]", "tail_shape", "beta_threshold[1]")))
    },
    .extract_weights = function(...) matrix(1, nrow = 1L, ncol = 1L),
    .extract_bulk_params = function(...) list(mean = matrix(1, nrow = 1L, ncol = 1L)),
    .package = "CausalMixGPD"
  )

  expect_error(check_glue_validity(bad_fit, grid = c(0.1, 0.5)), "Training X not found")
  expect_error(
    check_glue_validity(
      structure(
        list(
          spec = list(meta = list(backend = "sb", GPD = FALSE, has_X = FALSE), dispatch = list(backend = "sb", GPD = FALSE)),
          data = list(y = NULL)
        ),
        class = "mixgpd_fit"
      )
    ),
    "Training y not found"
  )

  bad_fit$data$X <- matrix(1, nrow = 1L, ncol = 1L)
  expect_error(check_glue_validity(bad_fit, grid = c(0.1, 0.5)), "power link requires numeric link_power")

  bad_fit$spec$dispatch$gpd$threshold <- list(mode = "constant")
  expect_error(check_glue_validity(bad_fit, grid = c(0.1, 0.5)), "threshold not found")

  testthat::local_mocked_bindings(
    .extract_draws_matrix = function(...) matrix(c(1, 0.5, 1, 1, 0.6, 1.1), nrow = 2L, byrow = TRUE, dimnames = list(NULL, c("w[1]", "tail_shape", "threshold"))),
    .package = "CausalMixGPD"
  )
  expect_error(check_glue_validity(bad_fit, grid = c(0.1, 0.5)), "tail_scale not found")
})

test_that("coverage-heavy internal progress helpers cover inline step and done branches", {
  progress_colorize <- getFromNamespace(".cmgpd_progress_colorize", "CausalMixGPD")
  progress_format <- getFromNamespace(".cmgpd_progress_format", "CausalMixGPD")
  progress_start <- getFromNamespace(".cmgpd_progress_start", "CausalMixGPD")
  progress_step <- getFromNamespace(".cmgpd_progress_step", "CausalMixGPD")
  progress_done <- getFromNamespace(".cmgpd_progress_done", "CausalMixGPD")

  progress_updates <- character(0)
  progress_messages <- character(0)

  testthat::local_mocked_bindings(
    .cmgpd_progress_write = function(text) {
      progress_updates <<- c(progress_updates, text)
      invisible(text)
    },
    .cmgpd_message = function(...) {
      progress_messages <<- c(progress_messages, paste0(..., collapse = ""))
      invisible(NULL)
    },
    .package = "CausalMixGPD"
  )

  expect_equal(progress_colorize("", step_index = 1L, enabled = TRUE), "")
  expect_equal(progress_colorize("step", step_index = 1L, enabled = TRUE), "<step>")
  expect_equal(progress_format(1L, 3L, "   ", label = "mix", color = TRUE), "<[mix] Working...>")

  ctx <- progress_start(total_steps = 0L, enabled = TRUE, quiet = FALSE, label = "mix")
  ctx$total <- 2L
  ctx$enabled <- TRUE
  ctx$color_enabled <- TRUE
  ctx$live_backend <- "inline"
  ctx$rendered <- FALSE

  progress_step(ctx, "First")
  progress_done(ctx, final_label = "Done")

  expect_equal(ctx$current, ctx$total)
  expect_false(length(progress_messages))
  expect_true(any(grepl("First", progress_updates, fixed = TRUE)))
  expect_true(any(grepl("Done", progress_updates, fixed = TRUE)))
  expect_true(length(progress_updates) >= 3L)
})

test_that("coverage-heavy gamma cdf guards handle non-finite intermediates", {
  cdf_mix <- as.numeric(pGammaMix(
    q = 1,
    w = c(0.5, 0.5),
    shape = c(1, NA_real_),
    scale = c(1, 1),
    lower.tail = 1,
    log.p = 0
  ))
  expect_true(is.finite(cdf_mix))
  expect_equal(cdf_mix, 0)

  cdf_mixgpd <- as.numeric(pGammaMixGpd(
    q = 2,
    w = c(0.5, 0.5),
    shape = c(1, NA_real_),
    scale = c(1, 1),
    threshold = 1,
    tail_scale = 1,
    tail_shape = 0.1,
    lower.tail = 1,
    log.p = 0
  ))
  expect_true(is.finite(cdf_mixgpd))
  expect_equal(cdf_mixgpd, 0)

  cdf_gpd <- as.numeric(pGammaGpd(
    q = 2,
    shape = NA_real_,
    scale = 1,
    threshold = 1,
    tail_scale = 1,
    tail_shape = 0.1,
    lower.tail = 1,
    log.p = 0
  ))
  expect_true(is.finite(cdf_gpd))
  expect_equal(cdf_gpd, 0)
})

test_that("coverage-heavy causal summary printers cover knitr fallback and raw summaries", {
  knitr_mode <- TRUE
  old_kable_opt <- getOption("causalmixgpd.knitr.kable")
  options(causalmixgpd.knitr.kable = TRUE)
  on.exit(options(causalmixgpd.knitr.kable = old_kable_opt), add = TRUE)

  testthat::local_mocked_bindings(
    .is_knitr_output = function() knitr_mode,
    .kable_table = function(x, ...) {
      if (is.data.frame(x)) {
        paste(names(x), collapse = "|")
      } else {
        paste(as.character(x), collapse = "\n")
      }
    },
    .knitr_asis = function(...) paste(unlist(list(...)), collapse = "\n"),
    .package = "CausalMixGPD"
  )

  effect_meta <- list(
    ps_enabled = TRUE,
    ps_scale = "logit",
    backend = list(trt = "sb", con = "crp"),
    kernel = list(trt = "normal", con = "gamma"),
    GPD = list(trt = TRUE, con = FALSE)
  )

  qte_draw_obj <- structure(
    list(
      type = "cqte",
      probs = c(0.1, 0.9),
      n_pred = 2L,
      level = 0.9,
      interval = "hpd",
      x = matrix(c(1, 0, 0, 1), nrow = 2L),
      ps = c(0.2, 0.8),
      meta = effect_meta,
      qte = list(
        fit = data.frame(index = 0.9, estimate = 0.8, lower = 0.5, upper = 1.1),
        draws = matrix(c(NA_real_, NA_real_, 0.7, 0.9), nrow = 2L)
      )
    ),
    class = "causalmixgpd_qte"
  )

  qte_df_obj <- structure(
    list(
      type = "qte",
      probs = c(0.1, 0.5, 0.9),
      n_pred = 8L,
      level = 0.95,
      interval = "credible",
      x = matrix(1, nrow = 8L, ncol = 1L),
      ps = rep(0.4, 8L),
      meta = effect_meta,
      qte = list(
        fit = data.frame(
          id = seq_len(8L),
          index = rep(c(0.1, 0.5, 0.9, 0.5), each = 2L),
          estimate = seq(0.2, 1.6, length.out = 8L),
          lower = seq(0.1, 1.5, length.out = 8L),
          upper = seq(0.3, 1.7, length.out = 8L)
        )
      )
    ),
    class = "causalmixgpd_qte"
  )

  qte_matrix_obj <- structure(
    list(
      type = "cqte",
      probs = c(0.25, 0.75),
      n_pred = 4L,
      level = 0.9,
      interval = "hpd",
      x = matrix(1, nrow = 4L, ncol = 1L),
      ps = rep(0.5, 4L),
      meta = effect_meta,
      qte = list(fit = NULL, draws = NULL),
      fit = matrix(c(1, 2, 3, 4, 5, 6, 7, 8), nrow = 4L, ncol = 2L)
    ),
    class = "causalmixgpd_qte"
  )

  ate_df_obj <- structure(
    list(
      type = "cate",
      n_pred = 7L,
      level = 0.95,
      interval = "credible",
      nsim_mean = 15L,
      x = matrix(1, nrow = 7L, ncol = 1L),
      ps = rep(0.6, 7L),
      meta = effect_meta,
      ate = list(
        fit = data.frame(
          id = seq_len(7L),
          estimate = seq(0.5, 1.7, length.out = 7L),
          lower = seq(0.3, 1.5, length.out = 7L),
          upper = seq(0.7, 1.9, length.out = 7L)
        )
      )
    ),
    class = "causalmixgpd_ate"
  )

  ate_vec_obj <- structure(
    list(
      type = "ate",
      n_pred = 5L,
      level = 0.9,
      interval = "hpd",
      nsim_mean = 12L,
      x = matrix(1, nrow = 5L, ncol = 1L),
      ps = rep(0.3, 5L),
      meta = effect_meta,
      ate = list(fit = NULL),
      fit = c(0.4, 0.6, 0.8, 1.0, 1.2),
      lower = c(0.2, 0.4, 0.6, 0.8, 1.0),
      upper = c(0.6, 0.8, 1.0, 1.2, 1.4)
    ),
    class = "causalmixgpd_ate"
  )

  qte_draw_sum <- summary(qte_draw_obj)
  expect_s3_class(qte_draw_sum, "summary.causalmixgpd_qte")
  expect_true(all(is.na(qte_draw_sum$quantile_summary[1, c("estimate_qte", "mean_qte", "ci_lower")])))
  expect_equal(qte_draw_sum$quantile_summary$estimate_qte[2], 0.8)

  qte_matrix_sum <- summary(qte_matrix_obj)
  expect_s3_class(qte_matrix_sum, "summary.causalmixgpd_qte")
  expect_equal(nrow(qte_matrix_sum$quantile_summary), 2L)

  ate_vec_sum <- summary(ate_vec_obj)
  expect_s3_class(ate_vec_sum, "summary.causalmixgpd_ate")
  expect_equal(ate_vec_sum$ate_stats$mean_ate, mean(ate_vec_obj$fit))
  expect_true(is.list(ate_vec_sum$ci_summary))

  qte_knitr <- print(qte_df_obj, max_rows = 2L)
  expect_true(is.character(qte_knitr))
  expect_match(qte_knitr, "PS scale: logit")
  expect_match(qte_knitr, "Credible interval: credible")

  qte_matrix_knitr <- print(qte_matrix_obj, max_rows = 2L)
  expect_true(is.character(qte_matrix_knitr))
  expect_match(qte_matrix_knitr, "\\(matrix: 4 x 2\\)")
  expect_match(qte_matrix_knitr, "\\.\\.\\. \\(2 more rows\\)")

  qte_summary_knitr <- print(qte_draw_sum, digits = 2L)
  expect_true(is.character(qte_summary_knitr))
  expect_match(qte_summary_knitr, "Model specification:")
  expect_match(qte_summary_knitr, "Interval: hpd")

  ate_knitr <- print(ate_df_obj, max_rows = 2L)
  expect_true(is.character(ate_knitr))
  expect_match(ate_knitr, "Posterior mean draws: 15")
  expect_match(ate_knitr, "Credible interval: credible")

  ate_vec_knitr <- print(ate_vec_obj, max_rows = 2L)
  expect_true(is.character(ate_vec_knitr))
  expect_match(ate_vec_knitr, "\\(vector: 5\\)")
  expect_match(ate_vec_knitr, "\\.\\.\\. \\(3 more\\)")

  ate_summary_knitr <- print(ate_vec_sum, digits = 2L)
  expect_true(is.character(ate_summary_knitr))
  expect_match(ate_summary_knitr, "Model specification:")
  expect_no_match(ate_summary_knitr, "Credible interval width:")
  expect_no_match(ate_summary_knitr, "ATE statistics:")

  knitr_mode <- FALSE
  expect_output(print(qte_df_obj, max_rows = 2L), "Credible interval: credible")
  expect_output(print(qte_matrix_obj, max_rows = 2L), "\\(matrix: 4 x 2\\)")
  expect_output(print(qte_draw_sum, digits = 2L), "Model specification:")
  expect_output(print(ate_df_obj, max_rows = 2L), "Posterior mean draws: 15")
  expect_output(print(ate_vec_obj, max_rows = 2L), "\\(vector: 5\\)")
  expect_output(print(ate_vec_sum, digits = 2L), "Model specification:")
  expect_no_output(print(ate_vec_sum, digits = 2L), "Credible interval width:")
})

test_that("coverage-heavy glue diagnostics cover grid defaults violation summaries and malformed draws", {
  check_glue_validity <- getFromNamespace("check_glue_validity", "CausalMixGPD")

  crp_fit <- structure(
    list(
      spec = list(
        meta = list(backend = "crp", GPD = TRUE, has_X = TRUE),
        dispatch = list(
          backend = "crp",
          GPD = TRUE,
          gpd = list(
            threshold = list(mode = "constant"),
            tail_scale = list(mode = "constant", value = 1.5)
          )
        )
      ),
      data = list(
        X = matrix(c(1, 0, 0, 1), nrow = 2L),
        y = c(0.1, 0.5, 0.9, 1.4)
      )
    ),
    class = "mixgpd_fit"
  )

  testthat::local_mocked_bindings(
    .get_dispatch = function(...) {
      list(
        bulk_params = c("mean"),
        d = function(x, w, mean, tail_shape = NULL, threshold = NULL, tail_scale = NULL, log = 0L) {
          out <- rep(0.1, length(x))
          out[2L] <- -0.2
          if (log == 1L) log(pmax(out, 1e-8)) else out
        },
        p = function(q, w, mean, tail_shape = NULL, threshold = NULL, tail_scale = NULL, lower.tail = 1L, log.p = 0L) {
          if (length(q) == 1L) {
            out <- if (q < threshold) 0.2 else 0.6
          } else {
            out <- c(0.2, 0.7, 0.5, 1.2)[seq_along(q)]
          }
          if (log.p == 1L) log(pmax(out, 1e-8)) else out
        }
      )
    },
    .extract_draws_matrix = function(...) {
      matrix(
        c(
          0.2, 0.4, 0.8, 0.1,
          0.3, 0.5, 0.9, 0.2,
          0.4, 0.6, 1.0, 0.3
        ),
        nrow = 3L,
        byrow = TRUE,
        dimnames = list(NULL, c("tail_shape", "threshold_1", "threshold_2", "mean[1]"))
      )
    },
    .extract_weights = function(...) matrix(c(0.7, 0.3, Inf, 0, 0.6, 0.4), nrow = 3L, byrow = TRUE),
    .extract_bulk_params = function(...) list(mean = matrix(c(0.1, 0.2, 0.3), nrow = 3L)),
    .package = "CausalMixGPD"
  )

  res <- check_glue_validity(crp_fit, newdata =matrix(c(1, 0, 0, 1), nrow = 2L), grid = NULL, n_draws = 5L, check_continuity = TRUE)
  expect_equal(res$n_checked_draws, 3L)
  expect_equal(res$grid_n, 200L)
  expect_true(res$violations$cdf_range > 0L)
  expect_true(res$violations$cdf_monotone > 0L)
  expect_true(res$violations$density_nonneg > 0L)
  expect_true(res$violations$continuity > 0L)
  expect_true(length(res$details$bad_draws) >= 1L)
  expect_true(length(res$details$examples) >= 1L)
  expect_false(all(unlist(res$pass), na.rm = TRUE))

  testthat::local_mocked_bindings(
    .extract_draws_matrix = function(...) matrix(1, nrow = 1L, ncol = 1L),
    .package = "CausalMixGPD"
  )
  expect_error(check_glue_validity(crp_fit, newdata =matrix(c(1, 0, 0, 1), nrow = 2L), grid = c(0, 1)), "Posterior draws not found or malformed")
})

test_that("coverage-heavy build-run helper edge modes cover spliced scalar and standard branches", {
  build_dimensions_from_spec <- getFromNamespace("build_dimensions_from_spec", "CausalMixGPD")
  build_monitors_from_spec <- getFromNamespace("build_monitors_from_spec", "CausalMixGPD")

  kernel_info <- list(
    param_types = list(mean = "location", scale = "scale", shape = "shape", sd = "sd", rate = "scale"),
    bulk_support = list(mean = "", scale = "positive_scale", shape = "positive_shape", sd = "positive_sd", rate = "positive_scale")
  )

  spliced_link_spec <- list(
    meta = list(backend = "spliced", N = 4L, P = 2L, components = 3L, GPD = TRUE, kernel = "normal", has_X = TRUE),
    plan = list(
      concentration = list(mode = "dist", dist = "lognormal", args = list(meanlog = 0.2, sdlog = 0.4)),
      bulk = list(
        mean = list(mode = "link", link = "power", link_power = 2, beta_prior = list(dist = "lognormal", args = list(meanlog = -0.2, sdlog = 0.3))),
        scale = list(mode = "dist", dist = "lognormal", args = list(meanlog = 0.1, sdlog = 0.2))
      ),
      ps = list(prior = list(dist = "lognormal", args = list(meanlog = -1, sdlog = 0.5))),
      gpd = list(
        threshold = list(mode = "link", link = "power", link_power = 2, beta_prior = list(dist = "lognormal", args = list(meanlog = -2, sdlog = 0.2))),
        tail_scale = list(mode = "link", link = "power", link_power = 1.5, beta_prior = list(dist = "lognormal", args = list(meanlog = -0.5, sdlog = 0.2))),
        tail_shape = list(mode = "link", link = "power", link_power = 3, beta_prior = list(dist = "lognormal", args = list(meanlog = -0.3, sdlog = 0.2)))
      )
    ),
    kernel_info = kernel_info
  )

  spliced_scalar_spec <- list(
    meta = list(backend = "spliced", N = 4L, P = 2L, components = 3L, GPD = TRUE, kernel = "normal", has_X = TRUE),
    plan = list(
      concentration = list(mode = "fixed", value = 0.8),
      bulk = list(
        shape = list(mode = "fixed", value = 0.4),
        sd = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1))
      ),
      gpd = list(
        threshold = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_scale = list(mode = "fixed", value = 1.3),
        tail_shape = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 0.3))
      )
    ),
    kernel_info = kernel_info
  )

  standard_link_spec <- list(
    meta = list(backend = "sb", N = 5L, P = 2L, components = 3L, GPD = TRUE, kernel = "normal", has_X = TRUE),
    plan = list(
      concentration = list(mode = "fixed", value = 1.2),
      bulk = list(
        mean = list(mode = "link", link = "power", link_power = 2, beta_prior = list(dist = "normal", args = list(mean = 0, sd = 1))),
        sd = list(mode = "fixed", value = 1)
      ),
      ps = list(prior = list(dist = "normal", args = list(mean = 0, sd = 1))),
      gpd = list(
        threshold = list(mode = "link", link = "power", link_power = 2, link_dist = list(dist = "lognormal"), beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.2))),
        sdlog_u = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
        tail_scale = list(mode = "link", link = "power", link_power = 1.5, beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.4))),
        tail_shape = list(mode = "link", link = "power", link_power = 3, beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.3)))
      )
    ),
    kernel_info = kernel_info
  )

  standard_scalar_spec <- list(
    meta = list(backend = "crp", N = 4L, P = 0L, components = 4L, GPD = TRUE, kernel = "gamma", has_X = FALSE),
    plan = list(
      concentration = list(mode = "dist", dist = "gamma", args = list()),
      bulk = list(
        shape = list(mode = "dist", dist = "gamma", args = list()),
        rate = list(mode = "fixed", value = 1)
      ),
      gpd = list(
        threshold = list(mode = "fixed", value = 1.1),
        tail_scale = list(mode = "dist", dist = "gamma", args = list()),
        tail_shape = list(mode = "dist", dist = "normal", args = list())
      )
    ),
    kernel_info = kernel_info
  )

  standard_no_ld_spec <- list(
    meta = list(backend = "sb", N = 4L, P = 1L, components = 3L, GPD = TRUE, kernel = "normal", has_X = TRUE),
    plan = list(
      concentration = list(mode = "fixed", value = 1),
      bulk = list(mean = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 1))),
      gpd = list(
        threshold = list(mode = "link", link = "identity", beta_prior = list(dist = "normal", args = list(mean = 0, sd = 0.2))),
        tail_scale = list(mode = "fixed", value = 1.1),
        tail_shape = list(mode = "fixed", value = 0.2)
      )
    ),
    kernel_info = kernel_info
  )

  sb_small_spec <- list(
    meta = list(backend = "sb", N = 4L, P = 0L, components = 2L, GPD = FALSE, kernel = "normal", has_X = FALSE),
    plan = list(
      concentration = list(mode = "fixed", value = 1),
      bulk = list(mean = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 1)))
    ),
    kernel_info = kernel_info
  )

  gating_bad_spec <- list(
    meta = list(backend = "sb", N = 4L, P = 0L, components = 3L, GPD = FALSE, kernel = "normal", has_X = FALSE),
    plan = list(
      concentration = list(mode = "fixed", value = 1),
      bulk = list()
    ),
    cluster = list(gating = TRUE),
    kernel_info = kernel_info
  )

  unknown_backend_spec <- list(
    meta = list(backend = "mystery", N = 4L, P = 1L, components = 3L, GPD = FALSE, kernel = "normal", has_X = TRUE),
    plan = list(
      concentration = list(mode = "fixed", value = 1),
      bulk = list(mean = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 1)))
    ),
    kernel_info = kernel_info
  )

  bad_const_spec <- standard_scalar_spec
  bad_const_spec$plan$concentration$mode <- "weird"

  bad_prior_spec <- standard_scalar_spec
  bad_prior_spec$plan$bulk$shape$mode <- "weird"

  bad_dims_spec <- standard_scalar_spec
  bad_dims_spec$plan$gpd$tail_shape$mode <- "weird"

  bad_mon_spec <- standard_no_ld_spec
  bad_mon_spec$plan$bulk$mean$mode <- "weird"

  mons_sp_link <- build_monitors_from_spec(spliced_link_spec, monitor_latent = TRUE)
  dims_sp_link <- build_dimensions_from_spec(spliced_link_spec)
  inits_sp_link <- build_inits_from_spec(spliced_link_spec, y = c(1, 2, 3, 4))
  const_sp_link <- build_constants_from_spec(spliced_link_spec)
  prior_sp_link <- build_prior_table_from_spec(spliced_link_spec)

  expect_true(all(c("alpha", "z[1:4]", "beta_mean[1:3,1:2]", "beta_ps_mean[1:3]", "scale[1:3]", "beta_threshold[1:3,1:2]", "beta_tail_scale[1:3,1:2]", "beta_tail_shape[1:3,1:2]") %in% mons_sp_link))
  expect_true(all(c("z", "beta_mean", "beta_ps_mean", "scale", "beta_threshold", "eta_threshold", "threshold_i", "beta_tail_scale", "eta_tail_scale", "tail_scale_i", "beta_tail_shape", "eta_tail_shape", "tail_shape_i") %in% names(dims_sp_link)))
  expect_true(all(c("alpha", "z", "beta_mean", "beta_ps_mean", "scale", "beta_threshold", "beta_tail_scale", "beta_tail_shape") %in% names(inits_sp_link)))
  expect_equal(unname(unlist(const_sp_link[c("N", "P", "components")])), c(4, 2, 3))
  expect_true(any(grepl("power=2", prior_sp_link$notes, fixed = TRUE)))
  expect_true(any(grepl("power=1.5", prior_sp_link$notes, fixed = TRUE)))
  expect_true(any(grepl("power=3", prior_sp_link$notes, fixed = TRUE)))

  mons_sp_scalar <- build_monitors_from_spec(spliced_scalar_spec, monitor_latent = TRUE)
  dims_sp_scalar <- build_dimensions_from_spec(spliced_scalar_spec)
  inits_sp_scalar <- build_inits_from_spec(spliced_scalar_spec, y = c(2, 3, 4, 5))
  const_sp_scalar <- build_constants_from_spec(spliced_scalar_spec)
  prior_sp_scalar <- build_prior_table_from_spec(spliced_scalar_spec)

  expect_true(all(c("alpha", "z[1:4]", "sd[1:3]", "threshold[1:3]", "tail_scale[1:3]", "tail_shape[1:3]") %in% mons_sp_scalar))
  expect_true(all(c("z", "sd", "threshold", "tail_scale", "tail_shape") %in% names(dims_sp_scalar)))
  expect_true(all(c("z", "sd", "threshold", "tail_shape") %in% names(inits_sp_scalar)))
  expect_false("tail_scale" %in% names(inits_sp_scalar))
  expect_equal(unname(unlist(const_sp_scalar[c("N", "P", "components")])), c(4, 2, 3))
  expect_true(any(prior_sp_scalar$parameter == "tail_scale" & prior_sp_scalar$mode == "fixed"))
  expect_true(any(prior_sp_scalar$parameter == "threshold" & prior_sp_scalar$mode == "dist"))

  mons_std_link <- build_monitors_from_spec(standard_link_spec, monitor_v = TRUE, monitor_latent = TRUE)
  dims_std_link <- build_dimensions_from_spec(standard_link_spec)
  inits_std_link <- build_inits_from_spec(standard_link_spec, y = rep(NA_real_, 5L))
  const_std_link <- build_constants_from_spec(standard_link_spec)
  prior_std_link <- build_prior_table_from_spec(standard_link_spec)

  expect_true(all(c("alpha", "w[1:3]", "z[1:5]", "v[1:2]", "beta_mean[1:3,1:2]", "beta_ps_mean[1:3]", "threshold[1:5]", "beta_threshold[1:2]", "sdlog_u", "beta_tail_scale[1:2]", "beta_tail_shape[1:2]") %in% mons_std_link))
  expect_true(all(c("v", "w", "z", "beta_mean", "beta_ps_mean", "threshold", "beta_threshold", "beta_tail_scale", "beta_tail_shape") %in% names(dims_std_link)))
  expect_true(all(c("v", "z", "beta_mean", "beta_ps_mean", "beta_threshold", "threshold", "sdlog_u", "beta_tail_scale", "beta_tail_shape") %in% names(inits_std_link)))
  expect_equal(inits_std_link$threshold, rep(1, 5L))
  expect_equal(inits_std_link$sdlog_u, 0.2)
  expect_equal(unname(unlist(const_std_link[c("N", "P", "components")])), c(5, 2, 3))
  expect_true(any(prior_std_link$mode == "link+dist"))
  expect_true(any(grepl("Lognormal", prior_std_link$notes, fixed = TRUE)))

  mons_std_scalar <- build_monitors_from_spec(standard_scalar_spec, monitor_latent = TRUE)
  dims_std_scalar <- build_dimensions_from_spec(standard_scalar_spec)
  inits_std_scalar <- build_inits_from_spec(standard_scalar_spec, y = c(1, 2, 3, 4))
  const_std_scalar <- build_constants_from_spec(standard_scalar_spec)
  prior_std_scalar <- build_prior_table_from_spec(standard_scalar_spec)

  expect_true(all(c("alpha", "z[1:4]", "shape[1:4]", "threshold", "tail_scale", "tail_shape") %in% mons_std_scalar))
  expect_true(all(c("z", "shape") %in% names(dims_std_scalar)))
  expect_true(all(c("alpha", "z", "shape", "threshold", "tail_scale", "tail_shape") %in% names(inits_std_scalar)))
  expect_equal(unname(unlist(const_std_scalar[c("N", "P", "components")])), c(4, 0, 4))
  expect_true(any(prior_std_scalar$parameter == "threshold" & prior_std_scalar$mode == "fixed"))
  expect_true(any(prior_std_scalar$parameter == "tail_scale" & prior_std_scalar$mode == "dist"))

  inits_std_no_ld <- build_inits_from_spec(standard_no_ld_spec, y = rep(NA_real_, 4L))
  inits_small <- build_inits_from_spec(sb_small_spec, y = c(1, 2, 3, 4))
  expect_equal(inits_std_no_ld$threshold, rep(1, 4L))
  expect_length(inits_small$v, 1L)

  expect_error(build_monitors_from_spec(gating_bad_spec, monitor_latent = TRUE), "Cluster gating requires P > 0")
  expect_error(build_dimensions_from_spec(gating_bad_spec), "Cluster gating requires P > 0")
  expect_error(build_inits_from_spec(gating_bad_spec), "Cluster gating requires P > 0")
  expect_error(build_monitors_from_spec(unknown_backend_spec), "Unknown backend")
  expect_error(build_dimensions_from_spec(unknown_backend_spec), "Unknown backend")
  expect_error(build_inits_from_spec(unknown_backend_spec), "Unknown backend")
  expect_error(build_constants_from_spec(bad_const_spec), "Invalid plan\\$concentration\\$mode")
  expect_error(build_prior_table_from_spec(bad_prior_spec), "Invalid bulk mode")
  expect_error(build_dimensions_from_spec(bad_dims_spec), "Invalid gpd\\$tail_shape mode")
  expect_error(build_monitors_from_spec(bad_mon_spec), "Invalid bulk plan mode")
})

test_that("coverage-heavy build-run helper guard rails cover remaining mode errors and sampler fallbacks", {
  build_dimensions_from_spec <- getFromNamespace("build_dimensions_from_spec", "CausalMixGPD")
  build_monitors_from_spec <- getFromNamespace("build_monitors_from_spec", "CausalMixGPD")

  kernel_info <- list(
    param_types = list(mean = "location", scale = "scale", weird = "other"),
    bulk_support = list(mean = "", scale = "", weird = "")
  )

  make_spec <- function(backend = "sb", P = 1L, GPD = FALSE, bulk = list(), gpd = list(),
                        concentration = list(mode = "fixed", value = 1), N = 4L, K = 3L,
                        kernel_info_override = kernel_info) {
    list(
      meta = list(
        backend = backend,
        N = N,
        P = P,
        components = K,
        GPD = GPD,
        kernel = "normal",
        has_X = P > 0L
      ),
      plan = list(
        concentration = concentration,
        bulk = bulk,
        gpd = gpd
      ),
      kernel_info = kernel_info_override
    )
  }

  make_gpd_only_spec <- function(backend, P, param, mode, N = 4L, K = 3L) {
    gpd <- setNames(list(list(mode = mode)), param)
    make_spec(backend = backend, P = P, GPD = TRUE, gpd = gpd, N = N, K = K)
  }

  link_p0_spec <- make_spec(
    backend = "sb",
    P = 0L,
    bulk = list(scale = list(mode = "link"))
  )
  expect_error(build_monitors_from_spec(link_p0_spec), "link-mode but P=0")
  expect_error(build_dimensions_from_spec(link_p0_spec), "link-mode but P=0")
  expect_error(build_inits_from_spec(link_p0_spec), "link-mode but P=0")

  init_type_spec <- make_spec(
    backend = "sb",
    P = 0L,
    bulk = list(
      scale = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)),
      weird = list(mode = "dist", dist = "normal", args = list(mean = 0, sd = 1))
    )
  )
  init_type <- build_inits_from_spec(init_type_spec, y = c(1, 2, 3, 4))
  expect_equal(init_type$scale, rep(1, 3L))
  expect_equal(init_type$weird, rep(0, 3L))

  bad_inits_conc <- init_type_spec
  bad_inits_conc$plan$concentration$mode <- "odd"
  expect_error(build_inits_from_spec(bad_inits_conc), "Invalid plan\\$concentration\\$mode")
  expect_error(build_prior_table_from_spec(bad_inits_conc), "Invalid plan\\$concentration\\$mode")

  spliced_thr_dist_na <- make_spec(
    backend = "spliced",
    P = 1L,
    GPD = TRUE,
    gpd = list(threshold = list(mode = "dist", dist = "gamma", args = list(shape = 2, rate = 1)))
  )
  expect_equal(build_inits_from_spec(spliced_thr_dist_na, y = rep(NA_real_, 4L))$threshold, rep(1, 3L))

  spliced_thr_fixed <- make_spec(
    backend = "spliced",
    P = 1L,
    GPD = TRUE,
    gpd = list(threshold = list(mode = "fixed", value = 1.2))
  )
  expect_false("threshold" %in% names(build_inits_from_spec(spliced_thr_fixed, y = c(1, 2, 3, 4))))

  spliced_tsh_fixed <- make_spec(
    backend = "spliced",
    P = 1L,
    GPD = TRUE,
    gpd = list(tail_shape = list(mode = "fixed", value = 0.2))
  )
  expect_false("tail_shape" %in% names(build_inits_from_spec(spliced_tsh_fixed, y = c(1, 2, 3, 4))))

  for (param in c("threshold", "tail_scale", "tail_shape")) {
    sp_p0 <- make_gpd_only_spec("spliced", 0L, param, "link")
    std_p0 <- make_gpd_only_spec("sb", 0L, param, "link")
    expect_error(build_monitors_from_spec(sp_p0), "P=0")
    expect_error(build_dimensions_from_spec(sp_p0), "P=0")
    expect_error(build_inits_from_spec(sp_p0), "P=0")
    expect_error(build_monitors_from_spec(std_p0), "P=0")
    expect_error(build_dimensions_from_spec(std_p0), "P=0")
    expect_error(build_inits_from_spec(std_p0), "P=0")
  }

  for (param in c("threshold", "tail_scale", "tail_shape")) {
    sp_bad <- make_gpd_only_spec("spliced", 1L, param, "weird")
    std_bad <- make_gpd_only_spec("sb", 1L, param, "weird")
    expect_error(build_monitors_from_spec(sp_bad), "Invalid gpd")
    expect_error(build_dimensions_from_spec(sp_bad), "Invalid gpd")
    expect_error(build_inits_from_spec(sp_bad), "Invalid gpd")
    if (param != "tail_shape") {
      expect_error(build_monitors_from_spec(std_bad), "Invalid gpd")
    }
    expect_error(build_dimensions_from_spec(std_bad), "Invalid gpd")
    if (param != "tail_shape") {
      expect_error(build_inits_from_spec(std_bad), "Invalid gpd")
    }
    expect_error(build_prior_table_from_spec(std_bad), "Invalid gpd")
  }

  bad_const_bulk <- make_spec(
    backend = "sb",
    P = 1L,
    bulk = list(mean = list(mode = "weird"))
  )
  expect_error(build_constants_from_spec(bad_const_bulk), "Invalid bulk plan mode")

  bad_const_sdlog <- make_spec(
    backend = "sb",
    P = 1L,
    GPD = TRUE,
    gpd = list(
      threshold = list(mode = "link", link_dist = list(dist = "lognormal")),
      sdlog_u = list(mode = "fixed", value = 0.2)
    )
  )
  expect_error(build_constants_from_spec(bad_const_sdlog), "sdlog_u must be dist-mode")

  bad_const_thr <- make_gpd_only_spec("sb", 1L, "threshold", "weird")
  bad_const_ts <- make_gpd_only_spec("sb", 1L, "tail_scale", "weird")
  bad_const_tsh <- make_gpd_only_spec("sb", 1L, "tail_shape", "weird")
  expect_error(build_constants_from_spec(bad_const_thr), "Invalid gpd\\$threshold mode")
  expect_error(build_constants_from_spec(bad_const_ts), "Invalid gpd\\$tail_scale mode")
  expect_error(build_constants_from_spec(bad_const_tsh), "Invalid gpd\\$tail_shape mode")

  std_tail_shape_fixed <- make_spec(
    backend = "sb",
    P = 1L,
    GPD = TRUE,
    gpd = list(tail_shape = list(mode = "fixed", value = 0.1))
  )
  expect_equal(unname(unlist(build_constants_from_spec(std_tail_shape_fixed)[c("N", "P", "components")])), c(4, 1, 3))
  prior_tail_shape_fixed <- build_prior_table_from_spec(std_tail_shape_fixed)
  expect_true(any(prior_tail_shape_fixed$parameter == "tail_shape" & prior_tail_shape_fixed$mode == "fixed"))

  conf_empty <- list(samplerConfs = NULL)
  expect_identical(.configure_samplers(conf_empty, spec = list()), conf_empty)

  conf_fallback <- list(
    samplerConfs = list(
      list(
        name = "RW",
        target = "beta_mean[1]",
        control = list(
          clusterVarInfo = list(
            clusterNodes = c("z[1]", "junk"),
            numNodesPerCluster = 2L
          )
        )
      )
    ),
    getModel = function() stop("no model"),
    model = list(
      getNodeNames = function(stochOnly = TRUE, includeData = FALSE) c("z[1]", "beta_mean[1]")
    ),
    getSamplers = function() list(list(target = "beta_mean[1]")),
    removeSamplers = function(...) invisible(NULL),
    addSampler = function(...) invisible(NULL)
  )
  conf_norm <- .configure_samplers(conf_fallback, spec = list(), z_update_every = 2L)
  expect_true(is.list(conf_norm$samplerConfs[[1]]$control$clusterVarInfo$clusterNodes))
  expect_equal(conf_norm$samplerConfs[[1]]$control$adaptInterval, 200L)
  expect_true(isTRUE(conf_norm$samplerConfs[[1]]$control$adaptScaleOnly))
})

strip_ansi <- function(x) {
  gsub("\u001b\\[[0-9;]*m", "", x, perl = TRUE)
}

test_that("progress helpers emit step messages when enabled", {
  ctx <- .cmgpd_progress_start(total_steps = 2L, enabled = TRUE, quiet = FALSE, label = "unit")
  msgs <- utils::capture.output(
    {
      .cmgpd_progress_step(ctx, "step one")
      .cmgpd_progress_step(ctx, "step two")
    },
    type = "message"
  )
  msgs <- strip_ansi(msgs)
  expect_true(any(grepl("^\\[unit\\] step one$", msgs)))
  expect_true(any(grepl("^\\[unit\\] step two$", msgs)))
  expect_silent(.cmgpd_progress_done(ctx))
})

test_that("progress helpers are silent when disabled", {
  ctx <- .cmgpd_progress_start(total_steps = 2L, enabled = FALSE, quiet = FALSE, label = "unit")
  expect_silent(.cmgpd_progress_step(ctx, "no-op"))
  expect_silent(.cmgpd_progress_done(ctx))
})

test_that("long-running APIs expose progress controls in formals", {
  expect_true("show_progress" %in% names(formals(predict.mixgpd_fit)))
  expect_true("show_progress" %in% names(formals(predict.causalmixgpd_causal_fit)))
  expect_true("show_progress" %in% names(formals(cqte)))
  expect_true("show_progress" %in% names(formals(cate)))
  expect_true("show_progress" %in% names(formals(qte)))
  expect_true("show_progress" %in% names(formals(qtt)))
  expect_true("show_progress" %in% names(formals(ate)))
  expect_true("show_progress" %in% names(formals(att)))
  expect_true("show_progress" %in% names(formals(ate_rmean)))
  expect_true("quiet" %in% names(formals(run_mcmc_causal)))
  expect_identical(formals(run_mcmc_bundle_manual)$quiet, FALSE)
})

test_that("run_mcmc_bundle_manual and predict honor progress toggles", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  set.seed(222)
  y <- abs(stats::rnorm(18)) + 0.1
  bundle <- build_nimble_bundle(
    y = y,
    backend = "sb",
    kernel = "normal",
    GPD = FALSE,
    components = 4,
    mcmc = mcmc_fast(seed = 222L)
  )

  expect_message(
    fit <- run_mcmc_bundle_manual(bundle, show_progress = TRUE, quiet = FALSE),
    "\\[mixgpd\\] Validating configuration"
  )
  active_msg <- utils::capture.output(
    active_out <- utils::capture.output(
      run_mcmc_bundle_manual(bundle, show_progress = TRUE, quiet = FALSE),
      type = "output"
    ),
    type = "message"
  )
  active_msg <- strip_ansi(active_msg)
  active_lines <- c(active_out, active_msg)
  expect_true(any(grepl("[mixgpd] Validating configuration", active_lines, fixed = TRUE)))
  expect_false(any(grepl("\\[[0-9]+/[0-9]+\\].*\\[[=-]+\\]\\s+[0-9]+%", active_lines)))
  expect_false(any(grepl("===== Monitors =====|running chain|Defining model", active_lines)))

  quiet_msg <- utils::capture.output(
    quiet_fit <- run_mcmc_bundle_manual(bundle, show_progress = FALSE, quiet = TRUE),
    type = "message"
  )
  quiet_out <- utils::capture.output(quiet_fit, type = "output")
  expect_false(any(grepl("\\[[0-9]+/[0-9]+\\]", c(quiet_msg, quiet_out))))

  expect_message(
    predict(fit, type = "quantile", index = 0.5, show_progress = TRUE),
    "\\[predict_mixgpd\\] Validating prediction inputs"
  )
  pred_quiet <- utils::capture.output(
    predict(fit, type = "quantile", index = 0.5, show_progress = FALSE),
    type = "message"
  )
  expect_false(any(grepl("Validating prediction inputs", pred_quiet, fixed = TRUE)))
})

test_that("causal and cluster runners emit family-labeled progress", {
  skip_if_not_test_level("ci")
  skip_if_not_installed("nimble")

  set.seed(223)
  n <- 24
  X <- cbind(x1 = stats::rnorm(n), x2 = stats::runif(n))
  A <- stats::rbinom(n, 1, 0.5)
  y <- abs(0.3 * X[, 1] + A + stats::rnorm(n)) + 0.1
  cb <- build_causal_bundle(
    y = y, X = X, A = A,
    backend = "sb", kernel = "normal", GPD = FALSE, components = 4,
    mcmc_outcome = mcmc_fast(seed = 223L),
    mcmc_ps = mcmc_fast(seed = 224L),
    PS = "logit"
  )
  causal_msgs <- strip_ansi(utils::capture.output(
    run_mcmc_causal(cb, show_progress = TRUE, quiet = FALSE),
    type = "message"
  ))
  expect_true(any(grepl("[causal] Validating causal MCMC configuration", causal_msgs, fixed = TRUE)))
  expect_true(any(grepl("[ps] Validating PS MCMC inputs", causal_msgs, fixed = TRUE)))
  expect_true(any(grepl("[mixgpd] Validating configuration", causal_msgs, fixed = TRUE)))

  dat <- data.frame(
    y = abs(stats::rnorm(18)) + 0.2,
    x1 = stats::rnorm(18),
    x2 = stats::runif(18)
  )
  cluster_bundle <- build_cluster_bundle(
    y ~ x1 + x2,
    data = dat,
    kernel = "normal",
    components = 4,
    type = "both",
    GPD = FALSE,
    mcmc = mcmc_fast(seed = 225L)
  )
  cluster_msgs <- strip_ansi(utils::capture.output(
    run_cluster_mcmc(cluster_bundle, show_progress = TRUE, quiet = FALSE),
    type = "message"
  ))
  expect_true(any(grepl("[cluster] Validating configuration", cluster_msgs, fixed = TRUE)))
})

#' Gamma mixture distribution
#'
#' Finite mixture of gamma components for positive-support bulk modeling. The scalar functions in
#' this topic are the compiled building blocks behind the gamma bulk kernel family.
#'
#' The mixture density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k f_{\Gamma}(x \mid \alpha_k, \theta_k),
#' \qquad x > 0,
#' }
#' with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use [gamma_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param shape,scale Numeric vectors of length \eqn{K} giving Gamma shape and scale parameters.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qGammaMix()` returns a numeric vector
#'   with the same length as `p`.
#'
#' @details
#' Under the package parameterization, each component has density
#' \eqn{f_\Gamma(x \mid \alpha,\theta) = x^{\alpha-1}\exp(-x/\theta) /
#' \{\Gamma(\alpha)\theta^\alpha\}} on \eqn{x>0}. The mixture CDF is therefore
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k F_\Gamma(x \mid \alpha_k,\theta_k).
#' }
#' Random generation first selects a component according to the normalized mixture weights and then
#' draws from the corresponding gamma distribution. Since finite gamma mixtures do not have closed
#' form quantiles, \code{qGammaMix()} obtains them numerically by inverting the mixture CDF.
#'
#' The analytical mean is
#' \deqn{
#' E(X) = \sum_{k=1}^K \tilde{w}_k \alpha_k \theta_k.
#' }
#' This expression is reused in posterior predictive mean calculations for gamma-based fits.
#'
#' @seealso [gamma_mixgpd()], [gamma_gpd()], [gamma_lowercase()], [build_nimble_bundle()],
#'   [kernel_support_table()].
#' @family gamma kernel families
#'
#' @examples
#' w <- c(0.55, 0.30, 0.15)
#' scale <- c(1.0, 2.5, 5.0)
#' shape <- c(2, 4, 6)
#'
#' dGammaMix(2.0, w = w, scale = scale, shape = shape, log = 0)
#' pGammaMix(2.0, w = w, scale = scale, shape = shape, lower.tail = 1, log.p = 0)
#' qGammaMix(0.50, w = w, scale = scale, shape = shape)
#' qGammaMix(0.95, w = w, scale = scale, shape = shape)
#' replicate(10, rGammaMix(1, w = w, scale = scale, shape = shape))
#' @rdname gamma_mix
#' @name gamma_mix
#' @aliases dGammaMix pGammaMix rGammaMix qGammaMix
#' @importFrom stats dgamma pgamma rgamma qgamma runif uniroot
NULL

#' @describeIn gamma_mix Gamma mixture density
#' @export
dGammaMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    K <- length(w)
    wsum <- sum(w)

    if (wsum <= 0.0) {
      if (log == 1) return(-1.0e300) else return(0.0)
    }

    ww <- w / wsum

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + ww[j] * dgamma(x, shape = shape[j], scale = scale[j], log = 0)
    }

    if (log == 1) return(log(s0))
    return(s0)
  }
)

#' @describeIn gamma_mix Gamma mixture distribution function
#' @export
pGammaMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    K <- length(w)
    wsum <- sum(w)

    if (wsum <= 0.0) {
      if (log.p == 1) return(-1.0e300) else return(0.0)
    }

    ww <- w / wsum

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + ww[j] * pgamma(q, shape = shape[j], scale = scale[j],
                                  lower.tail = 1, log.p = 0)
    }

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn gamma_mix Gamma mixture random generation
#' @export
rGammaMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) return(0.0)

    u <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0

    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (u <= cw) {
          idx <- j
          found <- 1
        }
      }
    }

    return(rgamma(1, shape = shape[idx], scale = scale[idx]))
  }
)

# ---- nimbleFunction aliases for internal use ----
dGammaMix_nf <- dGammaMix
pGammaMix_nf <- pGammaMix
rGammaMix_nf <- rGammaMix

#' @describeIn gamma_mix Gamma mixture quantile function
#' @export
qGammaMix <- function(p, w, shape, scale,
                      lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- 0; next }
    if (pi >= 1) { out[i] <- Inf; next }

    hi <- max(stats::qgamma(pi, shape = shape, scale = scale), na.rm = TRUE)
    if (!is.finite(hi) || hi <= 0) hi <- 1
    f0 <- as.numeric(pGammaMix(0, w = w, shape = shape, scale = scale, lower.tail = 1, log.p = 0) - pi)
    fhi <- as.numeric(pGammaMix(hi, w = w, shape = shape, scale = scale, lower.tail = 1, log.p = 0) - pi)
    iter <- 0L
    while (is.finite(fhi) && f0 * fhi > 0 && hi < 1e20 && iter < 60L) {
      hi <- hi * 2
      fhi <- as.numeric(pGammaMix(hi, w = w, shape = shape, scale = scale, lower.tail = 1, log.p = 0) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(hi) || hi <= 0 || !is.finite(fhi) || f0 * fhi > 0) {
      out[i] <- Inf
    } else {
      out[i] <- stats::uniroot(
        function(z) pGammaMix(z, w = w, shape = shape, scale = scale,
                              lower.tail = 1, log.p = 0) - pi,
        interval = c(0, hi),
        tol = tol, maxiter = maxiter
      )$root
    }
  }
  out
}

meanGammaMix <- function(w, shape, scale) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  alpha <- as.numeric(shape)
  theta <- as.numeric(scale)
  if (length(alpha) != length(ww) || length(theta) != length(ww)) return(NA_real_)
  if (any(!is.finite(alpha)) || any(!is.finite(theta)) || any(alpha <= 0) || any(theta <= 0)) return(NA_real_)
  sum(ww * alpha * theta)
}

meanGammaMixTrunc <- function(w, shape, scale, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  alpha <- as.numeric(shape)
  theta <- as.numeric(scale)
  u <- as.numeric(threshold)[1]
  if (length(alpha) != length(ww) || length(theta) != length(ww)) return(NA_real_)
  if (any(!is.finite(alpha)) || any(!is.finite(theta)) || any(alpha <= 0) || any(theta <= 0) || is.na(u)) return(NA_real_)
  if (is.infinite(u) && u > 0) return(meanGammaMix(w = ww, shape = alpha, scale = theta))
  if (u <= 0) return(0)
  sum(ww * alpha * theta * stats::pgamma(u, shape = alpha + 1, scale = theta))
}

#' Gamma mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to a gamma mixture bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}.
#' @param shape,scale Numeric vectors of length \eqn{K} giving Gamma shape and scale parameters.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qGammaMixGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' The gamma mixture governs the body of the distribution up to the threshold \eqn{u}. Beyond
#' \eqn{u}, only the remaining survival mass is modeled by the GPD, giving
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' f_{mix}(x), & x < u, \\
#' \{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' This is the positive-support analogue of the normal and lognormal splice families. Bulk quantiles
#' are still found by numerical inversion, while tail quantiles use the explicit GPD inverse.
#'
#' @seealso [gamma_mix()], [gamma_gpd()], [gpd()], [gamma_lowercase()], [dpmgpd()].
#' @family gamma kernel families
#' @examples
#' w <- c(0.55, 0.30, 0.15)
#' scale <- c(1.0, 2.5, 5.0)
#' shape <- c(2, 4, 6)
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dGammaMixGpd(4.0, w = w, scale = scale, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape, log = 0)
#' pGammaMixGpd(4.0, w = w, scale = scale, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape, lower.tail = 1, log.p = 0)
#' qGammaMixGpd(0.50, w = w, scale = scale, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape)
#' qGammaMixGpd(0.95, w = w, scale = scale, shape = shape,
#'             threshold = threshold, tail_scale = tail_scale,
#'             tail_shape = tail_shape)
#' replicate(10, rGammaMixGpd(1, w = w, scale = scale, shape = shape,
#'                           threshold = threshold,
#'                           tail_scale = tail_scale,
#'                           tail_shape = tail_shape))
#' @rdname gamma_mixgpd
#' @name gamma_mixgpd
#' @aliases dGammaMixGpd pGammaMixGpd rGammaMixGpd qGammaMixGpd
#' @importFrom stats runif uniroot
NULL

#' @describeIn gamma_mixgpd Gamma mixture + GPD tail density
#' @export
dGammaMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

          if (x < threshold) return(dGammaMix_nf(x, w, shape, scale, log))

          Fu <- pGammaMix_nf(threshold, w, shape, scale, 1, 0)
    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)

    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn gamma_mixgpd Gamma mixture + GPD tail distribution function
#' @export
pGammaMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

          if (q < threshold) return(pGammaMix_nf(q, w, shape, scale, lower.tail, log.p))

          Fu <- pGammaMix_nf(threshold, w, shape, scale, 1, 0)
    G  <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)

    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn gamma_mixgpd Gamma mixture + GPD tail random generation
#' @export
rGammaMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 shape = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pGammaMix_nf(threshold, w, shape, scale, 1, 0)
    u  <- runif(1, 0.0, 1.0)

    if (u < Fu) return(rGammaMix_nf(1, w, shape, scale))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn gamma_mixgpd Gamma mixture + GPD tail quantile function
#' @export
qGammaMixGpd <- function(p, w, shape, scale, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE,
                         tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- pGammaMix(threshold, w, shape, scale, 1, 0)
  out <- numeric(length(p))

  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qGammaMix(pi, w, shape, scale, lower.tail = TRUE, log.p = FALSE, tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}

#' Gamma with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' gamma bulk distribution.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param scale Numeric scalar scale parameter for the Gamma bulk.
#' @param shape Numeric scalar Gamma shape parameter.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are
#'   \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric tolerance for numerical inversion in \code{qGammaGpd}.
#' @param maxiter Maximum iterations for numerical inversion in \code{qGammaGpd}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars.
#'   `qGammaGpd()` returns a numeric vector with the same length as `p`.
#'
#' @details
#' This topic combines a single gamma bulk with a generalized Pareto exceedance model. If
#' \eqn{F_\Gamma(u)} is the bulk probability below the threshold, then the splice replaces the
#' upper tail by \eqn{\{1-F_\Gamma(u)\}g_{GPD}(x)} while leaving the lower region unchanged. The
#' resulting distribution is continuous at the threshold and preserves the gamma body exactly below
#' \eqn{u}.
#'
#' The ordinary mean is finite only when the GPD shape satisfies \eqn{\xi < 1}. For heavier tails,
#' predictive mean summaries should be replaced by restricted means or quantile summaries.
#'
#' @seealso [gamma_mix()], [gamma_mixgpd()], [gpd()], [gamma_lowercase()].
#' @family gamma kernel families
#'
#' @examples
#' scale <- 2.5
#' shape <- 4
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dGammaGpd(4.0, scale = scale, shape = shape,
#'          threshold = threshold, tail_scale = tail_scale,
#'          tail_shape = tail_shape, log = 0)
#' pGammaGpd(4.0, scale = scale, shape = shape,
#'          threshold = threshold, tail_scale = tail_scale,
#'          tail_shape = tail_shape, lower.tail = 1, log.p = 0)
#' qGammaGpd(0.50, scale = scale, shape = shape,
#'          threshold = threshold, tail_scale = tail_scale,
#'          tail_shape = tail_shape)
#' qGammaGpd(0.95, scale = scale, shape = shape,
#'          threshold = threshold, tail_scale = tail_scale,
#'          tail_shape = tail_shape)
#' replicate(10, rGammaGpd(1, scale = scale, shape = shape,
#'                        threshold = threshold,
#'                        tail_scale = tail_scale,
#'                        tail_shape = tail_shape))
#'
#' @rdname gamma_gpd
#' @name gamma_gpd
#' @aliases dGammaGpd pGammaGpd rGammaGpd qGammaGpd
NULL


#' @describeIn gamma_gpd Gamma + GPD tail density
#' @export
dGammaGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 shape = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    if (x < threshold) return(dgamma(x, shape = shape, scale = scale, log = log))

    Fu <- pgamma(threshold, shape = shape, scale = scale, lower.tail = 1, log.p = 0)
    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)

    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn gamma_gpd Gamma + GPD tail distribution function
#' @export
pGammaGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 shape = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    if (q < threshold) return(pgamma(q, shape = shape, scale = scale, lower.tail = lower.tail, log.p = log.p))

    Fu <- pgamma(threshold, shape = shape, scale = scale, lower.tail = 1, log.p = 0)
    G  <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)

    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn gamma_gpd Gamma + GPD tail random generation
#' @export
rGammaGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 shape = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pgamma(threshold, shape = shape, scale = scale, lower.tail = 1, log.p = 0)
    u  <- runif(1, 0.0, 1.0)

    if (u < Fu) return(rgamma(1, shape = shape, scale = scale))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn gamma_gpd Gamma + GPD tail quantile function
#' @export
#' @param tol Numeric tolerance for numerical inversion in \code{qGammaGpd}.
#' @param maxiter Maximum iterations for numerical inversion in \code{qGammaGpd}.
qGammaGpd <- function(p, shape, scale, threshold, tail_scale, tail_shape,
                      lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- stats::pgamma(threshold, shape = shape, scale = scale, lower.tail = TRUE, log.p = FALSE)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- stats::qgamma(pi, shape = shape, scale = scale, lower.tail = TRUE, log.p = FALSE)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}

# ==========================================================
# Lowercase vectorized R wrappers for Gamma kernels
# ==========================================================

#' Lowercase vectorized gamma distribution functions
#'
#' Vectorized R wrappers for the scalar gamma-kernel topics in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param shape,scale Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These wrappers are vectorized interfaces to the scalar gamma and gamma-plus-GPD routines. They
#' preserve the package's shape-scale parameterization and the same splice definition used in the
#' fitted-model prediction code. Quantile wrappers delegate to the scalar inversion code rather than
#' implementing separate approximations.
#'
#' @seealso [gamma_mix()], [gamma_mixgpd()], [gamma_gpd()], [bundle()], [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.55, 0.3, 0.15)
#' shp <- c(2, 4, 6)
#' scl <- c(1, 2.5, 5)
#'
#' # Gamma mixture
#' dgammamix(c(1, 2, 3), w = w, shape = shp, scale = scl)
#' rgammamix(5, w = w, shape = shp, scale = scl)
#'
#' # Gamma mixture + GPD
#' dgammamixgpd(c(2, 3, 4), w = w, shape = shp, scale = scl,
#'              threshold = 3, tail_scale = 0.9, tail_shape = 0.2)
#'
#' # Gamma + GPD (single component)
#' dgammagpd(c(2, 3, 4), shape = 4, scale = 2.5, threshold = 3,
#'           tail_scale = 0.9, tail_shape = 0.2)
#'
#' @name gamma_lowercase
#' @rdname gamma_lowercase
NULL

# ---- Gamma Mix lowercase wrappers ----

#' @describeIn gamma_lowercase Gamma mixture density (vectorized)
#' @export
dgammamix <- function(x, w, shape, scale, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dGammaMix_nf(xi, w = w, shape = shape, scale = scale, log = log_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma mixture distribution function (vectorized)
#' @export
pgammamix <- function(q, w, shape, scale, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pGammaMix_nf(qi, w = w, shape = shape, scale = scale,
                                                  lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma mixture quantile function (vectorized)
#' @export
qgammamix <- function(p, w, shape, scale, lower.tail = TRUE, log.p = FALSE,
                      tol = 1e-10, maxiter = 200) {
  qGammaMix(p, w = w, shape = shape, scale = scale, lower.tail = lower.tail,
            log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn gamma_lowercase Gamma mixture random generation (vectorized)
#' @export
rgammamix <- function(n, w, shape, scale) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rGammaMix_nf(1L, w = w, shape = shape, scale = scale)),
         numeric(1L))
}

# ---- Gamma Mix + GPD lowercase wrappers ----

#' @describeIn gamma_lowercase Gamma mixture + GPD density (vectorized)
#' @export
dgammamixgpd <- function(x, w, shape, scale, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dGammaMixGpd(xi, w = w, shape = shape, scale = scale,
                                                  threshold = threshold, tail_scale = tail_scale,
                                                  tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma mixture + GPD distribution function (vectorized)
#' @export
pgammamixgpd <- function(q, w, shape, scale, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pGammaMixGpd(qi, w = w, shape = shape, scale = scale,
                                                  threshold = threshold, tail_scale = tail_scale,
                                                  tail_shape = tail_shape,
                                                  lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma mixture + GPD quantile function (vectorized)
#' @export
qgammamixgpd <- function(p, w, shape, scale, threshold, tail_scale, tail_shape,
                         lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qGammaMixGpd(p, w = w, shape = shape, scale = scale, threshold = threshold,
               tail_scale = tail_scale, tail_shape = tail_shape,
               lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn gamma_lowercase Gamma mixture + GPD random generation (vectorized)
#' @export
rgammamixgpd <- function(n, w, shape, scale, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rGammaMixGpd(1L, w = w, shape = shape, scale = scale,
                                                          threshold = threshold, tail_scale = tail_scale,
                                                          tail_shape = tail_shape)),
         numeric(1L))
}

# ---- Gamma + GPD lowercase wrappers ----

#' @describeIn gamma_lowercase Gamma + GPD density (vectorized)
#' @export
dgammagpd <- function(x, shape, scale, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dGammaGpd(xi, shape = shape, scale = scale,
                                               threshold = threshold, tail_scale = tail_scale,
                                               tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma + GPD distribution function (vectorized)
#' @export
pgammagpd <- function(q, shape, scale, threshold, tail_scale, tail_shape,
                      lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pGammaGpd(qi, shape = shape, scale = scale,
                                               threshold = threshold, tail_scale = tail_scale,
                                               tail_shape = tail_shape,
                                               lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn gamma_lowercase Gamma + GPD quantile function (vectorized)
#' @export
qgammagpd <- function(p, shape, scale, threshold, tail_scale, tail_shape,
                      lower.tail = TRUE, log.p = FALSE) {
  qGammaGpd(p, shape = shape, scale = scale, threshold = threshold,
            tail_scale = tail_scale, tail_shape = tail_shape,
            lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn gamma_lowercase Gamma + GPD random generation (vectorized)
#' @export
rgammagpd <- function(n, shape, scale, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rGammaGpd(1L, shape = shape, scale = scale,
                                                       threshold = threshold, tail_scale = tail_scale,
                                                       tail_shape = tail_shape)),
         numeric(1L))
}

set_max_fails(Inf)
options(testthat.reporter = "summary")

is_pkg_check <- nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_")) ||
  nzchar(Sys.getenv("RCMDCHECK")) ||
  identical(tolower(Sys.getenv("NOT_CRAN", "")), "false")

cache_default <- if (is_pkg_check) "0" else "1"
Sys.setenv(DPMIXGPD_USE_CACHE = Sys.getenv("DPMIXGPD_USE_CACHE", cache_default))

is_covr_run <- isTRUE(getOption("covr", FALSE)) ||
  nzchar(Sys.getenv("DPMIXGPD_COVERAGE")) ||
  nzchar(Sys.getenv("COVERAGE")) ||
  nzchar(Sys.getenv("R_COVR"))

if (is_pkg_check && !nzchar(Sys.getenv("DPMIXGPD_TEST_LEVEL"))) {
  Sys.setenv(DPMIXGPD_TEST_LEVEL = "cran")
}

if (!is_pkg_check && is_covr_run &&
  !nzchar(Sys.getenv("DPMIXGPD_TEST_LEVEL"))) {
  Sys.setenv(DPMIXGPD_TEST_LEVEL = "ci")
}

if (exists("init_kernel_registry", mode = "function", inherits = TRUE)) {
  try(init_kernel_registry(), silent = TRUE)
}

.cache_enabled <- function() {
  val <- tolower(Sys.getenv("DPMIXGPD_USE_CACHE", "0"))
  val %in% c("1", "true", "yes")
}

.cache_dir <- function() {
  d <- Sys.getenv("DPMIXGPD_CACHE_DIR", "")
  if (!nzchar(d)) {
    is_check <- nzchar(Sys.getenv("_R_CHECK_PACKAGE_NAME_")) ||
      nzchar(Sys.getenv("RCMDCHECK")) ||
      identical(tolower(Sys.getenv("NOT_CRAN", "")), "false")
    d <- if (is_check) {
      file.path(tempdir(), "CausalMixGPD_test_cache")
    } else {
      file.path("tests", "testthat", "_cache")
    }
  }
  dir.create(d, recursive = TRUE, showWarnings = FALSE)
  d
}

.cache_hash <- function(x) {
  tf <- tempfile()
  writeLines(x, tf)
  h <- tools::md5sum(tf)
  unlink(tf)
  as.character(h)
}

.cache_path <- function(key) {
  file.path(.cache_dir(), paste0(key, ".rds"))
}

.cache_get <- function(key) {
  path <- .cache_path(key)
  if (!file.exists(path)) return(NULL)
  out <- tryCatch(readRDS(path), error = function(e) NULL)
  if (is.null(out)) {
    unlink(path)
    return(NULL)
  }
  out
}

.cache_set <- function(key, value) {
  path <- .cache_path(key)
  saveRDS(value, path)
  invisible(path)
}

#' Laplace (double exponential) mixture distribution
#'
#' Finite mixture of Laplace components for real-valued bulk modeling. The scalar functions in this
#' topic are the NIMBLE-compatible building blocks for Laplace-based kernels.
#'
#' The mixture density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k f_{Lap}(x \mid \mu_k, b_k),
#' }
#' with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use [laplace_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param location Numeric vector of length \eqn{K} giving component locations.
#' @param scale Numeric vector of length \eqn{K} giving component scales.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qLaplaceMix()` returns a numeric
#'   vector with the same length as `p`.
#'
#' @details
#' Each component is a Laplace law with density
#' \eqn{f(x \mid \mu,b) = (2b)^{-1}\exp\{-|x-\mu|/b\}}. The mixture CDF is the corresponding
#' weighted average of component CDFs, and random generation selects a component first and then
#' samples from that component. As with the other finite mixtures in the package, the quantile has
#' no closed form and is therefore obtained numerically.
#'
#' The analytical mean of the mixture is simply
#' \deqn{
#' E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k.
#' }
#' That is the formula used in downstream predictive mean calculations for Laplace-based fits.
#'
#' @seealso [laplace_MixGpd()], [laplace_gpd()], [laplace_lowercase()],
#'   [build_nimble_bundle()], [kernel_support_table()].
#' @family laplace kernel families
#'
#' @examples
#' w <- c(0.50, 0.30, 0.20)
#' location <- c(-1, 0.5, 2.0)
#' scale <- c(1.0, 0.7, 1.4)
#'
#' dLaplaceMix(0.8, w = w, location = location, scale = scale, log = FALSE)
#' pLaplaceMix(0.8, w = w, location = location, scale = scale,
#'            lower.tail = TRUE, log.p = FALSE)
#' qLaplaceMix(0.50, w = w, location = location, scale = scale)
#' qLaplaceMix(0.95, w = w, location = location, scale = scale)
#' replicate(10, rLaplaceMix(1, w = w, location = location, scale = scale))
#' @rdname laplace_mix
#' @name laplace_mix
#' @aliases dLaplaceMix pLaplaceMix rLaplaceMix qLaplaceMix
#' @importFrom stats runif uniroot
#' @importFrom nimble ddexp pdexp rdexp qdexp
NULL

#' @describeIn laplace_mix Laplace mixture density
#' @export
dLaplaceMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) {
      if (log == 1) return(-1.0e300) else return(0.0)
    }
    ww <- w / wsum

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + ww[j] * ddexp(x, location[j], scale[j], log = 0)
    }

    if (log == 1) return(log(s0))
    return(s0)
  }
)

#' @describeIn laplace_mix Laplace mixture distribution function
#' @export
pLaplaceMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) {
      if (log.p == 1) return(-1.0e300) else return(0.0)
    }
    ww <- w / wsum

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + ww[j] * pdexp(q, location[j], scale[j], lower.tail = 1, log.p = 0)
    }

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn laplace_mix Laplace mixture random generation
#' @export
rLaplaceMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0.0) return(0.0)

    u <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0

    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (u <= cw) {
          idx <- j
          found <- 1
        }
      }
    }

    return(rdexp(1, location[idx], scale[idx]))
  }
)



#' @describeIn laplace_mix Laplace mixture quantile function
#' @export
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum iterations for \code{stats::uniroot}.
qLaplaceMix <- function(p, w, location, scale,
                        lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- -Inf; next }
    if (pi >= 1) { out[i] <- Inf; next }

    lo <- min(nimble::qdexp(pi, location, scale, lower.tail = TRUE, log.p = FALSE), na.rm = TRUE)
    hi <- max(nimble::qdexp(pi, location, scale, lower.tail = TRUE, log.p = FALSE), na.rm = TRUE)
    if (!is.finite(lo)) lo <- -1e20
    if (!is.finite(hi)) hi <- 1e20
    f_lo <- as.numeric(pLaplaceMix(lo, w = w, location, scale, lower.tail = TRUE, log.p = FALSE) - pi)
    f_hi <- as.numeric(pLaplaceMix(hi, w = w, location, scale, lower.tail = TRUE, log.p = FALSE) - pi)
    iter <- 0L
    while (is.finite(f_lo) && f_lo > 0 && lo > -1e20 && iter < 60L) {
      step <- max(1, abs(lo))
      lo <- lo - step
      f_lo <- as.numeric(pLaplaceMix(lo, w = w, location, scale, lower.tail = TRUE, log.p = FALSE) - pi)
      iter <- iter + 1L
    }
    iter <- 0L
    while (is.finite(f_hi) && f_hi < 0 && hi < 1e20 && iter < 60L) {
      step <- max(1, abs(hi))
      hi <- hi + step
      f_hi <- as.numeric(pLaplaceMix(hi, w = w, location, scale, lower.tail = TRUE, log.p = FALSE) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(lo) || !is.finite(hi) || lo >= hi || !is.finite(f_lo) || !is.finite(f_hi) || f_lo * f_hi > 0) {
      out[i] <- NA_real_
    } else {
      out[i] <- stats::uniroot(
        function(z) pLaplaceMix(z, w = w,  location, scale,
                                lower.tail = TRUE, log.p = FALSE) - pi,
        interval = c(lo, hi),
        tol = tol, maxiter = maxiter
      )$root
    }
  }
  out
}

meanLaplaceMix <- function(w, location, scale) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  loc <- as.numeric(location)
  scl <- as.numeric(scale)
  if (length(loc) != length(ww) || length(scl) != length(ww)) return(NA_real_)
  if (any(!is.finite(loc)) || any(!is.finite(scl)) || any(scl <= 0)) return(NA_real_)
  sum(ww * loc)
}

meanLaplaceMixTrunc <- function(w, location, scale, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  loc <- as.numeric(location)
  scl <- as.numeric(scale)
  u <- as.numeric(threshold)[1]
  if (length(loc) != length(ww) || length(scl) != length(ww)) return(NA_real_)
  if (any(!is.finite(loc)) || any(!is.finite(scl)) || any(scl <= 0) || is.na(u)) return(NA_real_)
  if (is.infinite(u) && u > 0) return(meanLaplaceMix(w = ww, location = loc, scale = scl))
  comp <- ifelse(
    u < loc,
    0.5 * (u - scl) * exp((u - loc) / scl),
    loc - 0.5 * (u + scl) * exp(-(u - loc) / scl)
  )
  sum(ww * comp)
}


#' Laplace mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to a Laplace mixture
#' bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}.
#' @param location Numeric vector of length \eqn{K} giving component locations.
#' @param scale Numeric vector of length \eqn{K} giving component scales.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qLaplaceMixGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' This family keeps the Laplace mixture body below the threshold and replaces the upper tail with a
#' generalized Pareto exceedance model scaled by the residual survival mass at the threshold. The
#' tail density is therefore
#' \deqn{
#' f(x) = \{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), \qquad x \ge u.
#' }
#' Bulk quantiles are found numerically from the mixture CDF, and tail quantiles are computed by
#' rescaling the upper-tail probability and applying the GPD inverse.
#'
#' @seealso [laplace_mix()], [laplace_gpd()], [gpd()], [laplace_lowercase()], [dpmgpd()].
#' @family laplace kernel families
#'
#' @examples
#' w <- c(0.50, 0.30, 0.20)
#' location <- c(-1, 0.5, 2.0)
#' scale <- c(1.0, 0.7, 1.4)
#' threshold <- 1
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dLaplaceMixGpd(2.0, w = w, location = location, scale = scale,
#'               threshold = threshold, tail_scale = tail_scale,
#'               tail_shape = tail_shape, log = FALSE)
#' pLaplaceMixGpd(2.0, w = w, location = location, scale = scale,
#'               threshold = threshold, tail_scale = tail_scale,
#'               tail_shape = tail_shape, lower.tail = TRUE, log.p = FALSE)
#' qLaplaceMixGpd(0.50, w = w, location = location, scale = scale,
#'               threshold = threshold, tail_scale = tail_scale,
#'               tail_shape = tail_shape)
#' qLaplaceMixGpd(0.95, w = w, location = location, scale = scale,
#'               threshold = threshold, tail_scale = tail_scale,
#'               tail_shape = tail_shape)
#' replicate(10, rLaplaceMixGpd(1, w = w, location = location, scale = scale,
#'                             threshold = threshold,
#'                             tail_scale = tail_scale,
#'                             tail_shape = tail_shape))
#' @rdname laplace_MixGpd
#' @name laplace_MixGpd
#' @aliases dLaplaceMixGpd pLaplaceMixGpd rLaplaceMixGpd qLaplaceMixGpd
NULL

#' @describeIn laplace_MixGpd Laplace mixture + GPD tail density
#' @export
dLaplaceMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    if (x < threshold) return(dLaplaceMix(x, w, location, scale, log))

    Fu <- pLaplaceMix(threshold, w, location, scale, 1, 0)
    val <- (1.0 - Fu) * dGpd(x, threshold, scale = tail_scale, shape = tail_shape, log = 0)

    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn laplace_MixGpd Laplace mixture + GPD tail distribution function
#' @export
pLaplaceMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    # FIXED: below threshold must call pLaplaceMix(..., scale = scale), not tail_scale
    if (q < threshold) return(pLaplaceMix(q, w, location, scale, lower.tail, log.p))

    Fu <- pLaplaceMix(threshold, w, location, scale, 1, 0)
    G  <- pGpd(q, threshold, scale = tail_scale, shape = tail_shape, lower.tail = 1, log.p = 0)

    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn laplace_MixGpd Laplace mixture + GPD tail random generation
#' @export
rLaplaceMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 location = double(1),
                 scale = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pLaplaceMix(threshold, w, location, scale, 1, 0)
    u  <- runif(1, 0.0, 1.0)

    if (u < Fu) return(rLaplaceMix(1, w, location, scale))
    return(rGpd(1, threshold, scale = tail_scale, shape = tail_shape))
  }
)



#' @describeIn laplace_MixGpd Laplace mixture + GPD tail quantile function
#' @export
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum iterations for \code{stats::uniroot}.
qLaplaceMixGpd <- function(p, w, location, scale, threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE,
                           tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- pLaplaceMix(threshold, w, location, scale, lower.tail = TRUE, log.p = FALSE)
  out <- numeric(length(p))

  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qLaplaceMix(pi, w, location, scale, lower.tail = TRUE, log.p = FALSE,
                            tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g,  threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


#' Laplace with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' Laplace bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param location Numeric scalar location parameter for the Laplace bulk.
#' @param scale Numeric scalar scale parameter for the Laplace bulk.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Logical; if \code{TRUE}, return the log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qLaplaceGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' This topic pairs a single Laplace bulk distribution with a generalized Pareto exceedance tail.
#' The splice is continuous at the threshold because the tail density is multiplied by the Laplace
#' survival probability at that threshold.
#'
#' The ordinary mean exists only when the GPD tail has \eqn{\xi < 1}. If the fitted tail is too
#' heavy for an ordinary mean to exist, the package directs users to restricted means or quantiles
#' rather than returning an unstable mean summary.
#'
#' @seealso [laplace_mix()], [laplace_MixGpd()], [gpd()], [laplace_lowercase()].
#' @family laplace kernel families
#'
#' @examples
#' location <- 0.5
#' scale <- 1.0
#' threshold <- 1
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dLaplaceGpd(2.0, location, scale, threshold, tail_scale, tail_shape, log = FALSE)
#' pLaplaceGpd(2.0, location, scale, threshold, tail_scale, tail_shape,
#'            lower.tail = TRUE, log.p = FALSE)
#' qLaplaceGpd(0.50, location, scale, threshold, tail_scale, tail_shape)
#' qLaplaceGpd(0.95, location, scale, threshold, tail_scale, tail_shape)
#' replicate(10, rLaplaceGpd(1, location, scale, threshold, tail_scale, tail_shape))
#' @rdname laplace_gpd
#' @name laplace_gpd
#' @aliases dLaplaceGpd pLaplaceGpd rLaplaceGpd qLaplaceGpd
NULL

#' @describeIn laplace_gpd Laplace + GPD tail density
#' @export
dLaplaceGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 location = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))

    if (x < threshold) return(ddexp(x, location, scale, log = log))

    Fu <- pdexp(threshold, location, scale, lower.tail = 1, log.p = 0)
    val <- (1.0 - Fu) * dGpd(x, threshold, scale = tail_scale, shape = tail_shape, log = 0)

    if (log == 1) return(log(val))
    return(val)
  }
)

#' @describeIn laplace_gpd Laplace + GPD tail distribution function
#' @export
pLaplaceGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 location = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))

    if (q < threshold) return(pdexp(q, location, scale, lower.tail = lower.tail, log.p = log.p))

    Fu <- pdexp(threshold, location, scale, lower.tail = 1, log.p = 0)
    G  <- pGpd(q, threshold, scale = tail_scale, shape = tail_shape, lower.tail = 1, log.p = 0)

    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(cdf))
    return(cdf)
  }
)

#' @describeIn laplace_gpd Laplace + GPD tail random generation
#' @export
rLaplaceGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 location = double(0),
                 scale = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pdexp(threshold, location, scale, lower.tail = 1, log.p = 0)
    u  <- runif(1, 0.0, 1.0)

    if (u < Fu) return(rdexp(1, location, scale))
    return(rGpd(1, threshold, scale = tail_scale, shape = tail_shape))
  }
)



#' @describeIn laplace_gpd Laplace + GPD tail quantile function
#' @export
qLaplaceGpd <- function(p, location, scale, threshold, tail_scale, tail_shape, lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- nimble::pdexp(threshold,  location, scale, lower.tail = TRUE, log.p = FALSE)
  out <- numeric(length(p))

  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- nimble::qdexp(pi,  location, scale, lower.tail = TRUE, log.p = FALSE)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g,  threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


# ==========================================================
# Lowercase vectorized R wrappers for Laplace kernels
# ==========================================================

#' Lowercase vectorized Laplace distribution functions
#'
#' Vectorized R wrappers for the scalar Laplace-kernel topics in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param location,scale Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These helpers vectorize the scalar Laplace and Laplace-plus-GPD routines for interactive R use.
#' They retain the same location-scale parameterization and the same splice definition as the
#' uppercase functions. Quantiles continue to use the scalar root-finding or piecewise logic rather
#' than a separate approximation.
#'
#' @seealso [laplace_mix()], [laplace_MixGpd()], [laplace_gpd()], [bundle()],
#'   [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.3, 0.1)
#' loc <- c(0, 1, -2)
#' scl <- c(1, 0.9, 1.1)
#'
#' # Laplace mixture
#' dlaplacemix(c(-1, 0, 1), w = w, location = loc, scale = scl)
#' rlaplacemix(5, w = w, location = loc, scale = scl)
#'
#' @name laplace_lowercase
#' @rdname laplace_lowercase
NULL

# ---- Laplace Mix lowercase wrappers ----

#' @describeIn laplace_lowercase Laplace mixture density (vectorized)
#' @export
dlaplacemix <- function(x, w, location, scale, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLaplaceMix(xi, w = w, location = location, scale = scale, log = log_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace mixture distribution function (vectorized)
#' @export
plaplacemix <- function(q, w, location, scale, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLaplaceMix(qi, w = w, location = location, scale = scale,
                                                 lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace mixture quantile function (vectorized)
#' @export
qlaplacemix <- function(p, w, location, scale, lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  qLaplaceMix(p, w = w, location = location, scale = scale, lower.tail = lower.tail,
              log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn laplace_lowercase Laplace mixture random generation (vectorized)
#' @export
rlaplacemix <- function(n, w, location, scale) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLaplaceMix(1L, w = w, location = location, scale = scale)),
         numeric(1L))
}

# ---- Laplace Mix + GPD lowercase wrappers ----

#' @describeIn laplace_lowercase Laplace mixture + GPD density (vectorized)
#' @export
dlaplacemixgpd <- function(x, w, location, scale, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLaplaceMixGpd(xi, w = w, location = location, scale = scale,
                                                    threshold = threshold, tail_scale = tail_scale,
                                                    tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace mixture + GPD distribution function (vectorized)
#' @export
plaplacemixgpd <- function(q, w, location, scale, threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLaplaceMixGpd(qi, w = w, location = location, scale = scale,
                                                    threshold = threshold, tail_scale = tail_scale,
                                                    tail_shape = tail_shape,
                                                    lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace mixture + GPD quantile function (vectorized)
#' @export
qlaplacemixgpd <- function(p, w, location, scale, threshold, tail_scale, tail_shape,
                           lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qLaplaceMixGpd(p, w = w, location = location, scale = scale, threshold = threshold,
                 tail_scale = tail_scale, tail_shape = tail_shape,
                 lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn laplace_lowercase Laplace mixture + GPD random generation (vectorized)
#' @export
rlaplacemixgpd <- function(n, w, location, scale, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLaplaceMixGpd(1L, w = w, location = location, scale = scale,
                                                           threshold = threshold, tail_scale = tail_scale,
                                                           tail_shape = tail_shape)),
         numeric(1L))
}

# ---- Laplace + GPD lowercase wrappers ----

#' @describeIn laplace_lowercase Laplace + GPD density (vectorized)
#' @export
dlaplacegpd <- function(x, location, scale, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLaplaceGpd(xi, location = location, scale = scale,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace + GPD distribution function (vectorized)
#' @export
plaplacegpd <- function(q, location, scale, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLaplaceGpd(qi, location = location, scale = scale,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape,
                                                 lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn laplace_lowercase Laplace + GPD quantile function (vectorized)
#' @export
qlaplacegpd <- function(p, location, scale, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE) {
  qLaplaceGpd(p, location = location, scale = scale, threshold = threshold,
              tail_scale = tail_scale, tail_shape = tail_shape,
              lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn laplace_lowercase Laplace + GPD random generation (vectorized)
#' @export
rlaplacegpd <- function(n, location, scale, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLaplaceGpd(1L, location = location, scale = scale,
                                                         threshold = threshold, tail_scale = tail_scale,
                                                         tail_shape = tail_shape)),
         numeric(1L))
}

#' Validate bulk+tail glue for MixGPD predictive distribution
#'
#' This diagnostic checks whether the implied predictive distribution behaves like
#' a valid distribution (monotone CDF in \eqn{[0,1]}, nonnegative density, and
#' sensible behavior around the threshold when a GPD tail is enabled).
#'
#' The check is performed draw-by-draw on a user-specified grid. It is intended
#' for development, debugging, and CI (not for routine large-scale use).
#'
#' @param fit A \code{mixgpd_fit} object.
#' @param x Optional design matrix for conditional models. If \code{NULL}, uses training \code{X}.
#' @param grid Numeric evaluation grid. If \code{NULL}, defaults to a grid based on training \code{y}.
#' @param n_draws Number of posterior draws to check (sampled without replacement when possible).
#' @param tol Numerical tolerance for monotonicity/range checks.
#' @param check_continuity Logical; if \code{TRUE} and GPD is enabled, checks continuity at the threshold.
#' @param eps Small offset used for threshold continuity check.
#' @return A list with per-check pass/fail flags and summaries of violations.
#' @export
check_glue_validity <- function(fit,
                                x = NULL,
                                grid = NULL,
                                n_draws = 50L,
                                tol = 1e-8,
                                check_continuity = TRUE,
                                eps = 1e-6) {
  stopifnot(inherits(fit, "mixgpd_fit"))


  spec <- fit$spec %||% list()
  meta <- spec$meta %||% list()

  backend <- meta$backend %||% spec$dispatch$backend %||% "<unknown>"
  GPD <- isTRUE(meta$GPD %||% spec$dispatch$GPD)
  is_spliced <- identical(backend, "spliced")
  pred_backend <- if (backend %in% c("crp", "spliced")) "sb" else backend

  Xtrain <- fit$data$X %||% fit$X %||% NULL
  ytrain <- fit$data$y %||% fit$y %||% NULL

  has_X <- isTRUE(meta$has_X %||% (!is.null(Xtrain)))

  if (has_X) {
    if (is.null(x)) {
      if (is.null(Xtrain)) stop("Training X not found; provide 'x'.", call. = FALSE)
      X <- as.matrix(Xtrain)
    } else {
      X <- as.matrix(x)
    }
  } else {
    if (!is.null(x)) stop("Unconditional model: 'x' not allowed.", call. = FALSE)
    X <- NULL
  }

  if (is.null(grid)) {
    if (is.null(ytrain)) stop("Training y not found; provide 'grid'.", call. = FALSE)
    ytrain <- as.numeric(ytrain)
    lo <- stats::quantile(ytrain, probs = 0.001, na.rm = TRUE)
    hi <- stats::quantile(ytrain, probs = 0.999, na.rm = TRUE)
    grid <- seq(from = as.numeric(lo), to = as.numeric(hi), length.out = 200L)
  }
  grid <- as.numeric(grid)
  if (any(!is.finite(grid))) stop("'grid' must be finite numeric.", call. = FALSE)
  grid <- sort(unique(grid))
  G <- length(grid)

  fns <- .get_dispatch(fit, backend_override = pred_backend)
  d_fun <- fns$d
  p_fun <- fns$p
  bulk_params <- fns$bulk_params

  draw_mat <- .extract_draws_matrix(fit)
  if (is.null(draw_mat) || !is.matrix(draw_mat) || nrow(draw_mat) < 2L) {
    stop("Posterior draws not found or malformed in fitted object.", call. = FALSE)
  }
  S <- nrow(draw_mat)

  n_draws <- as.integer(n_draws)
  if (is.na(n_draws) || n_draws < 1L) stop("'n_draws' must be >= 1.", call. = FALSE)
  idx <- if (S <= n_draws) seq_len(S) else sample.int(S, size = n_draws, replace = FALSE)

  W_draws <- .extract_weights(draw_mat, backend = pred_backend)
  bulk_draws <- .extract_bulk_params(draw_mat, bulk_params = bulk_params)
  base_params <- names(bulk_draws)
  link_plan <- spec$dispatch$link_params %||% meta$link_params %||% list()
  if (!length(link_plan)) {
    bulk_plan <- spec$plan$bulk %||% list()
    for (nm in names(bulk_plan)) {
      ent <- bulk_plan[[nm]] %||% list()
      if (identical(ent$mode %||% "constant", "link")) {
        link_plan[[nm]] <- list(
          mode = "link",
          link = ent$link %||% "identity",
          link_power = ent$link_power %||% NULL
        )
      }
    }
  }
  link_params <- names(link_plan)

  # GPD pieces (constant or link)
  tail_shape <- NULL
  threshold_mat <- NULL
  threshold_scalar <- NULL
  tail_scale <- NULL
  spliced_gpd_draws <- list()
  spliced_gpd_link <- list()
  P <- if (!is.null(X)) ncol(X) else 0L
  n_x <- if (!is.null(X)) nrow(X) else 1L

  .apply_link <- function(eta, link, link_power = NULL) {
    link <- as.character(link %||% "identity")
    if (link == "identity") return(eta)
    if (link == "exp") return(exp(eta))
    if (link == "log") return(log(eta))
    if (link == "softplus") return(log1p(exp(eta)))
    if (link == "power") {
      if (is.null(link_power) || length(link_power) != 1L || !is.finite(as.numeric(link_power))) {
        stop("power link requires numeric link_power.", call. = FALSE)
      }
      pw <- as.numeric(link_power)
      return(eta ^ pw)
    }
    stop(sprintf("Unsupported link '%s'.", link), call. = FALSE)
  }

  .compute_link_eta <- function(s) {
    if (!length(link_params)) return(list())
    out <- list()
    for (nm in link_params) {
      plan <- link_plan[[nm]] %||% list()
      mode <- plan$mode %||% "constant"
      if (identical(mode, "link")) {
        if (is.null(X)) stop("link-mode requires X.", call. = FALSE)
        link <- plan$link %||% "identity"
        pw   <- plan$link_power %||% NULL

        beta_cols <- grep(paste0("^beta_", nm, "\\[[0-9]+,\\s*[0-9]+\\]$"), colnames(draw_mat), value = TRUE)
        if (length(beta_cols)) {
          idx1 <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+),\\s*([0-9]+)\\]$"), "\\1", beta_cols))
          idx2 <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+),\\s*([0-9]+)\\]$"), "\\2", beta_cols))
          Kb <- max(idx1, na.rm = TRUE)
          Pb <- max(idx2, na.rm = TRUE)
          beta_mat <- matrix(NA_real_, nrow = Kb, ncol = Pb)
          for (j in seq_along(beta_cols)) {
            beta_mat[idx1[j], idx2[j]] <- draw_mat[s, beta_cols[j]]
          }
          eta_mat <- tcrossprod(X, beta_mat)
          out[[nm]] <- .apply_link(eta_mat, link, pw)
        } else {
          beta_cols_1d <- grep(paste0("^beta_", nm, "\\[[0-9]+\\]$"), colnames(draw_mat), value = TRUE)
          if (length(beta_cols_1d)) {
            idx <- as.integer(sub(paste0("^beta_", nm, "\\[([0-9]+)\\]$"), "\\1", beta_cols_1d))
            ord <- order(idx)
            beta_cols_1d <- beta_cols_1d[ord]
            Kb <- ncol(W_draws)
            Pb <- P
            if (length(beta_cols_1d) == Kb * Pb) {
              beta_vec <- as.numeric(draw_mat[s, beta_cols_1d])
              beta_mat <- matrix(beta_vec, nrow = Kb, ncol = Pb)
              eta_mat <- tcrossprod(X, beta_mat)
              out[[nm]] <- .apply_link(eta_mat, link, pw)
            } else {
              beta_nm <- .indexed_block(draw_mat, paste0("beta_", nm), K = P)
              eta <- as.numeric(tcrossprod(X, beta_nm[s, , drop = FALSE]))
              out[[nm]] <- matrix(as.numeric(.apply_link(eta, link, pw)), nrow = n_x)
            }
          } else {
            beta_nm <- .indexed_block(draw_mat, paste0("beta_", nm), K = P)
            eta <- as.numeric(tcrossprod(X, beta_nm[s, , drop = FALSE]))
            out[[nm]] <- matrix(as.numeric(.apply_link(eta, link, pw)), nrow = n_x)
          }
        }
      } else {
        if (!(nm %in% colnames(draw_mat))) stop(sprintf("'%s' not found in posterior draws.", nm), call. = FALSE)
        out[[nm]] <- matrix(rep(as.numeric(draw_mat[s, nm]), n_x), nrow = n_x)
      }
    }
    out
  }

  if (GPD) {
    gpd_plan <- spec$dispatch$gpd %||% meta$gpd %||% list()

    if (is_spliced) {
      K_sp <- ncol(W_draws)
      for (nm in c("threshold", "tail_scale", "tail_shape")) {
        ent <- gpd_plan[[nm]] %||% list(mode = "dist")
        mode <- ent$mode %||% "dist"
        if (identical(mode, "link")) {
          if (is.null(X)) stop(sprintf("%s link-mode requires X.", nm), call. = FALSE)
          beta_arr <- .indexed_block_matrix(draw_mat, paste0("beta_", nm), K = K_sp, P = P, allow_missing = TRUE)
          if (is.null(beta_arr)) stop(sprintf("beta_%s not found in posterior draws.", nm), call. = FALSE)
          spliced_gpd_link[[nm]] <- list(
            beta = beta_arr,
            link = ent$link %||% if (identical(nm, "tail_shape")) "identity" else "exp",
            link_power = ent$link_power %||% NULL
          )
        } else {
          vals <- .indexed_block(draw_mat, nm, K = K_sp, allow_missing = TRUE)
          if (is.null(vals) && identical(mode, "fixed") && !is.null(ent$value)) {
            vals <- matrix(rep(as.numeric(ent$value), S * K_sp), nrow = S, ncol = K_sp)
          }
          if (is.null(vals)) stop(sprintf("%s not found in posterior draws.", nm), call. = FALSE)
          spliced_gpd_draws[[nm]] <- vals
        }
      }
    } else {
      thr_mode <- gpd_plan$threshold$mode %||% "constant"
      if (identical(thr_mode, "link")) {
        if (is.null(X)) stop("threshold link-mode requires X.", call. = FALSE)
        beta_thr <- .indexed_block(draw_mat, "beta_threshold", K = P)
        threshold_mat <- matrix(NA_real_, nrow = S, ncol = n_x)
        thr_link <- gpd_plan$threshold$link %||% "exp"
        thr_power <- gpd_plan$threshold$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(X, beta_thr[s, , drop = FALSE]))
          threshold_mat[s, ] <- as.numeric(.apply_link(eta, thr_link, thr_power))
        }
      } else {
        thr_cols <- grep("^threshold(\\b|_)", colnames(draw_mat), value = TRUE)
        if (length(thr_cols) == 0L && "threshold" %in% colnames(draw_mat)) thr_cols <- "threshold"
        if (length(thr_cols) == 0L) stop("threshold not found in posterior draws.", call. = FALSE)
        if (length(thr_cols) == 1L) {
          threshold_scalar <- as.numeric(draw_mat[, thr_cols])
        } else {
          threshold_scalar <- rowMeans(draw_mat[, thr_cols, drop = FALSE], na.rm = TRUE)
        }
      }

      has_beta_ts <- any(grepl("^beta_tail_scale\\[", colnames(draw_mat)))
      ts_mode <- gpd_plan$tail_scale$mode %||% if (has_beta_ts) "link" else "constant"
      if (identical(ts_mode, "link")) {
        if (is.null(X)) stop("tail_scale link-mode requires X.", call. = FALSE)
        beta_ts <- .indexed_block(draw_mat, "beta_tail_scale", K = P)
        tail_scale <- matrix(NA_real_, nrow = S, ncol = n_x)
        ts_link <- gpd_plan$tail_scale$link %||% "exp"
        ts_power <- gpd_plan$tail_scale$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(X, beta_ts[s, , drop = FALSE]))
          tail_scale[s, ] <- as.numeric(.apply_link(eta, ts_link, ts_power))
        }
      } else if ("tail_scale" %in% colnames(draw_mat)) {
        tail_scale <- as.numeric(draw_mat[, "tail_scale"])
      } else if (!is.null(gpd_plan$tail_scale$value)) {
        tail_scale <- rep(as.numeric(gpd_plan$tail_scale$value), S)
      } else {
        stop("tail_scale not found in posterior draws.", call. = FALSE)
      }

      has_beta_tsh <- any(grepl("^beta_tail_shape\\[", colnames(draw_mat)))
      tsh_mode <- gpd_plan$tail_shape$mode %||% if (has_beta_tsh) "link" else "constant"
      if (identical(tsh_mode, "link")) {
        if (is.null(X)) stop("tail_shape link-mode requires X.", call. = FALSE)
        beta_tsh <- .indexed_block(draw_mat, "beta_tail_shape", K = P)
        tail_shape <- matrix(NA_real_, nrow = S, ncol = n_x)
        tsh_link <- gpd_plan$tail_shape$link %||% "identity"
        tsh_power <- gpd_plan$tail_shape$link_power %||% NULL
        for (s in seq_len(S)) {
          eta <- as.numeric(tcrossprod(X, beta_tsh[s, , drop = FALSE]))
          tail_shape[s, ] <- as.numeric(.apply_link(eta, tsh_link, tsh_power))
        }
      } else if ("tail_shape" %in% colnames(draw_mat)) {
        tail_shape <- as.numeric(draw_mat[, "tail_shape"])
      } else if (!is.null(gpd_plan$tail_shape$value)) {
        tail_shape <- rep(as.numeric(gpd_plan$tail_shape$value), S)
      } else {
        stop("tail_shape not found in posterior draws.", call. = FALSE)
      }
    }
  }

  .threshold_at <- function(s, i) {
    if (!is.null(threshold_mat)) return(threshold_mat[s, i])
    threshold_scalar[s]
  }

  .tail_scale_at <- function(s, i) {
    if (is.matrix(tail_scale)) return(tail_scale[s, i])
    tail_scale[s]
  }

  .tail_shape_at <- function(s, i) {
    if (is.matrix(tail_shape)) return(tail_shape[s, i])
    tail_shape[s]
  }

  .build_args0_or_null <- function(s) {
    w_s <- as.numeric(W_draws[s, ])
    if (!all(is.finite(w_s))) return(NULL)
    args0 <- if (pred_backend == "sb") list(w = w_s) else list()
    for (nm in base_params) {
      v <- as.numeric(bulk_draws[[nm]][s, ])
      sup <- as.character((get_kernel_registry()[[meta$kernel %||% spec$kernel$key %||% "<unknown>"]] %||% list())$bulk_support[[nm]] %||% "")
      if (sup %in% c("positive_sd", "positive_scale", "positive_shape", "positive_location")) {
        if (!all(is.finite(v) & (v > 0))) return(NULL)
      } else if (!all(is.finite(v))) {
        return(NULL)
      }
      args0[[nm]] <- v
    }
    if (GPD && !is_spliced && !is.matrix(tail_shape)) {
      xi <- as.numeric(tail_shape[s])
      if (!is.finite(xi)) return(NULL)
      args0$tail_shape <- xi
    }
    args0
  }

  .support_ok <- function(nm, v) {
    sup <- as.character((get_kernel_registry()[[meta$kernel %||% spec$kernel$key %||% "<unknown>"]] %||% list())$bulk_support[[nm]] %||% "")
    if (sup %in% c("positive_sd", "positive_scale", "positive_shape", "positive_location")) {
      return(all(is.finite(v) & (v > 0)))
    }
    all(is.finite(v))
  }

  .draw_valid <- logical(S)
  for (s in seq_len(S)) {
    ok <- !is.null(.build_args0_or_null(s))
    if (ok && GPD) {
      if (is_spliced) {
        for (nm in names(spliced_gpd_draws)) {
          vals <- as.numeric(spliced_gpd_draws[[nm]][s, ])
          if (identical(nm, "tail_scale")) {
            ok <- all(is.finite(vals) & (vals > 0))
          } else {
            ok <- all(is.finite(vals))
          }
          if (!ok) break
        }
      } else {
        if (!is.null(threshold_mat)) {
          if (!all(is.finite(threshold_mat[s, ]))) ok <- FALSE
        } else {
          if (!is.finite(threshold_scalar[s])) ok <- FALSE
        }
        if (ok) {
          if (is.matrix(tail_scale)) {
            ok <- all(is.finite(tail_scale[s, ]) & (tail_scale[s, ] > 0))
          } else {
            ok <- is.finite(tail_scale[s]) && (tail_scale[s] > 0)
          }
        }
        if (ok) {
          if (is.matrix(tail_shape)) {
            ok <- all(is.finite(tail_shape[s, ]))
          } else {
            ok <- is.finite(tail_shape[s])
          }
        }
      }
    }
    .draw_valid[s] <- ok
  }

  spliced_scalar <- if (is_spliced && GPD) .get_dispatch_scalar(fit, backend_override = "crp") else NULL

  .row_component_value <- function(obj, i, k) {
    if (is.null(obj)) return(NA_real_)
    if (is.null(dim(obj))) return(as.numeric(obj[1L]))
    if (length(dim(obj)) != 2L) stop("Expected a scalar or matrix component block.", call. = FALSE)
    j <- if (ncol(obj) >= k && ncol(obj) > 1L) k else 1L
    as.numeric(obj[i, j])
  }

  .compute_spliced_gpd_eta <- function(s) {
    if (!length(spliced_gpd_link)) return(list())
    out <- list()
    for (nm in names(spliced_gpd_link)) {
      ent <- spliced_gpd_link[[nm]]
      beta_mat <- ent$beta[s, , , drop = TRUE]
      if (is.null(dim(beta_mat))) beta_mat <- matrix(beta_mat, nrow = ncol(W_draws), ncol = P)
      eta_mat <- tcrossprod(X, beta_mat)
      out[[nm]] <- .apply_link(eta_mat, ent$link, ent$link_power)
    }
    out
  }

  .spliced_gpd_value_at <- function(nm, s, i, k, eta_cache) {
    if (!is.null(eta_cache[[nm]])) return(.row_component_value(eta_cache[[nm]], i, k))
    vals <- spliced_gpd_draws[[nm]]
    if (is.null(vals)) return(NA_real_)
    as.numeric(vals[s, k])
  }

  .spliced_component_args_or_null <- function(s, i, k, link_eta, gpd_eta) {
    args <- list()
    for (nm in bulk_params) {
      if (nm %in% link_params) {
        vv <- .row_component_value(link_eta[[nm]], i, k)
      } else if (nm %in% base_params) {
        vv <- as.numeric(bulk_draws[[nm]][s, k])
      } else {
        next
      }
      if (!.support_ok(nm, vv)) return(NULL)
      args[[nm]] <- as.numeric(vv)
    }

    args$threshold <- .spliced_gpd_value_at("threshold", s, i, k, gpd_eta)
    args$tail_scale <- .spliced_gpd_value_at("tail_scale", s, i, k, gpd_eta)
    args$tail_shape <- .spliced_gpd_value_at("tail_shape", s, i, k, gpd_eta)
    if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
      return(NULL)
    }
    args
  }

  .spliced_component_args_list_or_null <- function(s, i, link_eta, gpd_eta) {
    K <- ncol(W_draws)
    out <- vector("list", K)
    for (k in seq_len(K)) {
      out[[k]] <- .spliced_component_args_or_null(s, i, k, link_eta = link_eta, gpd_eta = gpd_eta)
      if (is.null(out[[k]])) return(NULL)
    }
    out
  }

  .normalize_weights_or_null <- function(w) {
    w <- as.numeric(w)
    if (!all(is.finite(w))) return(NULL)
    w[w < 0] <- 0
    sw <- sum(w)
    if (!is.finite(sw) || sw <= 0) return(NULL)
    w / sw
  }

  .spliced_density_cdf_row <- function(s, i, yvals, quantity, link_eta, gpd_eta) {
    w_s <- .normalize_weights_or_null(W_draws[s, ])
    if (is.null(w_s)) return(rep(NA_real_, length(yvals)))
    comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
    if (is.null(comp_args)) return(rep(NA_real_, length(yvals)))

    out <- numeric(length(yvals))
    for (j in seq_along(yvals)) {
      yj <- yvals[j]
      vals <- vapply(seq_along(comp_args), function(k) {
        args <- comp_args[[k]]
        if (quantity == "density") {
          as.numeric(do.call(spliced_scalar$d, c(list(x = yj, log = 0L), args)))[1]
        } else {
          cdfv <- as.numeric(do.call(spliced_scalar$p, c(list(q = yj, lower.tail = 1L, log.p = 0L), args)))[1]
          pmin(pmax(cdfv, 0), 1)
        }
      }, numeric(1))
      out[j] <- sum(w_s * vals)
    }
    out
  }

  violations <- list(
    cdf_range = 0L,
    cdf_monotone = 0L,
    density_nonneg = 0L,
    continuity = 0L
  )

  details <- list(
    bad_draws = integer(0),
    examples = list()
  )

  for (s in idx) {
    if (!.draw_valid[s]) {
      violations$cdf_range <- violations$cdf_range + 1L
      details$bad_draws <- c(details$bad_draws, s)
      next
    }

    args0 <- .build_args0_or_null(s)
    if (is.null(args0)) {
      violations$cdf_range <- violations$cdf_range + 1L
      details$bad_draws <- c(details$bad_draws, s)
      next
    }
    link_eta <- .compute_link_eta(s)
    gpd_eta <- if (is_spliced && GPD) .compute_spliced_gpd_eta(s) else list()

    for (i in seq_len(n_x)) {
      if (is_spliced && GPD) {
        cdfv <- .spliced_density_cdf_row(s, i, grid, quantity = "cdf", link_eta = link_eta, gpd_eta = gpd_eta)
        dens <- .spliced_density_cdf_row(s, i, grid, quantity = "density", link_eta = link_eta, gpd_eta = gpd_eta)
      } else {
        args <- args0
        if (GPD) {
          args$threshold <- .threshold_at(s, i)
          args$tail_scale <- .tail_scale_at(s, i)
          args$tail_shape <- .tail_shape_at(s, i)
          if (!is.finite(args$threshold) || !is.finite(args$tail_scale) || args$tail_scale <= 0 || !is.finite(args$tail_shape)) {
            violations$cdf_range <- violations$cdf_range + 1L
            next
          }
        }
        if (length(link_params)) {
          bad <- FALSE
          for (nm in link_params) {
            vv <- as.numeric(link_eta[[nm]][i, ])
            if (!all(is.finite(vv))) {
              bad <- TRUE
              break
            }
            args[[nm]] <- vv
          }
          if (bad) {
            violations$cdf_range <- violations$cdf_range + 1L
            next
          }
        }
        cdfv <- as.numeric(do.call(p_fun, c(list(q = grid, lower.tail = 1L, log.p = 0L), args)))
        dens <- as.numeric(do.call(d_fun, c(list(x = grid, log = 0L), args)))
      }

      if (any(!is.finite(cdfv)) || min(cdfv, na.rm = TRUE) < -tol || max(cdfv, na.rm = TRUE) > 1 + tol) {
        violations$cdf_range <- violations$cdf_range + 1L
        if (length(details$examples) < 5L) details$examples <- c(details$examples, list(list(draw = s, row = i, check = "cdf_range")))
      }

      if (any(diff(cdfv) < -tol, na.rm = TRUE)) {
        violations$cdf_monotone <- violations$cdf_monotone + 1L
        if (length(details$examples) < 5L) details$examples <- c(details$examples, list(list(draw = s, row = i, check = "cdf_monotone")))
      }

      if (any(!is.finite(dens)) || min(dens, na.rm = TRUE) < -tol) {
        violations$density_nonneg <- violations$density_nonneg + 1L
        if (length(details$examples) < 5L) details$examples <- c(details$examples, list(list(draw = s, row = i, check = "density_nonneg")))
      }

      if (check_continuity && GPD) {
        if (is_spliced) {
          comp_args <- .spliced_component_args_list_or_null(s, i, link_eta = link_eta, gpd_eta = gpd_eta)
          if (!is.null(comp_args)) {
            u_vals <- unique(vapply(comp_args, function(args) as.numeric(args$threshold), numeric(1)))
            u_vals <- u_vals[is.finite(u_vals)]
            bad_continuity <- FALSE
            for (u in u_vals) {
              lr <- .spliced_density_cdf_row(s, i, c(u - eps, u + eps), quantity = "cdf", link_eta = link_eta, gpd_eta = gpd_eta)
              if (all(is.finite(lr)) && abs(lr[2] - lr[1]) > 1e-4) {
                bad_continuity <- TRUE
                break
              }
            }
            if (bad_continuity) {
              violations$continuity <- violations$continuity + 1L
              if (length(details$examples) < 5L) details$examples <- c(details$examples, list(list(draw = s, row = i, check = "continuity")))
            }
          }
        } else {
          u <- as.numeric(args$threshold)
          if (is.finite(u)) {
            left <- as.numeric(do.call(p_fun, c(list(q = u - eps, lower.tail = 1L, log.p = 0L), args)))
            right <- as.numeric(do.call(p_fun, c(list(q = u + eps, lower.tail = 1L, log.p = 0L), args)))
            if (is.finite(left) && is.finite(right) && abs(right - left) > 1e-4) {
              violations$continuity <- violations$continuity + 1L
              if (length(details$examples) < 5L) details$examples <- c(details$examples, list(list(draw = s, row = i, check = "continuity")))
            }
          }
        }
      }
    }
  }

  pass <- list(
    cdf_range = (violations$cdf_range == 0L),
    cdf_monotone = (violations$cdf_monotone == 0L),
    density_nonneg = (violations$density_nonneg == 0L),
    continuity = if (GPD && check_continuity) (violations$continuity == 0L) else NA
  )

  list(
    pass = pass,
    violations = violations,
    n_checked_draws = length(idx),
    n_x = n_x,
    grid_n = G,
    details = details
  )
}

#' Normal mixture distribution
#'
#' Finite mixture of normal components for real-valued bulk modeling. This topic provides the
#' scalar mixture density, CDF, RNG, and quantile functions used by the bulk-only normal kernel in
#' the package registry.
#'
#' With weights \eqn{w_k}, means \eqn{\mu_k}, and standard deviations \eqn{\sigma_k}, the mixture
#' density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k \phi(x \mid \mu_k, \sigma_k^2),
#' }
#' where \eqn{\tilde{w}_k = w_k / \sum_j w_j}. These uppercase functions are scalar
#' NIMBLE-compatible building blocks. For vectorized R usage, use [normal_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param mean,sd Numeric vectors of length \eqn{K} giving component means and standard deviations.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qNormMix()` returns a numeric vector
#'   with the same length as `p`.
#'
#' @details
#' If \eqn{F_k} denotes the \eqn{k}-th component CDF, then the mixture distribution function is
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k F_k(x)
#' = \sum_{k=1}^K \tilde{w}_k \Phi\left(\frac{x-\mu_k}{\sigma_k}\right).
#' }
#' Random generation first draws a component index with probability \eqn{\tilde{w}_k} and then
#' generates from the corresponding normal law. The quantile function has no closed form for a
#' general finite mixture, so \code{qNormMix()} solves \eqn{F(x)=p} numerically by bracketing the
#' root and applying \code{stats::uniroot()}.
#'
#' The mixture mean is
#' \deqn{
#' E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k,
#' }
#' which is the analytical mean used by the package when a normal-mixture draw contributes to a
#' posterior predictive mean calculation.
#'
#' @seealso [normal_mixgpd()], [normal_gpd()], [normal_lowercase()], [build_nimble_bundle()],
#'   [kernel_support_table()].
#' @family normal kernel families
#'
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' mean <- c(-1, 0.5, 2.0)
#' sd <- c(1.0, 0.7, 1.3)
#'
#' dNormMix(0.5, w = w, mean = mean, sd = sd, log = FALSE)
#' pNormMix(0.5, w = w, mean = mean, sd = sd,
#'         lower.tail = TRUE, log.p = FALSE)
#' qNormMix(0.50, w = w, mean = mean, sd = sd)
#' qNormMix(0.95, w = w, mean = mean, sd = sd)
#' replicate(10, rNormMix(1, w = w, mean = mean, sd = sd))
#' @rdname normal_mix
#' @name normal_mix
#' @aliases dNormMix pNormMix rNormMix qNormMix
#' @importFrom stats uniroot pnorm dnorm runif qnorm rnorm
NULL

#' @describeIn normal_mix Normal mixture density
#' @export
dNormMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    K <- length(w)

    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) {
      if (log == 1) return(log(eps)) else return(eps)
    }

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + (w[j] / wsum) * dnorm(x, mean[j], sd[j], 0)
    }

    if (s0 < eps) s0 <- eps
    if (log == 1) return(log(s0)) else return(s0)
  }
)

#' @describeIn normal_mix Normal mixture distribution function
#' @export
pNormMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    K <- length(w)

    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) {
      cdf0 <- 0.0
      if (lower.tail == 0) cdf0 <- 1.0
      if (log.p == 1) return(log(max(cdf0, eps))) else return(cdf0)
    }

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + (w[j] / wsum) * pnorm(q, mean[j], sd[j], 1, 0)
    }

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(max(cdf, eps))) else return(cdf)
  }
)

#' @describeIn normal_mix Normal mixture random generation
#' @export
rNormMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    K <- length(w)
    wsum <- 0.0
    for (j in 1:K) wsum <- wsum + w[j]
    if (wsum <= 0.0) return(0.0)

    u <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0
    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (u <= cw) {
          idx <- j
          found <- 1
        }
      }
    }
    return(rnorm(1, mean[idx], sd[idx]))
  }
)

#' @describeIn normal_mix Normal mixture quantile function
#' @export
qNormMix <- function(p, w, mean, sd,
                     lower.tail = TRUE, log.p = FALSE,
                     tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- -Inf; next }
    if (pi >= 1) { out[i] <- Inf; next }

    lo <- min(stats::qnorm(pi, mean = mean, sd = sd), na.rm = TRUE)
    hi <- max(stats::qnorm(pi, mean = mean, sd = sd), na.rm = TRUE)
    if (!is.finite(lo)) lo <- -1e20
    if (!is.finite(hi)) hi <- 1e20
    if (lo == hi) {
      out[i] <- lo
    } else {
      f_lo <- as.numeric(pNormMix(lo, w = w, mean = mean, sd = sd, lower.tail = 1, log.p = 0) - pi)
      f_hi <- as.numeric(pNormMix(hi, w = w, mean = mean, sd = sd, lower.tail = 1, log.p = 0) - pi)
      iter <- 0L
      while (is.finite(f_lo) && f_lo > 0 && lo > -1e20 && iter < 60L) {
        step <- max(1, abs(lo))
        lo <- lo - step
        f_lo <- as.numeric(pNormMix(lo, w = w, mean = mean, sd = sd, lower.tail = 1, log.p = 0) - pi)
        iter <- iter + 1L
      }
      iter <- 0L
      while (is.finite(f_hi) && f_hi < 0 && hi < 1e20 && iter < 60L) {
        step <- max(1, abs(hi))
        hi <- hi + step
        f_hi <- as.numeric(pNormMix(hi, w = w, mean = mean, sd = sd, lower.tail = 1, log.p = 0) - pi)
        iter <- iter + 1L
      }
      if (!is.finite(lo) || !is.finite(hi) || lo >= hi || !is.finite(f_lo) || !is.finite(f_hi) || f_lo * f_hi > 0) {
        out[i] <- NA_real_
      } else {
        out[i] <- stats::uniroot(
          function(z) pNormMix(z, w = w, mean = mean, sd = sd, lower.tail = 1, log.p = 0) - pi,
          interval = c(lo, hi),
          tol = tol, maxiter = maxiter
        )$root
      }
    }
  }
  out
}

meanNormMix <- function(w, mean, sd) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(mean)
  sig <- as.numeric(sd)
  if (length(mu) != length(ww) || length(sig) != length(ww)) return(NA_real_)
  if (any(!is.finite(mu)) || any(!is.finite(sig)) || any(sig <= 0)) return(NA_real_)
  sum(ww * mu)
}

meanNormMixTrunc <- function(w, mean, sd, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(mean)
  sig <- as.numeric(sd)
  u <- as.numeric(threshold)[1]
  if (length(mu) != length(ww) || length(sig) != length(ww)) return(NA_real_)
  if (is.na(u) || any(!is.finite(mu)) || any(!is.finite(sig)) || any(sig <= 0)) return(NA_real_)
  a <- (u - mu) / sig
  sum(ww * (mu * stats::pnorm(a) - sig * stats::dnorm(a)))
}

# -------------------------------
# Normal mixture + GPD tail
# -------------------------------

#' Normal mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to a normal mixture bulk.
#' This matches the structure used by the package's normal `mixgpd` kernels.
#'
#' If \eqn{F_{mix}} denotes the normal-mixture CDF, the spliced CDF is
#' \deqn{
#' F(x) =
#' \left\{
#' \begin{array}{ll}
#' F_{mix}(x), & x < u, \\
#' F_{mix}(u) + \{1 - F_{mix}(u)\} G(x), & x \ge u,
#' \end{array}
#' \right.
#' }
#' where `threshold = u` and \eqn{G} is the GPD CDF.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}.
#' @param mean,sd Numeric vectors of length \eqn{K} giving component means and standard deviations.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qNormMixGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' The construction keeps the normal mixture unchanged below the threshold \eqn{u} and replaces the
#' upper tail by a generalized Pareto exceedance model. Writing \eqn{F_{mix}(u)=p_u}, the spliced
#' density is
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' f_{mix}(x), & x < u, \\
#' \{1-p_u\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' This formulation preserves total probability because the GPD is attached only to the residual
#' survival mass above the bulk threshold.
#'
#' The quantile is piecewise. If \eqn{p \le p_u}, \code{qNormMixGpd()} inverts the bulk mixture
#' CDF; otherwise it rescales the tail probability to \eqn{(p-p_u)/(1-p_u)} and applies the
#' closed-form GPD quantile. That same piecewise logic is what the fitted-model prediction code
#' uses draw by draw.
#'
#' @seealso [normal_mix()], [normal_gpd()], [gpd()], [normal_lowercase()], [dpmgpd()].
#' @family normal kernel families
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' mean <- c(-1, 0.5, 2.0)
#' sd <- c(1.0, 0.7, 1.3)
#' threshold <- 2
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dNormMixGpd(3.0, w, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)
#' pNormMixGpd(3.0, w, mean, sd, threshold, tail_scale, tail_shape,
#'            lower.tail = TRUE, log.p = FALSE)
#' qNormMixGpd(0.50, w, mean, sd, threshold, tail_scale, tail_shape)
#' qNormMixGpd(0.95, w, mean, sd, threshold, tail_scale, tail_shape)
#' replicate(10, rNormMixGpd(1, w, mean, sd, threshold, tail_scale, tail_shape))
#' @rdname normal_mixgpd
#' @name normal_mixgpd
#' @aliases dNormMixGpd pNormMixGpd rNormMixGpd qNormMixGpd
#' @importFrom stats uniroot pnorm dnorm runif qnorm rnorm
NULL

#' @describeIn normal_mixgpd Normal mixture + GPD tail density
#' @export
dNormMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) return(dNormMix(x, w, mean, sd, log))

    Fu <- pNormMix(threshold, w, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn normal_mixgpd Normal mixture + GPD tail distribution function
#' @export
pNormMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) return(pNormMix(q, w, mean, sd, lower.tail, log.p))

    Fu <- pNormMix(threshold, w, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(max(cdf, eps))) else return(cdf)
  }
)

#' @describeIn normal_mixgpd Normal mixture + GPD tail random generation
#' @export
rNormMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 mean = double(1),
                 sd = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pNormMix(threshold, w, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rNormMix(1, w, mean, sd))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn normal_mixgpd Normal mixture + GPD tail quantile function
#' @export
qNormMixGpd <- function(p, w, mean, sd, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE,
                        tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- pNormMix(threshold, w, mean, sd, 1, 0)
  out <- numeric(length(p))

  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qNormMix(pi, w, mean, sd, lower.tail = TRUE, log.p = FALSE, tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}

# -------------------------------
# Single Normal + GPD tail
# -------------------------------

#' Normal with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' normal bulk component.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param mean Numeric scalar mean parameter for the Normal bulk.
#' @param sd Numeric scalar standard deviation for the Normal bulk.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qNormGpd()` returns a numeric
#'   vector with the same length as `p`.
#'
#' @details
#' This is the single-component version of [normal_mixgpd()]. If \eqn{\Phi_u =
#' \Phi((u-\mu)/\sigma)} denotes the normal bulk probability below the threshold, then the density
#' is
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' \phi(x \mid \mu, \sigma^2), & x < u, \\
#' (1-\Phi_u) g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' The distribution is continuous at \eqn{u} by construction, although the derivative generally
#' changes there because the tail is modeled by a different family.
#'
#' The ordinary mean exists only when the GPD tail satisfies \eqn{\xi < 1}. When that condition
#' fails, downstream mean prediction is intentionally blocked and the package directs the user to
#' restricted means or quantile-based summaries instead.
#'
#' @seealso [normal_mix()], [normal_mixgpd()], [gpd()], [normal_lowercase()].
#' @family normal kernel families
#' @examples
#' mean <- 0.5
#' sd <- 1.0
#' threshold <- 2
#' tail_scale <- 1.0
#' tail_shape <- 0.2
#'
#' dNormGpd(3.0, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)
#' pNormGpd(3.0, mean, sd, threshold, tail_scale, tail_shape,
#'         lower.tail = TRUE, log.p = FALSE)
#' qNormGpd(0.50, mean, sd, threshold, tail_scale, tail_shape)
#' qNormGpd(0.95, mean, sd, threshold, tail_scale, tail_shape)
#' replicate(10, rNormGpd(1, mean, sd, threshold, tail_scale, tail_shape))
#' @rdname normal_gpd
#' @name normal_gpd
#' @aliases dNormGpd pNormGpd rNormGpd qNormGpd
#' @importFrom stats uniroot pnorm dnorm runif qnorm rnorm
NULL

#' @describeIn normal_gpd Normal + GPD tail density
#' @export
dNormGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 mean = double(0),
                 sd = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) {
      dens <- dnorm(x, mean, sd, 0)
      if (dens < eps) dens <- eps
      if (log == 1) return(log(dens)) else return(dens)
    }

    Fu <- pnorm(threshold, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn normal_gpd Normal + GPD tail distribution function
#' @export
pNormGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 mean = double(0),
                 sd = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) {
      cdf0 <- pnorm(q, mean, sd, lower.tail, log.p)
      return(cdf0)
    }

    Fu <- pnorm(threshold, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G

    if (is.nan(cdf)) cdf <- 0.0
    if (cdf < 0.0) cdf <- 0.0
    if (cdf > 1.0) cdf <- 1.0
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) return(log(max(cdf, eps))) else return(cdf)
  }
)

#' @describeIn normal_gpd Normal + GPD tail random generation
#' @export
rNormGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 mean = double(0),
                 sd = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pnorm(threshold, mean, sd, 1, 0)
    if (Fu < 0.0) Fu <- 0.0
    if (Fu > 1.0) Fu <- 1.0

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rnorm(1, mean, sd))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn normal_gpd Normal + GPD tail quantile function
#' @export
qNormGpd <- function(p, mean, sd, threshold, tail_scale, tail_shape,
                     lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- stats::pnorm(threshold, mean = mean, sd = sd, lower.tail = TRUE, log.p = FALSE)
  out <- numeric(length(p))

  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- stats::qnorm(pi, mean = mean, sd = sd, lower.tail = TRUE, log.p = FALSE)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


# ==========================================================
# Lowercase vectorized R wrappers for Normal kernels
# ==========================================================

#' Lowercase vectorized normal distribution functions
#'
#' Vectorized R wrappers for the scalar normal-kernel topics in this file. These helpers are meant
#' for interactive use and examples rather than direct use inside NIMBLE code.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param mean,sd Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These wrappers vectorize the scalar normal-kernel routines for ordinary R use. They preserve the
#' same formulas, parameter meanings, and tail construction as the uppercase functions; the only
#' change is that \code{x}, \code{q}, \code{p}, and \code{n} may now be length greater than one.
#'
#' For the mixture quantile and splice quantile functions, the numerical and piecewise logic is
#' delegated directly to the corresponding scalar routine. As a result, the lowercase helpers are
#' faithful front ends rather than separate implementations.
#'
#' @seealso [normal_mix()], [normal_mixgpd()], [normal_gpd()], [bundle()], [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.25, 0.15)
#' mu <- c(-1, 0.5, 2)
#' sig <- c(1, 0.7, 1.3)
#'
#' # Normal mixture
#' dnormmix(c(0, 1, 2), w = w, mean = mu, sd = sig)
#' rnormmix(5, w = w, mean = mu, sd = sig)
#'
#' # Normal mixture + GPD
#' dnormmixgpd(c(1, 2, 3), w = w, mean = mu, sd = sig,
#'             threshold = 2, tail_scale = 1, tail_shape = 0.2)
#'
#' # Normal + GPD (single component)
#' dnormgpd(c(1, 2, 3), mean = 0.5, sd = 1, threshold = 2,
#'          tail_scale = 1, tail_shape = 0.2)
#'
#' @name normal_lowercase
#' @rdname normal_lowercase
NULL

# ---- Normal Mix lowercase wrappers ----

#' @describeIn normal_lowercase Normal mixture density (vectorized)
#' @export
dnormmix <- function(x, w, mean, sd, log = FALSE) {
  x <- as.numeric(x)

  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dNormMix(xi, w = w, mean = mean, sd = sd, log = log_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal mixture distribution function (vectorized)
#' @export
pnormmix <- function(q, w, mean, sd, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pNormMix(qi, w = w, mean = mean, sd = sd,
                                              lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal mixture quantile function (vectorized)
#' @export
qnormmix <- function(p, w, mean, sd, lower.tail = TRUE, log.p = FALSE,
                     tol = 1e-10, maxiter = 200) {
  qNormMix(p, w = w, mean = mean, sd = sd, lower.tail = lower.tail,
           log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn normal_lowercase Normal mixture random generation (vectorized)
#' @export
rnormmix <- function(n, w, mean, sd) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rNormMix(1L, w = w, mean = mean, sd = sd)),
         numeric(1L))
}

# ---- Normal Mix + GPD lowercase wrappers ----

#' @describeIn normal_lowercase Normal mixture + GPD density (vectorized)
#' @export
dnormmixgpd <- function(x, w, mean, sd, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dNormMixGpd(xi, w = w, mean = mean, sd = sd,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal mixture + GPD distribution function (vectorized)
#' @export
pnormmixgpd <- function(q, w, mean, sd, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pNormMixGpd(qi, w = w, mean = mean, sd = sd,
                                                 threshold = threshold, tail_scale = tail_scale,
                                                 tail_shape = tail_shape,
                                                 lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal mixture + GPD quantile function (vectorized)
#' @export
qnormmixgpd <- function(p, w, mean, sd, threshold, tail_scale, tail_shape,
                        lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qNormMixGpd(p, w = w, mean = mean, sd = sd, threshold = threshold,
              tail_scale = tail_scale, tail_shape = tail_shape,
              lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn normal_lowercase Normal mixture + GPD random generation (vectorized)
#' @export
rnormmixgpd <- function(n, w, mean, sd, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rNormMixGpd(1L, w = w, mean = mean, sd = sd,
                                                         threshold = threshold, tail_scale = tail_scale,
                                                         tail_shape = tail_shape)),
         numeric(1L))
}

# ---- Normal + GPD lowercase wrappers ----

#' @describeIn normal_lowercase Normal + GPD density (vectorized)
#' @export
dnormgpd <- function(x, mean, sd, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dNormGpd(xi, mean = mean, sd = sd,
                                              threshold = threshold, tail_scale = tail_scale,
                                              tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal + GPD distribution function (vectorized)
#' @export
pnormgpd <- function(q, mean, sd, threshold, tail_scale, tail_shape,
                     lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pNormGpd(qi, mean = mean, sd = sd,
                                              threshold = threshold, tail_scale = tail_scale,
                                              tail_shape = tail_shape,
                                              lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn normal_lowercase Normal + GPD quantile function (vectorized)
#' @export
qnormgpd <- function(p, mean, sd, threshold, tail_scale, tail_shape,
                     lower.tail = TRUE, log.p = FALSE) {
  qNormGpd(p, mean = mean, sd = sd, threshold = threshold,
           tail_scale = tail_scale, tail_shape = tail_shape,
           lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn normal_lowercase Normal + GPD random generation (vectorized)
#' @export
rnormgpd <- function(n, mean, sd, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rNormGpd(1L, mean = mean, sd = sd,
                                                      threshold = threshold, tail_scale = tail_scale,
                                                      tail_shape = tail_shape)),
         numeric(1L))
}

## Copy package namespace bindings into the test environment so tests can call
## unexported helpers (replaces eval-parse into .GlobalEnv; CRAN policy).
ns <- asNamespace("CausalMixGPD")
for (nm in ls(ns, all.names = TRUE)) {
  if (!nzchar(nm) || startsWith(nm, ".__")) {
    next
  }
  if (!exists(nm, envir = ns, inherits = FALSE)) {
    next
  }
  assign(nm, get(nm, envir = ns, inherits = FALSE))
}

#' Lognormal mixture distribution
#'
#' Finite mixture of lognormal components for positive-support bulk modeling. The scalar functions
#' in this topic are the NIMBLE-compatible building blocks for the lognormal bulk kernel family.
#'
#' The mixture density is
#' \deqn{
#' f(x) = \sum_{k = 1}^K \tilde{w}_k f_{LN}(x \mid \mu_k, \sigma_k),
#' \qquad x > 0,
#' }
#' with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use [lognormal_lowercase()].
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
#'   internally when needed.
#' @param meanlog,sdlog Numeric vectors of length \eqn{K} giving component log-means and log-standard deviations.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot}.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Density/CDF/RNG functions return numeric scalars. `qLognormalMix()` returns a numeric
#'   vector with the same length as `p`.
#'
#' @details
#' Each component satisfies \eqn{\log X_k \sim N(\mu_k,\sigma_k^2)}, so the mixture CDF is
#' \deqn{
#' F(x) = \sum_{k=1}^K \tilde{w}_k
#' \Phi\left(\frac{\log x-\mu_k}{\sigma_k}\right), \qquad x>0.
#' }
#' Random generation proceeds by drawing a component index with probability \eqn{\tilde{w}_k} and
#' then sampling from the corresponding lognormal law. Because a finite mixture of lognormals does
#' not admit a closed-form inverse CDF, \code{qLognormalMix()} computes quantiles by numerical
#' inversion.
#'
#' The analytical mixture mean is
#' \deqn{
#' E(X) = \sum_{k=1}^K \tilde{w}_k \exp(\mu_k + \sigma_k^2/2),
#' }
#' which is the expression used by the package whenever an ordinary predictive mean exists.
#'
#' @seealso [lognormal_mixgpd()], [lognormal_gpd()], [lognormal_lowercase()],
#'   [build_nimble_bundle()], [kernel_support_table()].
#' @family lognormal kernel families
#'
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' meanlog <- c(-0.2, 0.6, 1.2)
#' sdlog <- c(0.4, 0.3, 0.5)
#'
#' dLognormalMix(2.0, w = w, meanlog = meanlog, sdlog = sdlog, log = FALSE)
#' pLognormalMix(2.0, w = w, meanlog = meanlog, sdlog = sdlog,
#'              lower.tail = TRUE, log.p = FALSE)
#' qLognormalMix(0.50, w = w, meanlog = meanlog, sdlog = sdlog)
#' qLognormalMix(0.95, w = w, meanlog = meanlog, sdlog = sdlog)
#' replicate(10, rLognormalMix(1, w = w, meanlog = meanlog, sdlog = sdlog))
#' @rdname lognormal_mix
#' @name lognormal_mix
#' @aliases dLognormalMix pLognormalMix rLognormalMix qLognormalMix
#' @importFrom stats dlnorm plnorm rlnorm qlnorm runif uniroot
NULL

#' @describeIn lognormal_mix Lognormal mixture density
#' @export
dLognormalMix <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0) {
      if (log == 1) return(-Inf) else return(0.0)
    }

    s0 <- 0.0
    for (j in 1:K) {
      s0 <- s0 + (w[j] / wsum) * dlnorm(x, meanlog[j], sdlog[j], 0)
    }
    if (s0 < eps) s0 <- eps
    if (log == 1) return(log(s0)) else return(s0)
  }
)

#' @describeIn lognormal_mix Lognormal mixture distribution function
#' @export
pLognormalMix <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0) {
      if (log.p == 1) return(-Inf) else return(0.0)
    }

    cdf <- 0.0
    for (j in 1:K) {
      cdf <- cdf + (w[j] / wsum) * plnorm(q, meanlog[j], sdlog[j], 1, 0)
    }
    cdf <- max(min(cdf, 1.0), 0.0)
    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)

#' @describeIn lognormal_mix Lognormal mixture random generation
#' @export
rLognormalMix <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1)) {
    returnType(double(0))

    if (n != 1) return(0.0)
    K <- length(w)
    wsum <- sum(w)
    if (wsum <= 0) return(0.0)

    thresholdU <- runif(1, 0.0, wsum)
    cw <- 0.0
    idx <- 1
    found <- 0
    for (j in 1:K) {
      cw <- cw + w[j]
      if (found == 0) {
        if (thresholdU <= cw) {
          idx <- j
          found <- 1
        }
      }
    }
    return(rlnorm(1, meanlog[idx], sdlog[idx]))
  }
)

#' @describeIn lognormal_mix Lognormal mixture quantile function
#' @export
qLognormalMix <- function(p, w, meanlog, sdlog,
                          lower.tail = TRUE, log.p = FALSE,
                          tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= 0) { out[i] <- 0; next }
    if (pi >= 1) { out[i] <- Inf; next }

    hi <- max(stats::qlnorm(pi, meanlog = meanlog, sdlog = sdlog), na.rm = TRUE)
    if (!is.finite(hi) || hi <= 0) hi <- 1
    f0 <- as.numeric(pLognormalMix(0, w = w, meanlog = meanlog, sdlog = sdlog, 1, 0) - pi)
    fhi <- as.numeric(pLognormalMix(hi, w = w, meanlog = meanlog, sdlog = sdlog, 1, 0) - pi)
    iter <- 0L
    while (is.finite(fhi) && f0 * fhi > 0 && hi < 1e20 && iter < 60L) {
      hi <- hi * 2
      fhi <- as.numeric(pLognormalMix(hi, w = w, meanlog = meanlog, sdlog = sdlog, 1, 0) - pi)
      iter <- iter + 1L
    }
    if (!is.finite(hi) || hi <= 0 || !is.finite(fhi) || f0 * fhi > 0) {
      out[i] <- Inf
    } else {
      out[i] <- stats::uniroot(
        function(z) as.numeric(pLognormalMix(z, w = w, meanlog = meanlog, sdlog = sdlog, 1, 0)) - pi,
        interval = c(0, hi),
        tol = tol, maxiter = maxiter
      )$root
    }
  }
  out
}

meanLognormalMix <- function(w, meanlog, sdlog) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(meanlog)
  sig <- as.numeric(sdlog)
  if (length(mu) != length(ww) || length(sig) != length(ww)) return(NA_real_)
  if (any(!is.finite(mu)) || any(!is.finite(sig)) || any(sig <= 0)) return(NA_real_)
  sum(ww * exp(mu + 0.5 * sig^2))
}

meanLognormalMixTrunc <- function(w, meanlog, sdlog, threshold) {
  ww <- .mean_norm_mix_weights(w)
  if (is.null(ww)) return(NA_real_)
  mu <- as.numeric(meanlog)
  sig <- as.numeric(sdlog)
  u <- as.numeric(threshold)[1]
  if (length(mu) != length(ww) || length(sig) != length(ww)) return(NA_real_)
  if (any(!is.finite(mu)) || any(!is.finite(sig)) || any(sig <= 0) || is.na(u)) return(NA_real_)
  if (u <= 0) return(0)
  z <- (log(u) - mu - sig^2) / sig
  sum(ww * exp(mu + 0.5 * sig^2) * stats::pnorm(z))
}

#' Lognormal mixture with a GPD tail
#'
#' Spliced bulk-tail family formed by attaching a generalized Pareto tail to a lognormal mixture
#' bulk.
#' Let \eqn{F_{mix}} be the Lognormal mixture CDF. The spliced CDF is
#' \eqn{F(x)=F_{mix}(x)} for \eqn{x<threshold} and
#' \eqn{F(x)=F_{mix}(threshold) + \{1-F_{mix}(threshold)\}G(x)} for \eqn{x\ge threshold}, where \eqn{G}
#' is the GPD CDF for exceedances above \code{threshold}.
#'
#' The density, CDF, and RNG are implemented as \code{nimbleFunction}s. The quantile is an R function:
#' it uses numerical inversion in the bulk region and the closed-form GPD quantile in the tail.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param w Numeric vector of mixture weights of length \eqn{K}.
#' @param meanlog,sdlog Numeric vectors of length \eqn{K} giving component log-means and log-standard deviations.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#' @param tol Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.
#' @param maxiter Integer maximum number of iterations for \code{stats::uniroot}.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars.
#'   `qLognormalMixGpd()` returns a numeric vector with the same length as `p`.
#'
#' @details
#' Let \eqn{F_{mix}} be the lognormal-mixture CDF and let \eqn{u} denote the threshold. The splice
#' uses the bulk law below \eqn{u} and attaches a GPD to the residual survival mass above
#' \eqn{u}. The density therefore becomes
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' f_{mix}(x), & x < u, \\
#' \{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' The quantile is computed piecewise: bulk quantiles are obtained numerically from the mixture CDF,
#' whereas tail quantiles use the closed-form GPD inverse after rescaling the upper-tail
#' probability.
#'
#' @seealso [lognormal_mix()], [lognormal_gpd()], [gpd()], [lognormal_lowercase()], [dpmgpd()].
#' @family lognormal kernel families
#'
#' @examples
#' w <- c(0.60, 0.25, 0.15)
#' meanlog <- c(-0.2, 0.6, 1.2)
#' sdlog <- c(0.4, 0.3, 0.5)
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dLognormalMixGpd(4.0, w = w, meanlog = meanlog, sdlog = sdlog,
#'                 threshold = threshold, tail_scale = tail_scale,
#'                 tail_shape = tail_shape, log = FALSE)
#' pLognormalMixGpd(4.0, w = w, meanlog = meanlog, sdlog = sdlog,
#'                 threshold = threshold, tail_scale = tail_scale,
#'                 tail_shape = tail_shape, lower.tail = TRUE, log.p = FALSE)
#' qLognormalMixGpd(0.50, w = w, meanlog = meanlog, sdlog = sdlog,
#'                 threshold = threshold, tail_scale = tail_scale,
#'                 tail_shape = tail_shape)
#' qLognormalMixGpd(0.95, w = w, meanlog = meanlog, sdlog = sdlog,
#'                 threshold = threshold, tail_scale = tail_scale,
#'                 tail_shape = tail_shape)
#' replicate(10, rLognormalMixGpd(1, w = w, meanlog = meanlog, sdlog = sdlog,
#'                               threshold = threshold,
#'                               tail_scale = tail_scale,
#'                               tail_shape = tail_shape))
#' @rdname lognormal_mixgpd
#' @name lognormal_mixgpd
#' @aliases dLognormalMixGpd pLognormalMixGpd rLognormalMixGpd qLognormalMixGpd
NULL

#' @describeIn lognormal_mixgpd Lognormal mixture + GPD tail density
#' @export
dLognormalMixGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) return(dLognormalMix(x, w, meanlog, sdlog, log))

    Fu <- pLognormalMix(threshold, w, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn lognormal_mixgpd Lognormal mixture + GPD tail distribution function
#' @export
pLognormalMixGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) return(pLognormalMix(q, w, meanlog, sdlog, lower.tail, log.p))

    Fu <- pLognormalMix(threshold, w, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G
    cdf <- max(min(cdf, 1.0), 0.0)

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)

#' @describeIn lognormal_mixgpd Lognormal mixture + GPD tail random generation
#' @export
rLognormalMixGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 w = double(1),
                 meanlog = double(1),
                 sdlog = double(1),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- pLognormalMix(threshold, w, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rLognormalMix(1, w, meanlog, sdlog))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn lognormal_mixgpd Lognormal mixture + GPD tail quantile function
#' @export
qLognormalMixGpd <- function(p, w, meanlog, sdlog, threshold, tail_scale, tail_shape,
                             lower.tail = TRUE, log.p = FALSE,
                             tol = 1e-10, maxiter = 200) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- as.numeric(pLognormalMix(threshold, w = w, meanlog = meanlog, sdlog = sdlog, 1, 0))
  Fu <- max(min(Fu, 1.0), 0.0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- qLognormalMix(pi, w, meanlog, sdlog,
                              lower.tail = TRUE, log.p = FALSE,
                              tol = tol, maxiter = maxiter)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}




#' Lognormal with a GPD tail
#'
#' Spliced family obtained by attaching a generalized Pareto tail above `threshold` to a single
#' lognormal bulk.
#'
#' @param x Numeric scalar giving the point at which the density is evaluated.
#' @param q Numeric scalar giving the point at which the distribution function is evaluated.
#' @param p Numeric scalar probability in \eqn{(0,1)} for the quantile function.
#' @param n Integer giving the number of draws. The RNG implementation supports \code{n = 1}.
#' @param meanlog Numeric scalar log-mean parameter for the Lognormal bulk.
#' @param sdlog Numeric scalar log-standard deviation for the Lognormal bulk.
#' @param threshold Numeric scalar threshold at which the GPD tail is attached.
#' @param tail_scale Numeric scalar GPD scale parameter; must be positive.
#' @param tail_shape Numeric scalar GPD shape parameter.
#' @param log Integer flag \code{0/1}; if \code{1}, return the log-density.
#' @param lower.tail Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.
#' @param log.p Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.
#'
#' @return Spliced density/CDF/RNG functions return numeric scalars. `qLognormalGpd()` returns a
#'   numeric vector with the same length as `p`.
#'
#' @details
#' This is the single-lognormal counterpart of [lognormal_mixgpd()]. If \eqn{F_{LN}(u)} denotes the
#' bulk probability below the threshold, then the spliced density is
#' \deqn{
#' f(x) =
#' \left\{
#' \begin{array}{ll}
#' f_{LN}(x \mid \mu,\sigma), & x < u, \\
#' \{1-F_{LN}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
#' \end{array}
#' \right.
#' }
#' The ordinary mean is finite only when the GPD tail has \eqn{\xi < 1};
#' otherwise the package requires restricted means or quantiles.
#'
#' @seealso [lognormal_mix()], [lognormal_mixgpd()], [gpd()], [lognormal_lowercase()].
#' @family lognormal kernel families
#'
#' @examples
#' meanlog <- 0.4
#' sdlog <- 0.35
#' threshold <- 3
#' tail_scale <- 0.9
#' tail_shape <- 0.2
#'
#' dLognormalGpd(4.0, meanlog, sdlog, threshold, tail_scale, tail_shape, log = FALSE)
#' pLognormalGpd(4.0, meanlog, sdlog, threshold, tail_scale, tail_shape,
#'              lower.tail = TRUE, log.p = FALSE)
#' qLognormalGpd(0.50, meanlog, sdlog, threshold, tail_scale, tail_shape)
#' qLognormalGpd(0.95, meanlog, sdlog, threshold, tail_scale, tail_shape)
#' replicate(10, rLognormalGpd(1, meanlog, sdlog, threshold, tail_scale, tail_shape))
#' @rdname lognormal_gpd
#' @name lognormal_gpd
#' @aliases dLognormalGpd pLognormalGpd rLognormalGpd qLognormalGpd
NULL

#' @describeIn lognormal_gpd Lognormal + GPD tail density
#' @export
dLognormalGpd <- nimble::nimbleFunction(
  run = function(x = double(0),
                 meanlog = double(0),
                 sdlog = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 log = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (x < threshold) {
      if (log == 1) return(dlnorm(x, meanlog, sdlog, 1)) else return(dlnorm(x, meanlog, sdlog, 0))
    }

    Fu <- plnorm(threshold, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    val <- (1.0 - Fu) * dGpd(x, threshold, tail_scale, tail_shape, 0)
    if (val < eps) val <- eps
    if (log == 1) return(log(val)) else return(val)
  }
)

#' @describeIn lognormal_gpd Lognormal + GPD tail distribution function
#' @export
pLognormalGpd <- nimble::nimbleFunction(
  run = function(q = double(0),
                 meanlog = double(0),
                 sdlog = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0),
                 lower.tail = integer(0, default = 1),
                 log.p = integer(0, default = 0)) {
    returnType(double(0))
    eps <- 1e-300

    if (q < threshold) return(plnorm(q, meanlog, sdlog, lower.tail, log.p))

    Fu <- plnorm(threshold, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    G <- pGpd(q, threshold, tail_scale, tail_shape, 1, 0)
    cdf <- Fu + (1.0 - Fu) * G
    cdf <- max(min(cdf, 1.0), 0.0)

    if (lower.tail == 0) cdf <- 1.0 - cdf
    if (log.p == 1) cdf <- log(max(cdf, eps))
    return(cdf)
  }
)

#' @describeIn lognormal_gpd Lognormal + GPD tail random generation
#' @export
rLognormalGpd <- nimble::nimbleFunction(
  run = function(n = integer(0),
                 meanlog = double(0),
                 sdlog = double(0),
                 threshold = double(0),
                 tail_scale = double(0),
                 tail_shape = double(0)) {
    returnType(double(0))

    if (n != 1) return(0.0)

    Fu <- plnorm(threshold, meanlog, sdlog, 1, 0)
    Fu <- max(min(Fu, 1.0), 0.0)

    uu <- runif(1, 0.0, 1.0)
    if (uu < Fu) return(rlnorm(1, meanlog, sdlog))
    return(rGpd(1, threshold, tail_scale, tail_shape))
  }
)

#' @describeIn lognormal_gpd Lognormal + GPD tail quantile function
#' @export
qLognormalGpd <- function(p, meanlog, sdlog, threshold, tail_scale, tail_shape,
                          lower.tail = TRUE, log.p = FALSE) {
  if (log.p) p <- exp(p)
  if (!lower.tail) p <- 1 - p
  p <- pmax(pmin(p, 1), 0)

  Fu <- stats::plnorm(threshold, meanlog = meanlog, sdlog = sdlog, lower.tail = TRUE, log.p = FALSE)
  Fu <- max(min(Fu, 1.0), 0.0)

  out <- numeric(length(p))
  for (i in seq_along(p)) {
    pi <- p[i]
    if (pi <= Fu) {
      out[i] <- stats::qlnorm(pi, meanlog = meanlog, sdlog = sdlog, lower.tail = TRUE, log.p = FALSE)
    } else {
      g <- if (Fu >= 1) 0 else (pi - Fu) / (1 - Fu)
      out[i] <- qGpd(g, threshold = threshold, scale = tail_scale, shape = tail_shape)
    }
  }
  out
}


# ==========================================================
# Lowercase vectorized R wrappers for Lognormal kernels
# ==========================================================

#' Lowercase vectorized lognormal distribution functions
#'
#' Vectorized R wrappers for the scalar lognormal-kernel topics in this file.
#'
#' @param x Numeric vector of quantiles.
#' @param q Numeric vector of quantiles.
#' @param p Numeric vector of probabilities.
#' @param n Integer number of observations to generate.
#' @param w Numeric vector of mixture weights.
#' @param meanlog,sdlog Numeric vectors (mix) or scalars (base+gpd) of component parameters.
#' @param threshold,tail_scale,tail_shape GPD tail parameters (scalars).
#' @param log Logical; if \code{TRUE}, return log-density.
#' @param lower.tail Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.
#' @param log.p Logical; if \code{TRUE}, probabilities are on log scale.
#' @param tol,maxiter Tolerance and max iterations for numerical inversion.
#'
#' @return Numeric vector of densities, probabilities, quantiles, or random variates.
#'
#' @details
#' These are direct vectorized wrappers around the scalar lognormal routines. They keep the same
#' parameterization, support restrictions, and bulk-tail splice, while allowing ordinary vector
#' inputs in R. Quantile wrappers continue to use the scalar inversion logic, so there is no
#' separate approximation layer in the lowercase API.
#'
#' @seealso [lognormal_mix()], [lognormal_mixgpd()], [lognormal_gpd()], [bundle()],
#'   [get_kernel_registry()].
#' @family vectorized kernel helpers
#'
#' @examples
#' w <- c(0.6, 0.3, 0.1)
#' ml <- c(0, 0.3, 0.6)
#' sl <- c(0.4, 0.5, 0.6)
#'
#' # Lognormal mixture
#' dlognormalmix(c(1, 2, 3), w = w, meanlog = ml, sdlog = sl)
#' rlognormalmix(5, w = w, meanlog = ml, sdlog = sl)
#'
#' # Lognormal mixture + GPD
#' dlognormalmixgpd(c(2, 3, 4), w = w, meanlog = ml, sdlog = sl,
#'                  threshold = 2.5, tail_scale = 0.5, tail_shape = 0.2)
#'
#' @name lognormal_lowercase
#' @rdname lognormal_lowercase
NULL

# ---- Lognormal Mix lowercase wrappers ----

#' @describeIn lognormal_lowercase Lognormal mixture density (vectorized)
#' @export
dlognormalmix <- function(x, w, meanlog, sdlog, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLognormalMix(xi, w = w, meanlog = meanlog, sdlog = sdlog, log = log_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal mixture distribution function (vectorized)
#' @export
plognormalmix <- function(q, w, meanlog, sdlog, lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLognormalMix(qi, w = w, meanlog = meanlog, sdlog = sdlog,
                                                   lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal mixture quantile function (vectorized)
#' @export
qlognormalmix <- function(p, w, meanlog, sdlog, lower.tail = TRUE, log.p = FALSE,
                          tol = 1e-10, maxiter = 200) {
  qLognormalMix(p, w = w, meanlog = meanlog, sdlog = sdlog, lower.tail = lower.tail,
                log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn lognormal_lowercase Lognormal mixture random generation (vectorized)
#' @export
rlognormalmix <- function(n, w, meanlog, sdlog) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLognormalMix(1L, w = w, meanlog = meanlog, sdlog = sdlog)),
         numeric(1L))
}

# ---- Lognormal Mix + GPD lowercase wrappers ----

#' @describeIn lognormal_lowercase Lognormal mixture + GPD density (vectorized)
#' @export
dlognormalmixgpd <- function(x, w, meanlog, sdlog, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLognormalMixGpd(xi, w = w, meanlog = meanlog, sdlog = sdlog,
                                                      threshold = threshold, tail_scale = tail_scale,
                                                      tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal mixture + GPD distribution function (vectorized)
#' @export
plognormalmixgpd <- function(q, w, meanlog, sdlog, threshold, tail_scale, tail_shape,
                             lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLognormalMixGpd(qi, w = w, meanlog = meanlog, sdlog = sdlog,
                                                      threshold = threshold, tail_scale = tail_scale,
                                                      tail_shape = tail_shape,
                                                      lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal mixture + GPD quantile function (vectorized)
#' @export
qlognormalmixgpd <- function(p, w, meanlog, sdlog, threshold, tail_scale, tail_shape,
                             lower.tail = TRUE, log.p = FALSE, tol = 1e-10, maxiter = 200) {
  qLognormalMixGpd(p, w = w, meanlog = meanlog, sdlog = sdlog, threshold = threshold,
                   tail_scale = tail_scale, tail_shape = tail_shape,
                   lower.tail = lower.tail, log.p = log.p, tol = tol, maxiter = maxiter)
}

#' @describeIn lognormal_lowercase Lognormal mixture + GPD random generation (vectorized)
#' @export
rlognormalmixgpd <- function(n, w, meanlog, sdlog, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLognormalMixGpd(1L, w = w, meanlog = meanlog, sdlog = sdlog,
                                                              threshold = threshold, tail_scale = tail_scale,
                                                              tail_shape = tail_shape)),
         numeric(1L))
}

# ---- Lognormal + GPD lowercase wrappers ----

#' @describeIn lognormal_lowercase Lognormal + GPD density (vectorized)
#' @export
dlognormalgpd <- function(x, meanlog, sdlog, threshold, tail_scale, tail_shape, log = FALSE) {
  x <- as.numeric(x)
  if (length(x) == 0L) return(numeric(0L))
  log_int <- as.integer(log)
  vapply(x, function(xi) as.numeric(dLognormalGpd(xi, meanlog = meanlog, sdlog = sdlog,
                                                   threshold = threshold, tail_scale = tail_scale,
                                                   tail_shape = tail_shape, log = log_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal + GPD distribution function (vectorized)
#' @export
plognormalgpd <- function(q, meanlog, sdlog, threshold, tail_scale, tail_shape,
                          lower.tail = TRUE, log.p = FALSE) {
  q <- as.numeric(q)
  if (length(q) == 0L) return(numeric(0L))
  lt_int <- as.integer(lower.tail)
  lp_int <- as.integer(log.p)
  vapply(q, function(qi) as.numeric(pLognormalGpd(qi, meanlog = meanlog, sdlog = sdlog,
                                                   threshold = threshold, tail_scale = tail_scale,
                                                   tail_shape = tail_shape,
                                                   lower.tail = lt_int, log.p = lp_int)),
         numeric(1L))
}

#' @describeIn lognormal_lowercase Lognormal + GPD quantile function (vectorized)
#' @export
qlognormalgpd <- function(p, meanlog, sdlog, threshold, tail_scale, tail_shape,
                          lower.tail = TRUE, log.p = FALSE) {
  qLognormalGpd(p, meanlog = meanlog, sdlog = sdlog, threshold = threshold,
                tail_scale = tail_scale, tail_shape = tail_shape,
                lower.tail = lower.tail, log.p = log.p)
}

#' @describeIn lognormal_lowercase Lognormal + GPD random generation (vectorized)
#' @export
rlognormalgpd <- function(n, meanlog, sdlog, threshold, tail_scale, tail_shape) {
  n <- as.integer(n)
  if (length(n) != 1L || is.na(n)) stop("'n' must be a single integer.", call. = FALSE)
  if (n <= 0L) return(numeric(0L))
  vapply(seq_len(n), function(i) as.numeric(rLognormalGpd(1L, meanlog = meanlog, sdlog = sdlog,
                                                          threshold = threshold, tail_scale = tail_scale,
                                                          tail_shape = tail_shape)),
         numeric(1L))
}

test_that("cluster summaries and print output use decreasing allocation order", {
  lbl <- structure(
    list(
      labels = c("10", "2", "2", "3", "3", "x", "x"),
      components = 4L,
      source = "train",
      data = data.frame(y = c(7, 6, 5, 4, 3, 2, 1))
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )

  sm <- summary(lbl)
  expect_equal(names(sm$cluster_sizes), c("2", "3", "x", "10"))
  expect_equal(as.integer(sm$cluster_sizes), c(2L, 2L, 2L, 1L))
  expect_match(paste(capture.output(print(lbl)), collapse = "\n"), "sizes     : 2:2, 3:2, x:2, 10:1", fixed = TRUE)
})

test_that("predicted cluster labels carry training reference for downstream methods", {
  fit <- structure(
    list(
      spec = list(meta = list()),
      samples = "fake_samples",
      cache_env = new.env(parent = emptyenv())
    ),
    class = c("dpmixgpd_cluster_fit", "list")
  )

  testthat::local_mocked_bindings(
    .cluster_samples_to_matrix = function(samples) {
      matrix(
        c(1, 1, 2, 1, 2, 2),
        nrow = 2,
        byrow = TRUE,
        dimnames = list(NULL, c("z[1]", "z[2]", "z[3]"))
      )
    },
    .cluster_draw_indices = function(n_draws, burnin = NULL, thin = NULL) seq_len(n_draws),
    .cluster_extract_z_from_matrix = function(draw_sub) {
      matrix(c(1L, 1L, 2L, 1L, 2L, 2L), nrow = 2, byrow = TRUE)
    },
    compute_psm = function(z_draws) diag(3),
    dahl_labels = function(z_draws, psm) list(labels = c(1L, 2L, 2L), K = 2L, draw_index = 1L),
    .cluster_compute_scores = function(z_draws, labels, psm) {
      matrix(c(0.9, 0.1, 0.2, 0.8, 0.1, 0.9), nrow = 3, byrow = TRUE)
    },
    .cluster_training_data_frame = function(fit) data.frame(y = c(10, 11, 12)),
    predict_labels_newdata = function(fit, newdata, burnin = NULL, thin = NULL) {
      list(
        labels = c(1L, 2L),
        scores = matrix(c(0.7, 0.3, 0.4, 0.6), nrow = 2, byrow = TRUE),
        data = data.frame(y = c(20, 21)),
        K = 2L
      )
    },
    .package = "CausalMixGPD"
  )

  nd_lbl <- predict(fit, newdata = data.frame(y = c(0.1, 0.2)), type = "label", return_scores = TRUE)

  expect_true(is.list(nd_lbl$train_reference))
  expect_equal(nd_lbl$train_reference$labels, c(1L, 2L, 2L))
  expect_equal(nd_lbl$train_reference$data$y, c(10, 11, 12))
  expect_s3_class(plot(nd_lbl, type = "summary", plotly = FALSE), "ggplot")
  expect_s3_class(plot(fit, which = "summary", plotly = FALSE), "ggplot")
})

test_that("cluster summary plots use boxplots ordered by selected clusters", {
  skip_if_not_installed("ggplot2")

  lbl <- structure(
    list(
      labels = c("2", "2", "3", "3", "3", "10"),
      components = 3L,
      source = "train",
      data = data.frame(y = c(1.0, 1.2, 2.0, 2.2, 2.4, 3.1))
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )

  p <- plot(lbl, type = "summary", top_n = 2L, order_by = "size", plotly = FALSE)
  geom_classes <- unique(vapply(p$layers, function(layer) class(layer$geom)[1], character(1)))
  built <- ggplot2::ggplot_build(p)
  fill_scale <- built$plot$scales$get_scales("fill")
  color_scale <- built$plot$scales$get_scales("colour")

  expect_equal(geom_classes, "GeomBoxplot")
  expect_equal(levels(p$data$cluster), c("3", "2"))
  expect_setequal(fill_scale$range$range, c("3", "2"))
  expect_setequal(color_scale$range$range, c("3", "2"))
})

test_that("newdata summary plots use newdata-only boxplots for populated clusters", {
  skip_if_not_installed("ggplot2")

  lbl <- structure(
    list(
      labels = c("2", "3", "3"),
      components = 3L,
      source = "newdata",
      data = data.frame(y = c(1.5, 2.1, 2.4)),
      train_reference = list(
        labels = c("2", "2", "3", "3", "10"),
        data = data.frame(y = c(1.0, 1.1, 2.0, 2.2, 3.3))
      )
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )

  p <- plot(lbl, type = "summary", top_n = 2L, order_by = "size", plotly = FALSE)
  geom_classes <- unique(vapply(p$layers, function(layer) class(layer$geom)[1], character(1)))
  built <- ggplot2::ggplot_build(p)
  fill_scale <- built$plot$scales$get_scales("fill")
  color_scale <- built$plot$scales$get_scales("colour")

  expect_equal(geom_classes, "GeomBoxplot")
  expect_equal(levels(p$data$cluster), c("3", "2"))
  expect_setequal(fill_scale$range$range, c("3", "2"))
  expect_setequal(color_scale$range$range, c("3", "2"))
})

test_that("newdata summary plots keep training cluster colors for displayed clusters", {
  skip_if_not_installed("ggplot2")

  lbl <- structure(
    list(
      labels = c("2", "10", "10"),
      components = 3L,
      source = "newdata",
      data = data.frame(y = c(1.5, 3.1, 3.4)),
      train_reference = list(
        labels = c("2", "2", "3", "3", "10"),
        data = data.frame(y = c(1.0, 1.1, 2.0, 2.2, 3.3))
      )
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )

  p <- plot(lbl, type = "summary", top_n = NULL, order_by = "label", plotly = FALSE)
  built <- ggplot2::ggplot_build(p)
  box_data <- built$data[[1]][order(built$data[[1]]$x), c("x", "fill", "colour")]
  pal <- getFromNamespace(".plot_palette", "CausalMixGPD")(3L)

  expect_equal(levels(p$data$cluster), c("2", "10"))
  expect_equal(unname(box_data$fill), unname(c(pal[1], pal[3])))
  expect_equal(unname(box_data$colour), unname(c(pal[1], pal[3])))
})

test_that("cluster size plots use distinct cluster colors and preserve training mapping", {
  skip_if_not_installed("ggplot2")

  lbl <- structure(
    list(
      labels = c("2", "10", "10"),
      components = 3L,
      source = "newdata",
      data = data.frame(y = c(1.5, 3.1, 3.4)),
      train_reference = list(
        labels = c("2", "2", "3", "3", "10"),
        data = data.frame(y = c(1.0, 1.1, 2.0, 2.2, 3.3))
      )
    ),
    class = c("dpmixgpd_cluster_labels", "list")
  )

  p <- plot(lbl, type = "sizes", top_n = NULL, order_by = "label", plotly = FALSE)
  built <- ggplot2::ggplot_build(p)
  bar_data <- built$data[[1]][order(built$data[[1]]$x), c("x", "fill", "colour")]
  pal <- getFromNamespace(".plot_palette", "CausalMixGPD")(3L)

  expect_equal(levels(p$data$cluster), c("2", "10"))
  expect_equal(unname(bar_data$fill), unname(c(pal[1], pal[3])))
  expect_equal(unname(bar_data$colour), unname(c(pal[1], pal[3])))
})

test_that("predict.mixgpd_fit accepts legacy x alias", {
  predict_method <- getFromNamespace("predict.mixgpd_fit", "CausalMixGPD")
  fake_fit <- structure(list(), class = "mixgpd_fit")
  seen <- new.env(parent = emptyenv())

  local_mocked_bindings(
    .validate_fit = function(object) TRUE,
    .predict_mixgpd = function(object, x = NULL, y = NULL, ...) {
      seen$x <- x
      seen$y <- y
      list(fit = data.frame(estimate = 1), type = "mean")
    },
    .package = "CausalMixGPD"
  )

  x_arg <- matrix(1, nrow = 1L, ncol = 1L)
  predict_method(fake_fit, x = x_arg, type = "mean", show_progress = FALSE)

  expect_equal(seen$x, x_arg)
  expect_null(seen$y)
  expect_error(
    predict_method(fake_fit, newdata = x_arg, x = x_arg, type = "mean", show_progress = FALSE),
    "Provide only one of 'newdata' or legacy 'x'"
  )
})

test_that("residuals plugin PIT forwards training X as newdata", {
  residuals_method <- getFromNamespace("residuals.mixgpd_fit", "CausalMixGPD")
  fake_fit <- structure(
    list(
      data = list(
        X = matrix(c(0, 1, 1, 0), nrow = 2, byrow = TRUE),
        y = c(0.2, 0.8)
      )
    ),
    class = "mixgpd_fit"
  )
  seen <- new.env(parent = emptyenv())

  local_mocked_bindings(
    predict.mixgpd_fit = function(object, newdata = NULL, y = NULL, type = c("density", "survival", "quantile", "sample", "mean", "rmean", "median", "fit"), ...) {
      seen$newdata <- newdata
      seen$y <- y
      list(fit = data.frame(survival = c(0.8, 0.4)))
    },
    .package = "CausalMixGPD"
  )

  pit <- residuals_method(fake_fit, type = "pit", pit = "plugin")

  expect_equal(as.numeric(pit), c(0.2, 0.6))
  expect_equal(seen$newdata, fake_fit$data$X)
  expect_equal(seen$y, fake_fit$data$y)
})

---
title: "One-Arm Regression Modeling with CausalMixGPD"
output: rmarkdown::html_vignette
bibliography: CausalMixGPD.bib
link-citations: true
vignette: >
  %\VignetteIndexEntry{One-Arm Regression Modeling with CausalMixGPD}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.8,
  out.width = "85%",
  fig.align = "center",
  message = FALSE,
  warning = FALSE
)
.pkg_extdata <- function(name) {
  f <- system.file("extdata", name, package = "CausalMixGPD")
  if (nzchar(f)) {
    return(f)
  }
  vdir <- knitr::current_input(dir = TRUE)
  pkg_root <- normalizePath(file.path(vdir, ".."), winslash = "/", mustWork = TRUE)
  alt <- file.path(pkg_root, "inst", "extdata", name)
  if (file.exists(alt)) {
    return(normalizePath(alt, winslash = "/", mustWork = TRUE))
  }
  stop("Missing extdata: ", name, ". Run data-raw/build_vignette_fits.R", call. = FALSE)
}
```

# Introduction

`CausalMixGPD` is designed for settings where the response distribution is not well represented by a simple parametric family. Many applications have heteroskedastic, skewed, or multimodal outcomes that are better modeled with flexible mixtures, together with upper-tail behavior that may require extrapolation. The `CausalMixGPD` package addresses this setting with a spliced distribution in which a Dirichlet process mixture (DPM) model for the core is optionally augmented with a generalized Pareto distribution (GPD) tail [@EscobarWest1995; @BalkemaDeHaan1974; @Pickands1975; @Coles2001].

A conditional outcome distribution is fitted first, and densities, survival functions, quantiles, means, and restricted means are then extracted from the fitted distribution. The software paper describes the one-arm workflow that the causal and clustering extensions build on.

This vignette walks through the workflow using a built-in simulated data set. It introduces the notation and model setup, describes the main one-arm workflow functions and their statistical targets, shows the interface through executable code, and summarizes the main customization options.

# Notation and statistical background

## Conditional distribution as the inferential target

Let $Y$ denote the response and let $X \in \mathbb{R}^p$ denote a predictor vector. For a fixed covariate value $x$, the conditional cumulative distribution function and conditional density are

$$
F(y \mid x) = \Pr(Y \le y \mid X = x), \qquad
f(y \mid x) = \frac{\partial}{\partial y} F(y \mid x),
$$

and the conditional quantile function is

$$
Q(\tau \mid x) = \inf\{y : F(y \mid x) \ge \tau\}, \qquad 0 < \tau < 1.
$$

In `CausalMixGPD`, this conditional distribution is the primary object of inference. Scalar summaries such as means or upper quantiles are derived from the fitted distribution rather than modeled separately. That design is important because it keeps different summaries internally coherent and avoids issues such as quantile crossing that can arise when quantiles are estimated one at a time.

## Bulk model: Dirichlet process mixtures

Below a tail threshold, the package uses a DPM regression model. If $k(\cdot \mid x; \theta)$ denotes a chosen kernel family, the bulk conditional density can be written as

$$
f_{\mathrm{DP}}(y \mid x) = \int k(y \mid x; \theta) \, dH(\theta)
= \sum_{j=1}^{\infty} w_j k(y \mid x; \theta_j),
$$

where $H$ denotes a random mixing distribution having a Dirichlet process prior such that the mixing proportions obey $w_j \ge 0$ and $\sum_{j} w_j = 1$, and the kernel parameters can be either fixed or covariate-dependent [@Ferguson1973; @Sethuraman1994; @Neal2000]. The mixture formulation provides much flexibility in dealing with multimodality, skewness, nonuniform local dispersion, and potential latent grouping without requiring any prespecification of the number of mixture components.

From an implementation standpoint, the bulk model specification is largely driven by four main inputs: `kernel`, `backend`, `components`, and `param_specs`. The first two control the kernel function $k(\cdot)$ and the underlying representation, respectively, the third is the internal truncation level, and the fourth decides whether the kernel parameters are fixed, drawn from priors, or regressed.

## Tail model: generalized Pareto exceedances

When the upper tail requires explicit modeling, the package replaces the part of the density above a threshold $u(x)$ by a GPD component. For exceedances over the threshold, define

$$
Z = Y - u(x) \quad \text{given } Y > u(x).
$$

Under the peaks-over-threshold framework, the conditional excess distribution is approximated by a GPD when the threshold is sufficiently high [@BalkemaDeHaan1974; @Pickands1975; @Coles2001]. For $y > u(x)$, with scale parameter $\sigma(x) > 0$ and shape parameter $\xi$, the GPD cumulative distribution function is

$$
F_{\mathrm{GPD}}(y \mid u(x), \sigma(x), \xi)
=
\begin{cases}
1 - \left(1 + \xi \frac{y-u(x)}{\sigma(x)}\right)^{-1/\xi}, & \xi \ne 0, \\
1 - \exp\left(-\frac{y-u(x)}{\sigma(x)}\right), & \xi = 0.
\end{cases}
$$

The support is $y \ge u(x)$ when $\xi \ge 0$, and $u(x) \le y \le u(x) - \sigma(x)/\xi$ when $\xi < 0$. The shape parameter controls the tail regime: negative values imply a finite upper endpoint, $\xi = 0$ gives an exponential-type tail, and positive values correspond to heavy tails.

## Spliced bulk-tail construction

The key statistical idea in the package is the spliced model. Let $F_{\mathrm{DP}}(\cdot \mid x; \Theta)$ and $f_{\mathrm{DP}}(\cdot \mid x; \Theta)$ denote the bulk CDF and density, and let $\Phi$ collect the tail parameters. Write

$$
p_u(x; \Theta) = F_{\mathrm{DP}}(u(x) \mid x; \Theta)
$$

for the bulk probability mass below the threshold. Then the spliced conditional CDF is

$$
F(y \mid x; \Theta, \Phi) =
\begin{cases}
F_{\mathrm{DP}}(y \mid x; \Theta), & y \le u(x), \\
p_u(x; \Theta) + \{1 - p_u(x; \Theta)\} F_{\mathrm{GPD}}(y \mid x; \Phi), & y > u(x),
\end{cases}
$$

and the corresponding conditional density is

$$
f(y \mid x; \Theta, \Phi) =
\begin{cases}
f_{\mathrm{DP}}(y \mid x; \Theta), & y \le u(x), \\
\{1 - p_u(x; \Theta)\} f_{\mathrm{GPD}}(y \mid x; \Phi), & y > u(x).
\end{cases}
$$

This formulation preserves the flexibility of the mixture in the central region while forcing the tail to obey an EVT-motivated form. The splice is continuous at $u(x)$ by construction, and the tail density is normalized by the bulk survival at the threshold.

The quantile function follows the same two-part logic. For quantile levels below the threshold mass, the package inverts the bulk CDF. For levels above the threshold mass, the probability index is rescaled to the exceedance scale:

$$
\widetilde\tau(x; \Theta) = \frac{\tau - p_u(x; \Theta)}{1 - p_u(x; \Theta)}.
$$

Then the upper-tail quantiles are obtained from the GPD quantile function evaluated at $\widetilde\tau(x; \Theta)$. This distinction explains why upper quantile predictions are one of the main places where a spliced fit can differ meaningfully from a bulk-only fit.

## Means, restricted means, and existence conditions

When the kernel mean exists and the tail satisfies $\xi < 1$, the conditional mean of the spliced model is finite. In that case the one-arm interface can report ordinary means through `predict(type = "mean")`. When the tail is too heavy for a finite mean or when the user prefers a bounded summary, the package provides restricted means through `predict(type = "rmean")`, which returns

$$
E\{\min(Y, c) \mid X = x\}
$$

for a user-supplied cutoff $c$. Restricted means are often easier to interpret in heavy-tailed settings because they remain finite and are less dominated by a few very large posterior draws. The software paper emphasizes this distinction explicitly in the one-arm prediction section.

# Function map for the one-arm workflow

The main exported functions used in this vignette are listed below.

- `dpmgpd()` fits a one-arm spliced DPM-GPD model.
- `dpmix()` fits the corresponding bulk-only DPM model.
- `summary()` reports posterior summaries from a fitted object.
- `params()` extracts parameter summaries in a more direct programmatic form.
- `plot()` provides diagnostic or summary graphics for fitted and predicted objects.
- `predict()` returns posterior summaries for densities, survival functions, quantiles, means, restricted means, medians, samples, and fitted values.

These are not separate modeling frameworks. They are different ways of interrogating the same posterior distribution after fitting a one-arm model.

# Package setup

```{r libraries}
library(CausalMixGPD)
library(ggplot2)
```

For a vignette, it is reasonable to use modest MCMC settings so the code remains runnable on ordinary hardware. In a substantive analysis, longer runs and sensitivity checks are recommended.

```{r mcmc-settings}
mcmc_vig <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2026
)
```

# Data

We use the bundled dataset `nc_posX100_p3_k2`, which contains a positive response and three predictors. It is convenient for illustrating both the bulk-only and spliced workflows.

```{r load-data}
data("nc_posX100_p3_k2", package = "CausalMixGPD")
onearm_dat <- data.frame(
  y = nc_posX100_p3_k2$y,
  nc_posX100_p3_k2$X
)

```

A quick visual check helps motivate flexible modeling.

```{r exploratory-plots, fig.height=4.5, fig.width=7}
p1 <- ggplot(onearm_dat, aes(x = y)) +
  geom_histogram(aes(y = after_stat(density)), bins = 25,
                 fill = "grey85", colour = "grey35") +
  geom_density(linewidth = 0.9) +
  labs(x = "y", y = "Density",
       title = "Observed response distribution") +
  theme_minimal()

p2 <- ggplot(onearm_dat, aes(sample = y)) +
  stat_qq() +
  stat_qq_line(colour = 2) +
  labs(title = "Normal Q-Q plot for the response") +
  theme_minimal()

p1
p2
```

The response is positive and the empirical distribution is skewed. That makes the dataset a natural candidate for positive-support kernels such as `gamma`, `lognormal`, or `invgauss`. If the scientific question involves high quantiles or tail probabilities, a spliced GPD tail can also be justified.

# Fitting a spliced one-arm model

We begin with `dpmgpd()`, which fits the bulk mixture together with an active GPD tail. The article presents this as the main one-arm interface for the spliced workflow, with the formula describing the conditional outcome model and the backend selecting the latent DPM representation.

To keep vignette build time predictable, outputs from this vignette are **loaded from** `inst/extdata/one_arm_outputs.rds` (created by `tools/.Rscripts/build_vignette_fits.R`). This contains only the prediction/summary objects needed to render the figures and tables, rather than the full fitted MCMC objects. The code chunks below show the package calls you would run in a full analysis; the rendered tables and figures are produced from the shipped artifacts in companion chunks with hidden source.

```{r fit-spliced-display, eval=FALSE}
fit_spliced <- dpmgpd(
  formula = y ~ x1 + x2 + x3,
  data = onearm_dat,
  backend = "sb",
  kernel = "lognormal",
  components = 5,
  mcmc = mcmc_vig
)
```

```{r fit-spliced, echo=FALSE}
one_arm_out <- readRDS(.pkg_extdata("one_arm_outputs.rds"))
```

A fitted one-arm object supports posterior summaries and diagnostic plots. To avoid storing the full fit, we include static diagnostics generated offline.

```{r summary-spliced-display, eval=FALSE}
summary(fit_spliced)
```

```{r summary-spliced, echo=FALSE}
fsum <- system.file("extdata", "one_arm_fit_summary.txt", package = "CausalMixGPD")
if (!nzchar(fsum)) {
  alt <- normalizePath(
    file.path(knitr::current_input(dir = TRUE), "..", "inst", "extdata", "one_arm_fit_summary.txt"),
    winslash = "/",
    mustWork = FALSE
  )
  if (file.exists(alt)) fsum <- alt
}
if (nzchar(fsum) && file.exists(fsum)) {
  cat(paste(readLines(fsum, warn = FALSE, encoding = "UTF-8"), collapse = "\n"))
} else {
  cat("Precomputed fit summary not found; run tools/.Rscripts/build_vignette_fits.R.\n")
}
```

```{r plot-spliced-trace-display, eval=FALSE}
plot(fit_spliced, family = c("traceplot", "density"), params = "alpha")
```

```{r plot-spliced-trace, echo=FALSE, fig.height=5.2, fig.width=7}
knitr::include_graphics("assets/one_arm_alpha_trace.png")
```

```{r plot-spliced-density-display, eval=FALSE}
# Posterior density for monitored parameters (e.g. alpha); see plot.mixgpd_fit().
plot(fit_spliced, family = "density", params = "alpha")
```

```{r plot-spliced-density, echo=FALSE, fig.height=5, fig.width=7}
knitr::include_graphics("assets/one_arm_alpha_density.png")
```

The exact parameter names stored by the object depend on the chosen kernel and backend, but the general interpretation is stable. The concentration parameter controls the effective complexity of the mixture, while the monitored bulk and tail parameters determine the conditional distribution used for all later prediction tasks. Use `params(fit_spliced)` when you need the posterior draws as a matrix for custom summaries.

# Posterior prediction from the spliced model

## Conditional density and survival summaries

The package article notes that density and survival summaries are returned by evaluating the corresponding functional draw by draw and then summarizing over retained MCMC draws. We illustrate that at three representative predictor profiles formed from empirical quartiles.

```{r newdata-grid-display, eval=FALSE}
x_new <- as.data.frame(lapply(
  onearm_dat[, c("x1", "x2", "x3")],
  stats::quantile,
  probs = c(0.25, 0.50, 0.75),
  na.rm = TRUE
))
rownames(x_new) <- c("q25", "q50", "q75")
x_new
```

```{r newdata-grid, echo=FALSE}
x_new <- one_arm_out$x_new
x_new
```

```{r density-prediction-display, eval=FALSE}
y_grid <- seq(min(onearm_dat$y), quantile(onearm_dat$y, 0.99), length.out = 120)
pred_dens <- predict(
  fit_spliced,
  newdata = x_new,
  y = y_grid,
  type = "density",
  interval = "credible",
  level = 0.95,
  store_draws = FALSE
)
plot(pred_dens, type = "density", facet = "covariate")
```

```{r density-prediction, echo=FALSE}
pred_dens <- one_arm_out$pred_dens
plot(pred_dens, type = "density", facet = "covariate")
```

```{r survival-prediction-display, eval=FALSE}
y_grid <- seq(min(onearm_dat$y), quantile(onearm_dat$y, 0.99), length.out = 120)
pred_surv <- predict(
  fit_spliced,
  newdata = x_new,
  y = y_grid,
  type = "survival",
  interval = "credible",
  level = 0.95,
  store_draws = FALSE
)
plot(pred_surv)
```

```{r survival-prediction, echo=FALSE}
pred_surv <- one_arm_out$pred_surv
plot(pred_surv)
```

These summaries show how the fitted conditional distribution changes across covariate profiles. In practice, that helps determine whether predictors act mostly through a location shift, through dispersion, or through upper-tail behavior.

## Quantiles, means, and restricted means

The one-arm workflow is especially useful when the scientific target is not just the density itself but one or more functionals of the conditional distribution. The article highlights quantiles and mean-type summaries as central examples, with restricted means remaining available even when the ordinary mean is unstable or undefined under a heavy tail.

```{r quantile-prediction-display, eval=FALSE}
pred_quant <- predict(
  fit_spliced,
  newdata = x_new,
  type = "quantile",
  index = c(0.25, 0.50, 0.75, 0.90, 0.95),
  interval = "credible",
  level = 0.95,
  store_draws = FALSE
)
```

```{r quantile-prediction, echo=FALSE}
pred_quant <- one_arm_out$pred_quant
```

```{r quantile-prediction-plot-display, eval=FALSE}
plot(pred_quant)
```

```{r quantile-prediction-plot, echo=FALSE, fig.height=5.5, fig.width=7}
plot(pred_quant)
```

```{r mean-prediction-display, eval=FALSE}
pred_mean <- predict(
  fit_spliced,
  newdata = x_new,
  type = "mean",
  interval = "hpd",
  level = 0.90,
  store_draws = FALSE
)
```

```{r mean-prediction, echo=FALSE}
pred_mean <- one_arm_out$pred_mean
```

```{r mean-prediction-plot-display, eval=FALSE}
plot(pred_mean)
```

```{r mean-prediction-plot, echo=FALSE, fig.height=5.2, fig.width=7}
plot(pred_mean)
```

```{r rmean-prediction-display, eval=FALSE}
cutoff_val <- as.numeric(stats::quantile(onearm_dat$y, 0.95))
pred_rmean <- predict(
  fit_spliced,
  newdata = x_new,
  type = "rmean",
  cutoff = cutoff_val,
  interval = "hpd",
  level = 0.90,
  store_draws = FALSE
)
```

```{r rmean-prediction, echo=FALSE}
cutoff_val <- one_arm_out$cutoff_val
```

```{r rmean-table-display, eval=FALSE}
knitr::kable(pred_rmean$fit_df, format = "html", row.names = FALSE, digits = 4)
```

```{r rmean-table, echo=FALSE}
knitr::kable(
  one_arm_out$pred_rmean_fit_df,
  format = "html",
  row.names = FALSE,
  digits = 4
)
```

The restricted mean is often a good complement to upper-tail quantiles. It preserves the original response scale, is always finite for finite cutoffs, and gives a more stable summary when a few extreme draws can dominate the ordinary mean.

# Comparing bulk-only and spliced fits

The package also provides a bulk-only wrapper, `dpmix()`, which removes the GPD tail while leaving the general predictive interface intact. The article presents this as the natural special case obtained when the tail component is switched off.

```{r fit-bulk-only-display, eval=FALSE}
fit_bulk <- dpmix(
  formula = y ~ x1 + x2 + x3,
  data = onearm_dat,
  backend = "sb",
  kernel = "lognormal",
  components = 5,
  mcmc = mcmc_vig
)
```

```{r fit-bulk-only, echo=FALSE}
cat("Bulk-only fit computed offline; see quantile comparison below.\n")
```

```{r compare-quantiles-display, eval=FALSE}
quant_bulk_fit_df <- predict(
  fit_bulk,
  newdata = x_new,
  type = "quantile",
  index = c(0.90, 0.95),
  interval = "credible",
  level = 0.95,
  store_draws = FALSE
)$fit_df
quant_spliced_fit_df <- predict(
  fit_spliced,
  newdata = x_new,
  type = "quantile",
  index = c(0.90, 0.95),
  interval = "credible",
  level = 0.95,
  store_draws = FALSE
)$fit_df
```

```{r compare-quantiles, echo=FALSE}
quant_bulk_fit_df <- one_arm_out$quant_bulk_fit_df
quant_spliced_fit_df <- one_arm_out$quant_spliced_fit_df
```

```{r compare-quantiles-plot-display, eval=FALSE}
qb <- quant_bulk_fit_df
qb$model <- "dpmix (bulk only)"
qs <- quant_spliced_fit_df
qs$model <- "dpmgpd (spliced)"
qc <- rbind(qb, qs)
qc$profile <- rownames(x_new)[as.integer(qc$id)]
ggplot(qc, aes(x = profile, y = estimate, colour = model, group = model)) +
  geom_line(linewidth = 0.8) +
  geom_point(size = 2) +
  facet_wrap(~index, scales = "free_y", ncol = 2) +
  labs(
    x = "Covariate profile",
    y = "Posterior mean quantile",
    title = "Upper quantiles: bulk-only vs spliced tail"
  ) +
  theme_minimal() +
  theme(axis.text.x = element_text(angle = 20, hjust = 1))
```

```{r compare-quantiles-plot, echo=FALSE, fig.height=5, fig.width=7}
qb <- quant_bulk_fit_df
qb$model <- "dpmix (bulk only)"
qs <- quant_spliced_fit_df
qs$model <- "dpmgpd (spliced)"
qc <- rbind(qb, qs)
qc$profile <- rownames(x_new)[as.integer(qc$id)]
ggplot(qc, aes(x = profile, y = estimate, colour = model, group = model)) +
  geom_line(linewidth = 0.8) +
  geom_point(size = 2) +
  facet_wrap(~index, scales = "free_y", ncol = 2) +
  labs(
    x = "Covariate profile",
    y = "Posterior mean quantile",
    title = "Upper quantiles: bulk-only vs spliced tail"
  ) +
  theme_minimal() +
  theme(axis.text.x = element_text(angle = 20, hjust = 1))
```

This comparison is one of the most revealing diagnostics in a heavy-tailed application. Bulk-only and spliced models can look very similar in the center of the distribution, yet their upper-tail quantile predictions can differ materially because only the spliced model imposes an explicit EVT-based tail form.

# Analysis summary

The one-arm workflow is easiest to interpret when the fitted conditional distribution is the main object of interest. In this example, the positive-support mixture captures the main structure of the response across covariate profiles, while the optional GPD tail changes how upper-tail mass is extrapolated. The density and survival summaries describe the overall shape of the fitted conditional distribution, the quantile summaries highlight profile-specific location and tail changes, and the restricted mean provides a bounded summary on the original response scale.

From a practical modeling standpoint, `dpmix()` and `dpmgpd()` are two related model classes with the same summary interface. The choice between them should be driven by the role of the upper tail in the analysis. If the target is mainly bulk prediction, a bulk-only mixture may be adequate. If inference on extreme upper quantiles, tail probabilities, or means sensitive to upper-tail behavior matters, the spliced formulation is often more appropriate.

# Customization options for one-arm models

This section collects the main arguments that users typically customize in one-arm analyses. The package appendix provides the same controls in compact reference form.

## Model specification options

`kernel` selects the bulk family. Positive-support choices include `"gamma"`, `"lognormal"`, `"invgauss"`, and `"amoroso"`; real-line kernels include `"normal"`, `"laplace"`, and `"cauchy"`. The support of the response should guide this choice.

`backend` determines how the DPM is represented internally. The main options are `"sb"` for stick-breaking and `"crp"` for a Chinese restaurant process representation.

`components` controls the truncation size used for the finite approximation to the infinite mixture.

`GPD` is implicit in the wrapper choice: `dpmgpd()` activates the spliced tail, whereas `dpmix()` fits the bulk-only model.

## Parameter and prior customization

`param_specs` is the central argument for advanced model control. Each kernel or tail parameter can be assigned one of three modes: a fixed value, a prior-driven random value, or a covariate-linked regression structure. This is the main route for changing whether a threshold, scale, or kernel parameter is held constant or allowed to vary with predictors.

`alpha_random` controls whether the DPM concentration parameter is treated as random under its prior or fixed.

`epsilon` controls downstream truncation and summary behavior for very small mixture components.

## MCMC and monitoring options

`mcmc` contains the iteration controls `niter`, `nburnin`, `thin`, `nchains`, and `seed`.

`monitor` chooses how much of the posterior state is retained. The usual values are `"core"` and `"full"`.

`monitor_latent = TRUE` requests latent allocation labels in the retained output, and `monitor_v = TRUE` stores stick-breaking fractions for SB fits.

## Prediction options

`predict()` supports `type = "density"`, `"survival"`, `"quantile"`, `"mean"`, `"rmean"`, `"median"`, `"sample"`, and `"fit"`.

For density and survival summaries, the argument `y` supplies the evaluation grid.

For quantile summaries, the argument `index` supplies the probability levels.

For restricted means, the argument `cutoff` defines the truncation point and `nsim_mean` can control Monte Carlo evaluation when applicable.

`interval` can be set to `"credible"`, `"hpd"`, or `NULL`, and `level` controls the posterior interval coverage level.

# Session information

```{r session-info}
sessionInfo()
```

---
title: "Predictor-Dependent Clustering with CausalMixGPD"
output: rmarkdown::html_vignette
bibliography: CausalMixGPD.bib
link-citations: true
vignette: >
  %\VignetteIndexEntry{Predictor-Dependent Clustering with CausalMixGPD}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 7,
  fig.height = 4.8,
  out.width = "85%",
  fig.align = "center",
  message = FALSE,
  warning = FALSE
)

.pkg_extdata <- function(name) {
  f <- system.file("extdata", name, package = "CausalMixGPD")
  if (nzchar(f)) {
    return(f)
  }
  vdir <- knitr::current_input(dir = TRUE)
  pkg_root <- normalizePath(file.path(vdir, ".."), winslash = "/", mustWork = TRUE)
  alt <- file.path(pkg_root, "inst", "extdata", name)
  if (file.exists(alt)) {
    return(normalizePath(alt, winslash = "/", mustWork = TRUE))
  }
  stop("Missing extdata: ", name, ". Run tools/.Rscripts/build_vignette_fits.R", call. = FALSE)
}
```

# Introduction

The clustering workflow within `CausalMixGPD` relies on Bayesian nonparametric mixture modeling similarly to the density estimation and causal workflows. However, the goal of inference shifts from predicting future responses based on the fitted model to inferring the underlying latent partition created by the mixing process. Thus, the outputs include the posterior similarity matrix (PSM), the cluster labels for some representative observations, cluster sizes, and cluster memberships scores [@Binder1978; @Dahl2006].

This aspect of `CausalMixGPD`'s functionality becomes crucial due to its capacity to discover subgroups using the latent random partition introduced by the Dirichlet process mixture and build flexible outcome models in each found cluster. Clustering is mentioned as one of the three core features of the package along with one-arm modeling and causal inference in the software article.

This vignette provides an overview of the clustering workflow of the package. First, we introduce the latent-partition notation and the types of predictor dependency that the package allows. Next, we run a clustering procedure on a built-in dataset, summarize the resulting random partition in terms of label-invariant objects, and predict cluster labels for some new observations. Finally, we provide a list of customizable elements specific to clustering.

# Notation and short theory background

## Mixture models and latent cluster labels

Let $Y$ denote the response and let $X$ denote the predictor vector. In a mixture model, the conditional density can be written abstractly as

$$
f(y \mid x) = \sum_{j=1}^{\infty} w_j(x) k(y \mid x, \theta_j),
$$

where the weights may or may not depend on predictors and the component-specific parameters may or may not depend on predictors. Introduce a latent allocation variable $z_i$ for observation $i$, with $z_i = j$ meaning that observation $i$ is assigned to component $j$. Then the conditional response model can be written as

$$
y_i \mid z_i, x_i, \{\theta_j\} \sim k(\cdot \mid x_i, \theta_{z_i}).
$$

The collection $\Pi = \{z_1, \ldots, z_n\}$ determines a random partition of the observations. In Bayesian nonparametric clustering, the posterior distribution of this partition is the central inferential object.

## Why labels are not directly interpretable

Mixture component labels are not identifiable across MCMC draws. One posterior draw may label a cluster as component 1 while another labels the same substantive group as component 4. Because of this label-switching phenomenon, cluster summaries should not be based on raw component labels alone.

Instead, the package reports label-invariant summaries. The most important one is the posterior similarity matrix, whose $(i,\ell)$ entry is

$$
S_{i\ell} = \Pr(z_i = z_\ell \mid \text{data}).
$$

This is estimated by averaging co-clustering indicators over retained posterior draws. A representative hard partition is then obtained using Dahl's least-squares criterion [@Dahl2006].

## Predictor dependence in clustering

The package supports three distinct clustering modes, each corresponding to a different way in which predictors can affect the mixture model.

### Weight dependence

In the weight-dependent mode,

$$
f(y \mid x) = \sum_{j=1}^{\infty} w_j(x) k(y \mid \theta_j).
$$

Here the cluster-specific kernel parameters are shared, but the prevalence of the clusters changes with predictors. This is useful when the goal is to let covariates modify which latent groups are more likely without changing the within-group model form.

### Parameter dependence

In the parameter-dependent mode,

$$
f(y \mid x) = \sum_{j=1}^{\infty} w_j k(y \mid x, \theta_j).
$$

The weights are global, but each cluster has its own predictor-response relationship. This is useful when clusters represent different regression regimes.

### Weight and parameter dependence

In the most flexible mode,

$$
f(y \mid x) = \sum_{j=1}^{\infty} w_j(x) k(y \mid x, \theta_j).
$$

Here both cluster prevalence and within-cluster behavior can vary with predictors.

## Main clustering summaries returned by the package

The clustering wrappers produce fitted objects whose predictions are cluster-focused rather than density-focused.

- `predict(type = "psm")` returns the posterior similarity matrix.
- `predict(type = "label")` returns representative labels and, when available, cluster membership scores.
- `summary()` reports cluster sizes and profile summaries.
- `plot()` can display the PSM, cluster sizes, cluster certainty, and representative-cluster summaries.

These summaries allow the user to distinguish between strong clustering structure and uncertain partitions. In many applications, the PSM is more informative than a single hard partition because it shows which observations cluster together consistently and which do not.

# Function map for the clustering workflow

The main exported functions used in this vignette are:

- `dpmix.cluster()` for bulk-only clustering.
- `dpmgpd.cluster()` for clustering with an active GPD tail.
- `predict()` with `type = "psm"` or `type = "label"`.
- `summary()` for cluster sizes and profiles.
- `plot()` for PSM heatmaps and cluster-summary graphics.
- `cluster_profiles()` for extracting the profile table shown by `summary()`.

Across the package, the public workflow is organized around a small set of
model-fitting wrappers and standard S3 methods. One-arm density models use
`dpmix()` and `dpmgpd()`, causal models use `dpmix.causal()` and
`dpmgpd.causal()`, and clustering models use `dpmix.cluster()` and
`dpmgpd.cluster()`. The lower-level bundle constructors are available for
advanced workflows, while code-generation helpers are internal implementation
details. Fitted objects are inspected through `print()`, `summary()`,
`predict()`, `plot()`, `params()`, `fitted()`, and `residuals()` where those
methods are meaningful for the object class.

# Package setup

```{r libraries}
library(CausalMixGPD)
library(ggplot2)
```

```{r mcmc-settings}
mcmc_vig <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2026
)
# Used when regenerating clustering vignette artifacts (tools/.Rscripts/build_vignette_fits.R).
mcmc_clust <- list(
  niter = 1200,
  nburnin = 300,
  thin = 2,
  nchains = 2,
  seed = 2029
)
```

# Data

We use the bundled dataset `nc_realX100_p3_k2`, which contains a real-valued response and three predictors.

```{r load-data}
data("nc_realX100_p3_k2", package = "CausalMixGPD")
dat_cl <- data.frame(
  y = nc_realX100_p3_k2$y,
  nc_realX100_p3_k2$X
)
```

For illustration, we form a simple train/test split so that we can show both training-cluster summaries and out-of-sample label prediction. The chunk below shows the sampling code used for the shipped artifacts; the rendered row counts come from `inst/extdata/clustering_outputs.rds` in a hidden chunk.

```{r split-data-display, eval=FALSE}
set.seed(123)
idx_train <- sample(seq_len(nrow(dat_cl)), size = 80)
train_dat <- dat_cl[idx_train, ]
test_dat <- dat_cl[-idx_train, ]
nrow(train_dat)
nrow(test_dat)
```

```{r split-data, echo=FALSE}
clust_out <- readRDS(.pkg_extdata("clustering_outputs.rds"))
train_dat <- clust_out$train_dat
test_dat <- clust_out$test_dat

nrow(train_dat)
nrow(test_dat)
```

A quick visualization of the response helps motivate the use of a flexible mixture model.

```{r exploratory-plots, fig.height=4.8, fig.width=7}
ggplot(train_dat, aes(x = y)) +
  geom_histogram(aes(y = after_stat(density)), bins = 25,
                 fill = "grey85", colour = "grey35") +
  geom_density(linewidth = 0.9) +
  labs(x = "y", y = "Density", title = "Training-sample response distribution") +
  theme_minimal()
```

# Fitting a clustering model

The bulk-only clustering wrapper is `dpmix.cluster()`. It uses the same general mixture machinery as the one-arm interface, but returns a clustering-oriented fitted object.

```{r fit-cluster-display, eval=FALSE}
fit_cluster <- dpmix.cluster(
  formula = y ~ x1 + x2 + x3,
  data = train_dat,
  kernel = "laplace",
  type = "both",
  components = 8,
  mcmc = mcmc_clust
)
```

```{r fit-cluster, echo=FALSE}
fit_cluster <- clust_out$fit_cluster
```

The choice `type = "both"` requests dependence in both the mixing weights and the component-specific parameters. In practice, this is the most flexible clustering specification because it allows both cluster prevalence and cluster-specific regression behavior to vary with covariates.

# Posterior similarity matrix

The first label-invariant summary is the PSM.

```{r psm-prediction-display, eval=FALSE}
z_train_psm <- predict(fit_cluster, type = "psm")
z_train_psm
```

```{r psm-prediction, echo=FALSE}
z_train_psm <- clust_out$psm_obj
z_train_psm
```

```{r psm-plot-display, eval=FALSE}
plot(z_train_psm, type = "summary")
```

```{r psm-plot, echo=FALSE, fig.height=5.2, fig.width=7}
plot(z_train_psm, type = "summary")
```

The block structure in the PSM is the main diagnostic for cluster separation. Darker within-block regions indicate pairs of observations that co-cluster consistently across posterior draws, while diffuse or weak block boundaries indicate uncertainty about the partition.

# Representative training partition

The next step is to extract representative cluster labels for the training sample.

```{r label-prediction-train-display, eval=FALSE}
z_train_lab <- predict(fit_cluster, type = "label", return_scores = TRUE)
z_train_lab
```

```{r label-prediction-train, echo=FALSE}
z_train_lab <- clust_out$train_lab
z_train_lab
```

```{r label-plot-train-display, eval=FALSE}
plot(z_train_lab, type = "summary")
```

```{r label-plot-train, echo=FALSE, fig.height=5, fig.width=7}
plot(z_train_lab, type = "summary")
```

The representative partition provides a single interpretable clustering summary on the response scale. It is convenient for reporting, but it should be read together with the PSM because the latter conveys the posterior uncertainty in the clustering structure.

# Predicting labels for new observations

A distinctive feature of the package is that it also supports cluster prediction for new observations.

```{r label-prediction-test-display, eval=FALSE}
z_test <- predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
z_test
```

```{r label-prediction-test, echo=FALSE}
z_test <- clust_out$test_lab
z_test
```

```{r test-cluster-profiles-display, eval=FALSE}
summary(
  predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
)
cluster_profiles(
  predict(fit_cluster, newdata = test_dat, type = "label", return_scores = TRUE)
)
```

```{r test-cluster-profiles, echo=FALSE}
cluster_profiles(z_test)
```

```{r sizes-plot-test-display, eval=FALSE}
plot(z_test, type = "sizes")
```

```{r sizes-plot-test, echo=FALSE, fig.height=4.8, fig.width=7}
plot(z_test, type = "sizes")
```

This prediction step maps new observations into the space of the representative training clusters. That is useful in applications where the discovered subgroup structure needs to be carried forward to new units without refitting the entire model.

# Analysis summary

The clustering workflow in `CausalMixGPD` is most informative when the user interprets it as a posterior partition analysis rather than as a hard unsupervised classification algorithm. The PSM shows which observations reliably belong together, the representative labels provide a convenient summary partition, and the test-set predictions extend that partition to new observations.

The three predictor-dependence modes are scientifically meaningful rather than cosmetic. Weight dependence is useful when covariates mainly affect subgroup prevalence. Parameter dependence is useful when clusters correspond to different regression surfaces. The combined mode is appropriate when both mechanisms are plausible. In all three cases, the package reports label-invariant summaries so that posterior uncertainty about the partition is not hidden behind arbitrary component labels.

# Customization options for clustering models

The software appendix gives a compact map of the clustering-specific controls; the most important ones are summarized here.

## Clustering mode

`type = "weights"` places predictor dependence in the mixing weights.

`type = "param"` places predictor dependence in the component-specific kernel parameters.

`type = "both"` allows both forms of dependence simultaneously.

## Wrapper choice

`dpmix.cluster()` fits the bulk-only clustering model.

`dpmgpd.cluster()` adds an active GPD tail and should be considered when tail behavior is part of the clustering problem.

## Kernel, backend, and truncation

`kernel` selects the bulk family, subject to the support of the response.

`components` controls the truncation size. In clustering problems, under-truncation can hide meaningful subgroup structure, so this argument deserves explicit attention.

The package handles some backend details internally based on the clustering mode, but users still choose the scientific dependence structure through `type`.

## Monitoring and output controls

`monitor`, `monitor_latent`, and `monitor_v` control how much posterior state is retained.

`predict(type = "psm")` and `predict(type = "label")` determine whether the user wants a matrix-valued co-clustering summary or representative labels.

`return_scores = TRUE` can be used when membership scores or certainty summaries are needed.

`summary()` reports cluster sizes and within-cluster profiles, while `plot()` supports PSM heatmaps and label-summary graphics such as cluster sizes and certainty.

# Session information

```{r session-info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_dimensions_from_spec}
\alias{build_dimensions_from_spec}
\title{Build dimension declarations from a compiled model spec}
\usage{
build_dimensions_from_spec(spec)
}
\arguments{
\item{spec}{A compiled model specification produced by \code{compile_model_spec()}.}
}
\value{
Named list of dimensions (integer vectors). Scalars are omitted.
}
\description{
Returns a named list of array dimensions used by downstream builders
(inits/monitors/code generation). Dimensions are derived solely from
\code{spec$meta} and \code{spec$plan}. This function does not inspect data.
}
\details{
The model size is controlled by a single parameter: \code{components}.
For SB this is the truncation level of the stick-breaking mixture.
For CRP this is the maximum number of clusters represented in the finite model.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.causalmixgpd_causal_fit}
\alias{print.summary.causalmixgpd_causal_fit}
\title{Print a causal-model summary object}
\usage{
\method{print}{summary.causalmixgpd_causal_fit}(x, digits = 3, max_rows = 60, ...)
}
\arguments{
\item{x}{A \code{"summary.causalmixgpd_causal_fit"} object.}

\item{digits}{Number of digits to print in summary tables.}

\item{max_rows}{Maximum rows to print from each summary table.}

\item{...}{Unused.}
}
\value{
\code{x} invisibly.
}
\description{
Print a causal-model summary object
}
\details{
This is a formatter for the object returned by
\code{summary.causalmixgpd_causal_fit()}. It prints the propensity-score summary
first when that block is present, followed by the control and treated outcome
summaries on the same scale of posterior diagnostics.

No new computation is performed here. The method arranges the stored summary
tables so that the three fitted blocks can be inspected together.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{print.dpmixgpd_cluster_labels}
\alias{print.dpmixgpd_cluster_labels}
\title{Print cluster labels}
\usage{
\method{print}{dpmixgpd_cluster_labels}(x, ...)
}
\arguments{
\item{x}{Cluster labels object.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly.
}
\description{
Compact display for a representative clustering.
}
\details{
A \code{dpmixgpd_cluster_labels} object represents one partition, usually the Dahl
representative partition for the training data or the induced allocation of
\code{newdata} to those representative clusters. The printed output therefore
describes the selected labels and their sizes, not the full posterior
uncertainty over alternative partitions.

When the object comes from \code{predict(..., return_scores = TRUE)}, richer
assignment information is carried alongside the labels and can be inspected
with \code{\link[=summary.dpmixgpd_cluster_labels]{summary.dpmixgpd_cluster_labels()}} or
\code{\link[=plot.dpmixgpd_cluster_labels]{plot.dpmixgpd_cluster_labels()}}.
}
\seealso{
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=summary.dpmixgpd_cluster_labels]{summary.dpmixgpd_cluster_labels()}},
\code{\link[=plot.dpmixgpd_cluster_labels]{plot.dpmixgpd_cluster_labels()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cauchy.R
\name{cauchy_mix}
\alias{cauchy_mix}
\alias{dCauchyMix}
\alias{pCauchyMix}
\alias{rCauchyMix}
\alias{qCauchyMix}
\title{Cauchy mixture distribution}
\usage{
dCauchyMix(x, w, location, scale, log = 0)

pCauchyMix(q, w, location, scale, lower.tail = 1, log.p = 0)

rCauchyMix(n, w, location, scale)

qCauchyMix(
  p,
  w,
  location,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{location, scale}{Numeric vectors of length \eqn{K} giving component locations and scales.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qCauchyMix()} returns a numeric
vector with the same length as \code{p}.
}
\description{
Finite mixture of Cauchy components for symmetric heavy-tailed bulk modeling on the real line.
}
\details{
The mixture density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k f_C(x \mid \ell_k, s_k),
}
with normalized weights \eqn{\tilde{w}_k}. These scalar functions are NIMBLE-compatible; for
vectorized R usage, use \code{\link[=cauchy_mix_lowercase]{cauchy_mix_lowercase()}}.

The mixture CDF is the weighted average of component CDFs,
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k
\left\{\frac{1}{2} + \frac{1}{\pi}\arctan\left(\frac{x-\ell_k}{s_k}\right)\right\}.
}
Random generation first selects a component according to the normalized weights and then draws
from the chosen Cauchy law by inverse-CDF sampling.

Because each Cauchy component has undefined mean and variance, the mixture also lacks an
ordinary mean in general. That is why the package exposes Cauchy kernels for densities, CDFs,
quantiles, medians, survival functions, and restricted means, but not for ordinary predictive
means.
}
\section{Functions}{
\itemize{
\item \code{dCauchyMix()}: Cauchy mixture density

\item \code{pCauchyMix()}: Cauchy mixture distribution function

\item \code{rCauchyMix()}: Cauchy mixture random generation

\item \code{qCauchyMix()}: Cauchy mixture quantile function

}}
\examples{
w <- c(0.50, 0.30, 0.20)
location <- c(-2, 0, 3)
scale <- c(1.0, 0.7, 1.5)

dCauchyMix(0.5, w = w, location = location, scale = scale, log = FALSE)
pCauchyMix(0.5, w = w, location = location, scale = scale,
           lower.tail = TRUE, log.p = FALSE)
qCauchyMix(0.50, w = w, location = location, scale = scale)
qCauchyMix(0.95, w = w, location = location, scale = scale)
replicate(10, rCauchyMix(1, w = w, location = location, scale = scale))

}
\seealso{
\code{\link[=cauchy]{cauchy()}}, \code{\link[=cauchy_mix_lowercase]{cauchy_mix_lowercase()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.
}
\concept{cauchy kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{predict.causalmixgpd_causal_fit}
\alias{predict.causalmixgpd_causal_fit}
\title{Predict arm-specific and contrast-scale quantities from a causal fit}
\usage{
\method{predict}{causalmixgpd_causal_fit}(
  object,
  newdata = NULL,
  y = NULL,
  ps = NULL,
  id = NULL,
  type = c("mean", "quantile", "density", "survival", "prob", "sample"),
  p = NULL,
  index = NULL,
  nsim = NULL,
  interval = "credible",
  probs = c(0.025, 0.5, 0.975),
  store_draws = TRUE,
  nsim_mean = 200L,
  ncores = 1L,
  show_progress = TRUE,
  ...
)
}
\arguments{
\item{object}{A \code{"causalmixgpd_causal_fit"} object returned by
\code{run_mcmc_causal()}.}

\item{newdata}{Optional new data. If \code{NULL}, uses training design (if stored).}

\item{y}{Numeric vector of evaluation points (required for \code{type="density"} or \code{"survival"}).}

\item{ps}{Optional numeric vector of propensity scores aligned with
\code{newdata}. When provided, the supplied scores are used instead of
recomputing them from the stored PS model (needed only for custom inputs).}

\item{id}{Optional identifier for prediction rows. Provide either a column name
in \code{newdata} or a vector of length \code{nrow(newdata)}. The id column
is excluded from analysis.}

\item{type}{Prediction type:
\itemize{
\item \code{"mean"}: posterior predictive mean treatment effect
\item \code{"quantile"}: posterior predictive quantile treatment effect
\item \code{"density"}: arm-specific posterior predictive densities
\item \code{"survival"}: arm-specific posterior predictive survival functions
\item \code{"prob"}: arm-specific posterior predictive probabilities
\item \code{"sample"}: paired posterior predictive samples
}}

\item{p}{Numeric vector of probabilities for quantiles (required for \code{type="quantile"}).}

\item{index}{Alias for \code{p}; numeric vector of quantile levels.}

\item{nsim}{Number of posterior predictive samples when \code{type = "sample"}.}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{probs}{Quantiles for credible interval bands.}

\item{store_draws}{Logical; whether to store treatment-effect sample draws in
the returned object when \code{type = "sample"}.}

\item{nsim_mean}{Number of posterior predictive samples used by
simulation-based mean targets. Ignored for analytical
\code{type = "mean"}; still used for \code{type = "rmean"}.}

\item{ncores}{Number of CPU cores to use for parallel prediction (if supported).}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}

\item{...}{Additional arguments forwarded to per-arm
\code{\link{predict.mixgpd_fit}} calls.}
}
\value{
For \code{"mean"} and \code{"quantile"}, a causal prediction object
whose \code{$fit} component reports treated-minus-control posterior
summaries. For \code{"density"}, \code{"survival"}, and \code{"prob"},
the \code{$fit} component contains side-by-side treated and control
summaries evaluated on the supplied \code{y} grid. For \code{"sample"},
the returned object contains paired treated, control, and treatment-effect
posterior predictive samples. Sample outputs also include long-form data
frames \code{$fit_df}, \code{$trt_fit_df}, and \code{$con_fit_df} for
direct extraction.
}
\description{
\code{predict.causalmixgpd_causal_fit()} is the causal counterpart to
\code{\link{predict.mixgpd_fit}}. It coordinates the treated and control arm
predictions so that both sides use the same covariate rows and the same PS
adjustment.
}
\details{
For each prediction row \eqn{x}, the function evaluates arm-specific
posterior predictive quantities based on
\eqn{F_1(y \mid x)} and \eqn{F_0(y \mid x)}. Mean
and quantile outputs are returned on the treatment-effect scale, while
density, survival, and probability outputs retain both arm-specific curves.
For outcome kernels with a finite analytical mean, the mean path uses
analytical per-draw means; restricted means remain simulation-based.

If a PS model is stored in the fit, the same estimated score is supplied to
both arms unless the user overrides it with \code{ps}. This is the main
prediction entry point used internally by \code{\link{ate}}, \code{\link{qte}},
\code{\link{cate}}, and \code{\link{cqte}}.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb)
predict(fit, newdata = X[1:10, , drop = FALSE], type = "quantile", index = c(0.25, 0.5, 0.75))
predict(fit, newdata = X[1:10, ], type = "mean", interval = "hpd")  # HPD intervals
predict(fit, newdata = X[1:10, ], type = "mean", interval = NULL)   # No intervals
}
}
\seealso{
\code{\link{predict.mixgpd_fit}}, \code{\link{ate}},
\code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{qtt}
\alias{qtt}
\title{Quantile treatment effects standardized to treated covariates}
\usage{
qtt(
  fit,
  probs = c(0.1, 0.5, 0.9),
  newdata = NULL,
  y = NULL,
  interval = "credible",
  level = 0.95,
  show_progress = TRUE
)
}
\arguments{
\item{fit}{A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.}

\item{probs}{Numeric vector of probabilities in (0, 1) specifying the quantile levels
of the outcome distribution to estimate treatment effects at.}

\item{newdata}{Ignored for marginal estimands. If supplied, a warning is issued
and training data are used.}

\item{y}{Ignored for marginal estimands. If supplied, a warning is issued
and training data are used.}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{level}{Numeric credible level for intervals (default 0.95 for 95 percent CI).}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}
}
\value{
An object of class \code{"causalmixgpd_qte"} containing the QTT
summary, the probability grid, and the arm-specific predictive objects used
in the aggregation. The returned object includes a top-level
\code{$fit_df} data frame for direct extraction.
}
\description{
\code{qtt()} computes the quantile treatment effect on the treated.
}
\details{
The estimand is
\deqn{\mathrm{QTT}(\tau) = Q_1^{t}(\tau) - Q_0^{t}(\tau),}
where marginalization is over the empirical covariate distribution of the
treated units only.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
qtt(fit, probs = c(0.5, 0.9))
}
}
\seealso{
\code{\link{qte}}, \code{\link{cqte}}, \code{\link{att}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/1-registry.R
\name{get_tail_registry}
\alias{get_tail_registry}
\title{Get tail registry}
\usage{
get_tail_registry()
}
\value{
A list of tail metadata.
}
\description{
Get tail registry
}
\details{
The tail registry records the generalized Pareto splice used by bulk-tail
models. It stores the tail parameter names
\code{threshold}, \code{tail_scale}, and \code{tail_shape}, together with the support each
parameter must satisfy and the modeling modes the builders may assign to
them.

In mathematical terms, for a threshold \eqn{u} the upper tail is represented
with a generalized Pareto law for excesses above \eqn{u}. Accessing this
registry is useful when inspecting how the package encodes those tail
parameters before model compilation.
}
\examples{
init_kernel_registry()
tail <- get_tail_registry()
tail$params
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_ate}
\alias{summary.causalmixgpd_ate}
\title{Summarize an ATE-style effect object}
\usage{
\method{summary}{causalmixgpd_ate}(object, ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_ate"} object from \code{ate()}.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.causalmixgpd_ate"} with
\code{overall}, \code{ate_stats}, \code{effect_table},
\code{ci_summary}, \code{meta}, and the original \code{object}.
}
\description{
\code{summary.causalmixgpd_ate()} converts ATE, ATT, CATE, or restricted-mean
output into a tabular summary suitable for reporting.
}
\details{
The summary reorganizes posterior treatment-effect output on the mean scale.
Depending on the source object, the target estimand is a marginal ATE,
treated-standardized ATT, conditional ATE, or a restricted-mean contrast
based on
\deqn{\int_0^c \{S_1(t) - S_0(t)\}\,dt.}

The returned object stores overall metadata, effect tables, and interval
summaries in a reporting-friendly format. It does not refit the model or
recompute arm-specific predictions.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
a <- ate(fit, interval = "credible")
summary(a)
}
}
\seealso{
\code{\link{print.causalmixgpd_ate}},
\code{\link{plot.causalmixgpd_ate}}, \code{\link{ate}},
\code{\link{cate}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.causalmixgpd_causal_fit}
\alias{plot.causalmixgpd_causal_fit}
\alias{plot.mixgpd_fit}
\alias{plot.mixgpd_fitted}
\title{Plot the treated and control outcome fits from a causal model}
\usage{
\method{plot}{causalmixgpd_causal_fit}(x, arm = "treated", ...)

\method{plot}{mixgpd_fit}(x, family = "traceplot", params = NULL, nLags = 50, ...)

\method{plot}{mixgpd_fitted}(x, y = NULL, ...)
}
\arguments{
\item{x}{Object of class \code{mixgpd_fitted} from \code{fitted.mixgpd_fit()}.}

\item{arm}{Integer or character; \code{1} or \code{"treated"} for treatment,
\code{0} or \code{"control"} for control, or \code{"both"} for both arms.}

\item{...}{Additional arguments (ignored).}

\item{family}{Character vector of plot names (ggmcmc plot types)
or a single one. The default \code{"traceplot"} returns one diagnostic.
Use \code{"all"} to include all plots supported for the available number
of chains/parameters. \code{"auto"} is accepted as an alias for the
default single diagnostic. Supported types:
\itemize{
\item \code{"histogram"}: posterior histograms
\item \code{"density"}: posterior density curves
\item \code{"traceplot"}: MCMC trace plots
\item \code{"running"}: running mean plots
\item \code{"compare_partial"}: partial chain comparisons
\item \code{"autocorrelation"}: autocorrelation plots
\item \code{"crosscorrelation"}: cross-correlation matrix
\item \code{"Rhat"}: Gelman--Rubin R-hat (2+ chains)
\item \code{"grb"}: Gelman--Rubin--Brooks (2+ chains)
\item \code{"effective"}: effective sample size
\item \code{"geweke"}: Geweke diagnostic
\item \code{"caterpillar"}: caterpillar/forest plots
\item \code{"pairs"}: pairwise scatter plots (2+ params)
}}

\item{params}{Optional parameter selector:
\itemize{
\item character vector of parameter patterns (exact names
or partial matches)
\item a single regex string
(e.g. \code{"alpha|threshold|tail_"})
\item \code{NULL} (default): plots all monitored parameters
}}

\item{nLags}{Number of lags for autocorrelation (ggmcmc).}

\item{y}{Ignored; included for S3 compatibility.}
}
\value{
The result of the underlying plot call (invisibly).

Invisibly returns a named list of ggplot objects.

Invisibly returns a list with the two plots.
}
\description{
\code{plot.causalmixgpd_causal_fit()} is a convenience router to the
underlying one-arm diagnostic plots for the treated and control fits.

Uses ggmcmc to produce standard MCMC diagnostic plots. Works with 1+ chains.

S3 method for visualizing fitted values from \code{fitted.mixgpd_fit()}.
Produces a 2-panel figure: Q-Q plot and residuals vs fitted.
}
\details{
Each arm-specific outcome model is itself a \code{mixgpd_fit}, so this method
delegates to \code{plot.mixgpd_fit()} for the selected arm. With \code{arm = "both"},
it returns a named list of treated and control diagnostics so the two fitted
outcome models can be assessed side by side.

These are MCMC diagnostics for the nuisance outcome models, not plots of
causal estimands. Use \code{plot()} on objects from
\code{predict.causalmixgpd_causal_fit()}, \code{qte()}, or \code{ate()} when the goal is to
visualize treatment effects rather than chain behavior.

The supported plots diagnose posterior simulation quality rather than data fit.
Depending on the selected \code{family}, they show chain traces, marginal posterior
densities, autocorrelation, cross-correlation, running means, or Gelman-style
convergence summaries for the monitored parameters.

These graphics should be read before interpreting posterior summaries or
treatment-effect results. Poor mixing or strong autocorrelation in the MCMC
output can invalidate downstream summaries even when the fitted model itself
is correctly specified.

These diagnostics compare the fitted values implied by the posterior summary
on the training design against the observed responses. The first panel checks
how closely fitted and observed values align, while the second panel looks for
residual structure that would indicate lack of fit or remaining mean trends.

This method is distinct from posterior predictive simulation on new data. It
is a training-sample diagnostic built from \code{fitted.mixgpd_fit()} and the
corresponding residuals.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
plot(fit, family = c("traceplot", "density"))
}
}
\seealso{
\code{\link{plot.mixgpd_fit}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.compute_psm}
\alias{.compute_psm}
\title{Compute posterior similarity matrix}
\usage{
.compute_psm(z_matrix)
}
\arguments{
\item{z_matrix}{Integer matrix (iterations x N) of cluster assignments.}
}
\value{
Symmetric N x N matrix of co-clustering probabilities.
}
\description{
Computes the posterior similarity matrix (PSM) from a matrix of cluster
assignments. \code{PSM[i,j]} = probability that observations i and j are in the
same cluster.
}
\details{
If \eqn{z_i^{(s)}} denotes the cluster label of observation \eqn{i} at draw
\eqn{s}, then this helper computes
\deqn{\mathrm{PSM}_{ij} \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}).}
The resulting matrix is the basic posterior co-clustering summary used by the
Dahl representative partition and several cluster diagnostics.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal_lowercase}
\alias{normal_lowercase}
\alias{dnormmix}
\alias{pnormmix}
\alias{qnormmix}
\alias{rnormmix}
\alias{dnormmixgpd}
\alias{pnormmixgpd}
\alias{qnormmixgpd}
\alias{rnormmixgpd}
\alias{dnormgpd}
\alias{pnormgpd}
\alias{qnormgpd}
\alias{rnormgpd}
\title{Lowercase vectorized normal distribution functions}
\usage{
dnormmix(x, w, mean, sd, log = FALSE)

pnormmix(q, w, mean, sd, lower.tail = TRUE, log.p = FALSE)

qnormmix(
  p,
  w,
  mean,
  sd,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rnormmix(n, w, mean, sd)

dnormmixgpd(x, w, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)

pnormmixgpd(
  q,
  w,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qnormmixgpd(
  p,
  w,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rnormmixgpd(n, w, mean, sd, threshold, tail_scale, tail_shape)

dnormgpd(x, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)

pnormgpd(
  q,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qnormgpd(
  p,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

rnormgpd(n, mean, sd, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{mean, sd}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar normal-kernel topics in this file. These helpers are meant
for interactive use and examples rather than direct use inside NIMBLE code.
}
\details{
These wrappers vectorize the scalar normal-kernel routines for ordinary R use. They preserve the
same formulas, parameter meanings, and tail construction as the uppercase functions; the only
change is that \code{x}, \code{q}, \code{p}, and \code{n} may now be length greater than one.

For the mixture quantile and splice quantile functions, the numerical and piecewise logic is
delegated directly to the corresponding scalar routine. As a result, the lowercase helpers are
faithful front ends rather than separate implementations.
}
\section{Functions}{
\itemize{
\item \code{dnormmix()}: Normal mixture density (vectorized)

\item \code{pnormmix()}: Normal mixture distribution function (vectorized)

\item \code{qnormmix()}: Normal mixture quantile function (vectorized)

\item \code{rnormmix()}: Normal mixture random generation (vectorized)

\item \code{dnormmixgpd()}: Normal mixture + GPD density (vectorized)

\item \code{pnormmixgpd()}: Normal mixture + GPD distribution function (vectorized)

\item \code{qnormmixgpd()}: Normal mixture + GPD quantile function (vectorized)

\item \code{rnormmixgpd()}: Normal mixture + GPD random generation (vectorized)

\item \code{dnormgpd()}: Normal + GPD density (vectorized)

\item \code{pnormgpd()}: Normal + GPD distribution function (vectorized)

\item \code{qnormgpd()}: Normal + GPD quantile function (vectorized)

\item \code{rnormgpd()}: Normal + GPD random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.25, 0.15)
mu <- c(-1, 0.5, 2)
sig <- c(1, 0.7, 1.3)

# Normal mixture
dnormmix(c(0, 1, 2), w = w, mean = mu, sd = sig)
rnormmix(5, w = w, mean = mu, sd = sig)

# Normal mixture + GPD
dnormmixgpd(c(1, 2, 3), w = w, mean = mu, sd = sig,
            threshold = 2, tail_scale = 1, tail_shape = 0.2)

# Normal + GPD (single component)
dnormgpd(c(1, 2, 3), mean = 0.5, sd = 1, threshold = 2,
         tail_scale = 1, tail_shape = 0.2)

}
\seealso{
\code{\link[=normal_mix]{normal_mix()}}, \code{\link[=normal_mixgpd]{normal_mixgpd()}}, \code{\link[=normal_gpd]{normal_gpd()}}, \code{\link[=bundle]{bundle()}}, \code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{att}
\alias{att}
\title{Average treatment effects standardized to treated covariates}
\usage{
att(
  fit,
  newdata = NULL,
  y = NULL,
  type = c("mean", "rmean"),
  cutoff = NULL,
  interval = "credible",
  level = 0.95,
  nsim_mean = 200L,
  show_progress = TRUE
)
}
\arguments{
\item{fit}{A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.}

\item{newdata}{Ignored for marginal estimands. If supplied, a warning is issued
and training data are used.}

\item{y}{Ignored for marginal estimands. If supplied, a warning is issued
and training data are used.}

\item{type}{Character; type of mean treatment effect:
\itemize{
\item \code{"mean"} (default): ordinary mean ATE
\item \code{"rmean"}: restricted-mean ATE (requires \code{cutoff})
}}

\item{cutoff}{Finite numeric cutoff for restricted mean; required for
\code{type = "rmean"}, ignored otherwise.}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{level}{Numeric credible level for intervals (default 0.95 for 95 percent CI).}

\item{nsim_mean}{Number of posterior predictive draws used by simulation-based
mean targets. Ignored for analytical ordinary means.}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}
}
\value{
An object of class \code{"causalmixgpd_ate"} containing the ATT
summary, optional intervals, and the arm-specific predictive objects used
in the aggregation. The returned object includes a top-level
\code{$fit_df} data frame for direct extraction.
}
\description{
\code{att()} computes the average treatment effect on the treated.
}
\details{
The estimand is
\deqn{\mathrm{ATT} = E\{Y(1) - Y(0) \mid A = 1\},}
approximated by marginalizing over the empirical covariate distribution of
treated units.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
att(fit, interval = "credible", nsim_mean = 100)
}
}
\seealso{
\code{\link{ate}}, \code{\link{qtt}}, \code{\link{cate}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.truncate_draws_matrix_components}
\alias{.truncate_draws_matrix_components}
\title{Truncate component draws in a draws matrix}
\usage{
.truncate_draws_matrix_components(object, mat, epsilon)
}
\arguments{
\item{object}{A mixgpd_fit object.}

\item{mat}{Numeric matrix of draws (iter x parameters).}

\item{epsilon}{Numeric in [0,1). Truncation level.}
}
\value{
Numeric matrix with truncated components.
}
\description{
Truncate component draws in a draws matrix
}
\details{
Posterior draw matrices often contain more components than are effectively
needed for reporting. This helper applies the package truncation rule
draw-by-draw, keeping the retained component blocks, associated weights, and
any linked coefficient matrices aligned after reordering and truncation.

The bookkeeping attached to the returned matrix records both the cumulative
mass rule and the per-component weight rule so later summaries can report how
many components were effectively retained.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal_mixgpd}
\alias{normal_mixgpd}
\alias{dNormMixGpd}
\alias{pNormMixGpd}
\alias{rNormMixGpd}
\alias{qNormMixGpd}
\title{Normal mixture with a GPD tail}
\usage{
dNormMixGpd(x, w, mean, sd, threshold, tail_scale, tail_shape, log = 0)

pNormMixGpd(
  q,
  w,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rNormMixGpd(n, w, mean, sd, threshold, tail_scale, tail_shape)

qNormMixGpd(
  p,
  w,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}.}

\item{mean, sd}{Numeric vectors of length \eqn{K} giving component means and standard deviations.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qNormMixGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to a normal mixture bulk.
This matches the structure used by the package's normal \code{mixgpd} kernels.
}
\details{
If \eqn{F_{mix}} denotes the normal-mixture CDF, the spliced CDF is
\deqn{
F(x) =
\left\{
\begin{array}{ll}
F_{mix}(x), & x < u, \\
F_{mix}(u) + \{1 - F_{mix}(u)\} G(x), & x \ge u,
\end{array}
\right.
}
where \code{threshold = u} and \eqn{G} is the GPD CDF.

The construction keeps the normal mixture unchanged below the threshold \eqn{u} and replaces the
upper tail by a generalized Pareto exceedance model. Writing \eqn{F_{mix}(u)=p_u}, the spliced
density is
\deqn{
f(x) =
\left\{
\begin{array}{ll}
f_{mix}(x), & x < u, \\
\{1-p_u\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
This formulation preserves total probability because the GPD is attached only to the residual
survival mass above the bulk threshold.

The quantile is piecewise. If \eqn{p \le p_u}, \code{qNormMixGpd()} inverts the bulk mixture
CDF; otherwise it rescales the tail probability to \eqn{(p-p_u)/(1-p_u)} and applies the
closed-form GPD quantile. That same piecewise logic is what the fitted-model prediction code
uses draw by draw.
}
\section{Functions}{
\itemize{
\item \code{dNormMixGpd()}: Normal mixture + GPD tail density

\item \code{pNormMixGpd()}: Normal mixture + GPD tail distribution function

\item \code{rNormMixGpd()}: Normal mixture + GPD tail random generation

\item \code{qNormMixGpd()}: Normal mixture + GPD tail quantile function

}}
\examples{
w <- c(0.60, 0.25, 0.15)
mean <- c(-1, 0.5, 2.0)
sd <- c(1.0, 0.7, 1.3)
threshold <- 2
tail_scale <- 1.0
tail_shape <- 0.2

dNormMixGpd(3.0, w, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)
pNormMixGpd(3.0, w, mean, sd, threshold, tail_scale, tail_shape,
           lower.tail = TRUE, log.p = FALSE)
qNormMixGpd(0.50, w, mean, sd, threshold, tail_scale, tail_shape)
qNormMixGpd(0.95, w, mean, sd, threshold, tail_scale, tail_shape)
replicate(10, rNormMixGpd(1, w, mean, sd, threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=normal_mix]{normal_mix()}}, \code{\link[=normal_gpd]{normal_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=normal_lowercase]{normal_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other normal kernel families: 
\code{\link{normal_gpd}},
\code{\link{normal_mix}}
}
\concept{normal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0-base-distributions.R
\name{amoroso}
\alias{amoroso}
\alias{dAmoroso}
\alias{pAmoroso}
\alias{rAmoroso}
\alias{qAmoroso}
\title{Amoroso distribution}
\usage{
dAmoroso(x, loc, scale, shape1, shape2, log = 0)

pAmoroso(q, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)

qAmoroso(p, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE)

rAmoroso(n, loc, scale, shape1, shape2)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{loc}{Numeric scalar location parameter.}

\item{scale}{Numeric scalar scale parameter.}

\item{shape1}{Numeric scalar first shape parameter.}

\item{shape2}{Numeric scalar second shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. The quantile function returns a numeric
scalar or vector matching the length of \code{p}.
}
\description{
Scalar Amoroso utilities used as flexible positive-support base kernels and mixture components.
The package parameterization uses \code{loc}, \code{scale}, \code{shape1}, and \code{shape2}, allowing the family
to represent a broad range of skewed bulk shapes.
}
\details{
Writing \code{loc = a}, \code{scale = theta}, \code{shape1 = alpha}, and \code{shape2 = beta}, the transformed
quantity \eqn{((X - a) / \theta)^\beta} follows a gamma law with shape \eqn{\alpha}.

These uppercase NIMBLE-compatible functions are scalar (\code{x}/\code{q} and \code{n = 1}).
For vectorized R usage, use \code{\link[=base_lowercase]{base_lowercase()}}.

The Amoroso family used in the package is defined by the density
\deqn{
f(x) =
\left|\frac{\beta}{\theta}\right|
\frac{z^{\alpha \beta - 1}\exp(-z^\beta)}{\Gamma(\alpha)},
\qquad
z = \frac{x-a}{\theta},
}
on the side of the location parameter determined by the sign of \eqn{\theta}. Equivalently,
\eqn{Z = ((X-a)/\theta)^\beta} follows a Gamma distribution with shape \eqn{\alpha} and unit
scale. That representation explains why the quantile function is computed from a gamma quantile
and then mapped back through the inverse transformation.

The mean exists whenever \eqn{\alpha + 1/\beta} lies in the domain of the gamma function used by
the moment formula. In the package this family serves as a flexible positive-support bulk kernel
capable of reproducing gamma-like, Weibull-like, and other skewed shapes with a single
parameterization.
}
\section{Functions}{
\itemize{
\item \code{dAmoroso()}: Density Function of Amoroso Distribution

\item \code{pAmoroso()}: Distribution Function of Amoroso Distribution

\item \code{qAmoroso()}: Quantile Function of Amoroso Distribution

\item \code{rAmoroso()}: Sample generating Function of Amoroso Distribution

}}
\examples{
loc <- 0
scale <- 1.5
shape1 <- 2
shape2 <- 1.2

dAmoroso(1.0, loc, scale, shape1, shape2, log = 0)
pAmoroso(1.0, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)
qAmoroso(0.50, loc, scale, shape1, shape2)
qAmoroso(0.95, loc, scale, shape1, shape2)
replicate(10, rAmoroso(1, loc, scale, shape1, shape2))
}
\seealso{
\code{\link[=amoroso_mix]{amoroso_mix()}}, \code{\link[=amoroso_gpd]{amoroso_gpd()}}, \code{\link[=base_lowercase]{base_lowercase()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other base bulk distributions: 
\code{\link{InvGauss}},
\code{\link{cauchy}}
}
\concept{base bulk distributions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.wrap_scalar_p}
\alias{.wrap_scalar_p}
\title{Wrap scalar CDF to handle q/x naming and vector inputs.}
\usage{
.wrap_scalar_p(fun)
}
\description{
Wrap scalar CDF to handle q/x naming and vector inputs.
}
\details{
Different scalar CDF helpers use either \code{q} or \code{x} for their evaluation
argument. This wrapper normalizes those naming differences and then applies the
same vector-lifting strategy used elsewhere so prediction code can call the
resulting function consistently.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_ps_bundle}
\alias{print.causalmixgpd_ps_bundle}
\title{Print a propensity score bundle}
\usage{
\method{print}{causalmixgpd_ps_bundle}(x, code = FALSE, max_code_lines = 200L, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_ps_bundle"} object.}

\item{code}{Logical; if TRUE, print generated NIMBLE code for the PS model.}

\item{max_code_lines}{Integer; maximum number of code lines to print when \code{code=TRUE}.}

\item{...}{Unused.}
}
\value{
The input object (invisibly).
}
\description{
Print a propensity score bundle
}
\details{
A PS bundle is the pre-sampling representation of the treatment-assignment
model \eqn{e(x) = \Pr(A = 1 \mid X = x)}. Depending on the stored model type,
the latent linear predictor is later mapped to a probability through a logit
link, a probit link, or a naive Bayes factorization.

The printed output is limited to the structural PS choices because posterior
draws do not exist yet. Use this method as a quick check that the requested
treatment model was encoded correctly before fitting the full causal bundle.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_bundle}
\alias{print.causalmixgpd_bundle}
\title{Print a one-arm workflow bundle}
\usage{
\method{print}{causalmixgpd_bundle}(x, code = FALSE, max_code_lines = 200L, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_bundle"} object.}

\item{code}{Logical; if TRUE, print the generated NIMBLE model code.}

\item{max_code_lines}{Integer; maximum number of code lines to print when \code{code=TRUE}.}

\item{...}{Unused.}
}
\value{
The object \code{x}, invisibly.
}
\description{
\code{print.causalmixgpd_bundle()} gives a compact structural summary of the
pre-run bundle created by \code{\link{build_nimble_bundle}}.
}
\details{
The bundle is the compiled representation of the predictive model before
MCMC. For a bulk-only fit, the underlying target law is
\deqn{f(y \mid x) = \sum_{k=1}^{K} w_k(x) f_k(y \mid x, \theta_k).}
When a GPD tail is enabled, the same bulk mixture is spliced to a generalized
Pareto tail above the threshold recorded in the bundle specification.

\code{print()} is intentionally brief. It is meant to confirm that the stored
backend, kernel, truncation size, covariate structure, and code-generation
artifacts match the intended model before you compile and sample with
\code{\link{run_mcmc_bundle_manual}}.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = FALSE, components = 3)
print(bundle)
print(bundle, code = TRUE, max_code_lines = 30)
}
}
\seealso{
\code{\link{summary.causalmixgpd_bundle}}, \code{\link{mcmc}},
\code{\link{run_mcmc_bundle_manual}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/1-registry.R
\name{get_kernel_registry}
\alias{get_kernel_registry}
\title{Get kernel registry}
\usage{
get_kernel_registry()
}
\value{
A list of kernel metadata.
}
\description{
Get kernel registry
}
\details{
This accessor returns the registry created by \code{\link[=init_kernel_registry]{init_kernel_registry()}}. The
returned object is a named list keyed by kernel name. Each kernel definition
describes which bulk parameters are present, how those parameters may depend
on covariates, whether a GPD tail is allowed, and which density or mean
functions should be dispatched for the supported backends.

Downstream builders use this registry as the package-level reference for
kernel-specific implementation metadata. Reading it is appropriate when you
need to inspect what the package believes a kernel can do before constructing
or debugging a model specification.
}
\examples{
init_kernel_registry()
reg <- get_kernel_registry()
reg$normal$bulk_params
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma.R
\name{gamma_mix}
\alias{gamma_mix}
\alias{dGammaMix}
\alias{pGammaMix}
\alias{rGammaMix}
\alias{qGammaMix}
\title{Gamma mixture distribution}
\usage{
dGammaMix(x, w, shape, scale, log = 0)

pGammaMix(q, w, shape, scale, lower.tail = 1, log.p = 0)

rGammaMix(n, w, shape, scale)

qGammaMix(
  p,
  w,
  shape,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{shape, scale}{Numeric vectors of length \eqn{K} giving Gamma shape and scale parameters.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qGammaMix()} returns a numeric vector
with the same length as \code{p}.
}
\description{
Finite mixture of gamma components for positive-support bulk modeling. The scalar functions in
this topic are the compiled building blocks behind the gamma bulk kernel family.
}
\details{
The mixture density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k f_{\Gamma}(x \mid \alpha_k, \theta_k),
\qquad x > 0,
}
with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use \code{\link[=gamma_lowercase]{gamma_lowercase()}}.

Under the package parameterization, each component has density
\eqn{f_\Gamma(x \mid \alpha,\theta) = x^{\alpha-1}\exp(-x/\theta) /
\{\Gamma(\alpha)\theta^\alpha\}} on \eqn{x>0}. The mixture CDF is therefore
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k F_\Gamma(x \mid \alpha_k,\theta_k).
}
Random generation first selects a component according to the normalized mixture weights and then
draws from the corresponding gamma distribution. Since finite gamma mixtures do not have closed
form quantiles, \code{qGammaMix()} obtains them numerically by inverting the mixture CDF.

The analytical mean is
\deqn{
E(X) = \sum_{k=1}^K \tilde{w}_k \alpha_k \theta_k.
}
This expression is reused in posterior predictive mean calculations for gamma-based fits.
}
\section{Functions}{
\itemize{
\item \code{dGammaMix()}: Gamma mixture density

\item \code{pGammaMix()}: Gamma mixture distribution function

\item \code{rGammaMix()}: Gamma mixture random generation

\item \code{qGammaMix()}: Gamma mixture quantile function

}}
\examples{
w <- c(0.55, 0.30, 0.15)
scale <- c(1.0, 2.5, 5.0)
shape <- c(2, 4, 6)

dGammaMix(2.0, w = w, scale = scale, shape = shape, log = 0)
pGammaMix(2.0, w = w, scale = scale, shape = shape, lower.tail = 1, log.p = 0)
qGammaMix(0.50, w = w, scale = scale, shape = shape)
qGammaMix(0.95, w = w, scale = scale, shape = shape)
replicate(10, rGammaMix(1, w = w, scale = scale, shape = shape))
}
\seealso{
\code{\link[=gamma_mixgpd]{gamma_mixgpd()}}, \code{\link[=gamma_gpd]{gamma_gpd()}}, \code{\link[=gamma_lowercase]{gamma_lowercase()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}},
\code{\link[=kernel_support_table]{kernel_support_table()}}.

Other gamma kernel families: 
\code{\link{gamma_gpd}},
\code{\link{gamma_mixgpd}}
}
\concept{gamma kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_qte}
\alias{print.causalmixgpd_qte}
\title{Print a QTE-style effect object}
\usage{
\method{print}{causalmixgpd_qte}(x, digits = 3, max_rows = 6, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_qte"} object from \code{qte()}.}

\item{digits}{Number of digits to display.}

\item{max_rows}{Maximum number of estimate rows to display.}

\item{...}{Unused.}
}
\value{
The object \code{x}, invisibly.
}
\description{
\code{print.causalmixgpd_qte()} prints a compact summary for objects produced
by \code{\link{qte}}, \code{\link{qtt}}, or \code{\link{cqte}}.
}
\details{
These objects store posterior summaries of quantile treatment contrasts. In
the marginal case,
\deqn{\Delta(\tau) = Q_{Y^1}(\tau) - Q_{Y^0}(\tau).}
For \code{qtt()}, the same contrast is standardized to the treated covariate
distribution, and for \code{cqte()} it is evaluated conditionally at the supplied
covariate profiles.

The print method is intentionally compact: it reports the prediction setup and
the resulting effect table, but it does not attempt to reproduce all
posterior draws. Use \code{summary()} or \code{plot()} on the same object for more
structured reporting.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
q <- qte(fit, probs = c(0.25, 0.5, 0.75), interval = "credible")
print(q)
}
}
\seealso{
\code{\link{summary.causalmixgpd_qte}},
\code{\link{plot.causalmixgpd_qte}}, \code{\link{qte}},
\code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_bundle}
\alias{summary.causalmixgpd_bundle}
\title{Summarize a one-arm workflow bundle}
\usage{
\method{summary}{causalmixgpd_bundle}(object, ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_bundle"} object.}

\item{...}{Unused.}
}
\value{
An invisible list with elements \code{meta}, \code{priors}, and
\code{monitors}.
}
\description{
\code{summary.causalmixgpd_bundle()} prints the structural contents of a
bundle before MCMC is run.
}
\details{
The summary is meant for workflow validation rather than inference. It shows:
\itemize{
\item the model metadata (backend, kernel, components, covariates, GPD flag),
\item the prior/parameter table derived from \code{spec$plan},
\item the nodes that will be monitored during MCMC.
}

This is the recommended checkpoint after \code{\link{build_nimble_bundle}}
and before \code{\link{run_mcmc_bundle_manual}}.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = FALSE, components = 3)
summary(bundle)
}
}
\seealso{
\code{\link{build_nimble_bundle}}, \code{\link{print.causalmixgpd_bundle}},
\code{\link{run_mcmc_bundle_manual}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_causal_bundle}
\alias{print.causalmixgpd_causal_bundle}
\title{Print a causal workflow bundle}
\usage{
\method{print}{causalmixgpd_causal_bundle}(x, code = FALSE, max_code_lines = 200L, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_causal_bundle"} object.}

\item{code}{Logical; if TRUE, print generated NIMBLE code for each block.}

\item{max_code_lines}{Integer; maximum number of code lines to print when \code{code=TRUE}.}

\item{...}{Unused.}
}
\value{
The input object (invisibly).
}
\description{
\code{print.causalmixgpd_causal_bundle()} gives a compact structural summary
of the pre-run causal bundle created by \code{\link{build_causal_bundle}}.
}
\details{
A causal bundle collects three pre-MCMC building blocks: the optional
propensity-score model for \eqn{e(x) = \Pr(A = 1 \mid X = x)}, the control
outcome model for \eqn{Y^0}, and the treated outcome model for \eqn{Y^1}. The
printed output aligns those blocks side by side so the user can verify that
the treated and control outcome specifications are coherent before sampling.

No causal estimand is computed at this stage. The bundle only records the
structural assumptions that will later support estimands such as
\eqn{E(Y^1 - Y^0 \mid X = x)} or
\eqn{Q_{Y^1}(\tau \mid X = x) - Q_{Y^0}(\tau \mid X = x)}.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal")
print(cb)
}
}
\seealso{
\code{\link{summary.causalmixgpd_causal_bundle}},
\code{\link{run_mcmc_causal}}, \code{\link{ate}}, \code{\link{qte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.truncation_info}
\alias{.truncation_info}
\title{Summarize truncation results from draws}
\usage{
.truncation_info(object, epsilon = NULL)
}
\arguments{
\item{object}{A mixgpd_fit.}

\item{epsilon}{Numeric; optional override.}
}
\value{
List with k summary.
}
\description{
Summarize truncation results from draws
}
\details{
After \code{.extract_draws()} applies draw-level component truncation, the chosen
number of retained components is stored as an attribute. This helper condenses
that bookkeeping into min, median, and max summaries for the effective
retained component count under both the cumulative-mass and per-weight
criteria.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_predict_plots}
\alias{print.mixgpd_predict_plots}
\title{Print method for prediction plots}
\usage{
\method{print}{mixgpd_predict_plots}(x, ...)
}
\arguments{
\item{x}{Object of class \code{mixgpd_predict_plots}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns the input object.
}
\description{
Print method for prediction plots
}
\details{
Prediction plotting methods may return a single plot or a richer plot object
with an additional wrapper class. This print method temporarily drops that
wrapper class so the underlying graphics object uses its native print method.

The stored predictive summaries are not changed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.get_dispatch}
\alias{.get_dispatch}
\title{Resolve kernel dispatch functions
Dispatch returns vector-aware d/p/q and n-aware r via wrappers; do not mutate namespace.}
\usage{
.get_dispatch(spec_or_fit, backend_override = NULL, gpd_override = NULL)
}
\arguments{
\item{spec_or_fit}{mixgpd_fit or spec list}
}
\value{
List with d/p/q/r/mean/mean_trunc functions and bulk_params.
}
\description{
Resolve kernel dispatch functions
Dispatch returns vector-aware d/p/q and n-aware r via wrappers; do not mutate namespace.
}
\details{
This is the prediction-oriented companion to \code{.get_dispatch_scalar()}. It
starts from the same kernel dispatch lookup, then wraps the scalar functions
so they can accept vector inputs and the package's preferred argument naming
conventions in ordinary R evaluation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.predict_mixgpd}
\alias{.predict_mixgpd}
\title{Internal prediction engine: evaluate per posterior draw, then summarize.}
\usage{
.predict_mixgpd(
  object,
  x = NULL,
  y = NULL,
  ps = NULL,
  id = NULL,
  type = c("density", "survival", "quantile", "sample", "mean", "rmean", "median", "fit"),
  p = NULL,
  index = NULL,
  nsim = NULL,
  level = 0.95,
  interval = "credible",
  probs = c(0.025, 0.5, 0.975),
  store_draws = TRUE,
  nsim_mean = 200L,
  cutoff = NULL,
  ndraws_pred = NULL,
  chunk_size = NULL,
  show_progress = TRUE,
  ncores = 1L,
  sample_draw_idx = NULL
)
}
\description{
Project rules:
\itemize{
\item density/survival: either provide both (x,y) or neither (defaults to training X and training y).
\item quantile/sample/mean: y must be NULL; x may be provided (new X) or NULL (defaults to training X).
\item CRP predictions use posterior weights derived from z for each draw.
\item Stores per-draw results in object$cache$predict (environment) for reuse in treatment effects.
}
}
\details{
This is the main internal workhorse behind \code{predict.mixgpd_fit()} and the
causal effect helpers. It evaluates the requested predictive functional
separately for each retained posterior draw, using either explicit SB weights
or CRP weights reconstructed from latent cluster labels, and only then
collapses the draw-level results into posterior summaries.

The helper also manages caching of per-draw predictive quantities because
treatment-effect functions repeatedly reuse the same arm-specific predictive
draws. That cache avoids recomputation while keeping the public prediction
interface simple.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal_gpd}
\alias{normal_gpd}
\alias{dNormGpd}
\alias{pNormGpd}
\alias{rNormGpd}
\alias{qNormGpd}
\title{Normal with a GPD tail}
\usage{
dNormGpd(x, mean, sd, threshold, tail_scale, tail_shape, log = 0)

pNormGpd(
  q,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rNormGpd(n, mean, sd, threshold, tail_scale, tail_shape)

qNormGpd(
  p,
  mean,
  sd,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{mean}{Numeric scalar mean parameter for the Normal bulk.}

\item{sd}{Numeric scalar standard deviation for the Normal bulk.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qNormGpd()} returns a numeric
vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
normal bulk component.
}
\details{
This is the single-component version of \code{\link[=normal_mixgpd]{normal_mixgpd()}}. If \eqn{\Phi_u =
\Phi((u-\mu)/\sigma)} denotes the normal bulk probability below the threshold, then the density
is
\deqn{
f(x) =
\left\{
\begin{array}{ll}
\phi(x \mid \mu, \sigma^2), & x < u, \\
(1-\Phi_u) g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
The distribution is continuous at \eqn{u} by construction, although the derivative generally
changes there because the tail is modeled by a different family.

The ordinary mean exists only when the GPD tail satisfies \eqn{\xi < 1}. When that condition
fails, downstream mean prediction is intentionally blocked and the package directs the user to
restricted means or quantile-based summaries instead.
}
\section{Functions}{
\itemize{
\item \code{dNormGpd()}: Normal + GPD tail density

\item \code{pNormGpd()}: Normal + GPD tail distribution function

\item \code{rNormGpd()}: Normal + GPD tail random generation

\item \code{qNormGpd()}: Normal + GPD tail quantile function

}}
\examples{
mean <- 0.5
sd <- 1.0
threshold <- 2
tail_scale <- 1.0
tail_shape <- 0.2

dNormGpd(3.0, mean, sd, threshold, tail_scale, tail_shape, log = FALSE)
pNormGpd(3.0, mean, sd, threshold, tail_scale, tail_shape,
        lower.tail = TRUE, log.p = FALSE)
qNormGpd(0.50, mean, sd, threshold, tail_scale, tail_shape)
qNormGpd(0.95, mean, sd, threshold, tail_scale, tail_shape)
replicate(10, rNormGpd(1, mean, sd, threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=normal_mix]{normal_mix()}}, \code{\link[=normal_mixgpd]{normal_mixgpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=normal_lowercase]{normal_lowercase()}}.

Other normal kernel families: 
\code{\link{normal_mix}},
\code{\link{normal_mixgpd}}
}
\concept{normal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_qte}
\alias{summary.causalmixgpd_qte}
\title{Summarize a QTE-style effect object}
\usage{
\method{summary}{causalmixgpd_qte}(object, ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_qte"} object from \code{qte()}.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.causalmixgpd_qte"} with
\code{overall}, \code{quantile_summary}, \code{effect_table},
\code{ci_summary}, \code{meta}, and the original \code{object}.
}
\description{
\code{summary.causalmixgpd_qte()} converts QTE, QTT, or CQTE output into a
tabular summary suitable for reporting.
}
\details{
The summary reorganizes the posterior effect object into reporting tables. The
target estimand remains a quantile contrast,
\deqn{\Delta(\tau) = Q_{Y^1}(\tau) - Q_{Y^0}(\tau),}
with the appropriate marginal, treated-standardized, or conditional
interpretation depending on whether the source object came from \code{qte()},
\code{qtt()}, or \code{cqte()}.

Besides the effect table itself, the summary records the quantile grid, the
interval settings, and per-quantile distributional summaries when posterior
draws are available. This makes the object convenient for reporting and
downstream printing without recomputing the estimand.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
q <- qte(fit, probs = c(0.25, 0.5, 0.75), interval = "credible")
summary(q)
}
}
\seealso{
\code{\link{print.causalmixgpd_qte}},
\code{\link{plot.causalmixgpd_qte}}, \code{\link{qte}},
\code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.causalmixgpd_qte}
\alias{plot.causalmixgpd_qte}
\alias{plot.causalmixgpd_ate}
\title{Plot QTE-style effect summaries}
\usage{
\method{plot}{causalmixgpd_qte}(
  x,
  y = NULL,
  type = c("both", "effect", "arms"),
  facet_by = c("tau", "id"),
  plotly = getOption("CausalMixGPD.plotly", FALSE),
  ...
)

\method{plot}{causalmixgpd_ate}(
  x,
  y = NULL,
  type = c("both", "effect", "arms"),
  plotly = getOption("CausalMixGPD.plotly", FALSE),
  ...
)
}
\arguments{
\item{x}{Object of class \code{causalmixgpd_ate}.}

\item{y}{Ignored.}

\item{type}{Character; plot type:
\itemize{
\item \code{"both"} (default): returns a list with both
arm means and treatment-effect plots
\item \code{"effect"}: ATE curve/points with pointwise CI
error bars
\item \code{"arms"}: treated vs control mean with
pointwise CI error bars
}
For CATE objects, omitted \code{type} defaults to \code{"effect"}.}

\item{facet_by}{Character; faceting strategy when multiple
prediction points exist:
\itemize{
\item \code{"tau"} (default): facets by quantile level
\item \code{"id"}: facets by prediction point
}}

\item{plotly}{Logical; if \code{TRUE}, convert the \code{ggplot2} output to a
\code{plotly} / \code{htmlwidget} representation via \code{.wrap_plotly()}. Defaults
to \code{getOption("CausalMixGPD.plotly", FALSE)}.}

\item{...}{Additional arguments passed to ggplot2 functions.}
}
\value{
A list of ggplot objects with elements \code{trt_control} and \code{treatment_effect}
(if \code{type="both"}), or a single ggplot object (if \code{type} is \code{"effect"} or
\code{"arms"}).

A list of ggplot objects with elements \code{trt_control} and \code{treatment_effect}
(if \code{type="both"}), or a single ggplot object (if \code{type} is \code{"effect"} or
\code{"arms"}).
}
\description{
\code{plot.causalmixgpd_qte()} visualizes objects returned by
\code{\link{qte}}, \code{\link{qtt}}, and \code{\link{cqte}}. The
\code{type} parameter controls the plot style. When \code{type} is omitted,
\code{cqte()} objects default to \code{"effect"} and, when multiple
quantile levels are present, \code{facet_by = "id"}. Whenever quantile index
appears on the x-axis, it is shown as an ordered categorical axis with
equidistant spacing:
\itemize{
\item \code{"both"} (default): Returns a list with both \code{trt_control} (treated vs
control quantile curves) and \code{treatment_effect} (QTE curve) plots
\item \code{"effect"}: QTE curve vs quantile levels (\code{probs}) with pointwise CI error bars
\item \code{"arms"}: Treated and control quantile curves vs \code{probs}, with pointwise CI error bars
}

\code{plot.causalmixgpd_ate()} visualizes objects returned by
\code{\link{ate}}, \code{\link{att}}, \code{\link{cate}}, and
\code{\link{ate_rmean}}. The \code{type} parameter controls the plot style.
When \code{type} is omitted, \code{cate()} objects default to
\code{"effect"}:
\itemize{
\item \code{"both"} (default): Returns a list with both \code{trt_control} (treated vs
control means) and \code{treatment_effect} (ATE curve) plots
\item \code{"effect"}: ATE curve/points vs index/PS with pointwise CI error bars
\item \code{"arms"}: Treated mean vs control mean, with pointwise CI error bars
}
}
\details{
The effect view emphasizes the quantile contrast
\eqn{\tau \mapsto Q_{Y^1}(\tau) - Q_{Y^0}(\tau)}, while the arms view shows the
treated and control quantile functions that generate that contrast. For
conditional CQTE objects, faceting can separate covariate profiles so the
same quantile contrast is compared across prediction settings.

These graphics visualize posterior summaries of the effect object itself. They
are therefore downstream of model fitting and downstream of the causal
prediction step.

The effect panel visualizes the posterior summary of the treatment contrast on
the mean scale, namely \eqn{E(Y^1) - E(Y^0)} or its conditional or
treated-standardized analogue. The arms panel instead shows the treated and
control mean predictions whose difference defines that contrast.

For \code{cate()} objects, the x-axis follows the prediction profiles; otherwise it
uses the estimated propensity score when available or a simple index order.
This keeps the comparison aligned with how the effect object was standardized.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
X_new <- X[1:5, , drop = FALSE]
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
qte_result <- cqte(fit, probs = c(0.1, 0.5, 0.9), newdata = X_new)
plot(qte_result)  # CQTE default: effect plot (faceted by id when needed)
plot(qte_result, type = "effect")  # single QTE plot
plot(qte_result, type = "arms")    # single arms plot
}
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
X_new <- X[1:5, , drop = FALSE]
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
ate_result <- cate(fit, newdata = X_new, interval = "credible")
plot(ate_result)  # CATE default: effect plot
plot(ate_result, type = "effect")  # single ATE plot
plot(ate_result, type = "arms")    # single arms plot
}
}
\seealso{
\code{\link{qte}}, \code{\link{cqte}},
\code{\link{summary.causalmixgpd_qte}}.

\code{\link{ate}}, \code{\link{cate}},
\code{\link{summary.causalmixgpd_ate}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/laplace.R
\name{laplace_mix}
\alias{laplace_mix}
\alias{dLaplaceMix}
\alias{pLaplaceMix}
\alias{rLaplaceMix}
\alias{qLaplaceMix}
\title{Laplace (double exponential) mixture distribution}
\usage{
dLaplaceMix(x, w, location, scale, log = 0)

pLaplaceMix(q, w, location, scale, lower.tail = 1, log.p = 0)

rLaplaceMix(n, w, location, scale)

qLaplaceMix(
  p,
  w,
  location,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{location}{Numeric vector of length \eqn{K} giving component locations.}

\item{scale}{Numeric vector of length \eqn{K} giving component scales.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qLaplaceMix()} returns a numeric
vector with the same length as \code{p}.
}
\description{
Finite mixture of Laplace components for real-valued bulk modeling. The scalar functions in this
topic are the NIMBLE-compatible building blocks for Laplace-based kernels.
}
\details{
The mixture density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k f_{Lap}(x \mid \mu_k, b_k),
}
with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use \code{\link[=laplace_lowercase]{laplace_lowercase()}}.

Each component is a Laplace law with density
\eqn{f(x \mid \mu,b) = (2b)^{-1}\exp\{-|x-\mu|/b\}}. The mixture CDF is the corresponding
weighted average of component CDFs, and random generation selects a component first and then
samples from that component. As with the other finite mixtures in the package, the quantile has
no closed form and is therefore obtained numerically.

The analytical mean of the mixture is simply
\deqn{
E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k.
}
That is the formula used in downstream predictive mean calculations for Laplace-based fits.
}
\section{Functions}{
\itemize{
\item \code{dLaplaceMix()}: Laplace mixture density

\item \code{pLaplaceMix()}: Laplace mixture distribution function

\item \code{rLaplaceMix()}: Laplace mixture random generation

\item \code{qLaplaceMix()}: Laplace mixture quantile function

}}
\examples{
w <- c(0.50, 0.30, 0.20)
location <- c(-1, 0.5, 2.0)
scale <- c(1.0, 0.7, 1.4)

dLaplaceMix(0.8, w = w, location = location, scale = scale, log = FALSE)
pLaplaceMix(0.8, w = w, location = location, scale = scale,
           lower.tail = TRUE, log.p = FALSE)
qLaplaceMix(0.50, w = w, location = location, scale = scale)
qLaplaceMix(0.95, w = w, location = location, scale = scale)
replicate(10, rLaplaceMix(1, w = w, location = location, scale = scale))
}
\seealso{
\code{\link[=laplace_MixGpd]{laplace_MixGpd()}}, \code{\link[=laplace_gpd]{laplace_gpd()}}, \code{\link[=laplace_lowercase]{laplace_lowercase()}},
\code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other laplace kernel families: 
\code{\link{laplace_MixGpd}},
\code{\link{laplace_gpd}}
}
\concept{laplace kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{predict.mixgpd_fit}
\alias{predict.mixgpd_fit}
\title{Posterior predictive summaries from a fitted one-arm model}
\usage{
\method{predict}{mixgpd_fit}(
  object,
  newdata = NULL,
  y = NULL,
  ps = NULL,
  id = NULL,
  type = c("density", "survival", "quantile", "sample", "mean", "rmean", "median", "fit"),
  p = NULL,
  index = NULL,
  nsim = NULL,
  level = 0.95,
  interval = "credible",
  probs = c(0.025, 0.5, 0.975),
  store_draws = TRUE,
  nsim_mean = 200L,
  cutoff = NULL,
  ncores = 1L,
  show_progress = TRUE,
  ndraws_pred = NULL,
  chunk_size = NULL,
  parallel = FALSE,
  workers = NULL,
  ...
)
}
\arguments{
\item{object}{A fitted object of class \code{"mixgpd_fit"}.}

\item{newdata}{Optional new data. If \code{NULL}, uses training design (if stored).}

\item{y}{Numeric vector of evaluation points (required for \code{type="density"} or \code{"survival"}).}

\item{ps}{Optional numeric vector of propensity scores for conditional prediction.
Used when the model was fit with propensity score augmentation.}

\item{id}{Optional identifier for prediction rows. Provide either a column name
in \code{newdata} or a vector of length \code{nrow(newdata)}. The id column
is excluded from analysis.}

\item{type}{Prediction type:
\itemize{
\item \code{"density"}: Posterior predictive density f(y | x, data)
\item \code{"survival"}: Posterior predictive survival S(y | x, data) = 1 - F(y | x, data)
\item \code{"quantile"}: Posterior predictive quantiles Q(p | x, data)
\item \code{"sample"}: Posterior predictive samples Y^rep ~ f(y | x, data)
\item \code{"mean"}: Posterior predictive mean E(Y | x, data) (averaged over posterior parameter uncertainty)
\item \code{"rmean"}: Posterior predictive restricted mean \eqn{E[\min(Y, cutoff) \mid x, data]}
\item \code{"median"}: Posterior predictive median (quantile at p=0.5)
\item \code{"fit"}: Per-observation posterior predictive draws
}
Note: \code{type="mean"} returns the posterior predictive mean, which integrates over
parameter uncertainty. This differs from the mean of a single model distribution.}

\item{p}{Numeric vector of probabilities for quantiles (required for \code{type="quantile"}).}

\item{index}{Alias for \code{p}; numeric vector of quantile levels.}

\item{nsim}{Number of posterior predictive samples (for \code{type="sample"}).}

\item{level}{Credible level for credible intervals (default 0.95 for 95 percent intervals).}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile
intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{probs}{Quantiles for credible interval bands.}

\item{store_draws}{Logical; whether to store all posterior draws (for \code{type="sample"}).}

\item{nsim_mean}{Number of posterior predictive samples used by
simulation-based mean targets. Ignored for analytical
\code{type = "mean"}; still used for \code{type = "rmean"}.}

\item{cutoff}{Finite numeric cutoff for \code{type="rmean"} (restricted mean).}

\item{ncores}{Number of CPU cores to use for parallel prediction (if supported).}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}

\item{ndraws_pred}{Optional integer subsample of posterior draws for prediction speed.
If NULL and \code{nrow(newdata) > 20000}, defaults to 200.}

\item{chunk_size}{Optional row chunk size for large \code{newdata} prediction.
If NULL and \code{nrow(newdata) > 20000}, defaults to 10000.}

\item{parallel}{Logical; if TRUE, enable parallel prediction (alias for setting \code{ncores > 1}).}

\item{workers}{Optional integer worker count (alias for \code{ncores}).}

\item{...}{Unused.}
}
\value{
A list with elements:
\itemize{
\item \code{fit}: numeric vector/matrix for \code{type = "sample"}, otherwise a data frame with
\code{estimate}/\code{lower}/\code{upper} columns (posterior means over draws) plus any index
columns (e.g. \code{id}, \code{y}, \code{index}).
\item \code{fit_df}: a machine-readable data frame view of the prediction output. For
non-sample types this aliases \code{fit}; for \code{type = "sample"} it is a long-form
data frame with draw indices and sampled values.
\item \code{lower}, \code{upper}: reserved for backward compatibility (typically \code{NULL}).
\item \code{type}, \code{grid}: metadata.
}
}
\description{
\code{predict.mixgpd_fit()} is the central distributional prediction method
for fitted one-arm models.
}
\details{
The method works with posterior predictive functionals rather than raw model
parameters. Supported output types include:
\itemize{
\item \code{"density"} for \eqn{f(y \mid x)},
\item \code{"survival"} for \eqn{S(y \mid x) = 1 - F(y \mid x)},
\item \code{"quantile"} for \eqn{Q(\tau \mid x)},
\item \code{"mean"} for \eqn{E(Y \mid x)},
\item \code{"rmean"} for \eqn{E\{\min(Y, c) \mid x\}},
\item \code{"sample"} and \code{"fit"} for draw-level predictive output.
}

For spliced models these predictions integrate over both the DPM bulk and
the GPD tail using component-specific tail parameters, including link-mode
tail coefficients when present. For kernels with a finite analytical mean,
\code{type = "mean"} computes the posterior-draw mean analytically and then
summarizes those draw-level means across the posterior. The
\code{type = "rmean"} path remains a separate posterior predictive
simulation pipeline.

For kernels with an analytical mean, \code{type = "mean"} is computed
analytically within each posterior draw and then summarized over draws. For
GPD-tail fits this analytical path is used when the tail shape parameter
satisfies \eqn{\xi < 1}. If the mean does not exist analytically for the
chosen kernel or if any required GPD tail has \eqn{\xi \ge 1}, the ordinary
mean is undefined and the function errors with a message directing you to
\code{type = "rmean"} or other summaries that remain well defined.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
pr <- predict(fit, type = "quantile", p = c(0.5, 0.9))
pr_surv <- predict(fit, y = sort(y), type = "survival")
pr_cdf <- list(fit = 1 - pr_surv$fit)
# HPD intervals
pr_hpd <- predict(fit, type = "quantile", p = c(0.5, 0.9), interval = "hpd")
# No intervals
pr_none <- predict(fit, type = "quantile", p = c(0.5, 0.9), interval = NULL)
# Restricted mean (finite under heavy tails)
pr_rmean <- predict(fit, type = "rmean", cutoff = 10, interval = "credible")
}
}
\seealso{
\code{\link{summary.mixgpd_fit}}, \code{\link{fitted.mixgpd_fit}},
\code{\link{residuals.mixgpd_fit}}, \code{\link{predict.causalmixgpd_causal_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lognormal.R
\name{lognormal_lowercase}
\alias{lognormal_lowercase}
\alias{dlognormalmix}
\alias{plognormalmix}
\alias{qlognormalmix}
\alias{rlognormalmix}
\alias{dlognormalmixgpd}
\alias{plognormalmixgpd}
\alias{qlognormalmixgpd}
\alias{rlognormalmixgpd}
\alias{dlognormalgpd}
\alias{plognormalgpd}
\alias{qlognormalgpd}
\alias{rlognormalgpd}
\title{Lowercase vectorized lognormal distribution functions}
\usage{
dlognormalmix(x, w, meanlog, sdlog, log = FALSE)

plognormalmix(q, w, meanlog, sdlog, lower.tail = TRUE, log.p = FALSE)

qlognormalmix(
  p,
  w,
  meanlog,
  sdlog,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rlognormalmix(n, w, meanlog, sdlog)

dlognormalmixgpd(
  x,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

plognormalmixgpd(
  q,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qlognormalmixgpd(
  p,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rlognormalmixgpd(n, w, meanlog, sdlog, threshold, tail_scale, tail_shape)

dlognormalgpd(
  x,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

plognormalgpd(
  q,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qlognormalgpd(
  p,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

rlognormalgpd(n, meanlog, sdlog, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{meanlog, sdlog}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar lognormal-kernel topics in this file.
}
\details{
These are direct vectorized wrappers around the scalar lognormal routines. They keep the same
parameterization, support restrictions, and bulk-tail splice, while allowing ordinary vector
inputs in R. Quantile wrappers continue to use the scalar inversion logic, so there is no
separate approximation layer in the lowercase API.
}
\section{Functions}{
\itemize{
\item \code{dlognormalmix()}: Lognormal mixture density (vectorized)

\item \code{plognormalmix()}: Lognormal mixture distribution function (vectorized)

\item \code{qlognormalmix()}: Lognormal mixture quantile function (vectorized)

\item \code{rlognormalmix()}: Lognormal mixture random generation (vectorized)

\item \code{dlognormalmixgpd()}: Lognormal mixture + GPD density (vectorized)

\item \code{plognormalmixgpd()}: Lognormal mixture + GPD distribution function (vectorized)

\item \code{qlognormalmixgpd()}: Lognormal mixture + GPD quantile function (vectorized)

\item \code{rlognormalmixgpd()}: Lognormal mixture + GPD random generation (vectorized)

\item \code{dlognormalgpd()}: Lognormal + GPD density (vectorized)

\item \code{plognormalgpd()}: Lognormal + GPD distribution function (vectorized)

\item \code{qlognormalgpd()}: Lognormal + GPD quantile function (vectorized)

\item \code{rlognormalgpd()}: Lognormal + GPD random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.3, 0.1)
ml <- c(0, 0.3, 0.6)
sl <- c(0.4, 0.5, 0.6)

# Lognormal mixture
dlognormalmix(c(1, 2, 3), w = w, meanlog = ml, sdlog = sl)
rlognormalmix(5, w = w, meanlog = ml, sdlog = sl)

# Lognormal mixture + GPD
dlognormalmixgpd(c(2, 3, 4), w = w, meanlog = ml, sdlog = sl,
                 threshold = 2.5, tail_scale = 0.5, tail_shape = 0.2)

}
\seealso{
\code{\link[=lognormal_mix]{lognormal_mix()}}, \code{\link[=lognormal_mixgpd]{lognormal_mixgpd()}}, \code{\link[=lognormal_gpd]{lognormal_gpd()}}, \code{\link[=bundle]{bundle()}},
\code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{print.dpmixgpd_cluster_fit}
\alias{print.dpmixgpd_cluster_fit}
\title{Print a cluster fit}
\usage{
\method{print}{dpmixgpd_cluster_fit}(x, ...)
}
\arguments{
\item{x}{A cluster fit.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly.
}
\description{
Compact display for a fitted clustering object.
}
\details{
A fitted cluster object contains posterior draws for the latent labels and
associated component parameters. The printed header identifies the model
family that generated those draws, including whether the fit used a bulk-only
kernel or a spliced bulk-tail specification.

The printed \code{components} value is the truncation size used by the sampler. It
is not the same thing as the number of occupied clusters in the Dahl
representative partition, which is computed later by
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}} and summarized by
\code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}}.
}
\seealso{
\code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}}, \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}},
\code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.extract_nimble_code}
\alias{.extract_nimble_code}
\title{Extract nimbleCode from bundle code}
\usage{
.extract_nimble_code(code)
}
\description{
Extract nimbleCode from bundle code
}
\details{
Bundle objects may store generated NIMBLE code either directly as a
\code{nimbleCode} object or inside a small wrapper list used for package storage.
This helper normalizes those storage conventions so downstream code can work
with the underlying code object without repeatedly checking both cases.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_monitors_from_spec}
\alias{build_monitors_from_spec}
\title{Build default monitors from a compiled model spec}
\usage{
build_monitors_from_spec(spec, monitor_v = FALSE, monitor_latent = FALSE)
}
\arguments{
\item{spec}{A compiled model specification produced by \code{compile_model_spec()}.}

\item{monitor_v}{Logical; for SB, whether to also monitor stick breaks \code{v}.}

\item{monitor_latent}{Logical; whether to monitor latent cluster labels \code{z}.}
}
\value{
Character vector of node names to monitor.
}
\description{
Returns the character vector of node names to monitor in MCMC.
This is a pre-run builder used by \code{build_nimble_bundle()}.
}
\details{
Monitoring follows these rules:
\itemize{
\item Always monitor concentration \code{kappa} (whether fixed or stochastic).
\item SB: monitor \code{w[1:components]} and optionally \code{v[1:(components-1)]}.
\item CRP: monitor \code{z[1:N]}.
\item Bulk parameters:
\itemize{
\item dist/fixed: monitor \code{<param>[1:components]}
\item link: monitor \code{beta_<param>[1:components, 1:P]}
}
\item GPD (if enabled):
\itemize{
\item threshold: monitor scalar \code{threshold} when not link-mode; \code{threshold[1:N]} for link-mode
\item if threshold is link-mode: monitor \code{beta_threshold[1:P]}
\item if threshold uses LN link-dist default: monitor \code{sdlog_u}
\item tail_scale: if link-mode, monitor \code{beta_tail_scale[1:P]}
\item tail_shape: monitor scalar \code{tail_shape} (fixed or dist)
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_code_from_spec}
\alias{build_code_from_spec}
\title{Build NIMBLE model code from a compiled model spec}
\usage{
build_code_from_spec(spec)
}
\arguments{
\item{spec}{A compiled model specification produced by \code{compile_model_spec()}.}
}
\value{
A \code{nimbleCode} object.
}
\description{
Dispatches to the backend-specific code generators:
\itemize{
\item \code{build_code_sb_from_spec()} for stick-breaking (\code{"sb"})
\item \code{build_code_crp_from_spec()} for CRP and spliced (\code{"crp"}, \code{"spliced"})
}
}
\details{
The model size is controlled by \code{spec$meta$components} only.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_nimble_bundle}
\alias{build_nimble_bundle}
\title{Build the explicit one-arm NIMBLE bundle}
\usage{
build_nimble_bundle(
  y,
  X = NULL,
  ps = NULL,
  backend = c("sb", "crp", "spliced"),
  kernel,
  GPD = FALSE,
  components = 10L,
  param_specs = NULL,
  mcmc = list(niter = 2000, nburnin = 500, thin = 1, nchains = 1, seed = 1),
  epsilon = 0.025,
  alpha_random = TRUE,
  monitor = c("core", "full"),
  monitor_latent = FALSE,
  monitor_v = FALSE
)
}
\arguments{
\item{y}{Numeric outcome vector.}

\item{X}{Optional design matrix/data.frame (N x p) for conditional variants.}

\item{ps}{Optional numeric vector (length N) of propensity scores. When provided,
augments the design matrix for PS-adjusted outcome modeling.}

\item{backend}{Character; the Dirichlet process representation:
\itemize{
\item \code{"sb"}: stick-breaking truncation
\item \code{"crp"}: Chinese Restaurant Process
}}

\item{kernel}{Character kernel name (must exist in \code{get_kernel_registry()}).}

\item{GPD}{Logical; whether a GPD tail is requested.}

\item{components}{Integer >= 2. Single user-facing truncation parameter:
\itemize{
\item SB: number of mixture components used in stick-breaking truncation
\item CRP: maximum number of clusters represented in the finite NIMBLE model
}}

\item{param_specs}{Optional list with entries \code{bulk} and \code{tail} to override defaults.}

\item{mcmc}{Named list of MCMC settings (niter, nburnin, thin, nchains, seed). Stored in bundle.}

\item{epsilon}{Numeric in [0,1). For downstream summaries/plots/prediction we keep the
smaller k defined by either (i) cumulative mass >= 1 - epsilon or (ii) per-component
weights >= epsilon, then renormalize.}

\item{alpha_random}{Logical; whether the DP concentration parameter \eqn{\kappa} is stochastic.}

\item{monitor}{Character monitor profile:
\itemize{
\item \code{"core"} (default): monitors only the essential model parameters
\item \code{"full"}: monitors all model nodes
}}

\item{monitor_latent}{Logical; if TRUE, include latent cluster labels (\code{z}) in monitors.}

\item{monitor_v}{Logical; if TRUE and backend is SB, include stick breaks (\code{v}) in monitors.}
}
\value{
A named list of class \code{"causalmixgpd_bundle"}. Its primary
components are \code{spec}, \code{code}, \code{constants},
\code{dimensions}, \code{data}, \code{inits}, \code{monitors}, and stored
\code{mcmc} settings.
}
\description{
\code{build_nimble_bundle()} is the detailed constructor behind
\code{\link{bundle}} for one-arm models. It compiles the modeling plan into a
self-contained object holding code-generation inputs, initialization rules,
monitor policy, and stored MCMC defaults.
}
\details{
The returned bundle encodes a finite approximation to a Dirichlet process
mixture using either a stick-breaking (\code{"sb"}) or Chinese restaurant
process / spliced (\code{"crp"} / \code{"spliced"}) representation.

For the bulk-only model, the target likelihood is the DPM predictive law
\deqn{f(y \mid x) = \sum_{k=1}^{K} w_k(x) f_k(y \mid x, \theta_k).}
When \code{GPD = TRUE}, the bundle augments the bulk model with a threshold
\eqn{u(x)} and generalized Pareto tail above that threshold, producing the
spliced predictive distribution described in the manuscript vignette.

This function intentionally stops before model compilation and sampling.
Use \code{\link{run_mcmc_bundle_manual}} or \code{\link{mcmc}} to execute the
stored model definition.

The object contains:
\itemize{
\item compiled model \code{spec}
\item \code{nimbleCode} model code
\item \code{constants}, \code{data}, explicit \code{dimensions}
\item initialization function \code{inits} (stored as a function)
\item monitor specification
\item MCMC settings list (stored but not used for code generation)
}
}
\examples{
\donttest{
y <- abs(rnorm(25)) + 0.1
bundle <- build_nimble_bundle(
  y = y,
  backend = "sb",
  kernel = "normal",
  GPD = FALSE,
  components = 3,
  mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
)
bundle
}
}
\seealso{
\code{\link{bundle}}, \code{\link{run_mcmc_bundle_manual}},
\code{\link{predict.mixgpd_fit}}, \code{\link{kernel_support_table}},
\code{\link{get_kernel_registry}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.wrap_scalar_first_arg}
\alias{.wrap_scalar_first_arg}
\title{Wrap scalar first-argument functions to handle vector inputs.}
\usage{
.wrap_scalar_first_arg(fun, first_arg_name)
}
\description{
Wrap scalar first-argument functions to handle vector inputs.
}
\details{
Many low-level distribution helpers are scalar in their first argument. This
wrapper lifts such functions to vector inputs by evaluating the scalar
function repeatedly and combining the results into either a numeric vector or
a matrix, depending on the length of the original return value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.detect_first_present}
\alias{.detect_first_present}
\title{Detect the first present argument name in dots.}
\usage{
.detect_first_present(dots, candidates = c("q", "x"))
}
\description{
Detect the first present argument name in dots.
}
\details{
Some wrapped scalar functions accept either \code{q} or \code{x} as their first numeric
argument depending on the original API. This helper inspects \code{...} and returns
the first candidate name that is actually present so wrapper code can map user
input onto the target function signature.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lognormal.R
\name{lognormal_mixgpd}
\alias{lognormal_mixgpd}
\alias{dLognormalMixGpd}
\alias{pLognormalMixGpd}
\alias{rLognormalMixGpd}
\alias{qLognormalMixGpd}
\title{Lognormal mixture with a GPD tail}
\usage{
dLognormalMixGpd(
  x,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  log = 0
)

pLognormalMixGpd(
  q,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rLognormalMixGpd(n, w, meanlog, sdlog, threshold, tail_scale, tail_shape)

qLognormalMixGpd(
  p,
  w,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}.}

\item{meanlog, sdlog}{Numeric vectors of length \eqn{K} giving component log-means and log-standard deviations.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars.
\code{qLognormalMixGpd()} returns a numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to a lognormal mixture
bulk.
Let \eqn{F_{mix}} be the Lognormal mixture CDF. The spliced CDF is
\eqn{F(x)=F_{mix}(x)} for \eqn{x<threshold} and
\eqn{F(x)=F_{mix}(threshold) + \{1-F_{mix}(threshold)\}G(x)} for \eqn{x\ge threshold}, where \eqn{G}
is the GPD CDF for exceedances above \code{threshold}.
}
\details{
The density, CDF, and RNG are implemented as \code{nimbleFunction}s. The quantile is an R function:
it uses numerical inversion in the bulk region and the closed-form GPD quantile in the tail.

Let \eqn{F_{mix}} be the lognormal-mixture CDF and let \eqn{u} denote the threshold. The splice
uses the bulk law below \eqn{u} and attaches a GPD to the residual survival mass above
\eqn{u}. The density therefore becomes
\deqn{
f(x) =
\left\{
\begin{array}{ll}
f_{mix}(x), & x < u, \\
\{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
The quantile is computed piecewise: bulk quantiles are obtained numerically from the mixture CDF,
whereas tail quantiles use the closed-form GPD inverse after rescaling the upper-tail
probability.
}
\section{Functions}{
\itemize{
\item \code{dLognormalMixGpd()}: Lognormal mixture + GPD tail density

\item \code{pLognormalMixGpd()}: Lognormal mixture + GPD tail distribution function

\item \code{rLognormalMixGpd()}: Lognormal mixture + GPD tail random generation

\item \code{qLognormalMixGpd()}: Lognormal mixture + GPD tail quantile function

}}
\examples{
w <- c(0.60, 0.25, 0.15)
meanlog <- c(-0.2, 0.6, 1.2)
sdlog <- c(0.4, 0.3, 0.5)
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dLognormalMixGpd(4.0, w = w, meanlog = meanlog, sdlog = sdlog,
                threshold = threshold, tail_scale = tail_scale,
                tail_shape = tail_shape, log = FALSE)
pLognormalMixGpd(4.0, w = w, meanlog = meanlog, sdlog = sdlog,
                threshold = threshold, tail_scale = tail_scale,
                tail_shape = tail_shape, lower.tail = TRUE, log.p = FALSE)
qLognormalMixGpd(0.50, w = w, meanlog = meanlog, sdlog = sdlog,
                threshold = threshold, tail_scale = tail_scale,
                tail_shape = tail_shape)
qLognormalMixGpd(0.95, w = w, meanlog = meanlog, sdlog = sdlog,
                threshold = threshold, tail_scale = tail_scale,
                tail_shape = tail_shape)
replicate(10, rLognormalMixGpd(1, w = w, meanlog = meanlog, sdlog = sdlog,
                              threshold = threshold,
                              tail_scale = tail_scale,
                              tail_shape = tail_shape))
}
\seealso{
\code{\link[=lognormal_mix]{lognormal_mix()}}, \code{\link[=lognormal_gpd]{lognormal_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=lognormal_lowercase]{lognormal_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other lognormal kernel families: 
\code{\link{lognormal_gpd}},
\code{\link{lognormal_mix}}
}
\concept{lognormal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{globals}
\alias{globals}
\title{Internal global declarations}
\description{
Declares global variables to satisfy R CMD check when using non-standard
evaluation or generated column names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.wrap_scalar_r}
\alias{.wrap_scalar_r}
\title{Wrap scalar RNG to handle n > 1.}
\usage{
.wrap_scalar_r(fun)
}
\description{
Wrap scalar RNG to handle n > 1.
}
\details{
Random-generation helpers in the package are scalar-at-a-time. This wrapper
promotes them to the standard \code{n} interface by repeating the scalar generator
and returning either a numeric vector or a matrix of generated values,
depending on the length of one draw.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{print.dpmixgpd_cluster_psm}
\alias{print.dpmixgpd_cluster_psm}
\title{Print a cluster posterior similarity matrix}
\usage{
\method{print}{dpmixgpd_cluster_psm}(x, ...)
}
\arguments{
\item{x}{Cluster PSM object.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly.
}
\description{
Compact display for a posterior similarity matrix.
}
\details{
A posterior similarity matrix records pairwise co-clustering probabilities on
the training sample. Its \eqn{(i,j)} entry is the posterior probability that
observations \eqn{i} and \eqn{j} share the same latent cluster across the
retained MCMC draws.

The printed header reports only matrix size and bookkeeping information. Use
\code{\link[=summary.dpmixgpd_cluster_psm]{summary.dpmixgpd_cluster_psm()}} for numerical summaries and
\code{\link[=plot.dpmixgpd_cluster_psm]{plot.dpmixgpd_cluster_psm()}} for a visual heatmap.
}
\seealso{
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=summary.dpmixgpd_cluster_psm]{summary.dpmixgpd_cluster_psm()}},
\code{\link[=plot.dpmixgpd_cluster_psm]{plot.dpmixgpd_cluster_psm()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{dpmgpd.cluster}
\alias{dpmgpd.cluster}
\title{Fit a clustering-only bulk-tail model}
\usage{
dpmgpd.cluster(
  formula,
  data,
  type = c("weights", "param", "both"),
  default = "weights",
  mcmc = list(),
  ...
)
}
\arguments{
\item{formula}{Model formula. The response must be present in \code{data}.}

\item{data}{Data frame containing the response and optional predictors.}

\item{type}{Clustering mode:
\itemize{
\item \code{"weights"}: links mixture weights to predictors
\item \code{"param"}: links kernel parameters to predictors
\item \code{"both"}: links both weights and kernel parameters to predictors
}}

\item{default}{Default mode used when \code{type} is omitted.}

\item{mcmc}{MCMC control list passed into the cluster bundle.}

\item{...}{Additional arguments passed to \code{build_cluster_bundle()}, including kernel settings,
prior overrides, component counts, and monitoring controls.}
}
\value{
Object of class \code{dpmixgpd_cluster_fit}.
}
\description{
Variant of \code{\link[=dpmix.cluster]{dpmix.cluster()}} that augments the cluster kernel with a generalized Pareto tail.
This is the clustering analogue of the spliced bulk-tail workflow used by \code{\link[=dpmgpd]{dpmgpd()}}.
}
\details{
For observations above a component-specific threshold, the component density is spliced as
\deqn{
f(y) = (1 - F_{bulk}(u)) g_{GPD}(y \mid u, \sigma_u, \xi_u), \qquad y \ge u,
}
so cluster assignment can be informed by both central behavior and tail behavior.

This interface is preferable when cluster separation is driven by upper-tail differences rather
than bulk-only shape or location differences.
}
\examples{
\donttest{
data("nc_posX100_p3_k2", package = "CausalMixGPD")
dat <- data.frame(y = nc_posX100_p3_k2$y[1:20],
                  nc_posX100_p3_k2$X[1:20, , drop = FALSE])
fit <- dpmgpd.cluster(
  y ~ x1 + x2 + x3,
  data = dat,
  kernel = "gamma",
  type = "param",
  components = 3,
  mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1)
)
cluster_profiles(fit)
}
}
\seealso{
\code{\link[=dpmix.cluster]{dpmix.cluster()}}, \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}},
\code{\link[=dpmgpd]{dpmgpd()}}, \code{\link[=sim_bulk_tail]{sim_bulk_tail()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma.R
\name{gamma_mixgpd}
\alias{gamma_mixgpd}
\alias{dGammaMixGpd}
\alias{pGammaMixGpd}
\alias{rGammaMixGpd}
\alias{qGammaMixGpd}
\title{Gamma mixture with a GPD tail}
\usage{
dGammaMixGpd(x, w, shape, scale, threshold, tail_scale, tail_shape, log = 0)

pGammaMixGpd(
  q,
  w,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rGammaMixGpd(n, w, shape, scale, threshold, tail_scale, tail_shape)

qGammaMixGpd(
  p,
  w,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}.}

\item{shape, scale}{Numeric vectors of length \eqn{K} giving Gamma shape and scale parameters.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qGammaMixGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to a gamma mixture bulk.
}
\details{
The gamma mixture governs the body of the distribution up to the threshold \eqn{u}. Beyond
\eqn{u}, only the remaining survival mass is modeled by the GPD, giving
\deqn{
f(x) =
\left\{
\begin{array}{ll}
f_{mix}(x), & x < u, \\
\{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
This is the positive-support analogue of the normal and lognormal splice families. Bulk quantiles
are still found by numerical inversion, while tail quantiles use the explicit GPD inverse.
}
\section{Functions}{
\itemize{
\item \code{dGammaMixGpd()}: Gamma mixture + GPD tail density

\item \code{pGammaMixGpd()}: Gamma mixture + GPD tail distribution function

\item \code{rGammaMixGpd()}: Gamma mixture + GPD tail random generation

\item \code{qGammaMixGpd()}: Gamma mixture + GPD tail quantile function

}}
\examples{
w <- c(0.55, 0.30, 0.15)
scale <- c(1.0, 2.5, 5.0)
shape <- c(2, 4, 6)
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dGammaMixGpd(4.0, w = w, scale = scale, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape, log = 0)
pGammaMixGpd(4.0, w = w, scale = scale, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape, lower.tail = 1, log.p = 0)
qGammaMixGpd(0.50, w = w, scale = scale, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape)
qGammaMixGpd(0.95, w = w, scale = scale, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape)
replicate(10, rGammaMixGpd(1, w = w, scale = scale, shape = shape,
                          threshold = threshold,
                          tail_scale = tail_scale,
                          tail_shape = tail_shape))
}
\seealso{
\code{\link[=gamma_mix]{gamma_mix()}}, \code{\link[=gamma_gpd]{gamma_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=gamma_lowercase]{gamma_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other gamma kernel families: 
\code{\link{gamma_gpd}},
\code{\link{gamma_mix}}
}
\concept{gamma kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{summary.dpmixgpd_cluster_psm}
\alias{summary.dpmixgpd_cluster_psm}
\title{Summarize a cluster posterior similarity matrix}
\usage{
\method{summary}{dpmixgpd_cluster_psm}(object, ...)
}
\arguments{
\item{object}{Cluster PSM object.}

\item{...}{Unused.}
}
\value{
Summary list with matrix size and basic summaries of the similarity entries.
}
\description{
Summarize pairwise posterior co-clustering probabilities.
}
\details{
The diagonal of a posterior similarity matrix is always close to one, while off-diagonal values
near one indicate highly stable co-clustering across retained posterior draws.
}
\seealso{
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=plot.dpmixgpd_cluster_psm]{plot.dpmixgpd_cluster_psm()}},
\code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.get_nobs}
\alias{.get_nobs}
\title{Get number of observations used in fitting}
\usage{
.get_nobs(object)
}
\arguments{
\item{object}{A mixgpd_fit.}
}
\value{
Integer n.
}
\description{
Get number of observations used in fitting
}
\details{
The fitted object may carry the response in slightly different storage slots
depending on how it was built. This helper centralizes the lookup and returns
the effective training sample size used by summaries and print methods.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_ate}
\alias{print.causalmixgpd_ate}
\title{Print an ATE-style effect object}
\usage{
\method{print}{causalmixgpd_ate}(x, digits = 3, max_rows = 6, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_ate"} object from \code{ate()}.}

\item{digits}{Number of digits to display.}

\item{max_rows}{Maximum number of estimate rows to display.}

\item{...}{Unused.}
}
\value{
The object \code{x}, invisibly.
}
\description{
\code{print.causalmixgpd_ate()} prints a compact summary for objects
produced by \code{\link{ate}}, \code{\link{att}}, \code{\link{cate}}, or
\code{\link{ate_rmean}}.
}
\details{
These objects summarize posterior treatment contrasts on the mean scale. For
the marginal average treatment effect,
\deqn{\Delta = E(Y^1) - E(Y^0).}
\code{att()} changes the standardization target to the treated population,
\code{cate()} conditions on supplied covariate profiles, and \code{ate_rmean()} replaces
the ordinary mean by a restricted mean
\eqn{\int_0^c S_a(t)\,dt} up to the chosen truncation point.

The print method shows the main effect table and setup metadata, but it is not
a full diagnostic report. Use \code{summary()} for tabular summaries and \code{plot()}
for graphical inspection of the same treatment-effect object.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
a <- ate(fit, interval = "credible")
print(a)
}
}
\seealso{
\code{\link{summary.causalmixgpd_ate}},
\code{\link{plot.causalmixgpd_ate}}, \code{\link{ate}},
\code{\link{cate}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_fit_plots}
\alias{print.mixgpd_fit_plots}
\title{Print method for mixgpd_fit diagnostic plots}
\usage{
\method{print}{mixgpd_fit_plots}(x, ...)
}
\arguments{
\item{x}{Object of class \code{mixgpd_fit_plots}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns the input object.
}
\description{
Print method for mixgpd_fit diagnostic plots
}
\details{
Diagnostic plotting for \code{mixgpd_fit} can return a named collection of ggmcmc
graphics. This print method iterates through that collection and prints each
stored diagnostic plot with a section label so trace, density, and related
views can be read in order.

The method performs no additional posterior computation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_causal_fit_plots}
\alias{print.causalmixgpd_causal_fit_plots}
\title{Print method for paired causal-fit diagnostic plots}
\usage{
\method{print}{causalmixgpd_causal_fit_plots}(x, ...)
}
\arguments{
\item{x}{Object of class \code{causalmixgpd_causal_fit_plots}.}

\item{...}{Additional arguments passed to the stored plot-print methods.}
}
\value{
Invisibly returns the input object.
}
\description{
Print method for paired causal-fit diagnostic plots
}
\details{
When \code{plot.causalmixgpd_causal_fit()} is called with \code{arm = "both"}, the
result is a named pair of treated and control diagnostic-plot objects. This
print method renders those two stored plot collections one after the other so
arm-specific diagnostics remain clearly separated.

It is a formatting helper and does not recompute any diagnostics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.wrap_nimble_code}
\alias{.wrap_nimble_code}
\title{Wrap nimbleCode for bundle storage}
\usage{
.wrap_nimble_code(code)
}
\description{
Wrap nimbleCode for bundle storage
}
\details{
This helper is the inverse of \code{.extract_nimble_code()}. It stores a raw
\code{nimbleCode} object inside a lightweight list so bundle objects can carry code
alongside other metadata without ambiguity about the field layout.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\name{sim_bulk_tail}
\alias{sim_bulk_tail}
\title{Simulate positive bulk-tail data}
\usage{
sim_bulk_tail(n = 200, tail_prob = 0.12, seed = NULL)
}
\arguments{
\item{n}{Integer sample size.}

\item{tail_prob}{Approximate tail probability \eqn{\Pr(X > u)} used to split the sample into
bulk and tail draws.}

\item{seed}{Optional random seed for reproducibility.}
}
\value{
Numeric vector of length \code{n} containing positive outcomes sorted in ascending order.
}
\description{
Generate synthetic outcomes with a light-to-moderate bulk and a heavier upper tail.
The sample is assembled from a lognormal-gamma bulk and a shifted tail sample, then sorted.
This generator is intended for examples, help pages, and workflow checks
rather than as a formal generative model matching the full package
hierarchy exactly.
}
\details{
The generator approximates a spliced sample
\deqn{
X \sim (1 - \pi_u) F_{bulk} + \pi_u F_{tail},
}
where \eqn{\pi_u =} \code{tail_prob}. The bulk component is itself a simple two-component mixture,
while the tail component is a shifted positive distribution that produces larger values.

Use this helper when you need a fast toy sample for \code{\link[=bundle]{bundle()}}, \code{\link[=dpmix]{dpmix()}}, or \code{\link[=dpmgpd]{dpmgpd()}}.
It should not be interpreted as posterior predictive simulation from a fitted object.
}
\seealso{
\code{\link[=sim_causal_qte]{sim_causal_qte()}}, \code{\link[=sim_survival_tail]{sim_survival_tail()}}, \code{\link[=bundle]{bundle()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other simulation helpers: 
\code{\link{sim_causal_qte}()},
\code{\link{sim_survival_tail}()}
}
\concept{simulation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.causalmixgpd_ate}
\alias{print.summary.causalmixgpd_ate}
\title{Print an ATE summary}
\usage{
\method{print}{summary.causalmixgpd_ate}(x, digits = 3, ...)
}
\arguments{
\item{x}{A \code{"summary.causalmixgpd_ate"} object.}

\item{digits}{Number of digits to display.}

\item{...}{Unused.}
}
\value{
The object \code{x}, invisibly.
}
\description{
Print an ATE summary
}
\details{
This method formats the object returned by \code{summary.causalmixgpd_ate()}. It
prints the prediction design, interval settings, optional model metadata, and
the resulting treatment-effect table on the mean or restricted-mean scale.

The method is purely a reporting layer. All posterior aggregation has already
been completed by the corresponding summary constructor.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoroso.R
\name{amoroso_lowercase}
\alias{amoroso_lowercase}
\alias{damorosomix}
\alias{pamorosomix}
\alias{qamorosomix}
\alias{ramorosomix}
\alias{damorosomixgpd}
\alias{pamorosomixgpd}
\alias{qamorosomixgpd}
\alias{ramorosomixgpd}
\alias{damorosogpd}
\alias{pamorosogpd}
\alias{qamorosogpd}
\alias{ramorosogpd}
\title{Lowercase vectorized Amoroso distribution functions}
\usage{
damorosomix(x, w, loc, scale, shape1, shape2, log = FALSE)

pamorosomix(q, w, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE)

qamorosomix(
  p,
  w,
  loc,
  scale,
  shape1,
  shape2,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

ramorosomix(n, w, loc, scale, shape1, shape2)

damorosomixgpd(
  x,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

pamorosomixgpd(
  q,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qamorosomixgpd(
  p,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

ramorosomixgpd(
  n,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape
)

damorosogpd(
  x,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

pamorosogpd(
  q,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qamorosogpd(
  p,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

ramorosogpd(n, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{loc, scale, shape1, shape2}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar Amoroso-kernel topics in this file.
}
\details{
These are vectorized wrappers around the scalar Amoroso routines used internally by the package.
They preserve the same location-scale-shape parameterization and the same piecewise splice logic
for GPD tails. Quantile wrappers therefore continue to rely on the scalar numerical inversion or
scalar GPD inverse exactly as documented for the uppercase functions.
}
\section{Functions}{
\itemize{
\item \code{damorosomix()}: Amoroso mixture density (vectorized)

\item \code{pamorosomix()}: Amoroso mixture distribution function (vectorized)

\item \code{qamorosomix()}: Amoroso mixture quantile function (vectorized)

\item \code{ramorosomix()}: Amoroso mixture random generation (vectorized)

\item \code{damorosomixgpd()}: Amoroso mixture + GPD density (vectorized)

\item \code{pamorosomixgpd()}: Amoroso mixture + GPD distribution function (vectorized)

\item \code{qamorosomixgpd()}: Amoroso mixture + GPD quantile function (vectorized)

\item \code{ramorosomixgpd()}: Amoroso mixture + GPD random generation (vectorized)

\item \code{damorosogpd()}: Amoroso + GPD density (vectorized)

\item \code{pamorosogpd()}: Amoroso + GPD distribution function (vectorized)

\item \code{qamorosogpd()}: Amoroso + GPD quantile function (vectorized)

\item \code{ramorosogpd()}: Amoroso + GPD random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.3, 0.1)
locs <- c(0.5, 0.5, 0.5)
scls <- c(1, 1.3, 1.6)
s1 <- c(2.5, 3, 4)
s2 <- c(1.2, 1.2, 1.2)

# Amoroso mixture
damorosomix(c(1, 2, 3), w = w, loc = locs, scale = scls, shape1 = s1, shape2 = s2)
ramorosomix(5, w = w, loc = locs, scale = scls, shape1 = s1, shape2 = s2)

}
\seealso{
\code{\link[=amoroso_mix]{amoroso_mix()}}, \code{\link[=amoroso_mixgpd]{amoroso_mixgpd()}}, \code{\link[=amoroso_gpd]{amoroso_gpd()}}, \code{\link[=bundle]{bundle()}},
\code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma.R
\name{gamma_gpd}
\alias{gamma_gpd}
\alias{dGammaGpd}
\alias{pGammaGpd}
\alias{rGammaGpd}
\alias{qGammaGpd}
\title{Gamma with a GPD tail}
\usage{
dGammaGpd(x, shape, scale, threshold, tail_scale, tail_shape, log = 0)

pGammaGpd(
  q,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rGammaGpd(n, shape, scale, threshold, tail_scale, tail_shape)

qGammaGpd(
  p,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{shape}{Numeric scalar Gamma shape parameter.}

\item{scale}{Numeric scalar scale parameter for the Gamma bulk.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are
\eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric tolerance for numerical inversion in \code{qGammaGpd}.}

\item{maxiter}{Maximum iterations for numerical inversion in \code{qGammaGpd}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars.
\code{qGammaGpd()} returns a numeric vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
gamma bulk distribution.
}
\details{
This topic combines a single gamma bulk with a generalized Pareto exceedance model. If
\eqn{F_\Gamma(u)} is the bulk probability below the threshold, then the splice replaces the
upper tail by \eqn{\{1-F_\Gamma(u)\}g_{GPD}(x)} while leaving the lower region unchanged. The
resulting distribution is continuous at the threshold and preserves the gamma body exactly below
\eqn{u}.

The ordinary mean is finite only when the GPD shape satisfies \eqn{\xi < 1}. For heavier tails,
predictive mean summaries should be replaced by restricted means or quantile summaries.
}
\section{Functions}{
\itemize{
\item \code{dGammaGpd()}: Gamma + GPD tail density

\item \code{pGammaGpd()}: Gamma + GPD tail distribution function

\item \code{rGammaGpd()}: Gamma + GPD tail random generation

\item \code{qGammaGpd()}: Gamma + GPD tail quantile function

}}
\examples{
scale <- 2.5
shape <- 4
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dGammaGpd(4.0, scale = scale, shape = shape,
         threshold = threshold, tail_scale = tail_scale,
         tail_shape = tail_shape, log = 0)
pGammaGpd(4.0, scale = scale, shape = shape,
         threshold = threshold, tail_scale = tail_scale,
         tail_shape = tail_shape, lower.tail = 1, log.p = 0)
qGammaGpd(0.50, scale = scale, shape = shape,
         threshold = threshold, tail_scale = tail_scale,
         tail_shape = tail_shape)
qGammaGpd(0.95, scale = scale, shape = shape,
         threshold = threshold, tail_scale = tail_scale,
         tail_shape = tail_shape)
replicate(10, rGammaGpd(1, scale = scale, shape = shape,
                       threshold = threshold,
                       tail_scale = tail_scale,
                       tail_shape = tail_shape))

}
\seealso{
\code{\link[=gamma_mix]{gamma_mix()}}, \code{\link[=gamma_mixgpd]{gamma_mixgpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=gamma_lowercase]{gamma_lowercase()}}.

Other gamma kernel families: 
\code{\link{gamma_mix}},
\code{\link{gamma_mixgpd}}
}
\concept{gamma kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{cluster_profiles}
\alias{cluster_profiles}
\title{Extract Cluster Profiles}
\usage{
cluster_profiles(object, ...)
}
\arguments{
\item{object}{A cluster fit, cluster-label object, or corresponding summary object.}

\item{...}{Additional arguments passed to summary methods for fitted or label objects.}
}
\value{
A data frame of cluster-level descriptive summaries, or \code{NULL}
when profiles are unavailable.
}
\description{
Access descriptive cluster profiles without reaching into summary internals.
}
\seealso{
Other cluster workflow: 
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0-base-distributions.R
\name{base_lowercase}
\alias{base_lowercase}
\alias{dgpd}
\alias{pgpd}
\alias{qgpd}
\alias{rgpd}
\alias{dinvgauss}
\alias{pinvgauss}
\alias{qinvgauss}
\alias{rinvgauss}
\alias{damoroso}
\alias{pamoroso}
\alias{qamoroso}
\alias{ramoroso}
\alias{dcauchy_vec}
\alias{pcauchy_vec}
\alias{qcauchy_vec}
\alias{rcauchy_vec}
\title{Lowercase vectorized distribution functions (base kernels)}
\usage{
dgpd(x, threshold, scale, shape, log = FALSE)

pgpd(q, threshold, scale, shape, lower.tail = TRUE, log.p = FALSE)

qgpd(p, threshold, scale, shape, lower.tail = TRUE, log.p = FALSE)

rgpd(n, threshold, scale, shape)

dinvgauss(x, mean, shape, log = FALSE)

pinvgauss(q, mean, shape, lower.tail = TRUE, log.p = FALSE)

qinvgauss(
  p,
  mean,
  shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rinvgauss(n, mean, shape)

damoroso(x, loc, scale, shape1, shape2, log = FALSE)

pamoroso(q, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE)

qamoroso(p, loc, scale, shape1, shape2, lower.tail = TRUE, log.p = FALSE)

ramoroso(n, loc, scale, shape1, shape2)

dcauchy_vec(x, location, scale, log = FALSE)

pcauchy_vec(q, location, scale, lower.tail = TRUE, log.p = FALSE)

qcauchy_vec(p, location, scale, lower.tail = TRUE, log.p = FALSE)

rcauchy_vec(n, location, scale)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{threshold, scale, shape, mean, loc, shape1, shape2, location}{Distribution parameters (scalars).}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are
\eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{n}{Integer number of observations to generate.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers around the scalar base-kernel functions defined in this file. These
helpers are intended for interactive R use, examples, testing, and checking numerical behavior
outside compiled NIMBLE code.
}
\details{
The wrappers preserve the same parameterizations as the uppercase scalar functions, but accept
vector inputs for \code{x}, \code{q}, or \code{p} and allow \code{n > 1} for random generation.

Each lowercase helper is a vectorized R wrapper around the corresponding
uppercase scalar routine documented in this file. The wrapper keeps the same
parameterization and applies the scalar kernel repeatedly over the supplied
evaluation points or simulation index. These helpers are therefore
appropriate for interactive analysis, testing, and examples, whereas the
uppercase functions are the building blocks used inside NIMBLE model code.

The wrappers do not change the underlying theory. For example, \code{qgpd()} still uses the
closed-form GPD inverse, \code{qinvgauss()} still performs numerical inversion of the inverse
Gaussian CDF, and \code{qamoroso()} still maps a gamma quantile through the Amoroso
transformation. Random-generation wrappers call the corresponding scalar RNG repeatedly when
\code{n > 1}.
}
\section{Functions}{
\itemize{
\item \code{dgpd()}: GPD density (vectorized)

\item \code{pgpd()}: GPD distribution function (vectorized)

\item \code{qgpd()}: GPD quantile function (vectorized)

\item \code{rgpd()}: GPD random generation (vectorized)

\item \code{dinvgauss()}: Inverse Gaussian density (vectorized)

\item \code{pinvgauss()}: Inverse Gaussian distribution function (vectorized)

\item \code{qinvgauss()}: Inverse Gaussian quantile function (vectorized)

\item \code{rinvgauss()}: Inverse Gaussian random generation (vectorized)

\item \code{damoroso()}: Amoroso density (vectorized)

\item \code{pamoroso()}: Amoroso distribution function (vectorized)

\item \code{qamoroso()}: Amoroso quantile function (vectorized)

\item \code{ramoroso()}: Amoroso random generation (vectorized)

\item \code{dcauchy_vec()}: Cauchy density (vectorized)

\item \code{pcauchy_vec()}: Cauchy distribution function (vectorized)

\item \code{qcauchy_vec()}: Cauchy quantile function (vectorized)

\item \code{rcauchy_vec()}: Cauchy random generation (vectorized)

}}
\examples{
# GPD
dgpd(c(1.5, 2.0, 2.5), threshold = 1, scale = 0.8, shape = 0.2)
pgpd(c(1.5, 2.0), threshold = 1, scale = 0.8, shape = 0.2)
qgpd(c(0.5, 0.9), threshold = 1, scale = 0.8, shape = 0.2)
rgpd(5, threshold = 1, scale = 0.8, shape = 0.2)

# Inverse Gaussian
dinvgauss(c(1, 2, 3), mean = 2, shape = 5)
rinvgauss(5, mean = 2, shape = 5)

# Amoroso
damoroso(c(1, 2), loc = 0, scale = 1.5, shape1 = 2, shape2 = 1.2)
ramoroso(5, loc = 0, scale = 1.5, shape1 = 2, shape2 = 1.2)

# Cauchy
dcauchy_vec(c(-1, 0, 1), location = 0, scale = 1)
rcauchy_vec(5, location = 0, scale = 1)

}
\seealso{
\code{\link[=gpd]{gpd()}}, \code{\link[=InvGauss]{InvGauss()}}, \code{\link[=amoroso]{amoroso()}}, \code{\link[=cauchy]{cauchy()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}},
\code{\link[=kernel_support_table]{kernel_support_table()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/1-registry.R
\name{kernel_support_table}
\alias{kernel_support_table}
\title{Kernel support matrix}
\usage{
kernel_support_table(round = TRUE)
}
\arguments{
\item{round}{Logical; \code{TRUE} to replace logical values with symbols.}
}
\value{
data.frame with columns \code{kernel}, \code{gpd}, \code{covariates}, \code{sb}, \code{crp}.
}
\description{
Returns a data frame summarizing each kernel's supported features.
}
\details{
The returned table is a compact view of the registry contracts. Each row
corresponds to one kernel, while the logical columns indicate whether that
kernel can be paired with a GPD tail, whether covariate-linked parameter
models are defined, and whether the stick-breaking (\code{sb}) and
Chinese-restaurant (\code{crp}) backends are implemented.

This helper is intended for inspection and reporting rather than model
fitting. It is a quick way to verify that a requested combination of kernel,
tail, and backend is supported before calling higher-level workflow
constructors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.get_samples_mcmclist}
\alias{.get_samples_mcmclist}
\title{Safely coerce MCMC samples to coda::mcmc.list}
\usage{
.get_samples_mcmclist(object)
}
\arguments{
\item{object}{A mixgpd_fit.}
}
\value{
A coda::mcmc.list object.
}
\description{
Safely coerce MCMC samples to coda::mcmc.list
}
\details{
Downstream summary and plotting code relies on the \code{coda} interface. This
helper validates the fit, locates the stored posterior draws, and converts a
single-chain \code{mcmc} object into an \code{mcmc.list} so later code can treat the
one-chain and multi-chain cases uniformly.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_causal_bundle}
\alias{summary.causalmixgpd_causal_bundle}
\title{Summarize a causal workflow bundle}
\usage{
\method{summary}{causalmixgpd_causal_bundle}(object, code = FALSE, max_code_lines = 200L, ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_causal_bundle"} object.}

\item{code}{Logical; if TRUE, print generated NIMBLE code for each block.}

\item{max_code_lines}{Integer; maximum number of code lines to print when \code{code=TRUE}.}

\item{...}{Unused.}
}
\value{
The input object (invisibly).
}
\description{
\code{summary.causalmixgpd_causal_bundle()} is the bundle-level validation
checkpoint for the causal workflow.
}
\details{
This summary is meant to be read before posterior sampling. It reports the
stored propensity-score specification, the treated and control outcome model
definitions, and the sample split across treatment arms. In other words, it
verifies the model ingredients for the causal decomposition
\eqn{(e(x), f_0(y \mid x), f_1(y \mid x))}.

Since no MCMC has been run yet, the summary contains only structural
information. Posterior treatment-effect summaries become available after
\code{\link{run_mcmc_causal}} through functions such as \code{\link{ate}} and
\code{\link{qte}}.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal")
summary(cb)
}
}
\seealso{
\code{\link{print.causalmixgpd_causal_bundle}},
\code{\link{run_mcmc_causal}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_causal_fit}
\alias{summary.causalmixgpd_causal_fit}
\title{Summarize a fitted causal model}
\usage{
\method{summary}{causalmixgpd_causal_fit}(object, pars = NULL, ps_pars = NULL, probs = c(0.025, 0.5, 0.975), ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_causal_fit"} object.}

\item{pars}{Optional character vector of outcome-model parameters to
summarize in both treatment arms. Passed to \code{\link{summary.mixgpd_fit}}.}

\item{ps_pars}{Optional character vector of PS-model parameters to summarize.
If \code{NULL}, all monitored PS parameters are summarized.}

\item{probs}{Numeric vector of posterior quantiles to report.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.causalmixgpd_causal_fit"} with
elements \code{ps}, \code{outcome}, and \code{probs}.
}
\description{
\code{summary.causalmixgpd_causal_fit()} returns posterior summaries for the
fitted PS block (when present) and both arm-specific outcome models.
}
\details{
This summary stays at the model-parameter level. It aggregates posterior
summaries for the nuisance model \eqn{e(x)} and for the arm-specific outcome
models \eqn{f_0(y \mid x)} and \eqn{f_1(y \mid x)}, but it does not yet
collapse those pieces into treatment-effect functionals.

That separation is intentional. Parameters and treatment effects answer
different questions: \code{summary.causalmixgpd_causal_fit()} summarizes posterior
draws of the fitted model, whereas \code{ate()}, \code{att()}, \code{cate()}, \code{qte()},
\code{qtt()}, and \code{cqte()} transform those draws into causal contrasts.
}
\seealso{
\code{\link{print.causalmixgpd_causal_fit}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\docType{data}
\name{nc_pos200_k3}
\alias{nc_pos200_k3}
\title{Positive-Support Bulk-Only Data Set With Three Components}
\format{
A list with:
\describe{
\item{y}{Numeric outcome vector.}
\item{X}{NULL.}
\item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
\item{truth}{List with kernel, weights, params, threshold, tail_params.}
}
}
\usage{
nc_pos200_k3
}
\description{
Positive-support, bulk-only mixture dataset with K=3 components and no covariates.
Intended for non-causal bulk-only positive-kernel vignettes (gamma/lognormal/invgauss/amoroso).
}
\examples{
head(nc_pos200_k3$y)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{dpmix}
\alias{dpmix}
\title{Fit a one-arm Dirichlet process mixture without a GPD tail}
\usage{
dpmix(
  y = NULL,
  X = NULL,
  treat = NULL,
  data = NULL,
  mcmc = list(),
  formula = NULL,
  ...
)
}
\arguments{
\item{y}{Either a response vector or a bundle object.}

\item{X}{Optional design matrix/data.frame.}

\item{treat}{Optional binary treatment indicator. If supplied, this wrapper
errors; use \code{dpmix.causal()} for causal models.}

\item{data}{Optional data.frame used with \code{formula}.}

\item{mcmc}{Named list of run arguments passed to \code{mcmc()} (including
optional performance controls such as \code{parallel_chains},
\code{workers}, \code{timing}, and \code{z_update_every}).}

\item{formula}{Optional formula.}

\item{...}{Additional build arguments passed to \code{\link{build_nimble_bundle}}.}
}
\value{
A fitted object of class \code{"mixgpd_fit"}.
}
\description{
\code{dpmix()} is the one-step convenience wrapper for the bulk-only model.
It combines \code{\link{bundle}} and \code{\link{mcmc}} for one-arm data.
}
\details{
The fitted model targets the posterior predictive bulk distribution
\deqn{f(y \mid x) = \int f(y \mid x, \theta)\,d\Pi(\theta),}
without the spliced tail augmentation used by \code{\link{dpmgpd}}.

Use this wrapper when the outcome support is adequately modeled by the bulk
kernel alone. If you need threshold exceedance modeling or extreme-quantile
extrapolation, use \code{\link{dpmgpd}} instead.
}
\seealso{
\code{\link{build_nimble_bundle}},
\code{\link{bundle}}, \code{\link{dpmgpd}},
\code{\link{predict.mixgpd_fit}}, \code{\link{summary.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.get_epsilon}
\alias{.get_epsilon}
\title{Get epsilon value from object spec/meta or argument}
\usage{
.get_epsilon(object, epsilon = NULL)
}
\arguments{
\item{object}{A mixgpd_fit object.}

\item{epsilon}{Numeric; if provided, overrides object spec/meta.}
}
\description{
Get epsilon value from object spec/meta or argument
}
\details{
Many downstream summaries truncate mixture components according to the bundle
or fit-level \code{epsilon} setting. This helper centralizes that lookup so an
explicit function argument overrides the stored fit metadata, and the package
fallback is used only when neither source is present.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{ate_rmean}
\alias{ate_rmean}
\title{Restricted-mean ATE helper}
\usage{
ate_rmean(
  fit,
  newdata = NULL,
  cutoff,
  interval = "credible",
  level = 0.95,
  nsim_mean = 200L,
  show_progress = TRUE
)
}
\arguments{
\item{fit}{A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.}

\item{newdata}{Optional data.frame or matrix of covariates for prediction.
If \code{NULL}, uses the training covariates stored in \code{fit}.}

\item{cutoff}{Finite numeric cutoff for the restricted mean.}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{level}{Numeric credible level for intervals (default 0.95 for 95 percent CI).}

\item{nsim_mean}{Number of posterior predictive draws used by simulation-based
mean targets. Ignored for analytical ordinary means.}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}
}
\value{
A \code{"causalmixgpd_ate"} object computed via \code{\link{ate}}
for unconditional fits or \code{\link{cate}} for conditional fits. The
returned object includes a top-level \code{$fit_df} data frame for direct
extraction.
}
\description{
\code{ate_rmean()} is a convenience wrapper for restricted-mean treatment
effects when the ordinary mean is unstable or undefined.
}
\details{
The restricted-mean estimand replaces \eqn{Y(a)} by
\eqn{\min\{Y(a), c\}}, so the contrast remains finite even when the fitted
GPD tail implies \eqn{\xi \ge 1}.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         GPD = TRUE, components = 3,
                         mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb)
ate_rm <- ate_rmean(fit, cutoff = 10, interval = "credible")
}
}
\seealso{
\code{\link{ate}}, \code{\link{cate}}, \code{\link{predict.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0-base-distributions.R
\name{gpd}
\alias{gpd}
\alias{dGpd}
\alias{pGpd}
\alias{rGpd}
\alias{qGpd}
\title{Generalized Pareto distribution}
\usage{
dGpd(x, threshold, scale, shape, log = 0)

pGpd(q, threshold, scale, shape, lower.tail = 1, log.p = 0)

rGpd(n, threshold, scale, shape)

qGpd(p, threshold, scale, shape, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{threshold}{Numeric scalar threshold at which the GPD is attached.}

\item{scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}
}
\value{
\code{dGpd()} returns a numeric scalar density, \code{pGpd()} returns a numeric scalar CDF,
\code{rGpd()} returns one random draw, and \code{qGpd()} returns a numeric quantile.
}
\description{
Scalar generalized Pareto distribution (GPD) utilities for threshold exceedances above
\code{threshold}. These NIMBLE-compatible functions provide the tail component used by the spliced
bulk-tail families elsewhere in the package.
}
\details{
The parameterization is
\deqn{
G(x) = 1 - \left(1 + \xi \frac{x - u}{\sigma_u}\right)^{-1 / \xi}, \qquad x \ge u,
}
where \code{threshold = u}, \code{scale = sigma_u > 0}, and \code{shape = xi}. When \code{shape} approaches zero,
the distribution reduces to the exponential tail limit.

These uppercase NIMBLE-compatible functions are scalar (\code{x}/\code{q} and \code{n = 1}).
For vectorized R usage, use \code{\link[=base_lowercase]{base_lowercase()}}.

The associated density is
\deqn{
g(x) = \frac{1}{\sigma_u}
\left(1 + \xi \frac{x - u}{\sigma_u}\right)^{-1/\xi - 1},
}
on the support where \eqn{1 + \xi (x-u)/\sigma_u > 0}. When \eqn{\xi = 0}, this reduces to the
exponential density with mean excess \eqn{\sigma_u}. The GPD has finite mean only when
\eqn{\xi < 1}, and finite variance only when \eqn{\xi < 1/2}. Those existence conditions matter
for downstream predictive means in the package: spliced models with \eqn{\xi \ge 1} require
restricted means rather than ordinary means.

If a bulk distribution has CDF \eqn{F_{bulk}}, the package's spliced families use the tail construction
\deqn{
F(x) =
\left\{
\begin{array}{ll}
F_{bulk}(x), & x < u, \\
F_{bulk}(u) + \{1 - F_{bulk}(u)\} G(x), & x \ge u.
\end{array}
\right.
}
}
\section{Functions}{
\itemize{
\item \code{dGpd()}: Generalized Pareto density function

\item \code{pGpd()}: Generalized Pareto distribution function

\item \code{rGpd()}: Generalized Pareto random generation

\item \code{qGpd()}: Generalized Pareto quantile function

}}
\examples{
threshold <- 1
tail_scale <- 0.8
tail_shape <- 0.2

dGpd(1.5, threshold, tail_scale, tail_shape, log = 0)
pGpd(1.5, threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
qGpd(0.50, threshold, tail_scale, tail_shape)
qGpd(0.95, threshold, tail_scale, tail_shape)
replicate(10, rGpd(1, threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=base_lowercase]{base_lowercase()}}, \code{\link[=normal_gpd]{normal_gpd()}}, \code{\link[=lognormal_gpd]{lognormal_gpd()}}, \code{\link[=gamma_gpd]{gamma_gpd()}},
\code{\link[=InvGauss_gpd]{InvGauss_gpd()}}, \code{\link[=laplace_gpd]{laplace_gpd()}}, \code{\link[=amoroso_gpd]{amoroso_gpd()}}.
}
\concept{base tail distributions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0-base-distributions.R
\name{InvGauss}
\alias{InvGauss}
\alias{dInvGauss}
\alias{pInvGauss}
\alias{rInvGauss}
\alias{qInvGauss}
\title{Inverse Gaussian (Wald) distribution}
\usage{
dInvGauss(x, mean, shape, log = 0)

pInvGauss(q, mean, shape, lower.tail = 1, log.p = 0)

rInvGauss(n, mean, shape)

qInvGauss(
  p,
  mean,
  shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{mean}{Numeric scalar mean parameter \eqn{\mu>0}.}

\item{shape}{Numeric scalar shape parameter \eqn{\lambda>0}.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar giving the probability for the quantile.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
\code{dInvGauss()} returns a numeric scalar density, \code{pInvGauss()} returns a numeric scalar
CDF, \code{rInvGauss()} returns one random draw, and \code{qInvGauss()} returns a numeric quantile.
}
\description{
Scalar inverse Gaussian utilities under the \eqn{(\mu, \lambda)} parameterization, where
\code{mean = mu > 0} and \code{shape = lambda > 0}. These functions are used directly and as building
blocks for inverse-Gaussian mixtures and spliced inverse-Gaussian-plus-GPD families.
}
\details{
The density is
\deqn{
f(x) = \left(\frac{\lambda}{2 \pi x^3}\right)^{1/2}
\exp\left\{- \frac{\lambda (x - \mu)^2}{2 \mu^2 x}\right\}, \qquad x > 0.
}

These uppercase NIMBLE-compatible functions are scalar (\code{x}/\code{q} and \code{n = 1}).
For vectorized R usage, use \code{\link[=base_lowercase]{base_lowercase()}}.

The inverse Gaussian is the first-passage-time distribution of a Brownian motion with positive
drift. Under the \eqn{(\mu, \lambda)} parameterization used here, the mean is \eqn{E(X)=\mu} and
the variance is \eqn{\mathrm{Var}(X)=\mu^3/\lambda}. The implementation follows that
parameterization throughout the package, so inverse-Gaussian mixture and splice families inherit
the same interpretation.

The distribution function is evaluated through the standard normal representation
\deqn{
F(x) =
\Phi\left(\sqrt{\frac{\lambda}{x}}\left(\frac{x}{\mu}-1\right)\right) +
\exp\left(\frac{2\lambda}{\mu}\right)
\Phi\left(-\sqrt{\frac{\lambda}{x}}\left(\frac{x}{\mu}+1\right)\right),
}
and the quantile is obtained numerically because no simple closed form is available.
}
\section{Functions}{
\itemize{
\item \code{dInvGauss()}: Inverse Gaussian density function

\item \code{pInvGauss()}: Inverse Gaussian distribution function

\item \code{rInvGauss()}: Inverse Gaussian random generation

\item \code{qInvGauss()}: Inverse Gaussian quantile function

}}
\examples{
mean <- 2
shape <- 5

dInvGauss(2.0, mean, shape, log = 0)
pInvGauss(2.0, mean, shape, lower.tail = 1, log.p = 0)
qInvGauss(0.50, mean, shape)
qInvGauss(0.95, mean, shape)
replicate(10, rInvGauss(1, mean, shape))

}
\seealso{
\code{\link[=InvGauss_mix]{InvGauss_mix()}}, \code{\link[=InvGauss_gpd]{InvGauss_gpd()}}, \code{\link[=base_lowercase]{base_lowercase()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}}.

Other base bulk distributions: 
\code{\link{amoroso}},
\code{\link{cauchy}}
}
\concept{base bulk distributions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\docType{data}
\name{causal_alt_real500_p4_k2}
\alias{causal_alt_real500_p4_k2}
\title{Alternative Real-Line Causal Data Set With Four Predictors}
\format{
A list with:
\describe{
\item{y}{Numeric outcome vector.}
\item{A}{Binary treatment indicator (0/1).}
\item{X}{data.frame with x1-x4.}
\item{meta}{List with N, support, p, K0, K1, tail, exceed_frac.}
\item{truth}{List with kernel0, kernel1, params0, params1, tail_params.}
}
}
\usage{
causal_alt_real500_p4_k2
}
\description{
Causal dataset (N=500, p=4) with different real-line kernels by arm.
Intended for alternating-kernel causal vignettes (GPD=FALSE).
}
\examples{
head(causal_alt_real500_p4_k2$X)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{dpmgpd.causal}
\alias{dpmgpd.causal}
\title{Fit a causal two-arm Dirichlet process mixture with a spliced GPD tail}
\usage{
dpmgpd.causal(
  y = NULL,
  X = NULL,
  treat = NULL,
  data = NULL,
  mcmc = list(),
  formula = NULL,
  ...
)
}
\arguments{
\item{y}{Either a response vector or a causal bundle object.}

\item{X}{Optional design matrix/data.frame.}

\item{treat}{Binary treatment indicator.}

\item{data}{Optional data.frame used with \code{formula}.}

\item{mcmc}{Named list of run arguments passed to \code{mcmc()} (including
optional performance controls such as \code{parallel_arms},
\code{workers}, \code{timing}, and \code{z_update_every}).}

\item{formula}{Optional formula.}

\item{...}{Additional build arguments passed to \code{\link{build_causal_bundle}}.}
}
\value{
A fitted object of class \code{"causalmixgpd_causal_fit"}.
}
\description{
\code{dpmgpd.causal()} is the highest-level causal fitting wrapper. It builds
or accepts a causal bundle, runs posterior sampling for the treated and
control arms, and returns a single causal fit ready for prediction and effect
estimation.
}
\details{
The arm-specific predictive distributions
\eqn{F_1(y \mid x)} and \eqn{F_0(y \mid x)} inherit
the spliced bulk-tail structure. Downstream causal estimands are computed as
functionals of these two predictive laws, for example
\deqn{\mathrm{QTE}(\tau) = Q_1(\tau) - Q_0(\tau), \qquad
\mathrm{ATE} = E(Y_1) - E(Y_0).}
}
\examples{
\donttest{
data("causal_pos500_p3_k2", package = "CausalMixGPD")
idx <- seq_len(30)
fit <- dpmgpd.causal(
  y = causal_pos500_p3_k2$y[idx],
  X = causal_pos500_p3_k2$X[idx, , drop = FALSE],
  treat = causal_pos500_p3_k2$A[idx],
  kernel = "gamma",
  backend = "sb",
  components = 3,
  mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1,
              show_progress = FALSE, quiet = TRUE)
)
ate(fit, show_progress = FALSE)
}
}
\seealso{
\code{\link{build_causal_bundle}},
\code{\link{bundle}}, \code{\link{dpmix.causal}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{summary.dpmixgpd_cluster_fit}
\alias{summary.dpmixgpd_cluster_fit}
\title{Summarize a cluster fit}
\usage{
\method{summary}{dpmixgpd_cluster_fit}(
  object,
  burnin = NULL,
  thin = NULL,
  top_n = 5L,
  order_by = c("size", "label"),
  vars = NULL,
  ...
)
}
\arguments{
\item{object}{A cluster fit.}

\item{burnin}{Number of initial posterior draws to discard.}

\item{thin}{Keep every \code{thin}-th posterior draw.}

\item{top_n}{Number of populated clusters to profile when descriptive summaries are available.}

\item{order_by}{Ordering rule for descriptive cluster profiles:
\itemize{
\item \code{"size"}: decreasing cluster size
\item \code{"label"}: ascending cluster label
}}

\item{vars}{Optional character vector of numeric columns to summarize within each cluster.}

\item{...}{Unused.}
}
\value{
Summary list with the number of retained clusters, cluster sizes, optional
cluster-level descriptive summaries, and the burn-in/thinning settings used to construct the
summary.
}
\description{
Summarize the posterior clustering induced by the Dahl representative partition.
}
\details{
This summary is based on \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}} with \code{type = "label"}. The reported
cluster count \eqn{K^*} is the number of unique labels in the representative partition rather
than the number of components available in the truncated sampler.
}
\seealso{
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}},
\code{\link[=summary.dpmixgpd_cluster_labels]{summary.dpmixgpd_cluster_labels()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_fit}
\alias{print.mixgpd_fit}
\title{Print a one-arm fitted model}
\usage{
\method{print}{mixgpd_fit}(x, ...)
}
\arguments{
\item{x}{A fitted object of class \code{"mixgpd_fit"}.}

\item{...}{Unused.}
}
\value{
\code{x} invisibly.
}
\description{
\code{print.mixgpd_fit()} gives a compact header for a fitted one-arm model.
It is meant as a quick identity check rather than a full posterior summary.
}
\details{
The fitted object represents posterior draws from a bulk mixture model, or
from its spliced bulk-tail extension when \code{GPD = TRUE}. For the bulk part, the
predictive law has the mixture form
\deqn{f(y \mid x) = \sum_{k=1}^{K} w_k(x) f_k(y \mid x, \theta_k).}
When a GPD tail is active, exceedances above the threshold are instead routed
through the generalized Pareto tail attached to the same bulk mixture.

The print method reports only the model identity and basic metadata. Use
\code{summary()} for parameter-level posterior summaries, \code{predict()} for
predictive functionals, and \code{plot()} for chain diagnostics.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
print(fit)
}
}
\seealso{
\code{\link{summary.mixgpd_fit}}, \code{\link{params}},
\code{\link{predict.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{fitted.mixgpd_fit}
\alias{fitted.mixgpd_fit}
\title{Fitted values on the training design}
\usage{
\method{fitted}{mixgpd_fit}(
  object,
  type = c("mean", "median", "quantile"),
  p = 0.5,
  level = 0.95,
  interval = "credible",
  seed = 1,
  ...
)
}
\arguments{
\item{object}{A fitted object of class \code{"mixgpd_fit"} (must have covariates).}

\item{type}{Which fitted functional to return:
\itemize{
\item \code{"mean"}: posterior predictive mean
\item \code{"median"}: posterior predictive median
\item \code{"quantile"}: posterior predictive quantile
at level \code{p}
}}

\item{p}{Quantile level used when \code{type = "quantile"}.}

\item{level}{Credible level for confidence intervals (default 0.95 for 95 percent credible intervals).}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile
intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{seed}{Random seed used for deterministic fitted values.}

\item{...}{Unused.}
}
\value{
A data frame with columns for fitted values, optional intervals, and
residuals computed on the training sample.
}
\description{
\code{fitted.mixgpd_fit()} is a thin training-data wrapper around
\code{\link{predict.mixgpd_fit}} for conditional models.
}
\details{
The method returns posterior predictive fitted values on the observed design
matrix. It is available only when the fitted model stored covariates.
}
\examples{
\donttest{
# Conditional model (with covariates X)
y <- abs(stats::rnorm(25)) + 0.1
X <- data.frame(x1 = stats::rnorm(25), x2 = stats::runif(25))
bundle <- build_nimble_bundle(y = y, X = X, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
fitted(fit)
fitted(fit, level = 0.90)
fitted(fit, interval = "hpd")  # HPD intervals
fitted(fit, interval = NULL)   # No intervals
}
}
\seealso{
\code{\link{predict.mixgpd_fit}}, \code{\link{residuals.mixgpd_fit}},
\code{\link{plot.mixgpd_fitted}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_predict}
\alias{print.mixgpd_predict}
\title{Print Prediction Results}
\usage{
\method{print}{mixgpd_predict}(x, digits = 3, max_rows = 10L, ...)
}
\arguments{
\item{x}{A prediction object returned by \code{predict.mixgpd_fit()}.}

\item{digits}{Number of digits to print for numeric summaries.}

\item{max_rows}{Maximum rows to print when the prediction contains a data frame.}

\item{...}{Unused.}
}
\value{
The input object, invisibly.
}
\description{
Compact print method for objects returned by \code{\link{predict.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.extract_draws}
\alias{.extract_draws}
\title{Safely coerce MCMC samples to a numeric matrix}
\usage{
.extract_draws(
  object,
  pars = NULL,
  chains = c("stack", "first"),
  epsilon = NULL
)
}
\arguments{
\item{object}{A mixgpd_fit.}

\item{pars}{Optional character vector of parameter names to keep (exact match).}
}
\value{
Numeric matrix of draws (iter x parameters).
}
\description{
Safely coerce MCMC samples to a numeric matrix
}
\details{
This is the matrix-oriented companion to \code{.get_samples_mcmclist()}. It can
either stack all chains or keep only the first chain, always removes
stick-breaking \code{v} variables, applies the standard component truncation rule,
and optionally filters to an exact set of parameter names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{cqte.causalmixgpd_causal_fit}
\alias{cqte.causalmixgpd_causal_fit}
\alias{cqte}
\title{Conditional quantile treatment effects}
\usage{
\method{cqte}{causalmixgpd_causal_fit}(
  fit,
  probs = c(0.1, 0.5, 0.9),
  newdata = NULL,
  interval = "credible",
  level = 0.95,
  show_progress = TRUE
)
}
\arguments{
\item{fit}{A \code{"causalmixgpd_causal_fit"} object from \code{run_mcmc_causal()}.}

\item{probs}{Numeric vector of probabilities in (0, 1) specifying the quantile levels
of the outcome distribution to estimate treatment effects at.}

\item{newdata}{Optional data.frame or matrix of covariates for prediction.
If \code{NULL}, uses the training covariates stored in \code{fit}.}

\item{interval}{Character or NULL; type of credible interval:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile intervals
\item \code{"hpd"}: highest posterior density intervals
}}

\item{level}{Numeric credible level for intervals (default 0.95 for 95 percent CI).}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}
}
\value{
An object of class \code{"causalmixgpd_qte"} containing the CQTE
summary, the probability grid, and the treated/control prediction objects
used to construct the effect. The returned object includes a top-level
\code{$fit_df} data frame for direct extraction.
}
\description{
\code{cqte()} evaluates treated-minus-control predictive quantiles at
user-supplied covariate rows.
}
\details{
For each prediction row \eqn{x}, the conditional quantile treatment effect is
\deqn{\mathrm{CQTE}(\tau, x) = Q_1(\tau \mid x) -
Q_0(\tau \mid x).}

This estimand is available only for conditional causal models with
covariates. For marginal quantile contrasts over the empirical covariate
distribution, use \code{\link{qte}} or \code{\link{qtt}}.

If the fit includes a PS block, the same PS adjustment is applied to both arm
predictions before differencing.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         components = 3, mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb, show_progress = FALSE)
cqte(fit, probs = c(0.5, 0.9), newdata = X[1:5, , drop = FALSE])
cqte(fit, probs = c(0.5, 0.9), interval = "credible", level = 0.90)  # 90\% CI
cqte(fit, probs = c(0.5, 0.9), interval = "hpd")  # HPD intervals
cqte(fit, probs = c(0.5, 0.9), interval = NULL)   # No intervals
}
}
\seealso{
\code{\link{qte}}, \code{\link{qtt}}, \code{\link{cate}},
\code{\link{predict.causalmixgpd_causal_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.summarize_posterior}
\alias{.summarize_posterior}
\title{Summarize posterior draws for selected parameters}
\usage{
.summarize_posterior(object, pars = NULL, probs = c(0.025, 0.5, 0.975))
}
\arguments{
\item{object}{mixgpd_fit}

\item{pars}{character vector; if NULL uses all non-v parameters}

\item{probs}{quantiles to report}
}
\value{
data.frame with mean/sd/quantiles + ess/rhat where available
}
\description{
Summarize posterior draws for selected parameters
}
\details{
This helper powers the one-arm summary methods. It extracts the retained draw
matrix, chooses a default set of non-redundant parameters when \code{pars} is not
supplied, and computes posterior means, standard deviations, quantiles,
effective sample sizes, and Gelman diagnostics when available.

The resulting table is parameter oriented rather than prediction oriented. It
is the internal workhorse behind \code{summary.mixgpd_fit()}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{plot.dpmixgpd_cluster_bundle}
\alias{plot.dpmixgpd_cluster_bundle}
\alias{plot.dpmixgpd_cluster_fit}
\alias{plot.dpmixgpd_cluster_labels}
\alias{plot.dpmixgpd_cluster_psm}
\title{Plot a cluster bundle}
\usage{
\method{plot}{dpmixgpd_cluster_bundle}(x, plotly = getOption("CausalMixGPD.plotly", FALSE), ...)

\method{plot}{dpmixgpd_cluster_fit}(
  x,
  which = c("psm", "k", "sizes", "summary"),
  burnin = NULL,
  thin = NULL,
  psm_max_n = 2000L,
  top_n = 5L,
  order_by = c("size", "label"),
  plotly = getOption("CausalMixGPD.plotly", FALSE),
  ...
)

\method{plot}{dpmixgpd_cluster_labels}(
  x,
  type = c("sizes", "certainty", "summary"),
  top_n = 5L,
  order_by = c("size", "label"),
  plotly = getOption("CausalMixGPD.plotly", FALSE),
  ...
)

\method{plot}{dpmixgpd_cluster_psm}(
  x,
  psm_max_n = x$psm_max_n \%||\% 2000L,
  order_by = c("label", "hclust", "input"),
  plotly = getOption("CausalMixGPD.plotly", FALSE),
  ...
)
}
\arguments{
\item{x}{Cluster PSM object.}

\item{plotly}{Logical; if \code{TRUE}, convert the \code{ggplot2} output to a \code{plotly} /
\code{htmlwidget} representation via \code{.wrap_plotly()}. Defaults to
\code{getOption("CausalMixGPD.plotly", FALSE)}.}

\item{...}{Unused.}

\item{which}{Plot type:
\itemize{
\item \code{"psm"}: posterior similarity matrix heatmap
\item \code{"k"}: posterior number of occupied clusters
\item \code{"sizes"}: bar chart of representative cluster sizes
\item \code{"summary"}: cluster-specific response summaries
}}

\item{burnin}{Number of initial posterior draws to discard.}

\item{thin}{Keep every \code{thin}-th posterior draw.}

\item{psm_max_n}{Maximum allowed matrix size for plotting.}

\item{top_n}{Number of populated representative clusters to display for
\code{type = "sizes"} or \code{type = "summary"}. Use \code{NULL} to
display all populated clusters.}

\item{order_by}{Ordering rule for rows and columns:
\itemize{
\item \code{"label"}: order by representative cluster labels when available
\item \code{"hclust"}: order by hierarchical clustering of \code{1 - PSM}
\item \code{"input"}: preserve input order
}}

\item{type}{Plot type:
\itemize{
\item \code{"sizes"}: bar chart of representative cluster sizes
\item \code{"certainty"}: assignment certainty distribution
\item \code{"summary"}: cluster-specific response boxplots
}}
}
\value{
A \code{ggplot2} object or a \code{plotly}/\code{htmlwidget} object when \code{plotly = TRUE}.

A \code{ggplot2} object or a \code{plotly}/\code{htmlwidget} object when \code{plotly = TRUE}.

A \code{ggplot2} object or a \code{plotly}/\code{htmlwidget} object when \code{plotly = TRUE}.

A \code{ggplot2} object or a \code{plotly}/\code{htmlwidget} object when \code{plotly = TRUE}.
}
\description{
Produce a compact graphical summary of the cluster bundle metadata.

Visualize either the posterior similarity matrix, the posterior number of occupied clusters, the
size distribution of the representative clusters, or cluster-specific response summaries.

Visualize representative cluster sizes, assignment certainty, or cluster-specific response
summaries. For \code{type = "summary"}, the response view is shown as boxplots ordered by
cluster size or label. When \code{x} comes from \code{predict(..., newdata = ...)}, only clusters
represented in the new sample are displayed.

Heatmap of pairwise posterior co-clustering probabilities.
}
\details{
The bundle plot is a metadata display rather than an inferential graphic. It
mirrors the structural fields reported by \code{print()} and \code{summary()} in a
single panel so the pre-MCMC clustering specification can be reviewed in a
figure-oriented workflow or notebook.

Because the object has not been sampled yet, no representative partition or
posterior uncertainty is shown here. Use \code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}},
\code{\link[=plot.dpmixgpd_cluster_labels]{plot.dpmixgpd_cluster_labels()}}, or \code{\link[=plot.dpmixgpd_cluster_psm]{plot.dpmixgpd_cluster_psm()}} after
fitting when you need substantive clustering output.

This plot method exposes the main posterior diagnostics for clustering. The
\code{which = "k"} view tracks the number of occupied clusters across retained
draws, \code{which = "psm"} visualizes pairwise co-clustering probabilities,
\code{which = "sizes"} displays the size profile of the representative partition,
and \code{which = "summary"} shows response summaries conditional on the selected
representative labels.

The representative partition is obtained from
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}} using Dahl's least-squares rule. As a
result, the \code{sizes} and \code{summary} views describe that representative
clustering rather than the full posterior distribution over partitions.

This method visualizes the representative partition stored in a
\code{dpmixgpd_cluster_labels} object. The \code{sizes} view emphasizes the empirical
distribution of the selected clusters, the \code{certainty} view summarizes the
assignment scores \eqn{\max_k p_{ik}}, and the \code{summary} view compares the
attached response data across representative clusters.

For new-data prediction, the plots are always interpreted relative to the
representative training clusters. That is why only clusters observed in the
predicted sample are shown even though the training partition may contain
additional occupied groups.

The heatmap visualizes the matrix
\deqn{
\mathrm{PSM}_{ij} \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}),
}
so larger values indicate pairs of observations that are stably allocated to
the same cluster over the retained posterior draws.

Because the PSM is an \eqn{n \times n} object, plotting and even storing it
becomes expensive for large \code{n}. The \code{psm_max_n} argument is therefore a
deliberate guard against accidental quadratic memory use.
}
\seealso{
\code{\link[=summary.dpmixgpd_cluster_bundle]{summary.dpmixgpd_cluster_bundle()}}, \code{\link[=dpmix.cluster]{dpmix.cluster()}}, \code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}.

\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}},
\code{\link[=plot.dpmixgpd_cluster_psm]{plot.dpmixgpd_cluster_psm()}}, \code{\link[=plot.dpmixgpd_cluster_labels]{plot.dpmixgpd_cluster_labels()}}.

\code{\link[=summary.dpmixgpd_cluster_labels]{summary.dpmixgpd_cluster_labels()}}, \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}.

\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=summary.dpmixgpd_cluster_psm]{summary.dpmixgpd_cluster_psm()}},
\code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glue.R
\name{check_glue_validity}
\alias{check_glue_validity}
\title{Validate bulk+tail glue for MixGPD predictive distribution}
\usage{
check_glue_validity(
  fit,
  x = NULL,
  grid = NULL,
  n_draws = 50L,
  tol = 1e-08,
  check_continuity = TRUE,
  eps = 1e-06
)
}
\arguments{
\item{fit}{A \code{mixgpd_fit} object.}

\item{x}{Optional design matrix for conditional models. If \code{NULL}, uses training \code{X}.}

\item{grid}{Numeric evaluation grid. If \code{NULL}, defaults to a grid based on training \code{y}.}

\item{n_draws}{Number of posterior draws to check (sampled without replacement when possible).}

\item{tol}{Numerical tolerance for monotonicity/range checks.}

\item{check_continuity}{Logical; if \code{TRUE} and GPD is enabled, checks continuity at the threshold.}

\item{eps}{Small offset used for threshold continuity check.}
}
\value{
A list with per-check pass/fail flags and summaries of violations.
}
\description{
This diagnostic checks whether the implied predictive distribution behaves like
a valid distribution (monotone CDF in \eqn{[0,1]}, nonnegative density, and
sensible behavior around the threshold when a GPD tail is enabled).
}
\details{
The check is performed draw-by-draw on a user-specified grid. It is intended
for development, debugging, and CI (not for routine large-scale use).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{bundle}
\alias{bundle}
\title{Build the workflow bundle used by the package fitters}
\usage{
bundle(
  y = NULL,
  X = NULL,
  treat = NULL,
  data = NULL,
  formula = NULL,
  GPD = FALSE,
  ...
)
}
\arguments{
\item{y}{Either a response vector or an existing bundle.}

\item{X}{Optional design matrix/data.frame.}

\item{treat}{Optional binary treatment indicator.}

\item{data}{Optional data.frame used with \code{formula}.}

\item{formula}{Optional formula.}

\item{GPD}{Logical; include GPD tail in build mode.}

\item{...}{Additional arguments passed to \code{build_nimble_bundle()} or
\code{build_causal_bundle()}.}
}
\value{
A \code{"causalmixgpd_bundle"} for one-arm models or a
\code{"causalmixgpd_causal_bundle"} for causal models. The bundle stores
code-generation inputs, monitor policy, and default MCMC settings, but it
does not run MCMC.
}
\description{
\code{bundle()} is the main workflow constructor. It converts raw inputs,
a formula/data pair, or an already prepared bundle into the canonical
object consumed by \code{\link{mcmc}}, \code{\link{dpmix}},
\code{\link{dpmgpd}}, \code{\link{dpmix.causal}}, and
\code{\link{dpmgpd.causal}}.
}
\details{
For one-arm models the returned object represents a bulk Dirichlet process
mixture, optionally augmented with a spliced generalized Pareto tail. For
causal models the returned object contains two arm-specific outcome bundles
plus an optional propensity score block.

The workflow is:
\enumerate{
\item prepare a bundle with \code{bundle()},
\item run posterior sampling with \code{\link{mcmc}} or one of the
\code{dpmix*}/\code{dpmgpd*} wrappers,
\item inspect the fitted object with \code{\link{summary.mixgpd_fit}},
\code{\link{params}}, \code{\link{predict.mixgpd_fit}}, or the causal
estimand helpers.
}

Setting \code{GPD = TRUE} requests the spliced bulk-tail model with
conditional distribution
\deqn{F(y \mid x) = F_{\mathrm{bulk}}(y \mid x)\mathbf{1}\{y \le u(x)\} +
\left[p_u(x) + \{1 - p_u(x)\}F_{\mathrm{GPD}}(y \mid x)\right]\mathbf{1}\{y > u(x)\},}
where \eqn{p_u(x)} is the bulk probability below the threshold \eqn{u(x)}.

See the manuscript vignette for the DPM hierarchy, SB/CRP representations,
and the spliced bulk-tail construction used throughout the package.
}
\seealso{
\code{\link{build_nimble_bundle}}, \code{\link{build_causal_bundle}},
\code{\link{mcmc}}, \code{\link{dpmix}}, \code{\link{dpmgpd}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{plot.mixgpd_predict}
\alias{plot.mixgpd_predict}
\alias{plot.causalmixgpd_causal_predict}
\title{Plot prediction results}
\usage{
\method{plot}{mixgpd_predict}(x, y = NULL, ...)

\method{plot}{causalmixgpd_causal_predict}(x, y = NULL, ...)
}
\arguments{
\item{x}{Object of class \code{causalmixgpd_causal_predict}.}

\item{y}{Ignored.}

\item{...}{Additional arguments passed to ggplot2 functions.}
}
\value{
Invisibly returns the ggplot object.

A ggplot object or a list of ggplot objects.
}
\description{
Generates type-specific visualizations for prediction objects returned by
\code{predict.mixgpd_fit()}. Each prediction type produces a tailored plot:
\itemize{
\item \code{quantile}: Quantile indices vs estimates with credible intervals
\item \code{sample}: Histogram of samples with density overlay
\item \code{mean}: Histogram density with posterior mean vertical line and CI bounds
\item \code{density}: Density values vs evaluation points
\item \code{survival}: Survival function (decreasing y values)
}

S3 method for visualizing causal predictions from \code{predict.causalmixgpd_causal_fit()}.
For mean/quantile, plots treated/control and treatment effect versus PS (or index).
For \code{type = "sample"}, plots arm-level posterior predictive samples alongside
treatment-effect samples. For density/prob, plots treated/control values versus y.
}
\details{
The plotting method is tied to the predictive functional stored in the input
object. Quantile and mean outputs display posterior point summaries and
intervals, density and survival outputs show evaluated functions on the
supplied grid, and posterior samples are visualized as empirical predictive
draws.

In every case the plot reflects the quantity requested from
\code{predict.mixgpd_fit()} after integrating over the retained posterior draws. It
is therefore distinct from parameter-level summaries and from chain
diagnostics.

The causal prediction object carries arm-specific predictions together with the
implied contrast. For mean predictions, the contrast is
\eqn{m_1(x) - m_0(x)}. For quantile predictions, the contrast is
\eqn{Q_{Y^1}(\tau \mid x) - Q_{Y^0}(\tau \mid x)}. The plotting method keeps
those arm and contrast views synchronized.

Unlike \code{plot.causalmixgpd_causal_fit()}, which diagnoses MCMC behavior inside
the outcome models, this method visualizes predictive quantities after
posterior integration. It is therefore the natural plotting method once the
user has already accepted the fitted-model diagnostics.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)

# Quantile prediction with plot
pred_q <- predict(fit, type = "quantile", index = c(0.25, 0.5, 0.75))
plot(pred_q)

# Sample prediction with plot
pred_s <- predict(fit, type = "sample", nsim = 500)
plot(pred_s)

# Mean prediction with plot
pred_m <- predict(fit, type = "mean", nsim_mean = 300)
plot(pred_m)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.posterior_summarize}
\alias{.posterior_summarize}
\title{Summarize posterior draws (mean + quantiles)}
\usage{
.posterior_summarize(
  draws,
  probs = c(0.025, 0.5, 0.975),
  interval = "credible"
)
}
\arguments{
\item{draws}{Numeric vector, matrix, or array with draws in last dimension.}

\item{probs}{Numeric quantile probs.}

\item{interval}{Character or NULL; interval type:
\itemize{
\item \code{NULL}: no interval
\item \code{"credible"} (default): equal-tailed quantile
intervals
\item \code{"hpd"}: highest posterior density intervals
}}
}
\value{
List with estimate, lower, upper, and q.
}
\description{
Summarize posterior draws (mean + quantiles)
}
\details{
The last dimension of \code{draws} is interpreted as the posterior-draw dimension.
This helper collapses that dimension to posterior means and interval summaries,
while preserving the leading dimensions of the input object. It is used
throughout prediction and treatment-effect code to turn per-draw evaluations
into reported posterior summaries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cauchy.R
\name{cauchy_mix_lowercase}
\alias{cauchy_mix_lowercase}
\alias{dcauchymix}
\alias{pcauchymix}
\alias{qcauchymix}
\alias{rcauchymix}
\title{Lowercase vectorized Cauchy mixture distribution functions}
\usage{
dcauchymix(x, w, location, scale, log = FALSE)

pcauchymix(q, w, location, scale, lower.tail = TRUE, log.p = FALSE)

qcauchymix(
  p,
  w,
  location,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rcauchymix(n, w, location, scale)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{location, scale}{Numeric vectors of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar Cauchy mixture functions in this file.
}
\details{
These are vectorized R wrappers around the scalar Cauchy-mixture routines.
They retain the same location-scale parameterization and the same inverse-CDF
logic for simulation and quantiles. The lowercase functions do not alter the
heavy-tail theory of the underlying Cauchy components; they apply the scalar
routines elementwise to vector inputs in R.
}
\section{Functions}{
\itemize{
\item \code{dcauchymix()}: Cauchy mixture density (vectorized)

\item \code{pcauchymix()}: Cauchy mixture distribution function (vectorized)

\item \code{qcauchymix()}: Cauchy mixture quantile function (vectorized)

\item \code{rcauchymix()}: Cauchy mixture random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.3, 0.1)
loc <- c(-1, 0, 1)
scl <- c(1, 1.2, 2)

dcauchymix(c(-2, 0, 2), w = w, location = loc, scale = scl)
rcauchymix(5, w = w, location = loc, scale = scl)

}
\seealso{
\code{\link[=cauchy_mix]{cauchy_mix()}}, \code{\link[=cauchy]{cauchy()}}, \code{\link[=bundle]{bundle()}}, \code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{ess_summary}
\alias{ess_summary}
\title{Effective sample size summaries for fitted models}
\usage{
ess_summary(
  fit,
  params = NULL,
  per_chain = TRUE,
  wall_time = NULL,
  robust = TRUE,
  ...
)
}
\arguments{
\item{fit}{A \code{"mixgpd_fit"} or \code{"causalmixgpd_causal_fit"} object.}

\item{params}{Optional character vector of parameter names/patterns. If
\code{NULL}, a fixed canonical set is auto-resolved.}

\item{per_chain}{Logical; if \code{TRUE}, include per-chain ESS rows.}

\item{wall_time}{Optional numeric total MCMC time in seconds. If \code{NULL},
uses \code{fit$timing$mcmc} when available.}

\item{robust}{Logical; if \code{TRUE}, ignore missing parameters.}

\item{...}{Unused.}
}
\value{
Object of class \code{"mixgpd_ess_summary"} with elements
\code{table}, \code{overall}, and \code{meta}.
}
\description{
\code{ess_summary()} reports effective sample size diagnostics for posterior
draws, optionally scaled by wall-clock time.
}
\details{
This is a convergence and efficiency diagnostic, not a model summary. For
causal fits the function evaluates each outcome arm separately and tags the
rows accordingly.
}
\seealso{
\code{\link{summary.mixgpd_fit}}, \code{\link{plot.mixgpd_fit}},
\code{\link{params}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lognormal.R
\name{lognormal_gpd}
\alias{lognormal_gpd}
\alias{dLognormalGpd}
\alias{pLognormalGpd}
\alias{rLognormalGpd}
\alias{qLognormalGpd}
\title{Lognormal with a GPD tail}
\usage{
dLognormalGpd(x, meanlog, sdlog, threshold, tail_scale, tail_shape, log = 0)

pLognormalGpd(
  q,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rLognormalGpd(n, meanlog, sdlog, threshold, tail_scale, tail_shape)

qLognormalGpd(
  p,
  meanlog,
  sdlog,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{meanlog}{Numeric scalar log-mean parameter for the Lognormal bulk.}

\item{sdlog}{Numeric scalar log-standard deviation for the Lognormal bulk.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qLognormalGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
lognormal bulk.
}
\details{
This is the single-lognormal counterpart of \code{\link[=lognormal_mixgpd]{lognormal_mixgpd()}}. If \eqn{F_{LN}(u)} denotes the
bulk probability below the threshold, then the spliced density is
\deqn{
f(x) =
\left\{
\begin{array}{ll}
f_{LN}(x \mid \mu,\sigma), & x < u, \\
\{1-F_{LN}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
The ordinary mean is finite only when the GPD tail has \eqn{\xi < 1};
otherwise the package requires restricted means or quantiles.
}
\section{Functions}{
\itemize{
\item \code{dLognormalGpd()}: Lognormal + GPD tail density

\item \code{pLognormalGpd()}: Lognormal + GPD tail distribution function

\item \code{rLognormalGpd()}: Lognormal + GPD tail random generation

\item \code{qLognormalGpd()}: Lognormal + GPD tail quantile function

}}
\examples{
meanlog <- 0.4
sdlog <- 0.35
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dLognormalGpd(4.0, meanlog, sdlog, threshold, tail_scale, tail_shape, log = FALSE)
pLognormalGpd(4.0, meanlog, sdlog, threshold, tail_scale, tail_shape,
             lower.tail = TRUE, log.p = FALSE)
qLognormalGpd(0.50, meanlog, sdlog, threshold, tail_scale, tail_shape)
qLognormalGpd(0.95, meanlog, sdlog, threshold, tail_scale, tail_shape)
replicate(10, rLognormalGpd(1, meanlog, sdlog, threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=lognormal_mix]{lognormal_mix()}}, \code{\link[=lognormal_mixgpd]{lognormal_mixgpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=lognormal_lowercase]{lognormal_lowercase()}}.

Other lognormal kernel families: 
\code{\link{lognormal_mix}},
\code{\link{lognormal_mixgpd}}
}
\concept{lognormal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_ps_fit}
\alias{print.causalmixgpd_ps_fit}
\title{Print a propensity score fit}
\usage{
\method{print}{causalmixgpd_ps_fit}(x, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_ps_fit"} object.}

\item{...}{Unused.}
}
\value{
The input object (invisibly).
}
\description{
Print a propensity score fit
}
\details{
A propensity-score fit models the treatment assignment probability
\eqn{e(x) = \Pr(A = 1 \mid X = x)}. The printed header identifies which PS
family was fitted, but it intentionally omits coefficient-level summaries.

Use \code{summary()} on the same object when you need posterior means, spread, and
intervals for the monitored PS parameters. The compact print method is mainly
an identity check inside larger causal workflows.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/1-registry.R
\name{init_kernel_registry}
\alias{init_kernel_registry}
\title{Initialize kernel registries}
\usage{
init_kernel_registry()
}
\value{
Invisibly returns TRUE.
}
\description{
Creates/refreshes registries used by the model specification compiler and
code generators. Each kernel entry stores bulk parameters, supports, default
regression/link behavior, and distribution signatures for SB/CRP backends.
}
\details{
The kernel registry is the package-level contract that keeps model building,
prediction, and documentation aligned. Each entry records the natural bulk
parameters for one kernel, the support constraints they must satisfy, the
default covariate-link strategy, and the backend-specific distribution names
used when generating NIMBLE code.

The companion tail registry records the generalized Pareto tail parameters
\eqn{u} (threshold), \eqn{\sigma_u} (tail scale), and \eqn{\xi_u} (tail
shape) together with their support and allowed modeling modes. Calling
\code{init_kernel_registry()} makes those contracts available in the package
namespace so later builders can validate requests without duplicating lookup
logic.
}
\examples{
init_kernel_registry()
reg <- get_kernel_registry()
names(reg)
tail <- get_tail_registry()
tail$params
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.extract_z_matrix}
\alias{.extract_z_matrix}
\title{Extract cluster assignment matrix from MCMC samples}
\usage{
.extract_z_matrix(object)
}
\arguments{
\item{object}{A \code{mixgpd_fit} object.}
}
\value{
Integer matrix with rows = posterior draws, cols = observations.
}
\description{
Extracts the posterior draws of cluster assignments \code{z[1:N]} from a fitted
mixgpd_fit object and returns them as an integer matrix (iterations x N).
}
\details{
Cluster samplers store latent labels as separate monitored nodes \code{z[i]}. This
helper locates those nodes in every retained chain, orders them by observation
index, stacks the chains, and returns the result as one integer matrix that is
ready for PSM and representative-partition calculations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{residuals.mixgpd_fit}
\alias{residuals.mixgpd_fit}
\title{Residual diagnostics on the training design}
\usage{
\method{residuals}{mixgpd_fit}(
  object,
  type = c("raw", "pit"),
  fitted_type = c("mean", "median"),
  pit = c("plugin", "bayes_mean", "bayes_draw"),
  pit_seed = NULL,
  ...
)
}
\arguments{
\item{object}{A fitted object of class \code{"mixgpd_fit"} (must have covariates).}

\item{type}{Residual type:
\itemize{
\item \code{"raw"}: observed minus fitted values
\item \code{"pit"}: probability integral transform
residuals (see \code{pit} argument)
}}

\item{fitted_type}{For \code{type = "raw"}, use fitted means or medians.}

\item{pit}{PIT mode for \code{type = "pit"}:
\itemize{
\item \code{"plugin"}: plug-in PIT using the posterior mean CDF.
\item \code{"bayes_mean"}: Bayesian PIT using draw-wise CDFs averaged over draws.
\item \code{"bayes_draw"}: Bayesian PIT using a single draw-wise CDF per observation.
}
Bayesian PIT modes drop invalid posterior draws using the same validation
rules as prediction and attach diagnostics via \code{attr(res, "pit_diagnostics")}.}

\item{pit_seed}{Optional integer seed for reproducible \code{bayes_draw} sampling.}

\item{...}{Unused.}
}
\value{
Numeric vector of residuals with length equal to the training sample
size. PIT variants attach diagnostic metadata as attributes.
}
\description{
\code{residuals.mixgpd_fit()} computes residual diagnostics for conditional
fits on the original training data.
}
\details{
Raw residuals are based on posterior predictive fitted means or medians.
PIT residuals instead assess calibration through the posterior predictive CDF.
The plug-in PIT uses a posterior mean CDF, while the Bayesian PIT variants
work draw by draw.

This method is not available for unconditional models because no training
design matrix is stored for observation-specific fitted values.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
X <- data.frame(x1 = stats::rnorm(25), x2 = stats::runif(25))
bundle <- build_nimble_bundle(y = y, X = X, backend = "sb", kernel = "lognormal",
                             GPD = FALSE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
pit_plugin <- residuals(fit, type = "pit", pit = "plugin")
pit_bayes_mean <- residuals(fit, type = "pit", pit = "bayes_mean", pit_seed = 1L)
pit_bayes_draw <- residuals(fit, type = "pit", pit = "bayes_draw", pit_seed = 1L)
attr(pit_bayes_draw, "pit_diagnostics")
}
}
\seealso{
\code{\link{fitted.mixgpd_fit}}, \code{\link{predict.mixgpd_fit}},
\code{\link{plot.mixgpd_fitted}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{print.dpmixgpd_cluster_bundle}
\alias{print.dpmixgpd_cluster_bundle}
\title{Print a cluster bundle}
\usage{
\method{print}{dpmixgpd_cluster_bundle}(x, ...)
}
\arguments{
\item{x}{A cluster bundle.}

\item{...}{Unused.}
}
\value{
\code{x}, invisibly.
}
\description{
Compact display for a \code{dpmixgpd_cluster_bundle} before MCMC is run.
}
\details{
A cluster bundle is the pre-sampling representation of the latent partition
model. It stores the formula-derived design, kernel choice, truncation level,
and the rule by which predictors enter the clustering model, but it does not
yet contain posterior draws of the latent labels \eqn{z_1, \dots, z_n}.

\code{print()} is intentionally brief. It is meant to confirm that the bundle
matches the requested clustering structure before you run MCMC with
\code{run_cluster_mcmc()} or the higher-level wrappers \code{\link[=dpmix.cluster]{dpmix.cluster()}} and
\code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}.
}
\seealso{
\code{\link[=dpmix.cluster]{dpmix.cluster()}}, \code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}, \code{\link[=summary.dpmixgpd_cluster_bundle]{summary.dpmixgpd_cluster_bundle()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.dahl_representative}
\alias{.dahl_representative}
\title{Find Dahl representative clustering}
\usage{
.dahl_representative(z_matrix, PSM)
}
\arguments{
\item{z_matrix}{Integer matrix (iterations x N) of cluster assignments.}

\item{PSM}{Posterior similarity matrix (N x N).}
}
\value{
List with components: draw_index (integer), labels (integer vector),
K (number of clusters).
}
\description{
Identifies the posterior draw that minimizes squared distance to the
posterior similarity matrix, following Dahl (2006). Returns relabeled
cluster assignments as consecutive integers 1, 2, ..., K.
}
\details{
For each posterior draw, the helper forms its adjacency matrix and computes
the squared Frobenius distance to the PSM. The selected representative draw is
the one that minimizes that loss, which is Dahl's least-squares rule for
choosing one clustering from the posterior sample.
}
\references{
Dahl, D. B. (2006). Model-based clustering for expression data
via a Dirichlet process mixture model. In M. Vannucci, et al. (Eds.),
Bayesian Inference for Gene Expression and Proteomics (pp. 201-218).
Cambridge University Press.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/laplace.R
\name{laplace_gpd}
\alias{laplace_gpd}
\alias{dLaplaceGpd}
\alias{pLaplaceGpd}
\alias{rLaplaceGpd}
\alias{qLaplaceGpd}
\title{Laplace with a GPD tail}
\usage{
dLaplaceGpd(x, location, scale, threshold, tail_scale, tail_shape, log = 0)

pLaplaceGpd(
  q,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rLaplaceGpd(n, location, scale, threshold, tail_scale, tail_shape)

qLaplaceGpd(
  p,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{location}{Numeric scalar location parameter for the Laplace bulk.}

\item{scale}{Numeric scalar scale parameter for the Laplace bulk.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qLaplaceGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
Laplace bulk.
}
\details{
This topic pairs a single Laplace bulk distribution with a generalized Pareto exceedance tail.
The splice is continuous at the threshold because the tail density is multiplied by the Laplace
survival probability at that threshold.

The ordinary mean exists only when the GPD tail has \eqn{\xi < 1}. If the fitted tail is too
heavy for an ordinary mean to exist, the package directs users to restricted means or quantiles
rather than returning an unstable mean summary.
}
\section{Functions}{
\itemize{
\item \code{dLaplaceGpd()}: Laplace + GPD tail density

\item \code{pLaplaceGpd()}: Laplace + GPD tail distribution function

\item \code{rLaplaceGpd()}: Laplace + GPD tail random generation

\item \code{qLaplaceGpd()}: Laplace + GPD tail quantile function

}}
\examples{
location <- 0.5
scale <- 1.0
threshold <- 1
tail_scale <- 1.0
tail_shape <- 0.2

dLaplaceGpd(2.0, location, scale, threshold, tail_scale, tail_shape, log = FALSE)
pLaplaceGpd(2.0, location, scale, threshold, tail_scale, tail_shape,
           lower.tail = TRUE, log.p = FALSE)
qLaplaceGpd(0.50, location, scale, threshold, tail_scale, tail_shape)
qLaplaceGpd(0.95, location, scale, threshold, tail_scale, tail_shape)
replicate(10, rLaplaceGpd(1, location, scale, threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=laplace_mix]{laplace_mix()}}, \code{\link[=laplace_MixGpd]{laplace_MixGpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=laplace_lowercase]{laplace_lowercase()}}.

Other laplace kernel families: 
\code{\link{laplace_MixGpd}},
\code{\link{laplace_mix}}
}
\concept{laplace kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\name{sim_survival_tail}
\alias{sim_survival_tail}
\title{Simulate censored survival-style tail data}
\usage{
sim_survival_tail(n = 250, seed = NULL)
}
\arguments{
\item{n}{Integer sample size.}

\item{seed}{Optional random seed.}
}
\value{
Data frame containing observed time \code{time}, event indicator \code{status}, and covariates.
}
\description{
Generate event times, censoring times, an event indicator, and covariates for examples where
right tail behavior and positive support matter.
}
\details{
Event times are sampled from an exponential model with covariate-dependent mean, then censored
by an independent uniform censoring time. The observed time is
\deqn{
\tilde{T} = \min(T, C), \qquad \Delta = I(T \le C).
}

This helper is mainly for experimentation and stress-testing positive-support kernels; it does
not implement a dedicated survival model from the package API.
}
\seealso{
\code{\link[=sim_bulk_tail]{sim_bulk_tail()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other simulation helpers: 
\code{\link{sim_bulk_tail}()},
\code{\link{sim_causal_qte}()}
}
\concept{simulation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{run_mcmc_bundle_manual}
\alias{run_mcmc_bundle_manual}
\title{Run posterior sampling for a prepared one-arm bundle}
\usage{
run_mcmc_bundle_manual(
  bundle,
  show_progress = TRUE,
  quiet = FALSE,
  parallel_chains = FALSE,
  workers = NULL,
  timing = FALSE,
  z_update_every = NULL
)
}
\arguments{
\item{bundle}{A \code{causalmixgpd_bundle} from \code{build_nimble_bundle()}.}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}

\item{quiet}{Logical; if TRUE, suppress console status messages.
Set to FALSE to see progress messages during MCMC setup and execution.}

\item{parallel_chains}{Logical; run chains concurrently when \code{nchains > 1}.}

\item{workers}{Optional integer number of workers for parallel execution.}

\item{timing}{Logical; if TRUE, include stage timings (\code{build}, \code{compile}, \code{mcmc})
in \code{fit$timing}.}

\item{z_update_every}{Integer >= 1 controlling latent cluster-label update cadence.}
}
\value{
A fitted object of class \code{"mixgpd_fit"} containing posterior
draws, model metadata, and cached objects used by downstream S3 methods.
}
\description{
\code{run_mcmc_bundle_manual()} is the explicit runner for objects created by
\code{\link{build_nimble_bundle}}. It compiles the stored NIMBLE code,
executes MCMC, and returns a \code{"mixgpd_fit"} object.
}
\details{
The resulting fit supports posterior summaries of the model parameters as
well as posterior predictive functionals such as
\eqn{f(y \mid x)}, \eqn{S(y \mid x)},
\eqn{Q(\tau \mid x)}, and restricted means.

If \code{parallel_chains = TRUE}, chains are run concurrently when the stored
MCMC configuration uses more than one chain. If the bundle was built with
latent cluster labels monitored, the \code{z_update_every} argument controls how
frequently those latent indicators are refreshed during sampling.
}
\examples{
\donttest{
library(nimble)
y <- abs(rnorm(25)) + 0.1
bundle <- build_nimble_bundle(
  y = y,
  backend = "sb",
  kernel = "normal",
  GPD = FALSE,
  components = 3,
  mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
)
fit <- run_mcmc_bundle_manual(bundle, show_progress = FALSE)
fit
}
}
\seealso{
\code{\link{build_nimble_bundle}}, \code{\link{mcmc}},
\code{\link{summary.mixgpd_fit}}, \code{\link{predict.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invgauss.R
\name{InvGauss_mixgpd}
\alias{InvGauss_mixgpd}
\alias{dInvGaussMixGpd}
\alias{pInvGaussMixGpd}
\alias{rInvGaussMixGpd}
\alias{qInvGaussMixGpd}
\title{Inverse Gaussian mixture with a GPD tail}
\usage{
dInvGaussMixGpd(x, w, mean, shape, threshold, tail_scale, tail_shape, log = 0)

pInvGaussMixGpd(
  q,
  w,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rInvGaussMixGpd(n, w, mean, shape, threshold, tail_scale, tail_shape)

qInvGaussMixGpd(
  p,
  w,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}.}

\item{mean, shape}{Numeric vectors of length \eqn{K} giving component means and shapes.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars.
\code{qInvGaussMixGpd()} returns a numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to an inverse Gaussian
mixture bulk.
}
\details{
This family keeps the inverse-Gaussian mixture body below the threshold \eqn{u} and attaches a
generalized Pareto exceedance law to the residual survival probability above \eqn{u}. If
\eqn{F_{mix}(u)=p_u}, then the tail density is \eqn{(1-p_u)g_{GPD}(x \mid u,\sigma_u,\xi)}.

Quantile evaluation is piecewise. For probabilities at or below \eqn{p_u}, the function solves
the mixture inverse numerically; above \eqn{p_u}, it rescales the upper-tail probability and
applies the GPD inverse directly.
}
\section{Functions}{
\itemize{
\item \code{dInvGaussMixGpd()}: Inverse Gaussian mixture + GPD tail density

\item \code{pInvGaussMixGpd()}: Inverse Gaussian mixture + GPD tail distribution function

\item \code{rInvGaussMixGpd()}: Inverse Gaussian mixture + GPD tail random generation

\item \code{qInvGaussMixGpd()}: Inverse Gaussian mixture + GPD tail quantile function

}}
\examples{
w <- c(0.55, 0.30, 0.15)
mean <- c(1.0, 2.5, 5.0)
shape <- c(2, 4, 8)
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dInvGaussMixGpd(4.0, w = w, mean = mean, shape = shape,
               threshold = threshold, tail_scale = tail_scale,
               tail_shape = tail_shape, log = 0)
pInvGaussMixGpd(4.0, w = w, mean = mean, shape = shape,
               threshold = threshold, tail_scale = tail_scale,
               tail_shape = tail_shape, lower.tail = 1, log.p = 0)
qInvGaussMixGpd(0.50, w = w, mean = mean, shape = shape,
               threshold = threshold, tail_scale = tail_scale,
               tail_shape = tail_shape)
qInvGaussMixGpd(0.95, w = w, mean = mean, shape = shape,
               threshold = threshold, tail_scale = tail_scale,
               tail_shape = tail_shape)
replicate(10, rInvGaussMixGpd(1, w = w, mean = mean, shape = shape,
                             threshold = threshold,
                             tail_scale = tail_scale,
                             tail_shape = tail_shape))
}
\seealso{
\code{\link[=InvGauss_mix]{InvGauss_mix()}}, \code{\link[=InvGauss_gpd]{InvGauss_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=invgauss_lowercase]{invgauss_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other inverse-gaussian kernel families: 
\code{\link{InvGauss_gpd}},
\code{\link{InvGauss_mix}}
}
\concept{inverse-gaussian kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{dpmix.causal}
\alias{dpmix.causal}
\title{Fit a causal two-arm Dirichlet process mixture without a GPD tail}
\usage{
dpmix.causal(
  y = NULL,
  X = NULL,
  treat = NULL,
  data = NULL,
  mcmc = list(),
  formula = NULL,
  ...
)
}
\arguments{
\item{y}{Either a response vector or a causal bundle object.}

\item{X}{Optional design matrix/data.frame.}

\item{treat}{Binary treatment indicator.}

\item{data}{Optional data.frame used with \code{formula}.}

\item{mcmc}{Named list of run arguments passed to \code{mcmc()} (including
optional performance controls such as \code{parallel_arms},
\code{workers}, \code{timing}, and \code{z_update_every}).}

\item{formula}{Optional formula.}

\item{...}{Additional build arguments passed to \code{\link{build_causal_bundle}}.}
}
\value{
A fitted object of class \code{"causalmixgpd_causal_fit"}.
}
\description{
\code{dpmix.causal()} fits a causal model with separate treated and control
outcome mixtures and, when requested, a propensity score block. It is the
bulk-only companion to \code{\link{dpmgpd.causal}}.
}
\details{
The resulting fit supports conditional outcome prediction
\eqn{F_a(y \mid x)} for \eqn{a \in \{0,1\}}, followed by causal
functionals such as \code{\link{ate}}, \code{\link{qte}},
\code{\link{cate}}, and \code{\link{cqte}}.
}
\examples{
\donttest{
N <- 30
X <- data.frame(x1 = stats::rnorm(N), x2 = stats::runif(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N) + A + 0.5 * X$x1) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)

fit <- dpmix.causal(
  y = y, X = X, treat = A,
  backend = "sb", kernel = "normal",
  components = 3, mcmc = mcmc_small
)
summary(fit)
ate(fit, show_progress = FALSE)
}
}
\seealso{
\code{\link{build_causal_bundle}},
\code{\link{bundle}}, \code{\link{dpmgpd.causal}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.format_fit_header}
\alias{.format_fit_header}
\title{Format a short header for printing}
\usage{
.format_fit_header(x)
}
\arguments{
\item{x}{A mixgpd_fit.}
}
\value{
Character vector lines.
}
\description{
Format a short header for printing
}
\details{
This helper builds the short header used by \code{print.mixgpd_fit()}. It extracts
model identity, training size, truncation size, epsilon, and stored MCMC
settings into a compact character vector so higher-level print methods do not
duplicate formatting logic.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invgauss.R
\name{InvGauss_mix}
\alias{InvGauss_mix}
\alias{dInvGaussMix}
\alias{pInvGaussMix}
\alias{rInvGaussMix}
\alias{qInvGaussMix}
\title{Inverse Gaussian mixture distribution}
\usage{
dInvGaussMix(x, w, mean, shape, log = 0)

pInvGaussMix(q, w, mean, shape, lower.tail = 1, log.p = 0)

rInvGaussMix(n, w, mean, shape)

qInvGaussMix(
  p,
  w,
  mean,
  shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{mean, shape}{Numeric vectors of length \eqn{K} giving component means and shapes.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qInvGaussMix()} returns a numeric
vector with the same length as \code{p}.
}
\description{
Finite mixture of inverse Gaussian components for positive-support bulk modeling. Each component
is parameterized by \code{mean[j]} and \code{shape[j]}.
}
\details{
The scalar functions in this topic are the compiled building blocks for inverse-Gaussian bulk
kernels. For vectorized R usage, use \code{\link[=invgauss_lowercase]{invgauss_lowercase()}}.

The mixture distribution is
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k F_{IG}(x \mid \mu_k,\lambda_k),
}
where each inverse Gaussian component has mean \eqn{\mu_k} and variance \eqn{\mu_k^3/\lambda_k}.
Random generation selects a component using the normalized weights and then generates from the
corresponding inverse Gaussian law. Quantiles are computed numerically because the finite-mixture
inverse CDF is not available in closed form.

The analytical mixture mean is
\deqn{
E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k.
}
That expression is used by the package whenever inverse-Gaussian mixtures contribute to
posterior predictive means.
}
\section{Functions}{
\itemize{
\item \code{dInvGaussMix()}: Inverse Gaussian mixture density

\item \code{pInvGaussMix()}: Inverse Gaussian mixture distribution function

\item \code{rInvGaussMix()}: Inverse Gaussian mixture random generation

\item \code{qInvGaussMix()}: Inverse Gaussian mixture quantile function

}}
\examples{
w <- c(0.55, 0.30, 0.15)
mean <- c(1.0, 2.5, 5.0)
shape <- c(2, 4, 8)

dInvGaussMix(2.0, w = w, mean = mean, shape = shape, log = 0)
pInvGaussMix(2.0, w = w, mean = mean, shape = shape,
            lower.tail = 1, log.p = 0)
qInvGaussMix(0.50, w = w, mean = mean, shape = shape)
qInvGaussMix(0.95, w = w, mean = mean, shape = shape)
replicate(10, rInvGaussMix(1, w = w, mean = mean, shape = shape))
}
\seealso{
\code{\link[=InvGauss_mixgpd]{InvGauss_mixgpd()}}, \code{\link[=InvGauss_gpd]{InvGauss_gpd()}}, \code{\link[=invgauss_lowercase]{invgauss_lowercase()}},
\code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other inverse-gaussian kernel families: 
\code{\link{InvGauss_gpd}},
\code{\link{InvGauss_mixgpd}}
}
\concept{inverse-gaussian kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{dpmix.cluster}
\alias{dpmix.cluster}
\title{Fit a clustering-only bulk model}
\usage{
dpmix.cluster(
  formula,
  data,
  type = c("weights", "param", "both"),
  default = "weights",
  mcmc = list(),
  ...
)
}
\arguments{
\item{formula}{Model formula. The response must be present in \code{data}.}

\item{data}{Data frame containing the response and optional predictors.}

\item{type}{Clustering mode:
\itemize{
\item \code{"weights"}: links mixture weights to predictors
\item \code{"param"}: links kernel parameters to predictors
\item \code{"both"}: links both weights and kernel parameters to predictors
}}

\item{default}{Default mode used when \code{type} is omitted.}

\item{mcmc}{MCMC control list passed into the cluster bundle.}

\item{...}{Additional arguments passed to \code{build_cluster_bundle()}, including kernel settings,
prior overrides, component counts, and monitoring controls.}
}
\value{
Object of class \code{dpmixgpd_cluster_fit}.
}
\description{
Build and fit a Dirichlet-process mixture for clustering without causal estimands or posterior
prediction for a response surface. This interface focuses on latent partition recovery from a
formula specification and returns a cluster-fit object that can be summarized, plotted, or
converted into labels and posterior similarity matrices with \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}.
}
\details{
The fitted model targets a latent partition \eqn{z_1, \dots, z_n} with component-specific kernel
parameters. Depending on \code{type}, predictors can enter through the gating probabilities
\deqn{
\Pr(z_i = k \mid x_i) = \pi_k(x_i)
}
or through linked kernel parameters for each component. The returned fit stores posterior draws
of the latent cluster labels and associated parameters; the representative clustering is extracted
later by \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}} using Dahl's least-squares rule.

Use \code{type = "weights"} or \code{type = "both"} only when the formula includes predictors and when an
explicit number of \code{components} is supplied. Otherwise the builder stops before fitting.
}
\examples{
\donttest{
data("nc_realX100_p3_k2", package = "CausalMixGPD")
dat <- data.frame(y = nc_realX100_p3_k2$y[1:20],
                  nc_realX100_p3_k2$X[1:20, , drop = FALSE])
fit <- dpmix.cluster(
  y ~ x1 + x2 + x3,
  data = dat,
  kernel = "normal",
  type = "param",
  components = 3,
  mcmc = list(niter = 60, nburnin = 30, thin = 1, nchains = 1, seed = 1)
)
summary(fit)
}
}
\seealso{
\code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}, \code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}},
\code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}}, \code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}},
\code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=dpmix]{dpmix()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{run_mcmc_causal}
\alias{run_mcmc_causal}
\title{Run posterior sampling for a causal bundle}
\usage{
run_mcmc_causal(
  bundle,
  show_progress = TRUE,
  quiet = FALSE,
  parallel_arms = FALSE,
  workers = NULL,
  timing = FALSE,
  z_update_every = NULL
)
}
\arguments{
\item{bundle}{A \code{"causalmixgpd_causal_bundle"} from \code{build_causal_bundle()}.}

\item{show_progress}{Logical; if TRUE, print step messages and render progress where supported.}

\item{quiet}{Logical; if TRUE, suppress step messages and progress display.}

\item{parallel_arms}{Logical; if TRUE, run control and treated outcome arms in parallel.}

\item{workers}{Optional integer workers for parallel arm execution.}

\item{timing}{Logical; if TRUE, return arm and total timings in \code{$timing}.}

\item{z_update_every}{Integer >= 1 passed to arm-level outcome MCMC.}
}
\value{
A list of class \code{"causalmixgpd_causal_fit"} containing the
fitted treated/control outcome models, optional PS fit, the original
bundle, and timing metadata when requested.
}
\description{
\code{run_mcmc_causal()} executes the PS block (when enabled) and the two
arm-specific outcome models prepared by \code{\link{build_causal_bundle}},
then returns a single \code{"causalmixgpd_causal_fit"} object.
}
\details{
The fitted object contains the posterior draws needed to evaluate arm-level
predictive distributions \eqn{F_1(y \mid x)} and
\eqn{F_0(y \mid x)}, followed by marginal or conditional causal
contrasts. When \code{PS = FALSE} in the bundle, the PS block is skipped and
outcome prediction uses only the original covariates.
}
\examples{
\donttest{
N <- 25
X <- data.frame(x1 = stats::rnorm(N))
A <- stats::rbinom(N, 1, 0.5)
y <- abs(stats::rnorm(N)) + 0.1
mcmc_small <- list(niter = 100, nburnin = 50, thin = 1, nchains = 1, seed = 1)
cb <- build_causal_bundle(y = y, X = X, A = A, backend = "sb", kernel = "normal",
                         mcmc_outcome = mcmc_small, mcmc_ps = mcmc_small)
fit <- run_mcmc_causal(cb)
}
}
\seealso{
\code{\link{build_causal_bundle}}, \code{\link{mcmc}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_causal_fit}
\alias{print.causalmixgpd_causal_fit}
\title{Print a fitted causal model}
\usage{
\method{print}{causalmixgpd_causal_fit}(x, ...)
}
\arguments{
\item{x}{A \code{"causalmixgpd_causal_fit"} object.}

\item{...}{Unused.}
}
\value{
The input object (invisibly).
}
\description{
\code{print.causalmixgpd_causal_fit()} provides a compact overview of the
fitted treated/control outcome blocks and the PS component when present.
}
\details{
A fitted causal object combines posterior draws for the treated outcome model,
the control outcome model, and optionally the propensity-score model. Those
fitted blocks are the ingredients used later to evaluate causal estimands such
as \eqn{\mu_1(x) - \mu_0(x)} or
\eqn{Q_{Y^1}(\tau \mid x) - Q_{Y^0}(\tau \mid x)}.

The print method is deliberately high level. It identifies which models were
fitted and whether GPD tails are active, but it does not report posterior
summaries or treatment-effect estimates. Use \code{summary()}, \code{predict()}, or the
dedicated causal estimand helpers for inferential output.
}
\seealso{
\code{\link{summary.causalmixgpd_causal_fit}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\docType{data}
\name{nc_pos_tail200_k4}
\alias{nc_pos_tail200_k4}
\title{Positive-Support Tail Data Set With Four Components}
\format{
A list with:
\describe{
\item{y}{Numeric outcome vector.}
\item{X}{NULL.}
\item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
\item{truth}{List with kernel, weights, params, threshold, tail_params.}
}
}
\usage{
nc_pos_tail200_k4
}
\description{
Positive-support, tail-designed mixture dataset with K=4 components and no covariates.
Intended for GPD vignettes (gamma/lognormal/invgauss/amoroso with GPD=TRUE).
}
\examples{
head(nc_pos_tail200_k4$y)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{params}
\alias{params}
\title{Extract posterior mean parameters in natural shape}
\usage{
params(object, ...)
}
\arguments{
\item{object}{A fitted object of class \code{"mixgpd_fit"}.}

\item{...}{Unused.}
}
\value{
An object of class \code{"mixgpd_params"} (a named list). For
causal fits, \code{params()} returns a treated/control pair and includes
a \code{ps} block when a propensity-score model was fitted.
}
\description{
\code{params()} reshapes posterior mean summaries back into the parameter
layout implied by the fitted model specification.
}
\details{
This extractor is intended for structural inspection of the fitted model.
Scalar quantities remain scalar, component-specific parameters are returned as
vectors, and linked regression blocks are returned as matrices with covariate
names as columns when available. If propensity-score adjustment is active for
a linked bulk parameter, its coefficient is folded into the returned beta
matrix as a leading \code{"PropScore"} column.

For a spliced model, the extractor returns posterior means of the bulk
mixture parameters together with component-level threshold, tail-scale, and
tail-shape terms. When tail terms are link-mode, the corresponding
component-by-covariate beta blocks are returned.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
params(fit)
p <- params(fit)
}
}
\seealso{
\code{\link{summary.mixgpd_fit}}, \code{\link{predict.mixgpd_fit}},
\code{\link{ess_summary}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.mixgpd_fit}
\alias{summary.mixgpd_fit}
\title{Summarize posterior draws from a one-arm fitted model}
\usage{
\method{summary}{mixgpd_fit}(object, pars = NULL, probs = c(0.025, 0.5, 0.975), ...)
}
\arguments{
\item{object}{A fitted object of class \code{"mixgpd_fit"}.}

\item{pars}{Optional character vector of parameters to summarize. If NULL, summarize all (excluding v's).}

\item{probs}{Numeric vector of quantiles to report.}

\item{...}{Unused.}
}
\value{
An object of class \code{"mixgpd_summary"}.
}
\description{
\code{summary.mixgpd_fit()} computes posterior summaries for monitored model
parameters.
}
\details{
The returned table is a parameter-level summary of the posterior draws, not a
predictive summary. Use \code{\link{predict.mixgpd_fit}} for posterior
predictive quantities such as densities, survival probabilities, quantiles,
and means.

The summary respects the stored truncation metadata and reports WAIC if it
was requested during MCMC.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
summary(fit, pars = c("alpha", "threshold"))
}
}
\seealso{
\code{\link{print.mixgpd_fit}}, \code{\link{params}},
\code{\link{predict.mixgpd_fit}}, \code{\link{ess_summary}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.validate_fit}
\alias{.validate_fit}
\title{Validate a fitted object}
\usage{
.validate_fit(object)
}
\arguments{
\item{object}{A fitted object.}
}
\value{
Invisibly TRUE, otherwise errors.
}
\description{
Validate a fitted object
}
\details{
This is a lightweight structural check used by multiple internal helpers. It
verifies that the object inherits from \code{mixgpd_fit} and that posterior draws
are available in one of the expected storage locations before later code tries
to summarize or predict from the fit.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{summary.dpmixgpd_cluster_bundle}
\alias{summary.dpmixgpd_cluster_bundle}
\title{Summarize a cluster bundle}
\usage{
\method{summary}{dpmixgpd_cluster_bundle}(object, ...)
}
\arguments{
\item{object}{A cluster bundle.}

\item{...}{Unused.}
}
\value{
Summary list containing kernel choice, GPD flag, dimensions, component count, and
monitor configuration.
}
\description{
Report the modeling choices encoded in a cluster bundle before fitting.
}
\details{
This summary is a pre-flight check for the clustering workflow. It reports
the latent-partition design, the chosen kernel family, whether a GPD tail
will be spliced above the threshold, the effective sample and predictor
dimensions, and the monitor set that will be carried into MCMC.

Because no posterior simulation has occurred yet, the summary describes only
the assumed model structure. Quantities such as representative labels,
pairwise co-clustering probabilities, and cluster-specific summaries become
available only after the fitted object has been created and post-processed.
}
\seealso{
\code{\link[=print.dpmixgpd_cluster_bundle]{print.dpmixgpd_cluster_bundle()}}, \code{\link[=plot.dpmixgpd_cluster_bundle]{plot.dpmixgpd_cluster_bundle()}},
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{.onLoad}
\alias{.onLoad}
\title{Package hooks}
\usage{
.onLoad(libname, pkgname)
}
\description{
Internal package initialization.
}
\details{
The load hook initializes package-wide defaults that should exist as soon as
the namespace is attached. In particular, it ensures that the kernel and tail
registries are ready for later model-building code and sets the package plotly
option if the user has not already chosen one.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.causalmixgpd_qte}
\alias{print.summary.causalmixgpd_qte}
\title{Print a QTE summary}
\usage{
\method{print}{summary.causalmixgpd_qte}(x, digits = 3, ...)
}
\arguments{
\item{x}{A \code{"summary.causalmixgpd_qte"} object.}

\item{digits}{Number of digits to display.}

\item{...}{Unused.}
}
\value{
The object \code{x}, invisibly.
}
\description{
Print a QTE summary
}
\details{
This formatter displays the summary object returned by
\code{summary.causalmixgpd_qte()}. It reports the quantile grid, interval
configuration, model metadata when available, and the tabulated quantile
effect summaries.

No additional causal computations are performed here. The method simply turns
the stored summary tables into a readable report.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\name{sim_causal_qte}
\alias{sim_causal_qte}
\title{Simulate causal quantile-treatment-effect data}
\usage{
sim_causal_qte(n = 300, seed = NULL)
}
\arguments{
\item{n}{Integer sample size.}

\item{seed}{Optional random seed.}
}
\value{
List with components \code{y}, \code{t}, and \code{X}; \code{A} is included as a backward-compatible alias
for \code{t}.
}
\description{
Generate a treatment indicator, covariates, and a continuous outcome with both location and
tail heterogeneity. The resulting structure is intended for examples involving
\code{\link[=dpmix.causal]{dpmix.causal()}}, \code{\link[=dpmgpd.causal]{dpmgpd.causal()}}, \code{\link[=qte]{qte()}}, and \code{\link[=cqte]{cqte()}}.
}
\details{
Treatment assignment is generated from a logistic propensity score
\deqn{
\Pr(T = 1 \mid X) = \operatorname{logit}^{-1}(\eta(X)),
}
and the observed outcome combines baseline covariate effects, an average treatment shift, and
a covariate-dependent tail amplification for treated units. This produces data where marginal
and conditional quantile effects differ across the outcome distribution.

The returned list can be converted directly into the arguments expected by the causal fitting
wrappers after minor formatting.
}
\seealso{
\code{\link[=sim_bulk_tail]{sim_bulk_tail()}}, \code{\link[=dpmgpd.causal]{dpmgpd.causal()}}, \code{\link[=qte]{qte()}}, \code{\link[=cqte]{cqte()}}.

Other simulation helpers: 
\code{\link{sim_bulk_tail}()},
\code{\link{sim_survival_tail}()}
}
\concept{simulation helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_constants_from_spec}
\alias{build_constants_from_spec}
\title{Build constants list from a compiled model spec}
\usage{
build_constants_from_spec(spec)
}
\arguments{
\item{spec}{A compiled model specification produced by \code{compile_model_spec()}.}
}
\value{
Named list of constants.
}
\description{
Produces a named list of constants to pass into \code{nimbleModel}.
Constants include core sizes (\code{N}, \code{P}, \code{components}) and
hyperparameters for priors implied by \code{spec$plan}.
}
\details{
This function is pre-run only; it does not compile or execute NIMBLE.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.summary.causalmixgpd_ps_fit}
\alias{print.summary.causalmixgpd_ps_fit}
\title{Print a propensity-score summary object}
\usage{
\method{print}{summary.causalmixgpd_ps_fit}(x, digits = 3, max_rows = 60, show_ess = FALSE, ...)
}
\arguments{
\item{x}{A \code{"summary.causalmixgpd_ps_fit"} object.}

\item{digits}{Number of digits to print in summary tables.}

\item{max_rows}{Maximum rows to print from the summary table.}

\item{show_ess}{Logical; if \code{TRUE}, include the \code{ess} column when present.}

\item{...}{Unused.}
}
\value{
\code{x} invisibly.
}
\description{
Print a propensity-score summary object
}
\details{
This is a display method for the object returned by
\code{summary.causalmixgpd_ps_fit()}. It prints the PS model identity, the
effective data dimension used by that model, and the posterior summary table
for the monitored parameters.

The method does not recompute propensity scores or refit the model. It is a
formatting layer over already computed posterior summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/causal.R
\name{build_causal_bundle}
\alias{build_causal_bundle}
\title{Build a causal bundle (design + two outcome arms)}
\usage{
build_causal_bundle(
  y,
  X,
  A,
  backend = c("sb", "crp", "spliced"),
  kernel,
  GPD = FALSE,
  components = 10L,
  param_specs = NULL,
  mcmc_outcome = list(niter = 2000, nburnin = 500, thin = 1, nchains = 1, seed = 1),
  mcmc_ps = list(niter = 1000, nburnin = 250, thin = 1, nchains = 1, seed = 1),
  epsilon = 0.025,
  alpha_random = TRUE,
  ps_prior = list(mean = 0, sd = 2),
  include_intercept = TRUE,
  PS = "logit",
  ps_scale = c("logit", "prob"),
  ps_summary = c("mean", "median"),
  ps_clamp = 1e-06,
  monitor = c("core", "full"),
  monitor_latent = FALSE,
  monitor_v = FALSE
)
}
\arguments{
\item{y}{Numeric outcome vector.}

\item{X}{Design matrix or data.frame of covariates (N x P).}

\item{A}{Binary treatment indicator (length N, values 0/1).}

\item{backend}{Character; the Dirichlet process representation for outcome models:
\itemize{
\item \code{"sb"}: stick-breaking truncation
\item \code{"crp"}: Chinese Restaurant Process
\item \code{"spliced"}: CRP with GPD tail splicing
}
If length 2, the first entry is used for treated (\code{A=1}) and the
second for control (\code{A=0}).}

\item{kernel}{Character kernel name for outcome models (must exist in
\code{get_kernel_registry()}). If length 2:
\itemize{
\item first entry: used for treated (\code{A=1})
\item second entry: used for control (\code{A=0})
}}

\item{GPD}{Logical; include GPD tail for outcomes if TRUE. If length 2:
\itemize{
\item first entry: used for treated (\code{A=1})
\item second entry: used for control (\code{A=0})
}}

\item{components}{Integer >= 2; truncation parameter for outcome mixtures. If length 2:
\itemize{
\item first entry: used for treated (\code{A=1})
\item second entry: used for control (\code{A=0})
}}

\item{param_specs}{Outcome parameter overrides (same structure as
\code{build_nimble_bundle()}):
\itemize{
\item a single list: used for both arms
\item a list with \code{con} and \code{trt} entries: arm-specific overrides
}}

\item{mcmc_outcome}{MCMC settings list for the outcome bundles.}

\item{mcmc_ps}{MCMC settings list for the PS model.}

\item{epsilon}{Numeric in [0,1) used by outcome bundles for posterior truncation
summaries. If length 2:
\itemize{
\item first entry: used for treated (\code{A=1})
\item second entry: used for control (\code{A=0})
}}

\item{alpha_random}{Logical; whether the outcome-model DP concentration parameter \eqn{\kappa} is stochastic.}

\item{ps_prior}{Normal prior for PS coefficients. List with \code{mean} and \code{sd}.}

\item{include_intercept}{Logical; if TRUE, an intercept column is prepended to \code{X}
in the PS model.}

\item{PS}{Character or logical; controls propensity score estimation:
\itemize{
\item \code{"logit"} (default): Logistic regression PS model
\item \code{"probit"}: Probit regression PS model
\item \code{"naive"}: Gaussian naive Bayes PS model
\item \code{FALSE}: No PS estimation; outcome models use only \code{X}
}
The PS model choice is stored in bundle metadata for downstream use in
prediction and summaries.}

\item{ps_scale}{Scale used when augmenting outcomes with PS:
\itemize{
\item \code{"logit"}: augment on the logit (log-odds) scale
\item \code{"prob"}: augment on the probability scale
}}

\item{ps_summary}{Posterior summary for PS:
\itemize{
\item \code{"mean"}: posterior mean of propensity scores
\item \code{"median"}: posterior median of propensity scores
}}

\item{ps_clamp}{Numeric epsilon for clamping PS values to \eqn{(\epsilon, 1-\epsilon)}.}

\item{monitor}{Character monitor profile:
\itemize{
\item \code{"core"} (default): monitors only the essential model parameters
\item \code{"full"}: monitors all model nodes
}}

\item{monitor_latent}{Logical; whether to monitor latent cluster labels (\code{z}) in outcome arms.}

\item{monitor_v}{Logical; whether to monitor stick-breaking \code{v} terms for SB outcomes.}
}
\value{
A list of class \code{"causalmixgpd_causal_bundle"} containing the
design bundle, two outcome bundles, training data, arm indices, and
metadata required for posterior prediction and causal effect summaries.
}
\description{
\code{build_causal_bundle()} is the detailed constructor behind
\code{\link{bundle}} for causal analyses. It prepares:
\itemize{
\item a propensity score (PS) design block for \eqn{A \mid X},
\item a control-arm outcome bundle for \eqn{Y(0)},
\item a treated-arm outcome bundle for \eqn{Y(1)}.
}
}
\details{
The outcome bundles reuse the one-arm DPM plus optional GPD machinery. The
PS block provides a shared adjustment object used by
\code{\link{run_mcmc_causal}} and
\code{\link{predict.causalmixgpd_causal_fit}}.

The causal bundle encodes the two arm-specific predictive laws
\eqn{F_0(y \mid x)} and \eqn{F_1(y \mid x)}. Downstream causal estimands are
functionals of these two distributions:
\deqn{\mathrm{ATE} = E\{Y(1)\} - E\{Y(0)\}, \qquad
\mathrm{QTE}(\tau) = Q_1(\tau) - Q_0(\tau).}

When \code{PS} is enabled, the package estimates a propensity score model
\eqn{e(x) = \Pr(A = 1 \mid X = x)} and uses a posterior summary of that score
as an augmented covariate in the arm-specific outcome models. This mirrors
the workflow described in the manuscript vignette.
}
\examples{
\donttest{
set.seed(1)
N <- 25
X <- cbind(x1 = rnorm(N), x2 = runif(N))
A <- rbinom(N, 1, plogis(0.3 + 0.5 * X[, 1]))
y <- rexp(N) + 0.1

cb <- build_causal_bundle(
  y = y,
  X = X,
  A = A,
  backend = "sb",
  kernel = "gamma",
  GPD = TRUE,
  components = 3,
  PS = "probit"
)
}
}
\seealso{
\code{\link{bundle}}, \code{\link{run_mcmc_causal}},
\code{\link{predict.causalmixgpd_causal_fit}}, \code{\link{ate}},
\code{\link{qte}}, \code{\link{cate}}, \code{\link{cqte}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invgauss.R
\name{invgauss_lowercase}
\alias{invgauss_lowercase}
\alias{dinvgaussmix}
\alias{pinvgaussmix}
\alias{qinvgaussmix}
\alias{rinvgaussmix}
\alias{dinvgaussmixgpd}
\alias{pinvgaussmixgpd}
\alias{qinvgaussmixgpd}
\alias{rinvgaussmixgpd}
\alias{dinvgaussgpd}
\alias{pinvgaussgpd}
\alias{qinvgaussgpd}
\alias{rinvgaussgpd}
\title{Lowercase vectorized inverse Gaussian distribution functions}
\usage{
dinvgaussmix(x, w, mean, shape, log = FALSE)

pinvgaussmix(q, w, mean, shape, lower.tail = TRUE, log.p = FALSE)

qinvgaussmix(
  p,
  w,
  mean,
  shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rinvgaussmix(n, w, mean, shape)

dinvgaussmixgpd(
  x,
  w,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

pinvgaussmixgpd(
  q,
  w,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qinvgaussmixgpd(
  p,
  w,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rinvgaussmixgpd(n, w, mean, shape, threshold, tail_scale, tail_shape)

dinvgaussgpd(x, mean, shape, threshold, tail_scale, tail_shape, log = FALSE)

pinvgaussgpd(
  q,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qinvgaussgpd(
  p,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rinvgaussgpd(n, mean, shape, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{mean, shape}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar inverse-Gaussian-kernel topics in this file.
}
\details{
These functions are vectorized R front ends to the scalar inverse-Gaussian
and splice routines. They retain the \eqn{(\mu,\lambda)} parameterization
used everywhere else in the package and apply the scalar evaluator
repeatedly over the supplied input vector or draw index.
}
\section{Functions}{
\itemize{
\item \code{dinvgaussmix()}: Inverse Gaussian mixture density (vectorized)

\item \code{pinvgaussmix()}: Inverse Gaussian mixture distribution function (vectorized)

\item \code{qinvgaussmix()}: Inverse Gaussian mixture quantile function (vectorized)

\item \code{rinvgaussmix()}: Inverse Gaussian mixture random generation (vectorized)

\item \code{dinvgaussmixgpd()}: Inverse Gaussian mixture + GPD density (vectorized)

\item \code{pinvgaussmixgpd()}: Inverse Gaussian mixture + GPD distribution function (vectorized)

\item \code{qinvgaussmixgpd()}: Inverse Gaussian mixture + GPD quantile function (vectorized)

\item \code{rinvgaussmixgpd()}: Inverse Gaussian mixture + GPD random generation (vectorized)

\item \code{dinvgaussgpd()}: Inverse Gaussian + GPD density (vectorized)

\item \code{pinvgaussgpd()}: Inverse Gaussian + GPD distribution function (vectorized)

\item \code{qinvgaussgpd()}: Inverse Gaussian + GPD quantile function (vectorized)

\item \code{rinvgaussgpd()}: Inverse Gaussian + GPD random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.3, 0.1)
mu <- c(1, 1.5, 2)
lam <- c(2, 3, 4)

# Inverse Gaussian mixture
dinvgaussmix(c(1, 2, 3), w = w, mean = mu, shape = lam)
rinvgaussmix(5, w = w, mean = mu, shape = lam)

}
\seealso{
\code{\link[=InvGauss_mix]{InvGauss_mix()}}, \code{\link[=InvGauss_mixgpd]{InvGauss_mixgpd()}}, \code{\link[=InvGauss_gpd]{InvGauss_gpd()}}, \code{\link[=bundle]{bundle()}},
\code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invgauss.R
\name{InvGauss_gpd}
\alias{InvGauss_gpd}
\alias{dInvGaussGpd}
\alias{pInvGaussGpd}
\alias{rInvGaussGpd}
\alias{qInvGaussGpd}
\title{Inverse Gaussian with a GPD tail}
\usage{
dInvGaussGpd(x, mean, shape, threshold, tail_scale, tail_shape, log = 0)

pInvGaussGpd(
  q,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rInvGaussGpd(n, mean, shape, threshold, tail_scale, tail_shape)

qInvGaussGpd(
  p,
  mean,
  shape,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{mean}{Numeric scalar mean parameter \eqn{\mu>0}.}

\item{shape}{Numeric scalar shape parameter \eqn{\lambda>0}.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric tolerance for numerical inversion in \code{qInvGaussGpd}.}

\item{maxiter}{Maximum iterations for numerical inversion in \code{qInvGaussGpd}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars.
\code{qInvGaussGpd()} returns a numeric vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
inverse Gaussian bulk.
}
\details{
This is the one-component version of \code{\link[=InvGauss_mixgpd]{InvGauss_mixgpd()}}. The inverse Gaussian governs the bulk
region and the generalized Pareto governs exceedances over the threshold. The splice is
continuous at \eqn{u} because the GPD is scaled by the inverse-Gaussian survival probability at
the threshold.

The ordinary mean of the spliced law exists only when the GPD tail has \eqn{\xi < 1}. When that
condition fails, the package uses restricted means or quantile-based summaries instead of an
ordinary mean.
}
\section{Functions}{
\itemize{
\item \code{dInvGaussGpd()}: Inverse Gaussian + GPD tail density

\item \code{pInvGaussGpd()}: Inverse Gaussian + GPD tail distribution function

\item \code{rInvGaussGpd()}: Inverse Gaussian + GPD tail random generation

\item \code{qInvGaussGpd()}: Inverse Gaussian + GPD tail quantile function

}}
\examples{
mean <- 2.5
shape <- 6
threshold <- 3
tail_scale <- 0.9
tail_shape <- 0.2

dInvGaussGpd(4.0, mean = mean, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape, log = 0)
pInvGaussGpd(4.0, mean = mean, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape, lower.tail = 1, log.p = 0)
qInvGaussGpd(0.50, mean = mean, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape)
qInvGaussGpd(0.95, mean = mean, shape = shape,
            threshold = threshold, tail_scale = tail_scale,
            tail_shape = tail_shape)
replicate(10, rInvGaussGpd(1, mean = mean, shape = shape,
                          threshold = threshold,
                          tail_scale = tail_scale,
                          tail_shape = tail_shape))
}
\seealso{
\code{\link[=InvGauss_mix]{InvGauss_mix()}}, \code{\link[=InvGauss_mixgpd]{InvGauss_mixgpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=invgauss_lowercase]{invgauss_lowercase()}}.

Other inverse-gaussian kernel families: 
\code{\link{InvGauss_mix}},
\code{\link{InvGauss_mixgpd}}
}
\concept{inverse-gaussian kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lognormal.R
\name{lognormal_mix}
\alias{lognormal_mix}
\alias{dLognormalMix}
\alias{pLognormalMix}
\alias{rLognormalMix}
\alias{qLognormalMix}
\title{Lognormal mixture distribution}
\usage{
dLognormalMix(x, w, meanlog, sdlog, log = 0)

pLognormalMix(q, w, meanlog, sdlog, lower.tail = 1, log.p = 0)

rLognormalMix(n, w, meanlog, sdlog)

qLognormalMix(
  p,
  w,
  meanlog,
  sdlog,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{meanlog, sdlog}{Numeric vectors of length \eqn{K} giving component log-means and log-standard deviations.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qLognormalMix()} returns a numeric
vector with the same length as \code{p}.
}
\description{
Finite mixture of lognormal components for positive-support bulk modeling. The scalar functions
in this topic are the NIMBLE-compatible building blocks for the lognormal bulk kernel family.
}
\details{
The mixture density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k f_{LN}(x \mid \mu_k, \sigma_k),
\qquad x > 0,
}
with normalized weights \eqn{\tilde{w}_k}. For vectorized R usage, use \code{\link[=lognormal_lowercase]{lognormal_lowercase()}}.

Each component satisfies \eqn{\log X_k \sim N(\mu_k,\sigma_k^2)}, so the mixture CDF is
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k
\Phi\left(\frac{\log x-\mu_k}{\sigma_k}\right), \qquad x>0.
}
Random generation proceeds by drawing a component index with probability \eqn{\tilde{w}_k} and
then sampling from the corresponding lognormal law. Because a finite mixture of lognormals does
not admit a closed-form inverse CDF, \code{qLognormalMix()} computes quantiles by numerical
inversion.

The analytical mixture mean is
\deqn{
E(X) = \sum_{k=1}^K \tilde{w}_k \exp(\mu_k + \sigma_k^2/2),
}
which is the expression used by the package whenever an ordinary predictive mean exists.
}
\section{Functions}{
\itemize{
\item \code{dLognormalMix()}: Lognormal mixture density

\item \code{pLognormalMix()}: Lognormal mixture distribution function

\item \code{rLognormalMix()}: Lognormal mixture random generation

\item \code{qLognormalMix()}: Lognormal mixture quantile function

}}
\examples{
w <- c(0.60, 0.25, 0.15)
meanlog <- c(-0.2, 0.6, 1.2)
sdlog <- c(0.4, 0.3, 0.5)

dLognormalMix(2.0, w = w, meanlog = meanlog, sdlog = sdlog, log = FALSE)
pLognormalMix(2.0, w = w, meanlog = meanlog, sdlog = sdlog,
             lower.tail = TRUE, log.p = FALSE)
qLognormalMix(0.50, w = w, meanlog = meanlog, sdlog = sdlog)
qLognormalMix(0.95, w = w, meanlog = meanlog, sdlog = sdlog)
replicate(10, rLognormalMix(1, w = w, meanlog = meanlog, sdlog = sdlog))
}
\seealso{
\code{\link[=lognormal_mixgpd]{lognormal_mixgpd()}}, \code{\link[=lognormal_gpd]{lognormal_gpd()}}, \code{\link[=lognormal_lowercase]{lognormal_lowercase()}},
\code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other lognormal kernel families: 
\code{\link{lognormal_gpd}},
\code{\link{lognormal_mixgpd}}
}
\concept{lognormal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\docType{data}
\name{nc_posX100_p3_k2}
\alias{nc_posX100_p3_k2}
\title{Positive-Support Covariate Data Set With Three Predictors}
\format{
A list with:
\describe{
\item{y}{Numeric outcome vector.}
\item{X}{data.frame with x1-x3.}
\item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
\item{truth}{List with kernel, weights, params, threshold, tail_params.}
}
}
\usage{
nc_posX100_p3_k2
}
\description{
Positive-support dataset with covariates (p=3) and K=2 mixture components.
Intended for covariate and prediction vignettes (GPD=FALSE).
}
\examples{
head(nc_posX100_p3_k2$X)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{predict.dpmixgpd_cluster_fit}
\alias{predict.dpmixgpd_cluster_fit}
\title{Predict labels or similarity matrices from a cluster fit}
\usage{
\method{predict}{dpmixgpd_cluster_fit}(
  object,
  newdata = NULL,
  type = c("label", "psm"),
  burnin = NULL,
  thin = NULL,
  return_scores = FALSE,
  psm_max_n = 2000L,
  ...
)
}
\arguments{
\item{object}{A fitted cluster object.}

\item{newdata}{Optional new data containing the response and predictors required by the original
formula. New-data prediction is available only for \code{type = "label"}.}

\item{type}{Prediction target:
\itemize{
\item \code{"label"}: representative partition via
Dahl's least-squares rule
\item \code{"psm"}: posterior similarity matrix on the
training sample
}}

\item{burnin}{Number of initial posterior draws to discard.}

\item{thin}{Keep every \code{thin}-th posterior draw.}

\item{return_scores}{Logical; if \code{TRUE} and \code{type = "label"}, include the matrix of Dahl-cluster
assignment scores.}

\item{psm_max_n}{Maximum training sample size allowed for \code{type = "psm"}.}

\item{...}{Unused.}
}
\value{
A \code{dpmixgpd_cluster_labels} object when \code{type = "label"} or a
\code{dpmixgpd_cluster_psm} object when \code{type = "psm"}.
}
\description{
Convert posterior draws from a \code{dpmixgpd_cluster_fit} object into either a representative
clustering or a posterior similarity matrix (PSM). This is the main post-processing step for
the cluster workflow after \code{\link[=dpmix.cluster]{dpmix.cluster()}} or \code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}.
}
\details{
Let \eqn{z_i^{(s)}} denote the latent cluster label for observation \eqn{i} at posterior draw
\eqn{s}. The posterior similarity matrix is
\deqn{
\mathrm{PSM}_{ij} = \Pr(z_i = z_j \mid y) \approx \frac{1}{S} \sum_{s=1}^S I(z_i^{(s)} = z_j^{(s)}).
}
The returned label solution is the Dahl representative partition, obtained by choosing the draw
whose adjacency matrix is closest to the PSM in squared error.

For \code{newdata}, the function combines draw-specific component weights and component densities to
produce posterior assignment scores relative to the representative training clusters. Returned
\code{newdata} label objects also carry the training labels and response data needed for comparative
\code{plot(..., type = "summary")} displays. A PSM is not defined for \code{newdata}, so \code{type = "psm"}
is restricted to the training sample.

Computing the PSM is \eqn{O(n^2)} in the training sample size, so \code{psm_max_n} guards against
accidental large matrix allocations.
}
\seealso{
\code{\link[=dpmix.cluster]{dpmix.cluster()}}, \code{\link[=dpmgpd.cluster]{dpmgpd.cluster()}}, \code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}},
\code{\link[=plot.dpmixgpd_cluster_fit]{plot.dpmixgpd_cluster_fit()}}, \code{\link[=summary.dpmixgpd_cluster_labels]{summary.dpmixgpd_cluster_labels()}},
\code{\link[=summary.dpmixgpd_cluster_psm]{summary.dpmixgpd_cluster_psm()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_labels}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.get_dispatch_scalar}
\alias{.get_dispatch_scalar}
\title{Resolve kernel dispatch functions (scalar)
Dispatch returns raw scalar nimbleFunctions for codegen; do not wrap.}
\usage{
.get_dispatch_scalar(spec_or_fit, backend_override = NULL, gpd_override = NULL)
}
\arguments{
\item{spec_or_fit}{mixgpd_fit or spec list}
}
\value{
List with d/p/q/r/mean/mean_trunc functions and bulk_params.
}
\description{
Resolve kernel dispatch functions (scalar)
Dispatch returns raw scalar nimbleFunctions for codegen; do not wrap.
}
\details{
This helper resolves the density, distribution, quantile, random-generation,
and mean functions implied by a kernel, backend, and GPD setting. The result
is intentionally scalar and wrapper-free because it is used in code-generation
contexts where NIMBLE expects raw function objects rather than vectorized R
adapters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{summary.causalmixgpd_ps_fit}
\alias{summary.causalmixgpd_ps_fit}
\title{Summarize a propensity score fit}
\usage{
\method{summary}{causalmixgpd_ps_fit}(object, pars = NULL, probs = c(0.025, 0.5, 0.975), ...)
}
\arguments{
\item{object}{A \code{"causalmixgpd_ps_fit"} object.}

\item{pars}{Optional character vector of PS parameters to summarize. If
\code{NULL}, summarize all monitored parameters.}

\item{probs}{Numeric vector of posterior quantiles to report.}

\item{...}{Unused.}
}
\value{
An object of class \code{"summary.causalmixgpd_ps_fit"} with
elements \code{model} and \code{table}.
}
\description{
\code{summary.causalmixgpd_ps_fit()} returns posterior summaries for the
monitored PS-model parameters.
}
\details{
The summary is parameter based. For logit and probit models, it summarizes the
posterior draws of the coefficients that determine the latent linear
predictor, which is then mapped to \eqn{e(x)} by the chosen link function.
For the naive Bayes option, it summarizes the class-conditional parameters
used to factorize the treatment-assignment model.

This function does not compute fitted propensity scores for specific covariate
rows. It summarizes the posterior distribution of the PS model itself, which
is the nuisance model later used by causal prediction and treatment-effect
standardization.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.causalmixgpd_causal_predict_plots}
\alias{print.causalmixgpd_causal_predict_plots}
\title{Print method for causal prediction plots}
\usage{
\method{print}{causalmixgpd_causal_predict_plots}(x, ...)
}
\arguments{
\item{x}{Object of class \code{causalmixgpd_causal_predict_plots}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns the input object.
}
\description{
Print method for causal prediction plots
}
\details{
The causal prediction plotting methods can return either a single plot or a
named list of plots. This print method renders those stored plot objects in
sequence so both arm-level and contrast-level graphics appear in console or
notebook workflows without manual extraction.

It is a display helper only and does not modify the underlying prediction
summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data-datasets.R
\docType{data}
\name{nc_real200_k2}
\alias{nc_real200_k2}
\title{Real-Line Bulk-Only Data Set With Two Components}
\format{
A list with:
\describe{
\item{y}{Numeric outcome vector.}
\item{X}{NULL.}
\item{meta}{List with n, support, p, K_true, tail, exceed_frac, seed.}
\item{truth}{List with kernel, weights, params, threshold, tail_params.}
}
}
\usage{
nc_real200_k2
}
\description{
Real-line, bulk-only mixture dataset with K=2 components and no covariates.
Intended for non-causal bulk-only vignettes (normal/laplace/cauchy, GPD=FALSE).
}
\examples{
head(nc_real200_k2$y)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoroso.R
\name{amoroso_gpd}
\alias{amoroso_gpd}
\alias{dAmorosoGpd}
\alias{pAmorosoGpd}
\alias{rAmorosoGpd}
\alias{qAmorosoGpd}
\title{Amoroso with a GPD tail}
\usage{
dAmorosoGpd(
  x,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  log = 0
)

pAmorosoGpd(
  q,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rAmorosoGpd(n, loc, scale, shape1, shape2, threshold, tail_scale, tail_shape)

qAmorosoGpd(
  p,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{loc}{Numeric scalar location parameter of the Amoroso bulk.}

\item{scale}{Numeric scalar scale parameter of the Amoroso bulk.}

\item{shape1}{Numeric scalar first Amoroso shape parameter.}

\item{shape2}{Numeric scalar second Amoroso shape parameter.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric tolerance for numerical inversion in \code{qAmorosoGpd}.}

\item{maxiter}{Maximum iterations for numerical inversion in \code{qAmorosoGpd}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qAmorosoGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced family obtained by attaching a generalized Pareto tail above \code{threshold} to a single
Amoroso bulk.
}
\details{
This is the single-component Amoroso splice. The Amoroso law controls the distribution below the
threshold and the GPD controls exceedances above it, scaled so that the resulting CDF is
continuous at the threshold. The ordinary mean of the spliced law exists only when the tail
satisfies \eqn{\xi < 1}; otherwise users should rely on restricted means or quantile summaries.
}
\section{Functions}{
\itemize{
\item \code{dAmorosoGpd()}: Density Function of Amoroso Distribution with GPD Tail

\item \code{pAmorosoGpd()}: Cumulative Distribution Function of Amoroso Distribution with GPD Tail

\item \code{rAmorosoGpd()}: Random Generation for Amoroso Distribution with GPD Tail

\item \code{qAmorosoGpd()}: Quantile Function of Amoroso Distribution with GPD Tail

}}
\examples{
loc <- 0
scale <- 1.5
shape1 <- 2
shape2 <- 1.2
threshold <- 3
tail_scale <- 1.0
tail_shape <- 0.2

dAmorosoGpd(4.0, loc, scale, shape1, shape2,
           threshold, tail_scale, tail_shape, log = 0)
pAmorosoGpd(4.0, loc, scale, shape1, shape2,
           threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
qAmorosoGpd(0.50, loc, scale, shape1, shape2,
           threshold, tail_scale, tail_shape)
qAmorosoGpd(0.95, loc, scale, shape1, shape2,
           threshold, tail_scale, tail_shape)
replicate(10, rAmorosoGpd(1, loc, scale, shape1, shape2,
                         threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=amoroso_mix]{amoroso_mix()}}, \code{\link[=amoroso_mixgpd]{amoroso_mixgpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=amoroso_lowercase]{amoroso_lowercase()}}.

Other amoroso kernel families: 
\code{\link{amoroso_mix}},
\code{\link{amoroso_mixgpd}}
}
\concept{amoroso kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/0-base-distributions.R
\name{cauchy}
\alias{cauchy}
\alias{dCauchy}
\alias{pCauchy}
\alias{rCauchy}
\alias{qCauchy}
\title{Cauchy distribution}
\usage{
dCauchy(x, location, scale, log = 0)

pCauchy(q, location, scale, lower.tail = 1, log.p = 0)

rCauchy(n, location, scale)

qCauchy(p, location, scale, lower.tail = TRUE, log.p = FALSE)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{location}{Numeric scalar location parameter.}

\item{scale}{Numeric scalar scale parameter; must be positive.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}
}
\value{
\code{dCauchy()} returns a numeric scalar density, \code{pCauchy()} returns a numeric scalar CDF,
\code{rCauchy()} returns one random draw, and \code{qCauchy()} returns a numeric quantile.
}
\description{
Scalar Cauchy utilities implemented for NIMBLE compatibility. These functions support symmetric
heavy-tailed kernels on the real line and feed directly into the finite Cauchy mixture family.
}
\details{
The density is
\deqn{
f(x) = \frac{1}{\pi s \{1 + ((x - \ell)/s)^2\}},
}
where \code{location = ell} and \code{scale = s > 0}.

These uppercase NIMBLE-compatible functions are scalar (\code{x}/\code{q} and \code{n = 1}).
For vectorized R usage, use \code{\link[=base_lowercase]{base_lowercase()}}.

The Cauchy law is a stable heavy-tailed distribution with undefined mean and variance. That is
why the package allows the Cauchy kernel only as a bulk distribution and deliberately does not
pair it with GPD tails in the kernel registry. For predictive summaries,
ordinary means are not available under Cauchy kernels; medians, quantiles,
survival curves, and restricted means remain well defined.

The distribution function is
\deqn{
F(x) = \frac{1}{2} + \frac{1}{\pi}\arctan\left(\frac{x-\ell}{s}\right),
}
and the quantile is the corresponding inverse
\deqn{
Q(p) = \ell + s \tan\{\pi(p-1/2)\}.
}
}
\section{Functions}{
\itemize{
\item \code{dCauchy()}: Cauchy density function

\item \code{pCauchy()}: Cauchy distribution function

\item \code{rCauchy()}: Cauchy random generation

\item \code{qCauchy()}: Cauchy quantile function

}}
\examples{
location <- 0
scale <- 1.5

dCauchy(0.5, location, scale, log = 0)
pCauchy(0.5, location, scale, lower.tail = 1, log.p = 0)
qCauchy(0.50, location, scale)
qCauchy(0.95, location, scale)
replicate(10, rCauchy(1, location, scale))

}
\seealso{
\code{\link[=cauchy_mix]{cauchy_mix()}}, \code{\link[=base_lowercase]{base_lowercase()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other base bulk distributions: 
\code{\link{InvGauss}},
\code{\link{amoroso}}
}
\concept{base bulk distributions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.compute_cluster_probs}
\alias{.compute_cluster_probs}
\title{Compute cluster membership probabilities from PSM}
\usage{
.compute_cluster_probs(z_matrix, labels_representative, PSM)
}
\arguments{
\item{z_matrix}{Integer matrix (iterations x N) of cluster assignments.}

\item{labels_representative}{Integer vector of representative cluster labels.}

\item{PSM}{Posterior similarity matrix (N x N).}
}
\value{
N x K matrix of cluster membership probabilities.
}
\description{
For each observation, computes the probability of membership in each cluster
defined by the representative clustering, derived from the posterior
similarity matrix.
}
\details{
The representative labels define a reference partition with clusters
\eqn{C_1, \dots, C_K}. For each observation \eqn{i}, this helper averages the
posterior similarity scores \eqn{\mathrm{PSM}_{ij}} over members
\eqn{j \in C_k} to obtain a cluster-membership score for cluster \eqn{k}, and
then normalizes those scores to sum to one across clusters.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/laplace.R
\name{laplace_lowercase}
\alias{laplace_lowercase}
\alias{dlaplacemix}
\alias{plaplacemix}
\alias{qlaplacemix}
\alias{rlaplacemix}
\alias{dlaplacemixgpd}
\alias{plaplacemixgpd}
\alias{qlaplacemixgpd}
\alias{rlaplacemixgpd}
\alias{dlaplacegpd}
\alias{plaplacegpd}
\alias{qlaplacegpd}
\alias{rlaplacegpd}
\title{Lowercase vectorized Laplace distribution functions}
\usage{
dlaplacemix(x, w, location, scale, log = FALSE)

plaplacemix(q, w, location, scale, lower.tail = TRUE, log.p = FALSE)

qlaplacemix(
  p,
  w,
  location,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rlaplacemix(n, w, location, scale)

dlaplacemixgpd(
  x,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

plaplacemixgpd(
  q,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qlaplacemixgpd(
  p,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rlaplacemixgpd(n, w, location, scale, threshold, tail_scale, tail_shape)

dlaplacegpd(x, location, scale, threshold, tail_scale, tail_shape, log = FALSE)

plaplacegpd(
  q,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qlaplacegpd(
  p,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

rlaplacegpd(n, location, scale, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{location, scale}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar Laplace-kernel topics in this file.
}
\details{
These helpers vectorize the scalar Laplace and Laplace-plus-GPD routines for interactive R use.
They retain the same location-scale parameterization and the same splice definition as the
uppercase functions. Quantiles continue to use the scalar root-finding or piecewise logic rather
than a separate approximation.
}
\section{Functions}{
\itemize{
\item \code{dlaplacemix()}: Laplace mixture density (vectorized)

\item \code{plaplacemix()}: Laplace mixture distribution function (vectorized)

\item \code{qlaplacemix()}: Laplace mixture quantile function (vectorized)

\item \code{rlaplacemix()}: Laplace mixture random generation (vectorized)

\item \code{dlaplacemixgpd()}: Laplace mixture + GPD density (vectorized)

\item \code{plaplacemixgpd()}: Laplace mixture + GPD distribution function (vectorized)

\item \code{qlaplacemixgpd()}: Laplace mixture + GPD quantile function (vectorized)

\item \code{rlaplacemixgpd()}: Laplace mixture + GPD random generation (vectorized)

\item \code{dlaplacegpd()}: Laplace + GPD density (vectorized)

\item \code{plaplacegpd()}: Laplace + GPD distribution function (vectorized)

\item \code{qlaplacegpd()}: Laplace + GPD quantile function (vectorized)

\item \code{rlaplacegpd()}: Laplace + GPD random generation (vectorized)

}}
\examples{
w <- c(0.6, 0.3, 0.1)
loc <- c(0, 1, -2)
scl <- c(1, 0.9, 1.1)

# Laplace mixture
dlaplacemix(c(-1, 0, 1), w = w, location = loc, scale = scl)
rlaplacemix(5, w = w, location = loc, scale = scl)

}
\seealso{
\code{\link[=laplace_mix]{laplace_mix()}}, \code{\link[=laplace_MixGpd]{laplace_MixGpd()}}, \code{\link[=laplace_gpd]{laplace_gpd()}}, \code{\link[=bundle]{bundle()}},
\code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{gamma_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoroso.R
\name{amoroso_mix}
\alias{amoroso_mix}
\alias{dAmorosoMix}
\alias{pAmorosoMix}
\alias{rAmorosoMix}
\alias{qAmorosoMix}
\title{Amoroso mixture distribution}
\usage{
dAmorosoMix(x, w, loc, scale, shape1, shape2, log = 0)

pAmorosoMix(q, w, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)

rAmorosoMix(n, w, loc, scale, shape1, shape2)

qAmorosoMix(
  p,
  w,
  loc,
  scale,
  shape1,
  shape2,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions treat the weights
as non-negative and normalize them internally when needed.}

\item{loc}{Numeric vector of length \eqn{K} giving component locations.}

\item{scale}{Numeric vector of length \eqn{K} giving component scales. Negative values flip support
for the corresponding component.}

\item{shape1}{Numeric vector of length \eqn{K} giving the first Amoroso shape parameter for each component.}

\item{shape2}{Numeric vector of length \eqn{K} giving the second Amoroso shape parameter for each component.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Mixture density/CDF/RNG functions return numeric scalars. \code{qAmorosoMix()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Finite mixture of Amoroso components for flexible positive-support bulk modeling.
}
\details{
The mixture density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k f_{Amoroso}(x \mid a_k, \theta_k, \alpha_k, \beta_k),
}
with normalized weights \eqn{\tilde{w}_k}. These scalar functions are NIMBLE-compatible; for
vectorized R usage, use \code{\link[=amoroso_lowercase]{amoroso_lowercase()}}.

The Amoroso family is especially useful for positive-support data because it can reproduce a wide
range of skewed and heavy-right-tail shapes while remaining analytically tractable through its
gamma transformation. The mixture CDF is
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k F_{Amoroso}(x \mid a_k,\theta_k,\alpha_k,\beta_k),
}
and random generation proceeds by selecting a component and sampling from that component.

Closed-form mixture quantiles are not available, so \code{qAmorosoMix()} inverts the mixture CDF
numerically. The analytical mixture mean is the weighted average of the component means,
\eqn{a_k + \theta_k \Gamma(\alpha_k + 1/\beta_k) / \Gamma(\alpha_k)}, whenever those component
moments exist.
}
\section{Functions}{
\itemize{
\item \code{dAmorosoMix()}: Density Function of Amoroso Mixture Distribution

\item \code{pAmorosoMix()}: Cumulative Distribution Function of Amoroso Mixture Distribution

\item \code{rAmorosoMix()}: Random Generation for Amoroso Mixture Distribution

\item \code{qAmorosoMix()}: Quantile Function of Amoroso Mixture Distribution

}}
\examples{
w <- c(0.60, 0.25, 0.15)
loc <- c(0, 1, 2)
scale <- c(1.0, 1.2, 1.6)
shape1 <- c(2, 4, 6)
shape2 <- c(1.0, 1.2, 1.5)

dAmorosoMix(2.0, w, loc, scale, shape1, shape2, log = 0)
pAmorosoMix(2.0, w, loc, scale, shape1, shape2, lower.tail = 1, log.p = 0)
qAmorosoMix(0.50, w, loc, scale, shape1, shape2)
qAmorosoMix(0.95, w, loc, scale, shape1, shape2)
replicate(10, rAmorosoMix(1, w, loc, scale, shape1, shape2))
}
\seealso{
\code{\link[=amoroso_mixgpd]{amoroso_mixgpd()}}, \code{\link[=amoroso_gpd]{amoroso_gpd()}}, \code{\link[=amoroso_lowercase]{amoroso_lowercase()}},
\code{\link[=build_nimble_bundle]{build_nimble_bundle()}}, \code{\link[=kernel_support_table]{kernel_support_table()}}.

Other amoroso kernel families: 
\code{\link{amoroso_gpd}},
\code{\link{amoroso_mixgpd}}
}
\concept{amoroso kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/laplace.R
\name{laplace_MixGpd}
\alias{laplace_MixGpd}
\alias{dLaplaceMixGpd}
\alias{pLaplaceMixGpd}
\alias{rLaplaceMixGpd}
\alias{qLaplaceMixGpd}
\title{Laplace mixture with a GPD tail}
\usage{
dLaplaceMixGpd(
  x,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  log = 0
)

pLaplaceMixGpd(
  q,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rLaplaceMixGpd(n, w, location, scale, threshold, tail_scale, tail_shape)

qLaplaceMixGpd(
  p,
  w,
  location,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}.}

\item{location}{Numeric vector of length \eqn{K} giving component locations.}

\item{scale}{Numeric vector of length \eqn{K} giving component scales.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qLaplaceMixGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to a Laplace mixture
bulk.
}
\details{
This family keeps the Laplace mixture body below the threshold and replaces the upper tail with a
generalized Pareto exceedance model scaled by the residual survival mass at the threshold. The
tail density is therefore
\deqn{
f(x) = \{1-F_{mix}(u)\} g_{GPD}(x \mid u,\sigma_u,\xi), \qquad x \ge u.
}
Bulk quantiles are found numerically from the mixture CDF, and tail quantiles are computed by
rescaling the upper-tail probability and applying the GPD inverse.
}
\section{Functions}{
\itemize{
\item \code{dLaplaceMixGpd()}: Laplace mixture + GPD tail density

\item \code{pLaplaceMixGpd()}: Laplace mixture + GPD tail distribution function

\item \code{rLaplaceMixGpd()}: Laplace mixture + GPD tail random generation

\item \code{qLaplaceMixGpd()}: Laplace mixture + GPD tail quantile function

}}
\examples{
w <- c(0.50, 0.30, 0.20)
location <- c(-1, 0.5, 2.0)
scale <- c(1.0, 0.7, 1.4)
threshold <- 1
tail_scale <- 1.0
tail_shape <- 0.2

dLaplaceMixGpd(2.0, w = w, location = location, scale = scale,
              threshold = threshold, tail_scale = tail_scale,
              tail_shape = tail_shape, log = FALSE)
pLaplaceMixGpd(2.0, w = w, location = location, scale = scale,
              threshold = threshold, tail_scale = tail_scale,
              tail_shape = tail_shape, lower.tail = TRUE, log.p = FALSE)
qLaplaceMixGpd(0.50, w = w, location = location, scale = scale,
              threshold = threshold, tail_scale = tail_scale,
              tail_shape = tail_shape)
qLaplaceMixGpd(0.95, w = w, location = location, scale = scale,
              threshold = threshold, tail_scale = tail_scale,
              tail_shape = tail_shape)
replicate(10, rLaplaceMixGpd(1, w = w, location = location, scale = scale,
                            threshold = threshold,
                            tail_scale = tail_scale,
                            tail_shape = tail_shape))
}
\seealso{
\code{\link[=laplace_mix]{laplace_mix()}}, \code{\link[=laplace_gpd]{laplace_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=laplace_lowercase]{laplace_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other laplace kernel families: 
\code{\link{laplace_gpd}},
\code{\link{laplace_mix}}
}
\concept{laplace kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.extract_draws_matrix}
\alias{.extract_draws_matrix}
\title{Extract posterior draws as a numeric matrix (iter x parameters)}
\usage{
.extract_draws_matrix(object, drop_v = TRUE, epsilon = NULL)
}
\arguments{
\item{object}{A mixgpd_fit.}

\item{drop_v}{Logical; if TRUE, drop stick-breaking v parameters.}
}
\value{
Numeric matrix of draws.
}
\description{
Extract posterior draws as a numeric matrix (iter x parameters)
}
\details{
This helper stacks the posterior chains into one numeric matrix, optionally
removes stick-breaking \code{v} variables, and then applies the component
truncation rule controlled by \code{epsilon}. The result is the standardized draw
representation used by most internal summary and prediction helpers.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{mcmc}
\alias{mcmc}
\title{Run posterior sampling from a prepared bundle}
\usage{
mcmc(b, ...)
}
\arguments{
\item{b}{A non-causal or causal bundle.}

\item{...}{Optional MCMC overrides (\code{niter}, \code{nburnin}, \code{thin},
\code{nchains}, \code{seed}, \code{waic}) and runner controls
(\code{show_progress}, \code{quiet}).}
}
\value{
A fitted object of class \code{"mixgpd_fit"} or
\code{"causalmixgpd_causal_fit"}.
}
\description{
\code{mcmc()} is the generic workflow runner. It dispatches to
\code{\link{run_mcmc_bundle_manual}} for one-arm bundles and to
\code{\link{run_mcmc_causal}} for causal bundles.
}
\details{
This wrapper is useful when you want a two-stage workflow:
build first, inspect or modify the bundle, then sample. Named MCMC arguments
supplied through \code{...} override the settings stored in the bundle before
execution.

The returned fit represents posterior draws from the finite SB/CRP
approximation encoded in the bundle. Downstream summaries therefore target
posterior predictive quantities such as \eqn{f(y \mid x)},
\eqn{F(y \mid x)}, and derived treatment-effect functionals.
}
\seealso{
\code{\link{bundle}}, \code{\link{run_mcmc_bundle_manual}},
\code{\link{run_mcmc_causal}}, \code{\link{predict.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/amoroso.R
\name{amoroso_mixgpd}
\alias{amoroso_mixgpd}
\alias{dAmorosoMixGpd}
\alias{pAmorosoMixGpd}
\alias{rAmorosoMixGpd}
\alias{qAmorosoMixGpd}
\title{Amoroso mixture with a GPD tail}
\usage{
dAmorosoMixGpd(
  x,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  log = 0
)

pAmorosoMixGpd(
  q,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = 1,
  log.p = 0
)

rAmorosoMixGpd(
  n,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape
)

qAmorosoMixGpd(
  p,
  w,
  loc,
  scale,
  shape1,
  shape2,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions treat the weights
as non-negative and normalize them internally when needed.}

\item{loc}{Numeric vector of length \eqn{K} giving component locations.}

\item{scale}{Numeric vector of length \eqn{K} giving component scales.}

\item{shape1}{Numeric vector of length \eqn{K} giving the first Amoroso shape parameter for each component.}

\item{shape2}{Numeric vector of length \eqn{K} giving the second Amoroso shape parameter for each component.}

\item{threshold}{Numeric scalar threshold at which the GPD tail is attached.}

\item{tail_scale}{Numeric scalar GPD scale parameter; must be positive.}

\item{tail_shape}{Numeric scalar GPD shape parameter.}

\item{log}{Logical; if \code{TRUE}, return the log-density (integer flag \code{0/1} in NIMBLE).}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. For portability inside NIMBLE,
the RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot} in quantile inversion.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Spliced density/CDF/RNG functions return numeric scalars. \code{qAmorosoMixGpd()} returns a
numeric vector with the same length as \code{p}.
}
\description{
Spliced bulk-tail family formed by attaching a generalized Pareto tail to an Amoroso mixture
bulk. Let \eqn{F_{mix}} denote the Amoroso mixture CDF. The spliced CDF is
\eqn{F(x)=F_{mix}(x)} for \eqn{x<threshold} and
\eqn{F(x)=F_{mix}(threshold) + \left\{1-F_{mix}(threshold)\right\}G(x)} for \eqn{x\ge threshold}, where \eqn{G}
is the GPD CDF for exceedances above \code{threshold}.
}
\details{
The density, CDF, and RNG are implemented as \code{nimbleFunction}s for use in NIMBLE models.
The quantile function is an R function that uses numerical inversion in the bulk region and
the closed-form GPD quantile in the tail region.

The Amoroso mixture describes the bulk up to the threshold and the generalized Pareto describes
exceedances above it. If \eqn{F_{mix}(u)=p_u}, then the splice uses
\deqn{
f(x) =
\left\{
\begin{array}{ll}
f_{mix}(x), & x < u, \\
(1-p_u) g_{GPD}(x \mid u,\sigma_u,\xi), & x \ge u.
\end{array}
\right.
}
Bulk quantiles are computed numerically from the Amoroso mixture CDF and tail quantiles are
obtained from the GPD inverse after rescaling the tail probability.
}
\section{Functions}{
\itemize{
\item \code{dAmorosoMixGpd()}: Density Function of Amoroso Mixture Distribution with GPD Tail

\item \code{pAmorosoMixGpd()}: Cumulative Distribution Function of Amoroso Mixture Distribution with GPD Tail

\item \code{rAmorosoMixGpd()}: Random Generation for Amoroso Mixture Distribution with GPD Tail

\item \code{qAmorosoMixGpd()}: Quantile Function of Amoroso Mixture Distribution with GPD Tail

}}
\examples{
w <- c(0.60, 0.25, 0.15)
loc <- c(0, 1, 2)
scale <- c(1.0, 1.2, 1.6)
shape1 <- c(2, 4, 6)
shape2 <- c(1.0, 1.2, 1.5)
threshold <- 3
tail_scale <- 1.0
tail_shape <- 0.2

dAmorosoMixGpd(4.0, w, loc, scale, shape1, shape2,
              threshold, tail_scale, tail_shape, log = 0)
pAmorosoMixGpd(4.0, w, loc, scale, shape1, shape2,
              threshold, tail_scale, tail_shape, lower.tail = 1, log.p = 0)
qAmorosoMixGpd(0.50, w, loc, scale, shape1, shape2,
              threshold, tail_scale, tail_shape)
qAmorosoMixGpd(0.95, w, loc, scale, shape1, shape2,
              threshold, tail_scale, tail_shape)
replicate(10, rAmorosoMixGpd(1, w, loc, scale, shape1, shape2,
                            threshold, tail_scale, tail_shape))
}
\seealso{
\code{\link[=amoroso_mix]{amoroso_mix()}}, \code{\link[=amoroso_gpd]{amoroso_gpd()}}, \code{\link[=gpd]{gpd()}}, \code{\link[=amoroso_lowercase]{amoroso_lowercase()}}, \code{\link[=dpmgpd]{dpmgpd()}}.

Other amoroso kernel families: 
\code{\link{amoroso_gpd}},
\code{\link{amoroso_mix}}
}
\concept{amoroso kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wrappers.R
\name{dpmgpd}
\alias{dpmgpd}
\title{Fit a one-arm Dirichlet process mixture with a spliced GPD tail}
\usage{
dpmgpd(
  y = NULL,
  X = NULL,
  treat = NULL,
  data = NULL,
  mcmc = list(),
  formula = NULL,
  ...
)
}
\arguments{
\item{y}{Either a response vector or a bundle object.}

\item{X}{Optional design matrix/data.frame.}

\item{treat}{Optional binary treatment indicator. If supplied, this wrapper
errors; use \code{dpmgpd.causal()} for causal models.}

\item{data}{Optional data.frame used with \code{formula}.}

\item{mcmc}{Named list of run arguments passed to \code{mcmc()} (including
optional performance controls such as \code{parallel_chains},
\code{workers}, \code{timing}, and \code{z_update_every}).}

\item{formula}{Optional formula.}

\item{...}{Additional build arguments passed to \code{\link{build_nimble_bundle}}.}
}
\value{
A fitted object of class \code{"mixgpd_fit"}.
}
\description{
\code{dpmgpd()} is the one-step convenience wrapper for the spliced
bulk-tail model. It combines \code{\link{bundle}} and \code{\link{mcmc}} for
one-arm data.
}
\details{
This wrapper targets the posterior predictive distribution obtained by
combining a flexible bulk DPM with a generalized Pareto exceedance model
above the threshold \eqn{u(x)}. In the tail region the predictive density is
proportional to
\deqn{\{1 - p_u(x)\} f_{\mathrm{GPD}}(y \mid x), \qquad y > u(x),}
where \eqn{p_u(x)} is the posterior bulk mass below the threshold.

Use this wrapper when upper-tail behavior matters for inference, prediction,
or extrapolation of extreme quantiles and survival probabilities.
}
\seealso{
\code{\link{build_nimble_bundle}},
\code{\link{bundle}}, \code{\link{dpmix}},
\code{\link{predict.mixgpd_fit}}, \code{\link{summary.mixgpd_fit}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/internal.R
\name{.truncate_components_one_draw}
\alias{.truncate_components_one_draw}
\title{Truncate and reorder mixture components by cumulative weight mass}
\usage{
.truncate_components_one_draw(w, params, epsilon = 0.01)
}
\arguments{
\item{w}{Numeric vector of component weights (length K).}

\item{params}{Named list of numeric vectors, each length K (component-specific params).}

\item{epsilon}{Numeric in [0,1). Keep the smallest k s.t. cumweight >= 1-epsilon.}
}
\value{
A list with reordered+truncated weights/params and bookkeeping.
}
\description{
Truncate and reorder mixture components by cumulative weight mass
}
\details{
This helper operates on one posterior draw at a time. It first orders mixture
components by decreasing weight, then keeps the smallest effective subset of
components implied by the package truncation rule, and finally renormalizes
the retained weights so they sum to one.

The same permutation is applied to every component-specific parameter vector in
\code{params}, which keeps the retained parameter blocks aligned with the retained
weights.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_fitted_plots}
\alias{print.mixgpd_fitted_plots}
\title{Print method for fitted value plots}
\usage{
\method{print}{mixgpd_fitted_plots}(x, ...)
}
\arguments{
\item{x}{Object of class \code{mixgpd_fitted_plots}.}

\item{...}{Additional arguments (ignored).}
}
\value{
Invisibly returns the input object.
}
\description{
Print method for fitted value plots
}
\details{
The fitted-value diagnostic object stores two plot panels. This print method
renders them in sequence so both the observed-versus-fitted comparison and the
residual-versus-fitted comparison are shown together.

It is a display helper only and does not recompute fitted values or
residuals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal_mix}
\alias{normal_mix}
\alias{dNormMix}
\alias{pNormMix}
\alias{rNormMix}
\alias{qNormMix}
\title{Normal mixture distribution}
\usage{
dNormMix(x, w, mean, sd, log = 0)

pNormMix(q, w, mean, sd, lower.tail = 1, log.p = 0)

rNormMix(n, w, mean, sd)

qNormMix(
  p,
  w,
  mean,
  sd,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)
}
\arguments{
\item{x}{Numeric scalar giving the point at which the density is evaluated.}

\item{w}{Numeric vector of mixture weights of length \eqn{K}. The functions normalize \code{w}
internally when needed.}

\item{mean, sd}{Numeric vectors of length \eqn{K} giving component means and standard deviations.}

\item{log}{Integer flag \code{0/1}; if \code{1}, return the log-density.}

\item{q}{Numeric scalar giving the point at which the distribution function is evaluated.}

\item{lower.tail}{Integer flag \code{0/1}; if \code{1} (default), probabilities are \eqn{P(X \le q)}.}

\item{log.p}{Integer flag \code{0/1}; if \code{1}, probabilities are returned on the log scale.}

\item{n}{Integer giving the number of draws. The RNG implementation supports \code{n = 1}.}

\item{p}{Numeric scalar probability in \eqn{(0,1)} for the quantile function.}

\item{tol}{Numeric scalar tolerance passed to \code{stats::uniroot}.}

\item{maxiter}{Integer maximum number of iterations for \code{stats::uniroot}.}
}
\value{
Density/CDF/RNG functions return numeric scalars. \code{qNormMix()} returns a numeric vector
with the same length as \code{p}.
}
\description{
Finite mixture of normal components for real-valued bulk modeling. This topic provides the
scalar mixture density, CDF, RNG, and quantile functions used by the bulk-only normal kernel in
the package registry.
}
\details{
With weights \eqn{w_k}, means \eqn{\mu_k}, and standard deviations \eqn{\sigma_k}, the mixture
density is
\deqn{
f(x) = \sum_{k = 1}^K \tilde{w}_k \phi(x \mid \mu_k, \sigma_k^2),
}
where \eqn{\tilde{w}_k = w_k / \sum_j w_j}. These uppercase functions are scalar
NIMBLE-compatible building blocks. For vectorized R usage, use \code{\link[=normal_lowercase]{normal_lowercase()}}.

If \eqn{F_k} denotes the \eqn{k}-th component CDF, then the mixture distribution function is
\deqn{
F(x) = \sum_{k=1}^K \tilde{w}_k F_k(x)
= \sum_{k=1}^K \tilde{w}_k \Phi\left(\frac{x-\mu_k}{\sigma_k}\right).
}
Random generation first draws a component index with probability \eqn{\tilde{w}_k} and then
generates from the corresponding normal law. The quantile function has no closed form for a
general finite mixture, so \code{qNormMix()} solves \eqn{F(x)=p} numerically by bracketing the
root and applying \code{stats::uniroot()}.

The mixture mean is
\deqn{
E(X) = \sum_{k=1}^K \tilde{w}_k \mu_k,
}
which is the analytical mean used by the package when a normal-mixture draw contributes to a
posterior predictive mean calculation.
}
\section{Functions}{
\itemize{
\item \code{dNormMix()}: Normal mixture density

\item \code{pNormMix()}: Normal mixture distribution function

\item \code{rNormMix()}: Normal mixture random generation

\item \code{qNormMix()}: Normal mixture quantile function

}}
\examples{
w <- c(0.60, 0.25, 0.15)
mean <- c(-1, 0.5, 2.0)
sd <- c(1.0, 0.7, 1.3)

dNormMix(0.5, w = w, mean = mean, sd = sd, log = FALSE)
pNormMix(0.5, w = w, mean = mean, sd = sd,
        lower.tail = TRUE, log.p = FALSE)
qNormMix(0.50, w = w, mean = mean, sd = sd)
qNormMix(0.95, w = w, mean = mean, sd = sd)
replicate(10, rNormMix(1, w = w, mean = mean, sd = sd))
}
\seealso{
\code{\link[=normal_mixgpd]{normal_mixgpd()}}, \code{\link[=normal_gpd]{normal_gpd()}}, \code{\link[=normal_lowercase]{normal_lowercase()}}, \code{\link[=build_nimble_bundle]{build_nimble_bundle()}},
\code{\link[=kernel_support_table]{kernel_support_table()}}.

Other normal kernel families: 
\code{\link{normal_gpd}},
\code{\link{normal_mixgpd}}
}
\concept{normal kernel families}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods.R
\name{print.mixgpd_summary}
\alias{print.mixgpd_summary}
\title{Print a MixGPD summary object}
\usage{
\method{print}{mixgpd_summary}(x, digits = 3, max_rows = 60, show_ess = FALSE, ...)
}
\arguments{
\item{x}{A \code{"mixgpd_summary"} object.}

\item{digits}{Number of digits to print.}

\item{max_rows}{Maximum rows to print.}

\item{show_ess}{Logical; if \code{TRUE}, include the \code{ess} column when present.}

\item{...}{Unused.}
}
\value{
\code{x} invisibly.
}
\description{
Print a MixGPD summary object
}
\details{
This method formats the output of \code{summary.mixgpd_fit()}. It prints the model
metadata, any stored WAIC value, the effective truncation information induced
by \code{epsilon}, and the parameter-level posterior summary table.

The printed rows correspond to monitored posterior parameters. They are not
predictions of densities, quantiles, or means, which should instead be
obtained from \code{predict.mixgpd_fit()}.
}
\examples{
\donttest{
y <- abs(stats::rnorm(25)) + 0.1
bundle <- build_nimble_bundle(y = y, backend = "sb", kernel = "normal",
                             GPD = TRUE, components = 3,
                             mcmc = list(niter = 100, nburnin = 50, thin = 1, nchains = 1))
fit <- run_mcmc_bundle_manual(bundle)
summary(fit)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build-run.R
\name{build_inits_from_spec}
\alias{build_inits_from_spec}
\title{Build initial values from a compiled model spec}
\usage{
build_inits_from_spec(spec, seed = NULL, y = NULL, X = NULL)
}
\arguments{
\item{spec}{A compiled model specification produced by \code{compile_model_spec()}.}

\item{seed}{Optional seed (single integer or vector). If provided, the first element is used.}

\item{y}{Optional numeric vector of observed outcomes used for heuristic initializations.}

\item{X}{Optional numeric matrix of covariates used for link-mode parameter initializations.}
}
\value{
Named list of initial values.
}
\description{
Produces a list of initial values suitable for passing to \code{nimbleModel}.
The initial values are derived from \code{spec$plan} and are intended to be
stable and support-respecting (e.g., positive parameters start positive).
}
\details{
Notes:
\itemize{
\item Uses only \code{components} as the model size parameter.
\item SB: initializes stick breaks \code{v}; weights \code{w} are deterministic.
\item CRP: initializes memberships \code{z} in \code{1:components}.
\item Link-mode parameters initialize regression coefficients \code{beta_<param>}
with shape \code{components x P}.
\item Default GPD threshold under X is stochastic lognormal:
initializes \code{threshold[1:N]} and scalar \code{sdlog_u} (non-link thresholds are scalar).
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gamma.R
\name{gamma_lowercase}
\alias{gamma_lowercase}
\alias{dgammamix}
\alias{pgammamix}
\alias{qgammamix}
\alias{rgammamix}
\alias{dgammamixgpd}
\alias{pgammamixgpd}
\alias{qgammamixgpd}
\alias{rgammamixgpd}
\alias{dgammagpd}
\alias{pgammagpd}
\alias{qgammagpd}
\alias{rgammagpd}
\title{Lowercase vectorized gamma distribution functions}
\usage{
dgammamix(x, w, shape, scale, log = FALSE)

pgammamix(q, w, shape, scale, lower.tail = TRUE, log.p = FALSE)

qgammamix(
  p,
  w,
  shape,
  scale,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rgammamix(n, w, shape, scale)

dgammamixgpd(
  x,
  w,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  log = FALSE
)

pgammamixgpd(
  q,
  w,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qgammamixgpd(
  p,
  w,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE,
  tol = 1e-10,
  maxiter = 200
)

rgammamixgpd(n, w, shape, scale, threshold, tail_scale, tail_shape)

dgammagpd(x, shape, scale, threshold, tail_scale, tail_shape, log = FALSE)

pgammagpd(
  q,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

qgammagpd(
  p,
  shape,
  scale,
  threshold,
  tail_scale,
  tail_shape,
  lower.tail = TRUE,
  log.p = FALSE
)

rgammagpd(n, shape, scale, threshold, tail_scale, tail_shape)
}
\arguments{
\item{x}{Numeric vector of quantiles.}

\item{w}{Numeric vector of mixture weights.}

\item{shape, scale}{Numeric vectors (mix) or scalars (base+gpd) of component parameters.}

\item{log}{Logical; if \code{TRUE}, return log-density.}

\item{q}{Numeric vector of quantiles.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \eqn{P(X \le x)}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are on log scale.}

\item{p}{Numeric vector of probabilities.}

\item{tol, maxiter}{Tolerance and max iterations for numerical inversion.}

\item{n}{Integer number of observations to generate.}

\item{threshold, tail_scale, tail_shape}{GPD tail parameters (scalars).}
}
\value{
Numeric vector of densities, probabilities, quantiles, or random variates.
}
\description{
Vectorized R wrappers for the scalar gamma-kernel topics in this file.
}
\details{
These wrappers are vectorized interfaces to the scalar gamma and gamma-plus-GPD routines. They
preserve the package's shape-scale parameterization and the same splice definition used in the
fitted-model prediction code. Quantile wrappers delegate to the scalar inversion code rather than
implementing separate approximations.
}
\section{Functions}{
\itemize{
\item \code{dgammamix()}: Gamma mixture density (vectorized)

\item \code{pgammamix()}: Gamma mixture distribution function (vectorized)

\item \code{qgammamix()}: Gamma mixture quantile function (vectorized)

\item \code{rgammamix()}: Gamma mixture random generation (vectorized)

\item \code{dgammamixgpd()}: Gamma mixture + GPD density (vectorized)

\item \code{pgammamixgpd()}: Gamma mixture + GPD distribution function (vectorized)

\item \code{qgammamixgpd()}: Gamma mixture + GPD quantile function (vectorized)

\item \code{rgammamixgpd()}: Gamma mixture + GPD random generation (vectorized)

\item \code{dgammagpd()}: Gamma + GPD density (vectorized)

\item \code{pgammagpd()}: Gamma + GPD distribution function (vectorized)

\item \code{qgammagpd()}: Gamma + GPD quantile function (vectorized)

\item \code{rgammagpd()}: Gamma + GPD random generation (vectorized)

}}
\examples{
w <- c(0.55, 0.3, 0.15)
shp <- c(2, 4, 6)
scl <- c(1, 2.5, 5)

# Gamma mixture
dgammamix(c(1, 2, 3), w = w, shape = shp, scale = scl)
rgammamix(5, w = w, shape = shp, scale = scl)

# Gamma mixture + GPD
dgammamixgpd(c(2, 3, 4), w = w, shape = shp, scale = scl,
             threshold = 3, tail_scale = 0.9, tail_shape = 0.2)

# Gamma + GPD (single component)
dgammagpd(c(2, 3, 4), shape = 4, scale = 2.5, threshold = 3,
          tail_scale = 0.9, tail_shape = 0.2)

}
\seealso{
\code{\link[=gamma_mix]{gamma_mix()}}, \code{\link[=gamma_mixgpd]{gamma_mixgpd()}}, \code{\link[=gamma_gpd]{gamma_gpd()}}, \code{\link[=bundle]{bundle()}}, \code{\link[=get_kernel_registry]{get_kernel_registry()}}.

Other vectorized kernel helpers: 
\code{\link{amoroso_lowercase}},
\code{\link{base_lowercase}},
\code{\link{cauchy_mix_lowercase}},
\code{\link{invgauss_lowercase}},
\code{\link{laplace_lowercase}},
\code{\link{lognormal_lowercase}},
\code{\link{normal_lowercase}}
}
\concept{vectorized kernel helpers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cluster.R
\name{summary.dpmixgpd_cluster_labels}
\alias{summary.dpmixgpd_cluster_labels}
\title{Summarize cluster labels}
\usage{
\method{summary}{dpmixgpd_cluster_labels}(object, top_n = 5L, order_by = c("size", "label"), vars = NULL, ...)
}
\arguments{
\item{object}{Cluster labels object.}

\item{top_n}{Number of populated clusters to profile when attached data are available.}

\item{order_by}{Ordering rule for descriptive cluster profiles:
\itemize{
\item \code{"size"}: decreasing cluster size
\item \code{"label"}: ascending cluster label
}}

\item{vars}{Optional character vector of numeric columns to summarize within each cluster.}

\item{...}{Unused.}
}
\value{
Summary list containing cluster sizes, optional cluster-level descriptive summaries, and,
when available, assignment-certainty summaries.
}
\description{
Summarize a representative clustering for training data or new observations.
}
\details{
If score or probability matrices are attached, certainty is summarized by the rowwise maxima
\eqn{\max_k p_{ik}}, which quantify how strongly each observation is assigned to its selected
cluster. When the labels object also carries attached training or prediction data, the summary
includes descriptive mean/sd profiles for the first populated clusters.
}
\seealso{
\code{\link[=predict.dpmixgpd_cluster_fit]{predict.dpmixgpd_cluster_fit()}}, \code{\link[=plot.dpmixgpd_cluster_labels]{plot.dpmixgpd_cluster_labels()}},
\code{\link[=summary.dpmixgpd_cluster_fit]{summary.dpmixgpd_cluster_fit()}}.

Other cluster workflow: 
\code{\link{cluster_profiles}()},
\code{\link{dpmgpd.cluster}()},
\code{\link{dpmix.cluster}()},
\code{\link{plot.dpmixgpd_cluster_bundle}()},
\code{\link{predict.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_bundle}()},
\code{\link{print.dpmixgpd_cluster_fit}()},
\code{\link{print.dpmixgpd_cluster_labels}()},
\code{\link{print.dpmixgpd_cluster_psm}()},
\code{\link{summary.dpmixgpd_cluster_bundle}()},
\code{\link{summary.dpmixgpd_cluster_fit}()},
\code{\link{summary.dpmixgpd_cluster_psm}()}
}
\concept{cluster workflow}
